;;; --------------------------------------------------------------------------------------------------
;;;   				PROPAGATION ALGORITHM  
;;; 
;;; Esta archivo contiene los algoritmos de propagacion de la red Temporal segun (DOUBLE BELLMAN-FORD)
;;;
;;;
;;;			 LAST MODIFICATION: August 05, 2002
;;;
;;; ----------------------------------------------------------------------------------------------------



(in-package "COMMON-LISP-USER")

;;; ----------------------------------------------------------------------------------
;;;
;;; Get the negative cycle using the lower bound pointers, y estos ciclos negativos 
;;; pueden contener el origen
;;;
;;; ----------------------------------------------------------------------------------

(defun get-ncycle-origin-lb (time-map dc tpv tpu new-lb-v ub-v)

  (let* ((neg-cycle-list-lb NIL)
         (neg-cycle-list-ub NIL)
         (dcs (car (time_map-dc_sources time-map)))
         (n-tps (time_map-ntp time-map))
         (tp-from-s (edge-from dcs))
         (count-lb 0)
         (count-ub 0)
         ;;; (delta-lb 0)
         (delta-ub 0)
         (tpc tpu))

    ;;; LOOPS FOR LB CHAIN
    (loop
      ;;;(print "o-lb-lb")
      (if (or (>= count-lb n-tps) 
              (not (time_point-pl tpc))) (return NIL))
      (push (time_point-st_edge_lb tpc) neg-cycle-list-lb)
      (incf count-lb)
      (setf tpc (time_point-pl tpc)))

    (when (>= count-lb n-tps)
      (setf tpc tp-from-s)
      (setf neg-cycle-list-lb NIL)
      (loop
        ;;; (print "o-lb-lb-s")
        (push (time_point-st_edge_lb tpc) neg-cycle-list-lb)
        (if (equal (time_point-pl tpc) tp-from-s)
            (return NIL))
        (setf tpc (time_point-pl tpc))))

    (setf tpc tpv)

    ;;; LOOPS FOR UB CHAIN
    (loop
      ;;;(print "o-lb-ub")
      (if (or (>= count-ub n-tps)
              (not (time_point-pu tpc))) (return NIL))
      (push (time_point-st_edge_ub tpc) neg-cycle-list-ub)
      (incf count-ub)
      (setf tpc (time_point-pu tpc)))

    (when (>= count-ub n-tps)
      (setf tpc tp-from-s)
      (setf neg-cycle-list-ub NIL)
      (loop
        ;;;(print "o-lb-ub-s")
        (push (time_point-st_edge_ub tpc) neg-cycle-list-ub)
        (if (equal (time_point-pu tpc) tp-from-s)
            (return NIL))
        (setf tpc (time_point-pu tpc))))


    (if (>= count-lb n-tps)
        (progn
          (setf (time_map-ncycle-list-lb time-map) neg-cycle-list-lb)
          (setf (time_map-delta time-map) delta-ub))
      (if (>= count-ub n-tps)
          (progn
            (setf (time_map-ncycle-list-ub time-map) neg-cycle-list-ub)
            (setf (time_map-delta time-map) delta-ub))
        (progn
          (setf (time_map-ncycle-list-lb time-map) 
                (append (list dc) neg-cycle-list-lb))
          (setf (time_point-st_edge_lb tpv) dc)
          (setf (time_point-pl tpv) tpu)
          (setf (time_map-ncycle-list-ub time-map) neg-cycle-list-ub))))

    (setf (time_map-delta time-map) (- (+ new-lb-v ub-v)))

    (if (< (time_map-delta time-map) 0)
        (format T "~%TP-V=~A TP-U=~A NEW-LB-V=~D  UB-V=~D" 
                  tpu tpv  new-lb-v ub-v))

)) ;;; END *GET-NCYCLE-ORIGIN-LB*

;;; ----------------------------------------------------------------------------------
;;;
;;; Get the negative cycle using the upper bound pointers, y estos ciclos negativos 
;;; pueden contener el origen
;;;
;;; ----------------------------------------------------------------------------------


(defun get-ncycle-origin-ub (time-map dc tpv tpu new-ub-v lb-v)

  (let* ((neg-cycle-list-lb NIL)
         (neg-cycle-list-ub NIL)
         (dcs (car (time_map-dc_sources time-map)))
         (n-tps (time_map-ntp time-map))
         (tp-from-s (edge-from dcs))
         (count-lb 0)
         (count-ub 0)
         ;;; (delta-lb 0)
         ;;; (delta-ub 0)
         (tpc tpv))

    ;;; LOOPS FOR LB CHAIN
    (loop
      ;;;(print "o-ub-lb")
      (if (or (>= count-lb n-tps) 
              (not (time_point-pl tpc))) (return NIL))
      (push (time_point-st_edge_lb tpc) neg-cycle-list-lb)
      (incf count-lb)
      (setf tpc (time_point-pl tpc)))

    (when (>= count-lb n-tps)
      (setf tpc tp-from-s)
      (setf neg-cycle-list-lb NIL)
      (loop
        ;;;(print "o-ub-lb-s")
        (push (time_point-st_edge_lb tpc) neg-cycle-list-lb)
        (if (equal (time_point-pl tpc) tp-from-s)
            (return NIL))
        (setf tpc (time_point-pl tpc))))

    (setf tpc tpu)

    ;;; LOOPS FOR UB CHAIN
    (loop
      ;;;(print "o-ub-ub")
      (if (or (>= count-ub n-tps)
              (not (time_point-pu tpc))) (return NIL))
      (push (time_point-st_edge_ub tpc) neg-cycle-list-ub)
      (incf count-ub)
      (setf tpc (time_point-pu tpc)))

    (when (>= count-ub n-tps)
      (setf tpc tp-from-s)
      (setf neg-cycle-list-ub NIL)
      (loop
        ;;;(print "o-ub-ub-s")
        (push (time_point-st_edge_ub tpc) neg-cycle-list-ub)
        (if (equal (time_point-pu tpc) tp-from-s)
            (return NIL))
        (setf tpc (time_point-pu tpc))))


    (if (>= count-lb n-tps)
        (setf (time_map-ncycle-list-lb time-map) neg-cycle-list-lb)
      (if (>= count-ub n-tps)
          (setf (time_map-ncycle-list-ub time-map) neg-cycle-list-ub)
        (progn
          (setf (time_map-ncycle-list-ub time-map) 
                (append (list dc) neg-cycle-list-ub))
          (setf (time_point-st_edge_ub tpv) dc)
          (setf (time_point-pu tpv) tpu)
          (setf (time_map-ncycle-list-lb time-map) neg-cycle-list-lb))))

    (setf (time_map-delta time-map) (- (+ new-ub-v lb-v)))

    (if (< (time_map-delta time-map) 0)
        (format T "~%TP-V=~A TP-U=~A NEW-LB-V=~D  UB-V=~D" 
                  tpu tpv  new-ub-v lb-v))

)) ;;; END *GET-NCYCLE-ORIGIN-UB*

;;; ----------------------------------------------------------------------------------
;;;
;;; looks for lower bound negative cycles that do not contain the orign
;;;
;;; ----------------------------------------------------------------------------------

(defun get-ncycle-norigin-lb (time-map dc tpu tpv new-lb-v old-lb-v)

  (let ((neg-cycle-list (list dc))
        (tpc tpu))

    ;;; LOOP FOR LB CHAIN
    (loop
      ;;;(print "no-lb")
      (push (time_point-st_edge_lb tpc) neg-cycle-list)
      (if (equal (time_point-pl tpc) tpv)
          (return NIL))
      (setf tpc (time_point-pl tpc)))

    (setf (time_map-ncycle-list-lb time-map) neg-cycle-list)
    (setf (time_map-delta time-map) (- old-lb-v new-lb-v)))

) ;;; END *GET-NCYCLE-NORIGIN-LB*

;;; ----------------------------------------------------------------------------------
;;;
;;; looks for upper negative cycles that do not contain the orign
;;;
;;; ----------------------------------------------------------------------------------


(defun get-ncycle-norigin-ub (time-map dc tpu tpv new-ub-v old-ub-v)

  (let ((neg-cycle-list (list dc))
        (tpc tpu))

    ;;; LOOP FOR UB CHAIN
    (loop
      ;;;(print "no-ub")
      (push (time_point-st_edge_ub tpc) neg-cycle-list)
      (if (equal (time_point-pu tpc) tpv)
          (return NIL))
      (setf tpc (time_point-pu tpc)))

    (setf (time_map-ncycle-list-ub time-map) neg-cycle-list)
    (setf (time_map-delta time-map) (- old-ub-v new-ub-v)))

) ;;; END *GET-NCYCLE-NORIGIN-UB*

;;; ---------------------------------------------------------------------------


(defun init-list-q (list-sources)
  
  (let ((list-out NIL))
    
    (dolist (dcc list-sources list-out)
      (let ((tp-to (edge-to dcc))
	    (tp-from (edge-from dcc)))
	
	  (when (not (time_point-in-queue tp-to))
	    (setf (time_point-lb-prop tp-to) T)
            (setf (time_point-ub-prop tp-to) T)
	    (push tp-to list-out)
	    (setf (time_point-in-queue tp-to) T))
	
          (when (not (time_point-in-queue tp-from))
	    (setf (time_point-lb-prop tp-from) T)
            (setf (time_point-ub-prop tp-from) T)
	    (push tp-from list-out)
	    (setf (time_point-in-queue tp-from) T)))

))) ;;; END *INIT-LIST-QUEUE*

;;; ---------------------------------------------------------------------------


(defun delete-queue-marks (list-q)
  
  (dolist (tpc list-q)
    (setf (time_point-in-queue tpc) NIL)
    (setf (time_point-lb-prop tpc) NIL)
    (setf (time_point-ub-prop tpc) NIL)))

;;; ---------------------------------------------------------------------------


(defun delete-source-marks (list-sources)

   (dolist (dcs list-sources)
     (setf (edge-source dcs) NIL)
     (setf (edge-lb-prop dcs) NIL)
     (setf (edge-ub-prop dcs) NIL)))

;;; ---------------------------------------------------------------------------


(defun relax-out-bf (time-map dc lb-prop ub-prop)

   (let* ((tpu (edge-from dc))
          (tpv (edge-to dc))
          (old-ub-v (time_point-ub tpv))
          (old-lb-v (- (time_point-lb tpv)))
	  (new-ub-v (+ (time_point-ub tpu) (edge-ub dc)))
          (new-lb-v (- (- (time_point-lb tpu)) (edge-lb dc)))	  
          (tpv-change NIL))
   
     (if ub-prop
         (when (< new-ub-v old-ub-v)
           (if (< new-ub-v (time_point-lb tpv))
               (progn
	         (setf *fail* T)
                 (get-ncycle-origin-ub time-map dc tpv tpu new-ub-v old-lb-v))
	       (if (and *source-check* (edge-source dc))
		   (if (edge-ub-prop dc)
                       (progn
	                 (setf *fail* T)
                         (get-ncycle-norigin-ub 
                              time-map dc tpu tpv new-ub-v old-ub-v))
		     (setf (edge-ub-prop dc) T))))
           (when (not *fail*)
	     (setf *sum-of-nodes* (+ *sum-of-nodes* 1))
             (setf (time_point-ub tpv) new-ub-v)
	     (setf (time_point-ub-prop tpv) T)
             (setf tpv-change T)

	     (if (= (time_point-name tpv) 1)
	         (format T "~%TP-V=~A TP-U=~A OLD-UB=~D NEW-UB=~D"
		            tpv tpu old-ub-v new-ub-v))

             (setf (time_point-st_edge_ub tpv) dc)
             (setf (time_point-pu tpv) tpu))))

     (if (and lb-prop (not *fail*))
         (when (< new-lb-v old-lb-v)
           (if (> (- new-lb-v) (time_point-ub tpv))
               (progn
	         (setf *fail* T)
	         ;;;(format T "~%TP-V=~A TP-U=~A OLD-UB-V=~D NEW-LB-V=~D"
		 ;;;           tpv tpu (time_point-ub tpv) new-lb-v)
                 (get-ncycle-origin-lb 
                      time-map dc tpv tpu new-lb-v (time_point-ub tpv)))
	       (if (and *source-check* (edge-source dc))
		   (if (edge-lb-prop dc)
                       (progn
	                 (setf *fail* T)
                         (get-ncycle-norigin-lb 
                              time-map dc tpu tpv new-lb-v old-lb-v))
		    (setf (edge-lb-prop dc) T))))
           (when (not *fail*)
	     (setf *sum-of-nodes* (+ *sum-of-nodes* 1))
             (setf (time_point-lb tpv) (- new-lb-v))
	     (setf (time_point-lb-prop tpv) T)
             (setf tpv-change T)
             (setf (time_point-st_edge_lb tpv) dc)
             (setf (time_point-pl tpv) tpu))))
     tpv-change)

) ;;; END *RELAX-OUT-BF*

;;; ---------------------------------------------------------------------------


(defun relax-in-bf (time-map dc lb-prop ub-prop)

   (let* ((tpv   (edge-from dc))
          (tpu   (edge-to dc))
          (old-ub-v (time_point-ub tpv))
          (old-lb-v (- (time_point-lb tpv)))
	  (new-ub-v (- (time_point-ub tpu) (edge-lb dc)))
          (new-lb-v (+ (- (time_point-lb tpu)) (edge-ub dc)))	  
          (tpv-change NIL))

     (if ub-prop
         (when (< new-ub-v old-ub-v)
           (if (< new-ub-v (time_point-lb tpv))
               (progn
	         (setf *fail* T)
                 (get-ncycle-origin-ub time-map dc tpv tpu new-ub-v old-lb-v))
	       (if (and *source-check* (edge-source dc))
		   (if (edge-ub-prop dc)
                       (progn
	                 (setf *fail* T)
                         (get-ncycle-norigin-ub 
                              time-map dc tpu tpv new-ub-v old-ub-v))
		     (setf (edge-ub-prop dc) T))))
           (when (not *fail*)
	     (setf *sum-of-nodes* (+ *sum-of-nodes* 1))
             (setf (time_point-ub tpv) new-ub-v)
             (setf tpv-change T)
	     (setf (time_point-ub-prop tpv) T)

             (if (= (time_point-name tpv) 1)
	         (format T "~%TP-V=~A TP-U=~A OLD-UB=~D NEW-UB=~D"
		           tpv tpu old-ub-v new-ub-v))

             (setf (time_point-st_edge_ub tpv) dc)
             (setf (time_point-pu tpv) tpu))))

     (if (and lb-prop (not *fail*))
         (when (< new-lb-v old-lb-v)
           (if (> (- new-lb-v) (time_point-ub tpv))
               (progn
	         (setf *fail* T)
                 ;;;(format T "~%TP-V=~A TP-U=~A OLD-UB-V=~D NEW-LB-V=~D"
		 ;;;           tpv tpu (time_point-ub tpv) new-lb-v)
                 (get-ncycle-origin-lb 
                      time-map dc tpv tpu new-lb-v (time_point-ub tpv)))
	     (if (and *source-check* (edge-source dc))
		 (if (edge-lb-prop dc)
                     (progn
	               (setf *fail* T)
                       (get-ncycle-norigin-lb 
                            time-map dc tpu tpv new-lb-v old-lb-v))
		   (setf (edge-lb-prop dc) T))))
         (when (not *fail*)
	   (setf *sum-of-nodes* (+ *sum-of-nodes* 1))
           (setf (time_point-lb tpv) (- new-lb-v))
	   (setf (time_point-lb-prop tpv) T)
           (setf tpv-change T)
           (setf (time_point-st_edge_lb tpv) dc)
           (setf (time_point-pl tpv) tpu))))
     tpv-change)

) ;;; END *RELAX-IN-BF*

;;; ---------------------------------------------------------------------------


(defun add-tps-prop-lb (tp)

 (let ((listin (time_point-in tp))
       (listout (time_point-out tp))
       (d-list-visit NIL)
       (d-list-out NIL))
   
   (dolist (dcc listin)
     (let ((tp-from (edge-from dcc)))
  
       (setf *sum-of-nodes* (+ *sum-of-nodes* 1))
       (if (equal (time_point-pl tp-from) tp)
	   (push tp-from d-list-visit))
       (if (not (time_point-in-queue tp-from))
	   (push tp-from d-list-out))
       (setf (time_point-lb-prop tp-from) T)))

   (dolist (dcc listout)
     (let ((tp-to (edge-to dcc)))
  
       (setf *sum-of-nodes* (+ *sum-of-nodes* 1))
       (if (equal (time_point-pl tp-to) tp)
	   (push tp-to d-list-visit))
       (if (not (time_point-in-queue tp-to))
	   (push tp-to d-list-out))
       (setf (time_point-lb-prop tp-to) T)))

    (values d-list-visit d-list-out)))

;;; ---------------------------------------------------------------------------


(defun tps-tree-prop-lb (tp-start)

  (let ((list-tps-out (list tp-start))
        (list-visit (list tp-start))
	(tpc NIL))

    (setf *sum-of-nodes* (+ *sum-of-nodes* 1))
    (loop 
      (if (null list-visit) (return list-tps-out))
      (setf tpc (pop list-visit))
      
      (setf (time_point-in-queue tpc) T)
      (setf (time_point-lb-prop tpc) T)
      (setf (time_point-lb tpc) (- *infinity*))
	
      (multiple-value-bind (delta-list-visit delta-list-out)
			   (add-tps-prop-lb tpc)
	(setf list-visit (append list-visit delta-list-visit))
        (setf list-tps-out (append list-tps-out delta-list-out))))))

;;; ---------------------------------------------------------------------------


(defun add-tps-prop-ub (tp)

 (let ((listin (time_point-in tp))
       (listout (time_point-out tp))
       (d-list-visit NIL)
       (d-list-out NIL))
   
   (dolist (dcc listin)
     (let ((tp-from (edge-from dcc)))
  
       (setf *sum-of-nodes* (+ *sum-of-nodes* 1))
       (when (equal (time_point-pu tp-from) tp)
	 (push tp-from d-list-visit)
	 (if (= (time_point-name tp-from) 1)
	     (format T "~%TP-FROM=~A TP=~A" tp-from tp)))
       
       (if (not (time_point-in-queue tp-from))
	   (push tp-from d-list-out))
       (setf (time_point-ub-prop tp-from) T)))

   (dolist (dcc listout)
     (let ((tp-to (edge-to dcc)))
  
       (setf *sum-of-nodes* (+ *sum-of-nodes* 1))
       (if (equal (time_point-pu tp-to) tp)
	   (push tp-to d-list-visit))
       (if (not (time_point-in-queue tp-to))
	   (push tp-to d-list-out))
       (setf (time_point-ub-prop tp-to) T)))

    (values d-list-visit d-list-out)))

;;; ---------------------------------------------------------------------------


(defun tps-tree-prop-ub (tp-start)

  (let ((list-tps-out (list tp-start))
        (list-visit (list tp-start))
	(tpc NIL))
    (if (= (time_point-name tp-start) 1) (print "TP-START REF-POINT!"))
    (setf *sum-of-nodes* (+ *sum-of-nodes* 1))
    (loop 

      (if (null list-visit) (return list-tps-out))
      (setf tpc (pop list-visit))
      (if (= (time_point-name tpc) 1) (print "UPDATE REF-POINT!"))      
      (setf (time_point-in-queue tpc) T)
      (setf (time_point-ub-prop tpc) T)
      (setf (time_point-ub tpc) *infinity*)
	
      (multiple-value-bind (delta-list-visit delta-list-out)
			   (add-tps-prop-ub tpc)
	(setf list-visit (append list-visit delta-list-visit))
        (setf list-tps-out (append list-tps-out delta-list-out))))))

;;; ---------------------------------------------------------------------------


(defun init-tree-prop (tp-from tp-to)
    
  (append 
    (if (equal (time_point-pu tp-to) tp-from)
	(tps-tree-prop-ub tp-to)
      (if (equal (time_point-pu tp-from) tp-to)
	  (tps-tree-prop-ub tp-from)
	NIL))
    (if (equal (time_point-pl tp-to) tp-from)
	(tps-tree-prop-lb tp-to)
      (if (equal (time_point-pl tp-from) tp-to)
	  (tps-tree-prop-lb tp-from)
	NIL))))

;;; ---------------------------------------------------------------------------


(defun propagation-q (time-map list-init-queue)
  
  (setf *fail* NIL)

  (let ((list-q list-init-queue))

    (loop
      ;;; Exit condition
      (if (null list-q) (return T))

      (let* ((tpc (pop list-q))
	     (lb-prop (time_point-lb-prop tpc))
	     (ub-prop (time_point-ub-prop tpc))
	     (listin (time_point-in tpc))
             (listout (time_point-out tpc)))

	(setf (time_point-in-queue tpc) NIL)
	(setf (time_point-lb-prop tpc) NIL)
        (setf (time_point-ub-prop tpc) NIL)

        (dolist (dcc listin)
          (when (relax-in-bf time-map dcc lb-prop ub-prop)
	    (when (not (time_point-in-queue (edge-from dcc)))
	      (setf (time_point-in-queue (edge-from dcc)) T)
	      (setf list-q (append list-q (list (edge-from dcc)))))))

        (dolist (dcc listout)
          (when (relax-out-bf time-map dcc lb-prop ub-prop)	    
	    (when (not (time_point-in-queue (edge-to dcc)))
	      (setf (time_point-in-queue (edge-to dcc)) T)
	      (setf list-q (append list-q (list (edge-to dcc))))))))))

) ;;; END *PROPAGATION-Q*


(defun propagation (tm)

  (setf *fail* NIL)
  (setf (time_map-ncycle-list-lb tm) NIL)
  (setf (time_map-ncycle-list-ub tm) NIL)

  (let ((list-q (init-list-q (time_map-dc_sources tm))))

    (loop
      ;;; EXIT CONDITION
      (if *fail*
	  (progn
	    (delete-queue-marks list-q)
	    (delete-source-marks (time_map-dc_sources tm))
	    (setf (time_map-dc_sources tm) NIL)
	    (return NIL))
	  (if (null list-q)
	      (progn
	        (delete-source-marks (time_map-dc_sources tm))
	        (setf (time_map-dc_sources tm) NIL)
	        (return T))))

      (let* ((tpc (pop list-q))
	     (lb-prop (time_point-lb-prop tpc))
	     (ub-prop (time_point-ub-prop tpc))
	     (listin (time_point-in tpc))
             (listout (time_point-out tpc)))

	(setf (time_point-in-queue tpc) NIL)
	(setf (time_point-lb-prop tpc) NIL)
        (setf (time_point-ub-prop tpc) NIL)

        (dolist (dcc listin)
          (if (not *fail*)
            (when (relax-in-bf tm dcc lb-prop ub-prop)
	      (when (not (time_point-in-queue (edge-from dcc)))
	        (setf (time_point-in-queue (edge-from dcc)) T)
	        (setf list-q (append list-q (list (edge-from dcc))))))))

        (dolist (dcc listout)
          (if (not *fail*)
            (when (relax-out-bf tm dcc lb-prop ub-prop)
	      (when (not (time_point-in-queue (edge-to dcc)))
	        (setf (time_point-in-queue (edge-to dcc)) T)
	        (setf list-q (append list-q (list (edge-to dcc)))))))))

      ))
   
   ) ;;; END *PROPAGATION*


;;; --------------------------------------------------------------------------------------------------------------------
;;; Funcion igual que propagation para el algoritmo BELLMAN-FORD con actualizacion de interseccion entre 2 tp en caso de que haya 
;;; mas de un valor entre ellos.  
;;;   TIME_MAP-NCYCLE-LIST-LB
;;; --------------------------------------------------------------------------------------------------------------------

(defun propagate-BF (tm)

  (setf *fail* NIL)
  (setf (time_map-ncycle-list-lb tm) NIL)
  (setf (time_map-ncycle-list-ub tm) NIL)

  (update_all_bounds (time_map-dc_sources tm))
  (let ((list-q (init-list-q (time_map-dc_sources tm))))

    (loop
      ;;; EXIT CONDITION
      (if *fail*
	  (progn
	    (delete-queue-marks list-q)
	    (delete-source-marks (time_map-dc_sources tm))
	    (setf (time_map-dc_sources tm) NIL)
	    (return NIL))
	  (if (null list-q)
	      (progn
	        (delete-source-marks (time_map-dc_sources tm))
	        (setf (time_map-dc_sources tm) NIL)
	        (return T))))

      (let* ((tpc (pop list-q))
	     (lb-prop (time_point-lb-prop tpc))
	     (ub-prop (time_point-ub-prop tpc))
	     (listin (time_point-in tpc))
             (listout (time_point-out tpc)))

	(setf (time_point-in-queue tpc) NIL)
	(setf (time_point-lb-prop tpc) NIL)
        (setf (time_point-ub-prop tpc) NIL)

        (dolist (dcc listin)
          (if (not *fail*)
            (when (relax-in-bf tm dcc lb-prop ub-prop)
	      (when (not (time_point-in-queue (edge-from dcc)))
	        (setf (time_point-in-queue (edge-from dcc)) T)
	        (setf list-q (append list-q (list (edge-from dcc))))))))

        (dolist (dcc listout)
          (if (not *fail*)
            (when (relax-out-bf tm dcc lb-prop ub-prop)
	      (when (not (time_point-in-queue (edge-to dcc)))
	        (setf (time_point-in-queue (edge-to dcc)) T)
	        (setf list-q (append list-q (list (edge-to dcc)))))))))

      ))
   
   ) ;;; END *PROPAGATION*

;;; ---------------------------------------------------------------------------


(defun relax-an-edge (time-map delta dc)
  
  (setf (edge-ub dc) (+ (edge-ub dc) delta))
  ;;; (format T "~%DELTA=~D" delta)
  (propagation-q time-map (init-tree-prop (edge-from dc) (edge-to dc)))

) ;;; END *RELAX-AN-EDGE*

;;; ---------------------------------------------------------------------------


(defun remove-an-edge (tm dc)

  (setf (time_point-in (edge-to dc))
        (remove dc (time_point-in (edge-to dc))))
  (setf (time_point-out (edge-from dc))
        (remove dc (time_point-out (edge-from dc))))	

  (propagation-q tm (init-tree-prop (edge-from dc) (edge-to dc)))

  (setf (edge-to dc) NIL)
  (setf (edge-from dc) NIL)

) ;;; END *REMOVE-AN-EDGE*

;;; ---------------------------------------------------------------------------
;;; ---------------------------------------------------------------------------
;;; *END-OF-FILE*