;;; This file has all things that I need to rewrite from the usual
;;; behavior of Prodigy4.0

;;; It includes also the work by Sira Vegas.

(in-package "P4")

;; Until I recompile Prodigy again.
(defvar *incremental-pending-goals* nil)


;;;*************************************************************
;;; Redefinition of this function to allow it not to ask
;;;*************************************************************
(defun ask-for-more-solutions ()
  (cond (user::*dont-ask-multiple-sols-p* user::*multiple-solutions-flg*)
	(t 
	 (format t
		 "~% Do you want to search for another solution:~%  ~
		     0. No more solutions.~%  ~
		     1. Any other solution.~%  ~
		     2. Only different solutions.~%  ~
		     3. Only equal length or shorter solutions.~%  ~
		     4. Only different solutions of shorter or eq length.~%  ~
		     5. Only strictly shorter solutions.~%~
		     >> ")
	 (let ((answer (car (read-atoms))))
	   (case answer
	     (0 (format t  "~% BYE, BYE!!"))
	     (1 :any-other)
	     (2 :different)
	     (3 :shorter)
	     (4 :shorter-and-diff)
	     (5 :strict-shorter)
	     (t (break "set answer by hand")))))))

;;;******************************************************************
;;; This is just to get rid of the message of reading meta-predicate
;;;******************************************************************
(defun check-lhs-exp (exp)
  (cond ((not (listp exp)) 0)
	((member (car exp) '(user::and user::or))
	 (check-lhs-list (cdr exp)))
	((eq (car exp) 'user::~)
	 (check-lhs-exp (second exp)))
	((member (car exp)
		 (problem-space-all-preds *current-problem-space*))
	 ;; This probably won't work, because you don't have type
	 ;; information in control rules. Probably the best I can do
	 ;; is take the maximum of all templates with this predicate name.
	 (template-abstraction-level exp nil))
	;; Jim changed this to print out a meta predicate only once
	;; per load, because it might be easier to spot typos that way
	(t
	 (unless (member (car exp)
		      (getf (problem-space-plist *current-problem-space*)
			    :read-meta-preds))
	   (push (car exp)
		 (getf (problem-space-plist *current-problem-space*)
		       :read-meta-preds))
	   (if user::*printp*
	       (format t "Reading Meta predicate: ~A~%" (car exp))))
	 (if (member (car exp) '(user::known user::true-in-state))
	     (check-lhs-exp (second exp))
	     0))))

;;; **********************************************************************
;;;  To get rid of the message of loading the functions file
;;; **********************************************************************

(defun domain (&optional name
			 &key (path *world-path*) (load-domain t) (compile nil)
			 (control t)
			 ;; I changed the default to source until we fix
			 ;; domains to have compiled files for more
			 ;; than one platform.
			 (function :source))
  (declare (special *world-path* *directory-separator* *problem-path*))
  (cond (name
	 (let ((*print-case* :downcase)
	       (user::*load-verbose* nil)
	       (old-control-rule (macro-function 'control-rule)))
	   (setf *problem-path*
		 (cleanup-string
		  (format nil "~A~C~S~C"
			  path *directory-separator* name *directory-separator*)))
	   ;; If function loading is specified, check the file exists
	   ;; and load if so.
	   (when function
	     ;; I commented this out because it is confusing when
	     ;; there is no functions file.
	     ;; (format t "Loading the functions file.~%")
	     ;;
	     ;; Daniel
	     ;;    changed:     :verbose t
	     ;;
	     (load (cleanup-string
		    (format nil "~A~C~A~Cfunctions~A"
			    *world-path* *directory-separator* name
			    *directory-separator*
			    (if (eq function :source)
				(format nil ".~a" user::*lisp-extension*)
			        user::*binary-extension*)))
		   :verbose nil
		   :if-does-not-exist nil))
	   (unless control
	     (setf (macro-function 'control-rule)
		   #'large-wobbly-control-rule))
	   (load (cleanup-string
		  (format nil "~A~Cdomain.~A"
			  *problem-path* *directory-separator*
			  user::*lisp-extension*)))
	   (setf (macro-function 'control-rule) old-control-rule)
	   (when load-domain
	     (if user::*printp*
		 (format t "Running load-domain.~%"))
	     (load-domain :compile compile))))
	;; If no name is specified, provide a listing of the current path.
	(t
	 (unless (char= (elt path (1- (length path)))
			*directory-separator*)
	   (setf path
		 (concatenate 'string path
			      (string *directory-separator*))))
	 (format t "~%Possible domains with the path ~S~%" path)
	 (pprint (mapcar #'file-namestring (directory path))))))



;;; ********************************************************
;;; Purpose:
;;; Code to not add the goal choices if apply is selected.
;;; In order NOT to get this behavior, just
;;; set *no-goal-choices-if-apply-p* to nil

;;; ********************************************************
;;; Author: Manuela Veloso - November 1992
;;; Send bugs and suggestions to mmv@cs.cmu.edu
;;; Change Log: none yet
;;; *******************************************************


;;; This redefines the function expand-binding-or-applied-op
;;; Added code is marked - just two lines

(defun expand-binding-or-applied-op (node)
  "Expand a binding node"
  (declare (type a-or-b-node node))

  (let* ((next (compute-next-thing node))
	 (applicables
	  (cache a-or-b-node-applicable-ops-left
		 node
		 (abs-generate-applicable-ops
		  node next)))
	 (subgoals
;; Daniel removed all this from Manuela's fix, and moved it to the
;; case in this function. If we delete the goals before knowing
;; whether we are going to sub-goal or not, we might be removing the
;; solution
;;	  ;; **********************************
;;	  ;; New code
;;	  ;; **********************************
	  (if (and applicables *no-goal-choices-if-apply-p*)
	      (cache a-or-b-node-goals-left node
		     (setf (a-or-b-node-pending-goals node)
			   nil))
;;	      ;; **********************************
;;	      ;; End of new code
;;	      ;; **********************************
	      (cache a-or-b-node-goals-left node
		     (setf (a-or-b-node-pending-goals node)
			   (delete-if
			    #'(lambda (goal)
				(goal-loop-p node goal))
			    (abs-generate-goals node next)))))))
    (if (and (null applicables) (null subgoals))
	(close-node node :no-choices)
	(let ((what-next (choose-apply-or-subgoal node applicables subgoals)))
	  (case what-next
	    (:apply
	     (if applicables
		 (do-apply-op node applicables next)
		 (close-node node :no-choices)))
	    (:sub-goal
	     (if subgoals
		 (do-subgoal node subgoals next)
		 (close-node node :no-choices)))
	    (otherwise what-next))))))


;;;************************************************************
;;; From apply.lisp
;;; This is for having the same behavior with and without rules
;;;************************************************************

;; If all the children die for reasons like unsolvable goals and
;; there are no binding select or reject control rules, mark
;; this operator node as unuseable.
;; Daniel has deleted these two preconds from the and
;;         (null (problem-space-select-bindings *current-problem-space*))
;;         (null (problem-space-reject-bindings *current-problem-space*))

(defun check-term-operator-node (node child child-failure)
  (declare (type operator-node node)
	   (type binding-node child)
	   (ignore child child-failure)
	   (special *current-problem-space*))
  (if (and (null (operator-node-bindings-left node))
	   (every #'(lambda (bind-node)
		      (let ((term (getf (nexus-plist bind-node)
					:termination-reason)))
			(or (eq term :uses-unsolvable-goal)
			    (and (consp term)
				 (eq (car term) :causes-goal-loop)))))
		  (nexus-children node)))
      (close-node node :no-good-bindings)))

;; If all the children failed for reasons like no binding choices,
;; and there are no select or reject operator control rules, we can
;; determine that this goal would never have succeeded. This is a
;; bit tricky, though.
;; Daniel has deleted these two preconds from the and
;;         (null (problem-space-select-operators *current-problem-space*))
;;         (null (problem-space-reject-operators *current-problem-space*))

(defun check-term-goal-node (node child child-failure)
  (declare (type goal-node node)
	   (ignore child child-failure)
	   (special *current-problem-space*))
  (if (and (null (goal-node-ops-left node))
	   (every #'(lambda (op-node)
		      (let ((term (getf (nexus-plist op-node)
					:termination-reason)))
			(or (eq term ':no-good-bindings)
			    (and (consp term)
				 (eq (car term) :causes-goal-loop)))))
		  (nexus-children node)))
      (close-node node :no-good-operators)))


;;;  To solve the problem of backtrack-past-undoable-goal when there
;;;  is the same node twice in the list of introducing operators of a
;;;  goal node
;;; Also to solve the problem of marking as undoable a node that it is
;;; already in a solution path
;;;
;;; close-node is called when we find there are no more children to
;;; expand at a node. Delete the node from the list of expandable
;;; nodes in the problem space, and then maybe propagate the failure
;;; back up to do some dependency-directed backtracking type stuff.
;;; Returns the failure.

;;; This isn't perfect - for example if closing a node because it
;;; can't be refined is enough to tell us we could kill off a goal
;;; loop, we'll miss that. But hey, it's a start.

(defun close-node (node &optional failure)
  (declare (type nexus node)
	   (special *current-problem-space*))
  "Delete the node from the list of nodes."

  ;; this test stops infinite recursion.
  (when (not (getf (nexus-plist node) :under-deletion))
    (setf (getf (nexus-plist node) :under-deletion) t)

    ;; This bit removes this node.
    (delete-node-link node *current-problem-space*)
    (setf (problem-space-property :expandable-nodes)
	  (delete node (problem-space-property :expandable-nodes)))

    ;; Record the reason to close the node if we know it.

    ;; Daniel. Added the test to whether it is a solution node, so
    ;; that when finding multiple-sols, it will not change the label
    ;; of a solution leaf node
    
    (if (and failure
	     (not (eq :achieve (getf (nexus-plist node)
				     :termination-reason))))
	(setf (getf (nexus-plist node) :termination-reason)
	      failure))

    ;; end Daniel
    
    ;; Propagate the node closing if we can.
    ;; refine failures are propagated regardless of the node type.
    (if (eq failure :not-refineable)
	(let* ((deepest (elt (problem-space-property :deepest-point)
			     (1- (nexus-abs-level node))))
	       (ancestor (and deepest (not (eq deepest node))
			      (ancestor deepest node))))
	  (if ancestor
	      (close-node (nexus-parent node) :not-refineable)))

      (unless (problem-space-property :chronological-backtracking)
	(typecase node
	  (operator-node
	   (check-term-goal-node (nexus-parent node) node failure))
	  (binding-node
	   (check-term-operator-node (nexus-parent node) node failure))
	  (goal-node

;; Daniel
	   (setf (goal-node-introducing-operators node)
		 (remove-duplicates (goal-node-introducing-operators node)
				    :test #'eq))
;; end Daniel
	   
	   (check-term-intro-nodes (goal-node-introducing-operators node)
				   node failure)))))
    (setf (getf (nexus-plist node) :under-deletion) nil))
  failure)

;; Sira. 10/6/95
;; One more parameter needed to check if the function was called
;; from PRODIGY or from my code.

(defun form-all-bindings (bound-vars relevant-vars op bindings &optional done)
  (let ((unbound-vars
	  (if done 
	     (my-set-difference relevant-vars bound-vars)
	     (my-sort (set-difference relevant-vars bound-vars) op)))
	(res nil))
    (if unbound-vars
	(dolist (binding bindings res)
	  (setf res
		(append (get-some-bindings unbound-vars op binding)
			res)))
      bindings)))

;; The variables have to be bound in the same order as they were
(defun my-set-difference (set1 set2)
  (let ((result nil))
    (dolist (e set1 (reverse result))
      (if (not (member e set2 :test #'eq))
	  (push e result)))))


(defun get-some-bindings (unbound-vars op binding)
  (if (null unbound-vars)
      (list binding)
    (let ((new-bindings (form-new-bindings
			 (car unbound-vars) op binding))
	  (res nil))
      (if (null (cdr unbound-vars))
	  new-bindings
	(dolist (bind new-bindings res)
	  (setf res
		(nconc (get-some-bindings
			(cdr unbound-vars) op bind)
		       res)))))))


(defun form-new-bindings (var op binding)
  (if (infinite-type-var var op)
      (let ((inf-var-binds
	      (if (assoc var (car (getf (rule-plist op) :costs)))
		  (gen-binding-for-inf-var var binding op t)
		  (gen-binding-for-inf-var var binding op))))
	(cross-product (form-real-bindings var inf-var-binds)
		       (list binding)))
    (let ((var-binds (if (assoc var (car (getf (rule-plist op) :costs)))
			 (generate-values-for-var var op t)
			 (generate-values-for-var var op))))
      (form-cross-product-and-check-functions
       var op (form-real-bindings var var-binds)
       (list binding)))))



(defun infinite-type-var (var op)
  (declare (special *current-problem-space*))
  (let ((spec (second (or (assoc var (second (rule-precond-exp op)))
			  (assoc var (car (getf (rule-plist op) :costs)))
			  (assoc var (second (rule-effects op)))))))
    (and (listp spec)
	 (is-infinite-type-p (second spec) *current-problem-space*))))


(defun gen-binding-for-inf-var (var binding op &optional done)
  (let* ((spec  (if done
		(second (assoc var (car (getf (rule-plist op) :costs))))
		(second (assoc var (second (rule-effects op))))))
	 (generator (sublis binding (third spec))))
    (apply (car generator) (cdr generator))))


(defun generate-values-for-var (var op &optional done)
  (let ((spec (if done
	     (second (assoc var (car (getf (rule-plist op) :costs))))
	     (second (assoc var (second (rule-effects op)))))))
    (mapcan #'get-values-from-types (make-type-list spec))))



#|

;;; Vars are no longer needed
;;;*************************************************************
;;; Adding a new prop to control rules: vars.
;; This is the code to load control rules.  
(defmacro CONTROL-RULE (&body body)
;; so that it does not disturb me with the too long story  
  (check-control-rule (cons (car body) (cddr body)))
  (let ((var (gensym)))
    `(let ((,var (create-control-rule ',body)))
      (case (control-rule-type ,var)
	(:select-node
	 (add-rule-new ,var (problem-space-select-nodes
			     *current-problem-space*)))
	(:reject-node
	 (add-rule-new ,var (problem-space-reject-nodes
			     *current-problem-space*)))
	(:prefer-node
	 (add-rule-new ,var (problem-space-prefer-nodes
			     *current-problem-space*)))
	(:select-goal
	 (add-rule-new ,var (problem-space-select-goals
			     *current-problem-space*)))
	(:reject-goal
	 (add-rule-new ,var (problem-space-reject-goals
			     *current-problem-space*)))
	(:prefer-goal
	 (add-rule-new ,var (problem-space-prefer-goals
			     *current-problem-space*)))
	(:select-op
	 (add-rule-new ,var (problem-space-select-operators
			     *current-problem-space*)))
	(:reject-op
	 (add-rule-new ,var (problem-space-reject-operators
			     *current-problem-space*)))
	(:prefer-op
	 (add-rule-new ,var (problem-space-prefer-operators
			     *current-problem-space*)))
	(:select-binding
	 (add-rule-new ,var (problem-space-select-bindings
			     *current-problem-space*)))
	(:prefer-binding
	 (add-rule-new ,var (problem-space-prefer-bindings
			     *current-problem-space*)))
	(:reject-binding
	 (add-rule-new ,var (problem-space-reject-bindings
			     *current-problem-space*)))
	(:apply-or-subgoal
	 (add-rule-new ,var (problem-space-apply-or-subgoal
			     *current-problem-space*)))
	(:refine-or-expand
	 (add-rule-new ,var (problem-space-refine-or-expand
			     *current-problem-space*)))
	(otherwise
	 (error "~S is a an unknown type of control rule.~%" ',body))))))

;; this code creates a control rule data structure.
;; The then slot only gets everything after the name of the object
;; which this rule controls (e.g. GOAL, OP,...) since 
(defun create-control-rule (rule)
  (declare (list rule))
  (make-control-rule :name (first rule)
		     :plist `(vars ,(second rule))
		     :type (extract-type (fourth rule))
		     :if (second (third rule))
		     :then (fourth rule)))
|#

;;; ******************************************************************
;;;   And now things to store the fired control rules, plus the things
;;;   to control the firing of rules, so that it fires all selection rules
;;;   Also, they set a variable that contains the
;;;   current-control-rule. This is used in the type-of-object
;;;   meta-predicate to control the values of the unbound variables.
;;;
;;;   From matcher-interface.lisp
;;; ******************************************************************

;;;
;;; I don't think I need for now for nodes. If so, please refer to the
;;; wonderfully done below. :-)

;;; ******************************************************************
;;; Daniel This is for operators
;;; ******************************************************************

;;; Daniel. So that if *keep-default-only-p* is t, it will only leave
;;; the default as the only alternative. If the number of fired-crs
;;; before is equal to the number after, then the decision was not
;;; made by rules

;;; return all possible uninstantiated operators that can possibly achieve
;;; the given goal.  It's obtained by first running select rules, then
;;; reject rules. The copy-list here is so that we can use delete when
;;; we iterate through the rules, which should be more efficient.
(defun generate-operators (node goal)
  (declare (special *current-problem-space*))
  (let ((fired-crs user::*number-fired-crs*)
	(user::*candidate-operators*
	 (scr-select-ops
	  node goal (problem-space-select-operators *current-problem-space*))))
    (declare (special user::*candidate-operators*))
    (if (eq user::*candidate-operators* :big-lose)
	:big-lose
        (let ((candidate-ops
	       (scr-reject-ops node user::*candidate-operators*
			       (problem-space-reject-operators
				*current-problem-space*))))
	  (if (and user::*keep-default-only-p*
		   (= fired-crs user::*number-fired-crs*))
	      (list (car candidate-ops))
	      candidate-ops)))))

#|
;; In Prodigy, the else part of last if is:
	       (scr-reject-ops node user::*candidate-operators*
			       (problem-space-reject-operators
				*current-problem-space*))
|#

;;; look at select-operator control rules in sequence, stop at the first one
;;; that fires and return the selected operator, or if no rule fires, return
;;; all relevent operators by accessing slots in problem space.

;;; Daniel changed this so that all rules are fired
;;; Also, selects the op with greater number of votes. If a control
;;; rule fires more than once on the same op, those votes are also counted
(defun scr-select-ops (node goal ctrlrules)
  (let ((ops nil))
    (setf (getf (nexus-plist node) 'user::pending-goals) nil)
    (cond (*first-selection-p*
	   (setq ops (find-if #'(lambda (cr)
				  (let ((matchedp
					 (caar
					  (match-select-op node goal cr))))
				    (when matchedp
				      (if user::*store-fired-crs-p*
					  (push (cons cr matchedp)
						(getf (nexus-plist node)
						      'user::fired-crs)))
				      (incf user::*number-fired-crs*)
				      matchedp)))
			      ctrlrules)))
 
;;; ********** Principio de Jose Luis Heras. **********

	  (user::*use-probabilities* 
	   (setq ops (user::best-operators goal node ctrlrules)))
	  
;;; ********** Final de Jose Luis Heras. **********

	  (t (dolist (cr ctrlrules)
	       (dolist (selected-op (match-select-op node goal cr))
		 (let ((op (car selected-op))
		       (votes (cdr selected-op)))
		   (when op
		     (if user::*store-fired-crs-p*
			 (push (list cr op)
			       (getf (nexus-plist node) 'user::fired-crs)))
		     (incf user::*number-fired-crs*)
		     (cond ((member op ops :test #'eq)
			    (incf (getf (rule-plist op) 'votes) votes))
			   (t (push op ops)
			      (setf (getf (rule-plist op) 'votes) 1)))))))
	     (if ops
		 (sort ops #'> :key #'(lambda (op)
					(getf (rule-plist op) 'votes))))))
    (or ops (get-all-ops goal node) :big-lose)))

#|

(dolist (cr ctrlrules)
	       ;; the rest of the list is eq to the car
	       (let* ((selected-ops-votes (match-select-op node goal cr))
		      (selected-ops (car selected-ops-votes))
		      (op (car selected-ops))
		      (votes (cdr selected-ops-votes)))
		 (when op
		   (if user::*store-fired-crs-p*
		       (push (list cr op)
			     (getf (nexus-plist node) 'user::fired-crs)))
		   (incf user::*number-fired-crs*)
		   (cond ((member op ops :test #'eq)
			  (incf (getf (rule-plist op) 'votes) votes))
			 (t (push op ops)
			    (setf (getf (rule-plist op) 'votes) 1))))))
	     (if ops
		 (sort ops #'> :key #'(lambda (op)
					(getf (rule-plist op) 'votes))))

    (cond (*first-selection-p*
	   (setq ops (find-if #'(lambda (cr)
				  (let ((matchedp
					 (delete-duplicates
					  (match-select-op node goal cr))))
				    (when matchedp
				      (if user::*store-fired-crs-p*
					  (push (cons cr matchedp)
						(getf (nexus-plist node)
						      'user::fired-crs)))
				      (incf user::*number-fired-crs*)
				      matchedp)))
			      ctrlrules)))

	   (if (or *first-selection-p* (not user::*random-selection-p*))
	       ops
	       (let ((choice (random-choice ops)))
		 (if user::*keep-all-selections-p*
		     (cons choice (delete choice ops))
		     (list choice)))))
|#

(DEFUN random-choice (l)
  (nth (random (length l)) l))
				   

;;; match select-operator control-rule; if it fires, return set of operators
;;; specified in the right-hand side of the rule, otherwise, return nil.
;;; Daniel. My control rules can only have one operator, without variables.
(defun match-select-op (node goal rule)
  (declare (ignore goal)
	   (special *current-problem-space*))
  ;; Daniel
  (setf user::*current-control-rule* rule)
  ;; end Daniel
  (let ((initial-time (get-internal-run-time))
	(matched (match-lhs (control-rule-if rule) node)))
;; Daniel. To control time spent on matching each rule.
    (if user::*utilityp*
	(control-utility rule matched initial-time))
;; Daniel
;; deleted the call to delete-duplicates in the let, since it is
;; done now on the calling function. Also, removed the mapcar to make
;; it faster, since there are not variables in the rhs of my operator
;; rules. This is something not compatible with Prodigy!!
    (when matched
      (let* ((select-expr (fourth (control-rule-then rule)))
	     (selected-ops nil))
	(dolist (bindings matched)
	  (let ((selected-op (rule-name-to-rule (if (eq bindings t)
						    select-expr
						  (sublis bindings select-expr))
						*current-problem-space*)))
	    (if (cdr (assoc selected-op selected-ops :test #'eq))
		(incf (cdr (assoc selected-op selected-ops :test #'eq)))
	      (push (cons selected-op 1) selected-ops))))
	(if user::*print-rules-fired-p*
	    (format t "~%Firing select op rule ~S to get ~S"
		    (control-rule-name rule) (mapcar #'car selected-ops)))
	(output 3 t "~%Firing select op rule ~S to get ~S"
		(control-rule-name rule) (mapcar #'car selected-ops))
	selected-ops))))

;; ;; Daniel
;; ;; deleted the call to delete-duplicates in the let, since it is
;; ;; done now on the calling function. Also, removed the mapcar to make
;; ;; it faster, since there are not variables in the rhs of my operator
;; ;; rules. This is something not compatible with Prodigy!!
;;     (when matched
;;       (let* ((select-expr (fourth (control-rule-then rule)))
;; 	     (votes (length matched))
;; 	     (selected-ops (list (rule-name-to-rule select-expr
;; 						    *current-problem-space*))))
;; 	(if user::*print-rules-fired-p*
;; 	    (format t "~%Firing select op rule ~S to get ~S"
;; 		    (control-rule-name rule) selected-ops))
;; 	(output 3 t "~%Firing select op rule ~S to get ~S"
;; 		(control-rule-name rule) selected-ops)
;; 	(cons selected-ops votes)))

;;
;; Stores utility-based information
;;
(DEFUN control-utility (rule matchedp initial-time)
  (let ((le (gethash (control-rule-name rule) user::*cr-to-le-hash*)))
    (cond (le ;; there is induction
	   (incf (user::learning-episode-matching-time le)
		 (user::elapsed-time initial-time))
	   (incf (user::learning-episode-matchings le))
	   (if matchedp
	       (incf (user::learning-episode-firings le))))
	  (t ;; no induction (test phase or hamlet)
	   (incf (getf (control-rule-plist rule) :matching-time)
		 (user::elapsed-time initial-time))
	   (incf (getf (control-rule-plist rule) :matchings))
	   (if matchedp
	       (incf (getf (control-rule-plist rule) :firings)))))))


#|
(defun match-select-op (node goal rule)
  (declare (ignore goal)
	   (special *current-problem-space*))
  (let ((select-expr (fourth (control-rule-then rule)))
	(matched (match-lhs (control-rule-if rule) node)))
;; Daniel
;; deleted the call to delete-duplicates in the let, since it is
;; done now on the calling function	
    (when matched
      (let ((selected-ops
	     (mapcar #'(lambda (bindings)
			 (rule-name-to-rule
			  (if (eq bindings t)
			      select-expr
			    (sublis bindings select-expr))
			  *current-problem-space*))
		     matched)))
	(if user::*print-rules-fired-p*
	    (format t "~%Firing select op rule ~S to get ~S"
		    (control-rule-name rule) (car selected-ops)))
	(output 3 t "~%Firing select op rule ~S to get ~S"
		(control-rule-name rule) (car selected-ops))
	selected-ops)))
)

;;; Daniel. Changed so that it allows to control the number of fired
;;; control rules
(defun choose-operator (node ops)
  "Uses preference rules to pick a best operator."
  (declare (ignore node)
	   (special *current-problem-space*))
  (if (<= (length ops) 1)
      (car ops)
      (find-a-best-one
       ops (problem-space-prefer-operators *current-problem-space*)
       :operator)))
|#

;;; ******************************************************************
;;; Daniel This is for bindings
;;; ******************************************************************

;;; Daniel. So that if *keep-default-only-p* is t, it will only leave
;;; the default as the only alternative. If the number of fired-crs
;;; before is equal to the number after, then the decision was not
;;; made by rules


;;; I'm going to change the data structure for a binding from list of
;;; ob to (cons list-of-obj condi), so I can know from which
;;; conditional effect this binding is coming from.

;;; Currently fudging to avoid the issue of sub-bindings for abstractions.
;;; This is because a lot of extra power may come from only doing the
;;; matching once for the parts that are shared, and this is a tricky
;;; one that I must hash out with Mei.

;;; Changed 12/7/91 to do gathering step before making instantiated
;;; operators, and sorting afterwards. This should be more efficient.
;;; (It was really slow on a simple stripsworld problem.)

(defun generate-bindings (node goal op)
  (declare (special *current-problem-space*))
  (let* ((abs (nexus-abs-parent node))
	 (winner (if abs (nexus-winner abs)))
	 (instop (if winner (binding-node-instantiated-op winner))))

    ;; If there is a node from the abstraction level above, that
    ;; should be the only possibility.
    (cond (instop
	   (let ((res (copy-instantiated-op instop)))
	     ;; Copy because need different back pointer
	     (setf (instantiated-op-precond res) nil)
	     (list res)))
	  (t
	   (let* ((fired-crs user::*number-fired-crs*)
		  (user::*candidate-bindings*
		   (get-all-bindings node goal op
				     (if (negated-goal-p goal)
					  (operator-del-list op)
					  (operator-add-list op))))
		  (instantiated-ops
		   (convert-to-instantiated-ops
		    op user::*candidate-bindings*))
		  (*abslevel* (nexus-abs-level node)))
	     (declare (special user::*candidate-bindings* *abslevel*))
	     ;; Sort by conspiracy number if the switch is on.
	     ;; Daniel
	     (let ((alternatives
		    (if (problem-space-property :min-conspiracy-number)
			(progn
			  (mapc #'compute-conspiracy-number instantiated-ops)
			  (stable-sort instantiated-ops #'<
				       :key #'instantiated-op-conspiracy))
		        instantiated-ops)))
	       (if (and user::*keep-default-only-p*
			(= fired-crs user::*number-fired-crs*))
		   (list (car alternatives))
		   alternatives)))))))

#|
	     (if (problem-space-property :min-conspiracy-number)
		 (progn
		   (mapc #'compute-conspiracy-number instantiated-ops)
		   (stable-sort instantiated-ops #'<
				:key #'instantiated-op-conspiracy))
		 instantiated-ops)
|#

;;; Daniel changed this so that all rules are fired
;;;  rhs-bindings is what is returned by find-rhs-binding.
;;;  by calling match-select-bindings, it recurs through select-binding 
;;;  control rules.  returns bindings as soon as one control rule fires.  If
;;;  none of the rules fires, it returns all tuples that satisfy rhs-bindings.

;;; post-conditions is a list of all effect in either add list or del list of
;;; an operator. The list is sublisted by conditional.

;;; Daniel. Node added to the call of scr-select-bindings to reinit
;;; the pending goals property
(defun get-all-bindings (node goal op post-conditions)
  (declare (type nexus node)
	   (type literal goal)
	   (type operator op)
	   (list post-conditions))
  (if post-conditions
      (let ((rhs-bindings (find-rhs-bindings (caar post-conditions) goal))
	    (other-bindings
	     (get-all-bindings
	      node goal op
	      (if (null (cdar post-conditions))
		  (cdr post-conditions)
		  (cons (cdar post-conditions) (cdr post-conditions)))
	      )))
	(if (null rhs-bindings)
	    other-bindings
	    (let ((binding1 (scr-select-bindings op node rhs-bindings)))
	      (cond ((eq (car binding1) t)
		     (or other-bindings binding1))
		    (binding1
		     (append binding1 other-bindings))
		    (t
		     other-bindings)))))))

;;; Daniel. Node added to reinit the pending goals property
(defun scr-select-bindings (op node rhs-bindings)
  (setf (getf (nexus-plist node) 'user::pending-goals) nil)
  (let ((bindings (rete-gen-bindings op rhs-bindings)))
    (if bindings
	(form-final-bindings bindings rhs-bindings op))))

(defun rete-gen-bindings (op rhs-bindings)
  (let* ((foo (rule-generator op))
	 (data (if foo (funcall foo (change-to-real-pb (car rhs-bindings)))))
	 (all-simple-tests (rule-simple-tests op))
	 (all-unary-tests (rule-unary-tests op))
	 (all-join-tests (rule-join-tests op))
	 (all-neg-simple-tests (rule-neg-simple-tests op))
	 (all-neg-unary-tests (rule-neg-unary-tests op))
	 (all-neg-join-tests (rule-neg-join-tests op))
	 (number-of-rules (- (length all-unary-tests) 2)))

    (if (car all-unary-tests)
	(do* ((n number-of-rules (1- n))
	      (rule (get-n-control-rule n op)
		    (get-n-control-rule n op))
	      (initial-time (get-internal-run-time) (get-internal-run-time))
	      (unary-tests all-unary-tests (cdr unary-tests))
	      (join-tests all-join-tests (cdr join-tests))
	      (simple-tests  all-simple-tests  (cdr simple-tests))
	      (simple-test (car simple-tests) (car simple-tests))
	      (precedence (simple-match simple-test)  (simple-match simple-test))
	      (neg (if precedence
		       (neg-simple-match all-neg-simple-tests op))
		   (if precedence
		       (neg-simple-match all-neg-simple-tests op)))
	      (tuple (if neg (match (car unary-tests) 
				    (car join-tests)
				    all-neg-unary-tests
				    all-neg-join-tests
				    (cons neg data)
				    op))
		     (if neg (match (car unary-tests) 
				    (car join-tests) 
				    all-neg-unary-tests
				    all-neg-join-tests
				    (cons neg data)
				    op)))
	      (result nil))
		     
	     ((or (and tuple *first-selection-p*)
		  (null (cdr unary-tests)))
;;	      (format t "~%Tuple: ~a~%Result: ~a~%" tuple result)
	      ;; Daniel. To control time spent on matching each rule.
	      (if (and user::*utilityp* rule)
		  (control-utility rule tuple initial-time))
	      (if tuple
		  (update-fired-crs rule tuple))
	      (cond (*first-selection-p* tuple)
		    (result)
		    (t tuple)))
;;	  (format t "~%Tuple: ~a~%Result: ~a~%" tuple result)
	  (if (and user::*utilityp* rule)
	      (control-utility rule tuple initial-time))
	  (when tuple
	    (update-fired-crs rule tuple)
	    (dolist (el tuple)
	      (pushnew el result :test #'equal))))
	  
	;; if the op does not have vars, then there will not be any
	;; binding control rule.  don't worry about firing stuff.

      ;; Daniel. In case of operators without vars, we set the
      ;; user::*current-control-rule* variable to that distinct label
      
	(and (setf user::*current-control-rule* :no-vars)
	     (simple-match (car all-simple-tests))
	     (neg-simple-match all-neg-simple-tests op)))))

#|
		       (if (and user::*random-selection-p* result)
			   (let ((choice (random-choice result)))
			     (if user::*keep-all-selections-p*
				 (cons choice (delete choice result
						      :test #'equal))
			       (list choice)))
			   result)
|#

;;
;; Returns the nth rule of the bindings rules, and it sets the
;; variable user::*current-control-rule* needed for the type-of-object
;; meta-predicate
;;
(DEFUN get-n-control-rule (n op)
  (if (>= n 0)
      (let ((rule (nth n (rule-select-bindings-crs op))))
	(if rule
	    (setf user::*current-control-rule* rule)))))

(DEFUN update-fired-crs (cr tuple)
  (declare (special user::*current-node*))
  (when cr
    (if user::*store-fired-crs-p*
	(push (cons cr tuple)
	      (getf (p4::nexus-plist user::*current-node*) 'user::fired-crs)))
    (incf user::*number-fired-crs*)
    (if user::*print-rules-fired-p*
	(format t "~%Firing select binding rule ~S at node ~S"
		(control-rule-name cr)
		(nexus-name user::*current-node*)))
    (output 3 t "~%Firing select binding rule ~S at node ~S"
	    (control-rule-name cr)
	    (nexus-name user::*current-node*))))

#|
;;; Daniel. Changed so that it allows to control the number of fired
;;; control rules
(defun choose-binding (node bindings)
  "Uses preference rules to pick a best binding."
  (declare (type operator-node node) (list bindings))

  (let ((*the-operator* (operator-node-operator node)))
    (declare (special *the-operator* *current-problem-space*))
    (or (eq bindings t)
	(and (<= (length bindings) 1) (car bindings))
	(find-a-best-one
	 bindings (problem-space-prefer-bindings *current-problem-space*)
	 :bindings))))
|#

;;; ******************************************************************
;;; Daniel This is for goals
;;; ******************************************************************

;;; Daniel. So that if *keep-default-only-p* is t, it will only leave
;;; the default as the only alternative

(defun abs-generate-goals (node next)
  ;; "next" is the next step in the more abstract space
  (declare (type a-or-b-node node))
  (let ((fired-crs user::*number-fired-crs*)
	(goals (if next
		   (delete-if
		    #'(lambda (goal)
			;; delete only if it comes from a higher
			;; abstraction space, but is not the next step in
			;; the plan, (see comments above).
			(and (> (compute-abstraction-level
				 goal *current-problem-space*)
				(nexus-abs-level node))
			     (or (not (goal-node-p next))
				 (not (eq goal (goal-node-goal next))))))
		    (generate-goals node))
		   (generate-goals node))))
    (if (and user::*keep-default-only-p*
	     (= fired-crs user::*number-fired-crs*))
	(list (car goals))
        goals)))


;; Recurs through all select-goals control rules to select set of goals to work
;; on. Stop recuring at the first rule that fires.
;; If no control rules fire, return all pending-goals.

;;; Daniel changed it so that now it calls all rules
;;; Also, selects the goal with greater number of votes. If a control
;;; rule fires more than once on the same goal, those votes are also counted
;;; The pending goals are initialized to nil. They are used to cache
;;; the pending goals for target-goal meta-predicate
(defun scr-select-goals (node ctrlrules)
  (let ((goals nil))
    (setf (getf (nexus-plist node) 'user::pending-goals) nil)
    (cond (*first-selection-p*
	   (setq goals (find-if #'(lambda (cr)
				    (let ((matchedp
					   (delete-duplicates
					    (match-select-goal node cr))))
				      (when matchedp
					(if user::*store-fired-crs-p*
					    (push (cons cr matchedp)
						  (getf (p4::nexus-plist node)
							'user::fired-crs)))
					(incf user::*number-fired-crs*)
					matchedp)))
				ctrlrules)))

;;; ********** Principio de Jose Luis Heras. **********

	  (user::*use-probabilities*
	   (case (user::node-type node)
	     (user::binding-node
	      (if (not (eq (car (user::node-info node (user::node-type node)))
			   'prodigy4::*finish*))
		  (setq goals (user::best-goals node ctrlrules))))
	     (user::applied-op-node
	      (setq goals (user::best-goals node ctrlrules)))))

;;; ********** Final de Jose Luis Heras **********

	  (t (dolist (cr ctrlrules)
	       (dolist (goal (match-select-goal node cr))
		 (cond ((member goal goals :test #'eq)
			  (if (and user::*store-fired-crs-p*
				   (not (member cr (getf (nexus-plist node)
							 'user::fired-crs)
						:test #'eq :key #'car)))
			      (push (list cr goal)
				    (getf (nexus-plist node)
					  'user::fired-crs)))
			(incf (getf (literal-plist goal) 'votes)))
		       (t (push goal goals)
			  (if user::*store-fired-crs-p*
			      (let ((memberp (member cr
						     (getf (nexus-plist node)
							   'user::fired-crs)
						     :test #'eq :key #'car)))
				(if memberp
				    (push goal (cdar memberp))
				    (push (list cr goal)
					  (getf (nexus-plist node)
						'user::fired-crs)))))
			  (incf user::*number-fired-crs*)
			  (setf (getf (literal-plist goal) 'votes) 1)))))
	     (if goals
		 (sort goals #'> :key #'(lambda (goal)
					  (getf (literal-plist goal)
						'votes))))))
    (or goals
	(if *incremental-pending-goals*
	    (a-or-b-node-pending-goals node)
	    (give-me-all-pending-goals node)))))

#|
	(if (or *first-selection-p* (not user::*random-selection-p*))
	    goals
	    (let ((choice (random-choice goals)))
	      (if user::*keep-all-selections-p*
		  (cons choice (delete choice goals))
	          (list choice))))
|#

;; Match one select goal control rule.  Returns nil if the rule doesn't fire,
;; return the goals specified by the rule if the rule fires.
(defun match-select-goal (node rule)
  (declare (type nexus node)
	   (type control-rule rule))
  ;; Daniel
  (setf user::*current-control-rule* rule)
  ;; end Daniel
  (let ((initial-time (get-internal-run-time))
	(select-expr (fourth (control-rule-then rule)))
	(matched (match-lhs (control-rule-if rule) node)))
;; Daniel. To control time spent on matching each rule.
    (if user::*utilityp*
	(control-utility rule matched initial-time))
;; Daniel
;; deleted the call to delete-duplicates in the let, since it is
;; done now on the calling function	
    (if matched 
	(let ((selected-goals
	       (mapcar #'(lambda (bindings)
			   (let ((res (sublis bindings select-expr)))
			     (if (literal-p res)
				 res
			         (instantiate-consed-literal res))))
		       matched)))
	  (if user::*print-rules-fired-p*
	      (format t "~%Firing select goals rule ~S to get ~S"
		      (control-rule-name rule) selected-goals))
	  (output 3 t "~%Firing select goals rule ~S to get ~S"
		  (control-rule-name rule) selected-goals)
	  selected-goals))))

#|
;;;
;;; Daniel. For controling number of crs fired
;;;
(defun choose-goal (node goals)
  (declare (special *current-problem-space*)
	   (ignore node))
  (if (<= (length goals) 1)
      (car goals)
      (find-a-best-one
       goals (problem-space-prefer-goals *current-problem-space*) :goal)))
|#

;;; ******************************************************************
;;; Daniel This is for apply/subgoal
;;; ******************************************************************

;;; I change this to fire all rules.
(defun choose-apply-or-subgoal (node app-ops goals)
  "If there are no control rules, decides to apply as soon as it can."
  (declare (special *current-problem-space*))
;; Heras uses it
;; 	   (ignore goals))
;; Daniel (list app-ops goals)
;; not needed for now (setf (getf (nexus-plist node) 'user::pending-goals) nil)
  (cond ((null app-ops) :sub-goal)
	(t (let ((user::*applicable-op* (car app-ops))
		 (rules (problem-space-apply-or-subgoal
			 *current-problem-space*)))
	     (declare (special user::*applicable-op*))

;;; ******* Principio de Jose Luis Heras

	     (if user::*use-probabilities*
		 (user::select-apply-or-subgoal app-ops goals rules)
		 (if rules
		     (scr-choose-apply-or-subgoal node rules)
		     :apply))))))

#|
	     (when (or (eq user::*use-probabilities* 'rules)
		       (eq user::*use-probabilities* 'both))
		   (if rules
		       (setq salida (user::select-apply-or-subgoal 
				     app-ops goals rules))))
	     (if (and (or (eq user::*use-probabilities* 'rules)
			  (eq user::*use-probabilities* 'both))
		      salida)
		 salida
	       (if rules
		   (scr-choose-apply-or-subgoal node rules)
	         :apply))))))
|#		     
	       
;;; ****** Final de Jose Luis Heras
;; Antes de JL Heras:
;; 	     (if rules
;; 		 (scr-choose-apply-or-subgoal node rules)
;; 	         :apply)))))
  
(defun scr-choose-apply-or-subgoal (node rules)
  (declare (list app-ops rules))
  (if *first-selection-p*
      (or (some #'(lambda (rule)
		    (match-a-or-s-rule node rule))
		rules)
	  :apply)
      (let ((votes-apply 0)
	    (votes-subgoal 0))
	(dolist (rule rules)
	  (case (match-a-or-s-rule node rule)
	    (:apply (incf votes-apply))
	    (:sub-goal (incf votes-subgoal))
	    (t nil)))
	(if (> votes-subgoal votes-apply)
	    :sub-goal
	    :apply))))
			
(defun match-a-or-s-rule (node rule)
  (declare (type control-rule rule))
  ;; Daniel
  (setf user::*current-control-rule* rule)
  ;; end Daniel
  (let ((initial-time (get-internal-run-time))
	(matchedp (match-lhs (control-rule-if rule) nil)))
    ;; Daniel. To control time spent on matching each rule.
    (if user::*utilityp*
	(control-utility rule matchedp initial-time))
    (if matchedp
	(let ((decision (second (control-rule-then rule))))
;;Daniel
	  (if user::*store-fired-crs-p*
	      (pushnew (cons rule decision)
		       (getf (p4::nexus-plist node) 'user::fired-crs)
		       :test #'(lambda (cr-selection1 cr-selection2)
				 (and (eq (car cr-selection1)
					  (car cr-selection2))
				      (eq (cdr cr-selection1)
					  (cdr cr-selection2))))))
	  (incf user::*number-fired-crs*)
;;
	  (if user::*print-rules-fired-p*
	      (format t "~%Firing apply/subgoal ~S deciding ~S"
		      (control-rule-name rule) decision))
	  (output 3 t "~%Firing apply/subgoal ~S deciding ~S"
		  (control-rule-name rule) decision)
	  ;; There are no bindings on the right hand side, so the match
	  ;; that is returned is ignored.
	  (case decision
	    (user::sub-goal :sub-goal)
	    (user::apply   :apply))))))



;;; ******************************************************************
;;; Daniel This is for refine-or-expand
;;; ******************************************************************

;;; I think I don't need them
#|
(defun scr-choose-refine-or-expand (node rules)
  (cond ((null rules)
	 (if (getf (nexus-plist node) 'winning-child)
	     :refine
	     :expand))
	((match-r-or-e-rule node (car rules)))
	(t (scr-choose-refine-or-expand node (cdr rules)))))

(defun match-r-or-e-rule (node rule)
  (declare (type control-rule rule))
  ;; Daniel
  (setf user::*current-control-rule* rule)
  ;; end Daniel
  (when (match-lhs (control-rule-if rule) nil)
    (let ((decision (second (control-rule-then rule))))
;;Daniel
      (if user::*store-fired-crs-p*
	  (setf (getf (p4::nexus-plist node) 'user::fired-crs) rule))
	  (incf user::*number-fired-crs*)
;;
	  (if user::*print-rules-fired-p*
	      (format t "~%Firing apply/subgoal ~S deciding ~S"
		      (control-rule-name rule) decision))
	  (output 3 t "~%Firing apply/subgoal ~S deciding ~S"
	      (control-rule-name rule) decision)
      ;; There are no bindings on the right hand side, so the match
      ;; that is returned is ignored.
      (case decision
	((user::refine :refine) :refine)
	((user::expand :expand)  :expand)))))


;;; It is not needed anymore, since we account for them in the
;;; function fire-prefer
;;; Daniel. Allows to control the number of fired prefer crs.
;;;
(defun my-find-a-best-one (candidates rules type)
  "Allows to control the number of fired crs"
  (declare (list candidates rules)
	   (symbol type)		; :operator, :bindings, etc
	   (special *current-problem-space*))
  (multiple-value-bind (result firedp)
      (find-a-best-one candidates rules type)
    (if firedp
	(incf user::*number-fired-crs*))
    (values result firedp)))
|#

;;; Changed this so it wouldn't alter the order of the candidates, at
;;; least for goals and operators. It's harder to do for bindings, and
;;; I'm not so sure it matters.

;;; Daniel. Changed so that if pref is equal to preferee, it accounts
;;; for a firing of rule. It does not compute the utility, yet
(defun fire-prefer (rule pref bindings candidates preferee type)
  (let* ((real-preferred-things
	  (delete-duplicates
	   (translate-rule-symbols bindings pref candidates type)))
	 (firedp (and (= (length real-preferred-things) 1)
		      (equal (car real-preferred-things) preferee)))
	 (preferred-things (delete preferee real-preferred-things)))
    (cond (preferred-things
	   (incf user::*number-fired-crs*)
	   (output 3 t "~%Firing pref rule ~S for ~S over ~S"
		   (control-rule-name rule) preferred-things preferee))
	  (firedp 
	   (incf user::*number-fired-crs*)
	   (output 3 t "~%Firing pref rule ~S for ~S over all"
		   (control-rule-name rule) preferee))
	  (t nil))
    preferred-things))


;;;
;;; Daniel. This is because of the non existence of the "preferred
;;; over what" element in the prefer rules
;;;
(defun desired-binding (target object type)
  "Return a list containing the binding which would allow us to find something
preferred to object by the preference statement of this control rule, or nil."
  (declare (special *the-operator*))
  (if target   ;;; added by Daniel
      (ecase type
	;; No bindings for operators or nodes.
	(:operator
	 (cond ((strong-is-var-p target)
		(list (list (cons target object))))
	       ((eq (operator-name object) target) (list nil))))
	(:goal				; I assume the object is a literal.
	 (cond ((strong-is-var-p target)
		(list (list (cons target object))))
	       ;; Reject if the negated-ness doesn't match up.
	       ((not (or (and (eq (car target) 'user::~)
			      (literal-neg-goal-p object))
			 (and (not (eq (car target) 'user::~))
			      (literal-goal-p object))))
		nil)
	       (t
		(if (eq (car target) 'user::~)
		    (setf target (second target)))
		(let ((result (user::instance-of object target)))
		  (if (or (eq result t) (eq (car result) t))
		      (list nil)
		    result)))))
	;; For bindings, the target is specified as a binding list, and
	;; object is an instantiated-operator.
	(:bindings
	 (cond ((strong-is-var-p target)
		(list (list (cons target object))))
	       (t
		(let ((ordered-vars-or-objects
		       (sublis target (operator-vars *the-operator*))))
		  (list (unify (instantiated-op-values object)
			       ordered-vars-or-objects))))))
	(:node
	 (cond ((strong-is-var-p target) (list (list (cons target object))))
	       ((and (numberp target) (= (nexus-name object) target))
		(list nil)))))
      (list nil)))  ;;; added by Daniel

;;;**********************************************************************
;;; From the result.lisp to solve the problem of solutionp
;;;**********************************************************************

;;; Daniel. I changed what it returns in the case of multiple-sols.
(eval-when (load eval compile)
  (defmacro p4::solutionp (result all-solutions)
    `(and (if (p4::problem-space-property :multiple-sols)
	      (if ,result
		  (cdr ,all-solutions)
		  (or (cdar ,all-solutions)
		      (cdr ,all-solutions)))
	      (and (listp (car ,result))
		   (listp (caar ,result))
		   (eq (cdaar ,result) :achieve)))
	  t)))

(defun user::pp-result (result stream z)
  (declare (ignore z))

  (format stream "#<PRODIGY result: ~a, ~d secs, ~d nodes, ~d sol, ~d this-cost, ~d best-cost>"
	  (user::prodigy-result-solutionp result)
	  (user::prodigy-result-time result)
	  (user::prodigy-result-nodes result)
	  (user::prodigy-result-number-solutions result)
	
	  ;; Sira. 2/4/96.
	  ;; Adding to the print function the cost of the current solution
	  ;; and the best cost found.
	  (getf (user::prodigy-result-plist result) :this-solution-cost)
	  (getf (user::prodigy-result-plist result) :best-solution-cost)))

(in-package "P4")

;;;**********************************************************************
;;; From treeprint.lisp. For telling the quality of the solution.
;;;**********************************************************************

;;; This function is used to print the final solution out to the user.
;;; The plan is a list of instantiated-operators.

(defun announce-plan (plan)
  (declare (list plan))
  ;; Sira. 19/4/96
  ;; New variables to print statistics.
  (declare (special *node-counter* user::*this-cost* user::*best-cost* 
		    *my-number-sols* *my-time*))
  (let ((plevel (or (problem-space-property :*output-level*) 2)))
    (when (> plevel 0)
      (cond ((and (consp (car plan)) (consp (caar plan))
		  (eq (caaar plan) :partial-achieve))
	     (format t "~%I didn't completely solve the problem,")
	     (if (eq (second (car (second (car plan)))) :resource-bound)
		 (format t "~%because I exceeded a resource bound, ~S,"
			 (third (car (second (car plan))))))
	     (format t "~%but I solved ~S of the top-level goals with this plan:~%~%"
		     (length (third (caar plan)))))
	    (t
	     (format t "~%~%Solution:~%")))
      (let ((*print-case* :downcase))
	(dolist (op (cdr plan))
	  (princ #\Tab)
	  (brief-print-inst-op op)
	  (terpri)))
      (terpri)
      
      ;; Sira. 19/4/96
      ;; That's what I have added
      (let ((time (float (/ (- (get-internal-run-time) *my-time*)
			    internal-time-units-per-second))))
	(incf *my-number-sols*)


	(cond ((problem-space-property :make-span)
	       (when (or (= user::*DurationMax* user::*MakeSpan* most-positive-fixnum)
			 (not user::*no-finish*))
		     (setf user::*DurationMax* 0)
		     (setf user::*MakeSpan* 0))
	       (format t "#<THIS result: ~d secs, ~d nodes, ~d sol, ~d this-cost, ~d best-cost, ~d this-makespan, ~d best makespan>"
		       time *node-counter* *my-number-sols*
		       user::*this-cost* user::*best-cost*
		       user::*DurationMax* user::*MakeSpan*)
	      )
	      (t (format t "#<THIS result: ~d secs, ~d nodes, ~d sol, ~d this-cost, ~d best-cost>"
			 time *node-counter* *my-number-sols*
			 user::*this-cost* user::*best-cost*)))))))
          
;(format t "~%:make-span= ~d :max-make-span= ~d~%" (problem-space-property :make-span) (problem-space-property :max-makespan))

#|(if (= user::*DurationMax* user::*MakeSpan* most-positive-fixnum)
    (progn
	  (setf user::*DurationMax* 0)
	  (setf user::*MakeSpan* 0)
    );;fin del progn
);;fin del if
|#

;;;**********************************************************************
;;; From search.lisp. To get rid of the "Setting the depth bound" message
;;; plus the messages on solutions.
;;;**********************************************************************
;;; call again (main-search node last-node ...) where
;;;node: node to which backtrack  (Jim says: "the node most
;;;       recently selected for attention.")
;;;last-node-created: the last-node as returned by prev call to
;;;      main-search
;;;current-depth: get it from node or last-node-created (or is there
;;;     a global var for it?)
;;;depth-bound:I guess this is the initial value :depth-bound, but I
;;;can reset it here as Manuela does so the next solution is shorter 

(defun call-main-search (plan answer current-depth-bound)
  (let* ((last-node-created (cdar plan))
	 restart-node current-depth
	 ;; I don't know why but *current-depth* is 1 less
	 ;; than the depth that appears to the left of last node
	 ;; (new-depth-bound (1+ *current-depth*))
	 (new-depth-bound (1+ (compute-depth last-node-created))))
    (close-node last-node-created)
    ;;from Jim's comments in search.lisp, restart-node is
    ;;either a binding or applied-op node. Thus the decision
    ;;between subgoal or apply has to be made
    (setf restart-node
	  (choose-node last-node-created
		       (generate-nodes last-node-created)))
    (cond
      (restart-node
       (setf current-depth
	     (maintain-state-and-goals last-node-created restart-node))
       (main-search
	restart-node last-node-created current-depth
	;;set the new depth-bound (but better solutions
	;;could be longer!) so new solutions are of the
	;;same or smaller length
	(cond ((or (eq answer :shorter)
		   (eq answer :shorter-and-diff))
;;	       (format t "Setting new depth bound: ~A.~%"
;;		       new-depth-bound)
	       new-depth-bound)
	      ((eq answer :strict-shorter)
;;	       (format t "Setting new depth bound: ~A.~%"
;;		       (- new-depth-bound 3))
	       (- new-depth-bound 3))
	      (t current-depth-bound))))
      (t
       (if user::*printp*
	   (format t "~%There are no more solutions.~%"))))))

;;; Delete this when recompile the code for Prodigy4.0
(defun run (&key (current-problem (current-problem))
	         search-default
		 depth-bound
		 max-nodes
		 time-bound
		 output-level
		 ;;Sira. Two parameters added: cost-bound, and cost-type 
		 cost-bound
		 cost-type
		 ;;IPSS new parameters: makespan, max-makespan, min-makespan, input-file and resources
		 (make-span (problem-space-property :make-span))
		 min-makespan
		 max-makespan
		 ;;;(input-file (problem-space-property :input-file)) ;Marzo 2003 MD
		 (include-resources (problem-space-property :include-resources)) ;Marzo 2003 MD
		 num-TP
		 activity-distance
             ;;;; --------------------------------- MD ----------------------------
		 
		 (permute-application-order :not-set)
		 abstraction-level
		 (compute-abstractions
		  (problem-space-property :compute-abstractions))
                 (shortest (problem-space-property :shortest))
		 (multiple-sols (problem-space-property :multiple-sols)))
  "This function solves the problem! (Yeah, right.)
It initialises for problem solving and then calls main-search, which
manages everything."
  (declare (special *current-problem-space* *user-aborted-search*
		    *last-result-obtained*
		    user::*best-cost* user::*cost-label* user::*this-cost*
		    *my-time* *my-number-sols*))

  (setf *dont-ask-and-find-all* nil);;alicia

  (setf *user-aborted-search* nil) ; RDS 93/12/11
 
  ;; First check that load-domain has been called, and a problem loaded.
  (unless (problem-space-property :domain-loaded)
    (cerror "Optionally call load-domain and continue"
	    "It doesn't seem as though you called load-domain")
    (if (y-or-n-p "Call load-domain? ")
	(load-domain)
	(format t "If this breaks call (load-domain) and start again")))
  (unless current-problem
    (cerror "Load a problem and continue" "There is no specified problem.")
    (format t "Specify a file to load:~%")
    (load (read-line)))

  ;; Set the value for the output level for this run.
  (if (or (and (integerp output-level)
	       (>= output-level 0)
	       (<= output-level 3))
	  (and (null (problem-space-property :*output-level*))
	       (or output-level (setf output-level 2))))
      (setf (problem-space-property :*output-level*) output-level))

  ;; Ok, we carried on, so load the new problem
  ;; (also resetting the problem space)
  (load-problem current-problem (not compute-abstractions))

  ;; Now we annotate preconditions with their abstraction levels
  ;; so as to make planning at different levels faster.
  (add-abstraction-levels-to-operators *current-problem-space*)
  (cond ((numberp abstraction-level))
	((numberp (abs-level))
	 (setf abstraction-level (abs-level)))
	(t;; otherwise set to highest interesting level
	 (setf abstraction-level
	       (max-abs
		(getf (rule-plist
		       (rule-name-to-rule '*finish* *current-problem-space*))
		      :annotated-preconds)))))

  (let ((done-goal (instantiate-consed-literal (list 'done))))
    
    ;; Make "done" be the top level goal
    (setf (literal-goal-p done-goal) (list :top-level-goal))

    ;; We'll set the goal node by hand (see function top-nodes).
    (setf (problem-space-expanded-goals *current-problem-space*)
	  (list done-goal)))

  ;; If there are no control rules for nodes, or if a search-method
  ;; has been explicitly requested, set up here to miss out the usual
  ;; search control and go for a (hopefully) faster version.
  (if (and (null search-default)
	   (null (problem-space-select-nodes *current-problem-space*))
	   (null (problem-space-reject-nodes *current-problem-space*))
	   (null (problem-space-prefer-nodes *current-problem-space*)))
      (setf search-default :depth-first))
  (setf (problem-space-property :search-default) search-default)

  ;; Daniel. It was not set before
  (setf (problem-space-property :multiple-sols) multiple-sols)

   ;;MD para make-span, ficheros de entrada y el calculo de recursos
  (setf (problem-space-property :make-span) make-span)
  (setf (problem-space-property :max-make-span) max-makespan)
  (setf (problem-space-property :include-resources) include-resources)
  ;;;;(setf (problem-space-property :input-file) input-file)
  (setf (problem-space-property :num-TP) num-TP)
  (setf (problem-space-property :activity-distance) activity-distance)


  ;; Jim. Changed to only set the flag if this key is used in the call.
  (unless (eq permute-application-order :not-set)
    (setf (problem-space-property :permute-application-order)
	  permute-application-order))

  ;; Use the depth-bound installed in the problem space, or default to
  ;; 30 if none.
  (if (null depth-bound)
      (setf depth-bound
	    (or (problem-space-property :depth-bound) 30)))

  ;; Sira. 3/5/95.Giving value to user::*best-cost* and user::*cost-label*
  (if (null cost-bound)
      (setf user::*best-cost* most-positive-fixnum)
      (setf user::*best-cost* cost-bound))

  (setf user::*this-cost* 0)
  (if (null cost-type)
      (setf user::*cost-label* 'user::default)
      (setf user::*cost-label* cost-type))

 ;;New parameters: maximum makespan, resources, etc. MD Marzo 2003
  (cond ((problem-space-property :multiple-sols)
	 (cond (max-makespan
		(setf user::*DurationMax* max-makespan))
	       (t
		(setq max-makespan most-positive-fixnum)
		(setf user::*DurationMax* max-makespan)))
	 (if min-makespan
	     (setf user::*DurationMin* min-makespan)
	   (setf user::*DurationMin* 0)))
	(t
	 (if max-makespan
	     (setf user::*DurationMax* max-makespan)
	   (setf user::*DurationMax* most-positive-fixnum))
	 (if min-makespan
	     (setf user::*DurationMin* min-makespan)
	   (setf user::*DurationMin* 0))))

      (if num-TP
          (setf user::*num* num-TP)
      )

      ;; MD modificaciones para calcular o no recursos. Marzo 2003
      (if (problem-space-property :include-resources) ; Si no se pone recursos no se llama a la parte de recursos.
	  (setf user::*calculate-resources* include-resources) ;
	(setf user::*calculate-resources* nil) ; Se coge el recurso de los parametros del run
	)
 ;(setf user::*MakeSpan* max-makespan) ;Sea igual al valor maximo
      (setf user::*MakeSpan* user::*DurationMax*) ;Sea igual al valor maximo
   
;-----------------------------MD ----------------------------------


  (when (and (null (getf (problem-space-plist *current-problem-space*) 
			 :additive-costs)) 
	     (null (getf (problem-space-plist *current-problem-space*) 
			 :non-additive-costs)))
    (setf user::*costs-declared* nil))


  (setf *all-solutions* nil);;it may have some from previous runs

  ;; Compute the static part of the information for eager inference
  ;; rules. Has to be computed after function top-nodes is called,
  ;; because it may need the predicates added/deleted by static-eager
  ;; inference rules. 
  ;; This should be called before (top-nodes abstraction-level) is
  ;; called. 
  (dolist (rule (problem-space-eager-inference-rules *current-problem-space*))
    (unless (static-inference-rule-p rule)
      (setf (getf (rule-plist rule) :possible-bindings)
	    (get-possible-bindings rule))))

  ;; Install the first two nodes in the path, which are the operator
  ;; and null bindings for *finish*. Then run the recursive search
  ;; the binding node as the most recently selected for attention.
  (let* ((*node-counter* 0)
	 (start-node (top-nodes abstraction-level))
	 (initial-time (get-internal-run-time)))
    (declare (special *node-counter* *prodigy-handlers*)
	     (type nexus start-node))
    

    ;; Sira 19/4/96
    ;; getting the first time
    (setf *my-time* initial-time)
    (setf *my-number-sols* 0)

    (add-handlers max-nodes depth-bound time-bound)
    ;; (setf (problem-space-property :next-thing) nil)

    ;; initialise the pending goals list
    (setf *pending-goals* (list :pending-goals-head))

    ;; This signal provides a way for users to add code that gets
    ;; called at the start of any run. No data passed as yet.
    (prod-signal :start-run)
    
    ;; Ok, go for it.
    (let* ((intermediate
	    (cond ((and shortest multiple-sols)
		   ;; RDS 93/08/18.  User wants to search for shortest plan.
		   (run-shortest-multiple start-node depth-bound))
		  (shortest
		   (run-shortest start-node depth-bound))
		  (multiple-sols
		   (run-multiple-sols start-node depth-bound))
		  (t
		   (let* ((result (main-search
				   start-node start-node 2 depth-bound))
			  (plan (prepare-plan result)))
		     ;; Print the plan nicely
		     ;; Sira. 19/4/96
		     ;; Two new parameters to the function announce plan,
		     ;; to print the time, number of nodes, and costs.
		     (if plan (announce-plan plan))
		     (or plan result)))))
	   (time (float (/ (- (get-internal-run-time) initial-time)
			   internal-time-units-per-second)))
	   (solutionp (solutionp intermediate *all-solutions*))
	   (solutions (if solutionp
			  (get-solutions intermediate *all-solutions*)))
	   (number-solutions (length solutions)))

      ;; Set the fields of the result structure
      (setf (user::prodigy-result-problem user::*prodigy-result*) current-problem)
      (setf (user::prodigy-result-time user::*prodigy-result*) time)
      (setf (user::prodigy-result-nodes user::*prodigy-result*) *node-counter*)
      (setf (user::prodigy-result-solutionp user::*prodigy-result*) solutionp)
      
      ;; Sira. The solution we want not is not the first one, but the last-one,
      ;; because it is the one of the highest quality.
      (setf (user::prodigy-result-solution user::*prodigy-result*)
	    (car solutions))

      ;; Sira. Storing the cost of the solution
      (setf (getf (user::prodigy-result-plist user::*prodigy-result*)
		  :this-solution-cost) user::*this-cost*)

      (setf (getf (user::prodigy-result-plist user::*prodigy-result*)
		  :best-solution-cost) 
	    (if (< user::*best-cost* most-positive-fixnum)  
		user::*best-cost*
		'NONE))

      ;; They are reversed so that they are stored in the order in which
      ;; they were achieved
      (setf (user::prodigy-result-solutions user::*prodigy-result*)
	    (reverse solutions))
      (setf (user::prodigy-result-number-solutions user::*prodigy-result*)
	    number-solutions)
      (setf (user::prodigy-result-interrupt user::*prodigy-result*)
	    (car intermediate))
      (setf (user::prodigy-result-whole-tree-expanded-p user::*prodigy-result*)

    ; Changed the following so that it works with :shortest as well.
    ; RDS 93/12/11
            (cond ((problem-space-property :multiple-sols)
		   ;; Daniel, so that it is correctly computed
		   (whole-tree-expanded-p intermediate))
                  ((problem-space-property :shortest))
		  (t nil)))

      ;; Cumulative data
      (incf (user::prodigy-result-acc-runs user::*prodigy-result*))
      (incf (user::prodigy-result-acc-time user::*prodigy-result*) time)
      (incf (user::prodigy-result-acc-nodes user::*prodigy-result*) *node-counter*)
      (incf (user::prodigy-result-acc-solutions user::*prodigy-result*)
	    number-solutions)
      user::*prodigy-result*)))


(defun run-multiple-sols (start-node depth-bound)
  (do* ((result (main-search start-node start-node 2 depth-bound))
	     (plan (prepare-plan result)
	      (prepare-plan result))
	answer)
       (nil)
     
    (cond
      ((and plan			;only new solutions wanted
	    (and (or (eq answer :different)
		     (eq answer :shorter-and-diff))
		 (repeated-solution-p plan)))
       (setf result (call-main-search plan answer depth-bound)))

      ;;if plan is null it means that no sol was found 
      ((null plan) 
       (if user::*printp*
	   (format t "~%No solution found !!! ~A.~%" result))
       ;;Se inicializa las variables globales para la correcta ejecucion
       ;; Daniel. I am not sure this is the right thing to do. They should be done at the beginning, MD!
	    (initialise-plan)
       (store-solution result)
       (return result))			;terminate	   

      (t	  ;Print the plan nicely and store it
           
        ;;Si la solucion es *finish* no se muestra la rete, puesto que no se
       ;; ha creado.
       ;; Daniel. Only call this function if there is a rete (it is doing temporal reasoning)
         
        (if (and user::*no-finish* user::*rete*)
           ;;Se muestra la solucion de la rete.09/01/03 ----------------  MD para que no muestre los resultados
           (COMMON-LISP-USER::Visualizar-Red user::*rete* user::*num*)
        );fin del if
         
       (announce-plan plan)
       (store-solution plan)
       ;;and search some more if required
       (setf answer (ask-for-more-solutions))
       (cond
	 (answer
	  (setf result (call-main-search plan answer depth-bound))
	  (if (null result)
	     (progn
	      ;;found all solutions: note that :termination-reason
	      ;;will be the termination-reason of the last path, which
	      ;;is a failure even when other solutions have been
	      ;;found. (You can distinguish that case by checking
	      ;;*all-solutions*) Is this the behaviour we want?
	      (initialise-plan)
	      (return)
	      );fin del progn
	     );fin del if
	   );fin del condicion
	 (t;;problem was solved, but no more solutions are needed
	   
	  (return plan)))))))		;terminate
	  
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;	  
;;Redefinicion de la funcion add-handlers
	  
(defun add-handlers (max-nodes depth-bound time-bound)
  "Set up handlers to check termination criteria"
  (declare (ignore depth-bound)
	   (special *partial-satisfaction* *closest-to-goal*))

  (remove-prod-handler :achieve #'handle-solution)
  (define-prod-handler :achieve #'handle-solution)

  ;; This function is bound if we have loaded the code to salvage
  ;; partial plans on hitting resource bounds. This code is in
  ;; /contrib/partial-satisfaction.
  (when (and (fboundp 'maintain-best-plan))
    (remove-prod-handler :achieve-one-goal #'maintain-best-plan)
    (setf *closest-to-goal* nil)
    (when (and *partial-satisfaction*
	       (goal-is-conjunction-of-literals))
      (define-prod-handler :achieve-one-goal #'maintain-best-plan)))
    
  (remove-prod-handler :always #'max-check)
  (when (and max-nodes (numberp max-nodes))
    (eval `(defun max-check (sig)
	    (declare (ignore sig) (special *node-counter*))
	    (when (>= *node-counter* ,max-nodes)
	      (setf (problem-space-property :termination-reason)
		    :too-many-nodes)
		    
		    ;;Se inicializan las variables dela rete.lramos.12/01/03
		    (initialise-plan)
;; 		    (print 'ini)
		    
	      (output 1 t "~%Hit node limit (~S)~%" ,max-nodes)
	      '(:stop :resource-bound :node))))
    (compile 'max-check)
    (define-prod-handler :always #'max-check))

  (remove-prod-handler :always #'time-check)
  (when (and time-bound (numberp time-bound))
    (let* ((now (get-internal-run-time))
	   (stop-time (+ now (* time-bound internal-time-units-per-second))))
      (eval `(defun time-check (sig)
	      (declare (ignore sig))
	      (let ((time (get-internal-run-time)))
		(when (>= time ,stop-time)
		  (setf (problem-space-property :termination-reason)
			:out-of-time)
			
			;;Se inicializan las variables dela rete.lramos.12/01/03
		   (initialise-plan)
;; 		   (print 'ini)
		  
		  (output 1 t "~%Out of time after ~S seconds~%"
			  (/ (- time ,now) internal-time-units-per-second))
		  '(:stop :resource-bound :time))))))
    (compile 'time-check)
    (define-prod-handler :always #'time-check)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; This function gets called by the interrupt handler on solving a level.
(defun handle-solution (signal)
  (declare (cons signal)
	   (special *current-depth* *choose-this-node*))
  (let ((solution-node (cdr signal)))
    (declare (type nexus solution-node))
    (announce-solution solution-node *current-depth*)
    ;; If this is the ground (zero) level we're done.
    (cond ((zerop (nexus-abs-level solution-node)) ())
	   ;;Sira. This part only stops if there is no cost calculation
	   ;;(when (not user::*costs-declared*)
		 ;;(cons :stop :achieve)))
	  ;; Otherwise we mark up the solution path at this
	  ;; abstraction level with the "winning-child" property.
	  (t (setf (nexus-winner solution-node) :solution)
	     (let* ((first-new-node (mark-up-solution solution-node t))
		    (last-old-node
		     (if first-new-node
			 (car (nexus-abs-children
			       (nexus-parent first-new-node))))))
	       ;; Force recomputing the children of this node, and set
	       ;; it up for being chosen. This code is ugly.
	       (when last-old-node
		 (setf (a-or-b-node-goals-left last-old-node) :not-computed)
		 (setf (a-or-b-node-applicable-ops-left last-old-node)
		       :not-computed)
		 (delete-node-link last-old-node *current-problem-space*)
		 (push-previous last-old-node *current-problem-space*))
	       (setf *choose-this-node* last-old-node)
	       (unless (problem-space-property :forget-abstract-order)
		 (setf (problem-space-property :next-thing)
		       first-new-node)))))))



;;;**********************************************************************
;;; From the load-domain.lisp to define new syntax for operators.
;;;**********************************************************************

;; Changed last time 5/3/1995

;; Changed because I have added a new parameter to Check-syntax
(eval-when (load eval compile)
  (defmacro OPERATOR (&body body)
    (declare (special *current-problem-space*))
    "Does an expansion time syntax check and then reads the operator
into a data-structure."
    ;;Sira 5/3/95 the new parameter is 'operator
    (check-syntax body *current-problem-space* 'operator)
    `(push (create-operator ',body)
	   (problem-space-operators *current-problem-space*)))


  ;; Changed for the same reason as above

  (defmacro INFERENCE-RULE (&body body)
    (declare (special *current-problem-space*))
    "Does an expansion time syntax check and then reads the inference-rule
into a data-structure."
    ;; This time, the new parameter is 'rule (from inference-rule) 
    (check-syntax body *current-problem-space* 'rule)
    `(let ((a ',(create-operator body)))
       (setf (getf (operator-plist a) :inference-rule) t)
       (if (eq (getf (operator-plist a) :mode) :eager)
	   (push a
		 (problem-space-eager-inference-rules
		  *current-problem-space*))
	   (push a
		 (problem-space-lazy-inference-rules
		  *current-problem-space*)))))

  ;;This macro is needed to obtain the costs from the operator
  (defmacro rule-costs* (rule) `(assoc 'user::costs (cdr ,rule)))

  )

;; Changed for the same reason as above
(defun foperator (&rest op)
  (declare (special *current-problem-space*))
  "Checks the operator syntax and then reads the operator into a data
structure."
  ;; Sira 5/3/95 the new parameter is again 'operator
  (check-syntax op *current-problem-space* 'operator)
  (let ((real-op (create-operator op)))
    (setf (problem-space-operators *current-problem-space*)
	  (delete (operator-name real-op)
		  (problem-space-operators *current-problem-space*)
		  :key #'operator-name))
    (push real-op  (problem-space-operators *current-problem-space*))
    real-op))

(defun check-syntax (*checked-op* p-space type)
  (declare (type problem-space p-space)
	   (special *checked-op*))
  (setf p4::*vars-types* (make-hash-table))

  (let ((*arity-hash* (problem-space-arity-hash p-space))
	(*defined-vars* (check-variable-generators
			 (second (rule-preconds* *checked-op*))))
	(*all-preds* nil)         ; all the predicates in conditional of rule.
	(*non-static-preds* nil)
	(*effect-preds* nil)            ; collects preds in effects
	(name (rule-name* *checked-op*)))
    (declare (special *arity-hash* *all-preds* *non-static-preds*
		      *effect-preds* *defined-vars*))
    (check-name name)
    (check-preconds (rule-preconds* *checked-op*) name)
    (check-effects (rule-effects* *checked-op*) name)
    ;;Sira. Needed to check the new part of the operator
    (if (eq type 'operator)
	(check-costs (rule-costs* *checked-op*) name))
    (check-params (rule-params* *checked-op*) name)
    (check-mode (rule-mode* *checked-op*) name)

    ;; I don't know why I simply didn't use set difference here.
    (setf (problem-space-all-preds p-space)
	  (union
	   (union (problem-space-all-preds p-space)
		   *all-preds*)
	   *effect-preds*))))

(defun check-costs (*checked-part* name)
  (declare (special *defined-vars* *checked-part*))
  ;;Sira. 13/3/95. An operator may not have costs part 
  ;;Done in order not to re-write all the domains
  ;;written for prodigy 4.0
  (when *checked-part*
    (cond  ((< (length *checked-part*) 3)
	    (error "Costs list ~S is too short in ~S.~%" *checked-part* name))
	   ((> (length *checked-part*) 3)
	    (error "Costs list ~S is too long in ~S.~%" *checked-part* name)))
    (let ((*defined-vars*
	   ;; (append *defined-vars*
	   (check-costs-variable-generators (second *checked-part*)
					    *defined-vars*)));;)
      (declare (special *defined-vars*))
      (dolist (individual-cost (third *checked-part*))
	(check-individual-cost (cadr individual-cost))))))


;;; Daniel. 2004. Redefined it to account for the types allowed for each predicate
;; a var gen pair is a list of the form (<a> object) or
;; (<a> (and object (diff <a> <b>))) where <b> must already have been defined.

(defun check-var-gen-pair (pair defined-vars)
  (cond ((/= (length pair) 2)
	 (error "Error in length of ~S." pair))
	((not (strong-is-var-p (car pair)))
	 (error "~S not a variable in ~S." (car pair) pair)))
  
  (check-generator (cadr pair) (cons (car pair) defined-vars))
  (set-type-of-var (car pair) (cadr pair))
  (list (car pair)))

(defun set-type-of-var (var type-info)
  (declare (special p4::*vars-types*))
  (setf (gethash var p4::*vars-types*)
    (if (symbolp type-info)
	type-info
      (car type-info))))

;; Daniel. 2004. Idem as before.
(defun check-effect-var-gen-pair (pair defined-vars)
  (cond ((/= (length pair) 2)
	 (error "Error in length of ~S." pair))
	((not (strong-is-var-p (car pair)))
	 (error "~S not a variable in ~S." (car pair) pair)))
  
  (check-effect-generator (cadr pair) (cons (car pair) defined-vars))
  (set-type-of-var (car pair) (cadr pair))
  (list (car pair)))


;; This function has been copied from the original check-variable-generators
(defun check-costs-variable-generators (var-gen-list &optional other-vars)
  "Checks all the generators to be sure they are syntactically correct
and that variables are not used before they are defined."
  (let ((defined-vars other-vars))
    (dolist (var-gen-pair var-gen-list defined-vars)
      (setf defined-vars
	    (nconc defined-vars
		   (check-costs-var-gen-pair var-gen-pair defined-vars))))))

;; This one is the same as check-var-gen-pair
(defun check-costs-var-gen-pair (pair defined-vars)
  (cond ((/= (length pair) 2)
	 (error "Error in length of ~S." pair))
	((not (strong-is-var-p (car pair)))
	 (error "~S not a variable in ~S." (car pair) pair)))
  
  (check-costs-generator (cadr pair) (cons (car pair) defined-vars))
  (list (car pair)))

;; And this one is the same as check-effect-generator
(defun check-costs-generator (generator defined-vars)
  (declare (special *current-problem-space*))
  (cond ((symbolp generator);; should realy check to see if sym
	 (unless (is-type-p generator *current-problem-space*)
	   (error "~S is not a valid type.~%" generator))
	 (if (is-infinite-type-p generator *current-problem-space*)
	     (error
	      "~S is an infinite type, need a generator for this variable. " 
	      generator)))
	((member (car generator) '(user::and user::or user::not))
	 (check-complex-generator generator defined-vars))
	((member (car generator) '(lisp:funcall lisp:apply))
	 (check-generator-vars (cddr generator) defined-vars))
	;; This change added by Jim to fix an incompatibility between
	;; cmu lisp and the new definition of common lisp. Also note
	;; the swapped order of this clause and the previous one.
	((or (symbolp (car generator))
	     (and (listp (car generator))
		  (eq (caar generator) 'lambda))
	     (functionp (car generator)))
	 (check-generator-vars (cdr generator) defined-vars))
	(t
	 (error "~S is not a valid type, AND, OR, or NOT.~%"
		(car generator)))))

(defun check-individual-cost (element)
  (cond ((numberp element) t)
	((and (strong-is-var-p element) (member element *defined-vars*)) t)
	((symbolp element) t)
	((and (listp element) (function-or-operator (car element)))
	 (let ((result t))
	   (dolist (aux (cdr element) result)
	     (setf result (and result
			       (check-individual-cost aux)))))) 
	(t (error "Not a good expression ~a" element)))) 

(defun function-or-operator (token)
  (if (or (symbolp token)
	  (eq '+ token)
	  (eq '- token)
	  (eq '* token)
	  (eq '/ token))
      t
      (error "Not name of function or operator ~a" token)))


;; This section generates a data structure for each operator.
;; Sira.1-3-95 
;; I have modified this function for the new operator
(defun create-operator (op)
  (let* ((name (first op))
	 (params (assoc 'user::params (cdr op)))
	 (preconds (assoc 'user::preconds (cdr op)))
	 (effects (assoc 'user::effects (cdr op)))
	 ;;Sira.First change
	 (costs (assoc 'user::costs (cdr op))) 
	 (all-vars  (nconc
		     (mapcar #'car (second preconds))
		     (mapcar #'car (second effects))
		     (mapcan #'(lambda (eff)
				 (if (eq (car eff) 'user::forall)
				     (mapcar #'car (second eff))))
			     (third effects))))
	 ;; Sira 28/6/96.
	 ;; The vars which appear on the costs part do not have to be stored
	 ;; with the rest of the vars of the operator (Daniel wants
	 ;; it to be like that). So, they must be stored in the plist.
	 (costs-vars (mapcar #'car (second costs)))
	 ;; only there for inference rules.
	 (mode (or (second (assoc 'user::mode (cdr op))) :lazy)))
    ;; Sira 28/6/96 
    ;; The variables which appear on the costs part, can also appear
    ;; in the params list
    (if (not (subsetp (cdr params) (append all-vars costs-vars)))
	(error "~%params list has variables that are not in the~
 operator: ~S" name))
    (make-operator
     :name name
     :params params
     :precond-exp preconds
     :precond-vars (mapcar #'car (second preconds))
     ;; I assume this order in apply.lisp - don't change!
     :vars all-vars
     :binding-lambda nil       ; (make-binding-generator (car preconds))
     :add-list (make-effect-list (third effects) :add nil)
     :del-list (make-effect-list (third effects) :del nil)
     :effects effects
     :effect-var-map (match-effects-with-vars effects)
     :plist (append (list :costs-vars costs-vars)
		    (list :costs
			  (list (second costs) (third costs)))
		    (list :mode
			  (intern (symbol-name mode) :keyword))))))

;;;**********************************************************************
;;; From the load-problem.lisp
;;;**********************************************************************

(eval-when (load eval compile)
  ;; Sira. 1-3-95
  (defmacro vgoal (binding-generator preconds)
    `(foperator '*FINISH*
		'(user::params ,.(mapcar #'car binding-generator))
		'(user::preconds ,binding-generator ,preconds)
		;; Sira.Here is the change. I have to do this because *finish*
		;; operator must have its associated cost. Of course it will
		;; be 0.
		'(user::costs () ((user::*cost-label* 0)));; No costs need be defined.
		'(user::effects ();; No variables need be defined.
				((user::add (done))))))
  )

;;;**********************************************************************
;;; From the misc.lisp to get rid of the message Properties of n symbols...
;;;**********************************************************************

(defun remove-problem-space (p-space)
  (declare (type problem-space p-space)
	   (special *package*))
  "This will remove all references to problem spaces not in the list."

  (let ((object-indicator (problem-space-object p-space))
	(type-indicator  (problem-space-type p-space))
	(count 0))
    (declare (fixnum count))

    (do-symbols (sym *package*)
      (when (eq (symbol-package sym) *package*)
	(incf count)
	(remprop object-indicator sym)
	(remprop type-indicator sym)))
    (if user::*printp*
	(format t "~&Properties of ~D symbol~:P removed.~%" count))))

;;;**********************************************************************
;;; From the matcher
;;;**********************************************************************

;;;
;;; This is something Mei changed. I don't know in what moment she's
;;; going to put it permanetly on version4.0, so I keep it here
;;;
(defun process-one-meta-pred (pred vars)
    (declare (special *bucket-simple* *bucket-unary* *bucket-join*))
    (unless (eq (car pred) 'user::current-ops)
      (let* ((all-vars (all-vars pred))
	   ;;; all-vars is all the variables appear in pred. 

	     (len (length all-vars))
	     (num (number-of-common-vars all-vars vars))
	     (var-number
	      (cond 
		;; Mei 6/29/92
		;; when all the variables in pred are in vars, 
		;; then var number should be equal to len
		((= num len) len)
		;; else, when there is only one variable in vars, 
		;; (i.e. the variables in pred do not
		;; appear in vars), the var-number should be set to 1
		;; so this test is put into unary test. (since in this
		;; case, there will be no join tests)
		((= (length vars) 1) 1)
		(t 2)))

	   ;;; the number of elements of the intersection of all-vars
	   ;;; and vars, i.e. the number of op-variables of my defination.
	     
	     (place (if (= num len)
			(unless (zerop var-number)
			  (max-position all-vars vars))
		        (1- (length vars))))
	   ;;; place is a number showing at which place the pred
	   ;;; should be put in.  
	     
	   )
;	(format t "~% len ~S num ~S var-num ~S" len num var-number)
;	(break "in process-one-meta-pred")
	(case var-number
	  (0 (add-to-bucket-simple pred))
	  (1 (add-to-bucket-unary pred place))
	  (t (add-to-bucket-join pred place))))))


;;;**********************************************************************
;;; From the matcher also.
;;;**********************************************************************

;;
;; Daniel. Modified to handle the different-vars-p,
;; some-candidate-goals, and type-of-object meta-predicates, since
;; they need to know the bindings (that are really binding)
;;
(defun gen-values (expr bindings &optional type-declarations)
  (declare (special *current-problem-space*))
  (cond ((eq (car expr) 'user::type-of-object)
	 (user::type-of-object (sublis bindings (cadr expr))
			       (caddr expr) bindings))
	((eq (car expr) 'user::different-vars-p)
	 ;; it is being done now on the some-candidate-goals
	 ;; (not (user::equal-vars-values-p bindings))
	 t)
	((eq (car expr) 'user::some-candidate-goals)
	 ;; so that the select bindings control rules that have a
	 ;; separate matching would not select a binding with equal
	 ;; value for two different variables
	 (if (null (cdr expr))
	     (not (user::equal-vars-values-p bindings))
	     (if (not (user::equal-vars-values-p bindings))
		 (user::some-candidate-goals (sublis bindings (cadr expr))
					     bindings))))
	((member (car expr) (problem-space-all-preds *current-problem-space*))
	 (funcall #'user::known
		  (sublis bindings expr) type-declarations))
	(t  ; Rujith - subst. internal fn  (substitute-internal (car expr))
	 (apply (car expr)
		(sublis bindings (cdr expr))))))

#|
I deleted this because it is not needed for my code
((member (car expr) (problem-space-all-preds *current-problem-space*))
	 (funcall #'user::known
		  (sublis bindings expr) type-declarations))
|#	


(in-package "USER")

;;;
;;; This is a new function to test whether the vars have the same
;;; values
;;;
(DEFUN equal-vars-values-p (binding)
;;  (format t "~% equal-vars-values-p binding: ~a" binding)
  (do* ((bind-pairs binding (cdr bind-pairs))
	(equal-vars-p (and (not (numberp (cdar bind-pairs)))
			   (member (cdar bind-pairs) (cdr bind-pairs) :key #'cdr
				   :test #'eq))
		      (and (not (numberp (cdar bind-pairs)))
			   (member (cdar bind-pairs) (cdr bind-pairs) :key #'cdr
				   :test #'eq))))
      ((or (null bind-pairs)
	   equal-vars-p)
;;       (format t "~%    result: ~a" equal-vars-p)
       equal-vars-p)))


;;; ****************************************************************
;;;  From meta-predicates.lisp
;;; ****************************************************************
;;;
;;;  Redefinition of the real ones to be simpler and more effective. They
;;;  are a simplification of the real ones, since we don't use variables
;;;  nor lists of operators.

;;; Returns t if op is the current operator, can be used to generate.
;;; This used to be called current-op, but the name has been changed to
;;; keep it distinct from current-ops.
(defun current-operator (op)
  "Tests whether <op> is the current <op> for the current goal  at the
current node.  Can be used as a generator."
  (declare (special *current-node* *current-problem-space*))
  (eq (p4::operator-name (p4::operator-node-operator
			  (give-me-op-node *current-node*)))
      op))

;;; If not computed, I compute it
;;; Match op to a candidate operator for this node. Can be used as a
;;; generator. 
(defun candidate-operator (op)
  "Tests whether <op> is a member of the relevant ops being considered
at the current node.  Can be used as a generator."
  (declare (special *current-node* *current-problem-space*))
  (let* ((goal-node (give-me-goal-node *current-node*))
	 (candidates (p4::goal-node-ops-left goal-node)))
    (if (eq candidates :not-computed)
	(setq candidates (or (getf (p4::nexus-plist goal-node) 'candidate-operators)
			     (setf (getf (p4::nexus-plist goal-node) 'candidate-operators)
			       (p4::get-all-ops (p4::goal-node-goal *current-node*)
						*current-node*)))))
    (cond ((p4::strong-is-var-p op)
	   (mapcar #'(lambda (cand) (list (cons op cand))) candidates))
	  ((typep op 'p4::operator)
	   (if (member op candidates) t))
	  (t
	   (if (member (p4::rule-name-to-rule op *current-problem-space*)
		       candidates)
	       t)))))

;;; Match op to a candidate operator for this node. Can be used as a
;;; generator. 
;; (defun candidate-operator (op)
;;   "Tests whether <op> is a member of the relevant ops being considered
;; at the current node.  Can be used as a generator."
;;   (declare (special *current-node* *candidate-operators*
;; 		    *current-problem-space*))
;;   (and (member (p4::rule-name-to-rule op *current-problem-space*)
;; 	       (if (boundp '*candidate-operators*)
;; 		   *candidate-operators*
;; 		   (p4::goal-node-ops-left
;; 		    (give-me-goal-node *current-node*))))
;; 	 t))

;;
;; I need to redefine it since I need to check whether the vars are different
;;
;;; this should return list-of-binding-list; bindings is a binding-list.
;;;  Returns all the matchings of exp.

;;; check if obj is of type type.  accept both real prodigy obj/type
;;; and symbols.

;;;
;;; Daniel. If the obj is a var, it picks up an object of the same type
;;; that is not bound to any other variable. This is incomplete, since
;;; it will not generate bindings for all the possible instances, but
;;; it is much more efficient. Also, if there is no object that can be
;;; bound to that variable, there is no problem. Simply return t. This
;;; is the case in which it went through a some-candidate-goal and the
;;; variables of the rest of the goals do not have a value. This is
;;; ok. for control rules that are operator and decide, since they do
;;; not have variables in the rhs. With the rest, we have to make sure
;;; that the variable on obj does not belong to the variables in the
;;; rhs. If so, it returns nil.
;;;
;;; I will have to do something with the case in which there are free infinite vars.
;; 		 (if (string= "<INFINITE" (substring (format nil "~a" obj) 0 9))

(defun type-of-object (obj type binding)
  (declare (special *current-problem-space*))
  (cond ((p4::strong-is-var-p obj)
	 (if (eq *current-control-rule* :no-vars)
	     t
	     (let ((effects (cdr (p4::control-rule-then
				  *current-control-rule*))))
	       (if (eq (car effects) 'decide)
		   t
		   (case (cadr effects)
		     ((operators operator) t)
		     ((goal goals)
		      (if (member obj (cdaddr effects))
			  (get-unused-var obj type binding)
			  t))
		     (t			; bindings control rule
		      (if (member obj (caddr effects) :key #'cdr :test #'eq)
			  (get-unused-var obj type binding)
			  t)))))))
	((numberp obj) (or (eq type 'top-type)
			   (eq type :top-type)))
	(t (p4::object-type-p
	    (p4::make-real-object obj *current-problem-space*)
	    (p4::type-name-to-type type *current-problem-space*)))))

;;
;; Returns the first unused variable. This is not right, but we'll go
;; for it, until we find a negative example. The problem is that when
;; it uses the variable in the rhs of the control rule, it should get
;; all instantiations. Otherwise, we might have problems with
;; refinement.
;;
(DEFUN get-unused-var (obj type binding)
  (some #'(lambda (instance)
	    (if (not (member instance binding :key #'cdr :test #'eq))
		(list (list (cons obj instance)))))
	(all-the-instances type)))


(defun cost-from-pred (literal)
  (if (> (count-if #'p4::strong-is-var-p
		   literal)
	 1)
      (error "~% Too many unbound vars in the predicate generator ~S" literal)
      (let ((res (true-in-state literal)))
	(if (or (> (length res) 1)
		(< (length res) 1))
	    (error "~% Variable with more/less than one possible value ~S"
		   literal)
	    (or (eq res t)
		(mapcar #'cdar res))))))

#|
;;;
;;; Daniel. If the obj is a var, generates the bindings that have
;;; different values for the variables in the binding list
;;;
(defun user::type-of-object (obj type binding)
  (declare (special *current-problem-space*))
  (if (strong-is-var-p obj)
      (let ((new-binding nil))
	(dolist (instance (user::all-the-instances type)
			  (nreverse new-binding))
	  (if (not (member instance binding :key #'cdr :test #'eq))
	      (push (list (cons obj instance)) new-binding))))
    (let ((real-ob (make-real-object obj *current-problem-space*))
	  (real-type (type-name-to-type type *current-problem-space*)))
      (object-type-p real-ob real-type))))

;;; older version
      (let ((real-ob (make-real-object obj *current-problem-space*))
	    (real-type (if (typep type 'p4type)
			   type
			   (type-name-to-type type *current-problem-space*))))
	(if (null real-type)
	    (error "~% ~S is not a prodigy type." type)
	    (object-type-p real-ob real-type)))))
|#

