;;; ----------------------------------------------------------------------------------
;;;   				DATA DEFINITION  
;;; 
;;; Esta archivo contiene informacion de las estrucuturas de datos de la red Temporal
;;;
;;;
;;;			 LAST MODIFICATION: August 05, 2002
;;;
;;; ----------------------------------------------------------------------------------



(in-package "COMMON-LISP-USER")  ; El paquete por defecto 

;;; GLOBAL VARIABLES

(defvar *sum-of-edge* 0.0F+00)
(defvar *sum-of-nodes* 0.0F+00)

(defvar *horizon*  1000000)     ;Cuidado porque a veces infinito es cinco millones por problemas de iniciacion de la red. Habria que cambiar la funcion propagate.
(defvar *infinity* 5000000)

(defvar *fail* NIL)

(defvar *source-check* NIL)  ;;; NIL = no source checking  T = source checking
(defvar *switch-cycle* NIL)  ;;; NIL = no cycle-check      T = cycle-check

(defvar *prop-type* NIL)     ;;; NIL = queue               T = stack


;;; ----------------------------------------------------------------------------------
;;;
;;;   La estructura time_map representa la rete temporal
;;;
;;;
;;; ----------------------------------------------------------------------------------
 



(defstruct (time_map
           (:print-function
           (lambda (p s k)
           (format s "<tm-~D>" (time_map-name p) k))))
    
   (name NIL)                ;;; Name of the time-map

   (list_dcs NIL)            ;;; Associative list of elements <dc_id, dc_pointer>
   (list_tps NIL)            ;;; Associative list of elements <tp_id, tp_pointer>
   (stack_ndc NIL)           ;;; List working as a stack used to saved the state of the time map
                             ;;; Lista gestita a pila usata per salvare lo stato del time map.
   (stack_ntp NIL)           ;;; Idem.

   (ntp 0)		     ;;; Time point index
   (ndc 0)	             ;;; Distance constraint index

   (dc_sources ())           ;;; List of new and modified distance constraints
   (ref_point NIL)           ;;; Origin time point
   (horizon NIL)             ;;; Horizon time point
   (horizon-constraint NIL)  ;;; Constraint between ref_point and horizon
                             ;;; time point
   (ncycle-list-ub NIL)      ;;; Negative cycle detected by propagation alg.
   (ncycle-list-lb NIL)      ;;; Negative cycle detected by propagation alg.
   (delta 0)                 ;;; Length of the detected negative cycle.

)

;;; ----------------------------------------------------------------------------------
;;;
;;;   La estrucutura time_point representa los puntos en la red temporal
;;;
;;;
;;; ----------------------------------------------------------------------------------



(defstruct  (time_point

	    (:print-function
            (lambda (p s k)
              (format s "<tp-~D>" (time_point-name p) k))))

   (name NIL)          ;;; Time point's name
   (in-queue NIL)      ;;; Boolean mark which indicates the time points is
                       ;;; in the propagation queue.
   (in NIL)            ;;; Set of edges which enters in the time point.
   (out NIL)           ;;; Set of edges which leave the time point.

   (lb (- *infinity*)) ;;; lower bound of the interval associated to the time point
   (ub *infinity* )    ;;; upper bound of the interval associated to the time point
 
   (st_edge_lb NIL)    ;;; pointer to the lower bound tree
   (st_edge_ub NIL)    ;;; puntatore albero ricoprente per upper bounds

   (stack-lb NIL)      ;;; List working as a stack used to save the state of the time map
   (stack-ub NIL)      ;;; idem
   (stack_ste_lb NIL)  ;;; idem
   (stack_ste_ub NIL)  ;;; idem

   (pl NIL)	       ;;; precedence function lb 
   (pu NIL)            ;;; precedence function ub
   (stack-pl NIL)
   (stack-pu NIL)

   (lb-prop NIL)
   (ub-prop NIL) 
  )

;;; ----------------------------------------------------------------------------------
;;;
;;;   La estrucutura edge representa  los arcos que unen los time point.
;;;
;;;
;;; ----------------------------------------------------------------------------------


(defstruct  (edge

	    (:print-function
            (lambda (p s k)
              (format s "<edge-~D>" (edge-name p) k))))

   (name NIL)                ;;; Name of the edge.
   (from NIL)                ;;; Time-point origin of the edge.
   (to NIL)                  ;;; Time-point destination of the edge.
   (source NIL)              ;;; Mark which indicates the edge is a source of
                             ;;; propagation
   (in_queue NIL)            ;;; Mark which indicates the edge is in the propagation queue
   (lb-prop NIL)             ;;; 
   (ub-prop NIL)             ;;;
   (lb 0)                    ;;; Lower bound of the interval associated to the arc
   (ub *horizon*)            ;;; Idem for the upper bound.
   (stack-lb NIL)	     ;;; List working as a stack used to save the state of the time map
   (stack-ub NIL)            ;;; idem.
   (stack_molt NIL)          ;;; idem.
   (new_lb_bs NIL)           ;;; Lower bound list relative to the modifications introduced on the distance contraints
			     ;;; Lista dei lower bounds relativi alle modifiche introdotte sul distance constraint.
   (new_ub_bs NIL)           ;;; idem.

 )