;; This file creates a footprint of a solution. The solution path
;; should be a list of nodes of form:
;;       (p4::path-from-root (cdr (prodigy-result-interrupt (run))))
;; It will create a list of literals, each of which was in the initial
;; state, and relevant to finding the solution
;
;; This file depends on 
;;    print-rules.lisp (format-literals-to-list)
;;    preconds.lisp    (set-valid-preconds)

(in-package "USER")

;; Daniel. I don't know what this is for, but..
(defvar *short-circuit-preconditions*)

; make a list of all the applied op nodes in the solution path
(defun extract-applied-op-nodes (solution-path)
  (let ((app-op-nodes nil ))
    (dolist (node solution-path app-op-nodes)
      (if (p4::applied-op-node-p node)
	  (push node app-op-nodes)))))


; make list of instantiated effects for an applied-node-op-application
(defun get-instantiated-effects-applied (conditional-preconditions applied)
  (if conditional-preconditions
      (let ((dels nil) (adds nil))
	(dolist (pair (car conditional-preconditions))
	  (push (cdr pair) dels))
	(dolist (pair (second conditional-preconditions))
	  (push (cdr pair) adds))
	(list dels adds))
      (let* ((inst-op (p4::op-application-instantiated-op applied))
	     (op (p4::instantiated-op-op inst-op))
	     (vars (p4::rule-vars op))
	     (values (p4::instantiated-op-values inst-op))
	     (precond-bindings (p4::precond-bindings vars values)))
	(p4::compute-effects op precond-bindings applied inst-op))))

;; Takes a solution path (just any list of nodes), extracts all the
;; applied-op-nodes that are in that list, goes backward through
;; them, adding the preconditions of each operator. Note that
;; each applied-op-node may have several (op-application)s.
;;

(defun footprint (solution-path
                  &optional (goals (cdr (p4::problem-goal (current-problem))))
                            (last-node (car (last solution-path))))
;; 21/2/94 goals & last-node optional allows footprint of partial solutions
;; For Daniel
  (let ((app-op-nodes (extract-applied-op-nodes solution-path)))
    (if (eq (caar goals) 'and)
        (setq goals (cdr (car goals))))

    (dolist (app-op-node app-op-nodes)
      (set-valid-preconds last-node app-op-node)
      
      ; must look at all applied operators in order to cover inference rules
      (dolist (applied (p4::applied-op-node-applied app-op-node))
        (let* ((instop (p4::op-application-instantiated-op applied))
               (preconds-list   (format-literals-to-list
                                 (getf (p4::instantiated-op-plist instop)
                                       :satisfying-preconds)))
               (conditional-preconds (getf (p4::instantiated-op-plist instop)
                                           :conditional-preconditions))
	       (effects  (get-instantiated-effects-applied conditional-preconds applied))
               (del-list  (add-neg-signs (mapcar #'format-literals-to-list
                                                 (car effects))))
               (add-list (mapcar #'format-literals-to-list (second effects)))
               (newgoals     nil)
               (search-list nil)
               (relevant-cond-pre nil))
	  
          (dolist (goal goals)
            (if (eq '~ (car goal))
                (progn
                  (setq search-list del-list)
                  (setq relevant-cond-pre (car conditional-preconds)))
                (progn
                  (setq search-list add-list)
                  (setq relevant-cond-pre (second conditional-preconds))))

            (if (find goal search-list :test #'equalp)
                ; if goal is found, add preconditions to new goallist
                ;   must also add conditional preconditions if necessary
                ; else put this goal back into the new goallist
                (setq newgoals (nestedunion newgoals
                                      (nestedunion preconds-list
                                              (add-conditional-preconds
                                                    relevant-cond-pre
                                                    goal))))
                ;else
                (setq newgoals (nestedunion (list goal) newgoals)))
          )
          (setq goals newgoals)
        )))
    goals))

(defun add-conditional-preconds (conditional-preconds goal)
  (let ((newgoals nil))
    (if (and (boundp 'user::*short-circuit-preconditions*)
             user::*short-circuit-preconditions*)
        (do* ((conds conditional-preconds (cdr conds))
	      (pair       (car conds) (car conds))
              (test-match (equalp (format-literals-to-list (cdr pair)) goal)
                          (equalp (format-literals-to-list (cdr pair)) goal))
              (exit-loop  nil))

           (exit-loop ; exit condition
            nil)  ; return value
          (when test-match
	    (setq exit-loop t)
	    (setq newgoals
		  (nestedunion (format-literals-to-list (car pair))
			       newgoals))))
        (dolist (pair conditional-preconds)
          (if (equalp (format-literals-to-list (cdr pair)) goal)
              (setq newgoals (nestedunion (format-literals-to-list (car pair))
					  newgoals)))))
    newgoals))

(defun add-neg-signs (l)
   (mapcar #'(lambda (x) (list '~ x)) l))

(defun listify (preconds)
; change format of preconditions from #<> to lists
; only works in simplest of conjuctions.
  (let ((preconds-list nil))
    (dolist (pre (case (type-of preconds)
                          (cons (cdr preconds))
                          (t (list preconds)))   )
         (case (type-of pre)
           (p4::literal (push (format-literals-to-list pre) preconds-list))
           (cons (push (list '~ (format-literals-to-list (car (cdr pre))))
                                   preconds-list))))
    preconds-list))

(defun nestedunion (l1 l2)
  ; union of nested things
  (union l1 l2 :test #'equalp))

;;(load "/afs/cs/project/prodigy/version4.0/working/system/contrib/new-applyop.lisp")

