;;; This is the new version of Minnie for Prodigy 4.0. It is called
;;; HAMLET, for Heuristics Acquisition Method by Learning from sEarch Trees.
;;; There is always a doubt in our lifes in which we ask ourselfs:
;;;   is it right or wrong to do it =
;;;   to be or not to be

;;; Falta por hacer: meter el metapredicado some-candidate-applicable-op
;;; en las reglas de decide. Queda el problema de que no se puede controlar
;;; en PRODIGY que operador, de todos los aplicables, se debe ejecutar.


(in-package "USER")

;;******************************************************************
;;;
;;; Funciones
;;;
;;******************************************************************
  

;;
;; Needed things for starting up
;;

;;
;; To be called after loading the domain of before calling it with a
;; domain name
;;
(defun reset-domain-variables (initp &optional (domain-name
						(p4::problem-space-name
						 *current-problem-space*)))
  (if initp (setf *decision-tree* nil))
  (unless (eq *domain-name* domain-name)
    (setf *domain-name* domain-name)
    (setf *domain-directory* (format nil "~a~(~a~)/" *world-path* domain-name))
    (setf *examples-file* (format nil "~aexample-file.d" *domain-directory*))
    (setf *utility-output* (format nil "~autility.~a"
				   *domain-directory* *lisp-extension*))
    (setf *ff-output-directory* (format nil "/home/dborrajo/tmp-~(~a~)/" *domain-name*))
    (setf *hamlet-results-output* (format nil "~ahamlet-results.~a"
					  *domain-directory* *lisp-extension*))
    (setf *prodigy-results-output* (format nil "~aprodigy-results.~a"
					   *domain-directory* *lisp-extension*))
    (setf *hamlet-output* (format nil "~ahamlet-output.~a"
				  *domain-directory* *lisp-extension*))))


;;;
;;; To store the last node of the search
;;;
(DEFUN store-last-node (signal)
  (declare (ignore signal)
	   (special user::*current-node*))
  (setf user::*last-node* user::*current-node*))


;;******************************************************************
;; If *learnp*=t, it learns new control rules
;; If *learnp*=nil, it only labels the search tree
;;
;; New dimensions:
;; Learning points
;;  - Lazy: does not learn from default decisions
;;  - Eager: learns also from default decisions
;; Exhaustion of the search space
;;  - Total: it requires the expansion of the whole search space
;;  - Partial: it expands the search space up to the time limit
;; Then, we have four modes:
;;  - Lazy-total: not default and whole search space (very-lazy)
;;  - Eager-total: default and whole search space (lazy)
;;  - Lazy-partial: not default and partial search space (eager)
;;  - Eager-partial: default and partial search space (very-eager)
;;
;; eager-mode=eager it learns all possible success rules, but not when
;;                  the decision was made the first one
;; eager-mode=very-eager it learns all possible success rules
;; eager-mode=lazy it learns when all alternatives were
;;                 tried and failed even if they were the first ones
;; eager-mode=very-lazy it learns when all alternatives were
;;                      tried and failed, and they were not the first ones
;;
;; Old modes:
;; eager-mode=t it will learn all possible success rules, but not when the
;;          decision was made the first one. Now: eager
;; eager-mode=very it will learn all possible success rules. Now: very-eager
;; eager-mode=nil it will learn only when all other alternatives were
;;            tried and failed and were not the first ones. Now: very-lazy
;;
(defun aprender (eager-mode previous-cost-achieved print-data-p)
  (init-learning)
  (let* ((t-inicial (get-internal-run-time))
	 (whole-tree-expanded-p (whole-tree-expanded-p))
	 (result (learn-from-tree *real-t-tree* eager-mode
				  whole-tree-expanded-p)))
    (setf (node-num-nodes *real-t-tree*) (caddr result))
    (setf *number-ops* (car result))
    (setf *best-cost-achieved* (cadr result))
    (if (or *store-state-p*
	    (and *learnp* *optimal-learning-p* whole-tree-expanded-p
		 (or (not previous-cost-achieved)
		     (<= *best-cost-achieved* previous-cost-achieved))
		 (solutionp)))
	(learn-only-optimal-path *real-t-tree* t))
    (setf (prodigy-result-all-nodes) (caddr result))
    (if *printp* (pp-rtt))
    (when *learnp*
      (if *inductionp*
	  (learn-from-induction t *cr-file* *cr-list* nil nil t))
      (save-cr))
    (something-changed-?)
    (if print-data-p
	(print-learning-data (elapsed-time t-inicial)))))

;;
;; Inicializa las variables de aprendizaje
;;
(defun init-learning nil
  (if *active-learning-mode*
      (setf *active-learning-hash* (make-hash-table :test #'equal)))
  (when *learnp*
    (setf *deleted-rule-p* nil)
    (setf *initial-cr* *rules*)
    (setf *cr-list* nil))
;; The real top level node is the 2nd.
  (setf *real-t-tree* (p4::find-node 2))
  (setf *my-current-node* nil)
;;  (setf *number-nodes* 0)
  (setf *number-deduced-rules* 0)
  (setf *number-induced-rules* 0)
  (setf *number-refined-rules* 0)
  (setf *number-deleted-rules* 0)
  (setf *number-undeleted-rules* 0))

(DEFUN something-changed-? nil
  (setf *something-changed-p*
	(or *something-changed-p*
	    (> (+ *number-refined-rules* *number-induced-rules*) 0)
	    (/= (+ *number-deduced-rules*
		   *number-induced-rules*
		   *number-refined-rules*
		   (- *number-deleted-rules*)
		   *number-undeleted-rules*)
		0))))


;;
;; Returns the list of all learned control rules frames
;;  
(DEFUN all-control-rules (&optional (rules-types *learned-rules-types*))
  (declare (list rules-types))
  (let ((result nil))
    (dolist (access-function rules-types result)
      (setq result (append (funcall access-function *current-problem-space*)
			   result)))))

;;******************************************************************
;;
;; Procedure that labels the search tree, assigns credit, finds the shortest
;; solution, finds the best paths, etc.
;; Returns a list with three elements: (num-ops cost num-nodes)
;;
(defun learn-from-tree (node eager-mode whole-tree-expanded-p)
  (declare (type p4::nexus node))
  (let* ((children (reverse (p4::nexus-children node)))
	 (first-child (car children))
	 (a-or-b-p (p4::a-or-b-node-p node))
	 (first-goal-child (and a-or-b-p (find-if #'p4::goal-node-p children)))
	 (wrong-default-behavior nil)
	 (learn-goal-rule-p nil)
	 (best-solution most-positive-fixnum)
	 (number-nodes 0)
	 (best-cost-achieved most-positive-fixnum)
         (shortest-num-nodes most-positive-fixnum)
	 successes failures unknowns untrieds best-childs)
    (cond (children
	   (dolist (child children)
	     (let* ((num-ops-cost-num-nodes
		     (learn-from-tree child eager-mode whole-tree-expanded-p))
		    (num-ops (car num-ops-cost-num-nodes))
		    (cost (cadr num-ops-cost-num-nodes)) 
		    (num-nodes (caddr num-ops-cost-num-nodes)))
	       (incf number-nodes num-nodes)
	       (setf (node-num-nodes child) num-nodes)
	       (case (node-learning-type child)
		 (success (if (> num-ops 0)
                              (if *use-quality-p*
                          	  (cond ((< cost best-cost-achieved)
				         (setq best-childs (list child))
				         (setq best-solution num-ops)
                                         (setq shortest-num-nodes num-nodes)
				         (setq best-cost-achieved cost))
				        ((= cost best-cost-achieved)
                                         (cond ((< num-nodes
						   shortest-num-nodes)
                                                (setq shortest-num-nodes
						      num-nodes)
                                                (setq best-childs
						      (list child)))
                                               ((= num-nodes
						   shortest-num-nodes)
						(push child best-childs))
                                               (t nil)))
				        (t nil))
                          	  (cond ((< num-ops  best-solution)
				         (setq best-childs (list child))
				         (setq best-solution num-ops)
                                         (setq best-cost-achieved num-ops))
				        ((= num-ops best-solution)
				         (push child best-childs))
				        (t nil))))
			  (push child successes))
		 (failure (push child failures))
		 (unknown (push child unknowns))
		 (t nil))))

	   (setq untrieds
		 (case (node-type node)
		   (operator-node (p4::operator-node-bindings-left node))
		   (goal-node
		    ;; forward link used for regressing the preconds later
		    (let ((binding-node
			   (car (p4::goal-node-introducing-operators node))))
		      (if binding-node
			  (push node (getf (p4::nexus-plist binding-node)
					   'goal-nodes))))
		    (p4::goal-node-ops-left node))
		   (t (let ((goals-left (p4::a-or-b-node-goals-left node)))
			(append (p4::a-or-b-node-applicable-ops-left node)
				(if (listp goals-left)
				    goals-left
				    (list goals-left)))))))
	   
	   (incf number-nodes (length untrieds))
	   
;;         (if (and more-than-1-child-p
;;                  (or failures unknowns untrieds))
;;             (setf *treep* t))
;; Esto lo borro ya que no debe ser asi
;; Esto es una chapuza como una catedral, pero el
;; dependency-backtracking deja sin estudiar determinados nodos. Por
;; lo tanto, si multiple-sols-p=t, entonces se consideran fallos
	   (when whole-tree-expanded-p
	     (setq failures (append failures unknowns untrieds))
	     (setq unknowns nil)
	     (setq untrieds nil))

;; si no hay una solucion desde donde esta, mejor solucion no habra variado,
;; por lo que hay que devolver 0. Si no, si es un nodo de ejecucion de 
;; un op, hay que contar un op mas. Si no, hay que devolver la mejor solucion
;; Sira. Ahora ya no es asi. Se devuelve el coste del mejor camino, que 
;; estara acumulado en el ultimo nodo del mismo.
;; Sira. En este caso lo que devuelvo es la longitud
;; total del camino por el que voy ascendiendo.
;; No tiene sentido ahora almacenar el coste desde
;; el punto actual hasta el final, puesto que ya 
;; dispongo del coste total del camino a seguir.
;; Lo que habia antes era...(1+ best-solution))
	   (when *use-quality-p*
	     (setq best-cost-achieved
		   (if (= best-solution most-positive-fixnum)
		       0
		       best-cost-achieved)))

	   (setq best-solution
		 (cond ((= best-solution most-positive-fixnum) 0)
		       ((p4::applied-op-node-p node) (1+ best-solution))
		       (t best-solution)))

           (unless *use-quality-p*
	     (setq best-cost-achieved best-solution))

	   (setq wrong-default-behavior
		 (not (member first-child best-childs :test #'eq)))

	   (setq learn-goal-rule-p
		 (or (eq eager-mode 'eager-partial)
		     (and (eq eager-mode 'eager-total)
			  (not (or unknowns untrieds)))
		     (and a-or-b-p
			  (not (member first-goal-child best-childs
				       :test #'eq)))))
	   
;; there was a test of (if best-child ...)         
	   (setf (node-best-child-info node)
		 (list best-childs
		       nil
;; Sira. Creo que esto es asi, pero no estoy segura.
                       (if *use-quality-p*
		           (number-steps best-childs best-cost-achieved)
	   		   (number-steps best-childs best-solution))
		       learn-goal-rule-p))
	   (cond 
;; *** 1. Hay nodos en el subarbol que no han sido intentados.
;;        Si eager-mode=eager-partial o el primer hijo no es el elegido, entonces aprende
	    ((or unknowns untrieds)
	     (if (and *learnp*
		      successes
		      (or (eq eager-mode 'eager-partial)
			  (and (eq eager-mode 'lazy-partial)
			       wrong-default-behavior)))
		 (if *optimal-learning-p*
		     (setf (cadr (node-best-child-info node)) t)
		     (learn-all-best-childs best-childs node t t
					    learn-goal-rule-p)))
	     ;; Sira. Esto no me acaba de convencer
	     (setf (caddr (node-best-child-info node)) 0)
	     (setf (node-learning-type node)
		   (if (and successes
			    (or (eq eager-mode 'eager-partial)
				(eq eager-mode 'lazy-partial)))
		       'success
		       'unknown)))
	    
;; Esto no sirve ya que puede ser que sean falsos tambien los que quedan,
;; por lo que no se debe aprender, debido al problema de los bucles de estado
;; y meta
;;           (if failures
;;               (dolist (malo failures)
;;                       (learn-decision node malo nil)))           
;; *** 2. Hay de los dos: failures y successes
	    ((and failures successes)
	     (setf (node-learning-type node) 'success)
;; Se van a aprender reglas de control de select y no
;; hace falta aprender de reject (las de select se ejecutan antes).
;; Se aprende la regla que logra la solucion mas corta
	     (if (and *learnp*
		      (or (eq eager-mode 'eager-partial)
			  (eq eager-mode 'eager-total)
			  wrong-default-behavior))
		 (if *optimal-learning-p*
		     (setf (cadr (node-best-child-info node)) t)
		     (learn-all-best-childs best-childs node t t
					    learn-goal-rule-p))))

;; *** 3. Todos son failures. El malo realmente es uno de los antepasados
	    (failures
	     (setf (node-learning-type node) 'failure))

;; *** 4. Todos son successes (uno o varios depende de la variable
;; *multiple-solutions-p*). El padre es bueno. Aprende el mejor si se
;; quiere aprender optimamente, hay mas de una posibilidad y no era la
;; primera
	    (t
;;           (setf (node-best-child-info node) (cons best-child nil))
	     (setf (node-learning-type node) 'success)
	     (if (and *learnp*
		      (or (eq eager-mode 'eager-partial)
			  (eq eager-mode 'eager-total)
			  (and (cdr successes)
			       wrong-default-behavior)))
		 (if *optimal-learning-p*
		     (setf (cadr (node-best-child-info node)) t)
		     (learn-all-best-childs best-childs node t t
					    learn-goal-rule-p)))))

;;         (if (eq (node-learning-type node) 'success)
;;             (incf *number-nodes-solution-path*))

;; The procedure returns the best solution (the shortest number of
;; applicable ops to finish and if tie, the one with less number of
;; nodes in the subtree).
	   (list best-solution
		 (if *use-quality-p*
		     best-cost-achieved
		     best-solution)
		 (1+ number-nodes)))
	  
;; If the node does not have children:
	  (t (setf (node-learning-type node)
		   (if (success-path-p node)
		       'success
		       'failure))
	     (if (and (eq (node-learning-type node) 'success)
		      (p4::applied-op-node-p node)
                      (or (not *use-quality-p*)
                          (numberp (getf (p4::nexus-plist node) :g-value))))
;; Sira. DEVUELVO EL COSTE TOTAL DE ESTE CAMINO.
;; Esto es lo que se ha de propagar a lo largo
;; de todo el recorrido hacia la raiz del arbol.
		     (list 1
			   (if *use-quality-p*
			       (getf (p4::nexus-plist node) :g-value)
			       1)
			   1)
		     (list 0 0 1))))))

(defun solution-length (applied-op-node)
  (declare (ignore applied-op-node))
  1)

;;;
;;; It is only called if the variable *optimal-learning-p* is t. It
;;; will go down the tree going from best-child to best-child until
;;; there was not a success, or there were not children (leaf of the tree).
;;; Returns a list of new-les and an a-list of goals and regressed
;;; preconds for each one.
;;;
(DEFUN learn-only-optimal-path (node compute-state-p)
  (declare (type p4::nexus node)
	   (atom compute-state-p))
  (let* ((best-child-info (node-best-child-info node))
	 (best-childs (car best-child-info)))
    (cond (best-childs
	   (if *store-state-p*
	       (setf (getf (p4::nexus-plist node) 'state)
		     (get-state-for-node node)))
	   (learn-all-best-childs best-childs node compute-state-p
				  (cadr best-child-info)		      
				  (cadddr best-child-info)))
	  ((and (p4::nexus-children node) *store-state-p*)
	   (setf  (getf (p4::nexus-plist node) 'state)
		  (get-state-for-node node))
	   nil)
	  (t nil))))

;;
;; It returns best-solution if at least one of the best childs has its subtree
;; all known.
;;
(DEFUN number-steps (best-childs best-solution)
  (declare (list best-childs)
	   (integer best-solution))
  (if (some #'(lambda (best-child)
		(let ((child-info (node-best-child-info best-child)))
		  (if child-info
		      (> (caddr child-info) 0)
		      (null (p4::nexus-children best-child)))))
	    best-childs)
      best-solution
      0))

;;
;; Learns control rules for each of the best childs, and returns the
;; new-les together with the substitutions made to create the rules.
;; If the node is a a-or-b node and all best-childs are all
;; the possible goals, then it does not learn the corresponding select
;; goals for each one. Also, it does not learn for the first one.
;; It does not learn a decide-sub-goal rule if the corresponding
;; applied node does not clobber the decision made in the sibling goal
;; node.
;; Now (26/8/96) it also learns negative examples from the failure children
;;
(DEFUN learn-all-best-childs (best-childs node compute-state-p learning-node-p
					  learn-goal-rule-p)
  (declare (type p4::nexus node)
	   (list best-childs))
  (let* ((learnp (and learning-node-p
		      (not (eq node (find-node 2)))
		      (not (eq node (find-node 3)))))
	 (state (if learnp
		    (if compute-state-p
			(get-state-for-node node)
			(getf (p4::nexus-plist node) 'state))))
	 (applied-child (if (and learnp
				 (p4::a-or-b-node-p node))
			    (find-if #'p4::applied-op-node-p
				     (p4::nexus-children node)
				     :from-end t)))
	 (les-substitutions nil)
	 (goal-regressed-state-p nil)
	 (goals-alist nil)
	 (new-les nil)
	 (bad-children nil))
    (when learnp
      (dolist (child (p4::nexus-children node))
	(if (not (member child best-childs))
	    (push child bad-children)))
      (if (and bad-children *inductionp*)
	  (repair-overgeneralization node node
				     (list (cons nil bad-children)) 0)))
    (dolist (best-child best-childs)
      (let ((regressed-state (get-regressed-state node best-child learnp))
	    (les-goals (learn-only-optimal-path best-child compute-state-p)))
	;; if there are two or more ways to achieve a certain
	;; goal, we pick up the first one
	(dolist (goal-state (cdr les-goals))
	  (pushnew goal-state goals-alist :test #'equal :key #'car))
	(dolist (le (car les-goals))
	  (push le new-les))
	;; if there are two or more ways to achieve a certain
	;; goal, we pick up the first one
	(when (and (p4::goal-node-p node)
		   (not goal-regressed-state-p))
	  (setq goal-regressed-state-p t)
	  (pushnew (cons (get-target-goal node) regressed-state)
		   goals-alist :test #'equal :key #'car))
	(if (and learnp
		 (or (not (p4::goal-node-p best-child))
		     (not applied-child)
		     (clobbersp applied-child regressed-state)))
	    (dolist (le-goals (learn-decision node best-child state
					      regressed-state t
					      learn-goal-rule-p))
	      (push (cons *substitution* le-goals) les-substitutions)))))
    (dolist (le-substitution les-substitutions (cons new-les goals-alist))
      (let ((le (cadr le-substitution)))
	(push le new-les)
	(if (and *inductionp* *use-preferred-preconds-p*)
	    (set-preferred-preconds le (cddr le-substitution) goals-alist
				    (car le-substitution)))))))

;;
;; Returns the intersection of the preconds-pool of a le with the
;; union of the regressed-states of the some-candidate-goals on that
;; decision. It is useful for giving these preconds preference over
;; the rest. They are also removed from the pool
;;
(DEFUN set-preferred-preconds (le some-goals goals-alist substitution)
  (declare (type list some-goals goals-alist substitution)
	   (type learning-episode le))
  (let ((regressed-state (mapcar #'(lambda (goal)
				     (list 'true-in-state goal))
				 (regressed-states-union some-goals
							 goals-alist
							 substitution)))
	(pool (learning-episode-preconds-pool le))
	(all-vars nil))
    (setf (learning-episode-preconds-pool le) nil)
    ;; given that true-in-state come before the type-of-object
    (dolist (precond pool)
      (case (car precond)
	(true-in-state
	 (cond ((member precond regressed-state :test #'equal)
		(dolist (var (vars-in-precond (cdadr precond)))
		  (pushnew var all-vars :test #'eq))
		(push precond (learning-episode-preferred-preconds le)))
	       (t (push precond (learning-episode-preconds-pool le)))))
	(type-of-object
	 (if (member (cadr precond) all-vars :test #'eq)
	     (push precond (learning-episode-preferred-preconds le))
	     (push precond (learning-episode-preconds-pool le))))
	(t ;; it should not enter here
	 nil)))
    ;; we have to restore the order
    (setf (learning-episode-preconds-pool le)
	  (nreverse (learning-episode-preconds-pool le)))
    (setf (learning-episode-preferred-preconds le)
	  (nreverse (learning-episode-preferred-preconds le)))))
    
	  

;;
;; Computes a recursive union (there is no union for more than two sets)
;;
(DEFUN regressed-states-union (some-goals goals-alist substitution)
  (declare (type list some-goals goals-alist substitution))
  (if some-goals
      (nunion (sublis substitution
		      (cdr (assoc (car some-goals) goals-alist
				  :test #'equal)))
	      (regressed-states-union (cdr some-goals) goals-alist
				      substitution)
	      :test #'equal)))
  


;;;
;;; Applied-child is the applied-child that did not work, and
;;; regressed-state is the regressed-state of a goal-node sibling
;;; The applied-child child clobbers a sibling node if it dels
;;; something the sibling needs.
;;;
(DEFUN clobbersp (applied-child regressed-state)
  (let ((postconditions (car (p4::applied-op-node-applied applied-child))))
    (or (some #'(lambda (literal)
		  (let ((del (get-visible-form literal)))
		    (some #'(lambda (state)
			      (equal del state))
			  regressed-state)))
	      (p4::op-application-delta-dels postconditions))
	(some #'(lambda (literal)
		  (let* ((add (get-visible-form literal))
			 (negated-add-p (negatedp add))
			 (pos-assertion (get-pos-assertion add)))
		    (some #'(lambda (state)
			      (if negated-add-p
				  (if (not (negatedp state))
				      (equal pos-assertion state))
				  (if (negatedp state)
				      (equal (get-pos-assertion state) add))))
			  regressed-state)))
	      (p4::op-application-delta-adds postconditions)))))


;;; ************************************************************
;;;    Functions for computing the regressed-state
;;; ************************************************************

;;
;; Returns the footprint of the preconditions that were needed to
;; achieve a plan for the node
;;
;; It might be useful to consider storing the union of the footprint
;; of the node and the footprint of all some-candidate-goals
;;
;; I do not know why I was printing the Possibly wrong message, so I
;; just comment it for now.
;;
(DEFUN get-regressed-state (node best-child learnp)
  (declare (ignore learnp)
	   (type p4::nexus node best-child))
  (let* ((node-type (node-type node))
	 (goal-child-p (p4::goal-node-p best-child)))
    (footprint (if goal-child-p
		   (get-best-path best-child)
		   (get-best-path node))
	       (list
		(case node-type
		  (operator-node (get-target-goal (p4::nexus-parent node)))
		  (goal-node (get-target-goal node))
		  (t                    ;;(binding-node applied-op-node)
		   (cond (goal-child-p (get-target-goal best-child))
			 ((eq node-type 'binding-node)
			  ;;(if learnp (format t "~%Possibly wrong~%"))
			  (get-target-goal (get-grandpa node)))
			 (t ;;(if learnp (format t "~%Possibly wrong~%"))
			    (get-target-goal
			     (get-grandpa
			      (p4::instantiated-op-binding-node-back-pointer
			       (p4::applied-op-node-instantiated-op
				node)))))))))
	       (or *my-current-node*
		   (get-last-node *last-node*)))))


;;
;; Returns the best path from the node to the end of the path,
;; following the best-child pointers. It only takes the first one of
;; the best childs.
;;
(DEFUN get-best-path (node)
  (declare (type p4::nexus node))
  (let ((best-child (caar (node-best-child-info node))))
    (cond (best-child
	   (cons node (get-best-path best-child)))
	  ((p4::nexus-children node) nil)
	  (t (list node)))))

  
;;; ************************************************************
;;;    Funciones de Juan Pedro (no se hasta que punto, ya :-)
;;; ************************************************************

;;
;; Creates new rule(s), and returns a list of the form:
;;  ((le . some-candidate-goals) ...)
;; candidate-operator=candidate-op
;; current-goal=target-goal
;;
(DEFUN learn-decision (parent node state regressed-state successp
			      learn-goal-rule-p)
  (declare (type p4::nexus parent node)
	   (list state))
  (setf *number-vars-in-cr* 0)
  (let ((parent-type (node-type parent))
	(prior-goal (if *re-use-prior-goal-p* (get-prior-goal node))))
    (case parent-type
      (operator-node
       (let* ((operator (get-operator-name parent parent-type))
	      (target-goal (get-target-goal (p4::nexus-parent parent)))
	      (some-candidate-goals (get-some-candidate-goals
				     (get-grandpa parent)
				     target-goal))
	      (cr-type (if successp 'select 'reject)))
	 (list (cons (create-cr `((current-operator ,operator)
				  (current-goal ,target-goal)
				  ,@(if *re-use-prior-goal-p*
					`((prior-goal (,prior-goal))))
				  (some-candidate-goals ,some-candidate-goals))
				state regressed-state
				(list cr-type 'bindings
				      (get-bindings-from-node node))
				nil nil
				(node-num-nodes parent))
		     some-candidate-goals))))
      (goal-node
       (let* ((operator (get-operator-name node (node-type node)))
	      (target-goal (get-target-goal parent))
	      (some-candidate-goals (get-some-candidate-goals
				     (p4::nexus-parent parent)
				     target-goal)))
	 (list (cons (create-cr `(,@(if (not successp)
					`((candidate-operator ,operator)))
				  (current-goal ,target-goal)
				  ,@(if *re-use-prior-goal-p*
					`((prior-goal (,prior-goal))))
				  (some-candidate-goals ,some-candidate-goals))
				state regressed-state
				(list (if successp 'select 'reject)
				      'operators operator)
				nil
				(get-bindings-from-success-child node)
				(node-num-nodes parent))
		     some-candidate-goals))))
      ((binding-node applied-op-node)
       (let* ((type (node-type node))
	      (candidate-goal (if (eq type 'goal-node)
				  (get-target-goal node)))
	      (candidate-goals-for-a (list candidate-goal))
	      (candidate-goals-for-s (get-some-candidate-goals parent
							       candidate-goal))
	      (instantiated-op (get-applied-op parent t))
	      (applicable-op (if instantiated-op
				 (get-applicable-op instantiated-op)))
	      (les-some-goals nil))
	 (if applicable-op
	     (let ((subgoalp (or (and successp (eq type 'goal-node))
				 (and (not successp)
				      (eq type 'applied-op-node)))))
	       (if (or subgoalp *learn-apply-p*)
		   (push (cons (create-cr `((applicable-op ,applicable-op)
					    ,@(if *re-use-prior-goal-p*
						  `((prior-goal (,prior-goal))))
					    (some-candidate-goals
					     ,candidate-goals-for-a))
					  state regressed-state
					  (list (if subgoalp
						    'sub-goal
						    'apply))
					  instantiated-op nil
					  (node-num-nodes parent))
			       candidate-goals-for-a)
			 les-some-goals))))
	 (if (and learn-goal-rule-p candidate-goals-for-s
		  (eq type 'goal-node))
	     (push (cons (create-cr `(,@(if successp
					    `((target-goal ,candidate-goal))
					    `((candidate-goal ,candidate-goal)))
				      ,@(if *re-use-prior-goal-p*
					    `((prior-goal (,prior-goal))))
				      (some-candidate-goals
				       ,candidate-goals-for-s))
				    state regressed-state
				    (list (if successp 'select 'reject)
					  'goals candidate-goal)
				    nil nil
				    (node-num-nodes parent))
			 candidate-goals-for-s)
		   les-some-goals))
	 les-some-goals)))))


;;;****************************************************************
;;
;; Returns the state in the node. I do not like it, because it does
;; more work than I would like (computing goals and so on), but I
;; will leave it now in this way, so that it works
;;
(DEFUN get-state-for-node (node)
  (declare (type p4::nexus node))
  (let ((current-node (if *my-current-node*
			  *my-current-node*
			  (get-last-node *last-node*))))
    (p4::maintain-state-and-goals current-node node)
    (setf *my-current-node* node)
    (let ((state (real-give-me-nice-state)))
      ;;      (p4::maintain-state-and-goals node *my-current-node*)
      state)))

;;
;; Returns the last node of the search
;;
(DEFUN get-last-node (node)
  (declare (type p4::nexus node))
  (let ((children (p4::nexus-children node)))
    (if children
	(get-last-node (car children))
        node)))


;;
;; Returns the name of the first applied operator of the childs of parent
;; In the case of positivep=t, it returns the instantiated-op
;;
(DEFUN get-applied-op (parent positivep)
  (declare (type p4::nexus parent))
  (let ((appliedp (if positivep
		      (some #'(lambda (child)
				(and (eq (node-type child) 'applied-op-node)
				     (p4::a-or-b-node-applied child)))
			    (p4::nexus-children parent))
		      (p4::a-or-b-node-applicable-ops-left parent))))
    (if appliedp
	(if positivep
	    (p4::op-application-instantiated-op (car appliedp))
	    (get-applicable-op (car appliedp))))))


;;
;; Returns the prior goal from a set of preconds
;;
(defun find-prior-goal (preconds)
  (declare (list preconds))
  (find-if #'(lambda (precond)
	       (eq (car precond) 'prior-goal))
	   preconds))

;;
;; Returns the goal node for a a-or-b-node
;;
(DEFUN get-goal-node (node type)
  (declare (type p4::nexus node))
  (if (eq type 'binding-node)
      (get-grandpa node)
    (do ((parent (p4::nexus-parent node) (p4::nexus-parent parent)))
	((p4::binding-node-p parent)
	 (get-grandpa parent)))))

;;
;; Returns the list of vars related to the bindings of a success child
;; of an operator node.
;;
(DEFUN get-bindings-from-success-child (node)
  (declare (type p4::nexus node))
  (mapcar #'cdr
	  (get-bindings-from-node (find-if #'(lambda (child)
					       (eq (node-learning-type child)
						   'success))
					   (p4::nexus-children node)))))


;;;****************************************************************
;;
;; Function that creates the control rule
;;
;; Until I change hamlet to cope with infinite values, I will have to check
;;whether the effects have infinite vars unbound in the current-goal
;;meta-predicate, and, in that case, do not create the rule.
(DEFUN create-cr (preconds state regressed-state effects
			   instantiated-op bindings-ctes num-nodes)
  (declare (list preconds state effects)
	   (ignore bindings-ctes))
  (let ((condes (create-cr-preconds preconds state instantiated-op
				    (member (car effects) '(sub-goal apply)
					    :test #'eq))))
    (setf *substitution* nil)
    (setf *rule-id* (string-trim "#:G" (gensym)))
    (find-substitution (cons effects condes))
    (let* ((name (name-cr effects
			  (if (eq (cadr effects) 'bindings)
			      (cadar condes))))
	   (complete-preconds (nconc condes (new-constraints)))
	   (new-preconds (parametrize complete-preconds))
	   (prior-goal (find-prior-goal new-preconds))
	   (new-cr (p4::make-control-rule
		    :name name
		    :if (cons 'and new-preconds)
		    :then (cons 'then (parametrize effects)))))
      (push new-cr *cr-list*)

;;; ********* Principio de Jose Luis Heras
      (when (or (eq *probabilities-p* 'rules) 
		(eq *probabilities-p* 'both))
	(nueva-regla-prob name effects))
;;; ********* Final de Jose Luis Heras

      ;; For utility
      ;;       (when *utilityp*
      ;; 	(setf (getf (p4::control-rule-plist new-cr) :matching-time) 0)
      ;; 	(setf (getf (p4::control-rule-plist new-cr) :matchings) 0)
      ;; 	(setf (getf (p4::control-rule-plist new-cr) :num-nodes) num-nodes)
      ;; 	(setf (getf (p4::control-rule-plist new-cr) :firings) 0))
      ;;       ;; end utility

      (incf *number-deduced-rules*)

      ;; For learning with FOIL      
      (if *save-examples-p*
	  (save-example-for-foil (if (eq *save-examples-format* 'ebl)
				     (remove-not-in-regressed-state complete-preconds regressed-state)
				   complete-preconds)
				 effects t))

      (if (eq *active-learning-mode* 'rules)
	  (update-active-learning-info regressed-state))
      
      ;; for the k-distance heuristic (old): (delete-not-main-preconds new-cr
      ;; (parametrize bindings-ctes))
      (delete-not-main-preconds new-cr (parametrize regressed-state))
      (if (or *inductionp* *utilityp*)
	  (let* ((rule-preconds (and *inductionp* (get-real-preconds new-cr)))
		 (pool (and *inductionp*
			    (different-preconds new-preconds 
						rule-preconds nil)))
		 (new-le (make-learning-episode
			  :name (get-new-le-name)
			  :type 'deduced
			  :rule new-cr
			  :problem *problem-name*
			  :from nil
			  :prior-goal
			  (and *inductionp*
			       (list
				(list new-cr prior-goal
				      (get-types-of-prior-goal prior-goal
							       rule-preconds
							       pool))))
			  :learning-steps 1
			  :matching-time 0
			  :matchings 0
			  :num-nodes num-nodes
			  :firings 0
			  :preconds-pool pool)))
	    (push new-le *le-list*)
	    (if *inductionp*
		(update-frequencies (cons prior-goal rule-preconds)
				    (find-decision-tree-node new-cr)
				    t)
	      (format t "~2%Adding new control rule: ~a~%" name))
	    (setf (gethash name *cr-to-le-hash*) new-le))
	(format t "~2%Adding new control rule: ~a~%" name)))))

;;
;; It returns the true-in-state conds ordered by the number of
;; predicates that they share with previous conds. Next step: order
;; them according to the entropy
;;
(DEFUN sort-preconds (preconds)
  (do* ((conds preconds (cdr conds))
	(cond (car conds) (car conds))
	(vars-in-precond (vars-in-precond cond)
			 (vars-in-precond cond))
	(new-preconds nil)
	(vars nil))
      ((or (null conds)
	   (eq (car cond) 'some-candidate-goals))
       (nconc (reverse new-preconds) conds))
    (case (car cond)
      (true-in-state
       (do ((rest-conds conds (cdr rest-conds))
	    (state-conds nil))
	   ((or (null (cdr rest-conds))
		(eq (caadr rest-conds) 'some-candidate-goals)
		(eq (caadr rest-conds) 'type-of-object))
	    (setq conds rest-conds)
	    (do ((rest-state-conds (if (eq (caar rest-conds) 'true-in-state)
				       (cons (car rest-conds) state-conds)
				       state-conds)))
		((null (cdr rest-state-conds))
		 (if rest-state-conds
		     (push (car rest-state-conds) new-preconds)))
	      (let ((best-state
		     (car (sort (copy-list rest-state-conds) #'>
				:key #'(lambda (scond)
					 (count-vars-in-preconds (cdadr scond)
								 vars))))))
		(push best-state new-preconds)
		(setq rest-state-conds (remove best-state rest-state-conds
					       :test #'equal))
		(dolist (var (cdadr best-state))
		  (pushnew var vars :test #'eq)))))
	 (push (car rest-conds) state-conds)))
      ((current-operator candidate-operator)
       (push cond new-preconds))
      (t ; prior-goal some-candidate-goals target-goal applicable-op
         ; current-goal candidate-goal
       (dolist (var vars-in-precond)
	 (pushnew var vars :test #'eq))
       (push cond new-preconds)))))

;; It updates the frecuencies of literals, predicates, instances and their
;; types that belonged to the regressed states
(defun update-active-learning-info (regressed-state)
  (let ((type nil))
    (dolist (literal regressed-state)
      (update-element literal 'literal)
      (update-element (car literal) 'predicate)
      (dolist (arg (cdr literal))
	(setq type (get-object-type arg))
	(update-element type 'type)
	(update-element arg type)))))

(defun update-element (element type)
  (declare (special *active-learning-hash*))
  (if (gethash (cons type element) *active-learning-hash*)
      (incf (gethash (cons type element) *active-learning-hash*))
    (setf (gethash (cons type element) *active-learning-hash*) 1))
  ; so that I also have a count per type for the roulette
  (if (gethash type *active-learning-hash*)
      (incf (gethash type *active-learning-hash*))
    (setf (gethash type *active-learning-hash*) 1)))

;;;***************************************************************
;;
;; Returns the preconds of the control rule by adding the state into
;; the preconditions, and deleting the null preconds.
;; Since we don't have a generator for some-candidate-goals, it is
;; better to place this meta-predicate after the state.
;;
(defun create-cr-preconds (preconds state instantiated-op decidep)
  (declare (list preconds state))
  (let ((some-candidate-goals
	 (if (eq (caar (last preconds)) 'some-candidate-goals)
	     (last preconds)))
	(rest-of-preconds (butlast preconds)))
    (append (delete-null-preconds rest-of-preconds)
	    (instantiate-cr-state
	     (if decidep
		 (delete-redundant-state state instantiated-op)
	         state))
	    some-candidate-goals)))

;;
;; Returns a list of true-in-state meta-predicates, one for each
;; predicate of the state
;;
(DEFUN instantiate-cr-state (state)
  (declare (list state))
  (mapcar #'(lambda (literal)
	      (cons 'true-in-state (list literal)))
	  state))

;;
;; Returns the preconds that are not null
;;
(DEFUN delete-null-preconds (conds)
  (declare (list conds))
  (cond ((null conds) nil)
	((not (or (eq (caar conds) 'some-candidate-goals)
		  (cadar conds)))
	 (delete-null-preconds (cdr conds)))
	(t (cons (car conds) (delete-null-preconds (cdr conds))))))

;;
;; Removes from the state those assertions that belong to the applicable
;; operator of the preconds when the rule is decide, since, obviously,
;; they are true in the state if the operator is applicable in the state
;;
(defun delete-redundant-state (state instantiated-op)
  (declare (list state))
  (if instantiated-op
      (let ((operator-preconds
	     (mapcar #'(lambda (precond)
			 (get-visible-form precond))
		     (get-instantiated-op-preconds instantiated-op))))
	(delete-if #'(lambda (assertion)
		       (member assertion operator-preconds :test #'equal))
		   state))))

;;
;; Returns the preconditions of an instantiated-op
;;
(DEFUN get-instantiated-op-preconds (instantiated-op)
  (let ((preconds (copy-list (getf (p4::instantiated-op-plist instantiated-op)
				   :satisfying-preconds))))
    (dolist (effects-list (getf (p4::instantiated-op-plist instantiated-op)
				:conditional-preconditions)
			  preconds)
      (dolist (effect effects-list)
	(dolist (precond (car effect))
	  (pushnew precond preconds :test #'equal))))))

;;
;; Saves an example in a file for future learning by FOIL
;; bk is background knowledge (preconds)
;; example were the effects
;; It does not save the type-of-object of the objects that do not appear elsewhere
;;
(DEFUN save-example-for-foil (bk example positivep)
  (declare (special *save-examples-format*))
  (with-open-file (stream *examples-file* :direction :output
		   ;; 			  :element-type 'character
		   :if-exists :append
		   :if-does-not-exist :create)
    (let ((*print-pretty* t)
	  (ctes (find-ctes example)))
      (format stream "~2%(~d ~a (" (incf *example-number*) positivep)
      (dolist (meta-predicate bk)
	(cond ((eq (car meta-predicate) 'type-of-object)
	       (let ((cte (car (rassoc (cadr meta-predicate)
				       *substitution* :test #'eq))))
		 (if (member cte ctes)
		     (format stream "~a~%          "
			     (list 'type-of-object
				   cte
				   (caddr meta-predicate))))))
	      (t (dolist (cte (find-ctes meta-predicate))
		   (pushnew cte ctes :test #'eq))
		 (format stream "~a~%          " meta-predicate))))
      (format stream ")~%  ~a)" example))))

;; Returns all constants in a given meta-predicate or then part
(DEFUN find-ctes (conde)
  (declare (list conde))
  (let ((ctes nil))
    (case (car conde)
      ((prior-goal some-candidate-goals)
       (dolist (pred-op (cadr conde))
	 (dolist (cte (get-pos-assertion pred-op))
	   (pushnew cte ctes :test #'eq))))
      ((target-goal true-in-state applicable-op current-goal candidate-goal)
       (dolist (cte (get-pos-assertion (cadr conde)))
	 (pushnew cte ctes :test #'eq)))
      ((select reject)
       (case (car conde)
	 (bindings (dolist (cte (mapcar #'cdr (cadr conde)))
		     (pushnew cte ctes :test #'eq)))
	 (goals (dolist (cte (cadr conde))
		  (pushnew cte ctes :test #'eq)))
	 (t nil)))
      (t nil))
    ctes))

;;;*************************************************************** Returns the
;; substitution of vars for constants to be applied to the cr from the
;; preconds and the effects
;;
(DEFUN find-substitution (condes)
  (declare (list condes))
  (dolist (conde condes)
    (case (car conde)
      ((prior-goal some-candidate-goals)
       (dolist (pred-op (cadr conde))
	 (sustituir-pred-op pred-op)))
      ((target-goal true-in-state applicable-op current-goal candidate-goal)
       (sustituir-pred-op (cadr conde)))
      ((select reject)
       (case (cadr conde)
	 (bindings (dolist (cte (mapcar #'cdr (caddr conde)))
		     (push-new-bind-pair cte)))
	 (goals (sustituir-pred-op (caddr conde)))
	 (t nil)))
      (t nil))))


;;
;; For all constants in the pred-op (predicate or operator), it adds a
;; binding for it and pushes it into the *substitution* list
;;
(DEFUN sustituir-pred-op (pred-op)
  (dolist (cte (cdr (get-pos-assertion pred-op)))
    (push-new-bind-pair cte)))

;;
;; Pushes the pair (new-var . cte) in the list of substitutions
;;
(DEFUN push-new-bind-pair (cte)
  (if (not (member cte *substitution* :key #'car :test #'eq))
      (push (cons cte (new-var-name cte)) *substitution*)))

;;
;; Returns a new name for a variable in the control rule.
;;
(DEFUN new-var-name (cte)
  (incf *number-vars-in-cr*)
  (let ((objectp (and (symbolp cte)
		      (p4::object-name-to-object cte
						 *current-problem-space*))))
    (intern (format nil "~:@(<~a-~d-~d>~)"
		    (cond ((numberp cte) 'infinite)
			  (objectp (get-object-type cte))
			  (t 'dummy))
		    *rule-id*
		    *number-vars-in-cr*))))


;;;***************************************************************
;;
;; Returns the name of a new rule, according to the day, month, type
;; and gensym, and the problem
;;
(DEFUN name-cr (effects operator)
  (multiple-value-bind (sec min hor dia mes ano) (get-decoded-time)
    (declare (ignore sec min hor ano))
    (let ((car1 (case (truncate dia 10)
		  ( 0 '|a|)
		  ( 1 '|e|)
		  ( 2 '|o|)
		  ( 3 '|i|)))
	  (car2 (case (mod dia 10)
		  ( 0 '|l|)
		  ( 1 '|n|)
		  ( 2 '|r|)
		  ( 3 '|s|)
		  ( 4 '|b|)
		  ( 5 '|f|)
		  ( 6 '|k|)
		  ( 7 '|p|)
		  ( 8 '|t|)
		  ( 9 '|x|)))
	  type subtype)
      (cond ((cdr effects)
	     (setq type (car effects))
	     (setq subtype (case (cadr effects)
			     (bindings (format nil "bind-~a" operator))
			     (goals (caaddr effects))
			     (operators (caddr effects)))))
	    (t (setq type 'decide)
	       (setq subtype (car effects))))
      (intern (format nil "~:@(~a-~a-~a-~a~a~a-~a~)" type subtype
		      *problem-name* (month-string mes)
		      car1 car2 *rule-id*)))))

;; Para el TI Explorer: (substring (time:month-string mes) 0 3)
;; Para el TI Explorer: (substring (gensym) 1))))))


;;;***************************************************************
;;
;; Returns the list of type-of-object meta-predicates, according to
;; the list of variables and their current type
;;
(DEFUN new-constraints nil
  (mapcar #'(lambda (cte-var)
	      (list 'type-of-object (cdr cte-var)
		    (if (numberp (car cte-var))
			'top-type
			(get-object-type (car cte-var)))))
	  *substitution*))

;;;***************************************************************

;; It parametrizes by substitution of vars for constants
(DEFUN parametrize (expression)
  (declare (list expression))
  (sublis *substitution* expression))


;;;***************************************************************
;;
;;; Deletes the not important preconds (those true-in-state that are
;;; not in the regressed-state)
;;
(defun delete-not-main-preconds (new-cr regressed-state)
  (declare (type p4::control-rule new-cr))
  (setf (p4::control-rule-if new-cr)
	(cons 'and (main-preconds (get-real-preconds new-cr)
				  regressed-state
				  (get-real-effects new-cr)))))

;;
;; New version. It deletes the preconditions of the state that do not
;; belong to the regressed-state. It adds the prior-goal if
;; *learn-prior-goal-p* is t, or if the rule is a decide-sub-goal
;;
(DEFUN main-preconds (preconds regressed-state effects)
  (declare (list preconds regressed-state effects))
  (let ((main-preconds nil)
	(vars (if (eq (cadr effects) 'bindings)
		  (mapcar #'cdr (caddr effects))
		  (vars-in-precond effects))))
    (do* ((conds preconds (cdr conds))
	  (precond (car conds) (car conds)))
	((or (null conds)
	     (eq (car precond) 'type-of-object)))
      (when (and (or *learn-prior-goal-p*
		     (not (prior-goal-meta-pred-p precond)))
		 (or (not (eq (car precond) 'true-in-state))
		     (member (cadr precond) regressed-state :test #'equal)))
	(dolist (var (vars-in-precond precond))
	  (pushnew var vars :test #'eq))
	(push precond main-preconds)))
    (dolist (var vars (sort-preconds (nreverse main-preconds)))
      (push (find-type-of-precond var preconds t)
	    main-preconds))))

(DEFUN remove-not-in-regressed-state (complete-preconds regressed-state)
  (declare (list complete-preconds regressed-state))
  (remove-if #'(lambda (precond)
		 (and (eq (car precond) 'true-in-state)
		      (not (member (cadr precond) regressed-state :test #'equal))))
	     complete-preconds))
 

;;
;; Returns a list with the vars that appear in the precond
;;
(DEFUN vars-in-precond (precond)
  (declare (list precond))
  (cond ((null precond) nil)
	((listp (car precond))
	 (nconc (vars-in-precond (car precond))
		(vars-in-precond (cdr precond))))
	((p4::is-variable-p (car precond))
	 (cons (car precond)
	       (vars-in-precond (cdr precond))))
	(t (vars-in-precond (cdr precond)))))

;;
;; Returns the type-of-object meta-predicate for the var
;;
(defun find-type-of-precond (var preconds breakp)
  (declare (list preconds))
  (let ((foundp (find-if #'(lambda (precond)
			     (and (eq (car precond) 'type-of-object)
				  (eq (cadr precond) var)))
			 preconds :from-end t)))
    (or foundp
	(and breakp
	     (list 'type-of-object var (get-type-from-name var))))))

;; 	(and breakp (break) nil))))

(defun get-type-from-name (var)
  (let ((name (subseq (string-right-trim "0123456789->"
					 (format nil "~a" var))
		      1)))
    (intern (format nil "~:@(<~a>~)"
		    (if (and (eq (elt name 0) #\I)
			     (eq (elt name 1) #\-))
			(subseq name 2)
			name)))))

;;
;; Returns the type-of-object declarations of the vars in the
;; prior-goal that do not appear in the preconds of the rule. They are
;; needed later when adding a prior-goal to the preconds of an
;; induced rule.
;;
(DEFUN get-types-of-prior-goal (prior-goal preconds pool)
  (declare (list preconds pool prior-goal))
  (let ((vars-in-preconds (get-all-vars-from-types preconds))
	(types nil))
    (dolist (goal (cadr prior-goal) types)
      (dolist (var (cdr (get-pos-assertion goal)))
	(if (not (member var vars-in-preconds :test #'eq))
	    (push (find-type-of-precond var pool nil) types))))))

;;
;; Returns a list of the vars that appear as defined in the
;; type-of-object meta-predicate
;;
(DEFUN get-all-vars-from-types (preconds)
  (declare (list preconds))
  (let ((all-vars nil))
    (dolist (precond preconds all-vars)
      (if (eq (car precond) 'type-of-object)
	  (push (cadr precond) all-vars)))))


;;**********************************************************************
;;*     Functions that check for subsumed rules
;;**********************************************************************

;;
;; Calls the function deletablep to see if the new cr should be added,
;; removed, or there is any other rule that should be deleted.
;;
(DEFUN include-rule (new-cr control-rules)
  (declare (type p4::control-rule new-cr)
	   (list control-rules))
  (let* ((name1 (p4::control-rule-name new-cr))
	 (deletablep (or (some #'(lambda (rule)
				   (let ((delp (deletablep new-cr rule)))
				     (if (not (eq delp 'include))
					 (cons delp rule))))
			       control-rules)
			 '(include)))
	 (rule (cdr deletablep))
	 (name2 (if rule
		    (p4::control-rule-name rule))))
    (case (car deletablep)
      (include (format t "Adding new control rule: ~a~%" name1)
	       (push new-cr *cr-list*)
	       'include)
      (delete-1 (incf *number-deleted-rules*)
		(format t "~%Found subsumed rule ~a~%Deleting ~a~%"
			name2 name1)
		'delete-1)
      (delete-2	(format t "~%Found subsumed rule ~a" name1)
		(format t "~%Adding new control rule: ~a~%" name1)
		(incf *number-deduced-rules*)
		(push new-cr *cr-list*)
		(let ((le (gethash name2 *cr-to-le-hash*)))
		  (delete-cr rule nil le
			     (if (eq (learning-episode-type le) 'deduced)
				 t
			         'forever)))
		'delete-2)
      (t nil))))


;;
;; Returns include if new-cr1 should be added, delete-1 if new-cr should
;; be deleted (rule2 included in rule1), and delete-2 if rule2 should
;; be deleted (rule1 included in rule2)
;;
(DEFUN deletablep (new-cr rule)
  (declare (type p4::control-rule new-cr rule))
  (let* ((preconds1 (get-real-preconds new-cr))
	 (effects1 (get-real-effects new-cr))
	 (preconds2 (get-real-preconds rule))
	 (effects2 (get-real-effects rule)))
    (if (and (eq (car effects1) (car effects2))
	     (eq (cadr effects1) (cadr effects2))
	     (case (cadr effects1)
	       (operators (eq (caddr effects1) (caddr effects2)))
	       (goals (eq (caaddr effects1) (caaddr effects2)))
	       (t t)))
	(check-equality preconds1 preconds2 effects1 effects2)
        'include)))


;; It checks whether rule2 is included in rule1. If so, it removes rule1,
;; since it is overspecific. It returns t, so that the calling
;; function should remove it.
;; Otherwise, it checks whether rule1 is included in rule2. If so,
;; rule2 is deleted, since it is overspecific

(DEFUN check-equality (preconds1 preconds2 effects1 effects2)
  (declare (list preconds1 preconds2 effects1 effects2))
  (cond ((<= (length preconds2) (length preconds1))
	 (if (equal-control-rule-p preconds2 preconds1 effects2
				   effects1 'subset)
	     'delete-1
	     'include))
	((equal-control-rule-p preconds1 preconds2 effects1 effects2 'subset)
	 'delete-2)
	(t 'include)))

;;
;;; New version of equal-control-rule-p (much better): incremental matching
;;; It returns all the bindings that are coherent with the effects if
;;; includedp is nil. If not, it returns the first one
;;;
(DEFUN equal-control-rule-p (preconds1 preconds2 effects1 effects2
				       includedp)
  (declare (list preconds1 preconds2 effects1 effects2))
  (if includedp
      (new-get-bindings-of-preconds (sort-preconds-for-matching preconds1)
				    (sort-preconds-for-matching preconds2)
				    effects1 effects2
				    (list nil) includedp)
    (let* ((binding-effects (equal-effects-p effects1 effects2 nil))
	   (bindings (if binding-effects
			 (if (or (eq t binding-effects)
				 (not (car binding-effects)))
			     (list nil)
			   (list binding-effects)))))
      (if bindings
	  (new-get-bindings-of-preconds (sort-preconds-for-matching preconds1)
					(sort-preconds-for-matching preconds2)
					effects1 effects2
					bindings includedp)))))

;;; Raquel, april 2006: i add the next "if" to control errors when precond1 or precond2 are nil
(defun sort-preconds-for-matching (preconds)
  (sort preconds #'(lambda (precond1 precond2)
		     (if (and precond1 precond2)
			 (<= (position (car precond1) *hamlet-meta-predicates*)
			     (position (car precond2) *hamlet-meta-predicates*))
		         t))))

;;; Returns the new list of bindings. There are three cases: 1. binding is
;;; more general than a certain x in all-bindings it is not added 2. binding
;;; is more specific than a set of x in all-bindings the x in this set are
;;; deleted from all-bindings and binding is added 3. Otherwise, it is added
;;;
;;; The trick is that if it finds that there is one that is more general
;;; than binding, then it should not keep testing whether binding is more
;;; general than any other, since it can not happen that there is a
;;; binding more general/specific than another one in all-bindings
;;;
(DEFUN introduce-new-binding (binding all-bindings)
  (declare (list binding all-bindings))
  (if (car binding)
      (do* ((bindings all-bindings (cdr bindings))
	    (first-binding (car bindings) (car bindings))
	    (result nil)
	    (more-specific-p nil)
	    (more-general-p nil))
	  ((or (null bindings)
	       more-general-p)
	   (if more-general-p
	       all-bindings
	       (cons binding result)))
	(if (and (not more-specific-p)
		 (subsetp binding first-binding :test #'eq-binding-p))
	    (setq more-general-p t)
	    (unless (subsetp first-binding binding :test #'eq-binding-p)
	      (setq more-specific-p t)
	      (push first-binding result))))
      all-bindings))

;;
;;; It returns t if bind-pair1 is eq to bind-pair2
;;
(DEFUN eq-binding-p (bind-pair1 bind-pair2)
  (and (eq (car bind-pair1) (car bind-pair2))
       (eq (cdr bind-pair1) (cdr bind-pair2))))

;;;
;;; It returns t if binding1 and binding2 are equal binding lists
;;;
(DEFUN equal-bindings-p (binding1 binding2)
  (and (= (length binding1) (length binding2))
       (subsetp binding1 binding2 :test #'eq-binding-p)))

;;       (equal-set-p binding1 binding2 :test eq-binding-p))

;;
;; Returns a list of bindings:
;; (((<object2> . <object1>) .. (<object3> . <object2>)) ...), if
;; preconds1 is included in preconds2, or nil if not matching preconds
;;; The test-inclusion-p parameter allows to control whether the
;;; preconds1 should be less than preconds2 (included) or not.
;;; It can be:
;;;  - subset: test in get-bindings-some-or-prior-goals whether goals2
;;;            is a subset of goals1
;;;  - intersection: the same but testing whether they intersect
;;;  - nil: to know the intersection of both
;;
(DEFUN new-get-bindings-of-preconds (preconds1 preconds2 effects1 effects2
					       bindings test-inclusion-p)
  (declare (list preconds1 preconds2 bindings))
  (cond ((null preconds1) (end-matching effects1 effects2 bindings
					test-inclusion-p))
	((null preconds2) nil)
	(t (let ((meta-predicate (caar preconds1)))
	     (case meta-predicate
	       (true-in-state
		(cond ((eq meta-predicate (caar preconds2))
		       (new-get-bindings-of-state preconds1 preconds2
						  effects1 effects2
						  bindings test-inclusion-p))
		      (test-inclusion-p nil)
		      (t (new-get-bindings-of-preconds
			  (cdr preconds1) preconds2 effects1 effects2
			  bindings test-inclusion-p))))
	       (type-of-object
		(end-matching effects1 effects2
			      (test-bindings-of-types
			       preconds1 preconds2 bindings
			       (if test-inclusion-p
				   #'more-general-type-p
				   #'common-supertype))
			      test-inclusion-p))
	       ((candidate-operator current-operator)
		(if (equal (car preconds1) (car preconds2))
		    (new-get-bindings-of-preconds (cdr preconds1)
						  (cdr preconds2)
						  effects1 effects2
						  bindings test-inclusion-p)))
	       ((current-goal candidate-goal target-goal applicable-op)
		(if (eq meta-predicate (caar preconds2))
		    (new-get-bindings-of-some-goals preconds1 preconds2
						    effects1 effects2 bindings
						    test-inclusion-p)))
	       (some-candidate-goals
		(new-get-bindings-some-or-prior-goals preconds1 preconds2
						      effects1 effects2
						      bindings
						      test-inclusion-p nil))
	       (prior-goal
		(new-get-bindings-some-or-prior-goals preconds1 preconds2
						      effects1 effects2
						      bindings
						      test-inclusion-p t))
	       (t nil))))))


;;;
;;; Changed 13-07-95. If test-inclusion-p=nil, then it has already
;;; computed the bindings of effects at the beginning of checking
;;; whether the two rules match.
;;;
(DEFUN end-matching (effects1 effects2 bindings test-inclusion-p)
  (and bindings
       (if test-inclusion-p
	   (some #'(lambda (binding)
		     (let ((new-binding (last-binding effects1 effects2
						      binding)))
		       (if new-binding
			   (list new-binding))))
		 bindings)
	   bindings)))


(defun last-binding (effects1 effects2 binding)
  (let ((binding-effects (equal-effects-p effects1 effects2 binding)))
    (if binding-effects
	(if (or (eq t binding-effects)
		(not (car binding-effects)))
	    binding
	    (append binding-effects binding)))))
  
;;
;; Returns the bindings in case of being a goal-based meta-predicate
;;
(DEFUN new-get-bindings-of-some-goals (preconds1 preconds2 effects1 effects2
						 bindings test-inclusion-p)
  (declare (list preconds1 preconds2 bindings))
  (if test-inclusion-p
      (do ((aux-bindings bindings (cdr aux-bindings))
	   (last-bindings nil))
	  ((or (null aux-bindings) last-bindings)
	   last-bindings)
	(let* ((binding (car aux-bindings))
	       (new-binding (get-binding-of-goals (cadar preconds1)
						  (cadar preconds2)
						  binding))
	       (the-append (if new-binding
			       (my-append new-binding binding
					  #'eq-binding-p)))
	       (new-bindings (if the-append
				 (if (car the-append)
				     (list the-append)
				     the-append))))
	  (setq last-bindings
		(if new-bindings
		    (new-get-bindings-of-preconds (cdr preconds1)
						  (cdr preconds2)
						  effects1 effects2
						  new-bindings
						  test-inclusion-p)))))
      (let ((new-bindings
	     (mapcan #'(lambda (binding)
			 (let* ((new-binding (get-binding-of-goals
					      (cadar preconds1)
					      (cadar preconds2)
					      binding))
				(the-append (if new-binding
						(my-append new-binding binding
							   #'eq-binding-p))))
			   (if the-append
			       (if (car the-append)
				   (list the-append)
				   the-append))))
		     bindings)))
	(if new-bindings
	    (new-get-bindings-of-preconds (cdr preconds1) (cdr preconds2)
					  effects1 effects2
					  new-bindings test-inclusion-p)))))
				
;;;
;;; Returns a list of bindings from all the true-in-state in the preconds.
;;; This is more efficient than the clearer way of doing it by
;;; removing all meta-predicates of the preconds that are not state.
;;; The dummy thing is for the first cdr in the do.
;;;
;;; Incremental version
;;;
(DEFUN new-get-bindings-of-state (preconds1 preconds2 effects1 effects2
					    bindings test-inclusion-p)
  (declare (list preconds1 preconds2 bindings))
  (let ((p1 (cons 'dummy preconds1))
	(p2 (cons 'dummy preconds2))
	state1 state2 end1 end2)
    (do ()
	((or (null p1) end1))
      (setq p1 (cdr p1))
      (if (eq (caar p1) 'true-in-state)
	  (push (cadar p1) state1)
	(setq end1 t)))
    (do ()
	((or (null p2) end2))
      (setq p2 (cdr p2))
      (if (eq (caar p2) 'true-in-state)
	  (push (cadar p2) state2)
	(setq end2 t)))
    (if (or (not test-inclusion-p)
	    (<= (length state1) (length state2)))
	(get-all-bindings-i state1 state2 p1 p2 effects1 effects2 bindings
			    test-inclusion-p t))))

(DEFUN new-get-bindings-some-or-prior-goals (preconds1 preconds2 effects1
						       effects2 bindings
						       test-inclusion-p
						       prior-goal-p)
  (declare (list preconds1 preconds2 bindings))
  (let ((meta-predicate1 (caar preconds1))
	(meta-predicate2 (caar preconds2)))
    (cond ((or (and test-inclusion-p
		    (eq meta-predicate1 meta-predicate2))
	       (and (not test-inclusion-p)
		    (or (and (not prior-goal-p)
			     (eq meta-predicate2 'some-candidate-goals))
			(and prior-goal-p
			     (eq meta-predicate2 'prior-goal)))))
	   (new-get-bindings-some-or-prior-goals-1 preconds1 preconds2
						   effects1 effects2
						   bindings test-inclusion-p))
	  ((and (not prior-goal-p)
		(eq meta-predicate2 'true-in-state))
	   (do ((preconds preconds2 (cdr preconds)))
	       ((or (null preconds)
		    (not (eq (caar preconds) 'true-in-state)))
		(new-get-bindings-some-or-prior-goals preconds1 preconds
						      effects1 effects2
						      bindings test-inclusion-p
						      prior-goal-p))))
	  (t (new-get-bindings-of-preconds (cdr preconds1) preconds2
					   effects1 effects2
					   bindings test-inclusion-p)))))


;;
;; Incremental version
;;
(DEFUN new-get-bindings-some-or-prior-goals-1 (preconds1 preconds2
							 effects1 effects2
							 bindings
							 test-inclusion-p)
  (declare (list preconds1 preconds2 bindings))
  (let ((goals1 (cadar preconds1))
	(goals2 (cadar preconds2))
	(rest-preconds1 (cdr preconds1))
	(rest-preconds2 (cdr preconds2)))
    (if (or (not test-inclusion-p)
	    (eq test-inclusion-p 'intersection)
	    (<= (length goals2) (length goals1)))
	(if (and goals1 goals2)
	    (get-all-bindings-i goals1 goals2 rest-preconds1 rest-preconds2
				effects1 effects2
				bindings test-inclusion-p nil)
	    (if (or (and test-inclusion-p (not goals2)
			 (or (not goals1)
			     (member nil goals1 :test #'eq)))
		    (not test-inclusion-p))
		(purge-bindings (new-get-bindings-of-preconds rest-preconds1
							      rest-preconds2
							      effects1 effects2
							      bindings
							      test-inclusion-p)
				goals1))))))

(DEFUN reverse-bindings (bindings)
  (declare (list bindings))
  (mapcar #'(lambda (binding)
	      (reverse-binding binding))
	  bindings))


(DEFUN test-bindings-of-types (preconds1 preconds2 bindings test-function)
  (declare (list preconds1 preconds2 bindings))
;; (function test-function)
  (let ((result nil))
    (dolist (binding bindings result)
      (if (all-matching-types test-function preconds1 preconds2 binding)
	  (pushnew binding result :test #'equal-bindings-p)))))

(DEFUN all-matching-types (test-function preconds1 preconds2 binding)
  (declare (list preconds1 preconds2 binding))
;; (function test-function)
  (or (null binding)
      (and (funcall test-function preconds1 preconds2 (car binding))
	   (all-matching-types test-function preconds1
			       preconds2 (cdr binding)))))


;;
;; It returns the common supertype that the vars in the bind pair
;; have, in case they have one different than :top-type
;;
(DEFUN common-supertype (preconds1 preconds2 bind-pair)
  (declare (list preconds1 preconds2 bind-pair))
  (let ((type1 (p4::type-name-to-type (find-type-of-var preconds1
							(car bind-pair))
				      *current-problem-space*))
	(type2 (p4::type-name-to-type (find-type-of-var preconds2
							(cdr bind-pair))
				      *current-problem-space*)))
    (if (and type1 type2)
	(if (eq type1 type2)
	    (p4::p4type-name type1)
	    (let* ((all-supers1 (cons type1 (p4::p4type-all-parents type1)))
		   (all-supers2 (cons type2 (p4::p4type-all-parents type2)))
		   (first-common-type (find-common-type all-supers1
							all-supers2)))
	      (if first-common-type
		  (p4::p4type-name first-common-type)))))))


(DEFUN find-common-type (all-supers1 all-supers2)
  (declare (list all-supers1 all-supers2))
  (cond ((eq (car all-supers1) *top-type*) nil)
	((member (car all-supers1) all-supers2)
	 (car all-supers1))
	(t (find-common-type (cdr all-supers1) all-supers2))))

(DEFUN more-general-type-p (preconds1 preconds2 bind-pair)
  (declare (list preconds1 preconds2 bind-pair))
  (let ((type1 (find-type-of-var preconds1 (car bind-pair)))
	(type2 (find-type-of-var preconds2 (cdr bind-pair))))
    (or (eq type1 type2)
	(let ((all-supers2 (p4::p4type-all-parents
			    (p4::type-name-to-type type2
						   *current-problem-space*))))
	  (member type1 all-supers2 :test #'eq :key #'p4::p4type-name)))))


(DEFUN find-type-of-var (preconds var)
  (declare (list preconds))
  (let ((precond-of-var (member var preconds :key #'cadr)))
    (if precond-of-var
	(caddar precond-of-var))))

;;
;; Given two literals with variables, it returns the same as
;; new-binding if the predicates are the same, and the are both + or -
;;
(DEFUN get-binding-of-goals (goal1 goal2 binding)
  (declare (list goal1 goal2 binding))
  (let* ((negatedp1 (negatedp goal1))
	 (negatedp2 (negatedp goal2))
	 pos-goal1 pos-goal2
	 (matchp (cond ((and negatedp1 negatedp2)
			(setq pos-goal1 (cadr goal1))
			(setq pos-goal2 (cadr goal2))
			t)
		       ((and (not negatedp1) (not negatedp2))
			(setq pos-goal1 goal1)
			(setq pos-goal2 goal2)
			t)
		       (t nil))))
    (if (and matchp (eq (car pos-goal1) (car pos-goal2)))
	(new-binding (cdr pos-goal1) (cdr pos-goal2) binding))))

;;
;; Given two lists of vars and an initial binding, it returns
;; 1. (nil) if they match, but there are not variables
;; 2. nil if they do not match
;; 3. A binding if there is a match
;;
(DEFUN new-binding (vars1 vars2 binding)
  (declare (list vars1 vars2 binding))
  (cond ((and (null vars1) (null vars2)) '(nil))
	((or (null vars1) (null vars2)) nil)
	(t (let ((var1 (car vars1))
		 (var2 (car vars2)))
	     (if (p4::is-variable-p var1)
		 (let ((memberp (member var1 binding :test #'eq
					:key #'car)))
		   (if memberp
		       (if (eq (cdar memberp) var2)
			   (new-binding (cdr vars1) (cdr vars2) binding))
		       (let ((memberp2 (member var2 binding
					       :test #'eq :key #'cdr)))
			 (if (not memberp2)
			     (combine-bindings vars1 vars2 binding)))))
	         (if (eq var1 var2)
		     (new-binding (cdr vars1) (cdr vars2) binding)))))))

;;; This version allows to have a substitution list such as
;;; ((<x> . <y>)(<z> . <y>)), which is right according to the
;;; unification algorithm, but not according to the restriction in
;;; hamlet that the variables in the control rule all have to be different.

(DEFUN combine-bindings (vars1 vars2 binding)
  (declare (list vars1 vars2 binding))
  (let* ((new-bind (cons (car vars1) (car vars2)))
	 (new-binding (new-binding (cdr vars1) (cdr vars2)
				   (cons new-bind binding))))
    (if new-binding
	(if (car new-binding)
	    (cons new-bind new-binding)
	  (list new-bind)))))


;;
;; If test-inclusion-p is t, it
;; returns a list of bindings if goals1 is a matching subset of
;; goals2, giving the binding as input
;; (((<o1> . <o2>) ...) ((<o1> . <o3>) ...) ...)
;; If nil, it returns the same if they match somehow
;; Any binding might be of the form (NIL), which means that they match
;; but there are not variables.
;;
(DEFUN get-bindings-of-goals (goals1 goals2 binding test-inclusion-p)
  (declare (list goals1 goals2 binding))
  (if (and goals1 goals2)
      (let ((goal1 (car goals1))
	    (new-bindings nil))
	(dolist (goal2 goals2)
	  (let ((new-binding (get-binding-of-goals goal1 goal2 binding)))
	    (if new-binding
		(let ((bindings-of-goals
		       (get-bindings-of-goals
			(cdr goals1)
			(remove goal2 goals2 :test #'equal)
			(my-append new-binding binding #'eq-binding-p)
			test-inclusion-p)))
		  (dolist (binding-of-goals bindings-of-goals)
		    (pushnew binding-of-goals new-bindings
			     :test #'equal-bindings-p))))))
	(if test-inclusion-p
	    new-bindings
	    (let ((rest-bindings (get-bindings-of-goals (cdr goals1) goals2
							binding
							test-inclusion-p)))
	      (if (or new-bindings rest-bindings)
		  (my-append rest-bindings new-bindings #'equal-bindings-p)))))
    (if (and goals1 test-inclusion-p (not goals2))
	nil
        (list binding))))

;; If test-inclusion-p is t, it
;; returns a list of bindings if goals1 is a matching subset of
;; goals2, giving the binding as input
;; (((<o1> . <o2>) ...) ((<o1> . <o3>) ...) ...)
;; If nil, it returns the same if they match somehow
;; Any binding might be of the form (NIL), which means that they match
;; but there are not variables.
;;
;; Incremental version
;;
(DEFUN new-get-bindings-of-goals (goals1 goals2 rest1 rest2 effects1
					 effects2 binding
					 test-inclusion-p statep)
  (declare (list goals1 goals2 binding))
  (if (and goals1 goals2)
      (if test-inclusion-p
	  (get-bindings-of-goals-i goals1 goals2 rest1 rest2
				   effects1 effects2 binding
				   test-inclusion-p statep)
	  (get-bindings-of-goals-ni goals1 goals2 effects1 effects2
				    binding test-inclusion-p))
      (if (and goals1 test-inclusion-p (not goals2))
	  nil
          (list binding))))

(DEFUN get-bindings-of-goals-i (goals1 goals2 rest1 rest2 effects1 effects2
				       binding test-inclusion-p statep)
  (do* ((rest-goals2 goals2 (cdr rest-goals2))
	(goal2 (car rest-goals2) (car rest-goals2))
	(goal1 (car goals1))
	(new-bindings nil)
	(last-bindings nil))
      ((or (null rest-goals2) last-bindings)
       last-bindings)
    (let* ((new-binding (get-binding-of-goals goal1 goal2 binding)))
      (when new-binding
	(setq new-bindings
	      (new-get-bindings-of-goals (cdr goals1)
					 (remove goal2 goals2 :test #'equal)
					 rest1 rest2 effects1 effects2
					 (my-append new-binding binding
						    #'eq-binding-p)
					 test-inclusion-p statep))
	(setq last-bindings
	      (if statep
		  (if new-bindings
		      (new-get-bindings-of-preconds rest1 rest2
						    effects1 effects2
						    new-bindings
						    test-inclusion-p))
		  new-bindings))))))

(DEFUN get-bindings-of-goals-ni (goals1 goals2 effects1 effects2
					binding test-inclusion-p)
  (let ((goal1 (car goals1))
	(new-bindings nil))
    (dolist (goal2 goals2)
      (let ((new-binding (get-binding-of-goals goal1 goal2 binding)))
	(if new-binding
	    (let ((bindings-of-goals
		   (new-get-bindings-of-goals
		    (cdr goals1)
		    (remove goal2 goals2 :test #'equal)
		    nil nil effects1 effects2
		    (my-append new-binding binding #'eq-binding-p)
		    test-inclusion-p nil)))
	      (dolist (binding-of-goals bindings-of-goals)
		(pushnew binding-of-goals new-bindings
			 :test #'equal-bindings-p))))))
    (let ((rest-bindings (new-get-bindings-of-goals (cdr goals1) goals2
						    nil nil effects1 effects2
						    binding
						    test-inclusion-p nil)))
      (if new-bindings
	  (if rest-bindings
	      (if (equal rest-bindings (list binding))
		  new-bindings
		  (my-append rest-bindings new-bindings #'equal-bindings-p))
	      new-bindings)
	  rest-bindings))))
	      
;;
;; Returns the list of bindings that can match both sets of goals
;;
(DEFUN get-all-bindings (goals1 goals2 bindings test-inclusion-p)
  (declare (list goals1 goals2 bindings))
  (mapcan #'(lambda (binding)
	      (get-bindings-of-goals goals1 goals2
				     binding test-inclusion-p))
	  bindings))

;;
;; Returns the list of bindings that can match both sets of goals
;; Incremental version. If statep is t, we are getting the bindings of
;; the state. If not, we are getting them from some-candidate-goals or
;; prior-goal
;;
(DEFUN get-all-bindings-i (goals1 goals2 rest1 rest2 effects1 effects2 bindings
				  test-inclusion-p statep)
  (declare (list goals1 goals2 bindings))
  (if test-inclusion-p
      (if statep
	  (get-all-bindings-i-state goals1 goals2 rest1 rest2 effects1 effects2
				    bindings test-inclusion-p)
	  (get-all-bindings-i-some goals1 goals2 rest1 rest2 effects1 effects2
				   bindings test-inclusion-p))
      (if statep
	  (or (mapcan #'(lambda (binding)
			  (let ((new-bindings
				 (new-get-bindings-of-goals goals1 goals2
							    rest1 rest2
							    effects1 effects2
							    binding
							    test-inclusion-p
							    statep)))
			    (if new-bindings
				(new-get-bindings-of-preconds
				 rest1 rest2 effects1 effects2 new-bindings
				 test-inclusion-p))))
		      bindings)
	      bindings)
	  (let ((new-bindings
		 (reverse-bindings
		  (get-all-bindings goals2 goals1 (reverse-bindings bindings)
				    nil))))
	    (if new-bindings
		(purge-bindings (new-get-bindings-of-preconds
				 rest1 rest2 effects1 effects2 new-bindings
				 test-inclusion-p)
				goals1))))))

(DEFUN get-all-bindings-i-state (goals1 goals2 rest1 rest2 effects1 effects2
					bindings test-inclusion-p)
  (do ((aux-bindings bindings (cdr aux-bindings))
       (last-bindings nil))
      ((or (null aux-bindings) last-bindings)
       last-bindings)
    (let* ((binding (car aux-bindings))
	   (new-bindings (new-get-bindings-of-goals
			  goals1 goals2 rest1 rest2 effects1 effects2 binding
			  test-inclusion-p t)))
      (setq last-bindings
	    (if new-bindings
		(new-get-bindings-of-preconds rest1 rest2 effects1 effects2
					      new-bindings
					      test-inclusion-p))))))

;;
;; This is needed only for the incremental version of the function
;; new-get-bindings-some-or-prior-goals-1
;;
(defun get-all-bindings-i-some (goals1 goals2 rest-preconds1 rest-preconds2
				       effects1 effects2
				       bindings test-inclusion-p)
  (some #'(lambda (binding)
	    (let* ((reverse-binding (reverse-binding binding))
		   (new-bindings (new-get-bindings-of-goals
				  goals2 goals1 rest-preconds1 rest-preconds2
				  effects1 effects2
				  reverse-binding (eq test-inclusion-p
						      'subset)
				  nil))
		   (last-bindings
		    (find-if #'(lambda (new-binding)
				 (purge-bindings
				  (new-get-bindings-of-preconds
				   rest-preconds1 rest-preconds2
				   effects1 effects2
				   (list (reverse-binding new-binding))
				   test-inclusion-p)
				  goals1))
			     new-bindings)))
	      (if last-bindings
		  (list last-bindings))))
	bindings))


;;;
;;; It generates a new bind-pair for each binding list in which there is a
;;; bind pair such as (<object2> . <object1>), there is no bind pair
;;; for <object1>, and there is a goal in goals that has the variable
;;; <object1>
;;;
(DEFUN purge-bindings (bindings goals)
  (declare (list goals bindings))
  (mapcan #'(lambda (binding)
	      (let ((new-binding binding))
		(setf *number-vars-in-cr* 0)
		(setf *rule-id* (string-trim "#:G" (gensym)))
		(dolist (goal goals (list new-binding))
		  (dolist (var (cdr (get-pos-assertion goal)))
		    (if (and (member var binding :test #'eq :key #'cdr)
			     (not (member var binding :test #'eq :key #'car)))
			(push (cons var (new-var-from-new-var var))
			      new-binding))))))
	  bindings))


(DEFUN new-var-from-new-var (var)
  (incf *number-vars-in-cr*)
  (intern (format nil "~:@(<~a-~d-~d>~)"
		  (string-trim (cons #\- *trim-characters*) var)
		  *rule-id* *number-vars-in-cr*)))


;;
;; Returns t if both effects are equal. Now, we check whether the
;; bindings are also compatible. Returns (nil), t, or a binding of
;; variables not appearing in binding
;;
(DEFUN equal-effects-p (effects1 effects2 binding)
  (declare (list effects1 effects2 binding))
  (and (eq (car effects1) (car effects2))
       (or (and (null (cdr effects1)) (null (cdr effects2)))
	   (and (eq (cadr effects1) (cadr effects2))
		(case (cadr effects1)
		  (operators (eq (caddr effects1) (caddr effects2)))
		  (goals (get-binding-of-goals (caddr effects1)
					       (caddr effects2)
					       binding))
		  (bindings (new-binding (mapcar #'cdr (caddr effects1))
					 (mapcar #'cdr (caddr effects2))
					 binding))
		  (t nil))))))

(DEFUN my-append (list1 list2 function)
  (declare (list list1 list2))
;; (function function)
  (cond	((and (null list1) (null list2)) nil)
	((null list1) list2)
	((null list2) list1)
	((and (car list1) (car list2))
	 (let ((result (copy-list list2)))
	   (dolist (e1 list1 result)
	     (pushnew e1 result :test function))))
	((car list2) list2)
	((car list1) list1)
	(t (cons nil nil))))


;;
;; Deletes a new control rule
;; Deleting mark can be either t, nil, or forever (see induction.lisp
;; for explanation, at the beginning)
;;
(defun delete-cr (cr node le deleting-mark)
  (declare (type p4::control-rule cr)
	   (type decision-node node)
	   (type learning-episode le))
  (setf *deleted-rule-p* t)
  (incf *number-deleted-rules*)
  (setf *cr-list*
	(remove-nil-from-rules (remove cr *cr-list* :test #'eq-name)))
  (setf *initial-cr*
	(remove-nil-from-rules (remove cr *initial-cr* :test #'eq-name)))
  (format t "Deleting rule ~a~%" (p4::control-rule-name cr))

;;; ************** Jose Luis Heras ***********
  (if (or (eq *probabilities-p* 'rules) 
	  (eq *probabilities-p* 'both))
      (borra-regla-prob cr node nil))
;;; ************** Fin Jose Luis Heras *********

  (when le
    (setf (learning-episode-deletedp le) deleting-mark)
    (if node
	(setf (decision-node-rules node)
	      (remove cr (decision-node-rules node) :test #'eq-name)))))


;;;
;;; High level function for deleting a rule
;;;
(DEFUN delete-rule (rule &optional (deleting-mark 'forever))
  (declare (type p4::control-rule rule))
  (if (not (p4::control-rule-p rule))
      (setq rule (from-name-to-rule-1 rule)))
  (delete-cr rule (find-decision-tree-node rule)
	     (gethash (p4::control-rule-name rule) *cr-to-le-hash*)
	     deleting-mark))


(DEFUN eq-name (cr1 cr2)
  (declare (type p4::control-rule cr1 cr2))
  (eq (p4::control-rule-name cr1)
      (p4::control-rule-name cr2)))


;;
;; It returns a control rule in version4 format if evalp is t. If stream is
;; different than nil, it also saves the definition in a file
;;
(defun save-prodigy-cr (cr stream)
  (declare (type p4::control-rule cr))
  (let ((preconds (cdr (p4::control-rule-if cr)))
	(effects (p4::control-rule-then cr)))
    (format stream "~2%(control-rule ~a" (p4::control-rule-name cr))
    (format stream "~%  (if (and ~(~a~)" (car preconds))
    (dolist (precond (cdr preconds))
      (format stream "~%           ~(~a~)" precond))
    (format stream "))")
    (if (and (= (length effects) 4)
	     (eq (caddr effects) 'bindings))
	(format stream "~%  (~(~a ~a ~a~%        ~a~)))"
		(car effects) (cadr effects) (caddr effects) (cadddr effects))
	(format stream "~%~(~a~))" effects))))

;;;***************************************************************
;; Function that refines rules, checking whether they are subsumed or equal
;; IMPORTANT: I'll have to refine this because the learning-episodes
;;
(defun refine-rules (&optional (cr-file (cr-file)) (refined-file (cr-file)) 
			       (domain-name *domain-name*))
  (domain domain-name :load-domain nil :function :compile)
  (when (probe-file cr-file)
    (load cr-file)
    (load-domain)
    (setf *cr-list* nil)
    (setf *number-deduced-rules* 0)
    (setf *number-induced-rules* 0)
    (setf *number-refined-rules* 0)
    (setf *number-deleted-rules* 0)
    (setf *number-undeleted-rules* 0)
    (dolist (control-rules-access *learned-rules-types*)
      (let ((cr-list nil))
	(dolist (control-rule (funcall control-rules-access
				       *current-problem-space*))
	  (let ((includedp (include-rule control-rule cr-list)))
	    (if (or (eq includedp 'include)
		    (eq includedp 'delete-2))
		(push control-rule cr-list))))))
    (when *cr-list*
      (with-open-file (stream refined-file :direction :output
;; 			      :element-type 'character
			      :if-exists :new-version
			      :if-does-not-exist :create)
	(format stream ";;; Refined rules from the file ~a" cr-file)
	(format stream "~%(in-package \"USER\")")
;;	 (sort *cr-list* #'string-lessp :key #'(lambda (cr)
;;						 (control-rule-name cr)))
	(dolist (frame-rc *cr-list*)
	  (save-prodigy-cr frame-rc stream)))
      t)))


;;; ***************************************************************
;;;    Some useful functions
;;; ***************************************************************

;; Prints the search tree
(defun pp-rtt (&optional (tree (find-node 2)) (depth -1)
			 only-success-path-p (stream t))
  (let ((learning-type (node-learning-type tree))
	(type (node-type tree)))
    (format stream (format nil "~~%~~~dt~~a ----> ~~a ~~a ----> ~~a"
		      (* 2 depth))
	    (node-info tree type)
	    type (p4::nexus-name tree)
	    learning-type)
    (when (or (not only-success-path-p) (eq learning-type 'success))
      (dolist (child (p4::nexus-children tree))
	(pp-rtt child (1+ depth) only-success-path-p stream)
	(case type
	  (goal-node
	   (dolist (other (p4::goal-node-ops-left tree))
	     (format stream (format nil "~~%~~~dt~~a --> ~~a --> ~~a"
			       (* 2 (1+ depth)))
		     (p4::operator-name other) 'operator-node 'untried)))
	  ((binding-node applied-op-node)
	   (dolist (other (p4::a-or-b-node-applicable-ops-left tree))
	     (format stream (format nil "~~%~~~dt~~a --> ~~a --> ~~a"
			       (* 2 (1+ depth)))
		     other 'applicable-op-node 'untried))
	   (dolist (other (p4::a-or-b-node-goals-left tree))
	     (format stream (format nil "~~%~~~dt~~a --> ~~a --> ~~a"
			       (* 2 (1+ depth)))
		     other 'goal-node 'untried)))
	  (operator-node 
	   (dolist (instantiated-op (p4::operator-node-bindings-left tree))
	     (format stream (format nil "~~%~~~dt~~a --> ~~a --> ~~a"
			       (* 2 (1+ depth)))
		     (mapcar #'(lambda (var cte)
				 (cons var cte))
			     (p4::operator-vars (p4::instantiated-op-op
						 instantiated-op))
			     (p4::instantiated-op-values instantiated-op))
		     'binding-node 'untried)))
	  (t
	   (dolist (other (p4::goal-node-ops-left tree))
	     (format stream (format nil "~~%~~~dt~~a --> ~~a --> ~~a"
			       (* 2 (1+ depth)))
		     (p4::operator-name other) 'operator-node 'untried))))))))

;;
;; Tree is the node from where to start
;; best-child-info-p if t will print the best children of each node, their
;;                   length to the best solution from them...
;; only-success-path-p if t will print only the success nodes and the failure
;;                   nodes that diverge one step from the success paths
;; only-tried-p if t will print only the expanded nodes
;; max-depth is the depth to which it should print the tree
;;
(defun pp-tree (&key (tree (find-node 2))
		     (best-child-info-p nil)
		     (only-success-path-p t)
		     (only-tried-p t)
		     (max-depth 100))
  (short-pp-rtt best-child-info-p only-success-path-p only-tried-p
		(if (p4::nexus-p tree)
		    tree
		    (find-node tree))
		max-depth -1 t))


(defun short-pp-rtt (best-child-info-p only-success-path-p only-tried-p
				       &optional (tree (find-node 2))
				       (max-depth 100)
				       (depth -1)
				       (stream t))
  (let ((learning-type (if (eq (node-learning-type tree) 'success) 's 'f))
	(type (node-type tree))
	(best-child-info (node-best-child-info tree)))
    (cond (best-child-info-p
	   (format stream (output-string type depth) (node-info tree type)
		   learning-type)
	   (format stream " ~a ~(~a~)"
		   (p4::nexus-name tree)
		   (cons (mapcar #'(lambda (best-child)
				     (node-info best-child
						(node-type best-child)))
				 (car best-child-info))
			 (caddr best-child-info))))
	  (t (format stream (output-string type depth)
		     (node-info tree type) (p4::nexus-name tree))
	     (format stream " ~(~a ~a~)" learning-type (caddr best-child-info))))
    (when (or (not only-success-path-p) (eq learning-type 's))
      ;; the reverse is to print in the expansion order
      (dolist (child (reverse (p4::nexus-children tree)))
	(if (< depth max-depth)
	    (short-pp-rtt best-child-info-p only-success-path-p only-tried-p
			  child max-depth (1+ depth) stream))
	(if (not only-tried-p)
	    (case type
	      (goal-node
	       (dolist (other (p4::goal-node-ops-left tree))
		 (format stream (output-string (node-type other) (1+ depth))
			 (p4::operator-name other) 'untried)))
	      ((binding-node applied-op-node)
	       (dolist (other (p4::a-or-b-node-applicable-ops-left tree))
		 (format stream (output-string (node-type other) (1+ depth))
			 other 'untried))
	       (dolist (other (p4::a-or-b-node-goals-left tree))
		 (format stream (output-string (node-type other) (1+ depth))
			 other 'untried)))
	      (operator-node 
	       (dolist (instantiated-op (p4::operator-node-bindings-left tree))
		 (format stream (output-string (node-type instantiated-op) (1+ depth))
			 (mapcar #'(lambda (var cte)
				     (cons var cte))
				 (p4::operator-vars (p4::instantiated-op-op
						     instantiated-op))
				 (p4::instantiated-op-values instantiated-op))
			 'untried)))
	      (t
	       (dolist (other (p4::goal-node-ops-left tree))
		 (format stream (output-string (node-type other) (1+ depth))
			 (p4::operator-name other) 'untried)))))))))

(DEFUN output-string (type depth)
  (if (eq type 'applied-op-node)
      (format nil "~~%~~~dt~~a ~~(~~a~~)" (* 2 (1+ depth)))
      (format nil "~~%~~~dt~~(~~a ~~a~~)" (* 2 (1+ depth)))))
      
;; Returns the information associated to the node
(defun node-info (node type)
  (case type
    ((operator-node applied-op-node) (get-operator-name node type))
    (goal-node (get-visible-form (p4::goal-node-goal node)))
    (binding-node (cons (get-operator-name node type)
			(convert-objects-in-constants
			 (mapcar #'cdr (get-bindings-from-node node)))))
    (t nil)))

(defun short nil
  (short-pp-rtt nil t t))
				  
;; Prints information about a certain node
(defun pp-rtn (node learning-type)
  (let ((type (node-type node)))
    (format t "~%~a Type of node: ~a  Contents: ~a  Learning type: ~a"
	    (p4::nexus-name node)
	    type (node-info node type) learning-type)))

(DEFUN learning-points (&optional (node (find-node 2)))
  (let* ((info (node-best-child-info node))
	 (lps (if (cadr info) (list node))))
    (if info
	(dolist (child (p4::nexus-children node) lps)
	  (let ((lp (learning-points child)))
	    (if lp
		(setq lps (nconc lps lp)))))))) 

;;;
;;; Funcion para devolver tres letras del mes
;;;
(defun month-string (mes)
  (nth (1- mes) '("Ene" "Feb" "Mar" "Abr" "May" "Jun" "Jul" "Ago" "Sep" "Oct" 
		  "Nov" "Dec")))

;;
;; Saves the control rules on a file given by the global variable *cr-file*
;;
(DEFUN save-cr nil
  (let ((new-crs (remove-duplicates (if (or *deleted-rule-p* *bootstrapp*)
					(append *cr-list* *initial-cr*)
				        *cr-list*))))
    (if new-crs
	(with-open-file (stream (cr-file) :direction :output
;; 				:element-type 'character
				:if-exists (if (or *deleted-rule-p* *bootstrapp*)
					       :new-version
					       :append)
				:if-does-not-exist :create)
          (format stream "~2%;;; New control rules of problem ~a" *problem-name*)
          (format stream "~%(in-package \"USER\")")
	  (if *printp*
	      (format t "~%New control rules learned:"))
	  (dolist (cr new-crs)
	    (save-prodigy-cr cr stream)
	    (if *printp* (pp-cr cr)))))
    (if *break-aprender-p* (break))
    (setf *cr-list* (if (or *deleted-rule-p* *bootstrapp*)
			new-crs
		        (delete-duplicates (append new-crs *initial-cr* nil)
					   :test #'eq)))
    (setf *rules* *cr-list*)))

;;; prior way of computing new number of rules:
;;; 		   (- (length *cr-list*) *number-induced-rules*)

(DEFUN print-learning-data (learning-time)
  (with-open-file (stream *hamlet-output*
			  :direction :output :if-exists :append
			  :if-does-not-exist :create)
     (format stream "~%        (learning-~:@(~a~) ~d ~d ~d ~d ~d ~f)"
	     *problem-name* *number-deduced-rules*
	     *number-induced-rules* *number-refined-rules*
	     *number-deleted-rules* *number-undeleted-rules*
	     learning-time)))
;;;
;;; Simpler version for saving all rules in *cr-list*
;;;
(DEFUN save-rules (&optional (file (cr-file)))
  (declare (pathname file))
  (with-open-file (stream file :direction :output
;; 			  :element-type 'character
			  :if-exists :new-version
			  :if-does-not-exist :create)
     (format stream "~2%;;; New control rules of problem ~a" *problem-name*)
     (format stream "~%(in-package \"USER\")")
     (if (and *printp* *cr-list*)
	 (format t "~%New control rules learned:"))
     (setf *cr-list* (remove-duplicates *cr-list* :test #'eq))
     (setf *rules* *cr-list*)
     (dolist (new-cr *cr-list*)
       (save-prodigy-cr new-cr stream)
       (if *printp* (pp-control-rule new-cr)))))


;;
;; Reads from a file and prints the control rules on that file
;;
(DEFUN read-and-pprint (&optional (file (cr-file)))
  (with-open-file (stream file :direction :input :if-does-not-exist nil)
      (if stream
	  (do ((regla (read stream nil 'eof) (read stream nil 'eof)))
	      ((eq regla 'eof))
	    (pprint regla)))))

;;
;; Prints a control rule in a reasonable way
;;
(defun pp-cr (&optional (cr (car (all-control-rules))))
  (format t "~2%Name: ~a" (p4::control-rule-name cr))
    (format t "~%Preconds:")
    (dolist (precond (cdr (p4::control-rule-if cr)))
      (format t "~%~a" precond))
;; (format t "~2%Main preconds:~50tPreconds:")
;;    (do ((mp (get-slot cr 'main-preconds) (cdr mp))
;;	 (p (cdr (get-slot cr 'preconds)) (cdr p)))
;;	((null (or mp p)))
;;	(format t "~%~a~50t~a" (car mp) (car p)))
    (format t "~2%Effects: ~a" (p4::control-rule-then cr)))

;; If list-format-p then it will return a list of solutions in a visible lisp format. If not, it prints in stream
;; If all-solutions-p then it will return/print all solutions from a problem
(defun pp-solution (&key (list-format-p nil) (stream t) (all-solutions-p nil))
  (let ((solutions (if all-solutions-p
		       (prodigy-result-solutions *prodigy-result*)
		       (list (prodigy-result-solution *prodigy-result*)))))
    (cond (list-format-p (mapcar #'(lambda (solution)
				     (get-visible-solution solution))
				 solutions))
	  (t (terpri stream)
	     (dolist (solution solutions)
		 (pp-list (get-visible-solution solution) 1 stream)
	       (terpri stream))))))

;;;***************************************************************
;;;  Meta predicates
;;;

;;;
;;;  The new version tests for membership of any of the goals to the
;;;  pending goals. If a goal of goals is nil, it tests whether the
;;;  pending goals are nil
;;;  Now, it does not return t as soon as one binding is ok, but
;;;  returns all possible bindings for the pending goals

(DEFUN some-candidate-goals (goals binding)
  (declare (special *current-node* *current-problem-space*))
  (let ((goals2 (get-my-pending-goals *current-node* 'mixed))
	(bound-constants (mapcar #'cdr binding))
	(result nil)
	(positive-result nil))
    (if goals
	(if (p4::strong-is-var-p goals)
	    (list (list (cons goals goals2)))
	  (dolist (goal goals (or result positive-result))
	    (let ((bindings (if goal
				(find-matching goal goals2 bound-constants)
			      (null goals2))))
	      (when bindings
		(setq positive-result t)
		(if (consp bindings)
		    (dolist (new-binding bindings)
		      (pushnew new-binding result :test #'eq-binding-p)))))))
        (null goals2))))


(DEFUN find-matching (goal goals bound-constants)
  (let ((result nil)
	(positive-result nil))
    (dolist (goal2 goals (or result positive-result))
      (let ((matching (and (eq (car goal) (car goal2))
			   (my-matching (cdr goal) (cdr goal2)
					bound-constants nil))))
	(when matching
	  (setq positive-result t)
	  (if (consp matching)
	      (pushnew matching result :test #'eq-binding-p)))))))

(DEFUN my-matching (goal1 goal2 bound-constants new-binding)
  (cond ((and (null goal1) (null goal2))
	 (or new-binding t))
	((or (null goal1) (null goal2)) nil)
	((p4::is-variable-p (car goal1))
	 (if (not (member (car goal2) bound-constants :test #'eq))
	     (let ((new-bind-pair (cons (car goal1) (car goal2))))
	       (my-matching (sublis (list new-bind-pair) (cdr goal1))
			    (cdr goal2)
			    (cons (car goal2) bound-constants)
			    (cons new-bind-pair new-binding)))))
	((eq (car goal1) (car goal2))
	 (my-matching (cdr goal1) (cdr goal2) bound-constants new-binding))
	(t nil)))


;;;
;;; It tests whether the goals are a subset of the pending goals
;;;
(DEFUN subset-candidate-goals (goals)
  (declare (special *current-node* *current-problem-space*))
  (if goals
      (let ((goals1 (mapcar #'(lambda (goal)
				(cons (car goal)
				      (convert-objects-in-constants
				       (cdr goal))))
			    goals))
	    (goals2 (get-my-pending-goals *current-node* t)))
	(subsetp goals1 goals2 :test #'equal))
      t))

;;
;; Tests whether goal is one of the pending goals. Note that the goals
;; are not there yet, and I cache them. I also cache the target goals
;; that matched with the bindings that produced them so that
;; prior-goal can use them. Note that this meta-predicate is always
;; called from a a-or-b-node. Otherwise, it would have to have an if
;; like the one below
(DEFUN target-goal (goal)
  (declare (special *current-node* *current-problem-space*))
  (let ((bindings-prior-goals
	 (my-match-candidate-goal goal
				  (get-my-pending-goals *current-node* nil))))
    (setf (getf (p4::nexus-plist *current-node*) 'prior-goal)
	  (cdr bindings-prior-goals))
    (car bindings-prior-goals)))

;;;
;;; Adaptation of the match-candidate-goal of prodigy. It returns a
;;; list of (bindings . prior-goals)
;;; match-candidate-goal cannot return t. Only bindings or nil.
;;;
(DEFUN my-match-candidate-goal (exp goals)
  (let ((bindings nil)
	(prior-goals nil))
    (dolist (goal goals)
      (let* ((negatedp (member (car exp) '(~ not)))
	     (literal-goal (car (if negatedp
				    (p4::literal-neg-goal-p goal)
				  (p4::literal-goal-p goal))))
	     (binding-node (and literal-goal
				(p4::instantiated-op-binding-node-back-pointer
				 literal-goal)))
	     (match-res (if (or (not literal-goal)
				binding-node)
			    (if negatedp
				(instance-of goal (second exp))
			        (instance-of goal exp)))))
	(if match-res
	    (let* ((prior-goal-p
		    (if (eq (p4::nexus-name binding-node) 4)
			(let ((instantiated-preconds
			       (p4::binding-node-instantiated-preconds
				binding-node)))
			  (if (atom instantiated-preconds)
			      (and (eq goal instantiated-preconds)
				   goal)
			      (car (member goal instantiated-preconds))))))
		   (prior-goal (if prior-goal-p
				   (get-visible-form prior-goal-p)
				 (get-prior-goal binding-node))))
	      (if (listp (car prior-goal))
		  (dolist (pg prior-goal)
		    (pushnew pg prior-goals :test #'equal))
		(pushnew prior-goal prior-goals :test #'equal))
	      ;; match-res might be t sometimes, when no variables in
	      ;; target concept
	      (if (eq match-res t)
		  (if (not bindings)
		      (setq bindings (list nil)))
		  (dolist (binding match-res)
		    (pushnew binding bindings :test #'equal-bindings-p)))))))
    (cons (if (member nil bindings :test #'eq)
	      (or (delete-if #'not bindings)
		  t)
	      bindings)
	  prior-goals)))

;;
;; Returns a list of the pending goals. If visible-form-p is t, it
;; will return the visible form. If it is mixed, it will return the
;; visible form, but objects will be pointers to Prodigy objects,
;; instead of constants. Otherwise, it will return the pointer
;; to the structure.
;;
(DEFUN get-my-pending-goals (node visible-form)
  (let ((pending-goals nil))
    (dolist (goal (cache-pending-goals node) (nreverse pending-goals))
      (if (not (p4::goal-loop-p node goal))
	  (push (if visible-form
		    (if (eq visible-form t)
			(get-visible-form goal)
		        (get-mixed-visible-form goal))
		    goal)
		pending-goals)))))


;;;
;;; I init the slot at the beggining of the code for matching all the
;;; control-rules.
;;;
(DEFUN cache-pending-goals (node)
  (let ((pending-goals (getf (p4::nexus-plist node) 'pending-goals)))
    (or pending-goals
        (setf (getf (p4::nexus-plist node) 'pending-goals)
	      (p4::give-me-all-pending-goals node)))))


(DEFUN applicable-op (op)
  (declare (special *current-node* *current-problem-space*))
  (let ((applicable-op
	 (and (p4::a-or-b-node-p *current-node*)
	      (get-applicable-op (car (p4::a-or-b-node-applicable-ops-left
				       *current-node*))))))
    (if (and applicable-op
	     (eq (car op) (car applicable-op)))
	(if (cdr op)
	    (list (mapcar #'(lambda (var cte)
			      (cons var
				    (if (numberp cte)
					cte
					(p4::object-name-to-object
					 cte *current-problem-space*))))
			  (cdr op)
			  (cdr applicable-op)))
	    t))))


(DEFUN prior-goal (goals)
  (declare (special *current-node*))
  (if (p4::a-or-b-node-p *current-node*)
      (get-bindings-prior-goals-a-or-b goals)
      (get-bindings-prior-goals-g-or-op goals *current-node*)))

(DEFUN get-bindings-prior-goals-g-or-op (goals node)
  (let ((prior-goal (get-prior-goal node))
	(result nil)
	(positive-result nil))
    (dolist (goal goals (or result positive-result))
      (let ((bindings (prior-goal-single goal prior-goal)))
	(when bindings
	  (setq positive-result t)
	  (if (consp bindings)
	      (if result
		  (dolist (binding bindings)
		    (if binding
			(push binding result)))
		  (setq result bindings))))))))

(DEFUN prior-goal-single (goal prior-goal)
  (declare (special *current-problem-space*))
  (let ((new-binding (get-binding-of-goals (convert-objects-in-constants goal)
					   prior-goal nil)))
    (cond ((null new-binding) nil)
	  ((equal new-binding '(nil)) t)
	  (t (list (convert-names-into-objects new-binding))))))

(DEFUN get-prior-goal (node)
  (declare (type p4::nexus node))
  (or (getf (p4::nexus-plist node) 'prior-goal)
      (cache-prior-goal node)))

;;
;; Returns the prior-goal: the first goal node that is a top level goal
;;; I'm not sure whether the car will work alright, but...????
;;; IMPORTANT: all goals in the path will be prior goals, so the
;;; transfer will be immediate if we consider all at once when
;;; generating the rule. In that case, we might also have negative
;;; examples.
(DEFUN cache-prior-goal (node)
  (declare (type p4::nexus node))
  (setf (getf (p4::nexus-plist node) 'prior-goal)
	(if (p4::goal-node-p node)
	    (let ((literal (p4::goal-node-goal node)))
	      (if (top-level-goal-p node literal)
		  (get-visible-form literal)
		  (let* ((grandpa (get-grandpa
				   (car (p4::goal-node-introducing-operators
					 node))))
			 (literal-grandpa (p4::goal-node-goal grandpa)))
		    (if (top-level-goal-p grandpa literal-grandpa)
			(get-visible-form literal-grandpa)
		        (get-prior-goal grandpa)))))
	    (get-prior-goal (p4::nexus-parent node)))))

;;
;; Returns the prior-goal node: the first goal node that is a top level goal
;;; I'm not sure whether the car will work alright, but...????
(DEFUN get-prior-goal-node (node)
  (declare (type p4::nexus node))
  (if (p4::goal-node-p node)
      (let ((literal (p4::goal-node-goal node)))
	(if (top-level-goal-p node literal)
	    node
	    (let ((grandpa (get-grandpa
			    (car (p4::goal-node-introducing-operators node)))))
	      (if (top-level-goal-p grandpa literal)
		  grandpa
		  (get-prior-goal-node grandpa)))))
      (get-prior-goal-node (p4::nexus-parent node))))

;;
;; Handles the case when the current-node is a-or-b.
;;
(DEFUN get-bindings-prior-goals-a-or-b (prior-goals)
  (declare (special *current-node*))
  (let ((result nil)
	(positive-result nil)
	(prior-goals-of-tg (if (eq (p4::nexus-name *current-node*) 4)
			       (get-my-pending-goals *current-node* t)
			       (get-prior-goal *current-node*))))
    (dolist (prior-goal prior-goals (or result positive-result))
      (dolist (prior-goal-of-tg prior-goals-of-tg)
	(let ((bindings (prior-goal-single prior-goal prior-goal-of-tg)))
	  (when bindings
	    (setq positive-result t)
	    (if (consp bindings)
		(if result
		    (dolist (binding bindings)
		      (if binding
			  (push binding result)))
		    (setq result bindings)))))))))


;;;
;;; Changes all names in the bindings for their respective objects pointers
;;;
(DEFUN convert-names-into-objects (binding)
  (declare (special *current-problem-space*))
  (mapcar #'(lambda (bind-pair)
	      (cons (car bind-pair)
		    (if (or (p4::is-variable-p (cdr bind-pair))
			    (numberp (cdr bind-pair)))
			(cdr bind-pair)
		      (p4::object-name-to-object
		       (cdr bind-pair)
		       *current-problem-space*))))
	  binding))


;;; I will add it to the control rules for testing for different
;;; values for the variables ???
(DEFUN diff-objects (&rest objects)
  (not (apply #'eq objects)))

(DEFUN first-goal (goal-var)
  (list (list (cons goal-var (mapcar #'(lambda (literal)
					 (get-visible-form literal))
				     (candidate-goals))))))

;; New meta-predicate needed to test whether a goal is a previous goal of the node we are in.
;; I would have to test it.
(DEFUN previous-goal (goal)
  (declare (special *current-node*))
  (if (p4::a-or-b-node-p *current-node*)
      (get-bindings-previous-goal-a-or-b goal)
      (get-bindings-previous-goal-g-or-op goal *current-node*)))

(DEFUN get-bindings-previous-goal-a-or-b (previous-goal)
  (declare (special *current-node*))
  (let ((result nil)
	(positive-result nil)
	(previous-goals-of-tg (if (eq (p4::nexus-name *current-node*) 4)
				  (get-my-pending-goals *current-node* t)
				(get-previous-goals *current-node*))))
    (dolist (previous-goal-of-tg previous-goals-of-tg
	      (or result positive-result))
      (let ((bindings (prior-goal-single previous-goal previous-goal-of-tg)))
	(when bindings
	  (setq positive-result t)
	  (if (consp bindings)
	      (if result
		  (dolist (binding bindings)
		    (if binding
			(push binding result)))
		(setq result bindings))))))))

(DEFUN get-previous-goals (node)
  (declare (type p4::nexus node))
  (or (getf (p4::nexus-plist node) 'previous-goals)
      (cache-previous-goals node)))

(DEFUN cache-previous-goals (node)
  (declare (type p4::nexus node))
  (setf (getf (p4::nexus-plist node) 'previous-goals)
    (if (p4::goal-node-p node)
	(let ((literal (p4::goal-node-goal node)))
	  (if (top-level-goal-p node literal)
	      (list (get-visible-form literal))
	    (cons (get-visible-form literal)
		  (get-previous-goals
		   (get-grandpa (car (p4::goal-node-introducing-operators node)))))))
      (get-previous-goals (p4::nexus-parent node)))))

(DEFUN get-bindings-previous-goal-g-or-op (goal node)
  (let ((result nil)
	(positive-result nil))
    (dolist (previous-goal (get-previous-goals node) (or result positive-result))
      (let ((bindings (prior-goal-single goal previous-goal)))
	(when bindings
	  (setq positive-result t)
	  (if (consp bindings)
	      (if result
		  (dolist (binding bindings)
		    (if binding
			(push binding result)))
		(setq result bindings))))))))

;; Redefinition from gen-from-pred so that it works for non numeric vars
(defun a-gen-from-pred (literal &rest vars)
  (if (> (count-if #'(lambda (x) (p4::strong-is-var-p x)) vars) 1)
      (error "~% Too many unbound vars in ~S ~a" literal vars)
    (let ((res (true-in-state (cons literal vars))))
      (or (eq res t)
	  (mapcar #'(lambda (x) (cdar x)) res)))))


;;; ****************************************************************
;;; Other things

(DEFUN son (node)
  (car (p4::nexus-children node)))

(DEFUN nieto (node)
  (son (son node)))

(DEFUN equal-rule-p (rule1 rule2 test-inclusion-p)
  (if (not (p4::control-rule-p rule1))
      (setq rule1 (from-name-to-rule-1 rule1)))
  (if (not (p4::control-rule-p rule2))
      (setq rule2 (from-name-to-rule-1 rule2)))
  (equal-control-rule-p (get-real-preconds rule1) (get-real-preconds rule2)
			(get-real-effects rule1) (get-real-effects rule2)
			test-inclusion-p))

;;;
;;; Standard unification algorithm according to Rich&Knight
;;;
(DEFUN unify (exp1 exp2)
  (if (or (atom exp1) (atom exp2))
      (cond ((eq exp1 exp2) nil)
	    ((p4::is-variable-p exp1)
	     (list (cons exp1 exp2)))
	    ((p4::is-variable-p exp2)
	     (list (cons exp2 exp1)))
	    (t 'fail))
      (if (eq (car exp1) (car exp2))
	  (if (= (length exp1) (length exp2))
	      (let ((substitution nil))
		(do ((e1 exp1 (cdr e1))
		     (e2 exp2 (cdr e2)))
		    ((or (null e1)
			 (eq substitution 'fail))
		     substitution)
		  (let ((binding (unify (car e1) (car e2))))
		    (cond ((eq binding 'fail)
			   (setq substitution 'fail))
			  (t (setq e1 (sublis binding e1))
			     (setq e2 (sublis binding e2))
			     (setq substitution (append binding
							substitution)))))))
	      'fail)
	  'fail)))

;;;
;;; Standard unification algorithm according to Rich&Knight, modified
;;; so that Prodigy variables are constants
;;;
(DEFUN new-unify (exp1 exp2)
  (cond ((or (atom exp1) (atom exp2))
	 (if (eq exp1 exp2) nil 'fail))
	((and (listp (car exp1)) (listp (car exp2)))
	 (let ((binding (new-unify (car exp1) (car exp2))))
	   (if (eq binding 'fail)
	       'fail
	       (new-unify (sublis binding (cdr exp1))
			  (sublis binding (cdr exp2))))))
	((eq (car exp1) (car exp2))
	 (if (= (length exp1) (length exp2))
	     (let ((substitution nil))
	       (do ((e1 exp1 (cdr e1))
		    (e2 exp2 (cdr e2)))
		   ((or (null e1)
			(eq substitution 'fail))
		    substitution)
		 (let ((binding (new-unify (car e1) (car e2))))
		   (cond ((eq binding 'fail)
			  (setq substitution 'fail))
			 (t (setq e1 (sublis binding e1))
			    (setq e2 (sublis binding e2))
			    (setq substitution
				  (append binding substitution)))))))
	     'fail))
	(t 'fail)))

(defun pr (n)
  (pp-control-rule (nth n *rules*)))

(defun ple (n)
  (pp-le1 (gethash (p4::control-rule-name (nth n *rules*))
		   *cr-to-le-hash*)))


;;; ****************************************************************
;;;            Functions for execution of EvoCK rules
;;; ****************************************************************

(defun evock-filter (predicate sequence)
  "Devuelve una lista con aquellos elementos de la lista que cumplen el predicado"
  (let (filtered_list)
    (dolist (x sequence)
            (when (funcall predicate x) (push x filtered_list)))
    (values (reverse filtered_list))))

(defun get_list_of_operators_and_lazy_inference_rules ()
    ""
    (values
     (append
      ;; cdr needed to eliminate op-finish
      (p4::problem-space-operators *current-problem-space*)
      (p4::problem-space-lazy-inference-rules *current-problem-space*))))

(defun give_operator (op)
  "Returns an operator given its name"
  (let* ((operator_list (get_list_of_operators_and_lazy_inference_rules))
         (operator (find op operator_list
                         :test (lambda (x y) (eq x (p4::operator-name y))))))
    (values operator)))

(defun return-operator-check (tree)
  "Returns the goals that need this operator to fulfill themselves"

  ;;  (values (cdr (nth 2 (p4::operator-precond-exp (give_operator tree)))))

  (values
   (mapcar #'cadr
           ;; Filter only 'add' effects
           (evock-filter #'(lambda (x) (eq (car x) 'add)) 
                   (cdr (nth 2 (p4::operator-effects (give_operator tree))))
                   ))))

(defun appropriate-operator (operator)
  (if 
    (dolist (z (return-operator-check operator))
      (when (current-goal z) (return T)))
    T
    nil))
;;   (some #'(lambda (z)
;; 	    (current-goal z))
;; 	(return-operator-check operator)))

;; It returns all initial goals
(defun give-me-nice-initial-goals nil
  (or (getf (p4::nexus-plist (find-node 2)) :initial-goals)
      (setf  (getf (p4::nexus-plist (find-node 2)) :initial-goals)
	(let ((goals (p4::binding-node-instantiated-preconds (find-node 4))))
	  (mapcar #'(lambda (goal)
		      (get-visible-form goal))
		  (if (listp goals)
		      (if (eq (car goals) 'and)
			  (cdr goals)
			goals)
		    (list goals)))))))
