;;; This file contains the code for doing induction over the control rules.
;;; this version induces first over the induces rules, and then over
;;; the deduced ones

(in-package "USER")



;;; ******************************************************************
;;;   functions for building the initial decision tree
;;; ******************************************************************
;;;
;;; for now, i didn't build the reject and prefer parts.
;;;
(defun build-fixed-decision-tree nil
  (let ((root (make-decision-node :name 'root :children-type 'effect)))
    (setf (decision-node-parent root) root)
    (setf *decision-tree* root)
    (let* ((root-children-1 (build-children root (cons 'sub-goal
						       (if *learn-apply-p*
							   (list 'apply)))
					    'prior-goal nil))
	   (root-children (build-children root '(select reject prefer)
					  'effect))
	   (operators (get-all-operators))
	   (goals (get-all-goals))
	   (select-children-1 (build-children (first root-children)
					      '(bindings) 'current-operator))
	   (select-children (build-children (first root-children)
					    '(goals operators)
					    'effect))
	   (select-goals-children (build-children (first select-children)
						  goals 'prior-goal nil))
	   (select-operators-children (build-children (second select-children)
						      operators 'current-goal))
	   (select-bindings-children (build-children (first select-children-1)
						     operators 'current-goal))
	   (select-nietos-operators (build-nietos select-operators-children
						  goals 'prior-goal))
	   (select-nietos-bindings (build-nietos select-bindings-children
						 goals 'prior-goal))
	   (reject-children-1 (build-children (second root-children)
					      '(bindings) 'current-operator))
	   (reject-children (build-children (second root-children)
					    '(goals operators)
					    'effect))
	   (reject-goals-children (build-children (first reject-children)
						  goals 'prior-goal nil))
	   (reject-operators-children (build-children (second reject-children)
						      operators 'current-goal))
	   (reject-bindings-children (build-children (first reject-children-1)
						     operators 'current-goal))
	   (reject-nietos-operators (build-nietos reject-operators-children
						  goals 'prior-goal))
	   (reject-nietos-bindings (build-nietos reject-bindings-children
						 goals 'prior-goal))
	   (prefer-children-1 (build-children (third root-children)
					      '(bindings) 'current-operator))
	   (prefer-children (build-children (third root-children)
					    '(goals operators)
					    'effect))
	   (prefer-goals-children (build-children (first prefer-children)
						  goals 'prior-goal nil))
	   (prefer-operators-children (build-children (second prefer-children)
						      operators 'current-goal))
	   (prefer-bindings-children (build-children (first prefer-children-1)
						     operators 'current-goal))
	   (prefer-nietos-operators (build-nietos prefer-operators-children
						  goals 'prior-goal))
	   (prefer-nietos-bindings (build-nietos prefer-bindings-children
						 goals 'prior-goal))
	   (leafs (cons 'any-goal (if *divide-by-prior-goal-p* goals))))
      (build-nietos select-goals-children leafs nil)
      (build-nietos select-nietos-operators leafs nil)
      (build-nietos select-nietos-bindings leafs nil)
      (build-nietos reject-goals-children leafs nil)
      (build-nietos reject-nietos-operators leafs nil)
      (build-nietos reject-nietos-bindings leafs nil)
      (build-nietos prefer-goals-children leafs nil)
      (build-nietos prefer-nietos-operators leafs nil)
      (build-nietos prefer-nietos-bindings leafs nil)
      (build-children (first root-children-1) leafs nil nil)
      (if *learn-apply-p*
	  (build-children (second root-children-1) leafs nil nil))
      (values))))
	   
	   

(defun build-children (node alternatives children-type
			    &optional (fixedp t))
  (declare (type decision-node node)
	   (atom children-type fixedp)
	   (list alternatives))
;;  (format t "~%creating children of ~a~% alternatives: ~a~% children type: ~a"
;;	  (decision-node-name node) alternatives children-type)
  (setf (decision-node-children node)
	(append (mapcar #'(lambda (alternative)
			    (make-decision-node
			     :name alternative
			     :fixedp fixedp
			     :children-type children-type
			     :children nil
			     :parent node))
			alternatives)
		(decision-node-children node))))

(defun build-nietos (children alternatives children-type)
  (declare (atom children-type)
	   (list children alternatives))
  (let ((nietos nil))
    (dolist (child children nietos)
      (setq nietos (append (build-children child alternatives children-type
					   (not (eq children-type
						    'prior-goal)))
			   nietos)))))



;;;******************************************************************
;;;  general functions for inducing new rules
;;;******************************************************************

;;
;; if slow-induction-p is t, it will print the results in the same
;; list as it is printing. if not, it should save them separetely
;;
(defun learn-from-induction (slow-induction-p induced-file control-rules
					      savep initp test-negatives-p)
  (declare (string induced-file)
	   (list control-rules)
	   (atom slow-induction-p savep initp test-negatives-p))
  (init-induction initp)
  (if savep (format t "~%***************** Inducing...~%"))
  (let ((t-inicial (get-internal-run-time)))  
    (dolist (cr control-rules)
      (let ((preconds (get-real-preconds cr))
	    (effects (get-real-effects cr)))
	(classify-rule cr preconds effects *decision-tree* t
		       test-negatives-p)))
    (setf *cr-file* induced-file)
    (save-induction (elapsed-time t-inicial) slow-induction-p savep)))
  

(defun init-induction (initp)
  (declare (atom initp))
  (setf *top-type* (p4::type-name-to-type :top-type *current-problem-space*))
  (setf *cr-list* nil)
;; 2-8-94  
;;  (setf *number-induced-rules* 0)
;;  (setf *number-deleted-rules* 0)
  (if initp (clear-decision-tree *decision-tree*)))

(defun clear-decision-tree (node)
  (declare (type decision-node node))
  (setf (decision-node-rules node) nil)
  (dolist (child (decision-node-children node))
    (clear-decision-tree child)))


;;; ******************************************************************
;;;   functions that traverse the tree according to the new rule
;;; ******************************************************************

;;;
;;; if storep is t, the chain of functions will try to store it, and,
;;; therefore, induce over it. if nil, it will return the leaf node of
;;; the decision tree where it belongs
;;;
(defun classify-rule (cr preconds effects node storep test-negatives-p)
  (declare (type p4::control-rule cr)
	   (type decision-node node)
	   (atom storep test-negatives-p)
	   (list preconds effects))
  (if (eq (decision-node-children-type node) 'effect)
      (classify-effect cr preconds effects node storep test-negatives-p)
      (classify-preconds cr preconds effects node storep test-negatives-p)))



(defun classify-preconds (cr preconds effects node storep test-negatives-p)
  (declare (type p4::control-rule cr)
	   (type decision-node node)
	   (atom storep test-negatives-p)
	   (list preconds effects))
  (let* ((decision (decision-node-children-type node))
	 (fixedp (or (not (eq decision 'prior-goal))
		     (decision-node-fixedp node))))
    (if fixedp
	(let* ((precond-found (find-if #'(lambda (precond)
					   (eq (car precond) decision))
				       preconds))
	       (rest-preconds (if precond-found
				  (remove precond-found preconds
					  :test #'equal)))
	       (decision-names (if precond-found
				   (case decision
				     (current-operator (cdr precond-found))
				     (current-goal
				      (list (car (get-pos-assertion
						  (cadr precond-found)))))
				     (prior-goal
				      (mapcar #'(lambda (goal)
						  (car (get-pos-assertion
							goal)))
					      (cadr precond-found)))
				     (t 'error))))
	       (childs-found (if precond-found
				 (mapcar #'(lambda (name)
					     (find-decision-child name node))
					 decision-names))))
	  (if childs-found
	      (if storep
		  (dolist (child childs-found)
		    (if child
			(classify-preconds cr rest-preconds effects child
					   storep test-negatives-p)))
		  (classify-preconds cr rest-preconds
				     effects (find-if #'(lambda (child)
							  child)
						      childs-found)
				     storep test-negatives-p))))
        (let ((child (find-decision-child 'any-goal node)))
	  (if storep
	      (classify-rest cr child test-negatives-p)
	      child)))))

		     
(defun find-decision-child (decision-name node)		     
  (declare (type decision-node node)
	   (atom decision-name))
  (find-if #'(lambda (child)
	       (eq (decision-node-name child) decision-name))
	   (decision-node-children node)))


(defun classify-effect (cr preconds effects node storep test-negatives-p)
  (declare (type p4::control-rule cr)
	   (type decision-node node)
	   (atom storep test-negatives-p)
	   (list preconds effects))
  (let ((effect-found (find-if #'(lambda (child)
				   (eq (decision-node-name child)
				       (if (listp (car effects))
					   (caar effects)
					   (car effects))))
			       (decision-node-children node))))
    (if effect-found
	(if (eq (car effects) 'bindings)
	    (classify-rule cr preconds effects effect-found storep
			   test-negatives-p)
	    (classify-rule cr preconds (cdr effects) effect-found storep
			   test-negatives-p))
        'error)))


;;; ******************************************************************
;;;   functions for induction on the leafs
;;; ******************************************************************


(defun classify-rest (cr node test-negatives-p)
  (declare (type p4::control-rule cr)
	   (type decision-node node)
	   (atom test-negatives-p))
  (if (not (decision-node-needs-language-change-p node))
      (if (and test-negatives-p
	       (includes-negative-examples-p (get-real-preconds cr)
					     (get-real-effects cr)
					     (decision-node-negatives node)))
	  (let* ((rule-name (p4::control-rule-name cr))
		 (le (gethash rule-name *cr-to-le-hash*)))
	    (format t "Marking ~a as deleted~%" rule-name)
	    (incf *number-deleted-rules*)
	    (if (not (recover le node nil))
		(setf (learning-episode-deletedp le) 'forever)))
	  (let ((rules (get-decision-node-rules node)))
	    (if (or (not rules)
		    (not (induce-over-induced cr (get-real-preconds cr)
					      (get-real-effects cr)
					      node rules)))
		(store-deduced-rule cr node))))))

;;;
;;; in case it couldn't induce, it stores the rule on the node, if it
;;; doesn't match any negative example.
;;;
(defun store-deduced-rule (cr node)
  (declare (type p4::control-rule cr)
	   (type decision-node node))
  (let ((rule-name (p4::control-rule-name cr)))
    (if (eq (learning-episode-type (gethash rule-name *cr-to-le-hash*))
	    'induced)
	(format t "Readding ~a~%" rule-name)
        (format t "Adding ~a~%" rule-name))
;;; ********** Principio de Jose Luis Heras ************

    (if (or (eq *probabilities-p* 'rules) 
	    (eq *probabilities-p* 'both))
	(when (not (get-number-rule rule-name))
	  (if (eq (car (p4::control-rule-then cr)) 'then)
	      (nueva-regla-prob rule-name (cdr (p4::control-rule-then cr)))
	      (nueva-regla-prob rule-name (p4::control-rule-then cr)))))

;;; ********* Final de Jose Luis Heras **************
	     
    (if (>= (length (push cr (decision-node-rules node)))
	    (decision-node-threshold node))
	(setf *over-threshold-p* t))
    (push cr *cr-list*)))


;;
;; returns t if there has been an intersection, and it creates an
;; induced rule if so.
;; rules-bindings is of the form: ((rule1 rule2 other-bindings) ...)
;; last-binding has the last one, which it is going to be used for inducing
;;
(defun induce-over-induced (cr preconds1 effects1 node rules)
  (declare (type p4::control-rule cr)
	   (type decision-node node)
	   (list preconds1 effects1 rules))
  (if rules
      (let* ((rule (car rules))
	     (preconds2 (get-real-preconds rule))
	     (effects2 (get-real-effects rule))
	     (bindings (equal-control-rule-p preconds1 preconds2
					     effects1 effects2 nil))
	     (intersection (some-intersection cr preconds1 effects1
					      rule preconds2 effects2
					      node bindings)))
	(or intersection
	    (induce-over-induced cr preconds1 effects1 node (cdr rules))))))

;;
;; It tries all possible bindings, calling again induce-over-induced
;; and returns as soon as one returns t. I do not use some because I
;; need to use the variable in the form
;; if it finds an intersection but it is equal to any of the rules,
;; then it does not learn the rule, by returning t.
;; Now, it only considers *max-number-bindings* bindings, since it is
;; very costly to consider all possible bindings. I guess, I'll have
;; to check whether at the end there are many repeated rules.
;;
(defun some-intersection (cr preconds1 effects1 rule preconds2 effects2
			     node bindings)
  (declare (type p4::control-rule cr rule)
	   (type decision-node node)
	   (list preconds1 effects1 preconds2 effects2 bindings))
  (do* ((rest-bindings (if (<= (length bindings) *max-number-bindings*)
			   bindings
			   (select-n-random bindings *max-number-bindings*))
		       (cdr rest-bindings))
	(relevant-intersection-p nil))
      ((or (null rest-bindings)
	   relevant-intersection-p)
       (if (eq relevant-intersection-p 'deleted-old)
	   nil
	   relevant-intersection-p))
    (let* ((new-rule (find-relevant-intersection preconds1 preconds2
						 cr effects1 effects2
						 (car rest-bindings) node t))
	   (new-intersection (car new-rule)))
      (if new-intersection
	  (if (eq new-intersection 'equal)
	      (cond ((cdr new-rule)
		     (let* ((rule-name (p4::control-rule-name rule))
			    (le (gethash rule-name *cr-to-le-hash*))
			    (deducedp (eq (learning-episode-type le)
					  'deduced)))
		       ;; if the rule is a deduced one we cannot
		       ;; delete it forever, since when re-installing
		       ;; we might need it
		       (delete-cr rule node le (if deducedp t 'forever))
		       (setq relevant-intersection-p 'deleted-old)))
		    (t
		     (let* ((rule-name (p4::control-rule-name cr))
			    (le (gethash rule-name *cr-to-le-hash*)))
		       (format t "Equal rule ~a deleted~%" rule-name)

;;; ************** Jose Luis Heras ***********
		       (if (or (eq *probabilities-p* 'rules) 
			       (eq *probabilities-p* 'both))
			   (borra-regla-prob cr nil rule))
;;; ************** Fin Jose Luis Heras *********
		       
		       (incf *number-deleted-rules*)
		       (setf *deleted-rule-p* t)
		       (if node
			   (setf (decision-node-rules node)
				 (remove cr (decision-node-rules node) :test #'eq-name)))
		       (setq relevant-intersection-p
			     (setf (learning-episode-deletedp le) 'forever)))))
	      (let* ((new-effects (cadr new-rule))
		     (new-bindings (cddr new-rule))
		     (new-rules-bindings (list cr rule (cdr rest-bindings))))
		(setq relevant-intersection-p t)
		(learn-induced-rule new-rules-bindings new-intersection
				    new-effects new-bindings node t)))))))

#|
;; previous version
		(or (induce-over-induced cr new-intersection
					 new-effects node
					 (get-decision-node-rules node)
					 new-rules-bindings
					 new-binding)
		    (learn-induced-rule new-rules-bindings new-intersection
					new-effects new-binding node t))))))))


;; older version
 (learning-episode-negatives (gethash (p4::control-rule-name rule)
				      *cr-to-le-hash*))
		(or (induce-over-induced cr new-intersection
					 (sublis new-bindings ir-effects)
					 node
					 (get-decision-node-rules node)
					 new-rules-bindings)
		    (learn-induced-rule new-rules-bindings new-intersection
					(sublis new-bindings ir-effects)
					node t))))))))
|#

;;; It extracts n random elements from list
(defun select-n-random (list n)
  (let ((new-list nil))
    (dotimes (i n new-list)
      (let ((elem (nth (random (length list)) list)))
	(setq list (remove elem list :test #'equal))
	(push elem new-list)))))

;;;
;;; finds a relevant intersection by checking whether the intersection
;;; is different than the previous preconds and if it doesn't include
;;; negative examples. it returns:
;;;  (new-preconds new-effects new-binding new-types-binding)
;;; if it finds that the intersection is equal to any one of the
;;; previous preconds, it returns: (equal equal1p), where equal1p
;;; reflects whether it was the equal to the first one or not.
;;;
(defun find-relevant-intersection (preconds1 preconds2 rule1 effects1
					     effects2 binding
					     node check-equality-p)
  (declare (type p4::control-rule rule1)
	   (type decision-node node)
	   (atom check-equality-p)
	   (list preconds1 effects1 preconds2 effects2 binding))
  (let* ((intersection-bindings (intersection-and-bindings preconds1 preconds2
							   binding))
	 (intersection (car intersection-bindings))
	 (new-types-binding (cdr intersection-bindings))
	 (new-binding (merge-bindings binding new-types-binding))
	 (new-effects (sublis new-binding (get-real-effects rule1)))
	 (new-intersection (remove-old-types intersection
					     (get-all-vars intersection
							   new-effects)))
	 (new-rule (list new-intersection new-effects new-binding
			 new-types-binding)))
    (if new-intersection
	(let* ((equal1p (and (equal new-intersection (sublis new-binding
							     preconds1))
			     (equal-effects-p effects1 effects2 new-binding)))
	       (equal2p (or equal1p
			    (and (equal-set-p (sublis new-types-binding
						      preconds2)
					      new-intersection)
				 (equal-effects-p effects1 effects2
						  new-types-binding)))))
	  (if (or equal1p equal2p)
	      (if check-equality-p
		  (cons 'equal equal1p))
	      (if (not (includes-negative-examples-p
			new-intersection new-effects
			(decision-node-negatives node)))
		  new-rule))))))

#|
			  (equal-set-p (sublis (reverse-bindings
						(list new-binding))
					       preconds2)
				       new-intersection))))
|#

;;;
;;; Merges two binding lists, in such a way that the result will be as
;;; a sequential application of both bindings
;;;
(DEFUN merge-bindings (binding1 binding2)
  (declare (list binding1 binding2))
  (let ((result binding1))
    (dolist (bind-pair binding2 result)
      (let ((existsp (member (car bind-pair) result :test #'eq :key #'cdr)))
	(cond (existsp
	       (setq result (remove (car existsp) result :test #'equal))
	       (push (cons (caar existsp) (cdr bind-pair)) result))
	      (t (pushnew bind-pair result :test #'equal)))))))

;;
;; It returns a cons of the common preconds and the new bindings due
;; to new names of variables (type generalized)
;; IMPORTANT: Next, we have to compute the set of all-possible
;; prior-goals for a certain goal, and if the set on any-prior-goal is
;; equal to that one, then delete this precond.
;;
;; I changed into union in the some-candidate-goals because that is
;; the way to do. I also changed the some-candidate-goals
;; meta-predicate so that it tests whether any of the goals is a
;; pending goal
;;
;; It creates new var names, so that unification is easier
;;
;; When the number of some-candidate-goals is greater than a ratio of the
;; number of predicates in the domain, it removes the some-candidate-goals
;; metapredicate
;;
;;
(DEFUN intersection-and-bindings (plain-preconds1 preconds2 binding)
  (declare (list plain-preconds1 preconds2 binding))
  (setf *rule-id* (string-trim "#:G" (gensym)))
  (setf *number-vars-in-cr* 0)
  (let ((preconds1 (sublis binding plain-preconds1))
	(new-binding (create-new-binding binding))
	(common-preconds nil)
	(new-prior-goal nil)
	(new-some-candidate-goals nil)
	(candidate-goals1 nil)
	(prior-goal1 nil)
	(rest-preconds2 preconds2)
	(common-vars nil)
	(number-true-in-state 0))
    (dolist (precond1 preconds1)
      (setq preconds2 rest-preconds2)
      (do* ((p2 preconds2 (cdr p2))
	    (equalp (equal (car p2) precond1)
		    (equal (car p2) precond1)))
	  ((or (null p2) equalp)
	   (cond (equalp
		  (push precond1 common-preconds)
		  (if (eq (car precond1) 'true-in-state)
		      (incf number-true-in-state))
		  (setq rest-preconds2
			(remove (car p2) rest-preconds2 :test #'equal)))
		 (t (case (car precond1)
		      (prior-goal
		       (setq prior-goal1 precond1))
		      (some-candidate-goals
		       (setq candidate-goals1 precond1))
		      (type-of-object
		       (let* ((var (cadr precond1))
			      (supertype (common-supertype preconds1
							   rest-preconds2
							   (cons var var))))
			 (if supertype
			     (let ((new-var
				    (induced-var-name supertype)))
			       (push var common-vars)
			       ;; To delete previous assignations of var
			       (setq new-binding
			             (delete-if #'(lambda (bind-pair)
						    (eq (car bind-pair) var))
						new-binding))
			       (push (cons var new-var) new-binding)
			       (push `(type-of-object ,var ,supertype)
				     common-preconds))
			     (push precond1 common-preconds))))
		      (t nil)))))))

;;    (format t "~2%Preconds1: ~a~2%Rest Preconds2: ~a~2%common-preconds: ~a"
;;	    preconds1 rest-preconds2 common-preconds)
;; 	(true-in-state
;;	 (push (cadr precond2) states))

    (when (or (not *check-number-states*)
	      (>= number-true-in-state *state-threshold*))
      (dolist (precond2 rest-preconds2)
	(case (car precond2)
	  (prior-goal
	   (let* ((goals1 (if prior-goal1 (cadr prior-goal1)))
		  (goals2 (cadr precond2))
		  (union (union goals1 goals2 :test #'equal))
		  (all-vars (mapcar #'car new-binding))
		  (new-vars nil))
	     (cond ((< (length union)
		       (* *max-ratio-some-goals*
			  (length (p4::problem-space-all-preds
				   p4::*current-problem-space*))))
		    (dolist (goal union)
			(dolist (var (cdr (get-pos-assertion goal)))
			    (if (not (member var all-vars :test #'eq))
				(pushnew var new-vars :test #'eq))))
;;	     (setq new-vars (set-difference new-vars all-vars))
		    (setq new-binding
			  (append new-binding
				  (create-new-binding
				   (mapcar #'(lambda (new-var)
					       (cons new-var new-var))
					   new-vars))))		  
		    (setq new-prior-goal `(prior-goal ,union)))
		   (t (setq new-prior-goal 'none)))))
	  (some-candidate-goals
	   (let* ((goals1 (if candidate-goals1 (cadr candidate-goals1)))
		  (goals2 (cadr precond2))
		  (all-vars (mapcar #'car new-binding))
		  (union (my-union goals1 goals2))
		  (new-vars nil))
	     (cond ((< (length union)
		       (* *max-ratio-some-goals*
			  (length (p4::problem-space-all-preds
				   p4::*current-problem-space*))))
		    (dolist (goal union)
		      (dolist (var (cdr (get-pos-assertion goal)))
			(if (not (member var all-vars :test #'eq))
			    (pushnew var new-vars :test #'eq))))
;;	     (setq new-vars (set-difference new-vars all-vars))
		    (setq new-binding
			  (append new-binding
				  (create-new-binding
				   (mapcar #'(lambda (new-var)
					       (cons new-var new-var))
					   new-vars))))
		    (setq new-some-candidate-goals
			  `(some-candidate-goals ,union)))
		   (t (setq new-some-candidate-goals 'none)))))
	  (type-of-object
	   (if (not (member (cadr precond2) common-vars))
	       (push precond2 common-preconds)))
	  (t nil)))
      
      (cond ((eq new-prior-goal 'none)
	     (setq new-prior-goal nil))
	    ((and prior-goal1 (not new-prior-goal))
	     (setq new-prior-goal prior-goal1))
	    (t nil))
      (cond ((eq new-some-candidate-goals 'none)
	     (setq new-some-candidate-goals nil))
	    ((and candidate-goals1 (not new-some-candidate-goals))
	     (setq new-some-candidate-goals candidate-goals1))
	    (t nil))
      
      (build-induced-preconds new-binding common-preconds
			      (sublis new-binding new-prior-goal)
			      (sublis new-binding new-some-candidate-goals)))))

;;; it was push `(subset-candidate-goals...) and intersection instead
;;; of union

(DEFUN induced-var-name (type)
  (declare (atom type))
  (incf *number-vars-in-cr*)
  (intern (format nil "~:@(<I-~a-~d-~d>~)" type
		  *rule-id* *number-vars-in-cr*)))


;;
;; It reverses the preconds list, substitutes the old variables for
;; the new ones (after inducing), and introduces in their right place
;; the new preconds of subset-candidate-goals and any-prior-goal in
;; case they are created.
;;
(DEFUN build-induced-preconds (new-binding common-preconds prior-goal
					   some-candidate-goals)
  (declare (list new-binding common-preconds prior-goal some-candidate-goals))
  (let ((result nil)
	(introducedp nil))
    (dolist (common-precond common-preconds (cons result new-binding))
      (let ((precond (sublis new-binding common-precond)))
	(if (and (member (car precond)
			 '(current-goal candidate-goal target-goal))
		 prior-goal)
	    (push prior-goal result))
	(when (and (not introducedp)
		   (eq (car precond) 'true-in-state)
		   some-candidate-goals)
	  (setq introducedp t)
	  (push some-candidate-goals result))
	(push precond result)))))

;;;
;;; it was testing for eq with subset-candidate-goals and not
;;; some-candidate-goals.



#|
;;
;; Checking whether the length is equal 1 is to avoid learning
;; too general rules. Checking whether they are the same pos, if they
;; have different signs is equal to testing if a OR ~a is t.
;;
(DEFUN complete-set-p (uncommon1 uncommon2)
  (declare (list uncommon1 uncommon2))
  (and (= 1 (length uncommon1) (length uncommon2))
       (let* ((goal1 (car uncommon1))
	      (goal2 (car uncommon2))
	      (negated1p (negatedp goal1))
	      (negated2p (negatedp goal2))
	      (pos1 (get-pos-assertion goal1))
	      (pos2 (get-pos-assertion goal2)))
	 (cond ((or (and negated1p negated2p)
		    (and (not negated1p) (not negated2p)))
    		(complete-state-p pos1 pos2 *domain-complete-goal-pairs*))
	       ((equal pos1 pos2) t)
	       (t nil)))))

(DEFUN complete-state-p (state1 state2 completeness-set)
  (find-if #'(lambda (set)
	       (and (or (and (eq (car state1) (car set))
			     (eq (car state2) (cadr set)))
			(and (eq (car state1) (cadr set))
			     (eq (car state2) (car set))))
		    (funcall (caddr set) state1 state2)))
	   completeness-set))
  
(DEFUN same-first-var (state1 state2)
  (eq (cadr state1) (cadr state2)))

|#

;; (all-vars (get-all-vars new-preconds new-effects))
;; (remove-old-types new-preconds all-vars)
(DEFUN learn-induced-rule (crs-bindings new-preconds new-effects new-bindings
					node inducedp)
  (declare (type decision-node node)
	   (atom inducedp)
	   (list crs-bindings new-preconds new-effects new-bindings))
  (setq new-preconds (sort-preconds new-preconds))
  (let* ((new-name (induced-name new-effects
				 (if (eq (cadr new-effects) 'bindings)
				     (cadar new-preconds))))
	 (new-cr (p4::make-control-rule
		  :name new-name
		  :if (cons 'and new-preconds)
		  :then (cons 'then new-effects)))
	 (cr1 (car crs-bindings))
	 (cr2 (cadr crs-bindings))
	 (binding1 (car new-bindings))
	 (binding2 (cadr new-bindings))
	 (le1 (gethash (p4::control-rule-name cr1) *cr-to-le-hash*))
	 (le2 (gethash (p4::control-rule-name cr2) *cr-to-le-hash*))
	 (prior-goal1 (learning-episode-prior-goal le1))
	 (prior-goal2 (learning-episode-prior-goal le2))
	 (prior-goals (generate-induced-prior-goals binding1 prior-goal1
						    binding2 prior-goal2))
	 (all-prior-goals (mapcar #'cadr prior-goals))
	 (new-learning-episode (make-learning-episode
				:name (get-new-le-name)
				:type (if inducedp 'induced 'refined)
				:rule new-cr
				:problem *problem-name*
				:from nil
				:other-bindings (if (caddr crs-bindings)
						    crs-bindings)
				:binding-used (list (cons cr1 binding1)
						    (cons cr2 binding2))
				:prior-goal (create-new-prior-goals
					     prior-goals new-preconds)
				:learning-steps
				  (1+ (max (learning-episode-learning-steps
					    le1)
					   (learning-episode-learning-steps
					    le2)))
				:preconds-pool nil
				:matching-time
				  (+ (learning-episode-matching-time le1)
				     (learning-episode-matching-time le2))
				:matchings
				  (+ (learning-episode-matchings le1)
				     (learning-episode-matchings le2))
				:num-nodes
				  (/ (+ (learning-episode-num-nodes le1)
					(learning-episode-num-nodes le2))
				     2.0)
				:firings
				  (+ (learning-episode-firings le1)
				     (learning-episode-firings le2)))))
    (push new-learning-episode *le-list*)
    (update-frequencies (add-prior-goals all-prior-goals new-preconds)
			node t)
    (setf (gethash new-name *cr-to-le-hash*) new-learning-episode)
    (push new-cr *cr-list*)
    (format t "~%Inducing ~a~%" new-name)

;;; ********* Jose Luis Heras *********

    (when (or (eq *probabilities-p* 'rules) 
	      (eq *probabilities-p* 'both))
;; 	  (format t "~%*~%*~%*~%*~%*~%*~%Learn-induce-rule: ~a" new-effects)
      (nueva-regla-prob new-name new-effects))

;;; ********** Fin Jose Luis Heras ************

    (if *printp*
	(pp-control-rule new-cr))
    (incf *number-induced-rules*)
    (setf *deleted-rule-p* t)
    (push new-cr (decision-node-rules node))
    (if inducedp
	(update-decision-tree node crs-bindings new-learning-episode))
    ;; I hope this will solve the problem of not inducing over all
    ;; stored rules
    (induce-over-induced new-cr new-preconds new-effects node
			 (remove new-cr (decision-node-rules node)))
    new-learning-episode))

;; Daniel. Added on March 2001
;; Auxiliary function for computing new prior-goals of the induced rule from
;; the ones of previous rules. If there was a problem on a previous step
;; placing a nil in a rule, it removes the prior goals given that we are not
;; using them anymore.
(defun generate-induced-prior-goals (binding1 prior-goal1 binding2
					      prior-goal2)
  (if (and (listp prior-goal1)
	   (car prior-goal1)
	   (listp prior-goal2)
	   (car prior-goal2))
      (remove-duplicates (append (sublis binding1 prior-goal1)
				 (sublis binding2 prior-goal2))
			 :test #'equal)))
;;
;; Updates the decision-tree and the learning episodes of the old
;; rules from where it induced. The test for negatives used to be
;; equal-set-p
;;
(DEFUN update-decision-tree (node crs-binding new-learning-episode)
  (declare (type decision-node node)
	   (type learning-episode new-learning-episode)
	   (list crs-binding))
  (let ((from-crs nil))
    (dolist (old-cr (nreverse (butlast crs-binding)) (terpri))
      (unless (member old-cr from-crs)
	(push old-cr from-crs)
	(let ((old-le (gethash (p4::control-rule-name old-cr)
			       *cr-to-le-hash*)))
	  (push old-le (learning-episode-from new-learning-episode))
	  (delete-cr old-cr node old-le t))))))

;; Not needed now (after the second push before)
;;	  (if (and (eq (learning-episode-type old-le) 'induced)
;;		   (member old-cr *cr-list* :test #'eq-name))
;;	      (decf *number-induced-rules*))

;;;
;;; Updates the frequencies of appearance of preconds of the state and
;;; prior goal, as attributes, for the later computation of the
;;; information gain.
;;; It supposses that run has already been executed after the last
;;; call to domain. Otherwise, not all predicates of the domain will
;;; be there.
;;; Negative goals will be handled as positive goals.
;;;
(DEFUN update-frequencies (preconds node positivep)
  (declare (type decision-node node)
	   (atom positivep)
	   (list preconds))
  (do* ((conds preconds (cdr conds))
	(stop-p nil)
	(frequency-table-positive
	 (or (decision-node-frequency-table-positive node)
	     (setf (decision-node-frequency-table-positive node)
		   (make-hash-table))))
	(frequency-table-negative
	 (or (decision-node-frequency-table-negative node)
	     (setf (decision-node-frequency-table-negative node)
		   (make-hash-table))))
	(frequency-table-prior-goal-positive
	 (or (decision-node-frequency-table-prior-goal-positive node)
	     (setf (decision-node-frequency-table-prior-goal-positive node)
		   (make-hash-table))))
	(frequency-table-prior-goal-negative
	 (or (decision-node-frequency-table-prior-goal-negative node)
	     (setf (decision-node-frequency-table-prior-goal-negative node)
		   (make-hash-table)))))
      ((or (null conds) stop-p))
;;    (print (car conds))
    (case (caar conds)
      (true-in-state
       (let ((predicate (car (get-pos-assertion (cadar conds)))))
	 (if positivep
	     (if (gethash predicate frequency-table-positive)
		 (incf (gethash predicate frequency-table-positive))
		 (setf (gethash predicate frequency-table-positive) 1))
	     (if (gethash predicate frequency-table-negative)
		 (incf (gethash predicate frequency-table-negative))
		 (setf (gethash predicate frequency-table-negative) 1)))))
      (prior-goal
       (dolist (goal (cadar conds))
	 (let ((predicate (car (get-pos-assertion goal))))
	   (if positivep
	       (if (gethash predicate frequency-table-prior-goal-positive)
		   (incf
		    (gethash predicate frequency-table-prior-goal-positive))
		   (setf
		    (gethash predicate frequency-table-prior-goal-positive)
		    1))
	       (if (gethash predicate frequency-table-prior-goal-negative)
		   (incf
		    (gethash predicate frequency-table-prior-goal-negative))
		   (setf
		    (gethash predicate frequency-table-prior-goal-negative)
		    1))))))
      ((some-candidate-goals type-of-object)
       (setq stop-p t))
      (t nil))))

;; 	(preds (p4::problem-space-all-preds p4::*current-problem-space*))
;; 	(number-preds (length preds))
;; 	     (setf (decision-node-frequency-table node)
;; 		   (make-array (list (* 2 number-preds) 2)
;; 			       :initial-element 0
;; 			       :element-type 'integer))

;;;
;;; Adds the prior-goals to the preconds if they are not already
;;; there. Prior-goals is a list of just the prior-goals, without the rules
;;; 
(DEFUN add-prior-goals (prior-goals preconds)
  (declare (list prior-goals preconds))
  (let ((new-prior-goals (some #'(lambda (precond)
				   (if (prior-goal-meta-pred-p precond)
				       (copy-list (cadr precond))))
			       preconds)))
    (dolist (cond prior-goals)
      (dolist (goal (cadr cond))
	(pushnew goal new-prior-goals :test #'equal)))
    (if new-prior-goals
	(replace-prior-goal preconds (list 'prior-goal new-prior-goals))
        preconds)))


(DEFUN replace-prior-goal (preconds new-prior-goals)
  (declare (list new-prior-goals preconds))
  (cond ((null preconds) (list new-prior-goals))
	((eq (caar preconds) 'prior-goal)
	 (cons new-prior-goals (cdr preconds)))
	((member (caar preconds) (list 'true-in-state 'some-candidate-goals
				       'type-of-object))
	 (cons new-prior-goals preconds))
	(t (cons (car preconds) (replace-prior-goal (cdr preconds)
						    new-prior-goals)))))

;;
;; It adds the types of the variables that go away with the induction
;;
(DEFUN create-new-prior-goals (prior-goals new-preconds)
  (declare (list prior-goals new-preconds))
  (mapcar #'(lambda (cr-prior-goal-type-def)
	      (let ((cr (car cr-prior-goal-type-def))
		    (prior-goal (cadr cr-prior-goal-type-def))
		    (type-defs (caddr cr-prior-goal-type-def)))
		(list cr prior-goal
		      (get-types-of-prior-goal prior-goal new-preconds
					       (append (get-real-preconds cr)
						       type-defs)))))
	  prior-goals))
  
(DEFUN get-new-le-name nil
  (let ((new-name (read-from-string (format nil "LE-~d" *le-index*))))
    (incf *le-index*)
    new-name))


(DEFUN induced-name (effects operator)
  (declare (list effects)
	   (atom operator))
;;  (setf *rule-id* (string-trim "#:G" (gensym)))
  (intern (format nil "~:@(induced-~a~)" (name-cr effects operator))))

(DEFUN my-subset-p (rule1 rule2)
  (declare (list rule1 rule2))
  (equal-control-rule-p (car rule1) (car rule2)
			(cdr rule1) (cdr rule2) 'subset))


;;
;; Given the preconds and effects of a control rule, it returns all
;; vars that appear in all preconds and effects, except for the ones
;; in the type-of-object meta-predicate
;;
(DEFUN get-all-vars (preconds effects)
  (declare (list preconds effects))
  (let ((all-vars nil))
    (dolist (precond preconds)
      (case (car precond)
	((target-goal current-goal candidate-goal true-in-state applicable-op)
	 (dolist (var (cdadr precond))
	   (pushnew var all-vars :test #'eq)))
	((prior-goal some-candidate-goals)
	 (dolist (goal (cadr precond))
	   (dolist (var (cdr goal))
	     (pushnew var all-vars :test #'eq))))
	(t nil)))
    (case (cadr effects)
      (bindings (dolist (bind-pair (caddr effects) all-vars)
		  (pushnew (cdr bind-pair) all-vars :test #'eq)))
      (goals (dolist (var (cdaddr effects) all-vars)
	       (pushnew var all-vars :test #'eq)))
      (t all-vars))))

;;
;; It removes the type-of-object meta-predicates that do not have one
;; of the vars in other precond or effect
(DEFUN remove-old-types (preconds all-vars)
  (declare (list preconds all-vars))
  (remove-if #'(lambda (precond)
		 (and (eq (car precond) 'type-of-object)
		      (not (member (cadr precond) all-vars :test #'eq))))
	     preconds))


;;; ******************************************************************
;;;       Repairing over generalization functions
;;; ******************************************************************
;;;
;;; Tries to find whether there has been an overgeneralization. If
;;; there was not, then it returns nil. If it finds one, then it
;;; repairs the overgeneralization by finding the node in which one of
;;; the wrong rules were fired, and refining them. Returns t if there
;;; was a repair.
;;; Shorterp tells whether the old-tree was searched with
;;; shorter-or-equal (t), or strict-shorter (nil). If t, it helps a
;;; lot on finding more negative examples.
;;;
(DEFUN repairp (old-tree new-tree shorterp)
  (declare (type p4::nexus old-tree new-tree)
	   (atom shorterp))
  (setf *refined-target-concepts* nil)
  (if (all-control-rules)
      (repair-unoptimal-rules old-tree new-tree (get-internal-run-time)
			      shorterp nil)))

;;;
;;; If the best-child of old-tree is a child of the new tree, then it
;;; refines all rules that had a greater number of steps to achieve
;;; the solution than that node, given that it did not find before
;;; that the best child was a worse solution than the old solution.
;;; Also, if it finds that the best child of the old-tree is not a
;;; child of the new-tree, it refines all rules applied.
;;; If unoptimal-path-p is t, it is following an unoptimal-path. In
;;; that case, it changes the behavior of find-rules-nodes
;;;
(DEFUN repair-unoptimal-rules (old-tree new-tree t-inicial shorterp
					unoptimal-path-p)
  (declare (type p4::nexus old-tree new-tree)
	   (atom shorterp)
	   (number t-inicial))
  (let* ((node-type (node-type old-tree))
	 (old-best-child-info (node-best-child-info old-tree))
	 (old-best-childs (car old-best-child-info))
	 (old-number-ops (if old-best-childs
			     (caddr (node-best-child-info
				     (car old-best-childs))))))
;;    (format t "~%childs: ~a" old-best-childs)
;;    (format t "~%childs: ~{~%  ~a~}" old-best-childs)
;;    (format t "~%number ops: ~a" old-number-ops)
    (if (and old-best-childs
	     old-number-ops
	     (> old-number-ops 0))
	(let* ((best-goal-child-p (some #'p4::goal-node-p old-best-childs))
	       (children (p4::nexus-children new-tree))
	       (first-best-child
		(position-if #'(lambda (node)
				 (let ((info (node-best-child-info node)))
				   (and info
					(= (caddr info) old-number-ops))))
			     children :from-end t))
	       (rules-nodes (find-rules-nodes new-tree node-type
					      old-number-ops nil
					      best-goal-child-p
					      unoptimal-path-p))
	       (repairedp (repair-wrong-rules old-tree new-tree t-inicial
					      node-type shorterp
					      old-best-childs rules-nodes)))
	  (dolist (old-best-child old-best-childs repairedp)
	    (setq repairedp
		  (or (let* ((best-child (find-equal-node old-best-child
							  new-tree node-type))
			     (new-number-ops
			      (if best-child
				  (caddr (node-best-child-info best-child))))
			     (best-child-p
			      (and best-child new-number-ops
				   (= new-number-ops old-number-ops))))
			;; we continue searching under the nodes that
			;; were the best ones or were explored before
			;; the first best child in the new-tree
;; 			(format t "~%first-best-child: ~a~%best-child: ~a%"
;; 				first-best-child best-child)
			(if (and best-child
				 (or best-child-p
				     (not first-best-child)
				     (> (position best-child children
						  :from-end t)
					first-best-child)))
			    (repair-unoptimal-rules
			     old-best-child best-child t-inicial shorterp
			     (or unoptimal-path-p
				 (not best-child-p)))))
		      repairedp)))))))


;;
;; Returns t if repair-overgeneralization returns t after trying to
;; repair the wrong rules on rules-nodes
;;
(DEFUN repair-wrong-rules (old-tree new-tree t-inicial node-type
				    shorterp old-best-childs rules-nodes)
  (declare (type p4::nexus old-tree new-tree)
	   (atom node-type shorterp)
	   (list old-best-childs rules-nodes)
	   (number t-inicial))
  (let ((unknown-nodes nil)
	(learning-nodes nil))
    (dolist (rule-node rules-nodes)
      (dolist (node (cdr rule-node))
	(if (not (member node learning-nodes :test #'eq))
	    (if (member node unknown-nodes :test #'eq)
		(delete node (cdr rule-node))
		(cond ((or (and shorterp
				(not (member (find-equal-node node old-tree
							      node-type)
					     old-best-childs
					     :test #'eq)))
			   (and (not shorterp)
				(not (rules-applied-p node))))
		       (push node learning-nodes))
		      (t (push node unknown-nodes)
			 (delete node rule-node)))))))
    (if learning-nodes
	(repair-overgeneralization old-tree new-tree rules-nodes t-inicial))))

;;
;; Returns t if there was at least one rule applied below the node
;;
(DEFUN rules-applied-p (node)
  (declare (type p4::nexus node))
  (or (node-fired-crs node)
      (some #'rules-applied-p (p4::nexus-children node))))

(DEFUN all-fired-crs (node)
  (declare (type p4::nexus node))
  (let* ((fired-crs nil)
	 (fired-in-node (node-fired-crs node))
	 (fired-in-children
	  (dolist (child (p4::nexus-children node) fired-crs)
	    (dolist (fired-cr (all-fired-crs child))
	      (push fired-cr fired-crs)))))
    (if fired-in-node
	(cons (cons (p4::nexus-name node) fired-in-node)
	      fired-in-children)
        fired-in-children)))


;;;
;;; Adds a new rule to the list of rules to be saved
;;;
(DEFUN add-rule (new-rule test-negatives-p)
  (declare (atom test-negatives-p)
	   (list new-rule))
  (let ((all-rules *cr-list*))
    (learn-from-induction t *cr-file* (list new-rule) nil nil
			  test-negatives-p)
    ;; some rules might have been deleted because they are
    ;; more specific than the new rule
    (dolist (rule all-rules *cr-list*)
      (let ((le (gethash (p4::control-rule-name rule) *cr-to-le-hash*)))
	(if (and le (not (learning-episode-deletedp le)))
	    (push rule *cr-list*))))))

;;;
;;; Returns the list of lists of the form (rule child1 ... childn)
;;; that lead to worse solutions than the best one. If all-childs-p is
;;; t, then it returns all rules and childs.
;;; 
;;; If a rule (set of rules) selected a node that failed, or
;;; it was worse than the best, and there were not rules
;;; applied below it, or the old tree was expanded, and the
;;; corresponding node was not one of the best ones, then it
;;; is a negative example.
;;;
;;; If best-goal-child-p is t, it will not consider as negative any sub-goal
;;; rule, since it has at least one success goal node. In the case of
;;; apply rules, since Prodigy (this version) does not consider the
;;; application of several operators, there is no need for that.
;;;
;;; Another learning opportunity for negative examples is when there
;;; is a suboptimal child of the new tree, and there was a rule that
;;; selected that child, and there are not applied rules under the
;;; child (if not, it could have been because of those control rules).
;;;
;;; IMPORTANT: I'll have to work this as high priority.
;;; If the new node was a a-or-b-node AND
;;;    some rules selected to sub-goal AND
;;;    new node does not have children AND
;;;    old-best-child was an application
;;; Then learn a negative example
;;;
;;; This will require some coding, but it is greatly needed
;;;
(DEFUN find-rules-nodes (new-tree node-type old-number-ops all-childs-p
				  best-goal-child-p unoptimal-path-p)
  (declare (type p4::nexus new-tree)
	   (atom node-type all-childs-p best-goal-child-p)
	   (integer old-number-ops))
  (let ((rules-nodes nil)
	(rules-selections (node-fired-crs new-tree))
	(new-children (p4::nexus-children new-tree)))
    (if rules-selections
	(if new-children
	    (dolist (child new-children rules-nodes)
	      (let ((number-ops (caddr (node-best-child-info child)))
		    (child-decision (get-decision child node-type)))
		(if (or all-childs-p
			(and unoptimal-path-p
			     (not number-ops)
			     (not (eq (getf (p4::nexus-plist new-tree)
					    :termination-reason)
				      :depth-exceeded)))
			(and number-ops
			     (or (> number-ops old-number-ops)
				 (zerop number-ops))))
		    (dolist (cr (find-rules-of-decision rules-selections child
							child-decision
							node-type
							best-goal-child-p))
		      (let ((wrong-rule-p (member cr rules-nodes
						  :test #'eq :key #'car)))
			(if wrong-rule-p
			    (push child (cdar wrong-rule-p))
			    (push (list cr child) rules-nodes)))))))
	    ))))

#|
;;It can be that the rule fired, but the problem solver did not choose
;;it. For instance, in a binding node, it might fire a select goals cr,
;;but the problem solver decides to apply instead.

		      (progn (format t "~2%Problem ~a~%child-decision ~a~%"
				     *problem-name* child-decision)
			     (format t "~%All selections: ~a"
				     rules-selections)
			     (read))))))))))
|#

;;
;; Returns the rules (from the rules-selections) that selected the child
;;
(DEFUN find-rules-of-decision (rules-selections child child-decision node-type
						best-goal-child-p)
  (declare (type p4::nexus child)
	   (atom child-decision node-type best-goal-child-p)
	   (list rules-selections))
  (let ((child-type (node-type child))
	(rules nil))
    (dolist (cr-selection rules-selections rules)
      (let ((wrong-rule-p (case (cdr cr-selection)
			    (sub-goal (and (eq child-type 'goal-node)
					   (not best-goal-child-p)))
			    (apply (eq child-type 'applied-op-node))
			    (t (some #'(lambda (selection)
					 (equal-decision-p child selection
							   child-decision
							   node-type))
				     (cdr cr-selection))))))
	(if wrong-rule-p
	    (push (car cr-selection) rules))))))

;;;
;;; It returns t if the selection is equal to the decision made in the
;;; node
;;;
(DEFUN equal-decision-p (node selection decision parent-type)
  (declare (type p4::nexus node)
	   (atom selection decision parent-type))
  (case parent-type
    (operator-node (equal (mapcar #'cdr (convert-names-into-objects decision))
			  (car selection)))
    (goal-node (eq (p4::operator-name selection) decision))
    ((binding-node applied-op-node)
     (if (p4::goal-node-p node)
	 (or (eq selection :goal)
	     (equal (get-visible-form selection) decision))
         (eq selection :apply)))
    (t nil)))
    
;;;
;;; It returns the decision (in a visible form) of the node
;;;
(DEFUN get-decision (node parent-type)
  (declare (type p4::nexus node)
	   (atom parent-type))
  (case parent-type
    (operator-node (get-bindings-from-node node))
    (goal-node (get-operator-name node (node-type node)))
    ((binding-node applied-op-node)
     (if (p4::goal-node-p node)
	 (get-target-goal node)))
    (t nil)))

;;;
;;; It returns the child of new-tree that has the same decision as the
;;; old-child or nil if there is none.
;;;
(DEFUN find-equal-node (old-child new-tree parent-type)
  (declare (type p4::nexus old-child new-tree)
	   (atom parent-type))
  (let ((old-decision (get-decision old-child parent-type)))
    (find-if #'(lambda (child)
		 (case parent-type
		   (operator-node
		    (equal old-decision (get-bindings-from-node child)))
		   (goal-node
		    (eq old-decision
			(get-operator-name child (node-type child))))
		   ((binding-node applied-op-node)
		    (if (p4::goal-node-p old-child)
			(and (p4::goal-node-p child)
			     (equal old-decision (get-target-goal child)))
		        (eq (node-type child) 'applied-op-node)))
		   (t nil)))
	     (p4::nexus-children new-tree))))

	 
;;;
;;; Parent is the node where the decision was made according to a
;;; control rule and was wrong. rules-nodes is a list of lists where
;;; the cars are rules that made the wrong decision, and the cdrs are
;;; the childs that they generated
;;;
(DEFUN repair-overgeneralization (old-parent parent rules-nodes t-inicial)
  (declare (type p4::nexus old-parent parent)
	   (number t-inicial)
	   (list rules-nodes))
  (let ((state (instantiate-cr-state (get-state-for-node parent)))
	(new-negative-p nil)
	(wrong-rule-p t))
    (cond (rules-nodes
	   (dolist (rule-nodes rules-nodes)
	     (let* ((responsible-rule (car rule-nodes))
		    (children (cdr rule-nodes))
		    (decision-made
		     (if responsible-rule
			 (car (get-real-effects responsible-rule))))
		    (neg-rules (get-neg-rules old-parent parent children
					      decision-made state))
		    (rule-name (if responsible-rule
				   (p4::control-rule-name responsible-rule)))
		    (decision-node
		     (if responsible-rule
			 (find-decision-tree-node responsible-rule)))
		    (decision-nodes (if decision-node (list decision-node))))
	       (if (not responsible-rule)
		   (setq wrong-rule-p nil))
	       (dolist (neg-rule neg-rules)
		 (if (not responsible-rule)
		     (pushnew (setq decision-node
				    (classify-rule nil (car neg-rule)
						   (cdr neg-rule)
						   *decision-tree* nil nil))
			      decision-nodes :test #'eq))
		 (let ((subset-p (member neg-rule
					 (decision-node-negatives
					  decision-node)
					 :test #'my-subset-p)))
		   (cond (subset-p
			  (format t "~%Overgeneral negative example found in node ~a~%"
				  (p4::nexus-name old-parent)))
			 (t (push neg-rule
				  (decision-node-negatives decision-node))
			    (setq new-negative-p t)
			    (update-frequencies (car neg-rule)
						decision-node nil)
			    (format t "~%Adding negative example in node ~a~%"
				    (p4::nexus-name old-parent))))))
	       (dolist (refined-decision-node decision-nodes)
		 (when (or new-negative-p
			   (not (member refined-decision-node
					*refined-target-concepts*
					:test #'eq)))
		   (push refined-decision-node *refined-target-concepts*)
		   (when responsible-rule
		     (format t "~%***************** Revising...~%")
		     (format t "~%Refining ~a~%" rule-name))
		   (setf *over-threshold-p* nil)
		   (dolist (rule (cond (*refine-all-p*
					(setf (decision-node-rules refined-decision-node)
					      (remove nil
						      (decision-node-rules
						       refined-decision-node))))
				       (responsible-rule
					(list responsible-rule))
				       (t nil)))
		     (let* ((rule-name-aux (p4::control-rule-name rule))
			    (le (gethash rule-name-aux *cr-to-le-hash*))
			    (negatives
			     (decision-node-negatives refined-decision-node)))
		       (format t "~%Revisiting ~a~%" rule-name-aux)
		       (revisit-rule rule le refined-decision-node negatives
				     (not (eq rule-name rule-name-aux)))))
		   (when *over-threshold-p*
		     (dotimes (i *number-restructures*)
		       (restructure-node refined-decision-node))
		     (incf
		      (decision-node-threshold refined-decision-node)))))))
	   (if wrong-rule-p
	       (save-refinement (elapsed-time t-inicial) t))
	   t)
	  (t nil))))

#|
		   (if (eq rule responsible-rule)
		       (setq refinedp (revisit-rule rule le decision-node
						    negatives nil))
		       (revisit-rule rule le decision-node negatives t))))
	       (if refinedp
		   (when *over-threshold-p*
		     (dotimes (i *number-restructures*)
		       (restructure-node decision-node))
		     (incf (decision-node-threshold decision-node)))
		   (unclassification learning-episode))))
|#


;;;
;;; Tests whether the rule is still ok. If not, it tries to refine it.
;;; We first ask to whether it has been deleted, because it might have
;;; been deleted to be induced from.
;;;
(DEFUN revisit-rule (rule le decision-node negatives test-inclusion-p)
  (declare (type p4::control-rule rule)
	   (type learning-episode le)
	   (type decision-node decision-node)
	   (atom test-inclusion-p)
	   (list negatives))
  (if (and rule le
	   (or (not test-inclusion-p)
	       (includes-negative-examples-p (get-real-preconds rule)
					     (get-real-effects rule)
					     negatives)))
      (if (eq (learning-episode-type le) 'deduced)
	  (or (recover le decision-node nil) t)
	  (refine-rule le decision-node))))


;;;
;;; Reinduces the rules of a certain node, if the number of rules on
;;; that node is greater than its threshold.
;;; I don't know why, I'm getting a nil in between the rules of a
;;; node. Until I discover the bug, I'll delete all nils from the list
;;; of rules.
;;;
(DEFUN restructure-node (node)
  (declare (type decision-node node))
  (let* ((rules (shuffle (remove-nil-from-rules
			  (get-decision-node-rules node))))
	 (all-rules (set-difference *cr-list* rules :test #'eq)))
    (format t "~2%***************** Restructuring a node...~%")
    (setf (decision-node-rules node) nil)
    (learn-from-induction t *cr-file* rules nil nil nil)
    (setf *cr-list* (append *cr-list* all-rules))))

(DEFUN remove-nil-from-rules (rules)
  (remove nil rules))

;;;
;;; Returns a list of negative examples in the form of negative rules:
;;; (preconds . effects)
;;;
(DEFUN get-neg-rules (old-parent parent children decision-made state)
  (declare (type decision-node old-parent parent)
	   (list state children)
	   (atom decision-made))
  (let ((parent-type (node-type parent))
	(prior-goal (if *re-use-prior-goal-p* (get-prior-goal parent)))
	(result nil))
    (dolist (child children result)
      (let* ((neg-rule (give-me-neg-rule old-parent parent child parent-type
					 prior-goal state decision-made))
	     (preconds (cdr neg-rule))
	     (effects (car neg-rule)))
	(when preconds
	  (setf *substitution* nil)
	  (setf *rule-id* (string-trim "#:G" (gensym)))
	  (setf *number-vars-in-cr* 0)
	  (find-substitution neg-rule)
	  (let ((new-constraints (new-constraints)))
	    (if *save-examples-p*
		(save-example-for-foil (append preconds new-constraints)
				       effects nil))
	    (pushnew (cons (append (parametrize preconds) new-constraints)
			   (parametrize effects))
		     result :test #'my-subset-p)))))))


#|
;;;
;;; Parent is the node of the tree with rules that did not have the
;;; unborn-child node.
;;;
(DEFUN get-neg-preconds (parent unborn-child rule state)
  (let* ((parent-type (node-type parent))
	 (prior-goal (if *re-use-prior-goal-p* (get-prior-goal parent)))
	 (preconds (give-me-neg-preconds parent unborn-child parent-type
					 prior-goal state rule)))
    (setf *substitution* nil)
    (setf *number-vars-in-cr* 0)
    (find-substitution preconds)
    (append (parametrize preconds) (new-constraints))))
|#

;;;
;;; It returns the preconds of the rule of the negative example
;;; Decision-made can be either sub-goal, apply, select, or nil (if it is not
;;; learning from the failure of a control rule)
;;;
(DEFUN give-me-neg-rule (old-parent parent child parent-type
				    prior-goal state decision-made)
  (declare (type decision-node old-parent parent child)
	   (atom parent-type decision-made)
	   (list state prior-goal))
  (case parent-type
    (operator-node
     (let* ((operator (get-operator-name parent parent-type))
	    (target-goal (get-target-goal (p4::nexus-parent parent)))
	    (some-candidate-goals (get-some-candidate-goals
				   (get-grandpa old-parent)
				   target-goal)))
       (cons `(select bindings ,(get-bindings-from-node child))
	     `((current-operator ,operator)
	       (current-goal ,target-goal)
	       ,@(if *re-use-prior-goal-p* `((prior-goal (,prior-goal))))
	       ,@state
	       (some-candidate-goals ,some-candidate-goals)))))
    (goal-node
     (let* ((operator (get-operator-name child (node-type child)))
	    (target-goal (get-target-goal parent))
	    (some-candidate-goals (get-some-candidate-goals
				   (p4::nexus-parent old-parent)
				   target-goal)))
       (cons `(select operators ,operator)
	     `((current-goal ,target-goal)
	       ,@(if *re-use-prior-goal-p* `((prior-goal (,prior-goal))))
	       ,@state
	       (some-candidate-goals ,some-candidate-goals)))))
    ((binding-node applied-op-node)
     (let* ((type (node-type child))
	    (subgoalp (eq type 'goal-node))
	    (candidate-goal (if subgoalp
				(get-target-goal child)))
	    (candidate-goals-for-s (get-some-candidate-goals old-parent
							     candidate-goal))
	    (candidate-goals-for-a (get-some-candidate-goals old-parent nil))
	    (applicable-op (get-applied-op parent nil))
	    (best-childs (if (not decision-made)
			     (car (node-best-child-info parent))))
	    (node-types nil))
       (if (not decision-made)
	   (dolist (child best-childs)
	     (push (node-type child) node-types)))
       (cond ((or (and (or (eq decision-made 'sub-goal)
			   (eq decision-made 'apply))
		       applicable-op
		       (or subgoalp *learn-apply-p*))
		  (and (not decision-made)
		       (or (and subgoalp
				(not (member 'goal-node node-types)))
			   (and *learn-apply-p*
				(not subgoalp)
				(not (member 'applied-op-node node-types))))))
	      (cons (list (if subgoalp 'sub-goal 'apply))
		    `((applicable-op ,applicable-op)
		      ,@(if *re-use-prior-goal-p*
			    `((prior-goal (,prior-goal))))
		      ,@state
		      (some-candidate-goals ,candidate-goals-for-a))))
	     ((and candidate-goals-for-s
		   subgoalp
		   (or (eq decision-made 'select)
		       (and (not decision-made)
			    (member 'goal-node node-types))))
	      (cons `(select goals ,candidate-goal)
		    `((target-goal ,candidate-goal)
		      ,@(if *re-use-prior-goal-p*
			    `((prior-goal (,prior-goal))))
		      ,@state
		      (some-candidate-goals ,candidate-goals-for-s))))
	     (t nil))))
    (t nil)))


#|
;; in the call for select goals:
		      ,@(if (and *re-use-prior-goal-p*
				 (not (eq operator 'p4::*finish*)))
			    `((prior-goal (,prior-goal))))

;;; It used to have a (current-operator ,operator) in the goals crs
(operator (get-operator-name parent parent-type))
|#

;;
;; It assumes the rule is wrong.
;; It tries another binding. If any of the bindings works, it learns
;; another rule. If not, it calls refine-parents.
;;
(DEFUN refine-rule (learning-episode node)
  (declare (type learning-episode learning-episode)
	   (type decision-node node))
;;  (setf (learning-episode-deletedp learning-episode) t)
  (let ((some-other-binding-p
	 (some-other-binding-p learning-episode node
			       (learning-episode-other-bindings
				learning-episode))))
    (if some-other-binding-p
	(let* ((other-bindings (car some-other-binding-p))
	       (intersection (cadr some-other-binding-p))
	       (new-effects (caddr some-other-binding-p))
	       (new-bindings (cdddr some-other-binding-p))
	       (cr1 (car other-bindings))
	       (cr2 (cadr other-bindings))
	       (new-other-bindings (list cr1 cr2 (cdaddr other-bindings)))
	       (old-cr (learning-episode-rule learning-episode))
	       (decision-leaf (find-decision-tree-node old-cr))
	       (new-learning-episode
		(learn-induced-rule new-other-bindings intersection
				    new-effects new-bindings
				    decision-leaf nil)))
	  (really-delete-rule learning-episode old-cr decision-leaf)
	  (propagate-changes learning-episode new-other-bindings)
;;	  (push learning-episode (learning-episode-from new-learning-episode))
	  (push (gethash (p4::control-rule-name cr2) *cr-to-le-hash*)
		(learning-episode-from new-learning-episode))
	  (push (gethash (p4::control-rule-name cr1) *cr-to-le-hash*)
		(learning-episode-from new-learning-episode))
	  t)
        (if (not (find-another-combination learning-episode node))
	    (refine-parents learning-episode node)))))

(DEFUN find-another-combination (learning-episode node)
  (declare (ignore learning-episode node))
  nil)

#|

;;;
;;; Checks whether intersecting rule2 and rule1 is better than rule1
;;; and rule2
;;;
(DEFUN find-another-combination (learning-episode node)
  (if (not (learning-episode-opposite-tried-p learning-episode))
      (let* ((old-le1 (car (learning-episode-from learning-episode)))
	     (old-le2 (cadr (learning-episode-from learning-episode)))
	     (cr1 (learning-episode-rule old-le1))
	     (cr2 (learning-episode-rule old-le2))
	     (re-order-p (some-binding-p learning-episode node
					 cr2 cr1 nil)))
	(setf (learning-episode-opposite-tried-p learning-episode) t)
	(if re-order-p
	    (let* ((old-cr (learning-episode-rule learning-episode))
		   (new-bindings (list (list r2 r1 (car re-order-p))))
		   (decision-leaf (find-decision-tree-node old-cr))
		   (new-learning-episode
		    (learn-induced-rule new-bindings (caadr re-order-p)
					(cadadr re-order-p)
					(car (cddr re-order-p))
					decision-leaf nil)))
	      (if (not (learning-episode-deletedp learning-episode))
		  (delete-cr old-cr decision-leaf learning-episode
			     (if (eq (learning-episode-type learning-episode)
				     'deduced)
				 t
			         'forever)))
	      (propagate-changes learning-episode new-bindings)
	      (push old-le1 (learning-episode-from new-learning-episode))
	      (push old-le2 (learning-episode-from new-learning-episode))
	      t)))))

|#

;;
;; It assumes the rule of the le is wrong.
;; It backs up to the previous rules. Tries to see if any of the preceding
;; rules don't cover the negative examples. If so, they are undeleted.
;;
(DEFUN refine-parents (le node)
  (declare (type learning-episode le)
	   (type decision-node node))
  (let ((recoverp nil))
    (dolist (from-le (learning-episode-from le))
      (setq recoverp (or (and from-le (recover-rule from-le node le))
			 recoverp)))
    (really-delete-rule le (learning-episode-rule le) node)
    recoverp))

#|
;;; Old version
;;
;; It backs up until there is the induced rule that created all
;; refinement. Tries to see if any of the preceding rules don't cover
;; the negative examples. If so, they are undeleted.
;; I take the last of froms, since they are the refined les.
;;
(DEFUN refine-parents (le node)
  (let ((rule (learning-episode-rule le)))
    (if (not (learning-episode-deletedp le))
	(delete-cr rule node le 'forever))
    (cond ((eq (learning-episode-type le) 'refined)
;;	 (setf (learning-episode-other-bindings le) nil)
;;	 (setf (learning-episode-deletedp le) 'forever)
	   (refine-parents (car (last (learning-episode-from le))) node))
	  (t (setf (learning-episode-other-bindings le) nil)
	     (let* ((from-les (learning-episode-from le))
		    (recover1p (recover-rule (car from-les) node le))
		    (recover2p (recover-rule (cadr from-les) node le)))
	       (or recover1p recover2p))))))

|#


;;;
;;; This is pretty difficult, but it is the only way I know now of
;;; doing it. The result is the union of two sets of negative examples
;;;
(DEFUN negatives-union (negatives1 negatives2)
  (declare (list negatives1 negatives2))
  (let ((result negatives1))
    (do* ((set2 negatives2 (cdr set2))
	  (negative2 (car set2) (car set2)))
	((null set2) result)
      (do* ((set1 result (cdr set1))
	    (negative1 (car set1) (car set1))
	    (included-1-in-2-p nil)
	    (included-2-in-1-p nil))
	  ((or (null set1)
	       (and included-1-in-2-p included-2-in-1-p))
	   (when (and (not included-1-in-2-p)
		      included-2-in-1-p)
	     (setq result (remove (car included-2-in-1-p) result
				  :test #'equal))
	     (push (cdr included-2-in-1-p) result)))
	(if (equal-control-rule-p (car negative1) (car negative2)
				  (cdr negative1) (cdr negative2) 'subset)
	    (setq included-1-in-2-p t))
	(if (equal-control-rule-p (car negative2) (car negative1)
				  (cdr negative2) (cdr negative1) 'subset)
	    (setq included-2-in-1-p (cons negative1 negative2)))))))


;;;
;;; The assumption is that child-le is wrong. 
;;;
(DEFUN recover-rule (le node child-le)
  (declare (type learning-episode le child-le)
	   (type decision-node node))
  (if (eq (learning-episode-type le) 'deduced)
      (recover child-le node le)
      (let ((rule (learning-episode-rule le)))
;;      (setf (learning-episode-other-bindings le) nil)
	(if (includes-negative-examples-p (get-real-preconds rule)
					  (get-real-effects rule)
					  (decision-node-negatives node))
	    (refine-rule le node)
	    (try-pool-or-create-pool child-le node le t 'intersection)))))


;;;
;;; It backtracks to the point to adding preconds to the rules.
;;;
(DEFUN recover (le node parent-le)
  (declare (type learning-episode le parent-le)
	   (type decision-node node))
  (if parent-le
      (let ((parent-rule (learning-episode-rule parent-le)))
	(if (includes-negative-examples-p (get-real-preconds parent-rule)
					  (get-real-effects parent-rule)
					  (decision-node-negatives node))
	    (let ((rule (learning-episode-rule le)))
	      ;; there is no need no know if it is a deduced one since
	      ;; it has a parent-le
	      (if (not (learning-episode-deletedp le))
		  (delete-cr rule node le 'forever))
	      ;; before: (refine-parents parent-le node)
	      (try-pool-or-create-pool
	       parent-le node nil nil
	       (if (eq (learning-episode-type parent-le) 'deduced)
		   'deduction
		   'intersection)))
	    (try-pool-or-create-pool le node parent-le t 'intersection)))
      (try-pool-or-create-pool le node nil nil 'deduction)))


;;
;; The input invariance is that parent-le, if different than nil, does
;; not cover the negative examples.
;;
;; Uses the pool of preconds, if there is one, or creates it, if there
;; is not. From that pool, it will pull one at a time to see if added
;; to the preconds of the more general (le), does not cover the
;; negative examples. The pool will usually be the difference between
;; the preconds of the control rule and the preconds of the previous
;; learning episode (or the whole set of preconds if the control rule
;; is deduced). In the case of deduced rules, it first tries to add
;; the preferred-preconds, that are the ones referring to the goal
;; regression of the some-candidate-goals. If this does not work, then
;; it begins adding from the whole set of preconds.
;;
;; Binding-used is only used the first time it is computed.
;; Op-type is the type of the inductive operator to recover from. For
;; now, there are only three types: deduction, intersection and
;; type-hierarchy.
;;
;; We could be better by passing along the preferred-preconds from
;; deduced to induced rules
;;
(DEFUN try-pool-or-create-pool (le node parent-le create-pool-p op-type)
  (declare (type learning-episode le parent-le)
	   (type decision-node node)
	   (atom create-pool-p op-type))
  (let* ((poolp (if (eq op-type 'deduction)
		    (learning-episode-preferred-preconds le)
		    (learning-episode-preconds-pool le)))
	 (rule (learning-episode-rule le))
	 (old-name (p4::control-rule-name rule))
	 (parent-rule (if parent-le (learning-episode-rule parent-le)))
	 (old-name-1 (if parent-rule (p4::control-rule-name parent-rule)))
	 (preconds-parent (if parent-rule (get-real-preconds parent-rule)))
	 (type-hierarchy-p (eq op-type 'type-hierarchy))
	 (deductionp (eq op-type 'deduction))
	 (binding-used (learning-episode-binding-used le))
	 (binding (some #'(lambda (rule-binding)
			    (if (and (car rule-binding)
				     (eq (p4::control-rule-name
					  (car rule-binding))
					 old-name-1))
				(cdr rule-binding)))
			binding-used))
	 (inverse-binding (if type-hierarchy-p
			      (reverse (reverse-binding binding))))
	 (second-parent (if (and parent-le type-hierarchy-p)
			    (some #'(lambda (binding-rule)
				      (if (not (eq (car binding-rule)
						   parent-rule))
					  (car binding-rule)))
				  binding-used)))
	 (preconds (if type-hierarchy-p
		       (specialize-hierarchy (sublis inverse-binding
						     (get-real-preconds rule))
					     preconds-parent
					     (if second-parent
						 (get-real-preconds
						  second-parent)))
		       (get-real-preconds rule)))
	 (pool (if type-hierarchy-p
		   (different-preconds preconds-parent preconds nil)
		   (or poolp
		       (and create-pool-p
			    (eq op-type 'intersection)
			    (different-preconds preconds-parent	preconds
						binding)))))
	 (effects (if type-hierarchy-p
		      (sublis inverse-binding (get-real-effects rule))
		      (get-real-effects rule)))
	 (first-prior-goals (if (and (eq op-type 'intersection) parent-le)
				(find-relevant-prior-goals
				 (learning-episode-prior-goal le)
				 (learning-episode-prior-goal parent-le))))
	 (new-preconds
	  (if type-hierarchy-p
	      (if (includes-negative-examples-p preconds effects
						(decision-node-negatives node))
		  (if pool
		      (find-discriminant-preconds preconds pool
						  first-prior-goals
						  effects node nil))
		  preconds)
	      (if pool
		  (find-discriminant-preconds preconds pool first-prior-goals
					      effects node deductionp))))
	 (new-binding
	  (and new-preconds
	       (setf *rule-id* (string-trim "#:G" (gensym)))
	       (create-new-binding
		(if type-hierarchy-p
		    inverse-binding
		    (if parent-le
			binding
			(mapcar #'(lambda (var)
				    (cons var var))
				(get-all-vars-from-types new-preconds))))))))
    (if *stop-repairing-p* (break))
    (cond (new-preconds
	   (let ((new-learning-step
		  (if parent-le
		      (1+ (max (learning-episode-learning-steps le)
			       (learning-episode-learning-steps
				parent-le)))
		      (1+ (learning-episode-learning-steps le)))))
	     (cond ((>= new-learning-step *max-learning-steps*)
		    (format t "Max learning steps reached for ~a~%" (p4::control-rule-name rule))
		    (really-delete-rule le rule node)
		    t)
		   (t 
		    (setq new-preconds
			  (sort-preconds (sublis new-binding new-preconds)))
		    (setq effects (sublis new-binding effects))
		    (let* ((name (rededuced-name new-preconds effects (not parent-le)))
			   (new-cr (p4::make-control-rule
				    :name name
				    :if (cons 'and new-preconds)
				    :then (cons 'then effects)))
			   (prior-goals (sublis new-binding
						(learning-episode-prior-goal le)))
			   (all-prior-goals (mapcar #'cadr prior-goals))
			   (new-le (make-learning-episode
				    :name (get-new-le-name)
				    :type (if (not parent-le) 'deduced 'refined)
				    :rule new-cr
				    :problem *problem-name*
				    :from (if parent-le (list parent-le))
				    :binding-used
				    (if parent-rule
					(pushnew (cons parent-rule new-binding)
						 binding-used :test #'equal))
				    :prior-goal
				    (create-new-prior-goals prior-goals new-preconds)
				    :learning-steps new-learning-step
				    :preferred-preconds nil
				    :preconds-pool
				    (if deductionp
					(sublis new-binding
						(learning-episode-preconds-pool le))
					(different-preconds (sublis new-binding pool)
							    new-preconds nil))
				    :matching-time
				    (if parent-le
					(max (learning-episode-matching-time le)
					     (learning-episode-matching-time
					      parent-le))
					(learning-episode-matching-time le))
				    :matchings
				    (if parent-le
					(max (learning-episode-matchings le)
					     (learning-episode-matchings parent-le))
					(learning-episode-matchings le))
				    :num-nodes
				    (if parent-le
					(max (learning-episode-num-nodes le)
					     (learning-episode-num-nodes parent-le))
					(learning-episode-num-nodes le))
				    :firings
				    (if parent-le
					(max (learning-episode-firings le)
					     (learning-episode-firings parent-le))
					(learning-episode-firings le)))))
		      (push new-le *le-list*)
		      (incf *number-refined-rules*)
	     
		      ;; So that we give more priority to things that worked out
	     
		      (update-frequencies (add-prior-goals all-prior-goals new-preconds)
					  node t)
		      (setf (gethash name *cr-to-le-hash*) new-le)
		      (format t "~%Recovering from ~(~a~) by creating pool from:~% ~a"
			      op-type old-name)
		      (if old-name-1
			  (format t " and~% ~a~%" old-name-1)
			  (format t " and~% the whole set of preconds~%"))
		      (really-delete-rule le rule node)
		      (add-rule new-cr nil))))))
	  (type-hierarchy-p
	   (format t "~2%This cannot happen!!!~%")
	   ;; (read)
	   (really-delete-rule le rule node)
	   (undelete-cr parent-le)
	   t)
	  (deductionp
	   ;; if it did not work to add all preconds from the
	   ;; preferred set, they are pushed into the pool, and the
	   ;; slot gets deleted
	   (let ((preferred-types nil))
	     (dolist (precond (learning-episode-preferred-preconds le))
	       (if (eq (car precond) 'true-in-state)
		   (push precond (learning-episode-preconds-pool le))
		   (push precond preferred-types)))
	     (setf (learning-episode-preferred-preconds le) nil)
	     (setf (learning-episode-preconds-pool le)
		   (append (learning-episode-preconds-pool le)
			   preferred-types)))
	   (try-pool-or-create-pool le node parent-le create-pool-p
				    'intersection))
	  (t
	   (cond ((or poolp (not create-pool-p))
		  (if (not (learning-episode-deletedp le))
		      (delete-cr rule node le
				 (if (eq (learning-episode-type le) 'deduced)
				     t
				     'forever)))
		  (if create-pool-p
		      (refine-parents parent-le node)))
		 (parent-le (try-pool-or-create-pool le node parent-le
						     create-pool-p
						     'type-hierarchy))
		 (t nil))))))

(defun really-delete-rule (le rule node)
  (if (not (learning-episode-deletedp le))
      (delete-cr rule node le
		 (if (eq (learning-episode-type le) 'deduced)
		     t
		     'forever))))
	  
#|
;; Rededuced-name was before:
    (or (not parent-le) (eq (learning-episode-type parent-le) 'deduced))
|#

;;
;; Changes the type-of-object declarations of the variables to the
;; ones without induction. Second-parent is the other parent of the
;; induction. Some variables come from that parent, so we have to
;; access their types from there
;;
(DEFUN specialize-hierarchy (preconds preconds-parent second-parent-preconds)
  (declare (list preconds preconds-parent second-parent-preconds))
  (let ((new-preconds nil))
    (dolist (cond preconds (nreverse new-preconds))
      (if (eq (car cond) 'type-of-object)
	  (push (or (find-type-of-precond (cadr cond) preconds-parent nil)
		    (find-type-of-precond (cadr cond) second-parent-preconds
					  t))
		new-preconds)
          (push cond new-preconds)))))

;;
;; It returns the list of cr-prior-goal-type-def lists that were also
;; in the parent
;;
;; cr-prior-goals is a list of the form:
;;  ((rule prior-goal type-of-object) ...)
;;
(DEFUN find-relevant-prior-goals (cr-prior-goals parent-prior-goals)
  (declare (list cr-prior-goals parent-prior-goals))
  (let ((prior-goals-types nil))
    (dolist (cr-prior-goal cr-prior-goals prior-goals-types)
      (if (member (car cr-prior-goal) parent-prior-goals
		  :key #'car :test #'eq)
	  (pushnew (cdr cr-prior-goal) prior-goals-types
		   :test #'(lambda (pg1 pg2)
			     (equal (get-pos-assertion pg1)
				    (get-pos-assertion pg2))))))))


;;; New version in which it computes the best predicate to add to the
;;; preconds based on the heuristic of most times appearing on the
;;; positive examples
;;;
;;; Returns the new set of preconds after adding the less number of
;;; preconds needed to not contain a negative example. Preconds
;;; are the ones of the rule, while pool is the ones we can add
;;;
;;; If use-all-preconds-p is t, it will add all preconds from new-pool
;;; (preferred-preconds) to the preconds.
;;;
(DEFUN find-discriminant-preconds (preconds old-pool prior-goals-types
					    effects node use-all-preconds-p)
  (declare (type decision-node node)
	   (list preconds old-pool prior-goals-types effects))
  (let* ((negatives (decision-node-negatives node))
	 (frequency-table-positive
	  (decision-node-frequency-table-positive node))
	 (frequency-table-prior-goal-positive
	  (decision-node-frequency-table-prior-goal-positive node))
	 (pool old-pool)
	 (new-pool nil)
	 (all-vars nil)
	 (type-defs nil)
	 (old-type-defs nil)
	 (before-preconds nil)
	 (after-preconds nil)
	 (after-preconds-assigned-p nil)
	 (state nil))
    (dolist (prior-goal-type prior-goals-types)
      (push (car prior-goal-type) pool)
      (setq type-defs (append type-defs (cadr prior-goal-type))))
    (dolist (cond pool)
      (case (car cond)
	(true-in-state
	 (push (cons cond
		     (or (gethash (car (get-pos-assertion (cadr cond)))
				  frequency-table-positive)
			 0))
	       new-pool))
	(prior-goal
	 (dolist (goal (cadr cond))
	   (push (cons (list 'prior-goal (list goal))
		       (or (gethash (car (get-pos-assertion goal))
				    frequency-table-prior-goal-positive)
			   0))
		 new-pool)))
	(type-of-object (push cond type-defs))
	(t nil)))
    (do ((conds preconds (cdr conds)))
	((null conds))
      (case (caar conds)
	(true-in-state (push (car conds) state))
	(prior-goal
	 (setq new-pool (remove-equal-goals new-pool conds))
	 (if (not after-preconds-assigned-p)
	     (push (car conds) before-preconds)))
	(some-candidate-goals
	 (when (not after-preconds-assigned-p)
	   (setq after-preconds conds)
	   (setq after-preconds-assigned-p t))
	 (setq new-pool (remove-equal-goals new-pool conds)))
	(type-of-object
	 (when (not after-preconds-assigned-p)
	   (setq after-preconds conds)
	   (setq after-preconds-assigned-p t))
	 (push (car conds) old-type-defs)
	 (push (cadar conds) all-vars))
	(t (if (not after-preconds-assigned-p)
	       (push (car conds) before-preconds)))))
    (setq before-preconds (nreverse before-preconds))
    (cond (new-pool
	   (setq new-pool
		 (sort new-pool
		       #'(lambda (cond1 cond2)
			   (more-significant-p cond1 cond2 all-vars))))
	   (if use-all-preconds-p
	       (try-preconds-set (mapcar #'car new-pool) before-preconds
				 state after-preconds effects negatives
				 type-defs old-type-defs)
	       (do ((i 1 (1+ i))
		    (new-preconds nil))
		   ((or (= i 3) new-preconds)
		    new-preconds)
		 (case i
		   (1 (setq new-preconds
			    (add-1-precond before-preconds state new-pool
					   after-preconds effects negatives
					   type-defs old-type-defs)))
		   (2 (setq new-preconds
			    (add-2-preconds before-preconds state new-pool
					    after-preconds effects negatives
					    type-defs old-type-defs)))
		   (t (format t "~2%I could not add only two things~%")
		      (read))))))
	  (t ;; the problem might be with the types
	     (add-type-precond before-preconds state after-preconds
			       effects negatives type-defs old-type-defs)))))

;; 	 (preds (p4::problem-space-all-preds p4::*current-problem-space*))
;; 	 (number-preds (length preds))
;; 		     (aref frequency-table
;; 			   (position (car (get-pos-assertion (cadr cond)))
;; 				     preds)
;; 			   0)  
;; 		       (aref frequency-table
;; 			     (+ number-preds
;; 				(position (car (get-pos-assertion goal))
;; 					  preds))
;; 			     0)
#|
;;; old version that computed the information gain
(DEFUN find-discriminant-preconds (preconds old-pool prior-goals-types
					    effects node)
  (declare (type decision-node node)
	   (list preconds old-pool prior-goal-types effects))
  (let* ((negatives (decision-node-negatives node))
	 (frequency-table (decision-node-frequency-table node))
	 (preds (p4::problem-space-all-preds p4::*current-problem-space*))
	 (number-preds (length preds))
	 (sums (compute-sums frequency-table number-preds))
	 (pool old-pool)
	 (new-pool nil)
	 (all-vars nil)
	 (type-defs nil)
	 (old-type-defs nil)
	 (before-preconds nil)
	 (after-preconds nil)
	 (after-preconds-assigned-p nil)
	 (state nil))
    (dolist (prior-goal-type prior-goals-types)
      (push (car prior-goal-type) pool)
      (setq type-defs (append type-defs (cadr prior-goal-type))))
    (dolist (cond pool)
      (case (car cond)
	(true-in-state
	 (push (cons cond
		     (compute-entropy frequency-table preds number-preds
				      (caadr cond) sums t))
	       new-pool))
	(prior-goal
	 (dolist (goal (cadr cond))
	   (push (cons (list 'prior-goal (list goal))
		       (compute-entropy frequency-table preds number-preds
					(car goal) sums nil))
		 new-pool)))
	(type-of-object (push cond type-defs))
	(t nil)))
    (do ((conds preconds (cdr conds)))
	((null conds))
      (case (caar conds)
	(true-in-state (push (car conds) state))
	(prior-goal
	 (setq new-pool (remove-equal-goals new-pool conds))
	 (if (not after-preconds-assigned-p)
	     (push (car conds) before-preconds)))
	(some-candidate-goals
	 (if (not after-preconds-assigned-p)
	     (setq after-preconds conds))
	 (setq new-pool (remove-equal-goals new-pool conds))
	 (setq after-preconds-assigned-p t))
	(type-of-object (if (not after-preconds-assigned-p)
			    (setq after-preconds conds))
			(push (car conds) old-type-defs)
			(push (cadar conds) all-vars))
	(t (if (not after-preconds-assigned-p)
	       (push (car conds) before-preconds)))))
    (setq before-preconds (nreverse before-preconds))
    (cond (new-pool
	   (setq new-pool (sort new-pool #'(lambda (cond1 cond2)
					     (more-significant-p cond1 cond2
								 all-vars))))
	   (do ((i 1 (1+ i))
		(new-preconds nil))
	       ((or (= i 3) new-preconds)
		new-preconds)
	     (case i
	       (1 (setq new-preconds
			(add-1-precond before-preconds state new-pool
				       after-preconds effects negatives
				       type-defs old-type-defs)))
	       (2 (setq new-preconds
			(add-2-preconds before-preconds state new-pool
					after-preconds effects negatives
					type-defs old-type-defs)))
	       (t (format t "~2%I could not add only two things~%")
		  (read)))))
	  (t ;; the problem might be with the types
	     (add-type-precond before-preconds state after-preconds
			       effects negatives type-defs old-type-defs)))))
|#

;; Deletes all prior goals that appear in the some-candidate-goals or
;; prior goals
(DEFUN remove-equal-goals (new-pool conds)
  (declare (list new-pool conds))
  (let ((goals (cadar conds)))
    (if goals
	(remove-if #'(lambda (precond-ig)
		       (let ((precond (car precond-ig)))
			 (and (eq (car precond) 'prior-goal)
			      (member (caadr precond) goals :test #'equal))))
		    new-pool)
        new-pool)))

;;
;; It gives preference to the ones that have more variables that are
;; already in the rule, then to the ones that have appeared more times
;; in the lhs of the learned rules, and then to the true-in-state
;;
(DEFUN more-significant-p (cond-frequency1 cond-frequency2 all-vars)
  (declare (list cond-frequency1 cond-frequency2 all-vars))
  (let* ((frequency-in-positives1 (cdr cond-frequency1))
	 (frequency-in-positives2 (cdr cond-frequency2))
	 (cond1 (car cond-frequency1))
	 (cond2 (car cond-frequency2))
	 (vars-in-cond1 nil)
	 (vars-in-cond2 nil))
    (cond ((eq (car cond1) (car cond2))
	   (cond ((eq (car cond1) 'true-in-state)
		  (setq vars-in-cond1 (cdadr cond1))
		  (setq vars-in-cond2 (cdadr cond2)))
		 (t (setq vars-in-cond1 (cdaadr cond1))
		    (setq vars-in-cond2 (cdaadr cond2))))
	   (let ((number-vars-precond1 (count-vars-in-preconds vars-in-cond1
							       all-vars))
		 (number-vars-precond2 (count-vars-in-preconds vars-in-cond2
							       all-vars)))
	     (cond ((> number-vars-precond1 number-vars-precond2))
		   ((= number-vars-precond1 number-vars-precond2)
		    (> frequency-in-positives1 frequency-in-positives2))
		   (t nil))))
	  ((eq (car cond1) 'true-in-state))
	  (t nil))))

#|
;; Old version
;; It gives preference to more times in the preconditions of positive
;; examples, the ones that have variables that are already in the
;; rule, and true-in-state's
;;
(DEFUN more-significant-p (cond-frequency1 cond-frequency2 all-vars)
  (declare (list cond-frequency1 cond-frequency2 all-vars))
  (let ((frequency-in-positives1 (cdr cond-frequency1))
	(frequency-in-positives2 (cdr cond-frequency2))
	(cond1 (car cond-frequency1))
	(cond2 (car cond-frequency2)))
    (cond ((> frequency-in-positives1 frequency-in-positives2))
	  ((< frequency-in-positives1 frequency-in-positives2) nil)
	  ((eq (car cond1) (car cond2))
	   (let ((vars-in-cond1 nil)
		 (vars-in-cond2 nil))
	     (cond ((eq (car cond1) 'true-in-state)
		    (setq vars-in-cond1 (cdadr cond1))
		    (setq vars-in-cond2 (cdadr cond2)))
		   (t (setq vars-in-cond1 (cdaadr cond1))
		      (setq vars-in-cond2 (cdaadr cond2))))
	     (> (count-vars-in-preconds vars-in-cond1 all-vars)
		(count-vars-in-preconds vars-in-cond2 all-vars))))
	  ((eq (car cond1) 'true-in-state))
	  (t nil))))


;; old version for information-gain
;;
;; It gives preference to less entropy, true-in-state, and the ones that
;; have variables that are already in the rule
;;
(DEFUN more-significant-p (cond-entropy1 cond-entropy2 all-vars)
  (declare (list cond-entropy1 cond-entropy2 all-vars))
  (let ((entropy1 (cdr cond-entropy1))
	(entropy2 (cdr cond-entropy2))
	(cond1 (car cond-entropy1))
	(cond2 (car cond-entropy2)))
    (cond ((< entropy1 entropy2))
	  ((> entropy1 entropy2) nil)
	  ((eq (car cond1) (car cond2))
	   (let ((vars-in-cond1 nil)
		 (vars-in-cond2 nil))
	     (cond ((eq (car cond1) 'true-in-state)
		    (setq vars-in-cond1 (cdadr cond1))
		    (setq vars-in-cond2 (cdadr cond2)))
		   (t (setq vars-in-cond1 (cdaadr cond1))
		      (setq vars-in-cond2 (cdaadr cond2))))
	     (> (count-vars-in-preconds vars-in-cond1 all-vars)
		(count-vars-in-preconds vars-in-cond2 all-vars))))
	  ((eq (car cond1) 'true-in-state))
	  (t nil))))
|#

(DEFUN count-vars-in-preconds (vars vars-in-preconds)
  (declare (list vars vars-in-preconds))
  (count-if #'(lambda (var)
		(member var vars-in-preconds :test #'eq))
	    vars))


#|
;; Not needed now
(DEFUN compute-entropy (frequency-table preds number-preds
					precond-name sums statep)
  (declare (list preds sums)
	   (array frequency-table)
	   (integer number-preds))
  (let* ((number-positives (aref frequency-table
				 (if statep
				     (position precond-name preds)
				     (+ number-preds
					(position precond-name preds)))
				 0))
	 (number-negatives (aref frequency-table
				 (if statep
				     (position precond-name preds)
				     (+ number-preds
					(position precond-name preds)))
				 1))
	 (ratio-pos (float (/ number-positives (car sums))))
	 (ratio-neg (float (/ number-negatives (cdr sums)))))
    (- (+ (if (zerop ratio-pos)
	      0
	      (* ratio-pos (log ratio-pos 2)))
	  (if (zerop ratio-neg)
	      0
	      (* ratio-neg (log ratio-neg 2)))))))

(DEFUN compute-sums (frequency-table number-preds)
  (declare (array frequency-table)
	   (integer number-preds))
  (let ((sums (cons 0 0)))
    (dotimes (i (* 2 number-preds) sums)
      (incf (car sums) (aref frequency-table i 0))
      (incf (cdr sums) (aref frequency-table i 1)))))
|#


;;;
;;; It returns the preconds if they do not cover the negative examples
;;;
(DEFUN not-covers-negatives (preconds effects negatives)
  (declare (list preconds effects negatives))
  (if (not (includes-negative-examples-p preconds effects negatives))
      preconds))


;;
;; Computes the set difference of two sets of preconds. It has in mind
;; the binding list used in preconds2
;;
;;;
;;; I have to fix this to allow computing a better set. F.i. in the
;;; case of some-candidate-goals, it should return the ones that are
;;; not already there. The same for any-prior-goal
;;;
(DEFUN different-preconds (preconds1 preconds2 binding-used)
  (declare (list preconds1 preconds2 binding-used))
  (let ((difference nil))
    (dolist (precond1 preconds1 (nreverse difference))
      (let* ((real-precond (sublis binding-used precond1))
	     (new-precond (different-precond real-precond preconds2)))
	(if new-precond
	    (push new-precond difference))))))

(DEFUN different-precond (precond1 preconds2)
  (declare (list precond1 preconds2))
  (let* ((meta-predicate1 (car precond1))
	 (new-precond
	  (some #'(lambda (precond2)
		    (case meta-predicate1
		      ((prior-goal some-candidate-goals)
		       (if (eq (car precond2) meta-predicate1)
			   (let ((different-goals
				  (set-difference (cadr precond1)
						  (cadr precond2)
						  :test #'equal)))
			     (if different-goals
				 (list meta-predicate1 different-goals)
			         (list nil)))))
		      ((true-in-state type-of-object)
		       (if (and (eq (car precond2) meta-predicate1)
				(equal (cdr precond2) (cdr precond1)))
			   (list nil)))
		      (t 
		       (if (eq (car precond2) meta-predicate1)
			   (if (equal (cdr precond2) (cdr precond1))
			       (list nil)
			       precond1)))))
		preconds2)))
    (if new-precond
	(if (car new-precond)
	    new-precond)
        precond1)))


;;; New version
;;;
;;; Returns the first preconds that do not cover the negative examples
;;; by adding one at a time state meta predicates.
;;;
(DEFUN add-1-precond (before-preconds state new-pool after-preconds
				      effects negatives
				      type-defs old-type-defs)
  (declare (list before-preconds state new-pool after-preconds
		 effects negatives type-defs old-type-defs))
  (some #'(lambda (cond-entropy)
	    (try-preconds-set (list (car cond-entropy)) before-preconds
			      state after-preconds effects negatives
			      type-defs old-type-defs))
	new-pool))

;;
;; Returns a new set of preconds if they do not cover the negative
;; examples after adding the new-conds.
;;
(DEFUN try-preconds-set (new-conds before-preconds state after-preconds
				   effects negatives type-defs old-type-defs)
  (declare (list new-conds before-preconds state after-preconds
		 effects negatives type-defs old-type-defs))
  (let ((new-set nil)
	(state-conds nil)
	(prior-goal-conds nil)
	(type-conds nil))
    (dolist (new-cond new-conds)
      (case (car new-cond)
	(true-in-state (push new-cond state-conds))
	(prior-goal (push new-cond prior-goal-conds))
	(type-of-object (push new-cond type-conds)
			(setq after-preconds
			      (change-type-same-var after-preconds
						    (cadr new-cond)
						    (caddr new-cond))))
	(t nil)))
    (setq new-set (append (if prior-goal-conds
			      (add-prior-goals prior-goal-conds
					       before-preconds)
			      before-preconds)
			  state-conds state after-preconds))
    (not-covers-negatives (append new-set
				  (add-type-of-new-vars
				   (append prior-goal-conds state-conds)
				   type-defs old-type-defs))
			  effects negatives)))

#|
    (setq new-preconds (not-covers-negatives new-set effects negatives))
    (if new-preconds
	(append new-preconds
		(add-type-of-new-vars (append prior-goal-conds state-conds)
				      type-defs old-type-defs)))
|#

;;
;; Removes the type-of-object meta-predicate for the var
;;
(DEFUN change-type-same-var (preconds var type)
  (declare (atom var type)
	   (list preconds))
  (cond ((null preconds) nil)
	((and (eq (caar preconds) 'type-of-object)
	      (eq (cadar preconds) var))
	 (cons (list (caar preconds) (cadar preconds) type)
	       (cdr preconds)))
	(t (cons (car preconds)
		 (change-type-same-var (cdr preconds) var type)))))

;;;
;;; Returns the first composition of two preconds that does not cover
;;; the negative examples.
;;;
(DEFUN add-2-preconds (before-preconds state new-pool after-preconds
				       effects negatives
				       type-defs old-type-defs)
  (declare (list before-preconds state new-pool after-preconds
		 effects negatives type-defs old-type-defs))
  (do* ((i 0 (1+ i))
	(length-pool (length new-pool))
	(precond1 (if (< i length-pool) (car (nth i new-pool)))
		  (if (< i length-pool) (car (nth i new-pool))))
	(new-preconds nil))
      ((or (= i length-pool)
	   new-preconds)
       new-preconds)
    (do* ((j 0 (1+ j))
	  (precond2 (if (and (/= i j) (< j length-pool))
			(car (nth j new-pool)))
		    (if (and (/= i j) (< j length-pool))
			(car (nth j new-pool)))))
	((or (= j length-pool)
	     new-preconds))
      (if precond2
	  (setq new-preconds
		(try-preconds-set (list precond1 precond2) before-preconds
				  state after-preconds effects negatives
				  type-defs old-type-defs))))))

;;;
;;; Returns the first preconds that do not cover the negative examples
;;; by adding one at a time the type-of-object meta-predicates.
;;;
(DEFUN add-type-precond (before-preconds state after-preconds effects
					 negatives type-defs old-type-defs)
  (declare (list before-preconds state after-preconds
		 effects negatives type-defs old-type-defs))
  (some #'(lambda (cond)
	    (try-preconds-set (list cond) before-preconds
			      state after-preconds effects negatives
			      type-defs old-type-defs))
	type-defs))

;;
;; Returns the new type-defs for the variables of the new-conds set
;; that were not already in the rule
;;
(DEFUN add-type-of-new-vars (new-conds type-defs old-type-defs)
  (declare (list new-conds type-defs old-type-defs))
  (let ((new-type-defs nil))
    (dolist (new-cond new-conds new-type-defs)
      (dolist (var (if (eq (car new-cond) 'true-in-state)
		       (cdadr new-cond)
		       (cdaadr new-cond)))
	(if (not (member var old-type-defs :test #'eq :key #'cadr))
	    (pushnew (car (member var type-defs :test #'eq :key #'cadr))
		     new-type-defs :test #'eq :key #'cadr))))))
  
#|
;;
;; I know this is not the minimal set, but it is a first approximation
;;
(DEFUN find-minimal-set (new-preconds state effects negatives)
  (do ((precond-deleted-p t)
       (remaining-state state))
      ((not precond-deleted-p)
       new-preconds)
    (setq precond-deleted-p nil)
    (setq state remaining-state)
    (dolist (precond state)
      (let ((new-conds (remove precond new-preconds :test #'equal)))
	(unless (includes-negative-examples-p new-conds effects negatives)
	  (setq new-preconds new-conds)
	  (setq remaining-state
		(remove precond remaining-state :test #'equal))
	  (setq precond-deleted-p t))))))

|#

;;
;; Returns a new name for a refinement. If the more specific rule
;; where it comes from is deduced, the name will be rededuced-x. If
;; not, it will be reinduced-x
;;
(DEFUN rededuced-name (preconds effects deducedp)
  (declare (atom deducedp)
	   (list preconds effects))
  (intern (format nil "~:@(re~a-~a~)"
		  (if deducedp 'deduced 'induced)
		  (name-cr effects (if (eq (cadr effects) 'bindings)
				       (cadar preconds))))))


;;;
;;; Propagates the other-bindings to the parents if the child is a
;;; refined le, and the parent is not a deduced le. Now it only
;;; propagates to the immediate parents.
;;;
(DEFUN propagate-changes (learning-episode new-other-bindings)
  (declare (type learning-episode learning-episode)
	   (list new-other-bindings))
  (if (eq (learning-episode-type learning-episode) 'refined)
      (dolist (from-le (learning-episode-from learning-episode))
	(if (and from-le (not (eq (learning-episode-type from-le) 'deduced)))
	    (setf (learning-episode-other-bindings learning-episode)
		  new-other-bindings)))))

;;
;; It returns the list of other-bindings, plus the one that worked,
;; if any element of the list of other-bindings creates
;; a relevant intersection. It really returns:
;;  (rest-other-crs-bindings new-preconds new-effects new-binding)
;;
(DEFUN some-other-binding-p (le node crs-binding)
  (declare (type learning-episode le)
	   (type decision-node node)
	   (list crs-binding))
  (if crs-binding
      (let* ((cr1 (car crs-binding))
	     (cr2 (cadr crs-binding))
	     (other-bindings (caddr crs-binding))
	     (some-binding-p (if other-bindings
				 (some-binding-p le node cr1 cr2
						 other-bindings))))
	(if some-binding-p
	    (cons (list cr1 cr2 (car some-binding-p))
		  (cdr some-binding-p))))))

;;
;; If any binding on the list other-bindings creates a
;; valid intersection of the preconds of rules 1 and 2, and does not
;; include any negative example, then it returns:
;;  (rest-bindings new-preconds new-effects new-bindings)
;;
(DEFUN some-binding-p (le node rule1 rule2 other-bindings)
  (declare (type learning-episode le)
	   (type decision-node node)
	   (type p4::control-rule rule1 rule2) 
	   (list other-bindings))
  (let ((new-rule (find-relevant-intersection (get-real-preconds rule1)
					      (get-real-preconds rule2)
					      rule1
					      (get-real-effects rule1)
					      (get-real-effects rule2)
					      (car other-bindings)
					      node nil)))
    (if new-rule
	(cons (cdr other-bindings) new-rule)
        (let ((rest-other-bindings (cdr other-bindings)))
	  (if rest-other-bindings
	      (some-binding-p le node rule1 rule2
			      rest-other-bindings))))))

;;;
;;; It returns t if the preconds-effects is a matching subset of any negative
;;; example
;;;
(DEFUN includes-negative-examples-p (preconds effects neg-examples)
  (declare (list preconds effects neg-examples))
  (let ((prior-goal-p (has-prior-goal-p preconds)))
    (some #'(lambda (neg-example)
	      (equal-control-rule-p preconds
				    (if prior-goal-p
				        (car neg-example)
					(remove-prior-goal (car neg-example)))
				    effects (cdr neg-example) 'intersection))
	  neg-examples)))
#|
;; previously
 (some #'(lambda (precond)
	      (prior-goal-meta-pred-p precond))
	 preconds)
|#

;;;
;;; Returns t if the preconds have any prior-goal related meta predicate
;;; It is the efficient procedure for doing it
;;;
(DEFUN has-prior-goal-p (preconds)
  (declare (list preconds))
  (cond ((null preconds) nil)
	((prior-goal-meta-pred-p (car preconds)) t)
	((member (caar preconds) '(true-in-state some-candidate-goals
						 type-of-object)
		 :test #'eq)
	 nil)
	(t (has-prior-goal-p (cdr preconds)))))

;;;
;;; Returns the preconds without any prior-goal related meta-predicate
;;;
(DEFUN remove-prior-goal (preconds)
  (declare (list preconds))
  (cond ((null preconds) preconds)
	((prior-goal-meta-pred-p (car preconds))
	 (cdr preconds))
	(t (cons (car preconds) (remove-prior-goal (cdr preconds))))))

;;;
;;; Just used for fast typing while debugging
;;;
(DEFUN fast-includes (rule-name)
  (declare (atom rule-name))
  (let ((rule (from-name-to-rule-1 rule-name)))
    (includes-negative-examples-p (get-real-preconds rule)
				  (get-real-effects rule)
				  (decision-node-negatives
				   (find-decision-tree-node rule)))))

;;;
;;; Undeletes a rule, replacing it into the appropriate structures
;;; Setting the variable *deleted-rule-p* is a trick so that it will
;;; save all rules and not only the new ones. If the number of rules
;;; is greater than the maximum number of rules allowed by the
;;; induction method, the flag *over-threshold-p* is set to t
;;;
(DEFUN undelete-cr (le)
  (declare (type learning-episode le))
  (let* ((cr (learning-episode-rule le))
	 (node (find-decision-tree-node cr)))
    (setf *deleted-rule-p* t)
    (incf *number-undeleted-rules*)
    (push cr *cr-list*)
    (format t "Undeleting ~a~%" (p4::control-rule-name cr))
;;; ******* Principio de Heras
    
    (if (or (eq *probabilities-p* 'rules)
	    (eq *probabilities-p* 'both))
	(borra-regla-prob (p4::control-rule-name cr) nil nil))

;;; ******* Final de Heras

    (setf (learning-episode-deletedp le) nil)
    (if (>= (length (push cr (decision-node-rules node)))
	    (decision-node-threshold node))
	(setf *over-threshold-p* t))
    t))

;;;
;;; High level function for undeleting a rule
;;;
(DEFUN undelete-rule (rule)
  (declare (type p4::control-rule rule))
  (if (not (p4::control-rule-p rule))
      (setq rule (from-name-to-rule-1 rule)))
  (undelete-cr (gethash (p4::control-rule-name rule) *cr-to-le-hash*)))



;;; ******************************************************************
;;;       Auxiliar functions
;;; ******************************************************************

;;;
;;; Saves the rules if savep is t, and prints information on the
;;; output file if *printp* is t.
;;;
(DEFUN save-induction (learning-time slow-induction-p savep)
  (declare (number learning-time)
	   (atom slow-induction-p savep))
  (if savep (save-rules))
  (if *printp*
      (with-open-file (stream *hamlet-output*
			      :direction :output :if-exists :append
			      :if-does-not-exist :create)
        (if slow-induction-p
	    (format stream "~%        (induced-~a ~d ~d ~d ~f)" *problem-name*
		    *number-induced-rules* *number-refined-rules*
		    *number-deleted-rules* learning-time)
	  (format stream "~%(setf *induction-data* '(induced-~a ~d ~d ~d ~f))"
		  *problem-name* *number-induced-rules* *number-refined-rules*
		  *number-deleted-rules* learning-time)))))

(DEFUN save-refinement (time save-rules-p)
  (declare (number time)
	   (atom save-rules-p))
  (if save-rules-p
      (save-rules))
  (if *printp*
      (with-open-file (stream *hamlet-output*
			      :direction :output :if-exists :append
			      :if-does-not-exist :create)
        (format stream "~%        (refined-~a ~d)" *problem-name* time))))


(DEFUN re-install-learning (&optional (domain 'logistics) (cr-file *cr-file*))
  (declare (atom domain))
  (format t "~%Recovering in the domain ~a~%" domain)
  (domain domain :function :compile)
  (reset-domain-variables t domain)
  (setf *cr-file* cr-file)
  (if (probe-file (cr-file))
      (load (cr-file))
      (load (format nil "~arules-backup.~a" *domain-directory*
		    *lisp-extension*)))
  (load-domain)
  (read-decision-tree)
  (read-learning-episodes domain)
  (domain domain)
  (load-domain))

;;
;; It deletes all rules learnt after and in the problem specified
;;
(DEFUN backtrack-to-problem (problem-name &optional (recoverp nil)
					  (delete-if-negative-p nil)
					  (domain 'logistics))
  (if recoverp (re-install-learning domain *cr-file*))
  (do* ((all-les *le-list* (cdr all-les))
	(le (car all-les) (car all-les))
	(problem-string (format nil "~a" problem-name)))
      ((or (null all-les)
	   (string< (format nil "~a" (learning-episode-problem le))
		    problem-string)))
    (let ((rule (learning-episode-rule le)))
      (when (or (and (learning-episode-deletedp le)
		     (or (and (listp rule)
			      (member (car rule) *cr-list*
				      :test #'(lambda (name rule2)
						(eq name
						    (p4::control-rule-name
						     rule2)))))
			 (and (atom rule)
			      (member rule *cr-list* :test #'eq))))
		(and (not (learning-episode-deletedp le))))
	(let* ((node (find-decision-tree-node (learning-episode-rule le)))
	       (negatives (if delete-if-negative-p
			      (decision-node-negatives node))))
	  (delete-cr (learning-episode-rule le) le node
		     (if (eq (learning-episode-type le) 'deduced)
			 t
		         'forever))
	  (dolist (from-le (learning-episode-from le))
	    (let ((rule (if (learning-episode-deletedp from-le)
			    (from-name-to-rule-1
			     (car (learning-episode-rule from-le)))
			    (learning-episode-rule from-le))))
	      (if (or (not delete-if-negative-p)
		      (not (includes-negative-examples-p
			    (get-real-preconds rule)
			    (get-real-effects rule)
			    negatives)))
		  (undelete-cr from-le)))))))))


(DEFUN save-decision-tree nil
  (with-open-file (stream (format nil "~adecision-tree.~a"
				  *domain-directory* *lisp-extension*)
			  :direction :output :if-exists :supersede
			  :if-does-not-exist :create)
    (format stream ";;; Decision tree")
    (format stream "~2%(setf *decision-tree-leaves*")
    (let ((*print-pretty* t))
      (format stream "~%     '~a)"
	      (get-all-decision-tree *decision-tree* nil)))))


(DEFUN get-all-decision-tree (node parents)
  (declare (type decision-node node)
	   (list parents))
  (let ((new-parents (cons (decision-node-name node) parents)))
    (if (decision-node-children node)
	(mapcan #'(lambda (child)
		    (get-all-decision-tree child new-parents))
		(decision-node-children node))
        (let ((rules (decision-node-rules node)))
	  (if rules
	      (list (list (reverse new-parents)
			  (from-rules-to-names rules)
			  (decision-node-negatives node)
			  (decision-node-needs-language-change-p node)
			  (get-frequency-table node))))))))

(DEFUN get-frequency-table (node)
  (declare (type decision-node node))
  (let* ((pos-table (decision-node-frequency-table-positive node))
	 (neg-table (decision-node-frequency-table-negative node))
	 (pos-prior-goal-table
	  (decision-node-frequency-table-prior-goal-positive node))
	 (neg-prior-goal-table
	  (decision-node-frequency-table-prior-goal-negative node))
	 (all-tables nil))
    (if (or pos-table neg-table pos-prior-goal-table neg-prior-goal-table)
	(dolist (predicate (p4::problem-space-all-preds
			    p4::*current-problem-space*)
			   all-tables)
	  (push (list predicate
		      (gethash predicate pos-table)
		      (gethash predicate neg-table)
		      (gethash predicate pos-prior-goal-table)
		      (gethash predicate neg-prior-goal-table))
		all-tables)))))

;; (DEFUN get-frequency-table (node)
;;   (declare (type decision-node node))
;;   (let* ((table (decision-node-frequency-table node))
;; 	 (number-preds (length (p4::problem-space-all-preds
;; 				    p4::*current-problem-space*)))
;; 	 (result nil)
;; 	 (result-aux nil))
;;     (if table
;; 	(dotimes (i 2 (nreverse result))
;; 	  (setq result-aux nil)
;; 	  (dotimes (j (* 2 number-preds) (push (nreverse result-aux) result))
;; 	    (push (aref table j i) result-aux))))))

(DEFUN read-frequency-table (preds-list node)
  (declare (list preds-list)
	   (type decision-node node))
  (when preds-list
      (setf (decision-node-frequency-table-positive node)
	    (make-hash-table))
    (setf (decision-node-frequency-table-negative node)
	  (make-hash-table))
    (setf (decision-node-frequency-table-prior-goal-positive node)
	  (make-hash-table))
    (setf (decision-node-frequency-table-prior-goal-negative node)
	  (make-hash-table))
    (dolist (predicate-frequencies preds-list)
      (let ((predicate (car predicate-frequencies)))
	(setf (gethash predicate
		       (decision-node-frequency-table-positive node))
	      (nth 1 predicate-frequencies))
	(setf (gethash predicate
		       (decision-node-frequency-table-negative node))
	      (nth 2 predicate-frequencies))
	(setf (gethash predicate
		       (decision-node-frequency-table-prior-goal-positive
			node))
	      (nth 3 predicate-frequencies))
	(setf (gethash predicate
		       (decision-node-frequency-table-prior-goal-negative
			node))
	      (nth 4 predicate-frequencies))))))
			 
	  

;; (DEFUN read-frequency-table (list)
;;   (declare (list list))
;;   (if list
;;       (let* ((pos-0 (car list))
;; 	     (pos-1 (cadr list))
;; 	     (number-preds-times-2 (length pos-0))
;; 	     (table (make-array (list number-preds-times-2 2)
;; 				:initial-element 0
;; 				:element-type 'integer)))
;; 	(dotimes (j number-preds-times-2)
;; 	  (setf (aref table j 0) (pop pos-0)))
;; 	(dotimes (j number-preds-times-2)
;; 	  (setf (aref table j 1) (pop pos-1)))
;; 	table)))


(DEFUN read-decision-tree nil
  (let ((decision-tree-file (format nil "~adecision-tree.~a"
				    *domain-directory* *lisp-extension*)))
    (build-fixed-decision-tree)
    (cond ((probe-file decision-tree-file)
	   (format t "~2%Reading decision tree.....~%")
	   (load decision-tree-file)
	   (dolist (leaf *decision-tree-leaves*)
	     (let ((decision-leaf (find-decision-leaf (cdar leaf) *decision-tree*)))
	       (setf (decision-node-rules decision-leaf)
		     (remove-nil-from-rules
		      (mapcar #'from-name-to-rule (second leaf))))
	       (setf (decision-node-negatives decision-leaf)
		     (third leaf))
	       (setf (decision-node-needs-language-change-p decision-leaf)
		     (fourth leaf))
	       (read-frequency-table (fifth leaf) decision-leaf))))
	  (t
 	   (format t "~2%Generating artificial decision tree.....~%")
	   (dolist (rule (all-control-rules))
	       (push rule (decision-node-rules (find-decision-tree-node rule))))))))

#|
      (setf (decision-node-deduced-rules decision-leaf)
	    (mapcar #'from-name-to-rule (third leaf))))))
|#


(DEFUN find-decision-leaf (path-list node)
  (declare (type decision-node node)
	   (list path-list))
  (if path-list
      (find-decision-leaf (cdr path-list)
			  (find-decision-child (car path-list) node))
      node))

(defun all-leafs (node)
  (declare (type decision-node node))
  (if (decision-node-children node)
      (mapcan #'(lambda (child) (all-leafs child))
	      (decision-node-children node))
      (list node)))


(DEFUN from-rules-to-names (rules)
  (declare (list rules))
  (let ((names nil))
    (dolist (rule rules (reverse names))
      (if (p4::control-rule-p rule)
	  (push (p4::control-rule-name rule) names)))))


(DEFUN pp-dn (node stream z)
  (declare (type decision-node node)
	   (stream stream)
	   (ignore z))
  (format stream "#<~a ~a" (decision-node-name node)
	  (decision-node-children-type node))
  (if (decision-node-children node)
      (format stream " ~a"
	      (mapcar #'(lambda (child) (decision-node-name child))
		      (decision-node-children node))))
  (let ((rules (decision-node-rules node)))
    (if rules
	(format stream " ~a"
		(mapcar #'(lambda (rule)
			    (if (p4::control-rule-p rule)
				(p4::control-rule-name rule)
			        'error))
			rules))))
  (format stream " ~d-NEGS>" (length (decision-node-negatives node))))


(defun pp-dnt (&optional (tree *decision-tree*) (profundidad 1))
  (declare (type decision-node tree)
	   (integer profundidad))
  (format t (format nil "~~%~~~dt~~a" (* 2 profundidad)) tree)
    (dolist (child (decision-node-children tree))
      (pp-dnt child (1+ profundidad))))


(DEFUN pp-control-rule (cr)
;;  (declare (type p4::control-rule cr))
  (if (not (p4::control-rule-p cr))
      (setq cr (from-name-to-rule-1 cr)))
  (format t "~%Rule ~a~%Preconds:" (p4::control-rule-name cr))
  (dolist (precond (get-real-preconds cr))
    (format t "~%  ~a" precond))
  (format t "~%Effects: ~a" (get-real-effects cr)))



;;;**************************************************
;;; Functions relative to learning episodes
;;;**************************************************

;;
;; It saves the learning episodes in order so that it can create them
;; properly when it recovers with read-learning-episodes
;;
(DEFUN save-learning-episodes (&optional (file (format nil
						       "~alearning-episodes.~a"
						       *domain-directory*
						       *lisp-extension*)))
  (when (probe-file (cr-file))
    (with-open-file (stream file
			    :direction :output :if-exists :supersede
			    :if-does-not-exist :create)
       (format stream ";;; Learning episodes")
       (format stream "~2%(setf *learning-episodes*~%     '(")
       (domain *domain-name*)
       (load (cr-file))
       (load-domain)
       (let ((le-array (make-array *le-index*))
	     (rules (all-control-rules)))
	 (maphash #'(lambda (cr-name le)
		      (declare (ignore cr-name))
		      (setf (svref le-array
				   (read-from-string
				    (subseq (format nil "~d"
						    (learning-episode-name le))
					    3)))
			    le))
		  *cr-to-le-hash*)
	 (dotimes (i *le-index*)
	   (let* ((le (svref le-array i))
		  (rule (if le (learning-episode-rule le))))
	     (if (and le rule)
		 (let ((deletedp (learning-episode-deletedp le))
		       (other-binding (learning-episode-other-bindings le))
		       (binding-used (learning-episode-binding-used le)))
		   (cond ((eq deletedp 'forever) (format stream "nil~%"))
			 (t (format stream
				    "(~a ~a ~a ~a~%          ~a~%          ~a"
				    (learning-episode-name le)
				    (learning-episode-type le)
				    (save-deleted-rule
				     rule
				     (or deletedp
					 (not (member rule rules :test #'eq))))
				    (get-from-les-names le)
				    (if (and other-binding
					     (car other-binding)
					     (cadr other-binding)) ;; borrar
					(list (p4::control-rule-name
					       (car other-binding))
					      (p4::control-rule-name
					       (cadr other-binding))
					      (caddr other-binding)))
				    deletedp)
			    (format stream "~%          ~a"
				    (learning-episode-preconds-pool le))
			    (format stream "~%          ~a"
				    (learning-episode-preferred-preconds le))
			    (format stream "~%          ~a"
				    (if (and binding-used
					     (caar binding-used)
					     (caadr binding-used)) ;; borrar
					(list (cons (p4::control-rule-name
						     (caar binding-used))
						    (cdar binding-used))
					      (cons (p4::control-rule-name
						     (caadr binding-used))
						    (cdadr binding-used)))))
			    (format stream "~%          ~a"
				    (save-prior-goal le))
			    (format stream "~%          ~a ~a ~d ~d ~d ~d ~d)~%       "
				    (learning-episode-opposite-tried-p le)
				    (learning-episode-problem le)
				    (learning-episode-learning-steps le)
				    (learning-episode-matching-time le)
				    (learning-episode-matchings le)
				    (learning-episode-num-nodes le)
				    (learning-episode-firings le)))))
	       (format stream "nil~%"))))
	 (format stream "))~%")))))

;;
;; It recovers the learning episodes from a file. It needs the rules
;; to be already loaded and load-domain evaluated
;; :negatives (nth 7 le)

(DEFUN read-learning-episodes (domain)
  (let ((le-file (format nil "~alearning-episodes.~a" *domain-directory*
			 *lisp-extension*)))
    (cond ((probe-file le-file)
	   (format t "~2%Reading learning episodes.....~%")
	   (clrhash *cr-to-le-hash*)
	   (setf *le-index* 0)
	   (setf *le-list* nil)
	   (load le-file)
	   (setf *rules* (all-control-rules))
	   (setf *cr-list* (copy-list *rules*))
	   (dolist (le *learning-episodes*)
	     (let* ((deletedp (nth 5 le))
		    (cr (nth 2 le))
		    (create-deleted-rule-p (or deletedp (listp cr))))
	       (if (and le create-deleted-rule-p)
		   (let* ((cr-name (if (listp cr) (car cr) cr))
			  (other-binding (nth 4 le))
			  (binding-used (nth 8 le))
			  (new-le
			   (make-learning-episode
			    :name (nth 0 le)
			    :type (nth 1 le)
			    :rule (if create-deleted-rule-p
				      (let ((rule (create-deleted-rule cr)))
					(push rule *cr-list*)
					rule)
				      (from-name-to-rule cr-name *cr-list*))
			    :problem (nth 11 le)
			    :from (from-names-to-pointers (nth 3 le))
			    :binding-used
			    (if binding-used
				(list (cons (from-name-to-rule (caar binding-used)
							       *cr-list*)
					    (cdar binding-used))
				      (cons (from-name-to-rule (caadr binding-used)
							       *cr-list*)
					    (cdadr binding-used))))
			    :prior-goal (read-prior-goal (nth 9 le))
			    :other-bindings
			    (if other-binding
				(list (from-name-to-rule (car other-binding) *cr-list*)
				      (from-name-to-rule (cadr other-binding) *cr-list*)
				      (caddr other-binding)))
			    :preconds-pool (nth 6 le)
			    :preferred-preconds (nth 7 le)
			    :deletedp deletedp
			    :learning-steps (nth 12 le)
			    :matching-time (nth 13 le)
			    :matchings (nth 14 le)
			    :num-nodes (nth 15 le)
			    :firings (nth 16 le)
			    :opposite-tried-p (nth 10 le))))
		     (push new-le *le-list*)
		     (setf (gethash cr-name *cr-to-le-hash*) new-le)))
	       (incf *le-index*)))
	   (setf *cr-list* nil))
	  (t
	   (format t "~2%Creating artificial learning episodes.....~%")
	   (create-artificial-les domain 'deduced)))))

(DEFUN save-prior-goal (le)
  (declare (type learning-episode le))
  (let ((output nil))
    (dolist (cr-prior-goal (learning-episode-prior-goal le) output)
      (let ((cr (car cr-prior-goal)))
	(if (and cr
		 (p4::control-rule-p cr))
	    (push (cons (p4::control-rule-name cr)
			(cdr cr-prior-goal))
		  output))))))

(DEFUN read-prior-goal (prior-goal)
  (declare (list prior-goal))
  (let ((input nil))
    (dolist (cr-prior-goal prior-goal input)
      (let ((name (from-name-to-rule (car cr-prior-goal) *cr-list*)))
	(if name
	    (push (cons name (cdr cr-prior-goal))
		  input))))))

(DEFUN save-deleted-rule (rule &optional (deletedp 'unknown))
  (declare (atom deletedp)
	   (type p4::control-rule rule))
  (let ((rule-name (p4::control-rule-name rule)))
    (if (eq deletedp 'unknown)
	(setq deletedp (learning-episode-deletedp (gethash rule-name
							   *cr-to-le-hash*))))
    (if deletedp
	(list rule-name (p4::control-rule-if rule)
	      (p4::control-rule-then rule))
        rule-name)))

(DEFUN get-from-les-names (episode)
  (declare (type learning-episode episode))
  (let ((les-names nil))
    (dolist (le (learning-episode-from episode) (nreverse les-names))
      (if le
	  (push (learning-episode-name le) les-names)))))



(DEFUN create-deleted-rule (rule)
  (declare (list rule))
  (p4::make-control-rule :name (first rule)
			 :if (second rule)
			 :then (third rule)))

(DEFUN from-name-to-rule (name &optional (rules (all-control-rules)))
  (declare (atom name)
	   (list rules))
  (find-if #'(lambda (cr)
	       (eq (p4::control-rule-name cr) name))
	   rules))

(DEFUN from-name-to-rule-1 (name)
  (declare (atom name))
  (learning-episode-rule (gethash name *cr-to-le-hash*)))

(DEFUN from-names-to-pointers (names-list)
  (declare (list names-list))
  (mapcar #'from-le-name-to-pointer names-list))


(DEFUN from-le-name-to-pointer (le-name)
  (declare (atom le-name))
  (find-if #'(lambda (le)
	       (eq (learning-episode-name le) le-name))
	   *le-list*))


;;; **************************************************
;;;           Auxiliary functions
;;; **************************************************

(DEFUN find-my-node (node name)
  (declare (atom name)
	   (type p4::nexus node))
  (if (eq (p4::nexus-name node) name)
      node
      (some #'(lambda (child)
		(find-my-node child name))
	    (p4::nexus-children node))))

(defun pp-le (episode stream z)
  (declare (type learning-episode episode)
	   (stream stream)
	   (ignore z))
  (format stream "#<~a ~a ~a ~a>" (learning-episode-name episode)
	  (learning-episode-type episode)
	  (learning-episode-rule episode)
	  (get-from-les-names episode)))


(defun pp-le1 (le-name-or-le)
;;  (declare (atom le-name-or-le))
  (let ((episode (if (learning-episode-p le-name-or-le)
		     le-name-or-le
		     (from-le-name-to-pointer le-name-or-le))))
    (format t "~2%Name: ~a  Type: ~a  Deletedp: ~a Problem: ~a~%Rule: ~a"
	    (learning-episode-name episode)
	    (learning-episode-type episode)
	    (learning-episode-deletedp episode)
	    (learning-episode-problem episode)
	    (learning-episode-rule episode))
    (format t "~%From les: ~a" (get-from-les-names episode))
    (format t "~%Preferred-Preconds: ~a~%Preconds-Pool: ~a~%Binding used: ~a"
	    (learning-episode-preferred-preconds episode)
	    (learning-episode-preconds-pool episode)
	    (learning-episode-binding-used episode))
    (format t "~%Other bindings: ~a~%Learning-steps: ~a  Matching-time: ~a"
	    (learning-episode-other-bindings episode)
	    (learning-episode-learning-steps episode)
	    (learning-episode-matching-time episode))
    (format t "  Matchings: ~a Number-nodes: ~a  Firings: ~a"
	    (learning-episode-matchings episode)
	    (learning-episode-num-nodes episode)
	    (learning-episode-firings episode))))

#|
      (format t "~%Negatives: ~a~%Deletedp: ~a"
	      (learning-episode-negatives episode)
	      (learning-episode-deletedp episode)))))
|#

(DEFUN pp-cr-les-hash nil
  (maphash #'(lambda (cr-name le)
	       (format t "~%~a~55t~a ~a ~a" cr-name
		       (learning-episode-name le)
		       (learning-episode-type le)
		       (learning-episode-deletedp le)))
	   *cr-to-le-hash*))

#|
;;; Auxiliary function for repairing. Here, there is
;;; another learning opportunity. When the search without rules has a
;;; node in the best path that it is not on the search tree with the
;;; rules, then it learns a decision from the old search tree.
;;;
;; in the else part of the last if
        (second-learning-opportunity old-tree)

(DEFUN repair-unoptimal-rules-1 (old-tree new-tree t-inicial
				 old-best-child old-number-ops node-type
				 shorterp old-best-childs)
  (declare (type p4::nexus old-tree new-tree old-best-child)
	   (number t-inicial)
	   (integer old-number-ops))
  (let ((best-child (find-equal-node old-best-child new-tree node-type)))
    (if best-child
	(let ((rules-nodes (find-rules-nodes new-tree node-type
					     old-number-ops nil))
	      (unknown-nodes nil)
	      (learning-nodes nil))
	  ;; if a rule (set of rules) selected a node that failed, or
	  ;; it was worse than the best, and there were not rules
	  ;; applied below it, or the old tree was expanded, and the
	  ;; corresponding node was not one of the best ones, then it
	  ;; is a negative example.
	  (dolist (rule-node rules-nodes)
	    (dolist (node (cdr rule-node))
	      (if (not (member node learning-nodes :test #'eq))
		  (if (member node unknown-nodes :test #'eq)
		      (delete node (cdr rule-node))
		      (let* ((old-node (find-equal-node node old-tree
							node-type)))
			(cond ((or (and shorterp
					(not (member old-node old-best-childs
						     :test #'eq)))
				   (not (rules-applied-p node)))
			       (push node learning-nodes))
			      (t (push node unknown-nodes)
				 (delete node rule-node))))))))
	  (if learning-nodes
	      (repair-overgeneralization old-tree new-tree
					 rules-nodes t-inicial))
	  (repair-unoptimal-rules old-best-child best-child t-inicial
				  shorterp))
        (repair-unfound-child old-tree new-tree t-inicial old-best-child
			      old-number-ops node-type))))


	 (best-child-info (if best-child
			      (node-best-child-info best-child))))

;;
;; Handles the case in which the new tree does not have a best child
;; of the old-tree.
;;
(DEFUN repair-unfound-child (old-tree new-tree t-inicial old-best-child
			     old-number-ops node-type)
  (declare (type p4::nexus old-tree new-tree)
	   (ignore old-best-child)
	   (number t-inicial)
	   (integer old-number-ops))
  (let* ((new-best-child-info (node-best-child-info new-tree))
	 (new-best-childs (car new-best-child-info))
	 (new-new-number-ops (caddr new-best-child-info)))
    (when (or (not new-best-childs)
	      (not new-new-number-ops)
	      (= 0 new-new-number-ops)
	      (> new-new-number-ops old-number-ops))
;;      (second-learning-opportunity old-tree)
      (repair-overgeneralization old-tree new-tree
				 (find-rules-nodes new-tree node-type
						   old-number-ops t)
				 t-inicial))))

;;
;; It follows tha past tree to learn optimally from it
;;
(DEFUN second-learning-opportunity (old-tree)
  (declare (type p4::nexus old-tree))
  (format t "~2%Second kind of learning opportunity~%")
  (setf *store-state-p* nil)
  (setf *learnp* t)
  (learn-from-tree old-tree nil (whole-tree-expanded-p))
  (dolist (new-le (learn-only-optimal-path old-tree nil))
    (add-rule (learning-episode-rule new-le) t)))
    

  (dolist (new-le (learn-decision old-tree old-best-child
				  (getf (p4::nexus-plist old-tree) 'state)
				  t))
    (add-rule (learning-episode-rule new-le) t))

;;; No longer needed because repair-unoptimal-rules does it at the
;;; same time as looking for unoptimal rules.
;;;
;;; It descends through the best-child structure until it doesn't find
;;; an equivalent best-child on the new-tree. Then, it calls
;;; repair-overgeneralization.
;;;
(DEFUN repair-different-node (old-tree new-tree t-inicial)
  (declare (type p4::nexus old-tree new-tree)
	   (number t-inicial))
  (let* ((old-best-child-info (node-best-child-info old-tree))
	 (old-best-childs (car old-best-child-info))
	 (node-type (node-type old-tree)))
    (some #'(lambda (old-best-child)
	      (let ((best-child (find-equal-node old-best-child
						 new-tree node-type)))
		(if best-child
		    (repair-different-node old-best-child best-child t-inicial)
		    (repair-overgeneralization old-tree new-tree
					       (find-rules-nodes
						new-tree (node-type new-tree)
						(caddr old-best-child-info) t)
					       t-inicial))))
	  old-best-childs)))

|#
