;; This is the initialization file for hamlet

(in-package "USER")

(defvar *truename* (namestring (truename "./")))
(setf *truename* (namestring (truename "./")))

(defvar *hamlet-path* (namestring (truename "./ipss/")))
(setf *hamlet-path* (namestring (truename "./ipss/")))

(defvar *ipss-path* *hamlet-path*)
(setf *ipss-path* *hamlet-path*)
(defvar *ipss-temporal-dir* (namestring (truename "./ipss/tmp/")))

(defvar *pltool-interface-dir* (format nil "~ainterface/pltool/" *ipss-path*))
(defvar *gtk-server-file* (format nil "~agtk-server.cfg" *pltool-interface-dir*))

(defvar *my-sayphi-pathname* (namestring (truename "./sayphi/")))
(setf *my-sayphi-pathname* (namestring (truename "./sayphi/")))

(defvar *my-cl-ppcre-pathname* (namestring (truename "./cl-ppcre-1.3.0/")))
(setf *my-cl-ppcre-pathname* (namestring (truename "./cl-ppcre-1.3.0/")))

(defvar *my-metricff-pathname* (namestring (truename "./metricff/")))
(setf *my-metricff-pathname* (namestring (truename "./metricff/")))

(defvar *my-sgplan-pathname* (namestring (truename "./sgplan/")))
(setf *my-sgplan-pathname* (namestring (truename "./sgplan/")))

(defvar *my-lpg-pathname* (namestring (truename "./LPG-td-1.0/")))
(setf *my-lpg-pathname* (namestring (truename "./LPG-td-1.0/")))

(defvar *my-manual-pathname* (namestring (truename "./manual/")))
(setf *my-manual-pathname* (namestring (truename "./manual/")))

(defvar *my-experimenter-pathname* (namestring (truename "./experimenter/")))
(setf *my-experimenter-pathname* (namestring (truename "./experimenter/")))

(defvar *my-prodigy-pathname* (namestring (truename "./prodigy/")))
(setf *my-prodigy-pathname* (namestring (truename "./prodigy/")))
(defvar *world-path* (format nil "~adomains/" *my-prodigy-pathname*))

(defvar *planner* 'prodigy)
(setf *planner* 'prodigy)

(defvar *language* 'es)
(setf *language* 'es)

;; I repeat this, so that it takes this new values if this file is reloaded

;; For faster code, less debugging:
(proclaim '(optimize (space 0) (compilation-speed 0) (safety 1) (speed 3)
		     (debug 0)))
;; For faster compilation, more debugging:
;; (proclaim '(optimize (compilation-speed 3) (safety 1) (speed 3) (debug 3)))


;;; The following line causes Allegro to do a tenured (full) GC when it should
#+ALLEGRO (setq EXCL::*GLOBAL-GC-BEHAVIOR* :AUTO)


;;; Stopping Allegro from warning about redefinitions makes loading files a
;;; lot faster
#+ALLEGRO (setf *enable-package-locked-errors* nil)
#+(or :Allegro-v4.1 :Allegro-v4.2 :Allegro-v4.3 :Allegro-v5.0)
(progn
  (setf (mp::global-symbol-value '*REDEFINITIONS-WARNINGS*) NIL)
  ;;the next four cause the compiler and loader not to generate source file and
  ;;cross-reference information which takes up much space & time:
  (setf *RECORD-SOURCE-FILE-INFO* NIL)
  (setf *LOAD-SOURCE-FILE-INFO* NIL)
  (setf *LOAD-XREF-INFO* NIL)
  (setf *RECORD-XREF-INFO* NIL)
  )

#+ALLEGRO
(setf TOP-LEVEL:*ZOOM-PRINT-LEVEL*  100
      TOP-LEVEL:*ZOOM-PRINT-LENGTH* 100
      TOP-LEVEL:*ZOOM-PRINT-CIRCLE* T
      TOP-LEVEL:*PRINT-LEVEL* 100
      TOP-LEVEL:*PRINT-LENGTH* 100)

#+(or ALLEGRO LISPWORKS)
(progn (setf *TRACE-PRINT-LEVEL* 100
	     *TRACE-PRINT-LENGTH* 100
	     *PRINT-LEVEL* 100
	     *PRINT-LENGTH* 100
	     *PRINT-PRETTY* T
	     *PRINT-CIRCLE* T
	     *PRINT-ARRAY* T
	     *PRINT-NICKNAME* NIL
	     *PRINT-STRUCTURE* T))

#+LISPWORKS
(progn (setf *enter-debugger-directly* t)
       (setf (editor:variable-value "Input Format Default")
	     '(:latin-1 :eol-style :lf))
       (setf *describe-print-level* 20
	     *step-print-level* 20
	     *trace-print-level* 20)
       (setf *compile-print* 0))



;;; ***************************************************************************
;;;          Loading Prodigy+Hamlet
;;; ***************************************************************************

;;; This is the file with all this Lisp functions that I use across applications

;; It returns a random element from list
(defun choose-one (list)
  (nth (random (length list)) list))

;; It returns a random element from list different from the one given as argument
(defun choose-diff-one (elem list)
  (let* ((number-elems (length list))
         (new-elem (nth (random number-elems)
                        list)))
    (if (and (> number-elems 1)
             (equal new-elem elem))
        (choose-diff-one elem list)
      new-elem)))

(defun diff (x y)
  (not (eq x y)))

;; It returns the euclidean distance between two points
(defun distance (x1 y1 x2 y2)
  (sqrt (+ (expt (- x1 x2) 2) (expt (- y1 y2) 2))))

;; It prints the contents of a hash table
(defun pp-hash-table (hash-table)
  (maphash #'(lambda (key val)
               (format t "~%~a: ~a" key val))
           hash-table))

;; It prints a list of elements in a stream with a tab
(defun pp-list (list &optional (tab 1) (ostream t) lowercasep)
  (dolist (literal list)
      (format ostream (format nil (if lowercasep
                                      "~~%~~~dt~~(~~a~~)"
                                      "~~%~~~dt~~a")
                              tab)
              literal)))

(defun load-prodigy4.0 ()
  (let ((*load-verbose* t))
    (load (format nil "~(~a~)~a" *my-prodigy-pathname* "loader"))
;; 		  #-LISPWORKS "loader.lsp"
;; 		  #+LISPWORKS "loader-lispworks.lsp"))
;; #-(or ALLEGRO CLISP LISPWORKS)
;;     (setf (working-directory) "/home/dborrajo/prodigy")
    (format t "~2%")))


;; If extension is "" (default), it will load the compiled files. Otherwise,
;; it will load the files ending with that string
(defun load-hamlet (&optional (extension "") (pathname *hamlet-path*))
  (load (format nil "~adefaults~a" pathname extension))
  (dolist (file *ipss-files*)
    (load (format nil "~a~(~a~)~a" pathname file extension)))
  (format t "~2%;;; IPSS loaded~2%"))

;; If extension is "" (default), it will load the compiled files. Otherwise,
;; it will load the files ending with that string
(defun load-ipss (&optional (extension "") (pathname *ipss-path*))
  (load (format nil "~adefaults~a" pathname extension))
  (dolist (file *ipss-files*)
    (load (format nil "~a~(~a~)~a" pathname file extension)))
  (format t "~2%;;; IPSS loaded~2%"))

(setf *print-length* most-positive-fixnum)
(setf *print-level* most-positive-fixnum)
#+LUCID (setf *debug-print-length* nil)
#+LUCID (setf *debug-print-level* nil)
#+(or ALLEGRO-V4.2 ALLEGRO-V4.3 ALLEGRO-V5.0)
  (progn (setf TOP-LEVEL::*PRINT-LEVEL* most-positive-fixnum)
	 (setf TOP-LEVEL::*PRINT-LENGTH* most-positive-fixnum))
(setf *print-pretty* t)

(load-prodigy4.0)
(load-ipss)
#+LISPWORKS (load (concatenate 'string *hamlet-path* "interface/lispworks/interface.lisp"))
#+(or ALLEGRO CLISP)
(defun pltool nil
  (setf *using-pltool-p* t)
  (load (format nil "~apltool" *pltool-interface-dir*)))

(load (format nil "~a~a"*my-cl-ppcre-pathname* "load.lisp"))

;; In case it is your first time, uncomment the following line and it will
;; recompile Hamlet for you

;; (load-and-compile *hamlet-files* :re-load-p nil)

;; I repeat this, so that it takes this new values if this file is reloaded
(setf *world-path* (namestring (truename "./prodigy/domains/")))

;; I have also implemented a mechanism to enter many lisps in sequence so that
;; each call to lisp will evaluate a user-defined set of functions. This
;; allows to learn from many different probsets and test, exiting and entering
;; lisp again, so that memory does not get exhausted. If you are going to use
;; that feature, ask me how to do it.
;; (load "/home/dborrajo/prodigy/domains/rocket/sigue.lisp")
