;; Copyright (c) 2004 James Bailey (dgym.REMOVE_THIS.bailey@gmail.com).
;;
;; Permission is hereby granted, free of charge, to any person obtaining a
;; copy of this software and associated documentation files (the "Software"), to
;; deal in the Software without restriction, including without limitation the
;; rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
;; sell copies of the Software, and to permit persons to whom the Software is
;; furnished to do so, subject to the following conditions:
;;
;; The above copyright notice and this permission notice shall be included in all
;; copies or substantial portions of the Software.
;;
;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
;; IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
;; AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
;; OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
;; SOFTWARE.

;; gtk-glade - this module exports glade-parse, which takes an xmls representation
;; of a glade save file, and generates all the widgets based on that information.
;; returns a hashtable of widget-ids (strings) to widget instances

(defpackage :gtk-glade
  (:use :common-lisp :gtk)
  (:export
    glade-parse))
(in-package :gtk-glade)


(defun get-attr (node name &optional (default ""))
  (let ((pair (assoc name (xmls:node-attrs node) :test #'string=)))
    (if pair
      (second pair)
      default)))

(defun get-props (node &optional (traduction-table nil))
  (let ((element nil))
  (loop with res = (make-hash-table :test 'equal)
        for child in (get-children node "property")
        finally return res
        do
	 (if (not (eq traduction-table nil))
	 	(if (or (equalp (get-attr child "name" "") "title") (equalp (get-attr child "name" "") "label"))
			(progn
	    		(setf element (gethash (first (xmls:node-children child)) traduction-table))
	    		(if (eq element nil)
	      			(setf (gethash (get-attr child "name" "") res)
		    		  (first (xmls:node-children child)))
	      			(setf (gethash (get-attr child "name" "") res) element)
	    		)  
			)
			(setf (gethash (get-attr child "name" "") res)
			  (first (xmls:node-children child)))
	 	)
	 	(setf (gethash (get-attr child "name" "") res)
              	  (first (xmls:node-children child)))
	  )
  )
)
)
      
;(defun get-props (node)
;  (loop with res = (make-hash-table :test 'equal)
;        for child in (get-children node "property")
;        finally return res
;        do
;        (setf (gethash (get-attr child "name" "") res)
;              (first (xmls:node-children child)))))

(defun get-int-prop (hash name)
  (parse-integer (gethash name hash "0")))

(defun get-bool-prop (hash name &optional (default nil))
  (string= (gethash name hash (if default "True" "False")) "True"))

(defun get-bool-prop-i (hash name &optional (default nil))
  (if (get-bool-prop hash name default)
    1
    0))

(defun get-children (node name)
  (remove-if (lambda (x) (not (and (consp x) (string= (car x) name))))
             (xmls:node-children node)))

(defun add-widget (child parent clas-hier node-hier &optional (traduction-table nil))
  ;;(pprint clas-hier)
  (macrolet ((with-packing-props 
               (&body body)

               `(let ((props (get-props (first (get-children (second node-hier) "packing")) traduction-table)))
                  ,@body)))
;    (print (first clas-hier))
    (case (first clas-hier)
      (|GtkTable|
        (let ((props (get-props (first (get-children (second node-hier) "packing")) traduction-table)))
          (gtk_table_attach_defaults parent child 
                                     (get-int-prop props "left_attach")
                                     (get-int-prop props "right_attach")
                                     (get-int-prop props "top_attach")
                                     (get-int-prop props "bottom_attach"))))
      ((|GtkVBox| |GtkHBox|)
       (with-packing-props
         (gtk_box_pack_start parent
                             child
                             (get-bool-prop-i props "expand")
                             (get-bool-prop-i props "fill")
                             (get-int-prop props "padding"))))

      (|GtkMenuItem|
        (gtk_menu_item_set_submenu parent child))
      (|GtkImageMenuItem|
        (gtk_menu_item_set_submenu parent child))
;     (|GtkButton|
;      (print node-hier))
;     (|GtkAlignment|
;      (print node-hier))
      (|GtkNotebook|
        (let ((cont (g_object_get_data parent "__child_store")))
          (cond 
            (cont
              (gtk_notebook_append_page parent cont child)
              (g_object_set_data parent "__child_store" nil))
            (t
              (g_object_set_data parent "__child_store" child)))))
      ;;raquel   
      (|GtkFixed|
         (let ((props (get-props (first (get-children (second node-hier) "packing")) traduction-table)))
      ;    (print (second node-hier))
      ;    (print "----------------------") 
      ;    (print props)
          (gtk_fixed_put parent child  (get-int-prop props "x")  (get-int-prop props "y"))))

      (otherwise
        (gtk_container_add parent child)))))

(defun glade-parse (node &optional (traduction-table nil) (parent nil) (clas-hier nil) (node-hier nil) (id-map (make-hash-table :test 'equal)))
  (let ((clas nil)
        (props nil)
        (new-widget nil)
        (args))
;  (print node)
    (when (consp node)

      (when (string= (xmls:node-name node) "widget")
        (setq clas (intern (get-attr node "class" "") :gtk-glade))
        (setq props (get-props node traduction-table))
;     (print clas)
;     (print props)

        (case clas
          (|GtkWindow|
            (setq new-widget (gtk_window_new 0))
            (gtk_window_set_title new-widget (gethash "title" props ""))
            (gtk_window_set_default_size 
                            new-widget 
                           (get-int-prop props "default_width")
                           (get-int-prop props "default_height"))

             (gtk_window_set_position new-widget 
                                      (get-int-enumerate-type (gethash "window_position" props) 
                                       '("GTK_WIN_POS_NONE"
                                         "GTK_WIN_POS_CENTER"
                                         "GTK_WIN_POS_MOUSE"
                                         "GTK_WIN_POS_CENTER_ALWAYS"
                                         "GTK_WIN_POS_CENTER_ON_PARENT"))))
         
          (|GtkVBox|
            (setq new-widget (gtk_vbox_new (get-bool-prop-i props "homogeneous")
                                           (length (get-children node "child")))))
          (|GtkHBox|
            (setq new-widget (gtk_hbox_new (get-bool-prop-i props "homogeneous")
                                           (length (get-children node "child")))))
           ;;raquel
           (|GtkFixed|
            (setq new-widget (gtk_fixed_new)))
          
           (|GtkTable|
            (setq new-widget (gtk_table_new 0
                                            (parse-integer (gethash "n_rows" props))
                                            (parse-integer (gethash "n_columns" props)))))
          (|GtkScrolledWindow|
            (setq new-widget (gtk_scrolled_window_new nil nil))
            (if (gethash "border_width" props)
                (gtk_container_set_border_width new-widget (parse-integer (gethash "border_width" props)))))
          (|GtkLabel|
            (setq new-widget (gtk_label_new (gethash "label" props ""))))

          (|GtkButton|
            (let ((label (gethash "label" props))) 
            (if (= 1 (get-bool-prop-i props "use_stock"))
               (setq new-widget (gtk_button_new_from_stock label))
               (if label
                  (setq new-widget (gtk_button_new_with_label label))
                  (setq new-widget (gtk_button_new))))))
           ;;raquel
          (|GtkToolButton|
            (let ((icon (gethash "icon" props ""))
                  (stock_id (gethash "stock_id" props ""))
                  (label (gethash "label" props "")))
              (if (equal icon "")
                 (if (equal stock_id "")
                    (setq new-widget (gtk_tool_button_new nil label))
                    (if (equal label "")
                       (setq new-widget (gtk_tool_button_new_from_stock stock_id))
                       (progn (setq new-widget (gtk_tool_button_new nil label))
                           (gtk_tool_button_set_stock_id new-widget stock_id)
                           (gtk_tool_button_set_icon_widget new-widget nil))))
                       
                 (setq new-widget (gtk_tool_button_new (gtk_image_new_from_file icon) label)))))
          ;;raquel
          (|GtkToggleButton|
            (setq new-widget (gtk_toggle_button_new_with_label (gethash "label" props ""))))
          ;;raquel
          (|GtkCheckButton|
            (setq new-widget (gtk_check_button_new_with_label (gethash "label" props "")))
            (gtk_toggle_button_set_active new-widget (get-bool-prop-i props "active")))
          ;;raquel
          (|GtkSpinButton|
            (let* ((adj-param (string-to-list (gethash "adjustment" props)))
                   (value (coerce (pop adj-param) 'double-float))
                   (lower (coerce (pop adj-param) 'double-float))
                   (upper (coerce (pop adj-param) 'double-float))
                   (step-increment (coerce (pop adj-param) 'double-float))
                   (page-increment (coerce (pop adj-param) 'double-float)) 
                   (page-size (coerce (pop adj-param) 'double-float))
                   (adjustment))
          (setq adjustment (gtk_adjustment_new value lower upper step-increment page-increment page-size))
          (setq new-widget (gtk_spin_button_new 
			        adjustment
                                  (coerce (read-from-string (gethash "climb_rate" props)) 'double-float ) 
                                  (parse-integer (gethash "digits" props))))
                 
            (gtk_spin_button_set_wrap new-widget (get-bool-prop-i props "wrap"))
            (gtk_spin_button_set_numeric new-widget (get-bool-prop-i props "numeric"))))
          ;;raquel
          (|GtkAlignment|
           (setq new-widget (gtk_alignment_new (coerce (read-from-string (gethash "xalign" props)) 'double-float)
                                               (coerce (read-from-string (gethash "yalign" props)) 'double-float) 
                                               (coerce (read-from-string (gethash "xscale" props)) 'double-float) 
                                               (coerce (read-from-string (gethash "yscale" props)) 'double-float)))) 
          (|GtkEntry|
            (setq new-widget (gtk_entry_new))
            (if (gethash "text" props)
              (gtk_entry_set_text new-widget (gethash "text" props))))
          ;;raquel
          (|GtkComboBox|
            (setq new-widget (gtk_combo_box_new_text))
             (let* ((items (gethash "items" props))
                    (list-items (string-to-list items)))
             (dolist (elem list-items)
		    (gtk_combo_box_append_text new-widget (symbol-name elem)))
             (gtk_combo_box_set_active new-widget 0)
             ))
           ;;raquel
           (|GtkImage|
             (let ((stock (gethash "stock" props))
		 (icon_size (gethash "icon_size" props)))
;            (print stock)
;            (print icon_size)
            (if stock
                 (setq new-widget (gtk_image_new_from_stock stock (parse-integer icon_size)))
                 (setq new-widget (gtk_image_new_from_file 
                                       (format nil "~a~a~a" user::*pltool-interface-dir* "pixmaps/" (gethash "pixbuf" props)))))))
           ;;raquel
           (|GtkImageMenuItem|
             (if (= 1 (get-bool-prop-i props "use_stock"))
                  (setq new-widget (gtk_image_menu_item_new_from_stock (gethash "label" props "") nil)) 
                  (setq new-widget (gtk_image_menu_item_new_with_label (gethash "label" props "" )))))
           ;;raquel
           (|GtkTreeView|
            (setq new-widget (gtk_tree_view_new))
            (gtk_tree_view_set_headers_visible new-widget 
                                               (get-bool-prop-i props "headers_visible")))
          (|GtkMenuBar|
            (setq new-widget (gtk_menu_bar_new)))
          ;;raquel
          (|GtkHPaned|
            (setq new-widget (gtk_hpaned_new)))
          (|GtkMenu|
            (setq new-widget (gtk_menu_new)))
          (|GtkMenuItem|
            (let ((label (gethash "label" props nil)))
              (setq new-widget 
                    (if label
                      (gtk_menu_item_new_with_mnemonic label)
                      (gtk_menu_item_new)))))
          ;;raquel
          (|GtkToolbar|
            (setq new-widget (gtk_toolbar_new)))
          ;;raquel
          (|GtkToolItem|
            (setq new-widget (gtk_tool_item_new)))
          ;;raquel
          (|GtkSeparatorMenuItem|
            (setq new-widget (gtk_separator_menu_item_new)))
           ;;raquel
          (|GtkVSeparator|
            (setq new-widget (gtk_vseparator_new)))
           ;;raquel
          (|GtkHSeparator|
            (setq new-widget (gtk_hseparator_new)))
          ;;raquel
          (|GtkTextView|
            (setq new-widget (gtk_text_view_new)))
          (|GtkNotebook|
            (setq new-widget (gtk_notebook_new)))

          (otherwise
            (format t "unhandled widget : ~a~%" clas)))   
         
         ;;raquel
         (if (and new-widget (gethash "width_request" props) (gethash "height_request" props))
           (gtk_widget_set_size_request new-widget (parse-integer (gethash "width_request" props))
                                                   (parse-integer (gethash "height_request" props)))))

      (when new-widget
        (setq new-widget new-widget)


        (when parent
          (add-widget new-widget parent clas-hier (cons node node-hier) traduction-table))

        (setf (gethash (get-attr node "id" "") id-map)
              new-widget)
        (setf (gethash new-widget id-map)
              (get-attr node "id" ""))

        ;; set up signals
        (loop for signal in (get-children node "signal") do
              (gtk_connect new-widget 
                           (get-attr signal "name") 
                           (let ((code (read-from-string (get-attr signal "handler"))))
                             (lambda (&rest args) (eval code))))))

      (loop with clases = (if clas 
                            (cons clas clas-hier)
                            clas-hier)
            with nodes = (cons node node-hier)
            for child in (xmls:node-children node) do
            (glade-parse child traduction-table (or new-widget parent) clases nodes id-map))

      id-map)))


;; raquel
;; this function converts a string with white spaces to a list of symbols/numbers
(defun string-to-list (string-list &optional (start 0))
(if string-list
  (let ((long (length string-list))
      (inic (multiple-value-bind (sym pos) (read-from-string string-list 
                :preserve-whitespace t :start start) (list sym pos))))
      (if (< (second inic) long)
          (let ((subresult (string-to-list string-list (second inic))))
               (push (car inic) subresult))
          (list (car inic))))))
                  
;; raquel
;; this function gives the integer of a enumerate type (string)
;; it is necesary that all-types (list of the posibles types) are in order
(defun get-int-enumerate-type (type all-types)
(if type 
   (position type all-types :test #'equal)
   0))



;;---------------------------------------------------------
;;  david montero
;; función que almacena el arbol de planes en un fichero
;;---------------------------------------------------------

(defun save-plans nil


)

   
 
