;;; This file contains the functions for high level control of Hamlet

(in-package "USER")

;;;***************************************************************
;;;;
;;;; Execution functions
;;;;
;;;***************************************************************

(defun my-run (time-bound multiple-sols-p my-output-level print-alts-p
			  linearp chronological-back-p random-behavior-p
			  depth-bound cost-bound cost-type)
  (pset :linear linearp)
  (pset :chronological-backtracking chronological-back-p)
  (pset :random-behaviour random-behavior-p)
  (pset :print-alts print-alts-p)
;;  (setf p4::*no-goal-choices-if-apply-p* linearp)
  (setf *number-fired-crs* 0)
  (if *use-quality-p*
      (run :output-level my-output-level
	   :multiple-sols multiple-sols-p
	   :time-bound time-bound
	   :depth-bound depth-bound
	   :cost-bound cost-bound
	   :cost-type cost-type
;;-------------- MD ----------- Marzo 2003
	   :make-span *make-span-p*
	   :include-resources *Res-value*
	   :min-makespan 0 ; *lower-bound-makespan-search*
	   :num-TP *numTP*
	   :activity-distance *Acti-dist*
	   :max-makespan *MS-value*)
;;-------------------------------------------
      (run :output-level my-output-level
	   :multiple-sols multiple-sols-p
	   :time-bound time-bound
	   :depth-bound depth-bound
;;-------------- MD ----------- Marzo 2003
	   :include-resources *Res-value*
	   :make-span *make-span-p*
 	   :min-makespan 0 ; *lower-bound-makespan-search*
	   :num-TP *numTP*
	   :activity-distance *Acti-dist*
	   :max-makespan *MS-value*
;;-------------------------------------------
)))


;;;
;;; Continuous can be:
;;;  nil: no continuous learning or testing (obsolete, will disappear)
;;;  (type old-solution-p old-cost-achieved old-time testingp shorterp)
;;;    type can be: without (1st pass, no rules), with (2nd pass, with
;;;    rules), t (no nothing)
;;;    shorterp tells if the old-tree was searched with shorter-or-equal
;;;
(defun my-init (problem learnp eager-mode time-bound multiple-sols-p
			continuous linearp reload-rules-p
			chronological-back-p random-behavior-p
			functions-not-loaded-p my-output-level print-alts-p
			depth-bound cost-bound cost-type)
  (when reload-rules-p
    (domain *domain-name* :load-domain nil :compile t
	    :function (if functions-not-loaded-p :compile))
    (load (cr-file) :if-does-not-exist nil)
    (load-domain))
  (let ((number-rules (length (union (all-control-rules)
				     (all-control-rules
				      *prefer-rules-types*))))
	(new-problem nil)
	(active-learning-p (numberp problem))
	(new-time-bound time-bound))
    (case number-rules
      (0 (format t "~%No rules to load~%"))
      (1 (format t "~%Loading 1 rule~%"))
      (t (format t "~%Loading ~d rules~%" number-rules)
	 (if *increment-time-p*
	     (setq time-bound (* time-bound
				 (1+ (truncate number-rules 10)))))))
    (multiple-value-setq (new-problem new-time-bound)
	(run-the-problem problem learnp time-bound multiple-sols-p my-output-level
		     print-alts-p linearp chronological-back-p
		     random-behavior-p depth-bound cost-bound cost-type eager-mode))
    (let* ((time (prodigy-result-time *prodigy-result*))
	   (nodes (prodigy-result-nodes *prodigy-result*))
	   (ac-time 0) (ac-nodes 0) (ac-all-nodes 0) (ac-nodes-solution 0)
	   (refinep (refinep continuous time))
	   (repeat-problem-p nil))
      (when learnp
	(format t "~%***************** Learning...~%")
	(remove-prod-handler :always #'store-last-node))
      (aprender (if learnp eager-mode 'eager-partial)
		(if (cdr continuous) (nth 2 continuous))
		(or (member (car continuous)
			    (list 'with 'without-induction)
			    :test #'eq)
		    (and *learnp*
			 (eq (car continuous) 'without))))
      ;; *number-ops* es cero cuando no se ha expandido el arbol por completo
      ;; solutionp es cierto cuando se encuentra una solucion
      (when (and (zerop *number-ops*) (solutionp))
	  (setf *best-cost-achieved* *best-cost*)
	(setf *number-ops*
	      (length (prodigy-result-solution *prodigy-result*))))
      (setf (prodigy-result-nodes-solution-path)
	    (if (and (numberp *number-ops*)
		     (> *number-ops* 0))
		(+ 3 (* *number-ops* 4))
	        0))
      (setq repeat-problem-p (and *repeat-problem-p*
				  (whole-tree-expanded-p)
				  ;; we will have to study this in terms of
				  ;; quality
				  (longer-first-path-p)))
      (setq refinep (or (and (not (member (car continuous)
					  (list 'without 'only-rules-testing
						'pure-testing)
					  :test #'eq))
			     *inductionp*
			     (or (and (repairp *old-tree* *real-t-tree*
					       (nth 5 continuous))
				      (not (eq eager-mode 'eager-partial)))
				 (and refinep *re-refine-p*)
				 (and repeat-problem-p
				      (learning-points-p *real-t-tree*))))
			(and repeat-problem-p
			     (eq (car continuous) 'without))))
      
;;; ********** Principio de Jose Luis Heras. **********

      (when (or (eq *probabilities-p* 'operators)
		(eq *probabilities-p* 'both))
	(create-probs *real-t-tree* nil))
	    
;;; ********** Final de Jose Luis Heras. **********

      (cond ((<= nodes 7)
	     (decf (prodigy-result-acc-nodes *prodigy-result*) nodes)
	     (setq nodes 7)
	     (incf (prodigy-result-acc-nodes *prodigy-result*) 7)
	     (setf (prodigy-result-nodes *prodigy-result*) 7)
	     (setf (prodigy-result-all-nodes) 7)
	     (setf (prodigy-result-nodes-solution-path) 7))
	    (t ;; since it also has into account the ficticious node 1
	     (decf nodes)
	     (decf (prodigy-result-acc-nodes *prodigy-result*))
	     (decf (prodigy-result-nodes *prodigy-result*))
	     (decf (prodigy-result-all-nodes))))
      (case (car continuous)
	((with pure-testing without-induction)
	 (decf (prodigy-result-acc-time *prodigy-result*) time)
	 (decf (prodigy-result-acc-nodes *prodigy-result*) nodes)
	 (unless refinep
	   (setq ac-time (incf (prodigy-result-acc-time-with) time))
	   (setq ac-nodes (incf (prodigy-result-acc-nodes-with) nodes))
	   (setq ac-all-nodes (incf (prodigy-result-acc-all-nodes-with)
				    (prodigy-result-all-nodes)))
	   (setq ac-nodes-solution
		 (incf (prodigy-result-acc-nodes-solution-path-with)
		       (prodigy-result-nodes-solution-path)))))
	(t
	 (setq ac-time (prodigy-result-acc-time *prodigy-result*))
	 (setq ac-nodes (prodigy-result-acc-nodes *prodigy-result*))
	 (setq ac-all-nodes (incf (prodigy-result-acc-all-nodes)
				  (prodigy-result-all-nodes)))
	 (setq ac-nodes-solution (incf (prodigy-result-acc-nodes-solution-path)
				       (prodigy-result-nodes-solution-path)))))
      (unless (solutionp)
	      (setf *DurationMax* 0)
	      (setf *MakeSpan* 0)
	      (p4::initialise-plan))
      (if (not *make-span-p*)
	  (setf *MakeSpan* *number-ops*))
      (output-results learnp new-time-bound (solutionp) (whole-tree-expanded-p)
		      *number-ops* time ac-time nodes ac-nodes
		      (prodigy-result-all-nodes) ac-all-nodes
		      (prodigy-result-nodes-solution-path) ac-nodes-solution
		      *number-fired-crs* number-rules *best-cost-achieved* *MakeSpan*
		      active-learning-p)
      (list refinep repeat-problem-p time new-problem))))


(defun output-results (learnp new-time-bound solutionp whole-tree-expanded-p
		       number-ops time ac-time nodes ac-nodes
		       all-nodes ac-all-nodes nodes-solution-path ac-nodes-solution
		       number-fired-crs number-rules best-cost-achieved make-span
		       active-learning-p)
  (cond ((and (eq *hamlet-output-format* 'spaces)
	      (not learnp))
	 (with-open-file (stream (if (= number-rules 0)
				     *prodigy-results-output*
				   *hamlet-results-output*)
			  :direction :output :if-exists :append
			  :if-does-not-exist :create)
	   (format stream "~%~a ~d ~a ~a ~d ~,3,,,F ~d ~d ~d ~d ~d"
		   *problem-name* new-time-bound solutionp whole-tree-expanded-p
		   number-ops time nodes all-nodes nodes-solution-path
		   number-fired-crs number-rules)
	   ;; o tal vez *best-cost*
	   (if *use-quality-p*
	       (format stream " ~,3,,,F" best-cost-achieved))
	   (format stream " ~,3,,,F" make-span)))
	((and (eq *hamlet-output-format* 'csv)
	      (not learnp))
	 (with-open-file (stream (if (= number-rules 0)
				     *prodigy-results-output*
				   *hamlet-results-output*)
			  :direction :output :if-exists :append
			  :if-does-not-exist :create)
	   (format stream "~%~a, ~d, ~a, ~a, ~d, ~,3,,,F, ~d, ~d, ~d, ~d, ~d"
		   *problem-name* new-time-bound solutionp whole-tree-expanded-p
		   number-ops time nodes all-nodes nodes-solution-path
		   number-fired-crs number-rules)
	   ;; o tal vez *best-cost*
	   (if *use-quality-p*
	       (format stream ", ~,3,,,F" best-cost-achieved))
	   (format stream " ~,3,,,F" make-span)))
	(t (with-open-file (stream *hamlet-output*
			    :direction :output :if-exists :append
			    :if-does-not-exist :create)
	     (format stream "~%        (~a ~a" *problem-name* learnp)
	     (format stream " ~d ~a ~a ~d ~,2,,,F ~,2,,,F ~d ~d ~d ~d ~d ~d ~d ~d"
		     new-time-bound solutionp whole-tree-expanded-p
		     number-ops time ac-time nodes ac-nodes
		     all-nodes ac-all-nodes nodes-solution-path ac-nodes-solution
		     number-fired-crs number-rules)
	     ;; o tal vez *best-cost*
	     (if *use-quality-p*
		 (format stream " ~,2,,,F" best-cost-achieved))
	     (format stream " ~,3,,,F" make-span)
	     (format stream ")"))))
;;   (with-open-file (ostream (format nil "~a-mio-out.txt" *domain-name*)
;; 		   :direction :output
;; 		   :if-exists :append
;; 		   :if-does-not-exist :create)
;;     (format ostream "~d~T~T~d~T~T~d~T~T~d~T~T~d~%" *problem-name* *MakeSpan* *number-ops* time *best-cost-achieved*)
;;     )
  (if active-learning-p
      (push (append (copy-tree *active-learning-parameters*)
		    (list (cons 'results
				(list *problem-name* learnp new-time-bound solutionp
				      whole-tree-expanded-p number-ops time nodes
				      nodes-solution-path number-fired-crs number-rules
				      (if *use-quality-p* best-cost-achieved 0) make-span
				      *something-changed-p*))))
	    *active-learning-trace*)))

(defun run-the-problem (problem learnp time-bound multiple-sols-p my-output-level
			print-alts-p linearp chronological-back-p
			random-behavior-p depth-bound cost-bound cost-type eager-mode)
  (declare (special *use-path-planning* *path-planning-dir* *solutions-hash* *dynamic-time-bound*
		    *current-problem-space*))
  (cond ((numberp problem) ;;active learning
	 (generate-and-run-problem problem learnp time-bound multiple-sols-p my-output-level
			print-alts-p linearp chronological-back-p
			random-behavior-p depth-bound cost-bound cost-type eager-mode))
	(t (my-load-problem problem)
	   (if learnp (define-prod-handler :always #'store-last-node))
	   ;;  
	   ;; someday I'll put this on, so that it doesn't run it again if it
	   ;; doesn't have rules and it already has the results of previous run  
	   ;;  (if (and continuous
	   ;;	   (eq (car continuous) 'with)
	   ;;	   (null (all-control-rules))))
	   ;;
	   (if *dynamic-time-bound*
	       (setq time-bound
		 (* time-bound
		    (length (cdr (p4::problem-goal *current-problem-space*))))))
	   (when *use-path-planning-p*
	     (translate-tp-to-pp problem (format nil "~atemp.lisp" *path-planning-dir*))
	     (setf *solutions-hash* (make-hash-table))
	     (busqueda :algoritmo 'a-star :dominio 'path-planning 
		       :problema "temp.lisp"))
	   (my-run time-bound multiple-sols-p my-output-level print-alts-p linearp
		   chronological-back-p random-behavior-p depth-bound cost-bound
		   cost-type)
	   (values problem time-bound))))

(defun generate-and-run-problem (problem learnp time-bound multiple-sols-p my-output-level
					 print-alts-p linearp chronological-back-p
					 random-behavior-p depth-bound cost-bound cost-type
					 eager-mode)
  (declare (special *active-learning-parameters* *domain-directory* *last-problem*))
  (do ((temp-file (concatenate 'string *domain-directory* "probsets/temp-problem.lisp"))
       (no-goals (cdr (assoc 'no-goals
			     (cdr (assoc 'parameters-alist *active-learning-parameters*)))))
       (max-goals (cdr (assoc 'max-goals 
			      (cdr (assoc 'parameters-alist *active-learning-parameters*)))))
       (types-info (cdr (assoc 'types-info *active-learning-parameters*)))
       (parameters-alist (cdr (assoc 'parameters-alist *active-learning-parameters*)))
       (new-problem nil)
       (unvalid-problems-with-solution 0)
       (valid-problem-p nil))
      (valid-problem-p (values new-problem time-bound))
    (incf (cdr (assoc 'generated-problems
		      (cdr (assoc 'parameters-alist *active-learning-parameters*)))))
    (when (or (and (eq *active-learning-mode* 'problem)
		   *last-problem*
		   (generate-problem-from-problem *last-problem*))
	      (and (eq *active-learning-mode* 'rules)
		   *last-problem*
		   (if (gethash 'type *active-learning-hash*)
		       (generate-problem-from-rules *last-problem*)
		     (generate-problem-from-problem *last-problem*)))
	      (generate-problem no-goals max-goals types-info "temp-problem" parameters-alist))
	(my-load-problem temp-file)
      (if learnp (define-prod-handler :always #'store-last-node))
      (setq new-problem (car *all-problems*))
      (setf *last-problem* new-problem)
      (let* ((new-problem-name nil)
	     (solution nil))
	(if *dynamic-time-bound*
	    (setq time-bound
		  (* time-bound
		     (length (cdr (p4::problem-goal *current-problem-space*))))))
	(cond ((test-problem new-problem time-bound multiple-sols-p my-output-level
			     print-alts-p linearp chronological-back-p
			     random-behavior-p depth-bound cost-bound cost-type eager-mode
			     (if *bootstrapp* (cr-file)))
	       (setq unvalid-problems-with-solution 0)
	       (setq valid-problem-p t)
	       (incf (cdr (assoc 'valid-problems
				 (cdr (assoc 'parameters-alist *active-learning-parameters*)))))
	       (setq solution (prodigy-result-solution *prodigy-result*))
	       (setq new-problem-name (intern (format nil "~:@(prob-~d~)" problem)))
	       (setf (cadadr (caddr new-problem)) new-problem-name)
	       (setf *problem-name* new-problem-name)
	       (with-open-file (ofile (concatenate 'string *domain-directory*
						   "probsets/active-learning."
						   *lisp-extension*)
				      :direction :output 
				      :if-exists :append)
		 (format ofile "~%~a"
			 (list new-problem-name *best-cost*
			       (if (listp solution)
				   (length solution)
				   1)
			       new-problem))))
	      (t (when *bootstrapp*
		     (if (solutionp)
			 (incf unvalid-problems-with-solution))
		   (when (> unvalid-problems-with-solution 10)
		       (setq unvalid-problems-with-solution 0)
		     (increase-difficulty)))))
	(delete-file temp-file)))))

;;; We have to refine if the problem was solved before and
;;; 1. the whole tree was expanded without finding a solution, OR
;;; 2. the whole tree was expanded and the number of ops is worse OR
;;; 3. the time spent in solving is greater in a max difference than
;;;    the other one, independently of whether it solved or not the problem
;;;    depending on the value of the variable *test-time-difference-p*	   
;;;
;;; The problem has to be solved wihout rules to repair, since,
;;; otherwise I don't know the best-child structure.
;;;	   
(DEFUN refinep (continuous time)
  (and *inductionp*
       (cdr continuous)
       (nth 1 continuous)
       (or (whole-tree-expanded-no-solution-p)
	   (and (or (nth 4 continuous) (whole-tree-expanded-p))
		;; Sira. Now we have to compare to *best-cost-achieved*
		;; instead of *number-ops*
		(> *best-cost-achieved* (nth 2 continuous)))
	   (and *test-time-difference-p*
		(> (- time (nth 3 continuous)) *max-time-difference*)))))

  
(DEFUN my-load-problem (problem)
  (setf *problem-name*
	(cond ((or (stringp problem) (pathnamep problem))
	       (load problem)
	       (my-problem-name problem 'string))
	      ((listp problem)
	       (eval problem)
	       (my-problem-name problem 'list))
	      (t (problem problem)
		 (my-problem-name problem 'atom)))))


(DEFUN my-problem-name (problem type)
  (case type
    (string (pathname-name problem))
    (list (cadadr (caddr problem)))
    (atom problem)
    (t nil)))
;; 		(subseq (namestring problem) 72)


;;; Now we learn from the run without rules, and refine when using the
;;; previous rules.
  
(DEFUN execute-one-problem (problem learnp eager-mode time-bound reload-rules-p
				    linearp multiple-sols-p continuous
				    functions-not-loaded-p depth-bound
				    cost-bound cost-type cycle-number)
  (declare (special p4::*all-solutions*))
  (let ((old-solution-p nil)
	(old-not-expanded-whole-p nil)
	(old-number-ops 0)
	(old-my-init nil)
	(old-time 0)
	(old-depth depth-bound)
	(shorterp nil)
	(old-best-cost-achieved cost-bound)
	(active-learning-p (numberp problem)))
    (when (and (listp problem) (not (eq (car problem) 'setf)))
      (setq old-depth (min old-depth (shorter-depth (caddr problem))))
      (setq old-best-cost-achieved (min old-best-cost-achieved (cadr problem)))
      (setq problem (cadddr problem)))
    (setf *old-tree* nil)
    (setf *real-t-tree* nil)
    (when (or *perform-makespan-dichotomic-search*
	      (and continuous (not (eq continuous 'only-rules-testing))))
      (domain *domain-name* :function (if functions-not-loaded-p :compile))
      (setq functions-not-loaded-p nil)
      (cond (*backtrack-with-costs-p*
	     (setf *multiple-solutions-flg* :different)
	     (setf *learnp* learnp))
	    (t (case continuous
		 (pure-testing
		  ;; This should be analysed in terms of quality
		  (setf *multiple-solutions-flg* :strict-shorter))
		 (without-induction
		  (setf *multiple-solutions-flg* :strict-shorter)
		  (setf *learnp* learnp))
		 (t (setq shorterp t)
		    (setf *learnp* learnp)
		    (setf *multiple-solutions-flg* :shorter)))))
      (format t "~%***************** Executing without rules...~%")
      (setf *store-state-p* (or learnp (eq continuous 'learn-testing)))
      (cond (*perform-makespan-dichotomic-search*
	     (setf *MS-value* *upper-bound-makespan-search*)
	     ;;	     (setf *MS-value* (/ (+ *upper-bound-makespan-search* *lower-bound-makespan-search*) 2.0))
	     (do* ((i 0 (1+ i)))
		 ((or (>= i *number-repetitions-makespan-dichotomic-search*)
		      (and (> i 0) (solutionp) (= *upper-bound-makespan-search* *lower-bound-makespan-search*))))
	       (incf old-time
		     (caddr (setq old-my-init (my-init problem learnp eager-mode 
						       (/ time-bound *number-repetitions-makespan-dichotomic-search*)
						       multiple-sols-p
						       (list 'without) linearp nil nil *random-behavior-p* nil
						       *hamlet-output-level* nil old-depth
						       cost-bound cost-type))))
	       (cond ((solutionp)
		      (setq problem (car (last old-my-init)))
		      (setq old-number-ops *number-ops*)
		      ;; if it could not expand the whole tree, and it can with the
		      ;; rules, we allow to learn in that case
		      (setq old-solution-p (solutionp))
		      (setq old-not-expanded-whole-p (not (whole-tree-expanded-p)))
		      (setf *upper-bound-makespan-search* *MakeSpan*)
		      (setf *MS-value* (/ (+ *upper-bound-makespan-search* *lower-bound-makespan-search*) 2.0)))
		     ((whole-tree-expanded-p)
		      (setf *lower-bound-makespan-search* *MS-value*)
		      (setf *MS-value* (/ (+ *upper-bound-makespan-search* *lower-bound-makespan-search*) 2.0)))
		     (t (setf i *number-repetitions-makespan-dichotomic-search*)))))
	    (t (setq old-my-init (my-init problem learnp eager-mode time-bound
					  multiple-sols-p
					  (list 'without) linearp nil nil *random-behavior-p* nil
					  *hamlet-output-level* nil old-depth
					  cost-bound cost-type))
	       (setq old-solution-p (solutionp))
	       (setq problem (car (last old-my-init)))
	       (when (and *follow-ff-p* 
			  (not old-solution-p)
			  (or learnp (eq continuous 'learn-testing)))
		 (format t "~%      ********* No solution found with Prodigy. Trying with FF....~%")
		 (generate-ff-solution *domain-name* problem)
		 (load (ff-cr-file))
		 ;; 	(if *best-first-search-p*
		 ;; 	    (load-a-star-rule))
		 (load-domain)
		 (my-load-problem problem)
		 (setq old-depth (min old-depth (shorter-depth (length *ff-solution*))))
		 (setq old-my-init 
		   (my-init problem learnp eager-mode time-bound
			    multiple-sols-p
			    (list 'without) linearp nil nil nil nil
			    *hamlet-output-level* nil old-depth 
			    cost-bound cost-type))
		 (setq old-solution-p (solutionp)))
	       (setq old-number-ops *number-ops*)
	       (setq old-time (caddr old-my-init))
	       (setq old-not-expanded-whole-p (not (whole-tree-expanded-p)))))
      ;; do not learn when *perform-makespan-dichotomic-search*=t because vars will not be ok.
      (setf *learnp* (and learnp old-not-expanded-whole-p))
      (if *print-solutions-p*
	  (print-the-solution))
      (if (or active-learning-p ;; active learning
	      (and old-solution-p (not old-not-expanded-whole-p)
		   *repeat-probset-p*
		   (not (member *problem-name* *solved-problems*
				:test #'eq :key #'car))))
	  (push (list *problem-name* *best-cost-achieved* *number-ops* problem)
		*solved-problems*))
      (when (and old-solution-p
		 (not (eq continuous 'pure-testing)))
	(setq old-depth (shorter-depth old-number-ops))
	(setq old-best-cost-achieved *best-cost-achieved*))
      (if (or (not (eq continuous 'pure-testing))
	      *using-pltool-p*)
	  (setf *old-tree* *real-t-tree*))
      (if *stop-after-problem-p* (break)))
    (if (and (not (eq continuous 'without-induction))
	     (not *perform-makespan-dichotomic-search*))
	(do ((repeatp t)
	     (repeat-whole-problem-p nil)
	     (old-optimal-learning *optimal-learning-p*))
	    ((or (and *never-repeat-p* (not (zerop cycle-number)))
		 (not repeatp)
		 (> cycle-number *max-number-cycles*)
		 repeat-whole-problem-p)
	     (setf *optimal-learning-p* old-optimal-learning)
	     (if (and (not *never-repeat-p*)
		      (not (member continuous
				   (list 'only-rules-testing 'pure-testing)
				   :test #'eq))
		      repeat-whole-problem-p
		      (< cycle-number *max-number-cycles*))
		 ;; Sira. Creo que esto es asi...pero no lo entiendo muy bien
		 (if *backtrack-with-costs-p*
		     (execute-one-problem problem learnp eager-mode time-bound
					  reload-rules-p linearp
					  multiple-sols-p continuous
					  functions-not-loaded-p
					  depth-bound
					  (if (zerop *best-cost-achieved*)
					      old-best-cost-achieved
					    *best-cost-achieved*)
					  cost-type (1+ cycle-number))
		   (execute-one-problem problem learnp eager-mode time-bound
					reload-rules-p linearp
					multiple-sols-p continuous
					functions-not-loaded-p
					(if (zerop *number-ops*)
					    (shorter-depth old-number-ops)
					  (shorter-depth *number-ops*))
					cost-bound cost-type
					(1+ cycle-number))))
	     (when *store-plans-p*
		 (push *real-t-tree* *all-trees*)
	       (push *old-tree* *all-trees*)
	       (push (prodigy-result-solutions *prodigy-result*) *all-solutions*)))
	  (setf *optimal-learning-p* t)
	  (incf cycle-number)
	  (format t "~%***************** Executing...~%")

	  ;; The output of my-init is whether to refine or not
	  ;; I can't make it strict-shorter always because, otherwise, it
	  ;; would consider the other solutions of the same length as
	  ;; negative examples
	  ;;(cond ((and *fast-learning-p*
	  ;;            (not (eq continuous 'pure-testing))
	  ;;            (not (eq continuous 'only-rules-testing)))))
	     ;;; I think I can do it now safely, since I already check
	     ;;; the needed things in the induction
	  (if *backtrack-with-costs-p*
	      (setf *multiple-solutions-flg* :different)
	    (setf *multiple-solutions-flg* :strict-shorter))
	  ;; it is not needed to store, since we can access it directly
	  (setf *store-state-p* nil)
	  (let ((my-init-result
		 ;; Sira. Creo que ahora esta bien
		 (if *backtrack-with-costs-p*
		     (my-init problem learnp eager-mode time-bound
			      multiple-sols-p
			      (if continuous
				  (if (or (eq continuous 'only-rules-testing)
					  (eq continuous 'pure-testing)
					  (eq continuous 'without-induction))
				      (list continuous)
				    (list 'with
					  (and old-solution-p t)
					  old-number-ops
					  old-best-cost-achieved
					  old-time
					  (eq continuous 'learn-testing)
					  shorterp))
				(list t))
			      linearp reload-rules-p
			      nil *random-behavior-p* functions-not-loaded-p
			      *hamlet-output-level*
			      nil depth-bound old-best-cost-achieved cost-type)
		   (my-init problem learnp eager-mode time-bound
			    multiple-sols-p
			    (if continuous
				(if (or (eq continuous 'only-rules-testing)
					(eq continuous 'pure-testing)
					(eq continuous 'without-induction))
				    (list continuous)
				  (list 'with
					(and old-solution-p t)
					;; Sira. Independientemente, siempre hay que usar esto.
					old-best-cost-achieved old-time
					(eq continuous 'learn-testing)
					shorterp))
			      (list t))
			    linearp reload-rules-p
			    nil *random-behavior-p* functions-not-loaded-p *hamlet-output-level*
			    nil old-depth cost-bound cost-type))))
	    (if *print-solutions-p*
		(print-the-solution))
	    ;; So that it does not loop in any problem...
	    (setq repeatp (and (car my-init-result)
			       (or (not old-not-expanded-whole-p)
				   (> *number-deduced-rules*
				      *number-deleted-rules*)
				   (> *number-induced-rules* 0))))
	    (setq repeat-whole-problem-p
	      (repeat-whole-problem-p (cadr my-init-result)
				      old-best-cost-achieved
				      old-not-expanded-whole-p)))
	  (if (and *utilityp* *print-utility-p*)
	      (pp-utility))
	  (when *stop-after-problem-p*
	    (setf *repeatp* repeatp)
	    (break)
	    (setq repeatp *repeatp*))
	  (setf p4::*all-solutions* nil)))))


;;;
;;; It returns t if my-init says to repeat the problem again or it did
;;; not explore the whole problem, solving the problem, without rules
;;; and it achieved a better solution with the rules
;;;
(DEFUN repeat-whole-problem-p (repeat-problem-p old-best-cost-achieved
						old-not-expanded-whole-p)
  (and (or (zerop old-best-cost-achieved)
	   (zerop *best-cost-achieved*))
       (or repeat-problem-p
	   (and (or *repeat-if-shorter-p*
		    (and (solutionp) (whole-tree-expanded-p)))
		(> old-best-cost-achieved *best-cost-achieved*)
		old-not-expanded-whole-p))))

;;;
;;; Returns a depth-bound that is enough for computing a solution of
;;; at most number-ops steps
;;; It adds 4 because there are some problems that could not be
;;; solved in only adding 3.
;;;
(DEFUN shorter-depth (number-ops)
  (if (= number-ops 1)
      8
      (+ 4 (* number-ops 4))))

;; Returns t if the first path explored was not the best one. It can be either
;; because the first path was a failure (since length-first-path returns the
;; most-positive-fixnum if the first path is a failure) or because the number
;; of nodes of the best path is greater than the one of the solution
;;
(DEFUN longer-first-path-p nil
  (declare (special *real-t-tree*))
  (let ((length-first-path (length-first-path *real-t-tree*)))
    (or (= length-first-path most-positive-fixnum)
	(< (prodigy-result-nodes-solution-path)
	   length-first-path))))

;;
;; Returns t if there has been any learning opportunity under node
;;
(DEFUN learning-points-p (node)
  (let* ((info (node-best-child-info node))
	 (lps (cadr info)))
    (and info
	 (or lps
	     (some #'(lambda (child)
		       (learning-points child))
		   (p4::nexus-children node))))))

;;
;; Returns the length of the first path explored beginning in node. It
;; returns a high value if the first path is a failure one
;;
(DEFUN length-first-path (node)
  (let ((last-child-p (car (last (p4::nexus-children node))))
	(best-info (node-best-child-info node)))
    (if (or (and best-info
		 (> (caddr best-info) 0))
	    (and (not last-child-p)
		 (eq :achieve (getf (p4::nexus-plist node)
				    :termination-reason))))
	(if last-child-p
	    (let ((child-length (length-first-path last-child-p)))
	      (if (< child-length most-positive-fixnum)
		  (1+ child-length)
		  most-positive-fixnum))
	    1)
	most-positive-fixnum)))

;;
;; Returns a list that is the first path beginning in node
;;
(DEFUN get-first-path (node)
  (declare (type p4::nexus node))
  (let ((first-child (car (last (p4::nexus-children node)))))
    (if first-child
	(cons node (get-first-path first-child))
        (list node))))

(DEFUN print-the-solution nil
  (with-open-file (ofile (concatenate 'string *domain-directory* *solutions-file*)
		   :direction :output :if-exists :append
		   :if-does-not-exist :create)
    (let ((solution (prodigy-result-solution *prodigy-result*)))
      (format ofile "~2%~a" (visible-solution (if (listp solution)
						  solution
						(list solution)))))))


(defun ejecutar-probset (learnp load-rules-p eager-mode problems time-bound
			 phase reload-rules-p linearp multiple-sols-p
			 continuous cost-type)
  (setf *load-verbose* nil)
  ;;  (my-init-prodigy-result)
  (setf *learnp* learnp)
  (let ((functions-not-loaded-p t))
    (unless reload-rules-p
      (cond (load-rules-p
	     (domain *domain-name* :load-domain nil
		     :compile t :function :compile)
	     (load (cr-file) :if-does-not-exist nil)
	     ;; 	     (if *best-first-search-p*
	     ;; 		 (load-a-star-rule))
	     (load-domain)
	     (if (member continuous '(only-rules-testing without-induction))
		 (init-for-utility)))
	    (t (domain *domain-name* :compile t :function :compile)))
      (setq functions-not-loaded-p nil)
      (setf *top-type* (p4::type-name-to-type :top-type
					      *current-problem-space*))
      (if (and *inductionp* (not *decision-tree*))
	  (build-fixed-decision-tree)))
    (if (and continuous learnp)
	(with-open-file (stream *hamlet-output*
			 :direction :output :if-exists :append
			 :if-does-not-exist :create)
	  (format stream "~2%(setf ~a~%      '(" phase)))
    ;; It will be a number when it is active learning
    (dolist (problem (if (numberp problems)
			 (list problems)
		       problems))
      (incf *problem-number*)
      (dotimes (i (if (and *random-behavior-p*
			   (not *perform-makespan-dichotomic-search*))
		      *number-repetitions-random-behavior* 
		    1))
	(execute-one-problem problem learnp eager-mode time-bound reload-rules-p
			     linearp multiple-sols-p
			     (if (eq continuous 'deduction)
				 t
			       continuous)
			     functions-not-loaded-p *max-depth*
			     *max-cost* cost-type 0)
	(if (zerop (length (prodigy-result-solution *prodigy-result*)))
	  (setf *unsolved-problems* (append *unsolved-problems* (list *problem-number*)))
	  )
	;; so that it deletes equal rules after every execution
	(if (eq continuous 'deduction)
	    (refine-rules))
	(setq functions-not-loaded-p nil))))
  (if (and continuous learnp)
      (with-open-file (stream *hamlet-output*
		       :direction :output :if-exists :append
		       :if-does-not-exist :create)
	(format stream "))"))))

;;
;; Learning-type can be
;; 1. interactive: it will ask for each problem, solve it and learn
;; 2. deduction: only learning by deduction
;; 3. deduction-induction: deduction followed by fast induction
;; 4. dynamic: dynamic learning
;; 5. ebl: learning only by deduction, computing utility. Later rules that do
;;         not have a higher utility than a threshold can be removed
;; 6. active: active learning. It will call the random generator at every
;;            cycle to obtain a new training problem. It also saves them.
;;            The random generator needs a random generator to exist for a
;;            given domain and is called with filter-problem-p=t, so it will
;;            only those problems that are known to be solvable and useful
;;
;; For eager-mode, there are two dimensions:
;; Learning points
;;  - Lazy: does not learn from default decisions
;;  - Eager: learns also from default decisions
;; Exhaustion of the search space
;;  - Total: it requires the expansion of the whole search space
;;  - Partial: it expands the search space up to the time limit
;; Then, we have four modes:
;;  - Lazy-total: not default and whole search space (very-lazy)
;;  - Eager-total: default and whole search space (lazy)
;;  - Lazy-partial: not default and partial search space (eager)
;;  - Eager-partial: default and partial search space (very-eager)
;;
;; probabilities=nil: no rules probabilities learning, or
;; probabilities=t: ops probabilities learning, or
;; probabilities=rules: rules  probabilities learning.
;; and use-probabilities is not used.
(DEFUN hamlet-learn (domain &key (probset "learning-1")
			    (learn-time-bound 200)
			    (eager-mode 'lazy-total)
			    (learning-type 'dynamic)
			    (re-init-p t)
			    (linearp nil)
;;; ********** Variables de Jose Luis Heras
			  (probabilities nil)
			  (use-probabilities nil)
;;; ********** Fin de variables de Jose Luis Heras
			    (cr-file nil)
			    (output-file "learn")
			    (make-span-p nil)
			    (max-MS most-positive-fixnum)
			    (include-resources-p nil)
			    (num-op 500)
			    (act-dist 10000)
			    (cost-bound nil)
			    (cost-type nil))

;;; Not needed. Handled as depth-bound (initial setting as *max-cost*)
  (declare (ignore cost-bound)
	   (special p4::*first-selection-p*))
;;; ********** Jose Luis Heras *******************

  (if (not cr-file)
      (setq cr-file
	(format nil "rules-~(~a-~a-~d.~a~)" eager-mode learning-type learn-time-bound
		*lisp-extension*)))

  (if (eq learning-type 'active)
      (if (eq *active-learning-mode* 'rules)
	  (setf *active-learning-hash* (make-hash-table :test #'equal)))
    (setf *active-learning-mode* nil))
  
  (if cost-type
      (setf *use-quality-p* t))
  (if (eq probabilities 'rules)
      (read-rules))
  (setf *use-probabilities* use-probabilities)
  (if (and (setf *probabilities-p* probabilities)
	   re-init-p)
      (init-probabilities domain t))

;;; ********** Fin Jose Luis Heras *******************

;; It is already true in defaults.lsp
;;   (setf *keep-all-selections-p* t)
  (if (not re-init-p)
      (re-install-learning domain cr-file))

  (setf *all-solutions* nil)
  (setf *all-trees* nil)

  (when *use-quality-p*
    (remove-prod-handler :always)
    (remove-prod-handler :achieve)
    (define-if-not-prod-handler :always #'p4::best-first-search-handler)
    (define-if-not-prod-handler :achieve #'p4::best-first-finish-handler))

  (setf *Res-value* include-resources-p)
  (setf *make-span-p* make-span-p)
  (when *make-span-p*
    ;; Only when I make them compatible, and they are not now unless *use-quality-p*
    (remove-prod-handler :always)
    (remove-prod-handler :achieve)
    (define-if-not-prod-handler :always #'p4::cost-ipss-handler-applied)
    (define-if-not-prod-handler :achieve #'p4::finish-ipss-handler))
  (setf *MS-value* max-MS)
  (setf *numTP* num-op)
  (setf *Acti-dist* act-dist)

  (case learning-type
    (deduction (setf *inductionp* nil))
    (ebl (setf *inductionp* nil)
	 (setq eager-mode 'eager-total)
	 (setf *utilityp* t))
    (t (setf *inductionp* t)))
  (setf p4::*first-selection-p* nil)
  (let ((starting-problem 1))
    (when (and (listp probset) (numberp (cadr probset)))
      (setq starting-problem (cadr probset))
      (setq probset (car probset)))
    (when *fast-learning-p*
;;    (setf *multiple-solutions-flg* :strict-shorter)
      (setf *test-time-difference-p* nil))
    (setf *cr-file* cr-file)
    (reset-domain-variables re-init-p domain)
    (setf *hamlet-results-output* (format nil "~(~ahamlet-~a.~a~)" *domain-directory*
					  output-file *lisp-extension*))
    (setf *prodigy-results-output* (format nil "~(~aprodigy-~a.~a~)" *domain-directory*
					   output-file *lisp-extension*))
    (setf *hamlet-output* (format nil "~(~a~a.~a~)" *domain-directory*
				  output-file *lisp-extension*))
    (if *repeat-probset-p*
	(load (format nil "~aprobsets/~(solved-~d~).~a" *domain-directory*
		      learn-time-bound *lisp-extension*)
	      :if-does-not-exist nil))
    (setf *optimal-learning-p* t)
    (domain domain :function :compile)
    (setf *valid-domain* 1)
    (when re-init-p
      (dotimes (i 10)
	(let ((filename (format nil
				"~aprobsets/~(solved-~d-~d.~a~)"
				*domain-directory*
				learn-time-bound
				i
				*lisp-extension*)))
	  (if (probe-file filename)
	      (delete-file filename))))
      (setf *rules* nil)
      (when *inductionp*
	(build-fixed-decision-tree)
	(clrhash *cr-to-le-hash*)
	(setf *le-list* nil)
	(setf *le-index* 0)))
    (case learning-type
      (interactive (do-interactive-learning eager-mode learn-time-bound linearp cost-type))

      (active (do-active-learning domain eager-mode learn-time-bound cost-type
				  linearp starting-problem)
	      (refine-rules))

      (t (setf *something-changed-p* nil)
	 (execute-and-learn probset eager-mode starting-problem
			    learn-time-bound (if (eq learning-type 'ebl)
						 'deduction
					       learning-type)
			    cost-type linearp)
;;; ******* Principio de Jose Luis Heras ******
	 (if (eq *probabilities-p* 'rules)
	     (write-rules))
;;; ****** Final de Jose Luis Heras
	 (repeat-probset-execution learn-time-bound eager-mode starting-problem
				   (if (eq learning-type 'ebl)
				       'deduction
				     learning-type)
				   cost-type linearp)
	 (case learning-type
	   (deduction-induction
	    (domain domain :function :compile)
	    (load (cr-file))
;; 	    (if *best-first-search-p*
;; 		(load-a-star-rule))
	    (load-domain)
	    (if (not *decision-tree*)
		(build-fixed-decision-tree))
	    (learn-from-induction nil (cr-file) (all-control-rules)
				  t nil nil))
	   (deduction
	    (refine-rules)
	    (setf *inductionp* t))
	   (ebl (setf *time-per-node*
		  (/ (* (prodigy-result-acc-time *prodigy-result*) 1000)
		     (prodigy-result-acc-nodes *prodigy-result*)))
		(refine-rules)
		(prune-rules-utility domain))
	   (t nil))))
    	   (setf *valid-control-rules* 1)))

(defun do-active-learning (domain eager-mode learn-time-bound cost-type
			   linearp starting-problem)
  (setf *last-problem* nil)
  (setf *active-learning-trace* nil)
  (setf *active-learning-parameters*
    (append (list (cons 'types-info *types-info*))
	    (list (cons 'parameters-alist *init-active-learning-parameters*))
	    (list (cons 'increase-difficulty-rules *init-active-learning-rules*))
	    nil))
  (with-open-file (ofile (concatenate 'string *domain-directory*
				      "probsets/active-learning."
				      *lisp-extension*)
		   :direction :output :if-exists :supersede :if-does-not-exist :create)
    (format ofile ";;; Problem set of the ~a domain" domain)
    (format ofile "~2%(setf *all-problems* '("))
  (load (concatenate 'string *domain-directory* "gen-probs"))
  (do ((i 0 (1+ i))
       (increased-difficulty-p t
			       (if (increase-difficulty-p)
				   (increase-difficulty)
				 t)))
      ((or (>= i *number-training-problems*)
	   (not increased-difficulty-p)))
    (setf *something-changed-p* nil)
    (execute-and-learn i eager-mode 0 learn-time-bound 'active cost-type linearp)
    (with-open-file (ofile (concatenate 'string *domain-directory*
					"trace-active-learning."
					*lisp-extension*)
		     :direction :output :if-exists :supersede :if-does-not-exist :create)
      (format ofile "~a~%" *active-learning-trace*))
    (if (eq *probabilities-p* 'rules)
	(write-rules)))
  (with-open-file (ofile (concatenate 'string *domain-directory*
				      "probsets/active-learning."
				      *lisp-extension*)
		   :direction :output :if-exists :append)
    (format ofile "))"))
  (setf *something-changed-p* t)
  (repeat-probset-execution learn-time-bound eager-mode starting-problem 'dynamic
			    cost-type linearp))

;; If we have been three problems without learning anything
(defun increase-difficulty-p nil
  (and (not (nth 13 (cdr (assoc 'results (car *active-learning-trace*)))))
       (not (nth 13 (cdr (assoc 'results (cadr *active-learning-trace*)))))
       (not (nth 13 (cdr (assoc 'results (caddr *active-learning-trace*)))))))

;; It returns t if there were more rules and nil otherwise
(defun increase-difficulty nil
  (declare (special *active-learning-parameters*))
  (let ((types (cdr (assoc 'types-info *active-learning-parameters*)))
	(type nil)
	(rules (pop (cdr (assoc 'increase-difficulty-rules *active-learning-parameters*)))))
    ;; so that it does not increase again in the next level if it does not
    ;; learn anything in the first try
    (if *active-learning-trace*
	(setf (nth 13 (cdr (assoc 'results (car *active-learning-trace*)))) t))
    (setf *last-problem* nil)
    (when rules
	(format t "Increasing difficulty by applying: ~a~%" rules)
	(dolist (what-to-increase rules t)
	    (setq type (cdr (assoc (car what-to-increase) types)))
	  (cond (type (incf (car type) (cadr what-to-increase))
		      (if (cadr type)
			  (incf (cadr type) (cadr what-to-increase))))
		(t ;; it must be number of goals
		 (incf (cdr (assoc 'no-goals (cdr (assoc 'parameters-alist
							 *active-learning-parameters*)))))
		 (incf (cdr (assoc 'max-goals (cdr (assoc 'parameters-alist
							  *active-learning-parameters*)))))))))))

(defun do-interactive-learning (eager-mode learn-time-bound linearp cost-type)
  (with-open-file (stream *hamlet-output*
			  :direction :output :if-exists :append
			  :if-does-not-exist :create)
    (format stream "(setf *interactive* '("))
  (do ((first-time-p t nil)
       (problem-file (ask-for-problem *fully-interactive-p*)
		     (ask-for-problem *fully-interactive-p*)))
      ((null problem-file)
       (with-open-file (stream *hamlet-output*
			       :direction :output :if-exists :append
			       :if-does-not-exist :create)
	 (format stream "))")))
    (if (not first-time-p)
	(save-rules))
    (ejecutar-probset t t eager-mode (list problem-file)
		      learn-time-bound nil t linearp t 'interactive cost-type)
    (break "~2%Type go if ready to continue~%")))

#|
  ;; there are still repetitions due to induce-over-induced
       
       (refine-rules (cr-file) (format nil "~arefined-rules.lisp"
				       *domain-directory*)))
       
      (t (format nil "~%Unknown learning type ~a" learning-type)))))
|#

;; I have changed it again so that it repeats it until I want to stop it.
;; 	       (setf *repeat-probset-p* nil)
(defun repeat-probset-execution (learn-time-bound eager-mode starting-problem learning-type
						  cost-type linearp)
  (do ((i 0 (1+ i)))
      ((or (not (and *repeat-probset-p* *something-changed-p*))
	   (= i *max-number-probset-repeats*)))
    ;; this is done until we divise a way of knowing when
    ;; we have effectively changed the set of control rules,
    ;; i.e. when a set is different than another.
    (setf *something-changed-p* nil)

    (with-open-file (stream (format nil
				    "~aprobsets/~(solved-~d-~d.~a~)"
				    *domain-directory*
				    learn-time-bound
				    i
				    *lisp-extension*)
		     :direction :output :if-exists :supersede
		     :if-does-not-exist :create)
      (let ((*print-pretty* t))
	(format stream "(setf *solved-problems*~%  '~(~a~))~%"
		*solved-problems*)
	(format stream
		"~%(setf *all-problems* *solved-problems*)~%")))
    (execute-and-learn (format nil "solved-~d-~d" learn-time-bound i)
		       eager-mode starting-problem
		       learn-time-bound learning-type cost-type linearp)))

;;
;; Executes and learns from a probset
;;
(DEFUN execute-and-learn (probset eager-mode starting-problem learn-time-bound
				  learning-type cost-type linearp)
  (if *utilityp* (init-for-utility))
  (let ((learn-set-name nil)
	(active-learning-p (eq learning-type 'active)))
    (dolist (ps (if (atom probset) (list probset) probset))
	(setq learn-set-name (format nil "~(*learn-~a-~d*~)" ps learn-time-bound))
      (if (not active-learning-p)
	  (load (format nil "~aprobsets/~(~a~).~a" *domain-directory* ps *lisp-extension*)))
      (my-init-prodigy-result)
      (ejecutar-probset t t eager-mode
			(if active-learning-p
			    probset
			    (nthcdr (1- starting-problem)
				    *all-problems*))
			learn-time-bound learn-set-name t linearp t
			(case learning-type
			  (interactive nil)
			  (deduction 'deduction)
			  (t t))
			cost-type)
      (setf *valid-problem* 1)
      (save-learning active-learning-p learn-time-bound ps))))

;; In case of active-learning, it saves a new rules file every *al-saving-rate* problems and
;; forces to write all rules everytime
(defun save-learning (active-learning-p learn-time-bound ps)
  (when (or *inductionp* *utilityp*)
      (if *utilityp* (save-utility))
    (let ((cr-file *cr-file*))
      (if active-learning-p
	  (setf *deleted-rule-p* t))
      (setf *cr-file* (if active-learning-p
			  (format nil "rules-~d-~d.~a" learn-time-bound
				  (floor ps *al-saving-rate*)
				  *lisp-extension*)
			  (format nil "rules-~a-~d.~a" ps learn-time-bound *lisp-extension*)))
      (save-cr)
      (setf *cr-file* cr-file))
    (if *inductionp*
	(save-decision-tree))
    (save-learning-episodes)))
;;
;; Inits the structure of prodigy-result
;;
(defun my-init-prodigy-result nil
  (init-prodigy-result)
  (setf (prodigy-result-acc-time-with) 0)
  (setf (prodigy-result-acc-nodes-with) 0)
  (setf (prodigy-result-all-nodes) 0)
  (setf (prodigy-result-acc-all-nodes) 0)
  (setf (prodigy-result-acc-all-nodes-with) 0)
  (setf (prodigy-result-nodes-solution-path) 0)
  (setf (prodigy-result-acc-nodes-solution-path) 0)
  (setf (prodigy-result-acc-nodes-solution-path-with) 0))
  

;;;
;;; It runs without the learned rules and with them.
;;; If probset-p is t, it will consider probsets as a set of probsets,
;;; and will load them from the probsets directory. If not, it will
;;; consider them as individual problems, and will load them from the
;;; probs directory.
;;;
;;; probabilities=nil: no ops. probabilities learning, or
;;; probabilities=operators: ops. probabilities learning.
;;; and use-probabilities=nil: do not use any probabilities,
;;;     use-probabilities=operators: use ops. probabilities,
;;;     use-probabilities=rules: use rules probabilities, or
;;;     use-probabilities=both: use ops. and rules probabilities.
;;;
(defun hamlet-test (domain probsets
		    &key (test-time-bound 200)
			 (refinep nil)
			 (run-without-p t)
			 (multiple-sols-p nil)
			 (linearp nil)
			 (cr-file (format nil "induced-rules.~a"
					  *lisp-extension*))
;;; ********** Variables de Jose Luis Heras
			 (probabilities nil)
			 (use-probabilities nil)
;;; ********** Fin de variables de Jose Luis Heras
			 (output-file nil)
			 (probset-p t)
			 (backtrack-with-costs-p nil)
			 ;; ********* VAriables de MD ****************************
			 (make-span-p nil)
			 (max-MS most-positive-fixnum)
			 (include-resources-p nil)
			 (num-op 500)
			 (act-dist 10000)
			 ;; *********************** FIN MD ***********************
			 (output-format 'lisp)
			 (store-fired-rules-p nil)
			 (cost-bound nil)
			 (cost-type nil))

;;; idem as in hamlet-learn  
  (declare (ignore cost-bound)
	   (special p4::*first-selection-p*))
  (setf *all-solutions* nil)
  (setf *all-trees* nil)
  (let ((starting-problem 1))
    (when (and (listp probsets) (numberp (cadr probsets)))
      (setq starting-problem (cadr probsets))
      (setq probsets (car probsets)))
    (setf *backtrack-with-costs-p* backtrack-with-costs-p)
    (setf *use-quality-p* backtrack-with-costs-p)
    (when *use-quality-p*
      (remove-prod-handler :always)
      (remove-prod-handler :achieve)
      (define-if-not-prod-handler :always #'p4::best-first-search-handler)
      (define-if-not-prod-handler :achieve #'p4::best-first-finish-handler))
    ;; ---------------------MD Marzo 2003 --------------
    (setf *Res-value* include-resources-p)
    (setf *make-span-p* make-span-p)
    (when *make-span-p*
      ;; It does not make sense to have both.
      (when *use-quality-p*
	(remove-prod-handler :always)
	(remove-prod-handler :achieve))
      (define-if-not-prod-handler :always #'p4::cost-ipss-handler-applied)
      (define-if-not-prod-handler :achieve #'p4::finish-ipss-handler))
    (setf *MS-value* max-MS)
    (setf *numTP* num-op)
    (setf *Acti-dist* act-dist)
					;---------------------------------------------------
    (setf *keep-all-selections-p* t)
    (setf p4::*first-selection-p* nil)
    (setf *footprintdata* nil)
    (setf *hamlet-output-format* output-format)
    ;;  (setf *random-selection-p* nil)    
    (domain domain :compile t :function :compile)      
    (setf *valid-domain* 1)
    (reset-domain-variables nil domain)
    (setf *cr-file* cr-file)

;;; ********** Principio de Jose Luis Heras. **********

    (if (setf *probabilities-p* probabilities)
	(init-probabilities domain nil))
    (setf *use-probabilities* use-probabilities)
    (if *use-probabilities*
	(read-probs))

;;; ********** Fin de Jose Luis Heras. **********

    (cond (refinep
	   (setf *rules* nil)
	   (setf *inductionp* t)
	   (unless *decision-tree*
	     (read-decision-tree)
	     (read-learning-episodes domain)))
	  (t;; we can't repeat problems when testing
	   (setf *repeat-problem-p* nil)
	   ;; this is needed so that it will not access to the learning episodes
	   (setf *inductionp* nil)
	   (if (and *utilityp* run-without-p)
	       (create-artificial-les domain 'artificial))
	   (setf *store-fired-crs-p* store-fired-rules-p)))
    (setf *cr-list* nil)
    (setf *rules* nil)
    ;;  (if multiple-sols-p
    ;;      (setf *multiple-solutions-flg* :strict-shorter))
    (dolist (probset (if (listp probsets) probsets (list probsets)))
      (let* ((test-name (format nil "~a-~d" probset test-time-bound))
	     (test-set-name (format nil "~(*test-~a*~)" test-name))
	     (reload-rules-p  (or run-without-p refinep)))
	(if probset-p
	    (load (format nil "~aprobsets/~a.~a"
			  *domain-directory* probset
			  *lisp-extension*))
	  (setf *all-problems* (list probset)))
	(setf *hamlet-results-output* (format nil "~(~ahamlet-~a.~a~)"
					      *domain-directory*
					      (or output-file test-name)
					      *lisp-extension*))
	(setf *prodigy-results-output* (format nil "~(~aprodigy-~a.~a~)"
					       *domain-directory*
					       (or output-file test-name)
					       *lisp-extension*))
	(setf *hamlet-output* (format nil "~(~a~a.~a~)"
				      *domain-directory*
				      (or output-file test-name)
				      *lisp-extension*))
	(if (not (eq *hamlet-results-output* 'csv))
	    (with-open-file (file *hamlet-output* :direction :output
			     :if-exists :append
			     :if-does-not-exist :create)
	      (format file "~%;;; Tests with probset ~a, using rules in ~(~a~)" probset (cr-file))))
	;; When testing, utilities have to be added through all probsets if
	;; one wants to know the final utility for each rule.
	(if *utilityp* (init-for-utility))
	(my-init-prodigy-result)
	(ejecutar-probset nil (not reload-rules-p) nil 
			  (nthcdr (1- starting-problem) *all-problems*)
			  test-time-bound test-set-name reload-rules-p 
			  linearp multiple-sols-p
			  (if run-without-p
			      (if refinep
				  'learn-testing
				'pure-testing)
			    'only-rules-testing)
			  cost-type)
	(setf *valid-planner* 1)
	(setf *valid-problem* 1)
	(if *utilityp*
	    (save-utility))))))


;;
;; They are created to have a way of controlling utility through
;; several problems execution when there is no induction
;;
(DEFUN create-artificial-les (domain type)
  (let (new-le name preconds)
    (clrhash *cr-to-le-hash*)
    (setf *le-index* 0)
    (setf *le-list* nil)
    (domain domain :compile t :function :compile)
    (load (cr-file))
;;     (if *best-first-search-p*
;; 	(load-a-star-rule))
    (load-domain)
    (setf *rules* (all-control-rules))
    (setf *cr-list* (copy-list *rules*))
    (dolist (rule *rules*)
      (setq new-le
	    (make-learning-episode :name (get-new-le-name)
				   :type type
				   :rule rule
				   :problem "initial-problem"
				   :from nil
				   :prior-goal nil
				   :learning-steps 1
				   :matching-time 0
				   :matchings 0
				   :num-nodes 0
				   :firings 0
				   :preconds-pool nil))
      (setq name (p4::control-rule-name rule))
      (setf (gethash name *cr-to-le-hash*)
	    new-le)
      (incf *le-index*)
      (when (and (eq type 'deduced)
		 (or *inductionp* *utilityp*))
	  (store-deduced-rule rule (find-decision-tree-node rule))
	(incf *number-deduced-rules*)
	(setq preconds (get-real-preconds rule))
	(if *save-examples-p*
	    (save-example-for-foil preconds (get-real-effects rule) t))
	(push new-le *le-list*)
	(if *inductionp*
	    (update-frequencies preconds
				(find-decision-tree-node rule)
				t))))
    
    (setf *cr-list* nil)))

;;;
;;; Easy running of hamlet
;;; Problem can be either:
;;; 1. string (pathname to file)
;;;    "/usr/.../x.lisp"
;;; 2. atom (name of problem file that should be in the probs directory
;;;    x
;;; 3. list (the definition of the problem itself)
;;;      (setf (current-problem) (create-problem ...))
;;;
(DEFUN simple-hamlet (domain problem &key (rules-file "rules")
		                          (output-file "hamlet-output")
					  (load-previous-rules-p nil)
					  (multiple-sols-p nil)
					  (time-bound 200)
					  (eager-mode 'eager-partial)
					  (optimal-learning-p nil)
					  (refine-rules-p nil)
					  (linearp nil)
					  (cost-bound most-positive-fixnum)
					  (cost-type nil))
  (declare (ignore cost-bound)
	   (special p4::*first-selection-p*))
  (setf *keep-all-selections-p* t)
  (setf p4::*first-selection-p* nil)
  (setf *inductionp* nil)
  (setf *learnp* t)
  (setf *optimal-learning-p* optimal-learning-p)
  (setf *rules* nil)
  (domain domain :function :compile)
  (reset-domain-variables nil domain)
  (setf *hamlet-results-output* (format nil "~(~ahamlet-~a.~a~)" *domain-directory*
					output-file *lisp-extension*))
  (setf *prodigy-results-output* (format nil "~(~aprodigy-~a.~a~)" *domain-directory*
					 output-file *lisp-extension*))
  (setf *hamlet-output* (format nil "~(~a~a.~a~)" *domain-directory*
				output-file *lisp-extension*))
  (setf *cr-file* (format nil "~a.~a" rules-file *lisp-extension*))
  (if *utilityp* (init-for-utility))
  (my-init-prodigy-result)
;;  (when load-previous-rules-p
;;    (domain *domain-name* :function :compile)
;;    (load (cr-file))
;;    (load-domain))
;;  (with-open-file (stream *hamlet-output*
;;			  :direction :output :if-exists :append
;;			  :if-does-not-exist :create)
;;      (format stream "~2%(setf *learn-~d*~%      '(" time-bound))
  (ejecutar-probset t nil eager-mode (list problem)
		    time-bound (format nil "*learn-~d*" time-bound)
		    load-previous-rules-p linearp multiple-sols-p
		    'without-induction cost-type)
  (if refine-rules-p
      (refine-rules)))

#|
  (with-open-file (stream *hamlet-output*
			  :direction :output :if-exists :append
			  :if-does-not-exist :create)
      (format stream "))~%")))

  		    (if (and multiple-sols-p *optimal-learning-p*)
			'without-induction))
|#

(defun crunit (&optional (domain 'logistics))
  ;;; this overrides the value of *to-do* in case there is such a file
  (let ((filename (format nil "~a~(~a~)/to-do.~a"
			  *world-path* domain *lisp-extension*)))
    (if (probe-file filename)
	(load filename))
    ;; stops when it has done everything
    (when *to-do*
      (with-open-file (file filename
			    :direction :output :if-exists :supersede
			    :if-does-not-exist :create)
	(format file ";;; What rests to do~%(setf *to-do*~%")
	(let ((*print-pretty* t))
	  (format file "      '~(~s~))~%" (cdr *to-do*))))
      (eval (car *to-do*))
      (if *save-examples-p*
	  (with-open-file (file filename
				:direction :output :if-exists :append
				:if-does-not-exist :create)
	    (format file "~2%(setf *example-number* ~d)~%" *example-number*))))
    #+(or CLISP CMU LISPWORKS (and LUCID LCL4.0 AIX)) (quit)
    #+(and ALLEGRO (or SPARC RS6000 LINUX86))
    (exit)))


(DEFUN ask-for-problem (interactivep)
  (cond (interactivep
	 (format t "~2%New problem file (finish with nil):~%  ")
	 (read))
	(t (let ((prob (car *learn-probs*)))
	     (setf *learn-probs* (cdr *learn-probs*))
	     prob))))

;;; ************************************************************
;;;   Auxiliary functions
;;;

(defun load-hamlet-source (&optional (files *ipss-files*)
				     (load-path *ipss-path*))
  (dolist (file files)
    (load (format nil "~a~(~a~).~a" load-path file *source-extension*))))

(defun load-and-compile (&optional (files *ipss-files*)
			 &key (load-path *ipss-path*)
			      (compile-path *ipss-path*)
			      (re-load-p t))
  (if re-load-p
      (load-hamlet-source files load-path))
  (dolist (file files)
    (compile-file (format nil "~a~(~a~).~a" load-path file *source-extension*)
		  :output-file (format nil "~a~(~a~).~a" compile-path
				       file *binary-extension*))
    (load (format nil "~a~(~a.~a~)" compile-path file *binary-extension*))))

;; This allows to compile both Prodigy and Hamlet, and also loads the compiled
;; files
(defun load-and-compile-prodigy-hamlet
  (&optional (files *ipss-files*)
	     &key (load-path *ipss-path*)
	     (compile-path *ipss-path*)
	     (re-load-p t))
  (setf *load-prodigy-immediately* nil)
  (load (concatenate 'string *prodigy-root-directory* "loader."
		     *source-extension*))
  (load-source)
  (if re-load-p
      (load-hamlet-source files load-path))
  (compile-all)
  (p4-load-compiled)
  (load-hamlet-source files load-path)
  (dolist (file files)
    (compile-file (format nil "~a~(~a~).~a" load-path file *source-extension*)
		  :output-file (format nil "~a~(~a~).~a" compile-path
				       file *binary-extension*))
    (load (format nil "~a~(~a.~a~)" compile-path file *binary-extension*))))

;;
;; Only valid for learning data. If testing data, change the cddr for l.
;;
(defun prepare-for-removing (test main-problem failure-positions
				  &optional (print-partial-results-p nil)
				  (stream t)
				  (create-spoof-files-p nil)
				  (create-nodes-ratio-list-p nil))
  (let ((test1 nil)
	(test2 nil)
	(learning-time 0)
	(rest-list nil)
	(with nil))
    (do ((l test))
	((null l))
;;      (if (= (length (car l)) 2)
;;	  (setq l (cdr l)))
      (do* ((l-aux l (cdr l))
	    (first-l-aux (car l-aux) (car l-aux))
	    (length-first-l-aux (length first-l-aux) (length first-l-aux)))
	  ((= length-first-l-aux 16))
	(if (= length-first-l-aux 7)
	    (incf learning-time (nth 6 first-l-aux))))
      (push (car l) test1)
      (do ((l-aux (cddr l) (cdr l-aux)))
	  ((or (null l-aux)
	       (and (not (eq (caar l-aux) (caar test1)))
		    (= (length (car l-aux)) 16)))
	   (push with test2)
;;	   (print with)
;;	   (print learning-time)
;;	   (read)
	   (setq l rest-list)
	   (setq with nil rest-list nil))
	(case (length (car l-aux))
	  (16 (setq with (car l-aux))
	      (setq rest-list (cdr l-aux)))
	  (7 (incf learning-time (nth 6 (car l-aux))))
	  (t nil))))
    (cons learning-time
	  (remove-unsolved (nreverse test1) (nreverse test2) main-problem
			   failure-positions print-partial-results-p
			   stream create-spoof-files-p
			   create-nodes-ratio-list-p))))

#|
      (let ((with (car (member (caar l) (reverse (cdr l)) :test #'eq
			       :key #'car))))
	(push with test2)
	(setq l (member with (cdr l)))))
    (remove-unsolved (nreverse test1) (nreverse test2) main-problem
		     failure-positions stream create-spoof-files-p
		     create-nodes-ratio-list-p)))
|#

;;
;; High level function for automatic comparison of two test sets, and
;; printing on a file given by filename
;;
(DEFUN compare-test-sets (pathname-with
			  &key (filename (format nil "results.~a"
						 *lisp-extension*))
			  (domain 'logistics)
			  (goals-list (list 1 2 5 10 20 50))
			  (objects-list (list 5 10 20 50))
			  (member-function #'restricted-problems-p)
			  (pathname-without pathname-with)
			  (create-spoof-files-p nil)
			  (create-nodes-ratio-list-p nil))
  (reset-domain-variables nil domain)
  (with-open-file (stream (format nil "~a~(~a~)" *domain-directory* filename)
			  :direction :output :if-exists :supersede
			  :if-does-not-exist :create)
	(compare-all-results pathname-with
			     :pathname-without pathname-without
			     :print-partial-results-p t
			     :goals-list goals-list
			     :objects-list objects-list
			     :member-function member-function
			     :stream stream
			     :create-spoof-files-p create-spoof-files-p
			     :create-nodes-ratio-list-p
			       create-nodes-ratio-list-p)))


;;
;; Extracts all comparison between the test data of not using the
;; rules stored in the files in the directory pathname-with, and the
;; data using the rules in the directory pathname-without.
;; If print-partial-results-p is t, it will print comparison data for all
;; test sets. If nil, only the partial and the final ones will be printed.
;; The rest of arguments are explained in remove-unsolved
;;
(DEFUN compare-all-results (pathname-with
			    &key (pathname-without pathname-with)
			    (goals-list (list 1 2 5 10 20 50))
			    (objects-list (list 5 10 20 50))
			    (member-function #'restricted-problems-p)
			    (print-partial-results-p nil)
			    (stream t)
			    (create-spoof-files-p nil)
			    (create-nodes-ratio-list-p nil))
  (let ((results nil)
	(same-test-file-p (equal pathname-with pathname-without)))
    (dolist (goals goals-list)
      (let ((partial-results nil))
	(dolist (objects objects-list)
	  (when (funcall member-function goals objects)
	    (let* ((time (* 150 (1+ (truncate goals 10))))
		   (test-name (format nil "t-~d-~d-~d" goals objects time))
		   (without-filename
		    (format nil "~a~(~a.~a~)" pathname-without test-name
			    *lisp-extension*)))
	      (when (probe-file without-filename)
		(load without-filename)
		(let* ((var-name
			(intern (format nil "~:@(*test-~a*~)" test-name)))
		       (test-without (eval var-name))
		       (with-filename (format nil "~a~(~a.~a~)"
					      pathname-with test-name
					      *lisp-extension*)))
		  (if (not same-test-file-p)
		      (load with-filename))
		  (push (if same-test-file-p
			    (results-from-test test-without 'both 3
					       print-partial-results-p
					       stream
					       create-spoof-files-p
					       create-nodes-ratio-list-p)
			    (results-from-test-1 test-without (eval var-name)
						 'both 3
						 print-partial-results-p
						 stream
						 create-spoof-files-p
						 create-nodes-ratio-list-p))
			partial-results))))))
	(when partial-results
	  (setq partial-results (reverse partial-results))
	  (print-results partial-results nil
			 (format nil "Partial from ~a goals" goals)
			 stream)
	  (setq results (append partial-results results)))))
    (print-results (nreverse results) nil
		   (format nil "Total from all test sets") stream)))

(defun restricted-problems-p (goals objects)
  (not (or (and (= goals 20)
		(= objects 5))
	   (and (= goals 50)
		(or (= objects 5)
		    (= objects 10))))))

(defun less-goals-p (goals objects)
  (<= goals objects))

;;
;; Prints the results of the tests on the list test that contains both
;; with and without
;;
(DEFUN results-from-test (test main-problem failure-positions
				  &optional (print-partial-results-p nil)
				  (stream t)
				  (create-spoof-files-p nil)
				  (create-nodes-ratio-list-p nil))
  (let ((test-without nil)
	(test-with nil))
    (do ((l test (cdr l)))
	((null l))
      (push (car l) test-without)
      (setq l (cdr l))
      (push (car l) test-with))
    (remove-unsolved (nreverse test-without) (nreverse test-with)
		     main-problem failure-positions print-partial-results-p
		     stream create-spoof-files-p create-nodes-ratio-list-p)))

;;
;; Prints the results of the tests on the list tests
;;
(DEFUN results-from-test-1 (test-both test-with main-problem failure-positions
				  &optional (print-partial-results-p nil)
				  (stream t)
				  (create-spoof-files-p nil)
				  (create-nodes-ratio-list-p nil))
  (let ((test-without nil))
    (do ((l test-both (cddr l)))
	((null l))
      (push (car l) test-without))
    (remove-unsolved (nreverse test-without) test-with
		     main-problem failure-positions print-partial-results-p
		     stream create-spoof-files-p create-nodes-ratio-list-p)))

;;
;; Prints the results from learning from all learning episodes
;;
(DEFUN results-from-learn (learns main-problem &optional
				  (failure-positions (list 3 4))
				  (print-partial-results-p nil)
				  (stream t)
				  (create-spoof-files-p nil)
				  (create-nodes-ratio-list-p nil))
  (print-results (mapcar #'(lambda (learn)
			     (prepare-for-removing learn main-problem
						   failure-positions
						   print-partial-results-p
						   stream create-spoof-files-p
						   create-nodes-ratio-list-p))
			 learns)
		 t (format nil "Total from learning") stream))

;;
;; Prints results from several runs
;;
(DEFUN print-results (results learningp comment stream)
  (let ((total-time (cons 0 0))
	(total-nodes (cons 0 0))
	(total-acc-nodes (cons 0 0))
	(total-sol-nodes (cons 0 0))
	(total-unsolved (cons 0 0))
	(total-solved 0)
	(total-better (cons 0 0))
	(total-number-ops (cons 0 0))
	(total-number-probs 0)
	(total-learning-time 0))
    (dolist (result results)
      (let ((learning-time (if learningp (car result)))
	    (wo (car result))
	    (w (cadr result)))
	(when learningp
	  (setq result (cdr result))
	  (incf total-learning-time learning-time))
	(incf (car total-time) (car wo))
	(incf (cdr total-time) (car w))
	(incf (car total-nodes) (nth 1 wo))
	(incf (cdr total-nodes) (nth 1 w))
	(incf (car total-acc-nodes) (nth 2 wo))
	(incf (cdr total-acc-nodes) (nth 2 w))
	(incf (car total-sol-nodes) (nth 3 wo))
	(incf (cdr total-sol-nodes) (nth 3 w))
	(incf (car total-unsolved) (nth 2 result))
	(incf (cdr total-unsolved) (nth 3 result))
	(incf total-solved (nth 4 result))
	(incf (car total-number-ops) (nth 5 result))
	(incf (cdr total-number-ops) (nth 6 result))
	(incf (car total-better) (nth 7 result))
	(incf (cdr total-better) (nth 8 result))
	(incf total-number-probs (nth 9 result))))
    (format stream "~2%*****************************************************")
    (format stream "~2%;;; Total results (~a)" comment)
    (format stream "~%;;;   Number solved by both: ~d (~6,2F%)"
	    total-solved (* 100.0 (/ total-solved total-number-probs)))
    (format stream "~%;;;   Number of total problems: ~d" total-number-probs)
    (format stream "~%;;;   Number unsolved by without: ~d (~6,2F%)"
	    (car total-unsolved)
	    (* 100.0 (/ (car total-unsolved) total-number-probs)))
    (format stream "~%;;;   Number unsolved by with: ~d (~6,2F%)"
	    (cdr total-unsolved)
	    (* 100.0 (/ (cdr total-unsolved) total-number-probs)))
    (format stream "~%;;;   Number total ops without: ~d"
	    (car total-number-ops))
    (format stream "~%;;;   Number total ops with: ~d"
	    (cdr total-number-ops))
    (format stream
	    "~%;;;   Number without less ops than with: ~d (~6,2F% solved)"
	    (car total-better)
	    (if (not (zerop total-solved))
		(* 100.0 (/ (car total-better) total-solved))))
    (format stream
	    "~%;;;   Number with less ops than without: ~d (~6,2F% solved)"
	    (cdr total-better)
	    (if (not (zerop total-solved))
		(* 100.0 (/ (cdr total-better) total-solved))))
    (format stream "~%;;;   Time: ~d (wo) ~d (w) ~d (ratio)"
	    (car total-time) (cdr total-time)
	    (if (zerop (cdr total-time))
		0
		(float (/ (car total-time) (cdr total-time)))))
    (format stream "~%;;;   Number nodes: ~d (wo) ~d (w) ~d (ratio)"
	    (car total-nodes) (cdr total-nodes)
	    (if (zerop (cdr total-nodes))
		0
	      (float (/ (car total-nodes) (cdr total-nodes)))))
    (format stream
	    "~%;;;   Number nodes+alternatives: ~d (wo) ~d (w) ~d (ratio)"
	    (car total-acc-nodes) (cdr total-acc-nodes)
	    (if (zerop (cdr total-acc-nodes))
		0
	      (float (/ (car total-acc-nodes) (cdr total-acc-nodes)))))
    (format stream "~%;;;   Number optimum nodes: ~d (wo) ~d (w) ~d (ratio)~%"
	    (car total-sol-nodes) (cdr total-sol-nodes)
	    (if (zerop (cdr total-sol-nodes))
		0
	      (float (/ (car total-sol-nodes) (cdr total-sol-nodes)))))
    (if learningp
	(format stream ";;;   Total learning time: ~d~%" total-learning-time))
    (format stream
	    "~2%*****************************************************")))
  


;;
;; Removes the problems that have not be solved and returns statistics
;; main-problem-time-bound can be:
;;   list: (main-problem time-bound)
;;   atom: main-problem
;; main-problem can be:
;;  both: leaves only the ones that were solved by the two
;;  first: leaves only the ones that were solved only by the second
;;  second: leaves only the ones that were solved only by the first
;; failure-positions is a list of positions to test. Usually, 3
;; (solutionp) or 4 (whole-tree-expanded-p)
;;
(defun remove-unsolved (test1 test2 main-problem-time-bound
			      failure-positions
			      print-partial-results-p stream
			      &optional create-spoof-files-p
			      create-nodes-ratio-list-p)
  (declare (ignore create-spoof-files-p))
  (let ((result1 nil)
	(result2 nil)
	(last-result1 nil)
	(last-result2 nil)
	(main-problem nil)
	(time-bound most-positive-fixnum)
	(minus-acc-time-1 0)
	(minus-acc-nodes-1 0)
	(minus-acc-all-nodes-1 0)
	(minus-acc-solution-nodes-1 0)
	(minus-acc-time-2 0)
	(minus-acc-nodes-2 0)
	(minus-acc-all-nodes-2 0)
	(minus-acc-solution-nodes-2 0)
	(number-solved 0)
	(number-unsolved1 0)
	(number-unsolved2 0)
	(number-ops-1 0)
	(number-ops-2 0)
	(number-better-1 0)
	(nodes-ratio-list-1 nil)
	(nodes-ratio-list-2 nil)
	(number-better-2 0))
    (cond ((listp main-problem-time-bound)
	   (setq main-problem (car main-problem-time-bound))
	   (setq time-bound (cadr main-problem-time-bound)))
	  (t (setq main-problem main-problem-time-bound)))
    (if (atom failure-positions)
	(setq failure-positions (list failure-positions)))
    (mapc #'(lambda (problem1 problem2)
	      (let* ((test1 (failure-positions problem1 failure-positions
					       time-bound))
		     (test2 (failure-positions problem2 failure-positions
					       time-bound))
		     (test (case main-problem
			     (both (and test1 test2))
			     (first test1)
			     (second test2)
			     (t nil)))
		     (copy-problem1 (copy-list problem1))
		     (copy-problem2 (copy-list problem2)))
		(cond (test1
		       (cond (test2
			      (incf number-ops-1 (nth 5 copy-problem1))
			      (incf number-ops-2 (nth 5 copy-problem2))
			      (cond ((> (nth 5 copy-problem2)
					(nth 5 copy-problem1))
				     (incf number-better-1))
				    ((> (nth 5 copy-problem1)
					(nth 5 copy-problem2))
				     (incf number-better-2))
				    (t nil)))
			      (t (incf number-unsolved2))))
		      (t (incf number-unsolved1)
			 (if (not test2)
			     (incf number-unsolved2))))
		(cond (test
		       (when create-nodes-ratio-list-p
			 (push (/ (nth 8 copy-problem1) (nth 12 copy-problem1)
				  1.0)
			       nodes-ratio-list-1)
			 (push (/ (nth 8 copy-problem2) (nth 12 copy-problem2)
				  1.0)
			       nodes-ratio-list-2))
		       (incf number-solved)
		       (decf (nth 7 copy-problem1) minus-acc-time-1)
		       (decf (nth 7 copy-problem2) minus-acc-time-2)
		       (decf (nth 9 copy-problem1) minus-acc-nodes-1)
		       (decf (nth 9 copy-problem2) minus-acc-nodes-2)
		       (decf (nth 11 copy-problem1) minus-acc-all-nodes-1)
		       (decf (nth 11 copy-problem2) minus-acc-all-nodes-2)
		       (decf (nth 13 copy-problem1) minus-acc-solution-nodes-1)
		       (decf (nth 13 copy-problem2) minus-acc-solution-nodes-2)
		       (push copy-problem1 result1)
		       (push copy-problem2 result2))
		      (t (incf minus-acc-time-1 (nth 6 problem1))
			 (incf minus-acc-time-2 (nth 6 problem2))
			 (incf minus-acc-nodes-1 (nth 8 problem1))
			 (incf minus-acc-nodes-2 (nth 8 problem2))
			 (incf minus-acc-all-nodes-1 (nth 10 problem1))
			 (incf minus-acc-all-nodes-2 (nth 10 problem2))
			 (incf minus-acc-solution-nodes-1 (nth 12 problem1))
			 (incf minus-acc-solution-nodes-2
			       (nth 12 problem2))))))
	  test1 test2)
    (setq last-result1 (car result1))
    (setq last-result2 (car result2))
    (setq result1 (reverse result1))
    (setq result2 (reverse result2))
    (when print-partial-results-p
      (format stream
	      "~2%;;; Results:~%;;;   Number solved by both: ~d" number-solved)
      (format stream "~%;;;   Number of total problems: ~d" (length test1))
      (format stream "~%;;;   Number unsolved by without: ~d" number-unsolved1)
      (format stream "~%;;;   Number unsolved by with: ~d" number-unsolved2))
    (when (and last-result1 last-result2)
      (when print-partial-results-p
	(format stream "~%;;;   Number total ops without: ~d" number-ops-1)
	(format stream "~%;;;   Number total ops with: ~d" number-ops-2)
	(format stream "~%;;;   Number without less ops than with: ~d"
		number-better-1)
	(format stream "~%;;;   Number with less ops than without: ~d"
		number-better-2)
	(format stream "~%;;;   Time: ~d (wo) ~d (w)"
		(nth 7 last-result1) (nth 7 last-result2))
	(format stream "~%;;;   Number nodes: ~d (wo) ~d (w)"
		(nth 9 last-result1) (nth 9 last-result2))
	(format stream "~%;;;   Number nodes+alternatives: ~d (wo) ~d (w)"
		(nth 11 last-result1) (nth 11 last-result2))
	(format stream "~%;;;   Number optimum nodes: ~d (wo) ~d (w)~%"
		(nth 13 last-result1) (nth 13 last-result2)))
      
      (when create-nodes-ratio-list-p
	(setq nodes-ratio-list-1 (reverse nodes-ratio-list-1))
	(setq nodes-ratio-list-2 (reverse nodes-ratio-list-2))
;;      (format stream "~%;;; Nodes-ratio (wo): ~a" nodes-ratio-list-1)
	;;      (format stream "~%;;; Nodes-ratio (w): ~a" nodes-ratio-list-2)
	))

    (list (general-info last-result1) (general-info last-result2)
	  number-unsolved1 number-unsolved2 number-solved
	  number-ops-1 number-ops-2
	  number-better-1 number-better-2 (length test1))))

;; 	(write-to-spoof (list (cons "without" nodes-ratio-list-1)
;; 			      (cons "with" nodes-ratio-list-2))
;; 			"nodes-ratio.spoof"
;; 			"Number of problems"
;; 			"Number nodes/Number nodes solution path")

;;       (when create-spoof-files-p
;; 	(write-to-spoof (list (cons "without"
;; 				    (create-list-from-pos result1 7))
;; 			      (cons "with" (create-list-from-pos result2 7)))
;; 			"time.spoof"
;; 			"Number of problems"
;; 			"Time (seconds)")
;; 	(write-to-spoof (list (cons "without"
;; 				    (create-list-from-pos result1 9))
;; 			      (cons "with" (create-list-from-pos result2 9)))
;; 			"nodes.spoof"
;; 			"Number of problems"
;; 			"Number of nodes explored")
;; 	(write-to-spoof (list (cons "without"
;; 				    (create-list-from-pos result1 11))
;; 			      (cons "with"
;; 				    (create-list-from-pos result2 11)))
;; 			"all-nodes.spoof"
;; 			"Number of problems"
;; 			"Number of nodes generated")
;; 	(write-to-spoof (list (cons "without"
;; 				    (create-list-from-pos result1 13))
;; 			      (cons "with"
;; 				    (create-list-from-pos result2 13)))
;; 			"nodes-solution.spoof"
;; 			"Number of problems"
;; 			"Number of nodes in solution path"))
;;       

(DEFUN general-info (list)
  (if list
      (mapcar #'(lambda (position) (nth position list))
	      (list 7 9 11 13))
      (list 0 0 0 0)))

(DEFUN failure-positions (problem failure-positions time-bound)
  (and (<= (nth 6 problem) time-bound)
       (every #'(lambda (pos)
		  (nth pos problem))
	      failure-positions)))

(DEFUN sort-problems (probs-list results position &optional
		       (output-file (format nil "~aprobsets/sorted-~a.~a"
					    *domain-directory* results
					    *lisp-extension*)))
  (setq results (eval results))
  (sort probs-list #'(lambda (prob1 prob2)
		       (let* ((name1 (my-problem-name prob1 'list))
			      (name2 (my-problem-name prob2 'list))
			      (result1 (member name1 results :key #'car))
			      (result2 (member name2 results :key #'car)))
			 (< (nth position result1) (nth position result2)))))
  (with-open-file (stream output-file :direction :output
			  :if-exists :supersede :if-does-not-exist :create)
     (let ((*print-pretty* t))
       (format stream ";;; Sorted problems~%(setq sorted '~a)~%" probs-list))))

(DEFUN shuffle-problems (file)
  (let ((filename (format nil "~(~aprobsets/~a.~a~)" *domain-directory* file
			  *lisp-extension*)))
    (load filename)
    (setf *all-problems* (shuffle *all-problems*))
    (with-open-file (stream (format nil "~(~aprobsets/shuffled-~a.~a~)"
				    *domain-directory* file *lisp-extension*)
			    :direction :output :if-exists :supersede
			    :if-does-not-exist :create)
	(format stream ";;; Shuffled problems from ~a" filename)
	(let ((*print-pretty* t))
	  (format stream "~2%(setf *all-problems* '~a)" *all-problems*)))))

;;; from Manuela
(defun shuffle (l)
  (do ((result nil)
       (laux l))
      ((null laux) result)
    (let ((nelt (nth (random (length laux)) laux)))
      (push nelt result)
      (setq laux (remove-if #'(lambda (el) (eq el nelt)) laux)))))

#|
(defun shuffle (l)       
  (let ((res (copy-list l))
	(length (length l)))
    (dotimes (i length res)
      (let ((nelt (nth (random length) res)))
	(setq res (cons nelt (remove nelt res)))))))

|#

(DEFUN save-all-best-results nil
  (dolist (goals '(1 2 5 10 20 50))
    (dolist (objects '(5 10 20 50))
      (unless (or (and (= goals 20)
		       (= objects 5))
		  (and (= goals 50)
		       (or (= objects 5)
			   (= objects 10))))
	(let* ((time (* 150 (1+ (truncate goals 10))))
	       (test-name (format nil "t-~d-~d-~d" goals objects time)))
	  (load (format nil "~(~alast-results/results-final-rules/~a.~a~)"
			*domain-directory* test-name *lisp-extension*))
	  (let* ((var-name (intern (format nil "~:@(*test-~a*~)" test-name)))
		 (test-without (eval var-name)))
	    (load (format nil "~(/../schubert/usr/dborrajo/domains/logistics/fall94-results/~a~)" 
			  test-name))
	    (save-best-results test-name test-without (eval var-name))))))))

;;
;; Saves in a file a list of problem descriptions with the best
;; results in a given test set
;;
(DEFUN save-best-results (testname &rest lists)
  (declare (special *best-results*))
  (let ((file (format nil "~(~abest-results~).~a" *domain-directory*
		      *lisp-extension*))
	(old-test-results nil)
	(simple-lists nil)
	(num-problems (apply #'min (mapcar #'length lists))))
    (if (probe-file file)
	(load file))
    (when *best-results*
      (setq old-test-results
	    (find-if #'(lambda (test-name)
			 (eq test-name testname))
		     *best-results* :key #'car))
      (when old-test-results
	(setq lists (cons old-test-results lists))
	(setq *best-results* (remove old-test-results *best-results*))
	(setq old-test-results nil)))
    (dolist (list lists)
      (if (or (> (length list) num-problems)
	      (eq (caar list) (caadr list)))
	  (let ((list-without nil)
		(list-with nil))
	    (do ((l list (cdr l)))
		((null l)
		 (push (nreverse list-without) simple-lists)
		 (push (nreverse list-with) simple-lists))
	      (push (car l) list-without)
	      (setq l (cdr l))
	      (push (car l) list-with)))
	  (push list simple-lists)))
    (dotimes (i num-problems)
      (let* ((i-elements (mapcar #'(lambda (list) (nth i list)) simple-lists))
	     (first-element (car i-elements)))
	(push (list (nth 0 first-element)
		    (nth 1 first-element)
		    (nth 2 first-element)
		    (some #'(lambda (l) (nth 3 l)) i-elements)
		    (some #'(lambda (l) (nth 4 l)) i-elements)
		    (get-min-or-0 5 i-elements)
		    (get-min-or-0 6 i-elements)
		    (apply #'min (mapcar #'(lambda (l) (nth 7 l)) i-elements))
		    (apply #'min (mapcar #'(lambda (l) (nth 8 l)) i-elements))
		    (apply #'min (mapcar #'(lambda (l) (nth 9 l)) i-elements))
		    (apply #'min (mapcar #'(lambda (l) (nth 10 l)) i-elements))
		    (apply #'min (mapcar #'(lambda (l) (nth 11 l)) i-elements))
		    (get-min-or-0 12 i-elements)
		    (get-min-or-0 13 i-elements)
		    (apply #'max (mapcar #'(lambda (l) (nth 14 l)) i-elements))
		    (apply #'max
			   (mapcar #'(lambda (l) (nth 15 l)) i-elements)))
	      old-test-results)))
    (push (cons testname (nreverse old-test-results))
	  *best-results*)
    (with-open-file (stream file :direction :output :if-exists :supersede
			    :if-does-not-exist :create)
       (format stream ";;; Best results for test~%")
       (format stream "(setf *best-results*~%")
       (let ((*print-pretty* t))
	 (format stream "  '~a)" *best-results*)))))

;;
;; Returns the min of the nth elements of the list i-elements, or 0 if
;; the min is 0
;;
(DEFUN get-min-or-0 (n i-elements)
  (let ((min (apply #'min (mapcar #'(lambda (l)
				      (if (zerop (nth n l))
					  most-positive-fixnum
					  (nth n l)))
				  i-elements))))
    (if (< min most-positive-fixnum)
	min
        0)))

#|
;;; New version
  
(DEFUN execute-one-problem (problem learnp eager-mode time-bound reload-rules-p
				    linearp multiple-sols-p continuous
				    functions-not-loaded-p)
  (let ((previous-solution-p nil)
	(old-number-ops 0)
	(old-time 0)
	(old-depth 1000)
	(shorterp nil))
;;    (when *not-trace-trees-p*)
    (setf *old-tree* nil)
    (setf *real-t-tree* nil)
    (if (and continuous (not (eq continuous 'without-induction))
	     (not (eq continuous 'only-rules-testing)))
	(let ((old-learnp *learnp*))
	  ;; this is done to gain space
	  (domain *domain-name* :function (if functions-not-loaded-p :compile))
	  (setq functions-not-loaded-p nil)
	  (setf *learnp* nil)
	  (setf *optimal-learning-p* t)
	  (cond (*fast-learning-p*
		 ;; I'm not sure about this yet
		 (setq shorterp t)
		 (setf *multiple-solutions-flg* :shorter))
		((eq continuous 'pure-testing)
		 (setf *multiple-solutions-flg* :strict-shorter))
		(t nil))
	  (format t "~%***************** Executing without rules...~%")
	  (setf *store-state-p* (or learnp (eq continuous 'learn-testing)))
	  (setq old-time
		(cdr (my-init problem t 'eager-partial time-bound
			      multiple-sols-p
			      (list 'without) linearp nil nil nil nil
			      0 nil old-depth)))
	  (setq old-number-ops *number-ops*)
	  (setf *learnp* old-learnp)
	  (setq previous-solution-p (solutionp))
	  (if (and previous-solution-p
		   (not (eq continuous 'pure-testing)))
	      (setq old-depth
		    (if (= old-number-ops 1)
			8
		        (+ 3 (* old-number-ops 4)))))
	  (if (not (eq continuous 'pure-testing))
	      (setf *old-tree* *real-t-tree*))))
    (do ((repeatp t))
	((not repeatp))
      (format t "~%***************** Executing...~%")

      ;; The output of my-init is whether to refine or not
      ;; I can't make it strict-shorter always because, otherwise, it
      ;; would consider the other solutions of the same length as
      ;; negative examples
      (cond ((eq continuous 'without-induction)
	     (setf *multiple-solutions-flg* :strict-shorter))
	    ((and *fast-learning-p*
		  (not (eq continuous 'pure-testing))
		  (not (eq continuous 'only-rules-testing)))
	     ;;; I think I can do it now safely, since I already check
	     ;;; the needed things in the induction
	     (setf *multiple-solutions-flg* :strict-shorter))
	    (t nil))
      ;; it is not needed to store, since we can access it directly
      (setf *store-state-p* nil)
      (setq repeatp
	    (car (my-init problem learnp eager-mode time-bound multiple-sols-p
			  (if continuous
			      (if (or (eq continuous 'only-rules-testing)
				      (eq continuous 'pure-testing)
				      (eq continuous 'without-induction))
				  (list continuous)
				  (list 'with
					(and previous-solution-p t)
					old-number-ops
					old-time
					(eq continuous 'learn-testing)
					shorterp))
			    (list t))
			  linearp reload-rules-p
			  nil nil functions-not-loaded-p 0 nil old-depth)))
      (if *stop-after-problem-p* (break))
      (setf p4::*all-solutions* nil))))
|#

