;;; ******************************************************************
;;;
;;;   Macros of induction.lisp
;;;

(in-package "USER")

(eval-when (eval load compile)

;;
;; Creates a new binding list with new vars
;;	   
  (DEFMACRO create-new-binding (binding)
    ` (mapcar #'(lambda (bind-pair)
		  (cons (cdr bind-pair)
			(new-var-from-new-var (cdr bind-pair))))
	      ,binding))
  
  (DEFMACRO prior-goal-meta-pred-p (precond)
    `(eq (car ,precond) 'prior-goal))

;;     `(or (eq (car ,precond) 'prior-goal)
;;          (eq (car ,precond) 'any-prior-goal)))

  (DEFMACRO get-all-operators nil
    `(let ((all-the-ops nil))
       (dolist (the-op (p4::problem-space-operators *current-problem-space*))
	 (push (p4::operator-name the-op) all-the-ops))
       (dolist (the-op (p4::problem-space-lazy-inference-rules
			*current-problem-space*)
		       (cons '*finish* (nreverse all-the-ops)))
	 (push (p4::operator-name the-op) all-the-ops))))

;;        (cons '*finish*
;; 	     (mapcar #'(lambda (op)
;; 			 (p4::operator-name op))
;; 		     (p4::problem-space-operators *current-problem-space*)))))
       

  (DEFMACRO get-all-goals nil
    `(cons 'done (p4::problem-space-all-preds *current-problem-space*)))

;
;; Returns the fired control rule for the node if there was any
;  
  (DEFMACRO node-fired-crs (node)
    `(getf (p4::nexus-plist ,node) 'fired-crs))

  
  (DEFMACRO my-union (goals1 goals2)
    `(cond ((and ,goals1 ,goals2)
	    (union ,goals1 ,goals2 :test #'equal))
	   ((and ,goals1 (null ,goals2))
	    (my-push ,goals1))
	   ((and ,goals2 (null ,goals1))
	    (my-push ,goals2))
	   (t nil)))

  (DEFMACRO my-push (list)
    `(if (member nil ,list :test #'eq)
	 ,list
         (cons nil ,list)))

  (DEFMACRO my-mapcar (function list)
    `(let ((result nil))
       (dolist (element ,list (nreverse result))
	 (push (funcall ,function element)
	       result))))

;;;
;;; Returns all rules of a decision node
;;;  
  (DEFMACRO get-decision-node-rules (node)
    `(decision-node-rules ,node))

;;;
;;; It returns the node in which a rule should be classified in the
;;; decision tree. It is useful for finding out the node without
;;; storing it
;;; 
  (DEFMACRO find-decision-tree-node (rule)
    `(classify-rule ,rule (get-real-preconds ,rule) (get-real-effects ,rule)
		    *decision-tree* nil nil))

)

#|
  (DEFMACRO relevant-intersection-p (intersection)
    `(> (length (remove-if-not #'(lambda (precond)
				   (eq (car precond) 'true-in-state))
			       ,intersection))
	0))
|#

;;********************************************************************
;;;
;;; Macros of hamlet.lisp
;;;

(eval-when (eval load compile)

  (DEFMACRO get-real-preconds (cr)
    `(let ((true-preconds (p4::control-rule-if ,cr)))
       (if (eq (car true-preconds) 'and)
	   (cdr true-preconds)
	   true-preconds)))

  (DEFMACRO get-real-effects (cr)
    `(cdr (p4::control-rule-then ,cr)))

  (DEFMACRO reverse-binding (binding)
    `(mapcar #'(lambda (bind-pair)
		 (cons (cdr bind-pair) (car bind-pair)))
	     ,binding))
  
  (DEFMACRO negatedp (condition)
    `(eq (car ,condition) `~))

  (DEFMACRO get-pos-assertion (assertion)
    `(if (negatedp ,assertion) (cadr ,assertion) ,assertion))

;;
;; Returns a list of visible forms
;;;
   (DEFMACRO real-give-me-nice-state nil
     `(mapcar #'(lambda (lit)
		  (get-visible-form lit))
	       (p4::give-me-nice-state)))
;
;; Returns the object-type of a given object
;
   (DEFMACRO get-object-type (object)
     `(p4::p4type-name (p4::prodigy-object-type
			  (p4::object-name-to-object ,object
			       *current-problem-space*))))


;
;; Returns the type of the node. It is not very straightforward, but..
;

  (DEFMACRO node-type (node)
    `(cond ((p4::goal-node-p ,node) 'goal-node)
	   ((p4::operator-node-p ,node) 'operator-node)
	   ((p4::binding-node-p ,node) 'binding-node)
	   ((p4::applied-op-node-p ,node) 'applied-op-node)
	   (t 'instantiated-op)))

;
;; Returns the value of the learning type for the node. It can be:
;;  success, failure, or unknown
;  
  (DEFMACRO node-learning-type (node)
    `(getf (p4::nexus-plist ,node) 'learning-type))

;
;; Returns the value of the learning type for the node. It can be:
;;  success, failure, or unknown
;  
  (DEFMACRO node-best-child-info (node)
    `(getf (p4::nexus-plist ,node) 'best-child-info))

;
;; Returns the number of nodes under that node in the search tree
;  
  (DEFMACRO node-num-nodes (node)
    `(getf (p4::nexus-plist ,node) 'num-nodes))
  
  
;;
;; Returns the grandfather of node
;;
   (DEFMACRO get-grandpa (node)
     `(p4::nexus-parent (p4::nexus-parent ,node)))

;;
;; I know this is "ad hoc" but it is more efficient
;;   
   (DEFMACRO top-level-goal-p (node literal)
     `(or (member (p4::nexus-name ,node) '(2 5))
	  (eq (p4::nexus-name (p4::nexus-parent ,node)) 4)
	  (let ((instantiate-preconds
		 (p4::binding-node-instantiated-preconds (find-node 4))))
	    (if (listp instantiate-preconds)
		(member ,literal instantiate-preconds)
		(eq ,literal instantiate-preconds)))))
;; 	  (and (eq (p4::nexus-name ,node) 4)
;; 	       (member ,literal
;; 		       (p4::binding-node-instantiated-preconds ,node)))))
   
;;
;; Returns t if node is a top-level-goal
;;
;;  (DEFMACRO top-level-goal-p (literal-node)
;;    `(or (eq (p4::literal-name ,literal-node) 'p4::done)
;;	 (let ((ins-op (or (car (p4::literal-goal-p ,literal-node))
;;		       (car (p4::literal-neg-goal-p ,literal-node)))))
;;	   (and ins-op
;;		(eq 'p4::*finish*
;;		    (p4::rule-name (p4::instantiated-op-op ins-op)))))))
  
;; 	 (member :top-level-goal (p4::literal-goal-p ,literal-node) :test #'eq)
;;	 (member :top-level-goal (p4::literal-neg-goal-p ,literal-node)
;;		 :test #'eq)

  
;
;; Returns the operator name of the node
;
  (DEFMACRO get-operator-name (node type)
    `(p4::operator-name (case ,type
			  (operator-node (p4::operator-node-operator ,node))
			  ((applied-op-node binding-node)
			   (p4::instantiated-op-op
			    (p4::a-or-b-node-instantiated-op ,node)))
			  (instantiated-op (p4::instantiated-op-op ,node))
			  (t 'error))))

;;
;; Returns the common preconds of all success binding nodes children
;; of node
;;
   (DEFMACRO get-common-operator-preconds (node)
     `(let ((common-preconds nil))
	(dolist (binding-node (p4::nexus-children ,node) common-preconds)
	  (if (eq (node-learning-type binding-node) 'success)
	      (let ((binding-preconds (p4::binding-node-instantiated-preconds
					   binding-node)))
		(setq common-preconds
		      (if common-preconds
			  (intersection common-preconds
					(get-binding-preconds binding-preconds)
					:test #'equal)
			  (get-binding-preconds binding-preconds))))))))

;;
;; Returns the list of visible forms of the preconds of a binding node
;;
   (DEFMACRO get-binding-preconds (b-preconds)
     `(if (listp ,b-preconds)
	  (mapcar #'(lambda (precond)
	           (get-visible-form precond))
		      (cdr ,b-preconds))
	  (list (get-visible-form ,b-preconds))))
	 
	 
;;Nueva.lramos.18/02/03	  
 (DEFMACRO get-binding-preconds-in-visible-form (b-preconds)
    `(if (listp ,b-preconds)
	      (mapcar #'(lambda (precond)
	           (get-preconds-in-visible-form precond))
		      (cdr ,b-preconds))
		
	      (list (get-preconds-in-visible-form ,b-preconds))) )	  
	  
;;
;; Returns the list of pending goals (~ goals left on the node)
;; (given that node is a goal or applied node)
;;
  (DEFMACRO get-some-candidate-goals (node goal-to-be-removed)
    `(delete ,goal-to-be-removed
	     (mapcar #'(lambda (literal) (get-visible-form literal))
		     (all-goals-from-node ,node))
	     :test #'equal))

;;
;; Returns a list of goals nodes children of the node
;;
  (DEFMACRO all-goals-from-node (node)
    `(let ((goals-left (p4::a-or-b-node-goals-left ,node))
	   (children-goals nil))
       (append (if (listp goals-left) goals-left)
	       (dolist (child (p4::nexus-children ,node) children-goals)
		 (if (p4::goal-node-p child)
		     (push (p4::goal-node-goal child) children-goals)))
	       nil)))
				     
;; 
;; Returns the chosen goal on the node (given that node is a goal-node)
;;
  (DEFMACRO get-target-goal (node)
    `(get-visible-form (p4::goal-node-goal ,node)))


;;
;; It returns a list of the object names that appear in the seq  
  (DEFMACRO convert-objects-in-constants (seq)
    `(map 'list #'(lambda (arg) (convert-object-in-name arg)) ,seq))

;;
;; Returns the name of the object
;;
   (DEFMACRO convert-object-in-name (object)
     `(if (p4::prodigy-object-p ,object)
	  (p4::prodigy-object-name ,object)
	  ,object))

;;
;; Returns the visible form of a literal node
;;  
  (DEFMACRO get-visible-form (literal-node)
    `(if (p4::literal-neg-goal-p ,literal-node)
	 (cons '~ (list (get-pos-visible-form ,literal-node)))
         (get-pos-visible-form ,literal-node)))
          
   
(DEFMACRO get-preconds-in-visible-form (precond)  
   `(if (listp ,precond)
       (cons '~ (list (get-pos-visible-form (car(cdr precond))) ))
        (get-pos-visible-form ,precond)
    );fin del if
);fin del get-preconds-in-visible-form


;;
;; Returns a visible-form
;;
  (DEFMACRO get-pos-visible-form (literal-node)
    `(cons (p4::literal-name ,literal-node)
	   (convert-objects-in-constants
	    (p4::literal-arguments ,literal-node))))

;; Returns a visible solution
  (defmacro get-visible-solution (solution)
    `(mapcar #'(lambda (instop)
		 (get-applicable-op instop))
	     ,solution))

;;
;; Returns the visible form of a literal node, but with the objects as
;; Prodigy pointers
;;  
  (DEFMACRO get-mixed-visible-form (literal-node)
    `(if (p4::literal-neg-goal-p ,literal-node)
	 (cons '~ (list (get-mixed-pos-visible-form ,literal-node)))
         (get-mixed-pos-visible-form ,literal-node)))

;;
;; Returns a visible-form
;;
  (DEFMACRO get-mixed-pos-visible-form (literal-node)
    `(cons (p4::literal-name ,literal-node)
	   (map 'list #'(lambda (arg) arg)
		(p4::literal-arguments ,literal-node))))

  
;;
;; Returns the applicable operator.
;;
  (DEFMACRO get-applicable-op (instantiated-op)
    `(cons (get-operator-name ,instantiated-op 'instantiated-op)
	   (convert-objects-in-constants
	    (p4::instantiated-op-values ,instantiated-op))))

;
;; Returns the bindings of a binding node. I'm not sure whether this
;; will work.
;
  (DEFMACRO get-bindings-from-node (node)
    `(let ((instantiated-op (p4::binding-node-instantiated-op ,node)))
       (mapcar #'(lambda (var cte)
		   (cons var (convert-object-in-name cte)))
		(p4::operator-vars (p4::instantiated-op-op instantiated-op))
		(p4::instantiated-op-values instantiated-op))))
)
(in-package "P4")
;; Returns the bindings of a binding node. I'm not sure whether this
;; will work. Svegas.
;
(eval-when (eval load compile)
  (DEFMACRO get-bindings-from-basic-node (node)
    `(let ((instantiated-op (binding-node-instantiated-op ,node)))
       (mapcar #'(lambda (var cte) 
		   (cons var (user::convert-object-in-name cte)))
	       (operator-vars (instantiated-op-op instantiated-op))
	       (instantiated-op-values instantiated-op))))
  
  ;; Svegas.
  ;; Daniel. Taking the car might mean removing alternatives
  (DEFMACRO get-bindings-from-applied-node (node)
    `(let ((instantiated-op (applied-op-node-instantiated-op ,node)))
       (mapcar #'(lambda (elemen)
		   (cons (car elemen) 
			 (user::convert-object-in-name (cdr elemen))))   
	       (car 
		(form-all-bindings 
		 (ligadas (reverse 
			   (get-bindings-from-basic-node 
			    (instantiated-op-binding-node-back-pointer
			     (applied-op-node-instantiated-op node)))))
		 (append (operator-vars 
			  (instantiated-op-op instantiated-op)) 
			 (getf (operator-plist 
				(instantiated-op-op instantiated-op)) 
			       :costs-vars))
		 (instantiated-op-op instantiated-op)
		 (list (make-my-list  
			(operator-vars 
			 (instantiated-op-op instantiated-op)) 
			(instantiated-op-values instantiated-op)))
		 t)))))
  
    (DEFMACRO get-bindings-from-bindings-node (node)
    `(let ((instantiated-op (a-or-b-node-instantiated-op ,node)))
       (mapcar #'(lambda (elemen)
		   (cons (car elemen) 
			 (user::convert-object-in-name (cdr elemen))))   
	       (car 
		(form-all-bindings 
		 (ligadas (reverse (get-bindings-from-basic-node node)))
		 (append (operator-vars (instantiated-op-op instantiated-op)) 
			 (getf (operator-plist (instantiated-op-op instantiated-op)) 
			       :costs-vars))
		 (instantiated-op-op instantiated-op)
		 (list (make-my-list  
			(operator-vars (instantiated-op-op instantiated-op)) 
			(instantiated-op-values instantiated-op)))
		 t)))))

)

(in-package "USER")

(eval-when (eval load compile)
;
;; Checks if the node belongs to a success path
;
  (DEFMACRO success-path-p (node)
    `(eq (getf (p4::nexus-plist ,node) :termination-reason) :achieve))

;
;; Comprueba si dos conjuntos son iguales
;
  (DEFMACRO equal-set-p (set1 set2 &key (test 'equal))
    `(null (set-exclusive-or ,set1 ,set2 :test #',test)))

;
;; Tests whether it expanded the whole tree and did not find the solution
;
  (DEFMACRO whole-tree-expanded-p nil
    `(and (prodigy-result-whole-tree-expanded-p *prodigy-result*)
	  t))

;
;; Tests whether it expanded the whole tree and did not find the solution
;
  (DEFMACRO whole-tree-expanded-no-solution-p nil
    `(and (prodigy-result-whole-tree-expanded-p *prodigy-result*)
	  (not (prodigy-result-solutionp *prodigy-result*))))

;
;; Returns t if there is a solution path
;  
  (DEFMACRO solutionp nil
    `(prodigy-result-solutionp *prodigy-result*))

;
;; Returns the name of the control rules file
;
  (DEFMACRO cr-file nil
    `(format nil "~a~a" *domain-directory* *cr-file*))

  ;
;; Returns the name of the FF domain dependent control rules file
;
  (DEFMACRO ff-cr-file nil
    `(format nil "~a~a" *domain-directory* *ff-cr-file*))


;;
;; Returns the vars of a control rule
;;
   (DEFMACRO get-vars-from-cr (cr)
     `(getf (p4::control-rule-plist ,cr) 'vars))
)  

#|
;;; Old versions  
  (DEFMACRO solutionp nil
    `(if (p4::problem-space-property :multiple-sols)
         (if *result*
	     (cdr p4::*all-solutions*)
	     (cdar p4::*all-solutions*))
         (and (listp (car *result*))
	      (listp (caar *result*))
	      (eq (cdaar *result*) :achieve))))

;
;; Tests whether it expanded the whole tree and did not find the solution
;
  (DEFMACRO whole-tree-expanded-no-solution-p nil
    `(and (p4::problem-space-property :multiple-sols)
	  *result*
	  (eq (car *result*) :fail)
	  (not (cdr p4::*all-solutions*))))

;;
;; Tests whether it expanded the whole tree
;; Another way of doing this is by checking whether the time<time-bound
;  
  (DEFMACRO whole-tree-expanded-p nil
    `(and (p4::problem-space-property :multiple-sols)
	  (or (and *result*
		   (eq (car *result*) :fail))
	      (not *result*))))

;;
;; Returns the list of applicable operators. 
;;
  (DEFMACRO get-applicable-ops (node)
    `(mapcar #'(lambda (op) (get-operator-name op (node-type op)))
	      (append (remove-if-not #'p4::applied-op-node-p
				      (p4::nexus-children ,node))
		      (p4::a-or-b-node-applicable-ops-left ,node)
		      nil)))
|#

;;********************************************************************
;;;
;;; Macros of load-problems.lisp
;;;

(eval-when (load eval compile)

   (DEFMACRO prodigy-result-acc-time-with nil
     `(getf (prodigy-result-plist *prodigy-result*) 'acc-time-with))
     
   (DEFMACRO prodigy-result-acc-nodes-with nil
     `(getf (prodigy-result-plist *prodigy-result*) 'acc-nodes-with))
     
   (DEFMACRO prodigy-result-all-nodes nil
     `(getf (prodigy-result-plist *prodigy-result*) 'all-nodes))

   (DEFMACRO prodigy-result-acc-all-nodes nil
     `(getf (prodigy-result-plist *prodigy-result*) 'acc-all-nodes))
     
   (DEFMACRO prodigy-result-acc-all-nodes-with nil
     `(getf (prodigy-result-plist *prodigy-result*) 'acc-all-nodes-with))
     
   (DEFMACRO prodigy-result-nodes-solution-path nil
     `(getf (prodigy-result-plist *prodigy-result*) 'nodes-solution-path))

   (DEFMACRO prodigy-result-acc-nodes-solution-path nil
     `(getf (prodigy-result-plist *prodigy-result*) 'acc-nodes-solution-path))
     
   (DEFMACRO prodigy-result-acc-nodes-solution-path-with nil
     `(getf (prodigy-result-plist *prodigy-result*)
	    'acc-nodes-solution-path-with))
     
   (DEFMACRO elapsed-time (initial-time)
     `(float (/ (- (get-internal-run-time) ,initial-time)
		internal-time-units-per-second)))
;; )
;; 
;; (in-package "PRODIGY4")
;; (eval-when (load eval compile)

;;; From Sira
   (DEFMACRO additive (&rest labels) 
     (declare (special *current-problem-space* *costs-declared*))
     `(progn (setf *costs-declared* t)
	     (setf (getf (p4::problem-space-plist *current-problem-space*) 
			 :additive-costs) ',labels)))

   ;; From Sira
   (DEFMACRO non-additive (&rest labels) 
     (declare (special *current-problem-space* *costs-declared*))
     `(progn (setf *costs-declared* t)
	     (setf (getf (p4::problem-space-plist *current-problem-space*) 
			 :non-additive-costs) ',labels)))

   (defmacro problem-name (problem)
     `(cadr (cadr (caddr ,problem))))

   (defmacro problem-objects (problem)
     `(caddr (caddr ,problem)))

   (defmacro problem-state (problem)
     `(cadr (cddr (caddr ,problem))))

   (defmacro problem-goals (problem)
     `(cadr (cdddr (caddr ,problem))))
   
   (defmacro problem-objects-of-type (type objects)
     `(butlast (car (member ,type ,objects
			    :key #'(lambda (x) (car (last x)))))))

)

;; (export '(additive non-additive))


#|
   (DEFMACRO not-worse-than-before-p (continuous)
     `(and (cdr ,continuous)
	   (whole-tree-expanded-p)
	   (<= *number-ops* (nth 2 ,continuous)))))
|#

