;;; This code allows to load directly a PDDL domain description file and
;;; automatically translate it. It requires to have IPSS loaded and to have a
;;; directory with the name of the domain created under *world-path*

;;; It also translates problems. See below (find translate-problems definition)

;; Example of use:
;; 1. enter lisp and load IPSS; e.g. $ clisp -i ipss/init.lisp
;; 2. load this file; e.g. > (load "pddl2ipss")
;; 3. load a pddl domain file; e.g. (load "satellite.pddl")
;; this code will automatically generate a domain.lisp file in the directory *world-path*/domain-name/
;; so you will have to create first this path

(in-package "USER")

;; (load "~/ipss/init")

(eval-when (load eval compile)
  (defmacro define (what &rest arguments)
    `(if ,(eq (car what) 'domain)

;;;-------------------------sentencia anterior
;;	 (with-open-file (ofile ,(format nil "~a~(~a~)/domain.lisp" *world-path* (cadr what))

;;;-------------------------nueva sentencia

	 (with-open-file (ofile ,(format nil "~a/domain.lisp" *ipss-temporal-dir*)

				:direction :output
				:if-exists :new-version
				:if-does-not-exist :create)
	   (let ((predicates ',(find-argument arguments :predicates))
		 (functions ',(find-argument arguments :functions)))
	     (format ofile "(in-package \"USER\")")
	     (format ofile "~2%(create-problem-space '~a :current t)~%" ',(cadr what))
	     (translate-types-def ',(find-argument arguments :types) ofile)
	     (format ofile "~2%(infinite-type DURATION #'numberp)")
	     (dolist (function functions)
	       (format ofile "~%(infinite-type ~a #'numberp)" (car function)))
	     (format ofile "~2%(additive TIME #'numberp)")
	     (translate-actions-def ',(append (find-all-argument arguments :action)
					      (find-all-argument arguments :durative-action))
				    predicates functions ofile)
	   ))))
    )

(defun find-argument (domain argument)
  (cdr (assoc argument domain)))

(defun find-all-argument (domain argument)
  (remove-if-not #'(lambda (definition)
		     (eq (car definition) argument))
		 domain))

(defun translate-types-def (types ofile)
  (let ((read-types nil) (sub-types nil) 
	(top-types (remove '- types)))
    (do* ((prev-type (car types) type)
	  (rest-types (cdr types) (cdr rest-types))
	  (type (car rest-types) (car rest-types)))
	((null rest-types) nil)
      (cond ((eq prev-type '-)
	     (push (cons type (list (reverse read-types))) sub-types)
	     (setf read-types nil)
	     (setf rest-types (cdr rest-types))
	     (setf type (car rest-types)))
	    (t
	     (push prev-type read-types)
	     (setf top-types (remove prev-type top-types)))))
    (when (or top-types read-types)
	(dolist (this-type (append top-types read-types))
	    (format ofile "~%(ptype-of ~a :top-type)" this-type)))
    (dolist (this-subtype (reverse sub-types))
	(dolist (this-type (second this-subtype))
	    (format ofile "~%(ptype-of ~a ~a)" this-type (car this-subtype))))
    ))

;; It does not work properly for top-types
;; (defun translate-types-def (types ofile)
;;   (do* ((top-types nil)
;; 	(read-types nil)
;; 	(subtypes nil)
;; 	(types-def nil)
;; 	(rest-types types (cdr rest-types))
;; 	(type (car rest-types) (car rest-types)))
;;       ((null rest-types)
;;        (dolist (type (or read-types top-types))
;; 	 (format ofile "~%(ptype-of ~a :top-type)" type))
;;        (dolist (type-def types-def)
;; 	 (format ofile "~a" type-def)))
;;     (cond ((eq type '-)
;; 	   (setq subtypes nil)
;; 	   (dolist (read-type (reverse read-types))
;; 	     (setq top-types (remove read-type top-types))
;; 	     (push read-type subtypes))
;; 	   (setq read-types nil)
;; 	   (setq rest-types (cdr rest-types))
;; 	   (push (car rest-types) top-types)
;; 	   (dolist (subtype subtypes)
;; 	     (push (format nil "~%(ptype-of ~a ~a)" subtype (car rest-types))
;; 		   types-def)))
;; 	  (t (push type read-types)))))

(defun translate-actions-def (actions predicates functions ofile)
  (dolist (action actions)
      (format ofile "~2%(operator ~a" (cadr action))
    (let* ((params (compute-params (getf action :parameters)))
	   (preconds nil)
	   (effects nil)
	   (costs (getf action :duration)))
      (format ofile "~%   (params")
      (dolist (param params)
	(format ofile " ~(~a~)" (car param)))
      (format ofile ")")
      (setq preconds (compute-preconds (or (getf action :precondition)
					   (getf action :condition))
				       predicates
				       functions))
      (format ofile "~%   (preconds ~(~%      ~a~%      ~a~))" (append params *new-vars*) preconds)
      (setq effects (compute-effects (getf action :effect) predicates functions))
      (format ofile "~%   (effects ~(~%      ~a~%      ~a~))" (nreverse *new-vars*) effects)
      (if costs
	  (multiple-value-bind (variables cost-function)
	      (compute-costs costs functions)
	    (format ofile "~%   (costs ~(~%      ~a~%      ~a~))"
		    variables (or cost-function '((TIME <duration>)))))))
    (format ofile "~%)")))

(defun compute-params (params-def)
  (do ((params params-def
		(cdddr params))
	(params-types nil))
      ((null params) (nreverse params-types))
    (push (list (convert-pddl-var (car params))
		(caddr params))
	  params-types)))

(defun convert-pddl-var (pddl-var)
  (intern (format nil "~:@(<~a>~)" (string-left-trim '(#\?) (format nil "~a" pddl-var)))))

(defun compute-preconds (action-def predicates functions)
  (setf *new-vars* nil)
  (or (convert-preconds action-def (mapcar #'car predicates) (mapcar #'car functions))
      '(and)))

(defun convert-preconds (preconds predicates functions)
  (cond ((null preconds) nil)
	((or (and (eq (car preconds) 'over)
		  (eq (cadr preconds) 'all))
	     (and (eq (car preconds) 'at)
		  (or (eq (cadr preconds) 'start)
		      (eq (cadr preconds) 'end))))
	 (convert-preconds (caddr preconds) predicates functions))
	((member (car preconds) predicates) (convert-literal preconds))
	((member (car preconds) '(and or))
		 (cons (car preconds)
		       (convert-list (cdr preconds) predicates functions)))
	((eq (car preconds) 'not)
	 (cons '~ (convert-literal (cdr preconds))))
	(t 
	 (let* ((result (mapcar #'(lambda (precond)
				    (create-numeric-test precond functions))
				(cdr preconds)))
		(var-def (assoc (car result) *new-vars*)))
	   (cond (var-def
		  (setf *new-vars* (remove var-def *new-vars* :test #'equal))
		  (push (list (car result)
			      `(and ,@(cdadr var-def) (,(car preconds) ,@result)))
			*new-vars*)
		  nil)
		 (t (cons (car preconds) result)))))))

(defun create-numeric-test (preconds functions)
  (cond ((null preconds) nil)
	((numberp preconds) preconds)
	((atom preconds) (convert-pddl-var preconds))
	((member (car preconds) functions)
	 (let ((new-var (intern (format nil "~:@(<~a-~d>~)" (car preconds) (gensym)))))
	   (push (list new-var 
		       `(and ,(car preconds)
			     (gen-from-pred (,(car preconds) ,@(convert-vars (cdr preconds)) ,new-var))))
		 *new-vars*)
	   new-var))
	(t (cons (car preconds) (mapcar #'(lambda (precond)
					    (create-numeric-test precond functions))
					(cdr preconds))))))
		 

(defun convert-list (preconds predicates functions)
  (mapcan #'(lambda (precond)
	      (let ((result (convert-preconds precond predicates functions)))
		(if result
		    (list result))))
	  preconds))

(defun compute-effects (effects predicates functions)
  (setf *new-vars* nil)
  (setf *new-effects* nil)
  (let ((standard-effects (convert-effects effects (mapcar #'car predicates) (mapcar #'car functions))))
    (if (or (eq (car standard-effects) 'add)
	    (eq (car standard-effects) 'del))
	(setq standard-effects (list standard-effects)))
    (dolist (new-effect *new-effects*)
      (push new-effect standard-effects))
    (nreverse standard-effects)))

(defun convert-effects (effects predicates functions)
  (cond ((null effects) nil)
	((or (and (eq (car effects) 'over)
		  (eq (cadr effects) 'all))
	     (and (eq (car effects) 'at)
		  (or (eq (cadr effects) 'start)
		      (eq (cadr effects) 'end))))
	 (convert-effects (caddr effects) predicates functions))
	((member (car effects) '(and or))
	 (convert-effects-list (cdr effects) predicates functions))
	((eq (car effects) 'not)
	 (list 'del (convert-literal (cadr effects))))
	((member (car effects) predicates)
	 (list 'add (convert-literal effects)))
	((member (car effects) '(increase decrease assign))
	 (convert-change (cdr effects) (car effects) functions))
	(t nil)))

(defun convert-effects-list (effects predicates functions)
  (if effects
      (let ((result (convert-effects (car effects) predicates functions)))
	(if result
	    (cons result (convert-effects-list (cdr effects) predicates functions))
	  (convert-effects-list (cdr effects) predicates functions)))))
	  
(defun convert-change (effect direction functions)
  (let ((new-var (intern (format nil "~:@(<new-~a-~d>~)" (caar effect) (gensym))))
	(old-var (intern (format nil "~:@(<old-~a-~d>~)" (caar effect) (gensym)))))
    (push (list old-var 
		`(and ,(caar effect)
		      (gen-from-pred (,(caar effect) ,@(convert-vars (cdar effect)) ,old-var))))
	  *new-vars*)
    (push (list new-var 
		`(and ,(caar effect)
		      (ipss-change-value ,old-var ,(convert-quantity (cadr effect) functions) ',direction)))
	  *new-vars*)
    (push `(add (,@(convert-literal (car effect))
		   ,new-var))
	  *new-effects*)
    (push `(del (,@(convert-literal (car effect))
		   ,old-var))
	  *new-effects*)
    nil))

(defun convert-quantity (new-value functions)
  (cond ((numberp new-value) new-value)
	((atom new-value) 
	 (let ((var (convert-pddl-var new-value)))
	   (if (eq var '<duration>)
	       (format t "~%You will have to edit by hand the resulting file, since variable <duration> should only appear on the costs part, not on the effects. This might be because this quantity is really a cost function. If not, then we cannot handle it easily. One way would be to add all duration computation here also, apart from in the cost function."))
	     var))
	((member (car new-value) functions)
	 (let ((new-var (intern (format nil "~:@(<~a>~)" (car new-value)))))
	   (push (list new-var 
		       `(and ,(car new-value)
			     (cost-from-pred (,(car new-value) ,@(convert-vars (cdr new-value)) ,new-var))))
		 *new-vars*)
	   new-var))
	(t (cons (car new-value)
		 (mapcar #'(lambda (formula)
			     (convert-quantity formula functions))
			 (cdr new-value))))))

(defun compute-costs (costs functions)
  (setf *new-vars* nil)
  (if (and (eq (car costs) '=)
	   (eq (cadr costs) '?duration))
      (if (atom (caddr costs))
	  (values nil (list (list 'TIME (caddr costs))))
	(let ((result (convert-costs (caddr costs) (mapcar #'car functions))))
	  (if (p4::strong-is-var-p result)
	      (values (sublis (list (cons '<duration> result)) *new-vars*)
		      nil)
	    (values `(,@*new-vars*
		      (<duration> (and DURATION ,result)))
		    nil))))))

;; We still have to add the last argument for IPSS
(defun convert-costs (costs functions)
  (cond ((null costs) nil)
	((atom costs) costs)
	((member (car costs) functions)
	 (let ((new-var (intern (format nil "~:@(<~a-~d>~)" (car costs) (gensym)))))
	   (push (list new-var 
		       `(and ,(car costs)
			     (cost-from-pred (,(car costs) ,@(convert-vars (cdr costs)) ,new-var))))
		 *new-vars*)
	   new-var))
	(t (cons (car costs)
		 (mapcar #'(lambda (cost)
			     (convert-costs cost functions))
			 (cdr costs))))))

(defun convert-literal (literal)
  (cons (car literal)
	(convert-vars (cdr literal))))

(defun convert-vars (vars)
  (if vars
      (cons (convert-pddl-var (car vars))
	    (convert-vars (cdr vars)))))

;;; This must be included within IPSS code

(defun ipss-change-value (old new-value direction)
  (case direction
    (increase (list (+ old new-value)))
    (decrease (list (- old new-value)))
    (t (list new-value))))

;;; Translates problem descriptions in PDDL into IPSS format (for
;;; training and/or testing of Hamlet).

;; Example of use:
;; 1. enter lisp
;; 2. load this file
;; 3. place pddl problem files probxxx in ~/tmp/, for instance
;; 4. evaluate (translate-problems "~/tmp/" :domain 'blocksworld :file-prefix "prob*")
;; in ~/tmp/ you will find the corresponding IPSS problems

;; types is a list of non-root types
;; supertypes is a list of types that have subtypes
;; output-file-name is the output file
;; all-together-p is t, then all problems will be generated in a single file
;; unwanted-literals is a list of literals that we do not want to appear in the output-file. Usually NIL
(defun translate-problems (directory &key (domain 'logistics)
				     (file-prefix "prob*")
				     (types '(package location airport airplane truck city))
				     (supertypes '(location))
				     (output-file-name "AIPS02-")
				     (all-together-p t)
				     (unwanted-literals (if (eq domain 'driverlog)
							    '(path link time-to-drive
								   time-to-walk))))

;;--------------------------------parte de david montero--------------------------------------------------------------------
;; se ha modificado esta parte para que el fichero de problemas traducido se escriba en un temporal y no en el directorio 
;; de probs. Luego se le da al usuario la opcion de almacenar el fichero resultado donde desee
;;  (let ((ofile (format nil "~(~a~a~).lisp" directory output-file-name)))
    (let ((ofile (format nil "~(~a~a~).lisp" *ipss-temporal-dir* output-file-name)))
;;--------------------------------fin de david montero----------------------------------------------------------------------
    (if all-together-p
	(with-open-file (ostream ofile :direction :output :if-exists :supersede
				 :if-does-not-exist :create)
	  (format ostream "(setf *all-problems* '(")))
    (dolist (file (directory (format nil "~a~a" directory file-prefix)))
	(with-open-file (ostream (if all-together-p
				     ofile
				     (format nil "~(~a-~a~).lisp" ofile (pathname-name file)))
				 :direction :output
				 :if-exists (if all-together-p :append :supersede)
				 :if-does-not-exist :create)
	  (if (not all-together-p)
	      (format ostream "(setf *all-problems* '(~%"))
	  (with-open-file (istream file :direction :input)
	    (pp-list (list (translate_problem (read istream) types supertypes
							    unwanted-literals))
		     4 ostream t))
	  (if (not all-together-p)
	      (format ostream "))"))))
    (if all-together-p
	(with-open-file (ostream ofile :direction :output :if-exists :append)
	  (format ostream "))")))))

;; 	    (pp-list (list (change-names (translate_problem (read istream) types supertypes
;; 							    unwanted-literals)
;; 					 domain))
;; 		     4 ostream t))



(defun translate_problem (inputted_problem types supertypes unwanted-literals)
  (let ((problem-def (cdr inputted_problem))
	name objects state goals-list goal)
    (dolist (problem-part problem-def)
      (case (car problem-part)
	(problem (setq name (nth 1 problem-part)))
	(:objects (setq objects (cdr problem-part)))
	(:init (setq state (cdr problem-part)))
	(:goal
	 (setq goals-list (cdr problem-part))
	 (setq goal (if (and (listp (car goals-list))
			     (eq (caar goals-list) 'and))
			(cdar goals-list)
			goals-list)))
	(t nil)))
    (if (member '- objects)
	(translate-typed-problem name objects state goal unwanted-literals)
	(translate-untyped-problem name objects state goal types supertypes))))

;; This does not longer work, because people add new things
;;   (let* ((name (nth 1 (nth 1 inputted_problem)))
;; 	 (objects (cdr (nth 3 inputted_problem)))
;; 	 (state (cdr (nth 4 inputted_problem)))
;; 	 (goals-list (cdr (nth 5 inputted_problem)))
;; 	 (goal (if (and (listp (car goals-list))
;; 			(eq (caar goals-list) 'and))
;; 		   (cdar goals-list)
;; 		   goals-list)))

(defun translate-typed-problem (name objects state goal unwanted-literals)
  (let* ((count_types (count '- objects))
         (type_array (make-array count_types :initial-element nil))
         (current_type 0)
         (final_objects nil))
    (print name)
    (if (> count_types 0)
	(let ((length_of_objects (length objects)))
	  (do ((x 0 (+ x 1)))
	      ((>= x length_of_objects))
	    (let ((object (nth x objects)))
	      (cond ((eq object '-)
		     (push `(objects-are ,@(reverse (aref type_array current_type))
					 ,(nth (+ x 1) objects))
			   final_objects)
		     ;; 		   (push (nth (+ x 1) objects )
		     ;; 			 (aref type_array current_type))
		     (incf x) 
		     (incf current_type))
		    (t (push object (aref type_array current_type))))))))
    (let* ((state-without-functions (remove-unwanted-literals (remove-functions state) unwanted-literals))
	   (final_state (if (> (length state) 1)
			    `(and ,@state-without-functions)
			  (car state-without-functions)))
	   (final_goal (if (> (length goal) 1)
			   `(and ,@goal)
			 (car goal))))
      `(setf (current-problem)
	 (create-problem
	  (name ,name)
	  (objects ,@(reverse final_objects))
	  (state ,final_state)
	  (goal ,final_goal))))))


(defun translate-untyped-problem (name objects state goal types supertypes)
  (declare (ignore objects))
  (let ((types-info nil)
	(final_objects nil)
	(final_state nil))
    (print name)
    (dolist (type types)
      (push (cons type nil) types-info))
    (dolist (literal state)
      (let ((type-literal (assoc (car literal) types-info)))
	(if type-literal
	    (push (cadr literal)
		  (cdr type-literal))
	  (push literal final_state))))
    (setq final_state (remove-functions final_state))
    (setq final_state (if (> (length final_state) 1)
			  `(and ,@(reverse final_state))
			  (car final_state)))
    (setq goal (if (> (length goal) 1)
		   `(and ,@goal)
		   (car goal)))
    (dolist (type-info types-info)
      (push (if (member (car type-info) supertypes)
		(let ((subtypes-only nil))
		  (dolist (instance (cdr type-info))
		    (if (not (find-if #'(lambda (type-info-1)
					  (member instance (cdr type-info-1)))
				      (remove type-info types-info :test #'equal)))
			(push instance subtypes-only)))
		  `(,@subtypes-only ,(car type-info)))
		`(,@(reverse (cdr type-info)) ,(car type-info)))
	    final_objects))
    `(setf (current-problem)
       (create-problem
	(name ,name)
	(objects ,@final_objects)
	(state ,final_state)
	(goal ,goal)))))

(defun remove-functions (state)
  (let ((p4-state nil))
    (dolist (literal state)
      (push (if (eq (car literal) '=)
		`(,@(cadr literal) ,(caddr literal))
	      literal)
	    p4-state))
    (nreverse p4-state)))

(defun remove-unwanted-literals (state unwanted-literals)
  (remove-if #'(lambda (literal)
		 (member (car literal) unwanted-literals :test #'eq))
	     state))

;; In case we want to translate a representation of the logistics or blocksworld in another one.
(defun change-names (problem domain)
  (let ((substitution (case domain
			(logistics '((package . object)
				     (location . post-office)
				     (in-city . loc-at)
				     (in . inside)))
			(blocksworld '((block . object)
				       (ontable . on-table)
				       (handempty . arm-empty)))
			(t nil))))
    (if substitution
	(sublis substitution problem)
      problem)))
