;;;; This codes translates Hamlet rules into Foil input format
;;; It requires to load Hamlet first

(defun run-foil (domain example-file &optional output-file)
  (domain domain)
  (if (not output-file)
      (setq output-file (format nil "/tmp/~(~a~)-examples.d" domain)))
  (prodigy2foil domain example-file output-file)
  (eval `(shell ,(format nil "cd ~/ml/foil6/; foil6 -v0 < ~a" output-file))))

(defun prodigy2foil (domain example-file output-file &optional load-domain-p)
  (if load-domain-p
      (domain domain))
  (with-open-file (istream example-file 
		   :direction :input
		   :if-does-not-exist nil)
    (with-open-file (ostream output-file
		     :direction :output
		     :if-exists :supersede
		     :if-does-not-exist :create)
      (let ((type-of-object nil)
	    (true-in-state (make-array (list 2) :initial-element nil))
	    (prior-goal (make-array (list 2) :initial-element nil))
	    (some-candidate-goals (make-array (list 2) :initial-element nil))
	    (target-goal (make-array (list 2) :initial-element nil))
	    (true-in-state (make-array (list 2) :initial-element nil))
	    (applicable-op (make-array (list 2) :initial-element nil))
	    (current-goal (make-array (list 2) :initial-element nil))
	    (candidate-goal (make-array (list 2) :initial-element nil))
	    (candidate-operator (make-array (list 2) :initial-element nil))
	    (current-operator (make-array (list 2) :initial-element nil))
	    (states nil)
	    (select-operators (make-array (list 2) :initial-element nil))
	    (select-goals (make-array (list 2) :initial-element nil))
	    (select-bindings (make-array (list 2) :initial-element nil))
	    (sub-goals (make-array (list 2) :initial-element nil))
	    (last-current-op nil))
	(do* ((example (read istream nil :eof) (read istream nil :eof)))
	    ((eq example :eof))
	  (let ((state-number (nth 0 example))
		(positivep (nth 1 example))
		(thens (nth 3 example)))
	    (push state-number states)
	    (dolist (cond (nth 2 example))
	      (case (car cond)
		(type-of-object (setq type-of-object
				  (add-type-of-object cond type-of-object)))
		(true-in-state (setq true-in-state
				 (add-literal (cadr cond) true-in-state state-number positivep)))
		(prior-goal (setq prior-goal
			      (add-goals cond prior-goal state-number positivep)))
		(some-candidate-goals (setq some-candidate-goals
					(add-goals cond some-candidate-goals state-number positivep)))
		(target-goal (setq target-goal
			       (add-literal (cadr cond) target-goal state-number positivep)))
		(applicable-op (setq applicable-op
				 (add-literal (cadr cond) applicable-op state-number positivep)))
		(current-goal (setq current-goal
				(add-literal (cadr cond) current-goal state-number positivep)))
		(candidate-goal (setq candidate-goal
				  (add-literal (cadr cond) candidate-goal state-number positivep)))
		(candidate-operator (setq candidate-operator 
				      (add-operator (cadr cond) candidate-operator state-number positivep)))
		(current-operator
		 (setq last-current-op (cadr cond))
		 (setq current-operator 
		   (add-operator last-current-op current-operator state-number positivep)))
		(t nil)))
	    (case (car thens)
	      (select (case (cadr thens)
			(operators (setq select-operators (add-operator (caddr thens) select-operators state-number positivep)))
			(goals (setq select-goals (add-literal (caddr thens) select-goals state-number positivep)))
			(bindings (setq select-bindings (add-bindings (caddr thens) select-bindings
								      last-current-op state-number positivep)))
			(t nil)))
	      (reject)
	      (prefer)
	      (sub-goal (push state-number
			      (aref sub-goals (if positivep 0 1))))
	      (apply)
	      (t nil))))
	(write-state-numbers states ostream)
	(write-types type-of-object ostream)
	(write-tuples domain 'true-in-state true-in-state nil ostream)
	(write-tuples domain 'prior-goal prior-goal nil ostream)
	(write-tuples domain 'some-candidate-goals some-candidate-goals nil ostream)
	(write-tuples domain 'target-goal target-goal nil ostream)
	(write-tuples domain 'applicable-op applicable-op nil ostream)	
	(write-tuples domain 'current-goal current-goal nil ostream)
	(write-tuples domain 'candidate-goal candidate-goal nil ostream)
	(write-tuples domain 'candidate-operator candidate-operator nil ostream)
	(write-tuples domain 'current-operator current-operator nil ostream)
	(write-tuples domain 'select-operators select-operators t ostream)
	(write-tuples domain 'select-goals select-goals t ostream)
	(write-tuples domain 'select-bindings select-bindings t ostream)
	(write-sub-goal sub-goals ostream)
	(terpri ostream)))))


;;; **********************************************************************
;;;             Adding functions
;;; **********************************************************************

(defun add-type-of-object (cond type-of-object)
  (if (assoc (caddr cond) type-of-object)
      (pushnew (cadr cond)
	       (cdr (assoc (caddr cond) type-of-object))
	       :test #'eq)
    (push (list (caddr cond) (cadr cond))
	  type-of-object))
  type-of-object)

(defun add-literal (cond literals state-number positivep)
  (let ((predicate (car cond)))
    (if (assoc predicate (aref literals (if positivep 0 1)))
	(push (cons state-number (cdr cond))
	      (cdr (assoc predicate (aref literals (if positivep 0 1)))))
      (push (list predicate (cons state-number (cdr cond)))
	    (aref literals (if positivep 0 1))))
    literals))

(defun add-goals (cond goals state-number positivep)
  (dolist (goal (cadr cond))
    (setq goals (add-literal goal goals state-number positivep)))
  goals)

(defun add-operator (operator operators state-number positivep)
  (if (assoc operator (aref operators (if positivep 0 1)))
      (pushnew state-number
	       (cdr (assoc operator (aref operators (if positivep 0 1))))
	       :test #'eq)
    (push (list operator state-number)
	  (aref operators (if positivep 0 1))))
  operators)

(defun add-bindings (binding bindings operator state-number positivep)
  (let ((objs (mapcar #'cdr binding)))
    (if (assoc operator (aref bindings (if positivep 0 1)))
	(pushnew (cons state-number objs)
		 (cdr (assoc operator (aref bindings (if positivep 0 1))))
		 :test #'eq)
      (push (list operator (cons state-number objs))
	    (aref bindings (if positivep 0 1))))
    bindings))



;;; **********************************************************************
;;;             Writing functions
;;; **********************************************************************

(defun write-state-numbers (states ostream)
  (format ostream "#State: ~a" (car states))
  (dolist (state-number (cdr states))
    (format ostream ",~d" state-number))
  (format ostream ".~%"))

;; It has to include all objects in all their supertypes
(defun write-types (types ostream)
  (let ((supertypes-objs nil)
	(supertypes nil))
    (dolist (type types)
      (format ostream "#~@(~a~): " (car type))
      (do* ((objs (cdr type) (cdr objs))
	    (obj (car objs) (car objs)))
	  ((null objs))
	(if (cdr objs)
	    (format ostream "~a," obj)
	  (format ostream "~a" obj))
	(dolist (supertype
		    (p4::p4type-all-parents 
		     (p4::type-name-to-type (car type)
					    *current-problem-space*)))
	  (let ((supertype-name (p4::p4type-name supertype)))
	    (if (not (eq supertype-name :top-type))
		(if (assoc supertype-name supertypes-objs)
		    (pushnew obj (cdr (assoc supertype-name supertypes-objs))
			     :test #'eq)
		  (push (cons supertype-name (list obj))
			supertypes-objs))))))
      (format ostream ".~%"))
    (dolist (type supertypes-objs)
      (format ostream "#~@(~a~): ~a" (car type) (cadr type))
      (dolist (obj (cddr type))
	(format ostream ",~a" obj))
      (format ostream ".~%"))))
    

;; It could be easily included in the domain specification as in PDDL
(defun write-types-of-pred (domain meta-predicate predicate ostream)
  (format ostream "(State")
  (if (member meta-predicate '(current-operator candidate-operator select-operators))
      (format ostream ") #~%")
    (format ostream
	    (case domain
	      (logistics
	       (case predicate
		 (at-obj ",Object,Location) #--~%")
		 (at-truck ",Truck,Location) #--~%")
		 (at-airplane ",Airplane,Airport) #--~%")
		 (inside-truck ",Object,Truck) #--~%")
		 (inside-airplane ",Object,Airplane) #--~%")
		 (same-city ",Location,Location) #--~%")
		 (part-of ",Truck,City) #--~%")
		 (loc-at ",Location,City) #--~%")
		 (drive-truck ",Truck,Location,Location) #---~%")
		 (load-truck ",Object,Truck,Location) #---~%")
		 (unload-truck ",Object,Truck,Location) #---~%")
		 (load-airplane ",Object,Airplane,Airport) #---~%")
		 (unload-airplane ",Object,Airplane,Airport) #---~%")
		 (fly-airplane ",Airplane,Airport,Airport) #---~%")))))))

(defun write-tuples (domain meta-predicate true-in-state target-concept-p ostream)
  (dolist (pos-literal (aref true-in-state 0))
    (if target-concept-p
	(format ostream "~%~(~a-~a~)" meta-predicate (car pos-literal))
      (format ostream "~%*~(~a-~a~)" meta-predicate (car pos-literal)))
    (write-types-of-pred domain meta-predicate (car pos-literal) ostream)
    (write-literals (cdr pos-literal) ostream)
    (let ((neg-examples (cdr (assoc (car pos-literal) (aref true-in-state 1)))))
      (when neg-examples
	(if (member meta-predicate '(select-operators select-goals select-bindings))
	    (format ostream ";~%"))
	(write-literals neg-examples ostream)))
    (format ostream ".")))

(defun write-literals (states ostream)
  (dolist (state states)
    (cond ((listp state)
	   (format ostream "~a" (car state))
	   (dolist (obj (cdr state))
	     (format ostream ",~a" obj)))
	  (t (format ostream "~a" state)))
    (terpri ostream)))

(defun write-sub-goal (sub-goals ostream)
  (format ostream "~%sub-goal(State) #~%")
  (write-literals (aref sub-goals 0) ostream)
  (let ((neg-examples (aref sub-goals 1)))
    (when neg-examples
      (format ostream ";~%")
      (write-literals neg-examples ostream))
    (format ostream ".")))
