;;; ----------------------------------------------------------------------------------
;;;   				PROPAGATION ALGORITHM  
;;; 
;;; Esta archivo contiene los algoritmos de propagacion de la red Temporal
;;;
;;;
;;;			 LAST MODIFICATION: August 05, 2002
;;;
;;; ----------------------------------------------------------------------------------


(in-package "COMMON-LISP-USER")

;;; ----------------------------------------------------------------------------------
;;;
;;; The Arc-consistency propagation algorithm 
;;;
;;; ----------------------------------------------------------------------------------

(defun propagate-AC (tm)

 (setf *fail* NIL)

 (let* ((dc_sources (time_map-dc_sources tm))
        (single_source (if (null (cdr dc_sources)) T NIL)))

        (setf (time_map-dc_sources tm) ())
        (update_all_bounds dc_sources)
        (loop (if (null dc_sources) (return T))
              (let ((dcs (pop dc_sources)))
 
                    (cond (*fail* (smark_sources dc_sources)
                                  (setf (edge-source dcs) NIL)
                                  (return NIL))
                          (T (prop_to_network dcs single_source)
                             (setf (edge-source dcs) NIL)
                             (if *fail*
                                 (progn
                                   (smark_sources dc_sources) 
                                   (return NIL)))))))))

;;; ---------------------------------------------------------------------------


(defun update_all_bounds (dc_sources)

  (dolist (dcs dc_sources)
          (update_bounds_dc dcs)))

;;; ---------------------------------------------------------------------------


(defun smark_sources (list_sources)

   (dolist (dcs list_sources)
           (setf (edge-source dcs) NIL)))

;;; ---------------------------------------------------------------------------


(defun update_bounds_dc (dc)

   (let ((max_lb (maxlist (cons (edge-lb dc) (edge-new_lb_bs dc))))
         (min_ub (minlist (cons (edge-ub dc) (edge-new_ub_bs dc)))))

     (if (> max_lb min_ub)
	 (progn
	   ;(format T "~%FAIL!! LB=~D  UB=~D"  max_lb min_ub)
             (setq *fail* T))
             (progn
                (if (> max_lb (edge-lb dc))
                    (setf (edge-lb dc) max_lb))
                (if (< min_ub (edge-ub dc))
                    (setf (edge-ub dc) min_ub))
                (setf (edge-source dc) T)))
         (setf (edge-new_lb_bs dc) ())
         (setf (edge-new_ub_bs dc) ())))

;;; ---------------------------------------------------------------------------


(defun maxlist (listn)

 "Calcola il massimo tra i lower bound dei nuovi bound spec aggiunti."

  (let ((max 0))
       (loop (if (null listn) (return max))
             (if (> (eval (car listn)) max)
                 (setf max (eval (car listn)))
                 (pop listn)))))

;;; ---------------------------------------------------------------------------

(defun minlist (listn)

 "Calcula el minimo entre los upper bound de los nuevos bound spec aggiunti."

  (let ((min *horizon*))
       (loop (if (null listn) (return min))
             (if (< (eval (car listn)) min)
                 (setf min (eval (car listn)))
                 (pop listn)))))

;;; ---------------------------------------------------------------------------


(defun prop_to_network (dcs single_source)

   (let ((work_list (list dcs))
         (dc NIL)
         (tolb NIL)
         (toub NIL)
         (fromlb NIL)
         (fromub NIL))

         (loop (if (or *fail* (null work_list))
                   (return (smark work_list)))

           (setf *sum-of-edge* (+ *sum-of-edge* 1))
           (setf dc (pop work_list))
           (setf (edge-in_queue dc) NIL)
           (multiple-value-setq (tolb toub fromlb fromub)
	     (local_update dc (edge-from dc) (edge-to dc)))

           ; (if *fail*
           ;   (format T "FAIL BETWEEN ~A ~A" (edge-from dc) (edge-to dc)))

           (if (and *switch-cycle* (not single_source) (edge-source dc))
             (if (verify_cycle dc tolb toub fromlb fromub)
               (progn
                 (setf *fail* T)
                 (return (smark work_list))))) 
           (if (or tolb toub)
             (progn
               (setf work_list
                 (if *prop-type*
                   (append (add_list dc (edge-to dc) single_source) work_list)
                   (append work_list (add_list dc (edge-to dc) single_source)
			   )))
                   (when tolb
                     (setf (time_point-pl (edge-to dc)) (edge-from dc)) 
                     (setf (time_point-st_edge_lb (edge-to dc)) dc))
                   (when toub
                     (setf (time_point-pu (edge-to dc)) (edge-from dc)) 
                     (setf (time_point-st_edge_ub (edge-to dc)) dc))))

           (if (or fromlb fromub)
             (progn
               (setf work_list
                 (if *prop-type*
                  (append (add_list dc (edge-from dc) single_source) work_list)
                  (append work_list (add_list dc (edge-from dc) single_source)
                          )))
                 (when fromlb
                   (setf (time_point-pl (edge-from dc)) (edge-to dc)) 
                   (setf (time_point-st_edge_lb (edge-from dc)) dc))
                 (when fromub
                   (setf (time_point-pu (edge-from dc)) (edge-to dc)) 
		   (setf (time_point-st_edge_ub (edge-from dc)) dc))))))

);;fin de la funcion

;;; ---------------------------------------------------------------------------


(defun smark (l_work)

  (dolist (dci l_work)
     (setf (edge-in_queue dci) NIL)))

;;; ---------------------------------------------------------------------------


(defun verify_cycle (dc to_lb to_ub from_lb from_ub)

  (or
   (if to_lb
       (if (equal (time_point-st_edge_lb (edge-to dc)) dc)
           (verify_cycle_lb dc (edge-to dc)))
       (if from_lb
           (if (equal (time_point-st_edge_lb (edge-from dc)) dc)
               (verify_cycle_lb dc (edge-from dc)))))
   (if to_ub
       (if (equal (time_point-st_edge_ub (edge-to dc)) dc)
           (verify_cycle_ub dc (edge-to dc)))
       (if from_ub
           (if (equal (time_point-st_edge_ub (edge-from dc)) dc)
               (verify_cycle_ub dc (edge-from dc)))))))

;;; ---------------------------------------------------------------------------


(defun verify_cycle_lb (dc tp_mod)
    
  (let ((listvisit (push_next_lb dc tp_mod))
        (dcc NIL)
        (tpc tp_mod))

    (loop (if (null listvisit) (return NIL))
          (setf dcc (pop listvisit))
          (if (equal dcc dc) (return T))
          (cond ((equal dcc (time_point-st_edge_lb (edge-from dcc)))
                    (setf tpc (edge-from dcc)))
                ((equal dcc (time_point-st_edge_lb (edge-to dcc)))
                    (setf tpc (edge-to dcc)))
               (T (error "~%verify_cycle_lb: puntatori non corrispondenti")))
          (setf listvisit (append (push_next_lb dcc tpc) listvisit)))))

;;; ---------------------------------------------------------------------------


(defun verify_cycle_ub (dc tp_mod)
    
  (let ((listvisit (push_next_ub dc tp_mod))
        (dcc NIL)
        (tpc tp_mod))

    (loop (if (null listvisit) (return NIL))
          (setf dcc (pop listvisit))
          (if (equal dcc dc) (return T))
          (cond ((equal dcc (time_point-st_edge_ub (edge-from dcc)))
                    (setf tpc (edge-from dcc)))
                ((equal dcc (time_point-st_edge_ub (edge-to dcc)))
                    (setf tpc (edge-to dcc)))
               (T (error "~%verify_cycle_ub: puntatori non corrispondenti")))
          (setf listvisit (append (push_next_ub dcc tpc) listvisit)))))

;;; ---------------------------------------------------------------------------


(defun push_next_lb (dc tp)

 (let ((listin (remove dc (time_point-in tp)))
       (listout (remove dc (time_point-out tp)))
       (add_list NIL))

       (dolist (dcc listin)
               (if (equal dcc (time_point-st_edge_lb (edge-from dcc)))
                   (push dcc add_list)))

       (dolist (dcc listout)
               (if (equal dcc (time_point-st_edge_lb (edge-to dcc)))
                   (push dcc add_list)))
       add_list))

;;; ---------------------------------------------------------------------------


(defun push_next_ub (dc tp)

 (let ((listin (remove dc (time_point-in tp)))
       (listout (remove dc (time_point-out tp)))
       (add_list NIL))

       (dolist (dcc listin)
               (if (equal dcc (time_point-st_edge_ub (edge-from dcc)))
                   (push dcc add_list)))

       (dolist (dcc listout)
               (if (equal dcc (time_point-st_edge_ub (edge-to dcc)))
                   (push dcc add_list)))
       add_list))

;;; ---------------------------------------------------------------------------


(defun local_update (dc from to)

    (let* ((old_lb_to     (time_point-lb to))
           (old_ub_to     (time_point-ub to))
           (old_lb_from (time_point-lb from))
           (old_ub_from (time_point-ub from))

           (new_lb_to old_lb_to)
           (new_ub_to old_ub_to)
           (new_lb_from old_lb_from)
           (new_ub_from old_ub_from)

           (to_lb nil)
           (to_ub nil)
           (from_lb nil)
           (from_ub nil))

         ;Blocco per la propagazione del lower bound (lb).
         ;Forward lb.
          (setf new_lb_to 
                (max old_lb_to (+ old_lb_from (edge-lb dc))))
          (if (> new_lb_to old_ub_to)
              (setf *fail* T)           ;Propagazione fallita.
              (if (> new_lb_to old_lb_to)
                  (progn
                    (setf (time_point-lb to) new_lb_to)
                    (setf to_lb T))
                  ;Backward lb.
                  (progn
                     (setf new_lb_from
                           (max old_lb_from (- old_lb_to (edge-ub dc))))
                     (if (> new_lb_from old_ub_from)
                         (setf *fail* T) ;Propagazione fallita.
                         (if (> new_lb_from old_lb_from)
                             (progn
                                (setf (time_point-lb from) new_lb_from) 
                                (setf from_lb T)))))))

         ;Blocco per la propagazione dell'upper bound (ub).
         ;Forward ub.
          (setf new_ub_to
                (min old_ub_to (+ old_ub_from (edge-ub dc))))
          (if (< new_ub_to new_lb_to)
              (setf *fail* T)            ;Propagazione fallita.
              (if (< new_ub_to old_ub_to)
                  (progn
                    (setf (time_point-ub to) new_ub_to)
                    (setf to_ub T)) 
                  ;Backward ub.
                  (progn
                    (setf new_ub_from
                          (min old_ub_from (- old_ub_to (edge-lb dc))))
                    (if (< new_ub_from new_lb_from)
                        (setf *fail* T)  ;Propagazione fallita.
                        (if (< new_ub_from old_ub_from)
                            (progn
                               (setf (time_point-ub from) new_ub_from)
                               (setf from_ub T)))))))

         (values to_lb to_ub from_lb from_ub)))

;;; ---------------------------------------------------------------------------


(defun add_list (dc tp single_source)

  (let ((listin (remove dc (append (time_point-in tp)
				   (time_point-out tp))))
       (listout ())
       (dc_corr NIL))

       (loop (if (null listin) (return listout))
             (setf dc_corr (pop listin))
             (cond ((and *switch-cycle* single_source (edge-source dc_corr))
                      (setf *fail* T)
                      (return listout))
                   ((not (edge-in_queue dc_corr))
                      (setf (edge-in_queue dc_corr) T)
                      (push dc_corr listout))
                   (T nil)))))