;;; ----------------------------------------------------------------------------------
;;;   			REMOVE TEMPORAL CONSTRAINT ALGORITHM  
;;; 
;;; Esta archivo contiene 
;;;
;;;
;;;			 LAST MODIFICATION: August 05, 2002
;;;
;;; ----------------------------------------------------------------------------------

(in-package "COMMON-LISP-USER")

;;; ---------------------------------------------------------------------------
;;;
;;;
;;; ---------------------------------------------------------------------------


(defun prop_macro_cons (dm tp)
  (declare (ignore dm))
   (let ((queue (append (time_point-in tp) (time_point-out tp)))
         (dc NIL)
         (to_lb NIL)
         (to_ub NIL)
         (from_lb NIL)
         (from_ub NIL))

         (loop (if (null queue)
                   (return (format t "~%Rete propagata")))
              (setf dc (pop queue))
              (setf (edge-in_queue dc) NIL)
              (multiple-value-setq (to_lb to_ub from_lb from_ub)
                                                 (local_prop dc))
              (if (or to_lb to_ub (edge-source dc))
                (progn
                  (if (edge-source dc) (setf (edge-source dc) NIL))
                  (setf queue
                        (append queue (add_queue dc (edge-to dc))))
                     (if to_lb
                         (setf (time_point-st_edge_lb (edge-to dc)) dc))
                     (if to_ub
                         (setf (time_point-st_edge_ub (edge-to dc)) dc))))

              (if (or from_lb from_ub (edge-source dc))
                 (progn
                  (if (edge-source dc) (setf (edge-source dc) NIL))
                  (setf queue
                        (append queue (add_queue dc (edge-from dc))))
                     (if from_lb
                         (setf (time_point-st_edge_lb (edge-from dc)) dc))
                     (if from_ub
                         (setf (time_point-st_edge_ub (edge-from dc)) dc))))
          )
      )
   
   ) ;;; END *PROP_MACRO_CONS*

;;; ---------------------------------------------------------------------------


(defun prop_macro_loc (dm listinit)
  (declare (ignore dm))
   (let ((queue listinit)
         (dc nil)
         (to_lb nil)
         (to_ub nil)
         (from_lb nil)
         (from_ub nil))

         (loop (if (null queue)
                   (return))
              (setf dc (pop queue))
              (setf (edge-in_queue dc) NIL)
              (multiple-value-setq (to_lb to_ub from_lb from_ub)
                                                 (local_prop dc))
              (if (or to_lb to_ub)
                (progn
                  (setf queue
                        (append queue (add_queue dc (edge-to dc))))
                     (if to_lb
                         (setf (time_point-st_edge_lb (edge-to dc)) dc))
                     (if to_ub
                         (setf (time_point-st_edge_ub (edge-to dc)) dc))))

              (if (or from_lb from_ub)
                 (progn
                  (setf queue
                        (append queue (add_queue dc (edge-from dc))))
                     (if from_lb
                         (setf (time_point-st_edge_lb (edge-from dc)) dc))
                     (if from_ub
                         (setf (time_point-st_edge_ub (edge-from dc)) dc))))
          )
      )
 
   ) ;;; END *PROP_MACRO_LOC*

;;; ---------------------------------------------------------------------------


(defun reset_network_lb (tp)
    
  (let ((listvisit (add_list_lb nil tp))
        (listqueue)
        (dcc)
        (tpc tp))

    (setf listqueue (append listqueue (add_queue dcc tpc)))
    (setf (time_point-lb tpc) 0)
    (loop (if (null listvisit)
              (return listqueue))
          (setf dcc (pop listvisit))

          (cond ((equal dcc (time_point-st_edge_lb (edge-from dcc)))
                    (setf tpc (edge-from dcc)))
                ((equal dcc (time_point-st_edge_lb (edge-to dcc)))
                    (setf tpc (edge-to dcc)))
               (T (error "~%Reset_network_lb: puntatori non corrispondenti")))
          (setf listqueue (append listqueue (add_queue dcc tpc)))
          (setf (time_point-lb tpc) 0)
          (setf listvisit (append (add_list_lb dcc tpc) listvisit)))))

;;; ---------------------------------------------------------------------------


(defun reset_network_ub (tp)
    
  (let ((listvisit (add_list_ub nil tp))
        (listqueue)
        (dcc)
        (tpc tp))

    (setf listqueue (append listqueue (add_queue dcc tpc)))
    (setf (time_point-ub tpc) *horizon*)
    (loop (if (null listvisit)
              (return listqueue))
          (setf dcc (pop listvisit))

          (cond ((equal dcc (time_point-st_edge_ub (edge-from dcc)))
                    (setf tpc (edge-from dcc)))
                ((equal dcc (time_point-st_edge_ub (edge-to dcc)))
                    (setf tpc (edge-to dcc)))
               (T (error "~%Reset_network_ub: puntatori non corrispondenti")))
          (setf listqueue (append listqueue (add_queue dcc tpc)))  
          (setf (time_point-ub tpc) *horizon*)
          (setf listvisit (append (add_list_ub dcc tpc) listvisit)))))

;;; ---------------------------------------------------------------------------


(defun add_list_lb (dc tp)

 (let ((listin (remove dc (time_point-in tp)))
       (listout (remove dc (time_point-out tp)))
       (add_list))

       (dolist (dcc listin)
               (setf (edge-source dcc) T)
               (if (equal dcc (time_point-st_edge_lb (edge-from dcc)))
                   (push dcc add_list)))

       (dolist (dcc listout)
               (setf (edge-source dcc) T)
               (if (equal dcc (time_point-st_edge_lb (edge-to dcc)))
                   (push dcc add_list)))
       add_list))

;;; ---------------------------------------------------------------------------


(defun add_list_ub (dc tp)

 (let ((listin (remove dc (time_point-in tp)))
       (listout (remove dc (time_point-out tp)))
       (add_list))

       (dolist (dcc listin)
               (setf (edge-source dcc) T)
               (if (equal dcc (time_point-st_edge_ub (edge-from dcc)))
                   (push dcc add_list)))

       (dolist (dcc listout)
               (setf (edge-source dcc) T)
               (if (equal dcc (time_point-st_edge_ub (edge-to dcc)))
                   (push dcc add_list)))
       add_list))

;;; ---------------------------------------------------------------------------


(defun mark_network (dm)
                 
  (dolist (dc (time_map-list_dcs dm))
              (setf (edge-source (cdr dc)) T)))

;;; ---------------------------------------------------------------------------


(defun add_queue (dc tp)
 (let ((listin (remove dc (append (time_point-in tp) (time_point-out tp))))
       (listout ()))
       (dolist (dcc listin)
               (if (not (edge-in_queue dcc))
                   (progn
                     (setf (edge-in_queue dcc) T)
                     (push dcc listout))))
       (setf *sum-of-edge* (+ *sum-of-edge* (length listout)))
       listout))

;;; ---------------------------------------------------------------------------


(defun local_prop (dc)

   (let* ((to (edge-to dc))
          (from (edge-from dc))
          (old_lb_to     (time_point-lb to))
          (old_ub_to     (time_point-ub to))
          (old_lb_from (time_point-lb from))
          (old_ub_from (time_point-ub from))

          (new_lb_to  )
          (new_ub_to  )
          (new_lb_from)
          (new_ub_from)

          (to_lb nil)
          (to_ub nil)
          (from_lb nil)
          (from_ub nil))

          ;; Lower bounds propagation (lb).

          (setf  new_lb_to 
                 (max old_lb_to (+ old_lb_from (edge-lb dc))))
          (if (> new_lb_to old_lb_to)
              (progn
                (setf (time_point-lb to) new_lb_to)
                (setf to_lb T))
              (progn
                (setf new_lb_from
                      (max old_lb_from (- old_lb_to (edge-ub dc))))
                (if (> new_lb_from old_lb_from)
                    (progn
                      (setf (time_point-lb from) new_lb_from) 
                      (setf from_lb T)))))

          ;; Upper bounds propagation (ub)

          (setf new_ub_to
                (min old_ub_to (+ old_ub_from (edge-ub dc))))
          (if (< new_ub_to old_ub_to)
              (progn
                (setf (time_point-ub to) new_ub_to)
                (setf to_ub T))
              (progn
                (setf new_ub_from
                      (min old_ub_from (- old_ub_to (edge-lb dc))))
                (if (< new_ub_from old_ub_from)
                    (progn
                      (setf (time_point-ub from) new_ub_from)
                      (setf from_ub T)))))

          (values to_lb to_ub from_lb from_ub)))


;;; ---------------------------------------------------------------------------
;;;
;;; Elimina el puntero al arco dc de la rete temporal (dm)
;;;
;;; ---------------------------------------------------------------------------


(defun remove_edge (dm dc)

  (let ((tp_lb (cond ((equal dc (time_point-st_edge_lb (edge-from dc)))
                      (edge-from dc))
                     ((equal dc (time_point-st_edge_lb (edge-to dc)))
                      (edge-to dc))
                     (T nil)))
        (tp_ub (cond ((equal dc (time_point-st_edge_ub (edge-from dc)))
                      (edge-from dc))
                     ((equal dc (time_point-st_edge_ub (edge-to dc)))
                      (edge-to dc))
                     (T nil))))
      
    (setf (time_point-in (edge-to dc))
      (remove dc (time_point-in (edge-to dc))))
    (setf (time_point-out (edge-from dc))
      (remove dc (time_point-out (edge-from dc))))
	
    (setf (edge-from dc) NIL)
    (setf (edge-to dc) NIL)
    
    (setf (time_map-list_dcs dm)
        (remove (assoc (edge-name dc) (time_map-list_dcs dm))
		(time_map-list_dcs dm)))

        (let ((aff-edges-lb (if tp_lb
                               (progn
                                 (setf (time_point-st_edge_lb tp_lb) nil)
                                 (reset_network_lb tp_lb))))
              (aff-edges-ub (if tp_ub
                               (progn
                                 (setf (time_point-st_edge_ub tp_ub) nil)
                                 (reset_network_ub tp_ub)))))
	  
	
	(prop_macro_loc dm (append aff-edges-lb aff-edges-ub)))
     )
   ) ;;; END *REMOVE_EDGE*
				  
;;; --------------------------------------------------------------------------
;;; 
;;; Funcion para eliminar los edges pero se tienen que pasar los arcos. 
;;; Esta modificacion hace que se le puedan pasar los identificadores y se 
;;; encarga de buscar el edge asociado a ello. MD y Angelo
;;;
;;; --------------------------------------------------------------------------


(defun remove_edge_id (dm id_dc)
   
 (let ((dc (cdr (assoc id_dc (time_map-list_dcs dm)))))

  (remove_edge dm dc)  

   )
 ) 

;;; END *REMOVE_EDGE_ID*
				   
;;; ---------------------------------------------------------------------------
;;; ---------------------------------------------------------------------------
;;; *END-OF-FILE*




