;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Fichero: estructuras.lisp
;; Descripcion: contiene las estructuras para el algoritmo de amenazas
;; Historico:
;;      Version    Autor        Fecha        Cambio
;;      Creacion    M.D        25/08/02      --------
;;
;;       
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(in-package "USER")

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; Estructura que almacena la condicion que se establece entre 
;; dos operadores:    Al--R-->Ap 
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 (defstruct (LINK  (:print-function print-link))
   id1 			; id1 is a step (operator) that establishes condition
   condition 		; the condition stablished by id1 and needed by id2
   id2			; the condition is a precondition of step id2, the term been supported
   min                  ; minimal  time distance between 2 operators
   max                  ; maximal  time distance between 2 operators
 )

(defun print-link (self &optional (stream t) depth)
  (declare (ignore depth))
  (format stream "(LINK ~a ~a ~a)"
          (TimePoint-id (link-id1 self))
          (link-condition self)
          (TimePoint-id (link-id2 self))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; Estructura que almacena para cada operador 2 TP, correspondiente
;; al inicio y fin del operador. Se debe conocer la duracion.
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defstruct (TimePoint (:print-function print-TimePoint print-TPName))
   id				; integer id corresponding to the operator
   operator			; operator name
   preconds			; operator preconditions
   adds			; operator adds-list
   dels		      ; operator del-list
   start-id			; operator start id 
   end-id			; operator end id
   duration-min		; duration min
   duration-max		; duration max
   added			; if the time point has been added to the Network
   state			; state after applying the operator
   links-added		; the links added with the operator
   id-domain	      ; identify the operator (order appereance in the domain) with the resource
   res-decision		; Posible de-ordering decision in resource conflict
   act-decision		; Posible de-ordering decision in resource conflict
   links-res	      ; Links added by the resource
   ;TP-st 
   ;TP-end
 )

;(defun print-TimePoint (self &optional (stream t) depth)
(defun print-TP (self &optional (stream t) depth)
  (declare (ignore depth))
  (format stream "<Name= ~d ID= ~a Dur Min= ~d st= ~a ~%precon= ~a~%adds= ~d~% dels= ~d~%state = ~d~% Links = ~d ~%links-res = ~d>" 
					(TimePoint-operator self)
					(TimePoint-id self)
					(TimePoint-duration-min  self)
					(TimePoint-start-id self)
					(TimePoint-preconds self)
					(TimePoint-adds self)
					(TimePoint-dels self)
					;(TimePoint-added self)
					(TimePoint-state self)
					(TimePoint-links-added self)
					(TimePoint-links-res self)

))

 (defun print-TimePoint (self &optional (stream t) depth)
  (declare (ignore depth))
  (format stream " (ID = ~d Name = ~d Id-Res = ~d ~%Links-Res = ~d)" 
  		     (TimePoint-id self)
                 (TimePoint-operator self)				
                 (TimePoint-id-domain self)
		     (TimePoint-links-res self)
))




;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; Estructura que asocia el numero de veces que se utiliza un recurso con su nombre
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defstruct (RESOURCE  (:print-function print-resource))
   name 		; Resource Name 
   times 		; Times that the resource appears in the plan
   total		; Number of resources
 )

 (defun print-resource(self &optional (stream t) depth)
  (declare (ignore depth))
  (format stream "~% Name = ~d Times = ~d Total = ~d" 
  		     (RESOURCE-name self)
                 (RESOURCE-times self)				
                 (RESOURCE-total self)
))