;;; ----------------------------------------------------------------------------------
;;;   				SAVING AND RECOVERING FUNCTIONS  
;;; 
;;; Esta archivo contiene informacion para guardar y recuperar estados de la Red temporal
;;;
;;;
;;;			 LAST MODIFICATION: August 05, 2002
;;;
;;; ----------------------------------------------------------------------------------


(in-package "COMMON-LISP-USER")


(defun back_tp (atp)

  (let ((tp (cdr atp)))
    
    (setf (time_point-lb tp) (pop (time_point-stack-lb tp)))
    (setf (time_point-ub tp) (pop (time_point-stack-ub tp)))

    (setf (time_point-pl tp) (pop (time_point-stack-pl tp))) 
    (setf (time_point-pu tp) (pop (time_point-stack-pu tp)))

    (setf (time_point-st_edge_lb tp) (pop (time_point-stack_ste_lb tp)))
    (setf (time_point-st_edge_ub tp) (pop (time_point-stack_ste_ub tp)))

    )
  ) ;;; END *BACK_TP*

;;; ---------------------------------------------------------------------------


(defun back_dc (dcp)

   (let ((dc (cdr dcp)))
    
     (if (> (car (edge-stack_molt dc)) 1)
	 
         (progn
           (push (- (pop (edge-stack_molt dc)) 1) (edge-stack_molt dc))
           (setf (edge-lb dc) (car (edge-stack-lb dc)))
           (setf (edge-ub dc) (car (edge-stack-ub dc))))

       (progn
	 (pop (edge-stack_molt dc))
	 (setf (edge-lb dc) (pop (edge-stack-lb dc)))
	 (setf (edge-ub dc) (pop (edge-stack-ub dc))))
       )
     )
   ) ;;; END *BACK_DC*

;;; ---------------------------------------------------------------------------


(defun remove_dc (dc)

     (setf (time_point-in (edge-to dc)) 
           (remove dc (time_point-in (edge-to dc))))
     (setf (time_point-out (edge-from dc))
           (remove dc (time_point-out (edge-from dc))))

     (setf (edge-from dc) NIL)
     (setf (edge-to dc) NIL)

) ;;; END *REMOVE-DC*

;;; ---------------------------------------------------------------------------


(defun remove_tp (tp)
   
       (setf (time_point-in tp) NIL)
       (setf (time_point-out tp) NIL)

) ;;; END *REMOVE-TP* 

;;; ---------------------------------------------------------------------------
;;;
;;; Save the last consistent solution
;;;
;;; ---------------------------------------------------------------------------

(defun save-solution (tm)
  
   (push (time_map-ntp tm) (time_map-stack_ntp tm))
   (push (time_map-ndc tm) (time_map-stack_ndc tm))
    
   (dolist (atpp (time_map-list_tps tm))
      
      (let ((tpp (cdr atpp)))
                
         (push (time_point-lb tpp) (time_point-stack-lb tpp))
         (push (time_point-ub tpp) (time_point-stack-ub tpp))
      
         (push (time_point-pl tpp) (time_point-stack-pl tpp))
         (push (time_point-pu tpp) (time_point-stack-pu tpp))
      
         (push (time_point-st_edge_lb tpp) (time_point-stack_ste_lb tpp))
         (push (time_point-st_edge_ub tpp) (time_point-stack_ste_ub tpp))
       )
      )

   (dolist (adcc (time_map-list_dcs tm))
     
      (let ((dcc (cdr adcc)))
            
         (cond 
               ((null (edge-stack-lb dcc))  
        
                (push (edge-lb dcc) (edge-stack-lb dcc))
                (push (edge-ub dcc) (edge-stack-ub dcc))
                (push 1 (edge-stack_molt dcc)))
               
               ((and (equal (edge-lb dcc) (car (edge-stack-lb dcc)))
                     (equal (edge-ub dcc) (car (edge-stack-ub dcc))))
                 
                (push (+ 1 (pop (edge-stack_molt dcc))) (edge-stack_molt dcc)))
               
               (T
                 (push (edge-lb dcc) (edge-stack-lb dcc))
                 (push (edge-ub dcc) (edge-stack-ub dcc))
                 (push 1 (edge-stack_molt dcc)))
               )
         )
      )
   
) ;;; END *SAVE-SOLUTION*

;;; ---------------------------------------------------------------------------
;;; Restore the last consistent solution
;;;
;;; ---------------------------------------------------------------------------

(defun restore-solution (tm)
   
  
  (loop
    
     (when (equal (car (time_map-stack_ndc tm))
                  (caar (time_map-list_dcs tm)))
        
        (setf (time_map-ndc tm) (pop (time_map-stack_ndc tm)))
        (return T))
     
     (remove_dc (cdr (pop (time_map-list_dcs tm))))
   
     ) ;;; END LOOP
   

  (loop
       
     (when (equal (car (time_map-stack_ntp tm))
                  (caar (time_map-list_tps tm)))
        
        (setf (time_map-ntp tm) (pop (time_map-stack_ntp tm)))
        (return T))
     
      (remove_tp (cdr (pop (time_map-list_tps tm))))
   
      ) ;;; END LOOP
   
                
  (dolist (dcp (time_map-list_dcs tm))
    (back_dc dcp))

  (dolist (tpp (time_map-list_tps tm))
    (back_tp tpp))

  ) ;;; END *RESTORE-SOLUTION*