;;; ----------------------------------------------------------------------------------
;;;   				TIME MAP'S TOOLS   
;;; 
;;; Esta archivo contiene las funciones para la manejar Red Temporal
;;;
;;;
;;;			 LAST MODIFICATION: August 05, 2002
;;;
;;; ----------------------------------------------------------------------------------


(in-package "COMMON-LISP-USER")


;;; ----------------------------------------------------------------------------------
;;;
;;;   La estructura time_map representa la rete temporal
;;;
;;;
;;; ----------------------------------------------------------------------------------
 
(defun create_tp (tm)

  (let* ((ntp (incf (time_map-ntp tm)))
	 (tp (make-time_point :name ntp)))
    
     (setf (time_map-list_tps tm) (acons ntp tp (time_map-list_tps tm)))
     
     tp)
   
   ) ;;; END *CREATE_TP*

;;; ---------------------------------------------------------------------------


(defun gen_dc (tm from to)

   (let* ((ndc (incf (time_map-ndc tm)))
         (dc (make-edge
              :name ndc
              :from from
              :to to)))

      (push dc (time_point-out from))
      (push dc (time_point-in to))

     (setf (time_map-list_dcs tm) 
           (acons ndc dc (time_map-list_dcs tm)))
     dc)
   
  ) ;;; END *GEN_DC*

;;; ---------------------------------------------------------------------------


(defun add_bs_dc (dc lb ub)

  (push lb (edge-new_lb_bs dc))
  (push ub (edge-new_ub_bs dc))

  dc
  
  ) ;;; END *ADD_BS_DC*

;;; ---------------------------------------------------------------------------


(defun exist_dc (tp1 tp2)
 
  (let ((laux (time_point-out tp1)))
  
    (loop 

      (if (null laux) (return NIL))
      
      (if (equal tp2 (edge-to (car laux))) 
	  (return (car laux)))
      
      (pop laux))
    )
   
  ) ;;; END *EXIST_DC*

;;; ---------------------------------------------------------------------------
;;; 
;;; Crea los TP (los dos primeros los crea en la inicializacion)
;;;
;;; ---------------------------------------------------------------------------


(defun gen_tps (tm ntp)

  (loop  
    (if (eq ntp 0) 
       (return T))                        
         (create_tp tm)
         (decf ntp))
   
   ) ;;; END *GEN_TPS*

;;; ---------------------------------------------------------------------------


(defun relate_tp (tm from to lb ub)

   (let ((dc (exist_dc from to)))

     (cond 
      
      (dc (if (not (member dc (time_map-dc_sources tm)))
	  
	      (push (add_bs_dc dc lb ub) (time_map-dc_sources tm))
	    
	    (add_bs_dc dc lb ub))
	  
	  dc)

      (T (let ((dc-new (gen_dc tm from to)))

	   (push (add_bs_dc dc-new lb ub) (time_map-dc_sources tm))

	   dc-new))
      )
     )

   ) ;;; END *RELATE_TP*

;;; ---------------------------------------------------------------------------
;;; 
;;;  FUNCION QUE SE LE PASA el identificador en vez de los apuntadores y relaciona
;;;  2 TP (create the arc)
;;;
;;;  						AO y MD    4/6/02
;;; ---------------------------------------------------------------------------


(defun relate_tp_id (tm id_from id_to lb ub)

  
   (let* ((from (cdr (assoc id_from (time_map-list_tps tm))))
          (to   (cdr (assoc id_to   (time_map-list_tps tm))))
          (dc   (exist_dc from to)))

      (cond (dc (if (not (member dc (time_map-dc_sources tm)))
                   (push (add_bs_dc dc lb ub) (time_map-dc_sources tm))
                   (add_bs_dc dc lb ub))

             dc)
            
            (T (let ((dc-new (gen_dc tm from to)))
                  (push (add_bs_dc dc-new lb ub) (time_map-dc_sources tm))

                  dc-new))))
   
) ;;; END *RELATE_TP_ID*

;;; ----------------------------------------------------------------------------------
;;;
;;;   init-time-map crea la red temporal iniciando el horizonte con un millon y el origen con cero
;;;
;;;
;;; ----------------------------------------------------------------------------------

(defun init-time-map (name-tm)
  
  (let* ((tm (make-time_map 
	     :name name-tm))
          
	 (ref-tp (create_tp tm))
	 (horizon (create_tp tm)))

    (setf (time_point-ub ref-tp) 0)
    (setf (time_point-lb ref-tp) 0)
  
    (setf (time_map-ref_point tm) ref-tp)
    (setf (time_map-horizon tm) horizon)
  
    (setf (time_map-horizon-constraint tm)
      (relate_tp tm ref-tp horizon 0 *horizon*))
    tm
    )
  ) 

;;; ----------------------------------------------------------------------------------


(defun get-horizon (tm)
;;   (let*((horizon 0)
;; 	)

;;     (if (equal tm nil)
;; 	(setq horizon 0)
;;       (progn
	(time_point-lb (time_map-horizon tm))
					;(setq horizon time_point-lb)
;; 	);;fin progn
;;       );;fin if
;; 
;;     );;fin let*
)

;;; ----------------------------------------------------------------------------------