/*********************************************************************
 *
 * File: control_rules.c
 *
 * Description: implementation of routines that evaluate heuristic 
 *              using control rules and CLIPS.
 *              Modification: Susana Fernndez 2006. 
 *                            Control rules for selecting operators
 *
 *
 * Author: Raquel Fuentetaja 2005
 *
 *********************************************************************/

#include "ff.h"
#include "output.h"
#include "control_rules.h"
/* only for function LESS */
#include "relax.h"

/**************************/

#include "clips.h"

/*SFA Nov06*/
int gCR[5000];
int gnum_CR;

int gnum_index_action=0;
Bool CLIPS_ini=FALSE;

/***************************************************************************
  NAME         : cr_trace
  DESCRIPTION  : prints a trace of the control rules. 
  INPUTS       : void
  RETURNS      : void
  SIDE EFFECTS : print on stdout
  NOTES        : 
*****************************************************************************/
void cr_trace(char *trace)
{
  if ( gcmd_line.display_info == 130 || gcmd_line.display_info == 132 ) {
    printf("\n cr: %s",trace);
  }
}

void cr_trace2 (char *trace)
{
  /*
  if ( gcmd_line.display_info == 130 ) {
    printf("\n\ncr: %s",trace);
  }
  */
}

/****************************************************************************
  NAME         : initialize_clips
  DESCRIPTION  : initilize the clips environment
  INPUTS       : void
  RETURNS      : 1 if initialization process is correct.
  SIDE EFFECTS :
  NOTES        : 
******************************************************************************/
int initialize_clips(void)
{  
  char cr_file[MAX_LENGTH] = "";
  int loaded=TRUE;

  /* SFA Nov06, enforced_hill_climbing already has initialized CLIPS*/
  if (!CLIPS_ini) {
    cr_trace("initializing clips ..."); 
    sprintf(cr_file, "%s%s", gcmd_line.path, gcmd_line.cr_file_name);
    InitializeEnvironment();
    loaded=Load(cr_file);

    if (loaded == 0)
      fprintf(stdout, "\n\ncr: rules file couldn't be opened");
    else if (loaded ==-1)
            fprintf(stdout, "\n\ncr: errors ocurred while loading rules file");
          
    if ( gcmd_line.display_info == 130 ) {
       printf("\n\n");
       Watch("activations");
    }

    CLIPS_ini=TRUE;

  Reset();
  assert_static_facts_to_clips();
  assert_goals_to_clips();
  }
  return loaded;
}

/*****************************************************************************
  NAME         : assert_static_facts_to_clips
  DESCRIPTION  : 
  INPUTS       : void
  RETURNS      : void
  SIDE EFFECTS :
  NOTES        : 
*****************************************************************************/
int assert_static_facts_to_clips(void)
{  
  int i,j;

  /* STATIC PREDICATES */

  cr_trace("asserting into clips invariant predicates");
  for ( i = 0; i < gnum_predicates; i++ ) {
    /*  printf("\n\n%s:", gpredicates[i]);*/
    if ( !gis_added[i] &&	 !gis_deleted[i] ) {
      /* if it is a static fact */
      for ( j = 0; j < gnum_initial_predicate[i]; j++ ) {
	/*print_Fact( &(ginitial_predicate[i][j]));*/
	char fact[1000]="";
	print_Fact_to_string(&(ginitial_predicate[i][j]),fact,"TRUE-IN-STATE");
        /*we shouldn't assert EQUAL nor NOT-= facts*/
        if ((strstr(fact, "NOT-=")==NULL) && (strstr(fact,"EQUAL")==NULL)) {
	    cr_trace(fact);
	    AssertString(fact);
	}
      }
    }
  }

  /* RELEVANT FACTS */
 /* they are static, don't depend on the graph*/
  cr_trace("asserting into clips relevant facts");
  for ( i = 0; i < gnum_relevant_facts; i++ ) {
	char fact[1000]="";
	print_Fact_to_string( &(grelevant_facts[i]), fact, "RELEVANT-FACT");
	cr_trace(fact);
	AssertString(fact);
  }

  /* STATIC FUNCTIONS */
  cr_trace("asserting into clips invariant fluents");
  for ( i = 0; i < gnum_functions; i++ ) {
    /*    printf("\n\n%s:", gfunctions[i]);*/
    if ( !gis_changed[i] ) {
      for ( j = 0; j < gnum_initial_function[i]; j++ ) {
	char fact[1000]="";
	print_Fluent_and_Value_to_string(&(ginitial_function[i][j].fluent), 
                         ginitial_function[i][j].value, fact, "TRUE-IN-STATE");
	cr_trace(fact);
	AssertString(fact);
      }
    }
  }

  /* TYPES OF OBJECTS */
  /*we should check if the domain is typing*/
  cr_trace("asserting into clips types of objects");
  /*gtype_names[0]=ARTIFICIAL_ALL_OBJECTS, gtype_names[1]=OBJECT*/
  for ( i = 2; i < gnum_types; i++ ) {
    /*    printf("\n%d --> %s: ", i, gtype_names[i]); */
    char *type = gtype_names[i];  
    if (type != NULL) {
    char *either=strstr(type,"EITHER");
    char type1[100]="";
    char type2[100]="";
      if (either != NULL) sprintf(type,"EITHER~%s~%s", type1, type2);
      for ( j = 0; j < gtype_size[i]; j++ ) {
	int int_object=gtype_consts[i][j];
	char *object= gconstants[int_object];
	/*	printf("%d ", gtype_consts[i][j]);*/
	char fact[1000]=""; 
        if( either==NULL) {
	    sprintf (fact,"(TYPE-OF-OBJECT %s %s)", object, type);
            cr_trace(fact);
            AssertString(fact);}
        else {
	    sprintf (fact,"(TYPE-OF-OBJECT %s %s)", object, type1);
            cr_trace(fact);
            AssertString(fact);
	    sprintf (fact,"(TYPE-OF-OBJECT %s %s)", object, type2);
            cr_trace(fact);
            AssertString(fact);}
      }
    }
    }
  return 1;
}


/****************************************************************************
  NAME         : assert_goals_to_clips
  DESCRIPTION  : 
  INPUTS       : void
  RETURNS      : void
  SIDE EFFECTS :
  NOTES        : 
***************************************************************************/

int assert_goals_to_clips(void)
{
  int i;

  for (i = 0; i < gnum_logic_goal; i++ ) {
      char goal[1000]="";
      print_ft_name_to_string( glogic_goal[i], goal, "IN-CURRENT-GOALS");
      cr_trace(goal);
      AssertString(goal);
    }
    return 1;
}


/****************************************************************************
  NAME         : initialize_facts_info
  DESCRIPTION  : 
  INPUTS       : void
  RETURNS      : void
  SIDE EFFECTS :
  NOTES        : 
**************************************************************************/
void  initialize_facts_info(void)
{
  int i,j;

  for ( i = 0; i < gnum_relevant_facts; i++ ) {
    possible_positive_clips_facts[i]=NULL;
    positive_facts_in_state[i]=0;
  }

/*SFA*/
  for ( i = 0; i < gnum_functions; i++ ) {
    if (gis_changed[i]) {
      for(j=0; j< gnum_initial_function[i]; j++)
         possible_clips_fluents[i][j]=NULL;
   }
  }
}


/**************************************************************************
  NAME         : assert_state_to_clips (State *S)
  DESCRIPTION  : 
  INPUTS       : void
  RETURNS      : void
  SIDE EFFECTS :
  NOTES        : 
***************************************************************************/
void assert_state_to_clips(State *S)
{
 int i;

 cr_trace("\n\n Function assert_state_to_clips");
 for ( i = 0; i < S->num_F; i++ ) {
   assert_clips_fact(S->F[i]);
   positive_facts_in_state[S->F[i]]=1;
 }  

/* SFA Nov06. We added the fluents */
 assert_clips_fluents();

/*They are always the same
 assert_clips_relevant_facts();*/
}

void assert_clips_fluents() {
  int i,j;

  for ( i = 0; i < gnum_functions; i++ ) {
    /*    printf("\n\n%s:", gfunctions[i]);*/
    if ( gis_changed[i] ) {
      for ( j = 0; j < gnum_initial_function[i]; j++ ) {
	char fact[1000]="";
	print_Fluent_and_Value_to_string(&(ginitial_function[i][j].fluent), 
                   ginitial_function[i][j].value, fact, "TRUE-IN-STATE");
	cr_trace(fact);
	possible_clips_fluents[i][j]=AssertString(fact);
      }
    }
  }
}
  
void print_fl_name_to_string( int index, float value, char *output, char *head)
{

  int i;
  char aux[1000]="";

  if ( index < 0 ) {
    if ( index != -2 ) {
      printf("[UNDEF]");
    } else {
      strcpy(aux,"TOTAL-TIME");
    }
    return;
  }

  if ( grelevant_fluents_lnf[index] == NULL ) {
    /* this is a non-artificial "atomic" one
     * (or the mirrored version of one)     */
    sprintf(aux,"%s", grelevant_fluents_name[index]);
  } else {
    /* this only summarizes a LNF requirement */
    for ( i = 0; i < grelevant_fluents_lnf[index]->num_pF; i++ ) {
      print_fl_name_to_string(grelevant_fluents_lnf[index]->pF[i], value, 
                               output, head);
    }
  }

  sprintf(output,"(%s %s %.2f)", head, aux, value);
}

/*****************************************************************************
  NAME         : assert_clips_fact (int index)
  DESCRIPTION  : 
  INPUTS       : void
  RETURNS      : void
  SIDE EFFECTS :
  NOTES        : 
****************************************************************************/
void assert_clips_fact(int index)
{
   char buffer[1000]="";
   char fact[1000]="";

   if (possible_positive_clips_facts[index]==NULL)  {
      print_ft_name_to_string(index, fact, "TRUE-IN-STATE");
      possible_positive_clips_facts[index]= AssertString(fact);
      cr_trace(fact);
   }
}

/***************************************************************************
  NAME         : retract_clips_fact (int index)
  DESCRIPTION  : 
  INPUTS       : void
  RETURNS      : void
  SIDE EFFECTS :
  NOTES        : 
*****************************************************************************/
void retract_clips_fact(int index)
{ 
 
  if (possible_positive_clips_facts[index]!=NULL){
     Retract(possible_positive_clips_facts[index]);
     possible_positive_clips_facts[index]=NULL;
  }
}

/**************************************************************************
  NAME         : retract_clips_fact (int index)
  DESCRIPTION  : 
  INPUTS       : void
  RETURNS      : void
  SIDE EFFECTS :
  NOTES        : 
****************************************************************************/
void retract_clips_fluents()
{ 
  int i,j;

  for ( i = 0; i < gnum_functions; i++ ) {
    if (gis_changed[i]) {
      for(j=0; j< gnum_initial_function[i]; j++)
	if (possible_clips_fluents[i][j]) {
          Retract(possible_clips_fluents[i][j]);
	  possible_clips_fluents[i][j]=NULL;
	}
    }
  }
}



/***********************************************************************
  NAME         : get_h_from_control_rules
  DESCRIPTION  : call clips to get the heuristic value of a state
                 1) asserts the facts in the state
                 2) run clips
                 3) get the heuristic value
                 4) retract the facts in the state
  INPUTS       : The ff state
  RETURNS      : The heuristic value computed by the control rules. INFINITY if
                 no control rules have been fired
  SIDE EFFECTS :
  NOTES        : 
*************************************************************************/

float get_h_from_control_rules( State *S) 
{
  
  int fired_rules_number,i, first_salience, salience=0, 
      number_of_rules_to_fire=0;
  char buffer[1000]="";
  char *fluent_name;
  void *fact_pointer=NULL, *next_fact_pointer, *template;
  void * activation_pointer;
  float heuristic=INFINITY, min_heuristic=INFINITY, partial_heuristic;
  DATA_OBJECT result;


  /* initialize positive_facts_in_stat  */
  for ( i = 0; i < gnum_relevant_facts; i++ ) {
    positive_facts_in_state[i]=0;
  }
  /* assert new state */
  assert_state_to_clips(S);

  /* retract old facts not in state */
  for ( i = 0; i < gnum_relevant_facts; i++ ) {
    if (positive_facts_in_state[i]==0)
      retract_clips_fact(i);
  }

  /* new execution of all rules */
  /*new_computation_pointer=AssertString("(new-computation)");*/

  /* run clips */
   fired_rules_number=Run(-1L);

  if ( (gcmd_line.display_info == 130) || (gcmd_line.display_info == 132) )  {
    fprintf(stdout, "\n\n%s%d", "cr: number of fired control rules: \n", 
                     fired_rules_number);
  }
  /* end run clips  */
  /* take from clips the heuristic evaluation */
  /*if (fired_rules_number > 0) {*/
   
   fact_pointer=GetNextFact(fact_pointer);
    
   while (fact_pointer!=NULL) 
     {
	    next_fact_pointer=GetNextFact(fact_pointer);
            template=FactDeftemplate(fact_pointer);
	    if (template != NULL)  
	      {
		char *template_name;

		template_name=GetDeftemplateName(template);
		
		if (strcmp(template_name,"HEURISTIC-VALUE")== SAME)
		  {
		    heuristic=INFINITY;
		    /******************************************************/
		    /* Slot names are fluent names */
		    for ( i = 0; i < glnf_metric.num_pF; i++ ) {
		      if (glnf_metric.pF[i] == -2) {
			fluent_name="total-time";
		       }
		      else { 
			fluent_name=grelevant_fluents_name[glnf_metric.pF[i]];
		      }
		      if (!GetFactSlot(fact_pointer,fluent_name,&result)){
			 fprintf(stdout, "\ncr: warning: the metric fluent %s does not appear in control rules heuristic estimation. Assumed 0 \n",  fluent_name);
			 partial_heuristic=0.0;
		      }
		      else {
			partial_heuristic= DOToDouble(result);
		      }
		      if (heuristic == INFINITY){
			/* first time */
			heuristic= glnf_metric.pC[i] * partial_heuristic; 
		      }
		      else {
			heuristic= heuristic + glnf_metric.pC[i] * partial_heuristic; 
		      }
		    }

		    /******************************************************/

		    if (LESS(heuristic, min_heuristic))
		      {
			min_heuristic=heuristic;
		      }

		    if (gcmd_line.display_info == 130) {
		      cr_trace(" rules -->");
		      GetFactPPForm(buffer,1000,fact_pointer);
		      cr_trace(buffer);
		    }
		    /*Retract(fact_pointer);*/
		  }
	      }
	    fact_pointer=next_fact_pointer;
      }
   /*   }*/

  /*Retract(new_computation_pointer);*/
   return min_heuristic;
}

/*SFA Nov06. Control rules for selecting operators*/

/*************************************************************************
  NAME         : sus_get_operators_from_cr
  DESCRIPTION  : call clips to get operators from rules
                 1) asserts the facts in the state
                 2) run clips
                 3) get the operators from the rules
                 4) retract the facts in the state
                 5) Makes the intersection between the operators from the rules
                    and the operators find for ff (gA)
       The rules generate no-ordered facts: 
       (SELECT-OPERATOR (action "a") (vars "arg1" "arg2" ..."argn")).
       The instantiated operators are codified with integers and store in the
       variable gop_conn[index].action, where index is the integer used for 
       that action.
 
  INPUTS       : S: The ff state
                 lH: wether using gH or not
  RETURNS      : TRUE if gnum_CR > 0 otherwise FALSE
  SIDE EFFECTS :gCR: The intersection between operators from rules and 
                from ff (gA).
                If the intersection is nill it takes gA
                 gnum_CR: number of operators in gCR
                 
  NOTES   : It could easily extend to REJECT-OPERATOR and PREFER-OPERATOR rules
*****************************************************************************/
Bool sus_get_operators_from_cr(State *S, Bool lH) {
  
  int fired_rules_number,i, j, first_salience, salience=0, 
      number_of_rules_to_fire=0;
  char **rule_actions; 
  char ***rule_vars;
  int num_actions=0;
  int* in_rule_actions; 
  int *intersection;

  void *fact_pointer=NULL, *next_fact_pointer, *template;
  void * activation_pointer;

  if ( gcmd_line.display_info == 132 ) {
    printf("\n\n Function sus_get_operators_from_cr");
  }
 
  /* initialize gCR*/
  for (i=0; i< MAX_OPERATORS; i++)
    gCR[i]=0;
  gnum_CR=0;

  /* initialize positive_facts_in_stat  */
  for ( i = 0; i < gnum_relevant_facts; i++ ) {
    positive_facts_in_state[i]=0;
  }

  /* retract old fluents not in state */
  retract_clips_fluents();
  
  /* retract old facts not in state */
  for ( i = 0; i < gnum_relevant_facts; i++ ) {
    if (positive_facts_in_state[i]==0)
      retract_clips_fact(i);
  }

  /* assert new state */
  assert_state_to_clips(S);

  /* new execution of all rules */
  /*new_computation_pointer=AssertString("(new-computation)");*/

  /* run clips */

  if ( gcmd_line.display_info == 132 ) {
    printf("\n\n running CLIPS.....");
  }

     fired_rules_number=Run(-1L); 

  if ( gcmd_line.display_info == 132 ) {
    printf("\n\n cr: number of fired SELECT-OPERATOR control rules: %d \n",  fired_rules_number);
  }

  /* end run clips  */

  if (fired_rules_number == 0) return FALSE;

   /*Get memory for the actions and operators*/
   in_rule_actions=(int *) calloc(fired_rules_number,sizeof(int));
   rule_actions= (char **) calloc(fired_rules_number, sizeof (char *));
   rule_vars= (char ***) calloc(fired_rules_number, sizeof (char *));
   for (i=0; i< fired_rules_number; i++) {
      rule_actions[i]= (char *) calloc (MAX_NAME, sizeof (char *));
      rule_vars[i]= (char **) calloc (MAX_NAME, sizeof (char *));
      for (j=0; j<MAX_VARS; j++)
             rule_vars[i][j]=(char *) calloc(MAX_NAME,sizeof(char *));
  }

  /* fire only rules for selecting operators */

  
  /* take from clips the operators */
   fact_pointer=GetNextFact(fact_pointer); /*la primera es la intital-fact*/
    
   while (fact_pointer!=NULL) {

     next_fact_pointer=GetNextFact(fact_pointer);
     
     template=FactDeftemplate(fact_pointer);
	   
     if (template != NULL)  {
	char *template_name;

	template_name=GetDeftemplateName(template);
		
	if (strcmp(template_name,"SELECT-OPERATOR")== SAME) {

            DATA_OBJECT slotNames,slotValue;
            char *slot_name;
            void *mulPtr;
                    
            /*FactSlotNames(fact_pointer, &slotNames);*/
           /*It has (slot action) and (multislot vars)*/
	    /*slot_name,ValueToString(GetMFValue(slotNames,1)));*/
	    GetFactSlot(fact_pointer,"ACTION", &slotValue);
            /*if (slotValue == NULL)
              printf("\n Error in slot action of templete SELECT-OPERATOR");
	    else*/ 
            strcpy(rule_actions[num_actions], DOToString(slotValue));

	    /*slot_name=ValueToString(GetMFValue(slotNames,2));*/
	    GetFactSlot(fact_pointer,"VARS", &slotValue);
            mulPtr= GetValue(slotValue);
            /*if (slotValue == NULL)
              printf("\n Error in multislot vars of templete SELECT-OPERATOR");
	    else {*/
            for (i=GetDOBegin(slotValue);i<=GetDOEnd(slotValue);i++)
		 strcpy(rule_vars[num_actions][i-1], 
                        ValueToString(GetMFValue(mulPtr,i)));
            if ((in_rule_actions[num_actions]=
                 sus_get_index_action(rule_actions[num_actions],
                                      rule_vars[num_actions])) != -1)
		    num_actions++;
	    else 
               if ( gcmd_line.display_info == 132 ) {
                printf("\n cr: Action no found %s ",rule_actions[num_actions]);
                 for (j=0;j<MAX_VARS;j++)
                    printf("%s ", rule_vars[num_actions][j]);
	    }
            Retract(fact_pointer);
	 }
       }
       fact_pointer=next_fact_pointer;  
  }
   
  /*Make the intersection with gA*/

  /*Rules are changing order actions*/
   for (i=0; i<num_actions; i++)
     if (sus_in_gAH(in_rule_actions[i],lH)) 
       gCR[gnum_CR++]=in_rule_actions[i];

  if ( gcmd_line.display_info == 132 ) {
    printf("\n cr: Rules-fired %d: ", num_actions);
    for (i=0; i < num_actions;i++) {
      printf("\n    ");
      print_op_name(in_rule_actions[i]);
    }
    if (lH){
       printf("\n cr: gnum_H %d: ", gnum_H);
       for (i=0; i < gnum_H;i++) {
         printf("\n    ");
         print_op_name(gH[i]);
       }
    } else {
             printf("\n cr: gnum_A %d: ", gnum_A);
             for (i=0; i < gnum_A;i++) {
             printf("\n    ");
             print_op_name(gA[i]);
             }
    }
    printf("\n cr: Intersection %d", gnum_CR);
    for (i=0; i< gnum_CR; i++){
       printf("\n    ");
       print_op_name(gCR[i]);
    }
    printf("\n    ");
  }

   if (gnum_CR > 0 ) return TRUE;
   else return FALSE;

}


Bool sus_in_gAH(int v, Bool lH) {
  int i;
  
  if (lH) {
     for(i=0; i < gnum_H; i++)
       if (v==gH[i]) break;

     return  ( i==gnum_H) ? FALSE : TRUE;
  } else {

      for(i=0; i < gnum_A; i++)
          if (v==gA[i]) break;
 
      return  ( i==gnum_A) ? FALSE : TRUE;
  }
}

/*Return de number in gop_conn assigned to the instantiated action*/
/*-1 if error*/
int sus_get_index_action(char *action, char **vars) {

  Action *a;
  int i=-1,j,k;
 
  for(i=0;i<gnum_op_conn;i++) {
       a=gop_conn[i].action;
       if ( (a != NULL) && (a > 0x2 )){ 
         if (strcmp(a->name,action) == SAME) {
            for(j=0,k=0;j<a->num_name_vars;j++){
	      if (strcmp(gconstants[a->name_inst_table[j]],vars[j]) == SAME) k++;
          } 
       if (k >= (a->num_name_vars)) return i;
       }
       }
  }
 

  return  -1;
}


void sus_print_index_table(){
  int i,j;
  IndexAction *a;

  printf("\n SFA PRINT OP\n");
  for(i=0;i<gnum_op_conn;i++) {
    printf("%d: ", i);
    print_op_name(i);
    printf("\n ");
  }

  sus_print_types();
  printf("\n fin SFA \n"); 
}

void sus_print_types(){
  int i,j;

  for ( i = 0; i < gnum_types; i++ ) {
    printf("gtype_names[%d]=%s\n", i, gtype_names[i]);  
    for ( j = 0; j < gtype_size[i]; j++ ) {
        printf("  gconstants[gtype_consts[%d][%d]=%s\n", i,j,
                                             gconstants[gtype_consts[i][j]]);
      }}
}
