(in-package "USER")

;; Usage (translate-pdl-cr-to-clips "~/tmp/rules-prodigy.lisp" "~/tmp/rules-prodigy.clp")
(defun translate-pdl-cr-to-clips (ifile ofile)
  (with-open-file (istream ifile :direction :input)
    (with-open-file (ostream ofile :direction :output :if-exists :supersede :if-does-not-exist :create)
      (format ostream "(deftemplate SELECT-OPERATOR 
   (slot action
 	(type SYMBOL))
   (multislot vars
 	(type SYMBOL))
)")
      (do ((exp (read istream nil 'eof) (read istream nil 'eof)))
	  ((eq exp 'eof))
	(if (and (listp exp)
		 (eq (car exp) 'control-rule)
		 (eq (caddr (cadddr exp)) 'bindings))
	    (translate-rule exp ostream))))))

(defun translate-rule (cr stream)
  (let ((current-op nil))
    (format stream "~%(defrule ~a" (cadr cr))
    (setq current-op (translate-preconds (cdadr (caddr cr)) stream))
    (translate-effects (cadddr cr) current-op stream)
    (format stream "~%)~%" (cadr cr))))

(defun translate-preconds (preconds stream)
  (let ((current-op nil)
	(translation nil))
    (dolist (precond preconds)
	(case (car precond)
	    (current-operator (setq current-op (cadr precond)))
	  (current-goal (format stream "~%   (relevant-fact")
			(pp-list (setq translation (pdl-exp-to-clips (cadr precond))) 1 stream t t)
			(format stream ")")
			(format stream "~%   (not (true-in-state")
			(pp-list translation 1 stream t t)
			(format stream "))"))
	  (true-in-state (format stream "~%   (true-in-state")
			 (pp-list (pdl-exp-to-clips (cadr precond)) 1 stream t t)
			 (format stream ")"))
	  (type-of-object (format stream "~%   (type-of-object")
			  (pp-list (pdl-exp-to-clips (cdr precond)) 1 stream t t)
			  (format stream ")"))
	  (t nil)))
    current-op))

(defun pdl-exp-to-clips (exp)
  (mapcar #'(lambda (obj)
	      (if (p4::strong-is-var-p obj)
		  (intern (concatenate 'string "?" (string-trim "<>" (format nil "~a" obj))))
		  obj))
	  exp))

(defun translate-effects (effects current-op stream)
  (format stream "~2%=>~%   (assert (SELECT-OPERATOR (action ~a) (vars"
	  current-op)
  (pp-list (mapcar #'(lambda (binding) (car (pdl-exp-to-clips (list (cdr binding)))))
		   (cadddr effects))
	   1 stream t t)
  (format stream ")))"))


;; It prints a list of elements in a stream with a tab
(defun pp-list (list &optional (tab 1) (ostream t) lowercasep same-line-p)
 (dolist (literal list)
     (format ostream (format nil (concatenate 'string (if same-line-p " " "~~%")
                          (if lowercasep
                          "~~~dt~~(~~a~~)"
                          "~~~dt~~a"))
                 tab)
         literal)))


