(in-package "PRODIGY4")

(defun check-binding (oblist rule)
  "Check that the preconditions of the rule are satisfied with the
given bindings."
  (declare (list oblist)
	   (type operator rule))
  (check-binding-r oblist (operator-vars rule)
		   (third (operator-precond-exp rule)) rule t))

;;aperez 3/93 added rule argument to allow exists and foralls in
;;eager inference rules
(defun check-binding-r (oblist vars expr rule want-true-p)
  (cond ((or (and (eq (car expr) 'user::and) want-true-p)
	     (and (eq (car expr) 'user::or) (not want-true-p)))
	 (catch 'failed
	   (mapcan
	    #'(lambda (x) (or (check-binding-r oblist vars x rule want-true-p)
			      (throw 'failed nil)))
		   (cdr expr))))
	((or (and (eq (car expr) 'user::or) want-true-p)
	     (and (eq (car expr) 'user::and) (not want-true-p)))
	 (some #'(lambda (x) (check-binding-r oblist vars x rule want-true-p))
	       (cdr expr)))
	((eq (car expr) 'user::~)
	 (check-binding-r oblist vars (second expr) rule (not want-true-p)))

	;;aperez: allow exists and foralls in eager inference rules
	((or (and (eq (car expr) 'user::forall) want-true-p)
	     (and (eq (car expr) 'user::exists) (not want-true-p)))
	 (check-forall-binding oblist vars expr rule want-true-p))

	((or (and (eq (car expr) 'user::exists) want-true-p)
	     (and (eq (car expr) 'user::forall) (not want-true-p)))
	 (check-exists-binding oblist vars expr rule want-true-p))
		
	;; Anything else is assumed to be a predicate. I use the
	;; double negation to make sure that everything is either t or
	;; nil.
	((let ((lit (make-up-literal expr oblist vars)))
	   (if (eq (not (literal-state-p lit))
		   (not want-true-p))
	       (list lit))))))


;;This has to return the list of literals satisfied in the forall
;;(that is what check-binding should return to be stored as justification)

(defun check-forall-binding (oblist vars expr rule want-true-p)
  ;;checks the body with all possible bindings for the quantified variables.
  ;;assoc finds quantifier generator for this forall (there may be
  ;;several foralls in the operator) and then funcall gets all
  ;;the possible bindings for each the vars
  (do* ((gen (cdr (assoc (second expr)
			 (getf (rule-plist rule)
			       :quantifier-generators))))
	(data (if gen (funcall gen nil)))
	(choice (make-list (length data) :initial-element 0)
		(increment-choice choice data))
	(forall-bindings (choice-bindings expr data choice)
			 (choice-bindings expr data choice))
	this-binding-justif final-justif)
       ((null choice) (reverse final-justif))
    (setf this-binding-justif
	  (check-binding-r
	   (append oblist (mapcar #'cdr forall-bindings)) ;;objects
	   (append vars (mapcar #'car forall-bindings)) ;;vars
	   (third expr) rule want-true-p))
    (if this-binding-justif
	(push this-binding-justif final-justif)
	(return nil))))

(defun check-exists-binding (oblist vars expr rule want-true-p)
  (do* ((gen (cdr (assoc (second expr)
			 (getf (rule-plist rule)
			       :quantifier-generators))))
	(data (if gen (funcall gen nil)))
	(choice (make-list (length data) :initial-element 0)
		(increment-choice choice data))
	(exists-bindings (choice-bindings expr data choice)
			 (choice-bindings expr data choice))
	this-binding-justif)
       ((null choice) nil)
    (setf this-binding-justif
	  (check-binding-r
	   (append oblist (mapcar #'cdr exists-bindings)) ;;objects
	   (append vars (mapcar #'car exists-bindings)) ;;vars
	   (third expr) rule want-true-p))
    (if this-binding-justif (return this-binding-justif))))
  


