;; SAYPHI
;; Sistema de Aprendizaje Y Planificacion Heuristica Informativa
;; ______________________________________________________________
;; 
;; Td. Tomas de la Rosa
;; PLG - Planning & Learning Group
;; Universidad Carlos III de Madrid, Espaa
;; 09.12.2005
;; ______________________________________________________________
;; 

;; Loader File
;; Sorry if it seems like Prodigy loader :-)
;; This should automatically load SAYPHI, and compile it if there's no 
;; fas files

(in-package "COMMON-LISP-USER")

;;You should change this path to your work directory for SAYPHI 
;;The default is the directory in which the LISP system is called
(defvar *sayphi-root* (namestring (truename "./sayphi/")))

(defvar *system-dir* (concatenate 'string *sayphi-root* "system/"))
(defvar *bin-dir* (concatenate 'string *system-dir* 
			       #+ALLEGRO "allegro-bin/"
			       #+CLISP "compiled/"
			       #+SBCL "sbcl-bin/"))

(defvar *domains-dir* (concatenate 'string *sayphi-root* "domains/"))
(defvar *domain-dir* nil)


(defvar *say-lisp-extension* ".lisp")
(defvar *bin-extension* ".fas")

(defvar *auto-load* t)
(defvar *load-full* nil)

(defvar *sayphi-modules*
  '(("planner/" ("trace-error" "structures" "genericfuns" "parse-domain" "read-domain" 
		 "read-problem" "instantiate" "initmetric" "metric" "planning" "numheuristic" 
 		 "print" "ebl" "search"  "support" "runsets"))))


(defparameter *sayphi-modules-extra*
  '(("planner/" ("ebl" "tree-study" "rete" "training" "relational-search" "oversubscription"))))
  

(defvar *cbp-sayphi-modules*
  '(("cbr/" ("cbr-structures" "cbrgeneralfuns"  "casefs_interface" "property"
	     "utility-cases" "merge-case" "store-case" "select-replay"  "unify-typeseq"  
 	     "replay-case" "cbr-runsets" ))))


(defun load-sayphi (&key (load-source nil)
			 (modules *sayphi-modules*))
  (dolist (module modules)
    (dolist (file (cadr module))
      (if load-source
	  (load (concatenate 'string *system-dir* (car module) file *say-lisp-extension*))
	(load (concatenate 'string *bin-dir* file *bin-extension*))))))

(defun compile-all (&key (modules *sayphi-modules*))
  (dolist (module modules)
    (dolist (file (cadr module))
      (compile-file (concatenate 'string *system-dir* (car module) file  *say-lisp-extension*)
       :output-file (concatenate 'string *bin-dir* file *bin-extension*)))))


(defun load-cbp()
  (load-sayphi :load-source t :modules *cbp-sayphi-modules*)
  (compile-all :modules *cbp-sayphi-modules*)
  (load-sayphi :load-source nil :modules *cbp-sayphi-modules*)
  )

(when *auto-load*
  (progn 
    (setf *load-verbose* t)
    (load-sayphi :load-source t)
    (compile-all)
    (setf *load-verbose* nil)
    (load-sayphi :load-source nil)
    
    (when *load-full* 
      (setf *load-verbose* t)
      (load-sayphi :load-source t :modules *sayphi-modules-extra*)
      (compile-all :modules *sayphi-modules-extra*)
      (setf *load-verbose* nil)
      (load-sayphi :load-source nil :modules *sayphi-modules-extra*))
    
    (pp-welcome)
))



