;; SAYPHI Case-Based Reasoner
;; Learning Applicable Context
;; Td. 05.02.2007
;; _________________________________________________________________________________________


(defvar *applicable-context-table* nil)
(defvar *typed-context-integers* nil)
(defvar *typed-context-keys* nil)

(defmacro to-context-int (action-typed-relation)
  `(gethash ,action-typed-relation  *typed-context-integers*))

(defmacro to-context-key (context-integer)
  `(gethash ,context-integer *typed-context-keys*))

(defun acx-init ()
  (setf *applicable-context-table* (build-applicable-context-table))
  (build-context-integers))


(defun build-applicable-context-table () 
  (let ((acx-table (make-hash-table :test #'eq)))
    (dolist (i-action (dom-actions *pspace*) acx-table)
      (setf (gethash (action-name i-action) acx-table) 
	    (make-hash-table :test #'equal)))))


;; Fills the type-context-table with a correnponding integer key for faster comparisions during the search
(defun build-context-integers ()
  (setf *typed-context-integers* (make-hash-table :test #'equal))
  (setf *typed-context-keys* (make-hash-table))
  (let ((context-pos 0)
	(bsfield-action (byte 4 6))
	(bsfield-params (byte 6 0)))
    (dolist (iaction (dom-actions *pspace*))
      (let ((params (action-parameters iaction))
	    (action-pos (gethash (action-name iaction) *action-postable*)))
	(setf (ldb bsfield-action context-pos) action-pos)
	(dotimes (ipos (expt 2 (length params)))
	  (let ((typed-list nil)
		(param-len (1- (length params))))
	    (dotimes (iparam-pos (length params))
	      (when (logbitp iparam-pos ipos)
		(push (cdr (nth (- param-len iparam-pos) params)) typed-list)))
	    (push (action-name iaction) typed-list)
	    (setf (ldb bsfield-params context-pos) ipos)
	    (setf (gethash typed-list *typed-context-integers*) context-pos)
	    (setf (gethash context-pos *typed-context-keys*) typed-list)))))))
		     


(defun extract-applicable-context-cases()
  (when (and (solution-p *say-solution*)
	     (solution-found *say-solution*))
    (let* ((solseq (solution-path *say-solution*))
	   (context-node (car solseq))
	   (context-action (car (snode-plan-action context-node))))
      (dolist (x-node (rest solseq)  *applicable-context-table*)
	(let* ((siblings (snode-children (snode-parent x-node)))
	       (siblings-int (cons-siblings-int siblings))
	       (action-acx-table (gethash context-action *applicable-context-table*))
	       (path-typed-relation nil))
	  (setf path-typed-relation
		(create-path-type-relation context-node (get-trace-flatpath x-node)))
	  (unless (null path-typed-relation)
	    (cond ((null (gethash siblings-int action-acx-table))
		   (setf (gethash siblings-int action-acx-table)
			 (list (list (length path-typed-relation) path-typed-relation 1))))
		  (t
		   (setf (gethash siblings-int action-acx-table)
			 (merge-path-typed-relations (list (length path-typed-relation) path-typed-relation 1)
						     (gethash siblings-int action-acx-table)))
		   ))))
	(setf context-action (car (snode-plan-action x-node)))
	(setf context-node x-node)
	))))


(defun create-path-type-relation (context-node node-list)
  (let* ((direct-actions (get (dom-plist *pspace*) :direct-actions))
	 (this-context-objects (rest (snode-plan-action context-node)))
	 (path-type-relation nil))
;;     (format t "~% Extracting Context: ~% ~a ~a ~%" context-node node-list)
    (dolist (i-node node-list path-type-relation)
      (let ((action-def (gethash (car (snode-applied-action i-node)) direct-actions))
	    (action-type-relation (list (car (snode-plan-action i-node))))
	    (objpos 0))
;; 	(format t "~%__$ ~a ~a" (rest (snode-plan-action i-node)) this-context-objects)
	(dolist (i-object (rest (snode-plan-action i-node)))
	  (when (find i-object this-context-objects :test #'eq)
;; 	    (format t "~%__$>> ~a " i-object)
	    (setf action-type-relation (append action-type-relation 
					       (list (cdr (nth objpos (action-parameters action-def)))))))
	  (incf objpos))
	(setf path-type-relation (append path-type-relation (list (to-context-int action-type-relation))))
	(setf this-context-objects (rest (snode-plan-action i-node)))
	))))
	      

;; Does the same as get-path-type-relation but only with a node and its context-node
(defun get-action-typed-relation (context-node node)
  (let* ((direct-actions (get (dom-plist *pspace*) :direct-actions))
	 (this-context-objects (rest (snode-plan-action context-node)))
	 (action-def (gethash (car (snode-applied-action node)) direct-actions))
	 (action-typed-relation (list (car (snode-plan-action node))))
	 (objpos 0))
    
    (dolist (i-object (rest (snode-plan-action node)) action-typed-relation)
      (when (find i-object this-context-objects :test #'eq)
	(setf action-typed-relation (append action-typed-relation 
					    (list (cdr (nth objpos (action-parameters action-def)))))))
      (incf objpos))
	))


;; Get a sequence of the solution path from the current node until a node with a heuristic value
;; improvement is found.  Should not be called with the initial node.	      
(defun get-trace-flatpath (node)
  (let ((flatpath nil)
	(h-path (snode-h-value (snode-parent node)))
	(exit-found nil))
    (do ((i-node node (snode-selected i-node)))
	(exit-found (reverse flatpath))
	(push i-node flatpath)
	(when (< (snode-h-value i-node) h-path)
	  (setf exit-found t)))))
    

(defun merge-path-typed-relations (new-typed-relation stored-typed-relations)
  (let ((res-typed-relation)
	(found-relation nil))
  (dolist (i-relation stored-typed-relations)
    (when (and (not found-relation)
	       (= (car i-relation) (car new-typed-relation))
	       (equalp (second i-relation) (second new-typed-relation)))
      (incf (third i-relation))
      (setf found-relation t))
    (push i-relation res-typed-relation))
  (unless found-relation
    (push new-typed-relation res-typed-relation))
  (setf new-typed-relation (sort-path-relations res-typed-relation))
  ))


;; Sort the path-typed-relation ordering by length of the scaping sequence, and then by the most frequent sequences
(defun sort-path-relations (path-typed-relations)
  (setf path-typed-relations 
	(stable-sort path-typed-relations #'(lambda (x y)
				      (cond ((< (car x) (car y)) t)
					    ((= (car x) (car y))
					     (if (>= (third x) (third y)) t nil))
					    (t nil))))))

 
(defun pp-applicable-context()
  (format t "~% APPLICABLE PATH CONTEXT")
  (format t "~% ==================================================")
  (maphash #'(lambda (key action-context)
	       (format t "~% ~a Context:" key)
	       (pp-action-applicable-context action-context))
	   *applicable-context-table*))

(defun pp-action-applicable-context (action-context-table)
  (maphash #'(lambda (key typed-relations)
	       (format t "~%    Cx: ~a" (to-siblings-key key))
	       (dolist (i-relation typed-relations)
		 (format t "~%      -> len:~a  Freq:~a  Rel:" (car i-relation) (third i-relation))
		 (dolist (i-int (second i-relation))
		   (format t "~a" (to-context-key i-int)))))
	   action-context-table))


;; ====================================================================================
;; ACX Using path applicable context to order nodes in EHC
;; ===================================================================================

(defun context-ehc (init-node h-fn &optional (problem *current-problem*))
  (declare (ignore cost-fn))
  (let ((open-nodes nil) (open-children nil)
	(node init-node) (next-node nil) (visited nil)
	(current-h 0) (better-h-node nil) (flat-path-depth 1))
    (store-h-extras node h-fn)
    (setf current-h (snode-h-value node))
    (do* ((stop-this (stop-search node problem) (stop-search node problem)))
	 (stop-this (when-stop (cdr stop-this) node problem))

	 (expand-state node :helpful *say-helpful*)
	 (push node visited)

	 (setf better-h-node (get-context-heuristic node h-fn current-h flat-path-depth visited))
	 (cond ((snode-p better-h-node)
		(setf current-h (snode-h-value better-h-node))
		(setf node better-h-node)
		(setf open-nodes nil)
		(setf open-children nil)
		(setf flat-path-depth 1))
	       (t 
		(setf open-children (append (remove-if #'snode-closed (snode-children node)) 
					  open-children))
		(when (> *say-output* 1) (format t "  ~% [Expanding Breadth Level]"))
		(cond ((not (snode-p (setf next-node (pop open-nodes))))
;; 		     
		       (setf open-nodes (stable-sort open-children #'> :key #'snode-advise-weight))
		       (incf flat-path-depth)
 		       (when (null open-nodes) ;;Recuperando la poda por las helpful actions
			 (setf open-nodes (restore-nonhelpful node h-fn visited))
 			 (when (> *say-output* 1) 
			   (format t "  ~%!!Restoring NON Helpful Nodes.")))
		       (setf node (pop open-nodes))
		       (setf open-children nil))
		      (t 
		       (setf node next-node)))))       
	 (when (snode-p node)
	   (print-search-node node nil)
))))


(defun get-context-heuristic (node h-function current-h flat-path-depth visited-nodes &key (discrete *say-h-rounded*))
  (let ((better-found nil) (node-counter 0) (exit-depth (1- flat-path-depth)))
    (ck-sort-siblings node)
    (do ((i-node (next-context-advise node exit-depth)))
	((or better-found (null i-node)) nil)
	(cond ((not (find i-node visited-nodes :test #'equal-state))
	       (incf node-counter)
	       (store-h-extras i-node h-function)
	       (cond (discrete
		      (when (< (floor (snode-h-value i-node)) (floor current-h)) 
			(setf better-found i-node)
			(when *trace-ehc* (push node-counter *hit-node*))))
		     (t
		      (when (< (snode-h-value i-node) current-h) (setf better-found i-node))))) 
	      (t
	       (setf (snode-closed i-node) t))
	      
	      )
	(unless better-found 
	  (setf i-node (next-context-advise i-node exit-depth)))
	)
    (cond ((snode-p better-found) better-found)
	  (t
	   ;; In this case non of the advise is OK, the rest of the children are evaluated
    	   (dolist (i-child (snode-children node) better-found)
	     (when (null better-found)
	       (when (null (snode-h-value i-child))
		 (cond ((not (find i-child visited-nodes :test #'equal-state))
			(incf node-counter)
			(store-h-extras i-child h-function)
			(cond (discrete
			       (when (< (floor (snode-h-value i-child)) (floor current-h))
				 (setf better-found i-child)
				 (when *trace-ehc* (push node-counter *hit-node*))))
			      (t
			       (when (< (snode-h-value i-child) current-h)
				 (setf better-found i-child))))) 
		       (t
			(setf (snode-closed i-child) t)))
		 )))))))


;;This calculates the action-typed-relation for the node children
(defun map-action-typed-relation (node)
  (dolist (i-child (snode-children node))
    (setf (snode-typed-relation i-child)
	  (to-context-int (get-action-typed-relation node i-child)))))
  
(defun next-context-advise (node exit-depth)
  (when (snode-p (snode-parent node))
    (map-action-typed-relation node)
    (let* ((context-action (car (snode-plan-action node)))
	   (context-table (gethash context-action *applicable-context-table*))
	   (this-siblings-int (cons-siblings-int (snode-children node)))
	   (context-options (gethash this-siblings-int context-table))
	   (advise-node nil)
	   (advise-found nil))
;;         (format t "~%  __>> Context Options: [~a] ~a " context-action this-siblings-int)
;;         (format t "~%  __>>>  Options: ~a " context-options)
      (dolist (i-option context-options advise-node)
	(dolist (i-child (snode-children node) advise-node)
	  (unless advise-found
	    (when (null (snode-h-value i-child))
;;  	      (format t "~% ___>>>> Searching for Advise...~a ~a :: ~a ~a" 
;;  		      (to-context-key (snode-typed-relation i-child)) (snode-typed-relation i-child)  
;;  		      (car (second i-option))  (to-context-key (car (second i-option))))
;; 		      (nth exit-depth (second i-option))  (to-context-key (nth exit-depth (second i-option))))
	      (cond ((and 
;; 		      (numberp (nth exit-depth (second i-option)))
			 (eql (snode-typed-relation i-child)
			      (car (second i-option) )))
;; 			      (nth exit-depth (second i-option) )))
;; 	  	              Queda pendiente escoger del contexto superior para guiar el exit-depth 

		    (setf advise-found t)
		    (setf advise-node i-child)
		    (setf (snode-advise-weight advise-node) 2)
		    (when (> *say-output* 1)
		      (format t "~% ______>> Context-Advise: ~a ~a" 
			      this-siblings-int (to-context-key (snode-typed-relation advise-node)))
		      (format t "~% ______>>> ~a" (snode-plan-action advise-node)))
		    )))))))))



  


