;; SAYPHI Case-Based Reasoner
;; Running the Case-Based Reasoner
;; Td. 10.02.2006
;; _________________________________________________________________________________________

(defvar *expconv-rootprefix* "")
(defvar *expconv-testprefix* "")


(defun sol-trace(trace-info)
  (case trace-info
	(:first-hit (count 1 *hit-node*))
	(:average-hit (float (/ (apply #'+ *hit-node*)  (length *hit-node*))))
	(:guided-nodes *cbp-guided-nodes*)
	(:cbp-first-hit (count 1 *cbp-hit-node*))
	(:cbp-average-hit 
	 (if (> (length *cbp-hit-node*) 0)
	     (float (/ (apply #'+ *cbp-hit-node*)  (length *cbp-hit-node*)))
	     -1))
))

(defun set-cboption (option value)
  (cond ((eq option 'trace-used)
	 (setf *cbinfo-usedcases* (make-hash-table :test #'equal))
	 (setf *cboption-traceselected* value))))



(defun reset-cboptions ()
  (setf *cboption-traceselected* nil)
  (setf *cbp-store-utility* nil))


;; (defun delete-casebase()
;;   (setf *type-case-base* nil)
;;   (dolist (itype (get-all-types (dom-domtypes *pspace*)))
;;     (when (probe-file (format nil "~a/cases/~a.cbp" *domain-dir* itype))
;;       (delete-file (format nil "~a/cases/~a.cbp" *domain-dir* itype)))))

(defun delete-casebase()
  (setf *type-case-base* nil)
  (let ((dirfiles (directory (format nil "~acases/*.*" *domain-dir*))))
    (dolist (ifile dirfiles)
      (format t "~%~a" ifile)
      (when (probe-file ifile)  
	(delete-file ifile)))))

	     


(defun save-usedcases(out-file)
  (when (hash-table-p *cbinfo-usedcases*)
    (let ((file-path (format nil "~a/result/info_~a" *domain-dir* out-file)))
      (with-open-file (out-stream file-path :direction :output :if-exists 
				  :append :if-does-not-exist :create)
		      (format out-stream "CBR SAYPHI INFO USED CASES~%")
		      (pp-hash-table *cbinfo-usedcases* out-stream)))))


(defun format-solution (solution)
  (case *say-solution-format*
	(:values
	 (format nil "~t~a, ~t~4$, ~t~a, ~t~a, ~t~a, ~t~a, ~t~a"
		 (solution-found solution) (solution-total-time solution)
                 (solution-depth solution) (solution-num-nodes solution) (solution-evaluated-nodes solution)
		 (solution-total-cost solution) (solution-stop-reason solution)))

	(:extras
	 ;;agregando al final de los valores el numero de aciertos a la primera y el promedio de hit
	 (format nil "~t~a, ~t~a, ~t~a, ~t~a, ~t~a, ~t~a, ~t~a, ~t~a"
		 (solution-found solution) (solution-total-time solution)
                 (solution-depth solution) (solution-num-nodes solution) (solution-evaluated-nodes solution)
		 (solution-stop-reason solution)
		 (sol-trace :first-hit) (sol-trace :average-hit)
		 ))
	(:cbp-extras
	 ;;agregando al final de los valores el numero de aciertos a la primera y el promedio de hit
	 (format nil "~t~a, ~t~4$, ~t~a, ~t~a, ~t~a, ~t~a, ~t~a, ~t~a, ~t~a"
		 (solution-found solution) (solution-total-time solution)
                 (solution-depth solution) (solution-num-nodes solution) (solution-evaluated-nodes solution)
		 (solution-stop-reason solution)
		 (sol-trace :guided-nodes)   
		 (sol-trace :cbp-first-hit)
		 (sol-trace :cbp-average-hit)
		 ))
	(:tags-values
	 (format nil "Found:~a Time:~a Nodes:~a Depth:~a Stop:~a"
		 (solution-found solution) (solution-total-time solution)
		 (solution-num-nodes solution) (solution-depth solution)
		 (solution-stop-reason solution)))))


(defun cbp-runset (domain domain-file &key (algorithm *say-algorithm*)
		   (heuristic nil)
		   (runtype *cbp-runtype*)
		   (retrieve-option *cbp-retrieve-option*)
		   (max-solutions most-positive-fixnum)
		   (helpful *say-helpful*)
		   (w_g nil)
		   (w_h nil)
		   (try2-search nil)
		   (cost *say-costfn*) 
		   (timeout 5) (output-level 1)
		   (depthbound 2000) 
		   (probs-prefix "p") (save-result nil)
		   (store-case nil) (use-case nil)
		   (store-utility nil) 
		   (out-file nil)
		   (analyze-tree nil)
		   (experiment-var nil)
		   (trace-selected *cboption-traceselected*)
		   )
  (setf *say-output* output-level)
  (say-domain domain domain-file)
  (let* ((probdir (format nil "~aprobsets/" *domain-dir*))
	 (count-prob 1) (this-sol nil) 
	 (result-file (if (not (stringp out-file))
			  (format nil "~a/result/~aset-~a-~a-~d.csv"
				  *domain-dir* probs-prefix algorithm heuristic timeout)
			(format nil "~a/result/~a"
				*domain-dir* out-file)))
	 (wild-char #+SBCL "*.*"
		    #-SBCL "*"))
    (when trace-selected
      (setf *cbinfo-usedcases* (make-hash-table :test #'equal))
      (setf *cboption-traceselected* t))
    (when store-utility 
      (read-type-case-base)
      (setf *cbp-store-utility* t))

    (dolist (thisprob (dirfiles-sorted probdir (concatenate 'string probs-prefix wild-char)))
      (format t "~% Solving Problem ~a" (pathname-name thisprob))
      (read-pddl-problem thisprob)
      (setf *problem-file* (pathname-name thisprob))
      (if use-case
	  (progn 
	    (setf this-sol (plan-cbp :algorithm algorithm ;:try2-search try2-search
				     :runtype runtype :retrieve-option retrieve-option
				     :helpful helpful :w_g w_g :w_h w_h :max-solutions max-solutions
				     :heuristic heuristic :cost cost
				     :timeout timeout :depthbound depthbound)))
	(setf this-sol (plan :algorithm algorithm :try2-search try2-search :max-solutions max-solutions
			     :heuristic heuristic :cost cost
			     :helpful helpful :w_g w_g :w_h w_h
			     :timeout timeout :depthbound depthbound)))
      (when (and store-case (solution-p this-sol))
	(if (solution-found this-sol)
	    (store-case)
	    (format t "~% Skip Store-case")))

      (when (and (not store-case) store-utility)
	(compute-cbp-utility))

      (when save-result
	(with-open-file (out-stream result-file :direction :output :if-exists 
				    :append :if-does-not-exist :create)
	        (format out-stream "~d, ~a," count-prob (pathname-name thisprob))
		(format out-stream "~a~%" (format-solution this-sol))))
      (unless (null experiment-var)
	(push (list-solution count-prob (pathname-name thisprob) this-sol) 
	      (gethash experiment-var *experiment-results*)))
      
      (when analyze-tree
	(let ((ehc-tree (ehc-searchtree-analysis)))
	  (study-data-analyze-depth ehc-tree 'diffset runtype (get-problem-subset))
	  (study-data-analyze-depth ehc-tree 'testset runtype)
	  (unless (null experiment-var)
	    (push (list-treestudy count-prob (pathname-name thisprob) ehc-tree)
		  (gethash experiment-var *experiment-study*)))
	  
	  ))
      (incf count-prob)
      (say-mem-free-pointers)
      #+SBCL (sb-ext:gc :full t)
      (room nil)
    )
    (reset-cboptions)
))



;; =========================================================================
;; PARA EXPERIMENTOS DE CONVERGENCIA
;; =========================================================================


(defmacro convergence-key (root-prefix num)
  `(intern (concatenate 'string (string-upcase ,root-prefix) (format-groupprob ,num))))

(defmacro this-convresult (root-prefix num)
  `(let ((conv-key (intern (concatenate 'string (string-upcase ,root-prefix) (format-groupprob ,num)))))
     (gethash conv-key *experiment-results*)))

(defun create-conv-accumulators(fields-list root-prefix times)
  (let ((accumulators (make-hash-table :test #'eq)))
    (dotimes (i-time times accumulators)
      (setf (gethash (convergence-key root-prefix (1+ i-time)) accumulators)
	    (make-array (1+ (apply #'max fields-list)) :initial-element 0)))))


(defun convergence-problems-solved (out-file)
  (let ((file-path (format nil "~a/result/~a" *domain-dir* out-file)))
    (with-open-file (out-stream file-path :direction :output :if-exists 
				:append :if-does-not-exist :create)
      (format out-stream "SAYPHI EXPERIMENT - NUMBER OF PROBLEMS SOLVERD ~%")
      (maphash #'(lambda (test results)
		   (format out-stream "~%~a:~a" test
			   (count t results :key #'third)))
	       *experiment-results*
	       ))))


(defun convergence-data-toplot (root-prefix n-times out-file &key (fields '(3 4 6))
			         (acc-fields '(3 4 6)))
  (let ((file-path (format nil "~a/result/~a" *domain-dir* out-file))
	(acc-table (create-conv-accumulators acc-fields root-prefix n-times))
	(test-primary (this-convresult root-prefix 1))
	(acc-array nil))
    (with-open-file (out-stream file-path :direction :output :if-exists 
			      :supersede :if-does-not-exist :create)
       (dotimes (i (length test-primary))
	 (when (exp-same-solved i)
	   (format out-stream "~a ~t ~a ~t" (car (nth i test-primary)) (second (nth i test-primary)))
	   (dotimes (i-time n-times)
	     (setf acc-array (gethash (convergence-key root-prefix (1+ i-time)) acc-table))
	     (dolist (i-field fields)
	       (format out-stream "~a ~t" 
		       (format-plotfield (nth i-field (nth i (this-convresult root-prefix (1+ i-time)))))))
	     (dolist (i-field acc-fields)
	       (setf (aref acc-array i-field)
		     (+ (aref acc-array i-field) (nth i-field (nth i (this-convresult root-prefix (1+ i-time))))))
	       (format out-stream "~a ~t" 
		       (format-plotfield (aref acc-array i-field)))))
	   (format out-stream "~%"))))))
                  


(defun exp-countcase-training (domain dom-file num-times &key (timeout 5)
          (root-prefix *expconv-rootprefix*))
         
  (setf *say-output* 1)
  (say-domain domain dom-file)
  (let ((out-file (format nil "~a/result/countcases_~a_~a.txt" *domain-dir* domain root-prefix)))
    (with-open-file (out-stream out-file :direction :output :if-exists
				:supersede :if-does-not-exist :create)  
      (format out-stream "Train-set  ")
;;       (read-type-case-base)
;;       (maphash #'(lambda (type cases-table) (format out-stream "~a " type)) *type-case-base*)
      (dotimes (i-cycle num-times)
	(delete-casebase)
	(let* ((train-prefix (format nil "~a~a" root-prefix (format-groupprob (1+ i-cycle))))
	       (total-cases 0))
	  (dotimes (i-trainset (1+ i-cycle))
	    (let ((i-train-prefix (format nil "~a~a" root-prefix (format-groupprob (1+ i-trainset)))))
	      (cbp-runset domain dom-file :store-case t  :timeout timeout :probs-prefix i-train-prefix)))
	  
	  (read-type-case-base)
	  (format out-stream "~%~a " train-prefix)
	  (maphash #'(lambda (type cases-table)
		       (declare (ignore type))
		       (format out-stream "~a " (hash-table-count cases-table))
		       (setf total-cases (+ total-cases (hash-table-count cases-table))))
		   *type-case-base*)
	  )))))


(defun exp-convergence (domain dom-file num-times runtype &key (timeout *this-timeout*)
			       (test-prefix *expconv-testprefix*)
			       (root-prefix *expconv-rootprefix*))
			       
  (setf *say-output* 1)
  (setf *experiment-results* (make-hash-table))
  (say-domain domain dom-file)

  (dotimes (i-cycle num-times)
    (delete-casebase)
    (let ((train-prefix (format nil "~a~a" root-prefix (format-groupprob (1+ i-cycle)))))
      (dotimes (i-trainset (1+ i-cycle))
	(let ((i-train-prefix (format nil "~a~a" root-prefix (format-groupprob (1+ i-trainset)))))
	  (cbp-runset domain dom-file :store-case t  :timeout timeout :probs-prefix i-train-prefix)))
      
      (cbp-runset domain dom-file :runtype runtype :timeout timeout :probs-prefix test-prefix
		  :out-file (conc-outfile (concatenate 'string "converg-" train-prefix) "dat") 
		  :use-case t :save-result t :experiment-var (intern (string-upcase train-prefix)))))
  (exp-sort-results)
  (convergence-problems-solved (conc-outfile "convergence_probsolved" "txt"))
  (convergence-data-toplot root-prefix num-times (conc-outfile "convergence_data" "txt" ))
  )

