;; SAYPHI Case-Based Reasoner
;; Running the Case-Based Reasoner
;; Td. 10.02.2006
;; _________________________________________________________________________________________

(defvar *cbp-verbose* t)

(defparameter *cbinfo-usedcases* nil)
(defparameter *cboption-traceselected* nil)

;; Since I use more structures in the new properties version, i decided to put all structures
;; together

;; Holds the typed-substate and the applied action
(defstruct (casestep (:print-function casestep-print))
  (properties nil)
  (prop-extras nil)
  (app-action nil)
  (h-diff nil)
  (utility (cons 0 0))
)

(defstruct (instep (:include casestep)
		   (:print-function instep-print))
  (inst-substate nil))

;;Hold the sequence of typed-substate. (this is the case unit)
(defstruct (typedcase (:print-function typedcase-print))
  (case-no nil)
  (merged nil)
  (utility (cons 0 0))
  (rx-key nil)
  (metric nil)
  (sequence nil)
)


;;For the repaly table.  
(defstruct (replay (:print-function replay-print))
  (object nil)
  (case-no nil)
  (object-type nil)
  (goal-object-p nil)
  (sequence nil)
  (typedcase nil)
  (no-op-counter 0)
  (case-pointer 0)
  (used-in-step nil)
  (visited-substate nil)
  (active-utility (cons 0 0))
  (current-utility (cons 0 0))
  (rx-key)
)


(defun makenew-step (properties extras app-action &optional (h-diff 0))
  (make-casestep :properties properties
		 :prop-extras extras
		 :app-action app-action
		 :h-diff h-diff))


(defun list-to-casestep (liststep)
  (makenew-step (first liststep) (second liststep) (third liststep)
		(fourth liststep)))


(defun list-to-typedcase(listcase)
  (make-typedcase :case-no (get-vals 'case-no listcase)
		  :merged (get-vals 'merged listcase)
		  :utility (get-vals 'utility listcase)
		  :sequence (mapcar #'list-to-casestep (get-vals 'sequence listcase))
		  :rx-key (get-vals 'rx-key listcase)
		  :metric (get-vals 'metric listcase)
		  ))


(defun casestep-to-list (casestep)
  (list (casestep-properties casestep)
	(casestep-prop-extras casestep)
	(casestep-app-action casestep)
	(casestep-h-diff casestep)
	(casestep-utility casestep)))


(defun typedcase-to-list(typed-case)
  (list (list 'case-no (typedcase-case-no typed-case)) 
	(list 'merged (typedcase-merged typed-case))
	(list 'utility (typedcase-utility typed-case))
	(list 'rx-key (typedcase-rx-key typed-case))
	(list 'metric (typedcase-metric typed-case))
	(list 'sequence (mapcar #'casestep-to-list (typedcase-sequence typed-case)))
	))


(defun to-instep (casestep)
  (make-instep :properties (casestep-properties casestep)
	       :prop-extras (casestep-prop-extras casestep)
	       :app-action (casestep-app-action casestep)
	       :h-diff (casestep-h-diff casestep)
	       :utility (casestep-utility casestep)
))

(defun casestep-print (casestep stream z)
  (declare (type casestep casestep)
	   (stream stream)
	   (ignore z))
  (format stream "~%#<STEP: ~a ~a ~a [~a][~a]>" 
	  (casestep-properties casestep)
	  (casestep-prop-extras casestep)
	  (casestep-app-action casestep)
	  (casestep-h-diff casestep)
	  (casestep-utility casestep)))

(defun typedcase-print (typedcase stream z)
  (declare (type typedcase typedcase)
	   (stream stream)
	   (ignore z))
  (format stream "~%#<TYPEDCASE No.:~a  Length:~a  Utility:~a  Merged ~a  Metric ~a >"
	  (typedcase-case-no typedcase)
	  (length (typedcase-sequence typedcase))
	  (typedcase-utility typedcase)
	  (typedcase-merged typedcase)
	  (typedcase-metric typedcase)
	  ))


(defun instep-print (instep stream z)
  (declare (type instep instep)
	   (stream stream)
	   (ignore z))
  (format stream "~%#<INSTEP: ~a ~a ~a [~a][~a] ~a>" 
	  (instep-properties instep)
	  (instep-prop-extras instep)
	  (instep-app-action instep)
	  (instep-h-diff instep)
	  (instep-utility instep)
	  (instep-inst-substate instep)))

	  
(defun replay-print (replay stream z)
  (declare (type replay replay)
	   (stream stream)
	   (ignore z))
  (format stream "~%#<REPLAY [Type: ~a, Case-no ~a, Goal-Obj-p:~a>" 
	  (replay-object-type replay) (replay-case-no replay) (replay-goal-object-p replay))
  (format stream "~%          No-op-count: ~a" (replay-no-op-counter replay))
  (format stream "~%          Pointer: ~a" (replay-case-pointer replay))
  (format stream "~%          Utility: ~a      Active-Utility: ~a" (replay-current-utility replay)(replay-active-utility replay))
;;   (format stream "~%  SEQUENCE:~% ~a>" (replay-sequence replay))
)
