;; SAYPHI Case-Based Reasoner
;; Learning Typed Sequence in a Model Problem for
;; Metric Optimization
;; Td. 13.11.2007
;; _________________________________________________________________________________________


;; I take random goals of the problem to learn typed sequences for
;; a problem of less goals

(defparameter *keep-problem-goals* nil)

(defun select-random-goals (goals)
  (let ((new-goals (copy-state goals)))
    (maphash #'(lambda (pred bitmap)
		 (dotimes (ibit (length bitmap) new-goals)
		   (when (= 1 (sbit bitmap ibit))
		     (when (= 0 (random 2))
		       (setf (sbit (gethash pred new-goals) ibit) 0)))))
	     goals)
	new-goals))

(defun count-goals (goals)
  (let ((n-goals 0))
    (maphash #'(lambda (pred bitmap)
		 (declare (ignore pred))
		 (setf n-goals (+ n-goals (count 1 bitmap))))
	     goals)
    n-goals))
	     

(defun select-n-random-goals (goals n-max)
  (do ((new-goals (select-random-goals goals) 
		  (select-random-goals new-goals))) 
      ((<= (count-goals new-goals) n-max) new-goals)
    (format t "~% => Selecting random goals...")))
    
(defun set-random-subproblem (max-goals)
  (let ((new-goals (select-n-random-goals *keep-problem-goals* max-goals)))
    (setf (problem-goals *current-problem*) new-goals)
    (setf (problem-lit-goals *current-problem*) (pp-state new-goals 'list))))

  


(defun learn-subproblem-sequences (times &key (max-goals 3)
				              (plan-arguments nil))
  (setf *keep-problem-goals* (copy-state (problem-goals *current-problem*)))
  (let ((solution nil))
    (dotimes (i-time times)
      (set-random-subproblem max-goals)
      (format t "~%~% Learning Model Problem: Subproblem ~a" i-time)
      (format t "~% Goals: ~a~%" (problem-lit-goals *current-problem*))
      (setf solution (apply #'plan plan-arguments))
      (when (and (solution-p solution)
		 (solution-found solution))
	(store-case))
      
      ))
  (setf (problem-goals *current-problem*) *keep-problem-goals*)
  (setf (problem-lit-goals *current-problem*) (pp-state *keep-problem-goals* 'list))
  )

  

    