;; SAYPHI Case-Based Reasoner
;; Merging Type Cases
;; Td. 10.02.2006
;; _________________________________________________________________________________________


(defun list-merged (cases)
  (dolist (i-case cases)
    (format t "~% Case ~a, Len: ~a" (typedcase-case-no i-case) (length (typedcase-sequence i-case)))))

;; If the past step is not a casestep-type the properties and the extras must be on a list
(defun make-noop (past-properties noop-count &optional (h-diff 0))
  (cond ((casestep-p past-properties)
	 (make-casestep :properties (casestep-properties past-properties)
			:prop-extras (casestep-prop-extras past-properties)
			:app-action (cons 'no-op noop-count)
			:h-diff h-diff
			))
	(t
	 (make-casestep :properties (car past-properties)
			 :prop-extras (second past-properties)
			 :app-action (cons 'no-op noop-count)
			 :h-diff h-diff
			 ))))


(defun match-cases (i-typeseq new-typeseq)
  (do* ((diff-case-found nil) (merged-seq nil)
	(i-pos 0 (incf i-pos)) 
	(i-step (first i-typeseq) (nth i-pos i-typeseq))
	(new-pos 0 (incf new-pos))
	(new-step (first new-typeseq) (nth new-pos new-typeseq)))
       ((or (null i-step) (null new-step)) 
	(if (and (not diff-case-found) (null i-step) (null new-step))
	    (reverse merged-seq) nil))
       (let ((no-op-int 0) (merged-case-found nil))
	 (unless diff-case-found
	   (cond ((equal-substate i-step new-step)
		  (cond ((equal-app-action i-step new-step) 
			 (cond ((noop-step-p new-step)
				(setf no-op-int (min (cdr (casestep-app-action new-step)) 
						     (cdr (casestep-app-action i-step))))
				(push (make-noop new-step no-op-int) merged-seq)
				(setf merged-case-found t))
			       (t ;;are equal and normal operators
				(push new-step merged-seq))))
			((noop-step-p new-step)
			 (setf i-pos (- i-pos 1)))
			((noop-step-p i-step)
			 (setf new-pos (- new-pos 1)))
			(t  ;;are equal sub-state but different actions
			 (setf diff-case-found t))))
		 ((noop-step-p new-step)
		  (setf i-pos (- i-pos 1)))
		 ((noop-step-p i-step)
		  (setf new-pos (- new-pos 1)))
		 (t ;;completely difference 
		  (setf diff-case-found t)))))))



;; This takes the new type-seq and the stored ones to merge cases 
;; if there is the same but the no-ops
(defun merge-cases (new-case stored-cases)
  (let ((new-cases-stored nil)
	(new-case-p t))
    (cond ((> (length (typedcase-sequence new-case)) 1)
	   (dolist (i-stored stored-cases)
	     (cond ((and new-case-p
			 (equal (typedcase-metric new-case) (typedcase-metric i-stored)))
		    (let ((match-seq (match-cases (typedcase-sequence i-stored)
						  (typedcase-sequence new-case))))
		      (cond ((not (null match-seq))
			     (push (make-typedcase :case-no (typedcase-case-no i-stored)
						   :merged t
						   :rx-key (typedcase-rx-key i-stored)
						   :metric (typedcase-metric i-stored)
						   :sequence match-seq)
				   new-cases-stored)
			     (setf new-case-p nil))
			    (t
			     (push i-stored new-cases-stored)))))
		   (t
		    (push i-stored new-cases-stored))))
	   (when new-case-p
	     (setf (typedcase-case-no new-case) (1+ (length stored-cases)))
	     (push new-case new-cases-stored))
	   new-cases-stored)
	  (t stored-cases))))
