;; SAYPHI Case-Based Reasoner
;; Building Type Proporties
;; Td. 30.05.2006
;; _________________________________________________________________________________________


(defvar *obs-relevant-preds* nil)
(defvar *objs-property-maps* nil)
(defvar *property-table* nil)

;; Structure to store property features
(defstruct (property-descriptor
	    (:conc-name ppty-))
  (property nil)
  (pred nil)
  (pos nil)
  (arg nil)
  (attribute nil)
  (exchange nil)
  (static nil))


(defmacro object-relativepos-map (object pred-args pos)
  `(permut-bitmap (pred-relevant-bitmaps ,object ,pred-args ,pos 
					 (problem-inheritobjects *current-problem*))))

(defmacro to-property (pred pos)
 `(intern (format nil "~a~a" ,pred (1+ ,pos))))


(defun objstate-properties (object state)
  (let ((obj-maps (gethash object *objs-property-maps*))
	(obj-properties nil) (prop-extras nil))
    (when (hash-table-p obj-maps)
      (maphash #'(lambda (property pred-map)
		   (let* ((state-map (gethash (car pred-map) state))
			  (res-map nil) (num-props 0))
		     (when (simple-bit-vector-p state-map)
		       (setf res-map (bit-and state-map (cdr pred-map)))
		       (setf num-props (count 1 res-map))
		       (when (> num-props 0)
			 (push property obj-properties)
			 (push num-props prop-extras)))))
	       obj-maps))
    (list obj-properties prop-extras)))


(defun all-objs-properties(state)
  (let ((objects (problem-objects *current-problem*))
	(list-prop nil))
    (maphash #'(lambda (type objects)
		 (declare (ignore type))
		 (dolist (object objects)
		   (push (cons object (objstate-properties object state))
			 list-prop)))
	     objects)
    list-prop))



;; This is like objstate-properties but in this case I return the literasl of the object 
;; sub-state rather than the object properties
(defun objstate-literals (object state)
  (let ((obj-maps (gethash object *objs-property-maps*))
	(substate-literals nil))
    (maphash #'(lambda (property pred-map)
		 (declare (ignore property))
		 (let* ((state-map (gethash (car pred-map) state))
			(res-bitmap nil))
		   (when (simple-bit-vector-p state-map)
		     (setf res-bitmap (bit-and state-map (cdr pred-map)))
		     (dotimes (ibit (length res-bitmap))
		       (when (= 1 (sbit res-bitmap ibit))
			 (push (cons (car pred-map) ibit) substate-literals))))))
	     obj-maps)
    (remove-duplicates substate-literals :test #'equal)))


(defun create-pred-properties (pred args)
  (let ((pred-properties nil))
    (dotimes (i-arg (length args) pred-properties)
      (let ((property (intern (format nil "~a~d" pred (1+ i-arg)))))
	(push (cons property
		    (make-property-descriptor 
		     :property property
		     :pred pred
		     :arg (nth i-arg args)
		     :pos i-arg))
		    pred-properties)))))
			       
		   
(defun build-property-table ()
  (let ((property-table (make-hash-table :test #'eq)))
    (maphash #'(lambda (pred args)
		 (dolist (propdesc (create-pred-properties pred args))
		   (setf (gethash (car propdesc) property-table)
			 (cdr propdesc))))
	     (dom-predicates *pspace*))
    property-table)) 




(defun all-property-list (property-table)
  (let ((property-list nil))
    (maphash #'(lambda (property descriptor)
		 (declare (ignore descriptor))
		 (push property property-list))
	     property-table)
    property-list))


(defun get-object-relevant-preds (object)
  (let ((objects (problem-inheritobjects *current-problem*))
	(relevant-preds nil))
    (maphash #'(lambda (pred args)
		 (dolist (iarg args)
		   (when (and (find object (gethash (cdr iarg) objects))
			      (not (find pred relevant-preds)))
		     (push pred relevant-preds))))
	     (dom-predicates *pspace*))
    relevant-preds))


;; To save time a create a hash table with de relevant predicates for each object. Tbis is used
;; when computing the object sub-state
(defun build-objects-relevant-preds ()
  (let ((ob-relevant-preds (make-hash-table :test #'eq)))
    (maphash #'(lambda (type objects)
		 (declare (ignore type))
		 (dolist (iobject objects)
		   (setf (gethash iobject ob-relevant-preds)
			 (get-object-relevant-preds iobject))))
	     (problem-inheritobjects *current-problem*))
    ob-relevant-preds))


(defun is-prop-relevant (object prop-descriptor)
  (let* ((pred-arg (nth (ppty-pos prop-descriptor)
			(gethash (ppty-pred prop-descriptor) (dom-predicates *pspace*)))))
    (not (null (find object (gethash (cdr pred-arg) 
				     (problem-inheritobjects *current-problem*)))))))

(defun compound-properties (properties-extras)
  (let ((property-list (first properties-extras))
	(extras-list (second properties-extras)))
    (mapcar #'cons property-list extras-list)))
    


(defun property-object-statemap (object prop-descriptor)
  (let* ((pred (ppty-pred prop-descriptor))
	 (pred-args (gethash pred (dom-predicates *pspace*))))
     (cons pred (object-relativepos-map object pred-args (ppty-pos prop-descriptor)))))
	 

(defun get-object-property-maps (object property-table)
  (let ((ob-property-maps (make-hash-table :test #'eq)))
    (maphash #'(lambda (property descriptor)
		 (when (is-prop-relevant object descriptor)
		   (setf (gethash property ob-property-maps)
			 (property-object-statemap object descriptor))))
	     property-table)
    ob-property-maps))


(defun build-objects-property-maps (property-table)
  (let* ((objs-property-maps (make-hash-table :test #'eq)))
    (maphash #'(lambda (type objects)
		 (declare (ignore type))
		 (dolist (iobject objects)
		   (setf (gethash iobject objs-property-maps)
			 (get-object-property-maps iobject property-table))))
	     (problem-inheritobjects *current-problem*))
    objs-property-maps))


(defun init-properties()
  (setf *property-table* (build-property-table))
  (setf *obs-relevant-preds* (build-objects-relevant-preds))
  (setf *objs-property-maps* (build-objects-property-maps *property-table*))
)


