;; SAYPHI Case-Based Reasoner
;; Running the Case-Based Reasoner
;; Td. 10.02.2006
;; _________________________________________________________________________________________

(defvar *cbp-runtype* 'mixed)
(defvar *cbp-retrieve-option* 'default)
(defvar *cbp-trace* nil)

;;tracing CBPlanning
(defvar *cbp-guided-nodes* 0)
(defvar *cbp-hit-node* nil)
(defvar *cbp-hit-average* 0)

(defun reset-trace()
  (setf *cbp-guided-nodes* 0
	*cbp-hit-node* nil
	*cbp-hit-average* 0))
;;------------------------------------
(defun reset-usedstep ()
  (maphash #'(lambda (object obreplay)
	       (declare (ignore object))
	       (setf (replay-used-in-step obreplay) nil))
	   *replay-table*))



;; This function is common to all replaying algorithm, and should settle 
;; the instantiated replay table
(defun init-replay (i-node h-fn select-mode &key (sort-option nil) (retrieve-option nil))
  (setf *get-achieved-goals* t)
  (say-consed-bytes :mem-per-node (store-h-extras i-node h-fn))
  (setf *get-achieved-goals* nil)

  (cond ((eq select-mode 'allob)
	 (setf *replay-table* (select-cases-toreplay :replay-allobjs t :sort-option sort-option
						     :retrieve-option retrieve-option)))
	((eq select-mode 'instcase)
	 (setf *replay-table* (select-instcases-toreplay :replay-allobjs t :sort-option sort-option
							 :retrieve-option retrieve-option)))
	(t 
	 (setf *replay-table* (select-cases-toreplay :sort-option sort-option
						     :retrieve-option retrieve-option))))
  (instantiate-replay-table *h-achieved-goals* *replay-table*))


;; To order the open nodes by h-value and advise-weight
(defun lessthan-heuristic+advise (node-a node-b)
  (cond ((< (snode-h-value node-a)(snode-h-value node-b)) t)
	((= (snode-h-value node-a)(snode-h-value node-b)) 
	 (if (> (snode-advise-weight node-a) (snode-advise-weight node-b))
	     t nil))
	(t nil)))
	 

(defun update-active-utility (object improved-p)
  (let ((obreplay (object-replay object)))
    (incf (cdr (replay-active-utility obreplay)))
    (when improved-p
      (incf (car (replay-active-utility obreplay))))))


(defun advance-noop-type-case (object)
  (let* ((obreplay (object-replay object))
	 (case-step (get-case-step object)))
    (cond ((noop-step-p case-step)
	   (cond ((= (replay-no-op-counter obreplay) (cdr (instep-app-action case-step)))
		  (advance-type-case object))
		 (t 
		  (incf (replay-no-op-counter obreplay))))))))


(defun map-replay-extras (node)
  (dolist (i-child (snode-children node))
    (when (eq (snode-plist i-child) :candidate-for-replay)
      (setf (snode-h-value i-child) 
	    (floor (snode-h-value i-child)))
      (format t "~% Setting candiate for replay")
)))


(defun replay-table-list(sort-type)
  (let ((replay-list nil))
    (maphash #'(lambda (ob replay)
		 (declare (ignore ob))
		 (push replay replay-list))
	     *replay-table*)
    (cond ((eq sort-type 'seq-len) replay-list)
	  ((or (eq sort-type 'utility)
	       (eq sort-type 'step-utility))
	   (setf replay-list (stable-sort replay-list #'greater-current-step-utility)))
	  ((eq sort-type 'active-utility)
	   (setf replay-list (stable-sort replay-list #'greater-active-utility)))
	  (t 
	   replay-list))))


(defun sort-replay-table ()
  (let ((sorted-objects nil))
    (maphash #'(lambda (object obreplay)
		 (push (cons (- (length (replay-sequence obreplay))
				(replay-case-pointer obreplay))
			     object) sorted-objects))
	    *replay-table*)
    (stable-sort sorted-objects #'> :key #'car)))


(defun litsubstate-equal (current-substate visited-substate)
  (let ((equal-to-visited t))
    (dolist (i-literal current-substate equal-to-visited)
      (when equal-to-visited
	(unless (find i-literal visited-substate :test #'equal) 
	  (setf equal-to-visited nil))))))



;; Given an NODE, it looks up if any cases in REPLSAY-TABLE is relevant to replay.  
;; It returns the OBJECT of the corresponding case
(defun relevant-to-replay (node)
  (let* ((app-action (snode-plan-action node))
	 (relevant-obs (rest (snode-plan-action node))) 
	 (match-replay nil) (relevant-object nil))
    (dolist (iob relevant-obs relevant-object)
      (unless (or match-replay (null (object-replay iob)))
	(let ((case-step (get-case-step iob)))
;;  	  (format t "~%######-->> Matching Object ~a ~a" iob case-step)
	  (when (and (instep-p case-step)
		     (equal-stepaction (casestep-app-action case-step) app-action))
	    (let ((node-objsubstate nil) (node-objproperties nil))
;;  	      (format t "~%######-->> Equal actions ~a" (casestep-app-action case-step))
	      (cond ((not (null (instep-inst-substate case-step)))
		     (setf node-objsubstate (objstate-literals iob (snode-state node)))
		     ;; The instantiated seq is checked to be a subset of the substate. 
		     ;; It is done this way rather than the opposite because the seq are
		     ;; partially instantiated by the goal membership
		     (when (stepliterals-memberset node-objsubstate (instep-inst-substate case-step))
		       (setf relevant-object iob)
		       (setf match-replay t)))
		    (t ;; properties compared
		     (setf node-objproperties (car (objstate-properties iob (snode-state node))))
		     (when (stepcase-memberset (instep-properties case-step) node-objproperties)
		       (setf relevant-object iob)
		       (setf match-replay t))))
	      )))))))


;; Search in child-nodes until a relevant-to-replay node is found
(defun select-candidate-node (node h-function current-h visited-nodes 
				   &key (discrete *say-h-rounded*)
				        (replay-mode 'normal))
  (declare (ignore discrete))
  (let ((better-found nil) (case-object nil) (current-advise-weight (length (snode-children node)))
	(hit-counter 0))
    (dolist (i-child (snode-children node) better-found)
      (when (null better-found)
	(cond ((not (find i-child visited-nodes :test #'equal-state))
	       (setf case-object (relevant-to-replay i-child))
;; 	       (format t "~% __________## Advise for ~a " case-object)
	       (unless (null case-object)
		 (incf hit-counter)
		 (when *cbp-store-utility*
		   (save-step-advise case-object i-child))
		 (store-h-extras i-child h-function)
		 (cond ((< (snode-h-value i-child) current-h)
			(when *cbp-trace*
			  (push hit-counter *cbp-hit-node*))
			(setf better-found i-child)
			(markused-case case-object)
			(advance-type-case case-object))

		       ((and (eq replay-mode 'h-diff)
			     (>= (- current-h (snode-h-value i-child))
				 (casestep-h-diff (get-case-step case-object))))
			(when (> *say-output* 1) 
			  (format t "~% --*--> H-Diff Guided"))
			(when *cbp-trace*
			  (push hit-counter *cbp-hit-node*))
			(setf better-found i-child)
			(markused-case case-object)
			(advance-type-case case-object))

		       (t
			(setf (snode-advise-weight i-child) current-advise-weight)
			(decf current-advise-weight)

			(when (> *say-output* 1) 
			  (format t "~% --*--> Advise Not Acepted [~a]{~a}" 
				  case-object (snode-plan-action i-child))))
		       )))
	      (t
	       (setf (snode-closed i-child) t)))
	))))
  
;; ;; Like get-enforced-heuristic from the standard EHC. Just skipping already evaluated nodes
(defun select-nonreplay-node  (node h-function current-h &key (discrete *say-h-rounded*))
  (declare (ignore discrete))
  (let ((better-found nil))
    (dolist (i-child (snode-children node) better-found)
      (when (null better-found)
	(when (null (snode-h-value i-child))
	  (store-h-extras i-child h-function)
	  (when (< (snode-h-value i-child) current-h)
	    (setf better-found i-child)))))))


(defun refocus-replay (next-node)
  (let* ((app-action (snode-plan-action next-node))
	 (action-objects (cdr app-action)))
    (maphash #'(lambda(object obreplay)
		 (unless (replay-used-in-step obreplay)
		   (let ((current-step (get-case-step object))
			 (step-ahead nil))
		     (cond ((and (find object action-objects)
				 (instep-p current-step))
			    (when (equal-stepaction (instep-app-action current-step) app-action)
			      (advance-type-case object)))
			   ((and (find object action-objects)
				 (noop-step-p current-step))
			    (setf step-ahead (get-case-step-ahead object))
			    (when (and (instep-p step-ahead)
				       (equal-stepaction (instep-app-action step-ahead) app-action))
			      (advance-type-case object)
			      (advance-type-case object)))
			   (t 
			    (advance-noop-type-case object))))))
		   *replay-table*)))


(defun find-node-toreplay (object node)
  (let* ((case-step (get-case-step object))
	 (find-node nil))
    (when (instep-p case-step)
;;       (format t "~%_______>> Case-Step ~a" case-step)
      (dolist (ichild (snode-children node) find-node)
	(when (null find-node)
;; 	  (format t "~%____________>> Looking for: ~a ~a" object (snode-plan-action ichild))
	  (when (and (null (snode-h-value ichild))
		     (not (snode-closed ichild)))
	    
	    (when (and (equal-stepaction (casestep-app-action case-step) (snode-plan-action ichild))
		       (find object (snode-plan-action ichild)))
;; 	      (format t "~%##--->Equal step action ~a " case-step)
	      (let ((node-objsubstate nil) (node-objproperties nil))
		(cond ((not (null (instep-inst-substate case-step)))
		       (setf node-objsubstate (objstate-literals object(snode-state ichild)))
		       (when (stepliterals-memberset node-objsubstate (instep-inst-substate case-step))
			 (setf find-node ichild)))
		      (t ;; properties compared
		       (setf node-objproperties (car (objstate-properties object (snode-state ichild))))
;; 		       (format t "~% @@@ ---> ~a" node-objproperties) 
		       (when (and (stepcase-memberset (instep-properties case-step) node-objproperties)
				  (not (find node-objproperties (replay-visited-substate (object-replay object))
					     :test #'litsubstate-equal)))
			 ;; 		       (push node-objproperties (replay-visited-substate (object-replay object)))
			 (setf find-node ichild))))
		))))))))



;; Search in child-nodes until a relevant-to-replay node is found
(defun select-advise-next-node (node h-function current-h visited-nodes  
				     &key (replay-sort-option nil) (replay-mode 'normal))
  (declare (ignore replay-mode))
  (let ((found-candidate nil) (selected-node nil)  
	(temp-found nil) (better-found nil)
	(current-advise-weight (length (snode-children node)))) 
    (dolist (i-obreplay (replay-table-list replay-sort-option) selected-node)
      (let* ((object (replay-object i-obreplay)))
	(unless found-candidate
	  (unless (null (setf temp-found (find-node-toreplay object node)))
	    (cond ((not (find temp-found visited-nodes :test #'equal-state))
		   (when *cbp-store-utility*
		     (save-step-advise object temp-found))
		   (store-h-extras temp-found h-function)
		   (cond ((< (snode-h-value temp-found) current-h)
			  (setf better-found temp-found)
			  (setf found-candidate t)
			  (markused-case object)
			  (update-active-utility object t)
			  (advance-type-case object)
			  (when (> *say-output* 1) 
			    (format t "~% --*--> Advise Acepted [~a]{~a}" object (snode-plan-action better-found))))	    
			 (t
			  (setf (snode-advise-weight temp-found) current-advise-weight)
			  (decf current-advise-weight)
			  (update-active-utility object nil)
			  
			  (when (> *say-output* 1) 
			    (format t "~% --*--> Advise Not Acepted [~a]{~a}(~a)" object (snode-plan-action temp-found)
				    (snode-number temp-found))))))
		  (t
		   (setf (snode-closed temp-found) t)))))))
    (cond ((snode-p better-found) better-found)
	  (t 
	   (dolist (i-child (snode-children node) better-found)
	     (when (null better-found)
	       (when (and (null (snode-h-value i-child))
			  (not (snode-closed i-child)))
		 (cond ((not (find i-child visited-nodes :test #'equal-state))
			(store-h-extras i-child h-function)
;; 			(format t "~% --***--> Evaluating  ~a" (snode-number i-child))
			(cond ((< (snode-h-value i-child) current-h)
			       (setf better-found i-child)
			       (when (> *say-output* 1) 
				 (format t "~% --*--> Normal Selection {~a}" (snode-plan-action i-child))))
			      (t 
			       (setf (snode-advise-weight i-child) 0))))
		       (t
			(setf (snode-closed i-child) t))))))))))

	   


(defun cbp-candidate-nodes (node replay-table objects)
  (let ((candidate-nodes nil))
    (dolist (ichild (snode-children node) candidate-nodes)
      (cond ((relevant-to-replay ichild replay-table objects)
	     (push ichild candidate-nodes)
	     (format t "~%>> OK ~a" (snode-plan-action ichild)))
	    (t
	     (format t "~%>> NOT RELEVANT ~a" (snode-plan-action ichild))))
      )))


(defun cbp-breadth-open-nodes (open-children visited-nodes)
  (let ((current-h most-positive-fixnum)
	(open-nodes nil))
    (stable-sort open-children #'< :key #'snode-h-value)
    (setf current-h (snode-h-value (car open-children)))
  ;  (format t "~% Open Nodes h: ~a" current-h)
    (dolist (ichild open-children open-nodes)
      (unless (or (find ichild open-nodes :test #'equal-state)
		  (find ichild visited-nodes :test #'equal-state))
	;;Comento esta linea porque en FF se expanden todos los nodos sin importar si es mayor
;; 	la heuristica
	;;(when (= (snode-h-value ichild) current-h)
	  (push ichild open-nodes)))))




 
(defun supported-ehc (i-node h-fn cost-fn search-options)
  (declare (ignore cost-fn))
  (let ((problem *current-problem*) 
	(node i-node) (visited nil) (next-node nil) (current-h nil) (open-children nil) (open-nodes nil)
	(select-mode (search-option-value :select-mode search-options))
	(replay-mode (search-option-value :replay-mode search-options)))
    (init-replay i-node h-fn select-mode)
    (setf current-h (snode-h-value node))
    (print-search-node node nil)
    
    (do* ((stop-this (stop-search node problem) (stop-search node problem)))
	 (stop-this (when-stop (cdr stop-this) node problem))
	 
	 (expand-state node :helpful t)
	 (push node visited)
	 (reset-usedstep)
	 (setf next-node (select-candidate-node node h-fn current-h visited 
						:replay-mode replay-mode))
	 (cond ((snode-p next-node)
		(setf current-h (snode-h-value next-node))
		(setf node next-node)
		(setf open-nodes nil)
		(setf open-children nil)
		(when *cbp-trace*
		  (incf *cbp-guided-nodes*))
		(when (> *say-output* 1)
		  (print-search-node node nil)
		  (format t "-> Guided")))
	       (t 
		(setf next-node (select-nonreplay-node node h-fn current-h))
		(cond ((snode-p next-node)
		       (setf current-h (snode-h-value next-node))
		       (setf node next-node)
		       (setf open-nodes nil)
		       (setf open-children nil)
		       (print-search-node node nil))
		      (t 
		       (setf open-children (append (remove-if #'snode-closed (snode-children node)) 
					  open-children))
		       (cond ((not (snode-p (setf next-node (pop open-nodes))))
			      (when (> *say-output* 1) (format t "  ~% [Expanding Breadth Level]"))
			      (setf open-nodes (stable-sort open-children #'lessthan-heuristic+advise))
;; 			      (setf open-nodes (stable-sort open-children #'< :key #'snode-h-value))
			      (when (null open-nodes) ;;Recuperando la poda por las helpful actions
				(setf open-nodes (restore-nonhelpful node h-fn visited))
				(when (> *say-output* 1) 
				  (format t "  ~%!!Restoring NON Helpful Nodes.")))		      
			      (setf node (pop open-nodes))
			      (print-search-node node nil)
			      (setf open-children nil))
			     (t 
			      (setf node next-node)
			      (print-search-node node nil)
			      ))))))       
	 (when (snode-p node)
	   (refocus-replay node))
	 )))




(defun guided-ehc (i-node h-fn cost-fn search-options)
  (declare (ignore cost-fn))
  (let ((select-mode (search-option-value :select-mode search-options))
	(replay-mode (search-option-value :replay-mode search-options))
	(replay-sort-option (search-option-value :replay-sort-option search-options))
	(problem *current-problem*) 
	(node i-node) (visited nil) (next-node nil) (current-h nil) (open-children nil) (open-nodes nil))
    (init-replay i-node h-fn select-mode :sort-option replay-sort-option)
    (setf current-h (snode-h-value node))
    (print-search-node node nil)

    (do* ((stop-this (stop-search node problem) (stop-search node problem)))
	 (stop-this (when-stop (cdr stop-this) node problem))
	 
	 (expand-state node :helpful t)
	 (push node visited)
	 (reset-usedstep)
	 (setf next-node (select-advise-next-node node h-fn current-h visited 
						  :replay-mode replay-mode
						  :replay-sort-option replay-sort-option))
	 (cond ((snode-p next-node)
		(setf current-h (snode-h-value next-node))
		(setf node next-node)
		(setf open-nodes nil)
		(setf open-children nil)
		(print-search-node node nil))
	       (t 
		(setf open-children (append (remove-if #'snode-closed (snode-children node)) 
					  open-children))
		(cond ((not (snode-p (setf next-node (pop open-nodes))))
		       (setf open-nodes (stable-sort open-children #'lessthan-heuristic+advise))
		       (when (null open-nodes) ;;Recuperando la poda por las helpful actions
			 (setf open-nodes (restore-nonhelpful node h-fn visited))
			 (when (> *say-output* 1) 
			   (format t "  ~%!!Restoring NON Helpful Nodes.")))		      
		       (setf node (pop open-nodes))
		       (print-search-node node nil)
		       (setf open-children nil))
		      (t 
		       (setf node next-node)
		       (print-search-node node nil)
		       ))))
	 (when (snode-p node)
	   (refocus-replay node))
	 )))


(defun advance-case-node-pointers (knode object)
  (let ((hpointers-table (snode-replay-pointers knode)))
    (maphash #'(lambda (key pointer)
		 (cond ((eq object key)
			(incf (gethash key hpointers-table)))
		       (t
			(let ((current-step (get-nth-case-step key pointer)))
			  (cond ((or (and (casestep-p current-step)
					  (equal-stepaction (casestep-app-action current-step) (snode-plan-action knode))
					  (find key (cdr (snode-plan-action knode)))
;; 					  (instep-p current-step)
					  )
				     (noop-step-p current-step))
;; 				 (format t "~% >>> Case Pointers Advance Key ~a in Node ~a" key knode)
;; 				 (format t "~% >>>> current Step ~a" current-step) 
				 (incf (gethash key hpointers-table))))))))
		 hpointers-table)))
    

(defun copy-pointers-table (node)
  (when (hash-table-p (snode-replay-pointers node))
    (let ((hpointers-table (make-hash-table :test #'eq)))
      (maphash #'(lambda (key replay)
		   (setf (gethash key hpointers-table) replay))
	       (snode-replay-pointers node))
    hpointers-table)))
  
  
(defun create-void-pointers (node)
  (let ((hpointers-table (make-hash-table :test #'eq)))
    (maphash #'(lambda (key replay)
		 (declare (ignore replay))
		 (setf (gethash key hpointers-table) 0))
	     *replay-table*)
    (setf (snode-replay-pointers node) hpointers-table)))

	       
(defun kcbr-recommend (knode sort-option)
  (let ((is-recommended nil))
;;     (format t "~% En recommended No. ~a" (snode-number (snode-parent knode)))
;;     (pp-hash-table (snode-replay-pointers (snode-parent knode)))
    (setf (snode-replay-pointers knode)  (copy-pointers-table (snode-parent knode)))
    (dolist (i-obreplay (replay-table-list sort-option) is-recommended)
      (let* ((object (replay-object i-obreplay)))
	(unless is-recommended
	  (let* ((case-step (get-nth-case-step object (node-objcase-pointer (snode-parent knode) object))))
	    (when (instep-p case-step)
;;   	      (format t "~% ==>> Comparing for ~a" object) 
;;  	      (format t "~% ==>>>> Step ~a @@ ~a" case-step (snode-plan-action knode)) 
	      (when (and (equal-stepaction (casestep-app-action case-step) (snode-plan-action knode))
			 (find object (snode-plan-action knode)))
;;  		(format t "~% ==>>>>>> Action OK.")  
		(let ((node-objsubstate nil) (node-objproperties nil))
		(cond ((not (null (instep-inst-substate case-step)))
;;  		       (format t "~% ==>>>>>>> Comp. Obj Substate")  
		       (setf node-objsubstate (objstate-literals object (snode-state knode)))
		       (when (stepliterals-memberset node-objsubstate (instep-inst-substate case-step))
;;   			 (format t "~% ==>>>>>>> Step OK.")  
			 (setf is-recommended t))
		         (advance-case-node-pointers knode object))
		      (t ;; properties compared
;;  		       (format t "~% ==>>>>>>> Comp. Prop substate")  
		       (setf node-objproperties (car (objstate-properties object (snode-state knode))))
		       
		       (when (and (stepcase-memberset (instep-properties case-step) node-objproperties)
				  (not (find node-objproperties (replay-visited-substate (object-replay object))
					     :test #'litsubstate-equal)))
;;   			 (format t "~% ==>>>>>>> Step OK.")  
			 (setf is-recommended t)
			 (advance-case-node-pointers knode object)
			 )))	      
		)))))))))
		   
;; It return the number of parameter object that fullfills the recommendation
(defun match-recommend (knode)
  (let ((match-recommended 0))
    (setf (snode-replay-pointers knode)  (copy-pointers-table (snode-parent knode)))
    (dolist (i-obreplay (replay-table-list nil) match-recommended)
      (let* ((object (replay-object i-obreplay)))
	(let* ((case-step (get-nth-case-step object (node-objcase-pointer (snode-parent knode) object))))
	  (when (instep-p case-step)
	    (when (and (equal-stepaction (casestep-app-action case-step) (snode-plan-action knode))
		       (find object (snode-plan-action knode)))
	      (let ((node-objsubstate nil) (node-objproperties nil))
		(cond ((not (null (instep-inst-substate case-step)))
		       (setf node-objsubstate (objstate-literals object (snode-state knode)))
		       (when (stepliterals-memberset node-objsubstate (instep-inst-substate case-step))
			 (incf match-recommended)
		         (advance-case-node-pointers knode object)))
		      (t ;; properties compared
		       (setf node-objproperties (car (objstate-properties object (snode-state knode))))
		       (when (and (stepcase-memberset (instep-properties case-step) node-objproperties)
				  (not (find node-objproperties (replay-visited-substate (object-replay object))
					     :test #'litsubstate-equal)))
			 (incf match-recommended)
			 (advance-case-node-pointers knode object)
			 )))	      
		))))))))


    


(defun guided-beamsearch (init-node h-fn cost-fn search-options)
  (let ((retrieve-option (search-option-value :retrieve-option search-options))
	(select-mode (search-option-value :select-mode search-options))
;; 	(replay-mode (search-option-value :replay-mode search-options))
	(replay-sort-option (search-option-value :replay-sort-option search-options))
	(open-nodes nil) (node init-node) (repeated nil) 
	(visited (make-hash-table)) (open-hash (make-hash-table)) (current-h 0)
	(problem *current-problem*) (krecommended nil)
	(delayed-nodes nil) (level-kselected nil))

    (init-replay init-node h-fn select-mode :sort-option replay-sort-option 
		                            :retrieve-option retrieve-option)
    (create-void-pointers init-node)
;;     (pp-hash-table (snode-replay-pointers init-node))

    (when *trace-mem* (write-node-mem-info 'beam-bfs))
    
    (store-a-star-extras node cost-fn)
    (setf current-h (snode-h-value node))
    (do* ((stop-this (stop-search node problem) (stop-search node problem)))
	 (stop-this (when-stop (cdr stop-this) node problem))

      (setf repeated nil)
      (setf level-kselected nil)
      (setf delayed-nodes nil)

      (push node (gethash (snode-hash-code node) visited))

      (trace-search-extras (snode-number node))
      (expand-state node :helpful nil)

      (dolist (i-child (snode-children node))
	(setf repeated (find i-child (gethash (snode-hash-code i-child) visited) :test #'equal-state))
       	(cond ((snode-p repeated)
	       (cond ((better-g-path repeated i-child cost-fn)
		      (relink-g-parent i-child repeated cost-fn))
		     (t 
		      (setf (snode-closed i-child) t)))))
	
	(setf repeated (find i-child (gethash (snode-hash-code i-child) open-hash) :test #'equal-state))
	(cond ((snode-p repeated)
	       (cond ((better-g-path repeated i-child cost-fn)
		      (setf (snode-closed repeated) t))
		     (t 
		      (setf (snode-closed i-child) t)))))
       
	(unless (snode-closed i-child)
	  (store-h-extras i-child h-fn)
	  (store-a-star-extras i-child cost-fn)
	  (setf krecommended (kcbr-recommend i-child replay-sort-option))
	  (cond (krecommended
		 (push i-child open-nodes)
		 (push i-child (gethash (snode-hash-code i-child) open-hash))
		 (setf level-kselected t))
		(t
		 (push i-child delayed-nodes)))))

      (unless level-kselected
	(dolist (i-delayed delayed-nodes)
	  (push i-delayed open-nodes)
	  (push i-delayed (gethash (snode-hash-code i-delayed) open-hash))
	  ))
      
      (setf open-nodes (stable-sort (remove-if #'snode-closed open-nodes)
				    #'< :key #'snode-f-value)) 
;;       (pp-node-list open-nodes)
      (setf node (pop open-nodes))
      
      (when (snode-p node)
	(print-search-node node open-nodes))
      )))



(defun guided-bfs (init-node h-fn cost-fn search-options)
  (let ((retrieve-option (search-option-value :retrieve-option search-options))
	(select-mode (search-option-value :select-mode search-options))
;; 	(replay-mode (search-option-value :replay-mode search-options))
 	(replay-sort-option (search-option-value :replay-sort-option search-options))
	(open-nodes nil) (node init-node) (repeated nil) 
	(visited (make-hash-table)) (open-hash (make-hash-table)) (current-h 0)
	(problem *current-problem*) (krecommended nil)
	(delayed-nodes nil))

    (init-replay init-node h-fn select-mode :sort-option replay-sort-option
		                            :retrieve-option retrieve-option)
    (create-void-pointers init-node)
;;     (pp-hash-table (snode-replay-pointers init-node))

    (when *trace-mem* (write-node-mem-info 'beam-bfs))    
    
    (store-a-star-extras node cost-fn)
    (setf current-h (snode-h-value node))
    (do* ((stop-this (stop-search node problem) (stop-search node problem)))
	 (stop-this (when-stop (cdr stop-this) node problem))

      (setf repeated nil)
      
      (push node (gethash (snode-hash-code node) visited))

      (trace-search-extras (snode-number node))
      (expand-state node :helpful nil)

      (dolist (i-child (snode-children node))
	(setf repeated (find i-child (gethash (snode-hash-code i-child) visited) :test #'equal-state))
       	(cond ((snode-p repeated)
	       (cond ((better-g-path repeated i-child cost-fn)
		      (relink-g-parent i-child repeated cost-fn))
		     (t 
		      (setf (snode-closed i-child) t)))))
	
	(setf repeated (find i-child (gethash (snode-hash-code i-child) open-hash) :test #'equal-state))
	(cond ((snode-p repeated)
	       (cond ((better-g-path repeated i-child cost-fn)
		      (setf (snode-closed repeated) t))
		     (t 
		      (setf (snode-closed i-child) t)))))
       
	(unless (snode-closed i-child)
	  (store-h-extras i-child h-fn)
	  (store-a-star-extras i-child cost-fn)
	  (setf krecommended (kcbr-recommend i-child replay-sort-option))
	  (cond (krecommended
		 (push i-child open-nodes)
		 (push i-child (gethash (snode-hash-code i-child) open-hash)))
		(t
		 (push i-child delayed-nodes)))))
      
;;       (format t "~% Open:")
;;       (pp-node-list open-nodes)
;;       (format t "~% Delayed:")
;;       (pp-node-list delayed-nodes)

      (cond ((null open-nodes)
	     (setf delayed-nodes (stable-sort (remove-if #'snode-closed delayed-nodes) #'< :key #'snode-f-value))
	     (setf node (pop delayed-nodes)))
	    (t 
	     (setf open-nodes (stable-sort (remove-if #'snode-closed open-nodes) #'< :key #'snode-f-value)) 
	     (setf node (pop open-nodes))))
      
      (when (snode-p node)
	(print-search-node node open-nodes))
      )))



(defun guided-bnb (init-node h-function cost-fn search-options)
  (when (> *say-output* 0) (format t "~% Finding EHC solution for Upper-cost Bound"))
  (let ((first-sol (guided-ehc init-node h-function nil '((:select-mode 'allob))))
	(problem *current-problem*)
	(visited (make-hash-table)) (siblings nil)
	(hash-solutions (make-hash-table))
	(node init-node) (repeated nil)
	(upper-cost-bound nil)
	(max-solutions (search-option-value :max-solutions search-options))
	(helpful (search-option-value :say-helpful search-options)))
    
    
    (setf *hash-inconsistent-h* (make-hash-table))
    (setf *say-hash-solutions* hash-solutions)
    (cond ((and (solution-p first-sol)
		(solution-found first-sol) (> (solution-depth first-sol) 0))
	   (setf upper-cost-bound (solution-total-cost first-sol))
	   (setf (snode-closed (solution-last-node first-sol)) t)
	   (setf (gethash 1 hash-solutions) first-sol)

	   (store-a-star-extras node cost-fn)
	   (push node (gethash (snode-hash-code node) visited))
	   (when (> *say-output* 0) (format t "~% Running Branch & Bound [~a]" upper-cost-bound))
	   
	   (do* ((stop-this (stop-multiple node problem hash-solutions max-solutions) 
			    (stop-multiple node problem hash-solutions max-solutions)))
		((car stop-this) (when-multiple-stop (cdr stop-this) node problem)) 
	     (when (snode-closed node)  ;;Goals reached
	       (when (< (solution-total-cost (gethash (hash-table-count hash-solutions) hash-solutions)) upper-cost-bound)
		 (setf *say-solution* (gethash (hash-table-count hash-solutions) hash-solutions))
		 (setf upper-cost-bound (solution-total-cost *say-solution*))
		 (when (> *say-output* 0) 
		   (format t "~% => BnB >> Found new solution with Cost {~a}" (solution-total-cost *say-solution*)))
		 )
	       (setf node (snode-parent node)))

	     (unless (snode-expanded node) 
	       (expand-state node :helpful helpful))
	     (dolist (i-child (snode-children node))
	       (unless (snode-closed i-child)
		 (cond ((null (snode-h-value i-child))
			(setf repeated (find i-child (gethash (snode-hash-code i-child) visited) :test #'equal-state))
			(cond ((or (not (snode-p repeated))
				   (and (snode-p repeated)
					(not (better-g-path repeated i-child cost-fn))))
			       (store-h-extras i-child h-function)
			       
			       (push i-child (gethash (snode-hash-code i-child) visited)))
			      (t
			       (setf (snode-closed i-child) t)
			       (when (> *say-output* 1) 
				 (format t "~% == BnB ==>> Not evaluated repeated node ~a" (snode-number i-child)))
			       )))
		       (t
			(when (and (not (snode-closed i-child))
				   (not (find i-child (gethash (snode-hash-code i-child) visited) :key #'snode-number)))
			  (push i-child (gethash (snode-hash-code i-child) visited)))))
		 (when (and (not (snode-closed i-child)) (null (snode-f-value i-child)))
		   (store-a-star-extras i-child cost-fn))
		 (when (and (not (snode-closed i-child)) 
			    (not (integerp (snode-recommended i-child))))
		   (setf (snode-recommended i-child) (match-recommend i-child)))
		   ))
	     
	     (setf siblings (stable-sort (remove-if #'(lambda (i-sibling)
							(or (snode-closed i-sibling)
							    (= 0 (snode-recommended i-sibling))))
						    (snode-children node)) 
					 #'< :key #'snode-f-value))
	     (cond ((snode-p (car siblings))
		    (setf (snode-selected (snode-parent (car siblings))) (car siblings))
		    (setf node (car siblings)))
		   (t 
		    (setf (snode-selected node) nil)
		    (setf (snode-closed node) t)
		    (when (> *say-output* 1) (format t "~% == BnB ==>> Cost Pruning at node ~a" (snode-number node)))
		    (setf node (if (snode-p (snode-parent node)) (snode-parent node) nil))
		    
		    ))
	     
	     (when (snode-p node)
	       (print-search-node node nil))
	     ))
	  ;; 	  In case of no solution was found
	  (t first-sol))))


(defun plan-cbp (&key (algorithm nil) (runtype *cbp-runtype*)
		 (heuristic nil) (cost *say-costfn*)
		 (helpful *say-helpful*) (w_g nil) (w_h nil)
		 (retrieve-option *cbp-retrieve-option*)
		 (timeout *say-timeout*) (depthbound *say-depthbound*)
		 (select-mode 'allob) (replay-mode nil))
  (declare (special *current-problem*))
  (init-properties)
  (let ((problem *current-problem*)
	(i-node nil)
	(algorithm-fn (if (not (null algorithm)) (symbol-function algorithm) #'guided-ehc))
	(heuristic-fn (if (not (null heuristic)) (symbol-function heuristic) (say-plan-defaults 'heuristic)))
 	(cost-fn (symbol-function cost))
	(start-time (get-internal-run-time))
	(search-options nil))

    (set-sayp :say-timeout timeout)
    (set-sayp :say-depthbound depthbound)
    
    (setf *say-solution* nil)
    (reset-trace)
    (sayout-initialize)
    (setf i-node (initialize-current-problem))
    (set-duplicate-hashing)
    
    (setf (get (problem-plist problem) :initial-pre-time) start-time)
    (setf (get (problem-plist problem) :initial-search-time) (get-internal-run-time))
    
    ;;Inserting search options
    (push (list :retrieve-option retrieve-option) search-options)
    (push (list :select-mode select-mode) search-options)

    (push (list :helpful helpful) search-options)
    (when (numberp max-solutions) (push (list :max-solutions max-solutions) search-options))
    (when w_g (push (list :w_g w_g) search-options))
    (when w_h (push (list :w_h w_h) search-options))
 
    (when replay-mode 
      (push (list :replay-mode replay-mode) search-options))

    (case runtype
       (ehc-cbr-utility (push (list :replay-sort-option 'utility) search-options))
       (ehc-steputility (push (list :replay-sort-option 'step-utility) search-options))
       (ehc-caseutility (push (list :replay-sort-option 'case-utility) search-options))
       (ehc-cbractive   (push (list :replay-sort-option 'active-utility) search-options)))

     (funcall algorithm-fn i-node heuristic-fn cost-fn search-options)
      ))


;;        (ehc-cbmetric (guided-ehc i-node heuristic-fn :select-mode 'instcase))
;;        (ehc-cborder (guided-ehc i-node heuristic-fn :select-mode 'allob))
;;        (ehc-cbr-utility (guided-ehc i-node heuristic-fn :select-mode 'allob :replay-sort-option 'utility))
;;        (ehc-steputility (guided-ehc i-node heuristic-fn :select-mode 'allob :replay-sort-option 'step-utility))
;;        (ehc-caseutility (guided-ehc i-node heuristic-fn :select-mode 'allob :replay-sort-option 'case-utility))
;;        (ehc-cbractive (guided-ehc i-node heuristic-fn :select-mode 'allob :replay-sort-option 'active-utility))
;;        (ehc (supported-ehc i-node heuristic-fn))
;;        (ehc-cbr-allob (supported-ehc i-node heuristic-fn :select-mode 'allob))
;;        (ehc-h-diff (supported-ehc i-node heuristic-fn :select-mode 'allob :replay-mode 'h-diff))
;; 	  (eager-hc  (supported-hillclimbing i-node heuristic-fn))
;; 	  (crx-sibling (ckx-sibling-ehc i-node heuristic-fn))

