;; SAYPHI Case-Based Reasoner
;; Extracting Case from Current Problem
;; Td. 22.01.2007
;; _________________________________________________________________________________________

(defvar *crx-runtype* 'sibling)
(defvar *crx-siblings-match* nil)
(defvar *crx-hdiff-match* nil)
(defvar *action-postable* nil)
(defvar *siblings-ints* nil)
(defvar *siblings-keys* nil)

(defmacro name-action (node)
  `(car (snode-applied-action ,node)))

(defmacro to-siblings-key (siblings-int)
  `(gethash ,siblings-int *siblings-keys*))

(defun crx-init()
  (setf *action-postable* (build-action-position-table))
  (setf *crx-siblings-match* (make-hash-table))
  (setf *crx-hdiff-match* (make-hash-table))
  (build-siblings-integers)
  )

(defun build-action-position-table ()
  (let ((action-postable (make-hash-table :test #'eq))
	(pos-count 0))
    (dolist (i-action (dom-actions *pspace*) action-postable)
      (setf (gethash (action-name i-action) action-postable) pos-count)
      (incf pos-count))))

(defun build-siblings-integers()
  (setf *siblings-ints* (make-hash-table :test #'equal))
  (setf *siblings-keys* (make-hash-table))
  (let ((match-pos 0))
    (dotimes (ipos (expt 2 (length (dom-actions *pspace*))))
      (let ((action-list nil))
	(dolist (iaction (dom-actions *pspace*))
	  (let ((action-pos (gethash (action-name iaction) *action-postable*)))
	    (when (logbitp action-pos ipos)
	      (push (action-name iaction) action-list))))
	(unless (null action-list)
	  (setf action-list (reverse action-list))
	  (setf (gethash action-list *siblings-ints*) ipos)
	  (setf (gethash ipos *siblings-keys*) action-list))))))
      

(defun extract-rxplan-case()
  (when (and (solution-p *say-solution*)
	     (solution-found *say-solution*))
    (let* ((solseq (solution-path *say-solution*))
	   (htable-sibling-match *crx-siblings-match*)
	   (htable-hdiff-match *crx-hdiff-match*))
      (dolist (x-node solseq htable-sibling-match)
	(let* ((siblings (snode-children (snode-parent x-node)))
	       (siblings-action-int (cons-siblings-int siblings))
	       (selaction-array nil) (hdiff-array nil)
	       (this-action-pos (gethash (name-action x-node) *action-postable*)))
	  (cond ((null (gethash siblings-action-int htable-sibling-match))
		 (setf selaction-array (make-array (length (dom-actions *pspace*)) :initial-element 0))
		 (incf (aref selaction-array this-action-pos))
		 (setf (gethash siblings-action-int htable-sibling-match) selaction-array)
		 
		 (setf hdiff-array (make-array (length (dom-actions *pspace*)) :initial-element 0))
		 (when (< (snode-h-value x-node) (snode-h-value (snode-parent x-node)))
		   (incf (aref hdiff-array this-action-pos)))
		 (setf (gethash siblings-action-int htable-hdiff-match) hdiff-array))
		
		
		(t
		 (incf (aref (gethash siblings-action-int htable-sibling-match) this-action-pos))
		 (when (< (snode-h-value x-node) (snode-h-value (snode-parent x-node)))
		   (incf (aref (gethash siblings-action-int htable-hdiff-match) this-action-pos)))
		 )))))))
  
  
;; (defun cons-siblings-action-list (siblings)
;;   (let ((action-list nil))
;;     (dolist (i-brother siblings (reverse action-list))
;;       (unless (find (car (snode-applied-action i-brother)) action-list)
;; 	(push (car (snode-applied-action i-brother)) action-list)))))


(defun cons-siblings-int (siblings)
  (let ((siblings-int 0))
    (dolist (i-brother siblings siblings-int)
      (let ((action-pos (gethash (car (snode-plan-action i-brother)) *action-postable*)))
	(setf (ldb (byte 1 action-pos) siblings-int) 1)))))


(defun pp-sibling-matches ()
  (maphash #'(lambda (siblings-int applied-array)
	       (let ((hdiff-array (gethash siblings-int *crx-hdiff-match*)))
		 (format t "~%[~a]" (to-siblings-key siblings-int))
		 (dolist (i-action (dom-actions *pspace*))
		   (when (> (aref applied-array (gethash (action-name i-action) *action-postable*)) 0)
		     (format t "~%   <~a: ~a (~a)>" 
			     (action-name i-action) 
			     (aref applied-array (gethash (action-name i-action) *action-postable*))
			     (aref hdiff-array (gethash (action-name i-action) *action-postable*))
			 )))))
	   *crx-siblings-match*))


(defun crx-runset (domain &key (algorithm *say-algorithm*)
			  (heuristic *say-heuristic*)
			  (timeout 5) (output-level 1)
			  (probs-prefix "p") 
			  (add-sibling-match nil) (use-sibling-match nil)
			  (add-path-context nil) (use-path-context nil)
			  (save-result nil) (out-file nil)
			  (experiment-var nil)
			  (restart-crx nil))
  (let* ((probdir (format nil "~aprobsets/" *domain-dir*))
	 (count-prob 1) (this-sol nil)
	 (result-file (if (not (stringp out-file))
			  (format nil "~a/result/~aset-~a-~a-~d.csv"
				  *domain-dir* probs-prefix algorithm heuristic timeout)
			(format nil "~a/result/~a" *domain-dir* out-file)))
	 (wild-char #+SBCL "*.*"
		    #-SBCL "*"))
    (when restart-crx (crx-init) (acx-init))
    (dolist (thisprob (dirfiles-sorted probdir (concatenate 'string probs-prefix wild-char)))
      (format t "~% Solving Problem ~a" (pathname-name thisprob))
      (read-pddl-problem thisprob)
      (cond (use-path-context      
	     (setf this-sol (plan-crx :runtype 'context :algorithm algorithm :heuristic heuristic 
				      :timeout timeout)))
	    (use-sibling-match      
	     (setf this-sol (plan-crx :runtype 'sibling :algorithm algorithm :heuristic heuristic 
				      :timeout timeout)))
	    (t
	     (setf this-sol (plan :algorithm algorithm :heuristic heuristic 
				  :timeout timeout ))))

      (when add-sibling-match
	(format t "~% Storing Siblings Match...~%")
	(extract-rxplan-case))
      (when add-path-context
	(format t "~% Storing Path Applicable Context... ~%")
	(extract-applicable-context-cases))

      (when save-result
	(with-open-file (out-stream result-file :direction :output :if-exists :append :if-does-not-exist :create)
	        (format out-stream "~d, ~a," count-prob (pathname-name thisprob))
		(format out-stream "~a~%" (format-solution this-sol))))
      (unless (null experiment-var)
	(push (list-solution count-prob (pathname-name thisprob) this-sol) 
	      (gethash experiment-var *experiment-results*)))
      (incf count-prob)
    )
))


(defun plan-crx (&key (algorithm *say-algorithm*) (runtype *crx-runtype*)
		      (heuristic *say-heuristic*) (cost *say-costfn*)
		      (timeout *say-timeout*) (depthbound *say-depthbound*))
  (declare (special *current-problem*))
  (let ((problem *current-problem*)
	(i-node nil)
	(algorithm-fn (symbol-function algorithm))
	(heuristic-fn (symbol-function heuristic))
	(cost-fn (symbol-function cost))
	(start-time (get-internal-run-time))
	(sol nil))

    (set-sayp :say-timeout timeout)
    (set-sayp :say-depthbound depthbound)
    
    (setf *say-solution* nil)
    (reset-trace)
    (sayout-initialize)
    (setf i-node (initialize-current-problem))
    
    (setf (get (problem-plist problem) :initial-pre-time) start-time)
    (setf (get (problem-plist problem) :initial-search-time) (get-internal-run-time))
    
     (case runtype
	   ('sibling
	    (ckx-sibling-ehc i-node heuristic-fn))
	   ('context
	    (context-ehc i-node heuristic-fn))
	   ('hc-crx
	    (ckx-sibling-hc i-node heuristic-fn))
)))



;; ====================================================================================
;; CRX Using control knowlegde acquired from the relaxed plan and siblings information
;; ===================================================================================

(defun ckx-sibling-ehc (init-node h-fn
				  &optional (problem *current-problem*))
  (declare (ignore cost-fn))
  (let ((open-nodes nil) (open-children nil)
	(node init-node) (next-node nil) (visited nil)
	(current-h 0) (better-h-node nil))
    (store-h-extras node h-fn)
    (setf current-h (snode-h-value node))
    (do* ((stop-this (stop-search node problem) (stop-search node problem)))
	 (stop-this (when-stop (cdr stop-this) node problem))

	 (expand-state node :helpful *say-helpful*)
	 (push node visited)

	 (setf better-h-node (get-ckx-sibling-heuristic node h-fn current-h visited))

	 (cond ((snode-p better-h-node)
		(setf current-h (snode-h-value better-h-node))
		(setf node better-h-node)
		(setf open-nodes nil)
		(setf open-children nil))
	       (t 
		(setf open-children (append (remove-if #'snode-closed (snode-children node)) 
					  open-children))
		(when (> *say-output* 1) (format t "  ~% [Expanding Breadth Level]"))
		(cond ((not (snode-p (setf next-node (pop open-nodes))))
;; 		       (setf open-nodes (breadth-open-nodes open-children))
		       (setf open-nodes open-children)
 		       (when (null open-nodes) ;;Recuperando la poda por las helpful actions
			 (setf open-nodes (restore-nonhelpful node h-fn visited))
 			 (when (> *say-output* 1) 
			   (format t "  ~%!!Restoring NON Helpful Nodes.")))
		       (setf node (pop open-nodes))
		       (setf open-children nil))
		      (t 
		       (setf node next-node)))))       
	 (when (snode-p node)
	   (print-search-node node nil)
))))


(defun get-ckx-sibling-heuristic (node h-function current-h visited-nodes &key (discrete *say-h-rounded*))
  (let ((better-found nil) (node-counter 0)
	(sorted-children (ck-sort-siblings node)))
;; 	(sorted-children (ck-sort-siblings (snode-children node))))
    (dolist (i-child sorted-children better-found)
      (when (null better-found)
	(cond ((not (find i-child visited-nodes :test #'equal-state))
	       (incf node-counter)
	       (store-h-extras i-child h-function)
	       (cond (discrete
		      (when (< (floor (snode-h-value i-child)) (floor current-h))
			(setf better-found i-child)
			(when *trace-ehc*
			  (push node-counter *hit-node*))))
		     (t
		       (when (< (snode-h-value i-child) current-h)
			 (setf better-found i-child))))) 
	      (t
	       (setf (snode-closed i-child) t)))
	       ))))


(defun ck-sort-siblings (node)
  (let* ((sorted nil)
	 (this-actions-int (cons-siblings-int (snode-children node)))
	 (ck-array (gethash this-actions-int *crx-siblings-match*)))
    (when (> *say-output* 1)
;;        (format t "~%    ==>CRX Key Used ~a" this-action-list)
      )
    (cond ((not (null ck-array))
	   (setf (snode-children node) (stable-sort (snode-children node) #'(lambda (x y)
						  (> (aref ck-array (gethash (name-action x) *action-postable*))
						     (aref ck-array (gethash (name-action y) *action-postable*))))))
	  (when (> *say-output* 1)
;;  	    (format t "~%    =====>Array[~a] Sorted[~a]" 
;;  		    ck-array (cons-siblings-action-list (snode-children node)))
	    )
	    (snode-children node))
	  (t (snode-children node)))))

			      

(defun ckx-sibling-hc (init-node h-fn
				  &optional (problem *current-problem*))
  (declare (ignore cost-fn))
  (let ((open-nodes nil) (open-children nil)
	(node init-node) (next-node nil) (visited nil)
	(current-h 0) (better-h-node nil))
    (store-h-extras node h-fn)
    (setf current-h (snode-h-value node))
    (do* ((stop-this (stop-search node problem) (stop-search node problem)))
	 (stop-this (when-stop (cdr stop-this) node problem))

	 (expand-state node :helpful *say-helpful*)
	 (push node visited)

	 (setf better-h-node (get-ckx-sibling-heuristic node h-fn current-h visited))

	 (cond ((snode-p better-h-node)
		(setf current-h (snode-h-value better-h-node))
		(setf node better-h-node))
	       (t 
		(setf open-children (remove-if #'snode-closed (snode-children node)))
		(when (> *say-output* 1) (format t "  ~% [H-fn not improved...selecting next node"))
		(setf node (select-crx-next-node open-children))))
	 
	 (when (snode-p node)
	   (print-search-node node nil)
))))


(defun select-crx-next-node (children)
  (let* ((next-node (car children))
	 (h-value (snode-h-value next-node))
	 (name-action (car (snode-applied-action (car children)))))
    (dolist (i-child (rest children) next-node)
      (when (eq (car (snode-applied-action i-child)) name-action)
	(when (< (snode-h-value i-child) h-value)
	  (setf next-node i-child)
	  (setf h-value (snode-h-value next-node)))))))

	
    