;; SAYPHI Case-Based Reasoner
;; Selecting Replay Type Cases
;; Td. 10.02.2006
;; _________________________________________________________________________________________

(defvar *replay-allobjs* nil)
(defvar *replay-candidates* nil)

(defun get-cases-list (type sort-option case-base)
  (let ((cases-list nil))
    (maphash #'(lambda (case-no case)
		 (declare (ignore case-no))
		 (push case cases-list))
	     (gethash type case-base))
    (cond ((or (eq sort-option 'utility)
	       (eq sort-option 'active-utility)
	       (eq sort-option 'case-utility))
	   (setf cases-list (stable-sort cases-list #'greaterthan-greedy-utility)))
	  (t cases-list))))
  


(defun read-type-case-base()
  (setf *type-case-base* (make-hash-table :test #'eq))
  (maphash #'(lambda (type objects)
	       (declare (ignore objects))
	       (let ((type-cases-table (make-hash-table))
		     (stored-cases (mapcar #'list-to-typedcase (read-cbp-file type))))
		 (dolist (i-case stored-cases)
		   (setf (gethash (typedcase-case-no i-case) type-cases-table) i-case))
		 (setf (gethash type *type-case-base*) type-cases-table)))
	   (problem-objects *current-problem*)))


(defun read-object-case-base()
  (setf *object-case-base* (make-hash-table :test #'eq))
  (maphash #'(lambda (type objects)
	       (declare (ignore type))
	       (let ((object-cases-table (make-hash-table)))
		 (dolist (iob objects)
		   (let ((stored-cases (mapcar #'list-to-typedcase (read-cbp-file iob))))
		     (dolist (i-case stored-cases)
		       (setf (gethash (typedcase-case-no i-case) object-cases-table) i-case))
		     (setf (gethash iob *object-case-base*) object-cases-table)))))
	   (problem-objects *current-problem*)))



(defun greater-similarity-length (candidate-a candidate-b)
  (cond ((> (cdr candidate-a) (cdr candidate-b)) t)
	((= (cdr candidate-a) (cdr candidate-b)) 
	 (< (length (typedcase-sequence (car candidate-a)))
	    (length (typedcase-sequence (car candidate-b)))))
	(t nil)))
	


;; Compute the similarity between property sets (IF prob-properties is a SUBSET of store-properties)
;;    When the number of properties does not match it returns 1, 2 for whole match, 0 for nothing
;; (ie.   ((AT2) (2)) & ((AT2) (2)) -> 2
;;        ((AT2) (3)) & ((AT2) (2)) -> 1
;;        ((AT2) (2)) & ((IN2) (1)) -> 0
(defun stepcase-extras-memberset (store-properties-extras prob-properties-extras)
  (let ((store-extras (mapcar #'cons (car store-properties-extras) (cadr store-properties-extras)))
	(prob-extras (mapcar #'cons (car prob-properties-extras) (cadr prob-properties-extras)))
	(is-memberset 2))
    (dolist (i-prop prob-extras is-memberset)
      (when is-memberset
	(when (not (find i-prop store-extras :test #'equal))
	  (cond ((find (car i-prop) store-extras :test #'eq :key #'car)
		 (setf is-memberset 1))
		 (t
		  (setf is-memberset nil))))))))



(defun rxkey-memberset (prob-rxkey case-rxkey)
  (cond ((equalp prob-rxkey case-rxkey) 2)
	((equalp (mapcar #'car prob-rxkey) (mapcar #'car case-rxkey)) 1)
	(t 0)))


(defun select-object-case (object cases)
  (let ((objgoal-properties (car (objstate-properties object (problem-goals *current-problem*))))
	(objinit-properties (car (objstate-properties object (problem-init-state *current-problem*))))
	(selected-case nil)
	(partial-candidate nil))
    (dolist (icase cases)
      (let* ((sequence (typedcase-sequence icase)))
	(unless selected-case
	  (when (stepcase-memberset (casestep-properties (car (last sequence))) objgoal-properties)
	    (unless partial-candidate
	      (setf partial-candidate icase))
	    (when (stepcase-memberset (casestep-properties (first sequence)) objinit-properties)
	      (setf selected-case icase))))))
    (if selected-case selected-case partial-candidate)
    ))


(defun select-object-case-extra (object cases)
  (let ((objgoal-properties (objstate-properties object (problem-goals *current-problem*)))
	(objinit-properties (objstate-properties object (problem-init-state *current-problem*)))
	(match-result nil)
	(candidate-cases (mapcar #'(lambda (x)(cons x 0)) cases)))
    ;; Matching the goal typed sub-states
    (dolist (icandidate candidate-cases)
      (let* ((sequence (typedcase-sequence (car icandidate))))
	(setf match-result (stepcase-extras-memberset (list (casestep-properties (car (last sequence))) 
							    (casestep-prop-extras (car (last sequence))))
						      objgoal-properties))
	(when (numberp match-result)
	  (setf (cdr icandidate) match-result))))
        ;; Matching the init typed sub-states
    (dolist (icandidate candidate-cases)
      (when (> (cdr icandidate) 0)
	(let* ((sequence (typedcase-sequence (car icandidate))))
	  (setf match-result (stepcase-extras-memberset (list (casestep-properties (car sequence)) 
							      (casestep-prop-extras (car sequence)))
							objinit-properties))
	  (when (numberp match-result)
	    (setf (cdr icandidate) (+ (cdr icandidate) (* 2 match-result)))))))

;;     (format t "~% Candidate Cases ~a ~a " object candidate-cases)
    (setf candidate-cases (remove-if #'(lambda (x) (< (cdr x) 3)) candidate-cases))
    (unless (null candidate-cases)
     (setf candidate-cases (stable-sort candidate-cases #'greater-similarity-length))
     (setf (gethash object *replay-candidates*) candidate-cases)
     (car (first candidate-cases)))))



(defun select-object-case-rxkey (object cases)
  (let ((objgoal-properties (objstate-properties object (problem-goals *current-problem*)))
	(objinit-properties (objstate-properties object (problem-init-state *current-problem*)))
	(match-result nil)
	(goal-graph *h-achieved-goals*)
	(candidate-cases (mapcar #'(lambda (x)(cons x 0)) cases)))
    (dolist (icandidate candidate-cases)
      (let* ((sequence (typedcase-sequence (car icandidate))))
	(setf match-result (stepcase-extras-memberset (list (casestep-properties (car (last sequence))) 
							    (casestep-prop-extras (car (last sequence))))
						      objgoal-properties))
	(when (numberp match-result)
	  (setf (cdr icandidate) match-result))))
        ;; Matching the init typed sub-states
    (dolist (icandidate candidate-cases)
      (when (> (cdr icandidate) 0)
	(let* ((sequence (typedcase-sequence (car icandidate))))
	  (setf match-result (stepcase-extras-memberset (list (casestep-properties (car sequence)) 
							      (casestep-prop-extras (car sequence)))
							objinit-properties))
	  (when (numberp match-result)
	    (setf (cdr icandidate) (+ (cdr icandidate) (* 2 match-result)))))))

;;     (format t "~% Candidate Cases ~a ~a " object candidate-cases)
    (dolist (icandidate candidate-cases)
      (when (> (cdr icandidate) 3)  ;;like the remove step in retrieve-extra
	(setf match-result (rxkey-memberset (generate-rx-key object goal-graph) 
					    (typedcase-rx-key (car icandidate))))
	(when (numberp match-result)
	  (setf (cdr icandidate) (+ (cdr icandidate) (* 3 match-result))))))
	
   
     (unless (null candidate-cases)
       (setf candidate-cases (stable-sort candidate-cases #'greater-similarity-length))
       (setf (gethash object *replay-candidates*) candidate-cases) 
       (car (first candidate-cases)))
))



(defun select-cases-toreplay (&key (replay-allobjs *replay-allobjs*)
				   (trace-selected *cboption-traceselected*)
			           (retrieve-option nil)
				   (reload-cases nil)
				   (sort-option nil))
  (when (or (null *type-case-base*) reload-cases)
    (read-type-case-base))
  (setf *replay-candidates* (make-hash-table :test #'eq))
  (let* ((all-objects (get-all-objects))
	 (goals-objs (get-goals-objects))
	 (replay-objects (if replay-allobjs all-objects goals-objs))
	 (replay-table (make-hash-table :test #'eq)))
    (dolist (iob replay-objects replay-table)
      (let* ((itype (get-object-type iob))
	     (cases (get-cases-list itype sort-option *type-case-base*))
 	     (selected-case nil)
	     (ireplay nil))
	(cond ((eq retrieve-option 'rx-key)
	       (setf selected-case (select-object-case-rxkey iob cases)))
	      (t (setf selected-case (select-object-case-extra iob cases))))
	       
	(when (typedcase-p selected-case)
	  (setf ireplay (make-replay :object iob
				     :case-no (typedcase-case-no selected-case)
				     :object-type itype
				     :goal-object-p (when (find iob goals-objs) t)
			             :sequence (mapcar #'to-instep (typedcase-sequence selected-case))
				     :typedcase selected-case))
	  (setf (gethash iob replay-table) ireplay)
	  (when (and trace-selected (hash-table-p *cbinfo-usedcases*))
	    (let ((usedcase-key (cons itype (typedcase-case-no selected-case))))
	      (if (numberp (gethash usedcase-key *cbinfo-usedcases*))
		  (incf (gethash usedcase-key *cbinfo-usedcases*))
	      (setf (gethash usedcase-key *cbinfo-usedcases*) 1))))
	  )))))




(defun select-instcases-toreplay (&key (replay-allobjs *replay-allobjs*)
				   (reload-cases nil)
				   (sort-option nil))
  (when (or (null *object-case-base*) reload-cases)
    (read-object-case-base))
  (let* ((all-objects (get-all-objects))
	 (goals-objs (get-goals-objects))
	 (replay-objects (if replay-allobjs all-objects goals-objs))
	 (replay-table (make-hash-table :test #'eq)))
    (dolist (iob replay-objects replay-table)
      (let* ((cases (get-cases-list iob sort-option *object-case-base*))
	     (selected-case (select-object-case-extra iob cases))
	     (ireplay nil))
	(when (typedcase-p selected-case)
	  (setf ireplay (make-replay :object iob
				     :case-no (typedcase-case-no selected-case)
				     :object-type iob
				     :goal-object-p (when (find iob goals-objs) t)
			             :sequence (mapcar #'to-instep (typedcase-sequence selected-case))
				     :typedcase selected-case))
	  (setf (gethash iob replay-table) ireplay)
	  )))))






    

