;; SAYPHI Case-Based Reasoner
;; Extracting Case from Current Problem
;; Td. 10.02.2006
;; _________________________________________________________________________________________


(defun x_list-cases (cases)
  (dolist (i-case cases)
    (format t "~%  Case-no:~a Len: ~a" (typedcase-case-no i-case) (length (typedcase-sequence i-case)))))
	    
(defmacro nodes-h-difference (previous node)
  `(- (floor (snode-h-value ,node))
      (floor (snode-h-value ,previous))))

(defmacro is-parameter-object (object applied-action)
  `(when (find ,object (gaction-planaction ,applied-action)) t))


(defun applied-gactionpos (node)
  (let ((gaction (snode-applied-action node)))
    (cons (car (gaction-planaction gaction)) (gaction-int gaction))))


(defun generate-sequence (object solseq init-node)
  (let* ((case-seq nil) 
	 (no-op-count 0)
	 (temp-obsubstate nil) (next-obsubstate nil) 
	 (prev-node init-node)
	 (current-step nil))
    (setf temp-obsubstate (objstate-properties object (snode-state init-node)))
    (push (make-casestep :properties (car temp-obsubstate)
			 :prop-extras (second temp-obsubstate)
	                 :h-diff 0)
	  case-seq)
    
    (dolist (x-node solseq case-seq)
      (setf next-obsubstate (objstate-properties object (snode-state x-node)))
      (cond ((and (equalp next-obsubstate temp-obsubstate)
		  (not (is-parameter-object object (snode-applied-action x-node))))
	     (setf current-step (make-noop next-obsubstate no-op-count 0))
      	     (incf no-op-count))
 	    (t
 	     (when (casestep-p current-step)
 	       (setf case-seq (append case-seq (list current-step)))
 	       (setf current-step nil)
 	       (setf no-op-count 0))
	     (setf case-seq (append case-seq 
 				      (list (make-casestep 
 					     :properties (car next-obsubstate)
					     :prop-extras (second next-obsubstate)
 					     :app-action (applied-gactionpos x-node)
					     :h-diff (nodes-h-difference x-node prev-node)
					     ))))
 	     (setf temp-obsubstate next-obsubstate)))
      (setf prev-node x-node)

)))


;; It gets the relaxed-plan key to differentiate cases within middle steps
;; I need to repeat the init-node evaluation because I do not know if the achieved membership
;; has been computed
(defun generate-rx-key (object goal-graph)
  (let ((rx-key-list nil))
    (dotimes (ilayer (hash-table-count goal-graph) (reverse rx-key-list))
      (when (> ilayer 0)
	(push (compound-properties (objstate-properties object (gethash ilayer goal-graph)))
	      rx-key-list)))))


(defun get-init-achieved-graph (node)
  (setf *get-achieved-goals* t)
  (store-h-extras node (say-plan-defaults 'heuristic))
  (setf *get-achieved-goals* nil)
  *h-achieved-goals*)


;; The rx-key is the footprint of the relaxed properties transition
;; The prob-metric is the metric used solving the problem
(defun save-typed-case (type case-sequence &key (rx-key nil)
			                        (prob-metric nil))
			
  (let ((new-case nil) (all-cases nil)
	(stored-cases (mapcar #'list-to-typedcase (read-cbp-file type))))
    (setf new-case (make-typedcase :sequence case-sequence
				   :metric prob-metric
				   :rx-key rx-key))
    (setf all-cases (merge-cases new-case stored-cases))
    (unless (null all-cases)
      (setf all-cases (stable-sort all-cases #'< :key #'typedcase-case-no))
      (write-cbp-file type (mapcar #'typedcase-to-list all-cases)))))

;; It computes the relaxed key for an object.  In practice, the goal graph is computed once
;; and then a rx-key is generated for each object
(defun relaxed-footprint-key (object node)
  (let ((goal-graph (get-init-achieved-graph node)))
    (generate-rx-key object goal-graph)))


(defun store-case(&optional (instobj-seq nil))
  (when (and (solution-p *say-solution*)
	     (solution-found *say-solution*))
    (let* ((problem *current-problem*)
	   (solseq (solution-path *say-solution*))
	   (metric (second (problem-metric problem)))
	   (goal-graph (get-init-achieved-graph (problem-search-tree *current-problem*))))
      
      (init-properties)
      (maphash #'(lambda (type objects-of-type)
		   (dolist (iobject objects-of-type)
		     (let ((typeseq (generate-sequence iobject solseq (problem-search-tree *current-problem*)))
			   (rx-key (generate-rx-key iobject goal-graph)))
		     (cond (instobj-seq
			    (save-typed-case iobject typeseq :rx-key rx-key
					                     :prob-metric metric))
			   (t 
			    (save-typed-case type typeseq :rx-key rx-key 
					                  :prob-metric metric)))
		     )))
	       (problem-objects problem)))))

			      
