;; SAYPHI Case-Based Reasoner
;; Case Utility for Improving Time & Plan Quality
;; Td. 07.02.20076
;; _________________________________________________________________________________________

(defvar *cbp-store-utility* nil)
(defvar *cbp-case-w-greedy* 0.75)
(defvar *cbp-step-w-greedy* 0.5)
(defvar *cbp-active-w-greedy* 0.5)

;; Recieves a key in a list form with the type, the case number and the sequence position
(defmacro object-casestep-utility (step-key-list)
  `(let ((type (car, step-key-list))
	 (case-no (second ,step-key-list))
	 (pointer (third ,step-key-list)))
     (casestep-utility (nth pointer (typedcase-sequence (gethash case-no (gethash type *type-case-base*)))))))


;; This saves the recommendation for the node to later determine its utility
(defun save-step-advise (case-object node)
  (let ((replay (object-replay case-object)))
    (setf (snode-advise-step node)
	(list (replay-object-type replay)
	      (replay-case-no replay)
	      (1+ (replay-case-pointer replay))
	      (replay-object replay)))))


(defun compute-cbp-utility()
  (when (and (solution-p *say-solution*)
	     (solution-found *say-solution*))
    (dolist (i-node (solution-path *say-solution*))
      (let ((siblings (snode-children (snode-parent i-node))))
	(dolist (i-brother siblings)
	  (unless (null (snode-advise-step i-brother))
	    (when (> *say-output* 1)
	      (format t "~% Action: ~a  Advise-key:~a" (snode-plan-action i-brother) (snode-advise-step i-brother)))
	    (incf (cdr (object-casestep-utility (snode-advise-step i-brother))))
	    (incf (cdr (replay-current-utility (gethash (fourth (snode-advise-step i-brother)) *replay-table*))))
	    (when (= (snode-number i-brother) (snode-number i-node))
	      (incf (car (object-casestep-utility (snode-advise-step i-brother))))
	      (incf (car (replay-current-utility (gethash (fourth (snode-advise-step i-brother)) *replay-table*))))
	      )))))
    ;;   Reasigning the utility values of the global cases
    (maphash #'(lambda (object replay)
		 (declare (ignore object))
		 (let ((this-typedcase (gethash (replay-case-no replay) (gethash (replay-object-type replay) *type-case-base*))))
		   ;; 		 (format t "~% Previos Utility for ~a~a ~a" (replay-case-no replay) (replay-object-type replay) 
		   ;; 			 (typedcase-utility this-typedcase))
		   (setf (car (typedcase-utility this-typedcase))
			 (+ (car (typedcase-utility this-typedcase)) (car (replay-current-utility replay))))
		   (setf (cdr (typedcase-utility this-typedcase))
			 (+ (cdr (typedcase-utility this-typedcase)) (cdr (replay-current-utility replay))))
		   ;; 		 (format t "~% New Utility for ~a~a ~a" (replay-case-no replay) (replay-object-type replay) 
		   ;; 			 (typedcase-utility this-typedcase))
		   ))
	     *replay-table*)
  ))


(defun real-utility (utility-cons)
   (cond ((= 0 (cdr utility-cons)) 0)
	 (t (/ (car utility-cons) (cdr utility-cons)))))

(defun greaterthan-greedy-utility (case-a case-b)
  (let ((utility-a (real-utility (typedcase-utility case-a)))
	(utility-b (real-utility (typedcase-utility case-b))))
;;      If it does not reach the threshold the less used case is preferred
;;      If some utility reaches the threshold the case with better utility is preferred    
    (cond ((and (< utility-a *cbp-case-w-greedy*)
		(< utility-b *cbp-case-w-greedy*))
	   (< (cdr (typedcase-utility case-a)) (cdr (typedcase-utility case-b))))
	  (t (> utility-a utility-b)))))
	  
	   
(defun greater-current-step-utility (replay-a replay-b)
  (cond ((not (casestep-p (nth (1+ (replay-case-pointer replay-b)) (replay-sequence replay-b)))) t)
	((not (casestep-p (nth (1+ (replay-case-pointer replay-a)) (replay-sequence replay-a)))) nil)
	(t
	 (let* ((casestep-a (nth (1+ (replay-case-pointer replay-a)) (typedcase-sequence (replay-typedcase replay-a))))
		(casestep-b (nth (1+ (replay-case-pointer replay-b)) (typedcase-sequence (replay-typedcase replay-b))))
		(utility-a (real-utility (casestep-utility casestep-a)))
		(utility-b (real-utility (casestep-utility casestep-b))))	   
	   (cond ((and (< utility-a *cbp-step-w-greedy*)
		       (< utility-b *cbp-step-w-greedy*))
		  (< (cdr (casestep-utility casestep-a)) (cdr (casestep-utility casestep-b))))
		 (t (> utility-a utility-b)))))))


(defun greater-active-utility (replay-a replay-b)
  (cond ((not (casestep-p (nth (1+ (replay-case-pointer replay-b)) (replay-sequence replay-b)))) t)
	((not (casestep-p (nth (1+ (replay-case-pointer replay-a)) (replay-sequence replay-b)))) nil)
	(t
	 (let* ((utility-a (real-utility (replay-active-utility replay-a)))
		(utility-b (real-utility (replay-active-utility replay-b))))
	   (cond ((and (<= utility-a *cbp-active-w-greedy*)
		       (<= utility-b *cbp-active-w-greedy*))
		  (< (cdr (replay-active-utility replay-a)) (cdr (replay-active-utility replay-b))))
		 (t (> utility-a utility-b)))))))
	   

