;; SAYPHI Planner 
;; Domain Encoding & Instantiation 
;; Td. 14.09.2007
;; _________________________________________________________________________________________



;; A hash table holding the bitmap of semi-instantiated predicates
;; The value holds a cell cons (bitmap . n-initial-occurrence
(defparameter *semi-inst-pred* nil)
(defparameter *actions* nil)
(defparameter *factgraph-table* nil)
(defparameter *fluentgraph-table* nil)
(defparameter *total-gactions* 0)


(defun pred-true-able (pred)
  (cond ((predicate-is-added (gethash pred *predicates*)) t)
	(t nil)))

(defun pred-false-able (pred)
  (cond ((predicate-is-deleted (gethash pred *predicates*)) t)
	(t nil)))


;;==============================================================
;;Multiplying parameters for finding ground actions
;;==============================================================


(defun oppred-argpos (operator oppred)
  (cons (car oppred) (mapcar #'(lambda (x)
		   	 (position x (action-parameters operator) :test #'equal))
			     (cdr oppred))))


(defun oppred-substitution (operator oppred inst-args)
  (let ((pred-argpos (mapcar #'(lambda (x)
				 (position x (action-parameters operator) :test #'equal))
			     (cdr oppred))))
    (cons (car oppred) (mapcar #'(lambda (pos) (nth pos inst-args))
			       pred-argpos))
    ))

(defun create-oppred-argpos ()
  (dolist (iop (dom-actions *pspace*))
    (setf (action-prec-argpos iop) (mapcar #'(lambda (pred)
					      (oppred-argpos iop pred)) 
					  (action-preconditions iop)))
    (setf (action-adds-argpos iop) (mapcar #'(lambda (pred)
					      (oppred-argpos iop pred)) 
					  (action-adds iop)))
    (setf (action-dels-argpos iop) (mapcar #'(lambda (pred)
					      (oppred-argpos iop pred)) 
					  (action-dels iop)))))
    
		   
;; OP-inertia. It retursn a list of (pred pos-arg1 ... pos-argn)
(defun collect-op-inertiaprec (operator)
  (let ((op-inertia nil))
    (dolist (iprec (action-prec-argpos operator) op-inertia)
      (unless (pred-true-able (car iprec))
	(push iprec op-inertia)))))

;; @@-- Verificar si el estado inicial no ha sido modificado en *facts*
(defun verify-inertia-state (inertia ground-args &optional (facts *init-facts*))
  (when (gethash (cons (car inertia)
		       (mapcar #'(lambda (posarg) (nth posarg ground-args)) (cdr inertia)))
		 facts)

    t))

(defun verify-inertia-pos-state (inertia ground-args parampos &optional (facts *init-facts*))
  (when (gethash (cons (car inertia)
		       (mapcar #'(lambda (posarg) (nth (- posarg parampos) ground-args)) (cdr inertia)))
		 facts)

    t))





;;Returns the fact obtained from the substitution of the predicate positions in parameters
(defun literal-posparams (posargs instances)
  (cons (car posargs)
	(mapcar #'(lambda (posarg) (nth posarg instances)) (cdr posargs))))

;;Returns the pred-pos for op-ground predicates. It creates the fact if it doesn't exist
(defun set-op-ground-preds (literal)
  (let ((fact (gethash literal *facts*)))
    (if (null fact)
	 (fact-predpos (create-fact literal))
	 (fact-predpos fact))))
	  

(defun build-oppred-inst(op-preds instances)
  (let ((inst-preds nil))
    (dolist (ipred op-preds (reverse inst-preds))
      (push (set-op-ground-preds (literal-posparams ipred instances)) inst-preds))))
      
   
(defun create-gaction (operator instances)
  (make-gaction
   :planaction (cons (action-name operator) instances)
   :preconds (build-oppred-inst (action-prec-argpos operator) instances)
   :adds (build-oppred-inst (action-adds-argpos operator) instances)
   :dels (build-oppred-inst (action-dels-argpos operator) instances)
   ))


;; It veryfies if adds and dels are equal for marking the action as irrelevant
(defun find-relevant-effects (gaction)
  (let ((relevant nil))
    (cond ((= (length (gaction-adds gaction)) (length (gaction-dels gaction)))
	   (dolist (ieff (gaction-adds gaction) relevant)
	     (unless (find ieff (gaction-dels gaction) :test #'equal)
	       (setf relevant t))))
	  (t t))))
  
				

(defun add-element-multi-list (lists objects)
  (cond ((null lists) (mapcar #'list objects))
	(t
	 (let ((new-lists nil))
	   (dolist (i-ob objects new-lists)
	     (dolist (j-list lists)
	       (let ((new-inst (copy-list j-list)))
		 (push (push i-ob new-inst) new-lists))))))))

;; It removes the inertia list if the arguments are not present
(defun inertia-of-args (inertia-list total-args used-args)
  (remove-if #'(lambda (inertia) 
		 (some #'(lambda (x) (< x (- total-args used-args))) (cdr inertia)))
	     inertia-list))
  

;; It returs the object list of the specified parameter
;; I check out if it is a unary inertia for avoiding irrelevants objects
(defun objects-for-multiply (op inertia param)
  (let* ((list-objects (gethash (cdr param) (problem-inheritobjects *current-problem*))))
    (cond ((null inertia) list-objects)
	  (t 
	   (let* ((param-pos (position param (action-parameters op) :test #'equal))
		  (unary-inertia (remove-if #'(lambda (in)
				      (or  (not (= (length (cdr in)) 1))
					   (not (= (second in) param-pos)))) inertia))
		 (relevant-obs (copy-list list-objects)))
	     (dolist (uinertia unary-inertia relevant-obs)
	       (dolist (iob list-objects)
		 (unless (gethash (list (car uinertia) iob) *init-facts*)
		   (setf relevant-obs (remove iob relevant-obs))
		      ))))))))
    
	 

(defun multiexpand-op (op inertia expinst rparams viewed-inertia)
  (cond ((null rparams) expinst)
	 (t
	  (let* ((list-objects (objects-for-multiply op inertia (car rparams)))
		 (new-expinst (add-element-multi-list expinst list-objects))
		 (expand-iter (- (length (action-parameters op)) (length (cdr rparams))))
		 (inertia-positer (length (cdr rparams)))
		 (this-inertia (remove-if #'(lambda (in) (or (find in viewed-inertia :test #'equal)
							     (= (length (cdr in)) 1)))
					  (inertia-of-args inertia (length (action-parameters op)) expand-iter)))
		 (relevant-inst nil))
	    (dolist (instargs new-expinst)
	      (let ((args-relevant t))
		(dolist (inertia this-inertia)
		  (unless (verify-inertia-pos-state inertia instargs inertia-positer)
		    (setf args-relevant nil)))
		(when args-relevant
		  (push instargs relevant-inst))))
	    (multiexpand-op op inertia relevant-inst (cdr rparams) (append this-inertia viewed-inertia))
))))


(defun cost-operator-p (operator)
  (and (null (action-adds operator))
       (null (action-dels operator))
       (not (null (action-costs operator)))))

(defun multiply-inertia-operators ()
  (dolist (op (dom-actions *pspace*))
    (let* ((inertia-prec (collect-op-inertiaprec op))
	   (ground-args (multiexpand-op op inertia-prec nil (reverse (action-parameters op)) nil))
	   (cost-operator-p (cost-operator-p op)))
      (dolist (i-ground ground-args)
	(let ((ground-action (create-gaction op i-ground)))
	  (when (or cost-operator-p 
		    (find-relevant-effects ground-action))
	    (setf (gaction-operator ground-action) op)
	    (setf (gaction-int ground-action) *total-gactions*)
	    (incf *total-gactions*)
	    (setf (gaction-num-precs ground-action) (length (gaction-preconds ground-action)))
	    (push ground-action *actions*))))
      )))



;;=============================================================================
;; Instantiating Numeric Preconditions & Costs
;;=============================================================================

;; This three functions return as variables the real-functor vars and substitute the constants
(defun inst-gaction-funeval (operand gaction)
  (let ((istate (problem-init-state *current-problem*))
	(real-functors (dom-real-functors *pspace*)) (state-variable nil))
    (cond ((numberp operand) operand)
	  ((hkey-present (car operand) istate)
	   (cond ((not (find (car operand) real-functors))
		  (setf state-variable (fluent-state-value (oppred-substitution (gaction-operator gaction) operand
										(cdr (gaction-planaction gaction))) istate))
		  (if (numberp state-variable) state-variable 0))
		 (t
		  (let ((fluent-lit (oppred-substitution (gaction-operator gaction) 
							 operand (cdr (gaction-planaction gaction)))))
		    (fluent-predpos (gethash fluent-lit *fluents*))))))

	  (t 
	   (inst-funeval-aux operand gaction)
	   ))))


(defun map-gaction-inst-funeval (operand-list gaction)
  (mapcar (lambda (ioperand)
	    (inst-gaction-funeval ioperand gaction))
	  operand-list))


(defun inst-funeval-aux (ifun gaction)
  (let ((temp-operand (map-gaction-inst-funeval (cdr ifun) gaction)))
    (cond ((every #'numberp temp-operand)
	   (apply (car ifun) temp-operand))
 	  ((and (some #'numberp temp-operand)
		(some #'is-linearmath-operation temp-operand))
 	   (transform-fun-notation (car ifun) temp-operand))
	  (t 
	   (push (car ifun) temp-operand)))))




;; It's done after multiplying operators, but it could be integrated in the same algorithm.
(defun instantiate-precfuns ()
  (dolist (igaction *actions*)
    (setf (gaction-precfuns igaction)
	  (mapcar #'(lambda (ifun)
		      (inst-gaction-funeval (second ifun) igaction))
		  (action-precond-funs (gaction-operator igaction))))
    (setf (gaction-num-precfuns igaction) (length (gaction-precfuns igaction)))
    ))


(defun instantiate-costs ()
  (dolist (igaction *actions*)
    (setf (gaction-costs igaction)
	  (mapcar #'(lambda (icost)
		      (let ((fluent-lit (oppred-substitution (gaction-operator igaction) (second icost)
							     (cdr (gaction-planaction igaction)))))
;; 			(format t "~% >> Fluent: ~a" fluent-lit)
			(list (car icost)
			      (fluent-predpos (gethash fluent-lit *fluents*))
			      (cond ((inst-gaction-funeval (third icost) igaction))
				    (t 0)))))
		  (action-costs (gaction-operator igaction))))))



;;=============================================================================
;; Computing the connectivity graph
;;=============================================================================

(defun build-factgraph-table()
  (setf *factgraph-table* (make-hash-table :test #'eq))
  (setf *fluentgraph-table* (make-hash-table :test #'eq))
  (maphash #'(lambda (key statemap)
	       (cond ((bit-vector-p statemap)
		      (let ((pred-factgraph (make-array (length statemap))))
			(dotimes (i (length statemap))
			  (setf (aref pred-factgraph i) (make-factgraph)))
			
			(setf (gethash key *factgraph-table*) pred-factgraph)))
		     (t
		      (let ((pred-fluentgraph (make-array (length statemap))))
			(dotimes (i (length statemap))
			  (setf (aref pred-fluentgraph i) (make-fluentgraph)))
			(setf (gethash key *fluentgraph-table*) pred-fluentgraph)))))
	   (problem-init-state *current-problem*)))

;; Direct access to the factgraph and fluentgraph tables
(defmacro fact-of-predpos (pred-pos)
  `(aref (gethash (car ,pred-pos) *factgraph-table*) (cdr ,pred-pos)))
(defmacro fluent-of-predpos (pred-pos)
  `(aref (gethash (car ,pred-pos) *fluentgraph-table*) (cdr ,pred-pos)))


(defun fluent-of-precfun (precfun)
  (cond ((listp (second precfun)) (fluent-of-predpos (second precfun)))
	((listp (third precfun)) (fluent-of-predpos (third precfun)))))


(defun set-factgraph ()
  (dolist (iaction *actions*)
    (dolist (iprec (gaction-preconds iaction))
      (push iaction (fg-precond-of (fact-of-predpos iprec))))
    (dolist (iadd  (gaction-adds iaction))
      (push iaction (fg-added-by (fact-of-predpos iadd))))
    (dolist (idel (gaction-dels iaction))
      (push iaction (fg-deleted-by (fact-of-predpos idel))))))
						 
  
(defun set-fluentgraph ()
   (dolist (iaction *actions*)
     (let ((n-precfun 0))
       (dolist (iprecfun (gaction-precfuns iaction))
	 (push (cons iaction n-precfun) (flg-precfun-of (fluent-of-precfun iprecfun)))
	 (incf n-precfun)
	 ))
     (dolist (icost  (gaction-costs iaction))
       (push iaction (flg-changed-by (fluent-of-predpos (second icost)))))))
      

(defun new-range-layers-table (numvar-ranges)
  (let ((layers-table (make-hash-table)))
    (dolist (i-cut numvar-ranges layers-table)
      (setf (gethash i-cut layers-table) most-positive-fixnum))))


(defun set-fluentgraph-numvars-cut()
  (maphash #'(lambda (numvar fluentgraph-map)
	       (when (find numvar (dom-real-functors *pspace*))
		 (let ((var-range (gethash numvar (problem-numvars-range *current-problem*))))
		   (dotimes (i-flg (length fluentgraph-map))
		     (let ((flg (aref fluentgraph-map i-flg)))
		       (setf (flg-range-layers flg) (new-range-layers-table var-range)))))))
	       *fluentgraph-table*))



;;======================================================================

(defun instantiate-operators ()
  (setf *actions* nil)
  (setf *total-gactions* 0)
  (create-oppred-argpos)
  (multiply-inertia-operators)
  (build-factgraph-table)
  (set-factgraph)
  (instantiate-precfuns)
  (instantiate-costs)
  (set-fluentgraph)
  (split-numvars-range)
  (set-fluentgraph-numvars-cut)
)


;;============================================================================
;; Functions for Hashing Stated for Duplicate Search
;;============================================================================


(defun total-predschemas-bits ()
  (let ((total-bits 0))
    (maphash #'(lambda (key varmap)
		 (declare (ignore key))
		 (setf total-bits (+ total-bits (length varmap))))
	     (problem-patterns *current-problem*))
    total-bits))

(defun pred-for-hashing (nbits)
  (let ((add-lists (mapcar #'action-adds (dom-actions *pspace*)))
	(pred-selected nil) (pred-maxlowerbound nil))
    (dolist (i-addlist add-lists)
      (dolist (i-add i-addlist)
	(cond ((>= (length (gethash (car i-add) (problem-patterns *current-problem*))) nbits)
	       (setf pred-selected (car i-add)))
	      (t
	       (when (or (null pred-maxlowerbound) 
			  (> (length (gethash (car i-add) (problem-patterns *current-problem*)))
			     (length (gethash pred-maxlowerbound (problem-patterns *current-problem*)))))
		      (setf pred-maxlowerbound (car i-add)))
	       
	       ))))
    (if (not (null pred-selected)) pred-selected pred-maxlowerbound)))

      
(defun create-random-hash-matcher (nbits)
  (let ((hash-matcher (make-array nbits :element-type 'bit :initial-element 0)))
    (dotimes (ibit nbits hash-matcher)
      (when (= 1 (random 2))
	(setf (aref hash-matcher ibit) 1)))))
	


(defun set-duplicate-hashing ()
  (let* ((bits-hashing (min (round (/ (total-predschemas-bits) 5)) (logcount most-positive-fixnum)))
	(pred-for-hashing (pred-for-hashing bits-hashing)))
    (setf (get (problem-plist *current-problem*) :duphash-pred) pred-for-hashing)
    (setf (get (problem-plist *current-problem*) :duphash-matcher) 
	  (create-random-hash-matcher (min bits-hashing (length (gethash pred-for-hashing (problem-patterns *current-problem*))))))
    (setf *hash-duplicates* (make-hash-table))
    ))
