;; SAYPHI Planner 
;; Dealing with Metrics in Preconditions and Effects
;; Td. 08.09.2006
;; _________________________________________________________________________________________

(defun fluent-state-value (fluent-literal state)
  (let ((pred-pos (fluent-predpos (gethash fluent-literal *fluents*))))
  (aref (gethash (car pred-pos) state) (cdr pred-pos))))



(defun math-operation-p (expression)
  (and (listp expression)
       (not (null (find (car expression) '(+ - * /))))))


(defun precfun-operand-state-eval (iprecfun state)
  (cond ((numberp iprecfun) iprecfun)
	((hkey-present (car iprecfun) state)
	 (functor-pos-state-value (car iprecfun) (cdr iprecfun) state))
	((math-operation-p iprecfun)
	 (funcall (car iprecfun) 
		  (precfun-operand-state-eval (second iprecfun) state)
		  (precfun-operand-state-eval (third iprecfun) state)))
	))


(defun increase-total-time (state)
  (let ((current-total-time (functor-state-value 'total-time 'nil state 'nil)))
    (change-fun-state state 'total-time 0 (1+ current-total-time))))



(defun apply-gaction-costs (gaction state &key (relaxed nil))
  (increase-total-time state)
  (dolist (i-cost (gaction-costs gaction))
    (let ((current-value (numvar-predpos-state-value (second i-cost) state))
	  (delta-value (third i-cost))
	  (pred-pos (second i-cost)))
      (cond ((eq 'increase (car i-cost))
	     (change-fun-state state (car pred-pos) (cdr pred-pos) (+ current-value delta-value)))
	    ((eq 'assign (car i-cost))
	     (unless (= delta-value current-value)
	       (when (or (and relaxed (> delta-value current-value))
			 (not relaxed))
		 (change-fun-state state (car pred-pos) (cdr pred-pos) delta-value))))
	    ((eq 'decrease (car i-cost))
	     (when (not relaxed)
	       (change-fun-state state (car pred-pos) (cdr pred-pos) (- current-value delta-value)))))
	    )))


;; It computes the difference of the numeric variables for determining the cost of the
;; action that produced the second state
(defun diff-num-state (state next-state)
  (let ((num-state (copy-state-set (problem-patterns *current-problem*) 'real-fluents))) 
    (maphash #'(lambda (pred statemap)
		 (unless (simple-bit-vector-p statemap)
		   (let ((costmap (gethash pred num-state))
			 (nextmap (gethash pred next-state)))
		     (when (hkey-present pred num-state)
		       (dotimes (ipos (length statemap))
			 (when (numberp (aref statemap ipos))
			   (setf (aref costmap ipos)
				 (- (aref nextmap ipos) (aref statemap ipos)))))))))
	     state)
    num-state))

					   
(defun action-cost-metric (state next-state)
  (cond ((is-metric-domain)
	 (let ((cost-state (diff-num-state state next-state))
	       (metric (problem-metric *current-problem*)))
	   (operand-funeval (cadr metric) cost-state 'nil)))
	(t 1)))



(defun map-operand-funeval (operand-list state varpos)
  (mapcar (lambda (ioperand)
	    (operand-funeval ioperand state varpos))
	  operand-list))


(defun fun-eval (ifun state varpos)
   (apply (car ifun) (map-operand-funeval (cdr ifun) state varpos)))

      
(defun operand-funeval (operand state varpos)
  (cond ((numberp operand)
	 operand)
	((hkey-present (car operand) state)
	 (functor-state-value (car operand) (cdr operand) state varpos))
	(t 
	 (fun-eval operand state varpos))))


(defun compute-total-cost (node problem)
  (let ((metric (problem-metric problem))
	(last-state (snode-state node)))
    (cond ((null metric)
	   (snode-depth node))
	  (t
	   (operand-funeval (cadr metric) last-state 'nil)))))


 



