;; SAYPHI Planner 
;; Heuristic Functions for Numeric Optimization Task
;; I put this function in here to separate a litlle the code from the standard heuristics
;; Td. 2006-09-28
;; _________________________________________________________________________________________

;; ================================================================
;; Heuristic Functions for Metric Optimization
;; ================================================================


;; SAYPHI Planner 
;; Heuristic Functions 
;; The Relaxed Planning Graph
;; Td. 2007-10-08
;; _________________________________________________________________________________________

;; ================================================================
;; Heuristic Functions for Metric Optimization
;; ================================================================

(defparameter *h-max-layers* 8)
(defparameter *h-active-facts* nil)

(defparameter *h-active-flpatterns* nil)
(defparameter *h-active-fluents* nil)

(defparameter *h-achieved-facts* nil)
(defparameter *h-achieved-fluents* nil)

(defparameter *h-active-actions* nil)
(defparameter *actions-table* nil)
(defparameter *action-layers* nil)
(defparameter *achieved-layers* nil)
(defparameter *h-achieved-layers* nil)
(defparameter *focus-goals* nil)


(defvar *h-max-distance* 1000 "stops relax-plan search")
(defvar *get-achieved-goals* nil)
(defvar *h-achieved-goals* nil)

(defstruct (plangraph 
	    (:conc-name pg-))
  (facts-mship nil)
  (numvars-mship nil)
  (goal-layers (make-hash-table))
  (numgoal-layers (make-hash-table))
  (action-layers nil)
  (achiev-goals nil)
  (achiev-numgoals nil)
  (h-achieved nil)
  (focus-goals nil)
  (relaxed-plan nil))


;; It create an array of *actions* for direct access
(defun rebuild-action-table ()
  (setf *actions-table* (make-array (length *actions*)))
  (dolist (igaction *actions*)
    (setf (aref *actions-table* (gaction-int igaction))
	  igaction)))


(defun reset-state-facts (state)
  (let ((void-state (problem-patterns *current-problem*)))
    (maphash #'(lambda (key statemap)
		 (when (bit-vector-p statemap)
		   (bit-and statemap (gethash key void-state) t)))
	     state)))

(defun reset-action-table (atable)
  (dolist (iop *actions*)
    (setf (gaction-prec-h-gcount iop) 0)
    (setf (gaction-precfun-h-gcount iop) 0)
    (setf (gaction-applicable-h iop) nil)
    
    (setf (gaction-layer iop) *h-max-layers*))

  (dotimes (ibit (length atable))
    (setf (bit atable ibit) 0)))


(defun build-fluent-active-table (state)
  (let ((active-fluents (make-hash-table :test #'eq)))
    (maphash #'(lambda (key statemap)
		 (unless (bit-vector-p statemap)
		   (when (fluent-needed key) 
		     (setf (gethash key active-fluents)
			   (make-array (length statemap) :element-type 'bit
				       :initial-element 0))
		     (dotimes (i (length statemap))
		       (when (numberp (aref statemap i))
			 (setf (sbit (gethash key active-fluents) i) 1))))))
	     state)
    active-fluents
    ))

			       
					       
		 

(defun init-heuristics ()
  (rebuild-action-table)
  (setf *h-active-facts* (copy-state-set (problem-patterns *current-problem*)))
  (setf *h-active-flpatterns* (build-fluent-active-table (problem-init-state *current-problem*)))
  (setf *focus-goals* (copy-state (problem-patterns *current-problem*)))
  (setf *h-active-actions* (make-array (length *actions-table*) 
					       :element-type 'bit 
					       :initial-element 0))
  (setf *achieved-layers* (make-hash-table))
  (setf *h-achieved-layers* (make-hash-table))
)


(defun reset-h-factgraph (state)
  (maphash #'(lambda (key factmap)
	       (dotimes (pos (length factmap))
		 (cond ((true-in-state state key pos)
			(setf (fg-layer (aref factmap pos)) 0)
			(setf (fg-activated (aref factmap pos)) t))
		       (t
			(setf (fg-layer (aref factmap pos)) *h-max-layers*)
			(setf (fg-activated (aref factmap pos)) nil)))))
	   *factgraph-table*))


(defun reset-h-fluentgraph (state)
  (maphash #'(lambda (key fluentmap)
	       (dotimes (pos (length fluentmap))
		 (when (hash-table-p (flg-range-layers (aref fluentmap pos)))
		   (let ((state-value (numvar-state-value key state pos)))
		     (maphash #'(lambda (range layer)
				  (declare (ignore layer))
				  (cond ((> state-value range)
					 (setf (gethash range (flg-range-layers (aref fluentmap pos))) 0)
					 (setf (flg-activated (aref fluentmap pos)) t))
					(t
					 (setf (gethash range (flg-range-layers (aref fluentmap pos))) *h-max-layers*)
					 (setf (flg-activated (aref fluentmap pos)) nil))))
			      (flg-range-layers (aref fluentmap pos)))
		     ))))
	   *fluentgraph-table*))



(defun reset-plangraph (state)
  (setf *h-active-facts* (copy-state-set state 'facts))
  (setf *h-achieved-facts* (copy-state-set state 'facts))
  (setf *h-achieved-fluents* (copy-state-set state 'real-fluents))
  (setf *h-active-fluents* (copy-state *h-active-flpatterns*))
  (reset-h-factgraph state)
  (reset-h-fluentgraph state)
  (reset-state-facts *focus-goals*)
  (reset-action-table *h-active-actions*)
 
  (setf *action-layers* (make-hash-table))
)


(defmacro num-layers-graph ()
  `(hash-table-count *action-layers*))

(defun fluent-needed (fluent-pred)
  (when (or (find fluent-pred (dom-positive-vars *pspace*))
	    (find fluent-pred (dom-artificial-vars *pspace*))) t))

(defun max-needed (pred-pos)
   (apply #'max (gethash (car pred-pos) (problem-numvars-range *current-problem*))))


;; It returns the corresponding cost of a given numvar (pred-pos) looking through action costs
(defmacro cost-of-numvar-gaction (gaction pred-pos)
  `(find ,pred-pos (gaction-costs ,gaction) :test #'equal :key #'second))


(defun connect-fact-h (pred-pos layer)
  (let ((fg (fact-of-predpos pred-pos)))
   (dolist (iop (fg-precond-of fg))
      (unless (gaction-applicable-h iop)
	(incf (gaction-prec-h-gcount iop))
	(when (and (= (gaction-num-precs iop) (gaction-prec-h-gcount iop))
 		   (= (gaction-num-precfuns iop) (gaction-precfun-h-gcount iop)))
	  (setf (gaction-layer iop) layer)
	  (setf (gaction-applicable-h iop) t)
	  (setf (bit *h-active-actions* (gaction-int iop)) 1))
	
  ))))

(defun precfun-value-eval (precfun value)
  (> (funcall (car precfun)
	      (if (listp (second precfun)) value (second precfun))
	      (if (listp (third precfun)) value (third precfun))) 0)) 


;; It recieves the numeric goal in the form (*op num numvar) or (*op numvar num)
;; both equivalent to the (> numgoal 0). It return the value of numvar that makes true the expression
(defun solve-numtask-exp (expression)			  
  (cond ((numberp (second expression))
	 (split-cut-point (car expression) (second expression)))
	((numberp (third expression))
	 (split-cut-point (car expression) (third expression)))))


(defun precfun-old-achieved (precfun layer)
  (let ((flg (fluent-of-precfun precfun)))
    (< (gethash (solve-numtask-exp precfun) (flg-range-layers flg)) layer)))
       
    
(defun set-fluent-achiev-layer (precfun layer)
  (let ((flg (fluent-of-precfun precfun)))
    (setf (gethash (solve-numtask-exp precfun) (flg-range-layers flg)) layer)))


(defun connect-fluent-h (pred-pos layer state-value)
  (let ((flg (fluent-of-predpos pred-pos)))
    (dolist (iop-precfun-n (flg-precfun-of flg))
      (let* ((iop (car iop-precfun-n))
	     (precfun (nth (cdr iop-precfun-n) (gaction-precfuns iop))))
	(unless (gaction-applicable-h iop)
	  (when (precfun-value-eval precfun state-value)
	    (unless (precfun-old-achieved precfun (1- layer))
	      (incf (gaction-precfun-h-gcount iop))
	      (when (and (= (gaction-num-precs iop) (gaction-prec-h-gcount iop))
			 (= (gaction-num-precfuns iop) (gaction-precfun-h-gcount iop)))
		(set-fluent-achiev-layer precfun layer)
		(setf (gaction-layer iop) layer)
		(setf (gaction-applicable-h iop) t)
		(setf (bit *h-active-actions* (gaction-int iop)) 1))
	      
)))))))



(defmacro mark-action-graph (index layer)
  `(set-instance-map (gethash ,layer *action-layers*) ,index 1))


(defun update-numvars-level-achieved (pred-pos layer new-value)
  (let ((flg (fluent-of-predpos pred-pos)))
    (maphash #'(lambda (point rlayer)
		 (when (and (>= new-value point)
			    (< layer rlayer))
		   (setf (gethash point (flg-range-layers flg)) layer)))
	     (flg-range-layers flg))))
	   

(defun apply-gaction-relaxed-costs (gaction h-state layer)
  (dolist (i-cost (gaction-costs gaction))
    (when (fluent-needed (car (second i-cost)))
    (let ((current-value (numvar-predpos-state-value (second i-cost) h-state))
	  (delta-value (third i-cost))
	  (pred-pos (second i-cost)))
      (when (< current-value (max-needed pred-pos))
	(cond ((eq 'increase (car i-cost))
	       (unless (= delta-value 0)
		 (change-fun-state h-state (car pred-pos) (cdr pred-pos) (+ current-value delta-value))
		 (update-numvars-level-achieved pred-pos layer (+ current-value delta-value))))
	      ((eq 'assign (car i-cost))
	       (unless (= delta-value current-value)
		 (change-fun-state h-state (car pred-pos) (cdr pred-pos) delta-value)
		 (update-numvars-level-achieved pred-pos layer delta-value)))
	      ))))))
  

(defun connect-action-h (index layer)
  (let ((gaction (aref *actions-table* index)))
    (mark-action-graph index layer)
    (dolist (iadd (gaction-adds gaction))
      (let ((fg (fact-of-predpos iadd)))
	(unless (fg-activated fg)
	  (change-state *h-achieved-facts* (car iadd) (cdr iadd) 1)
	  (change-state *h-active-facts* (car iadd) (cdr iadd) 1)
	  (setf (fg-layer fg) layer)
	  (setf (fg-activated fg) t))))
    (apply-gaction-relaxed-costs gaction *h-achieved-fluents* layer) 
))



;;deciding if graphplan expansion should be stopped
;;By Reaching the goals of finding a fixpoint
(defun expand-stop (goals layer)
  (cond ((>= layer *h-max-layers*) t)
	(t (goals-reached *h-achieved-facts* goals))))
  



(defun expand-relaxedplangraph (state goals)
  (reset-plangraph state)
  (do ((layer 1 (incf layer)))
      ((expand-stop goals layer) (< layer *h-max-layers*))
    
    ;;Activating Facts
    (maphash #'(lambda (key statemap)
		 (dotimes (ibit (length statemap))
		   (when (= 1 (bit statemap ibit))
		     (connect-fact-h (cons key ibit) layer)
 		     (setf (bit statemap ibit) 0))))
	     *h-active-facts*)
    
    ;;Activating Fluents
    (maphash #'(lambda (key statemap)
		 (dotimes (ibit (length statemap))
		   (when (= 1 (bit statemap ibit))
		     (let ((state-value (numvar-state-value key *h-achieved-fluents* ibit)))
		       (connect-fluent-h (cons key ibit) layer state-value)
		       (when (> state-value (max-needed (cons key ibit)))
			 (setf (bit statemap ibit) 0))))))
	     *h-active-fluents*)
    
    (setf (gethash layer *action-layers*) 0)
    (dotimes (jbit (length *h-active-actions*))
      (when (= 1 (bit *h-active-actions* jbit))
	(connect-action-h jbit layer)
	(setf (bit *h-active-actions* jbit) 0)))
    
    ;;Reapplying action cost 
    (dotimes (k-layer (1- layer))
      (let ((pastlayer-actions (gethash k-layer *action-layers*)))
	(when (integerp pastlayer-actions)
	  (dotimes (ibit (integer-length pastlayer-actions))
	    (when (logbitp ibit pastlayer-actions)
	      (apply-gaction-relaxed-costs (aref *actions-table* ibit) *h-achieved-fluents* layer))))))
    ))
    

;; (defmacro fact-is-member (pred-pos layer fact-membership)
;;   `(= ,layer (aref (gethash (car ,pred-pos) ,fact-membership) (cdr ,pred-pos))))
(defmacro fact-is-member (pred-pos layer)
  `(= ,layer (fg-layer (aref (gethash (car ,pred-pos) *factgraph-table*) (cdr ,pred-pos)))))


;; Returns the number of the layer in which a literal is present
;; (defmacro layer-membership (pred-pos fact-membership)
;;   `(aref (gethash (car ,pred-pos) ,fact-membership) (cdr ,pred-pos)))
(defmacro layer-membership (pred-pos)
   `(fg-layer (aref (gethash (car ,pred-pos) *factgraph-table*) (cdr ,pred-pos))))


(defmacro achieved-in-graph (layer achieved-graph pred-pos)
  `(true-in-state (gethash ,layer ,achieved-graph) (car ,pred-pos) (cdr ,pred-pos)))



;; numgoal recibida como (> (FUEL . 0) 100)
(defmacro statevar-is-true (numgoal layer numvars-membership)
  `(= ,layer (layer-from-nummembership ,numgoal ,numvars-membership)))



;; It applies the corresponding action to a state of achieved goal layer
(defun apply-to-achiev (action achiev-goals)
  (apply-inst-effects achiev-goals (gaction-adds action) 1)
  (when (is-metric-domain)
    (apply-gaction-costs action achiev-goals :relaxed t)))



(defun apply-relaxed-plan (relaxed-plan)
  (let* ((applied-state (copy-state (problem-patterns *current-problem*))))
    (dolist (i-gaction relaxed-plan applied-state)
      (apply-to-achiev i-gaction applied-state))))


(defun goal-functor (numgoal)
  (cond ((numberp (second numgoal))
	 (third numgoal))
	((numberp (third numgoal))
	 (second numgoal))))


(defun instaction-difficulty (iaction)
  (let ((difficulty 0))
    (dolist (iprec (gaction-preconds iaction) difficulty)
      (setf difficulty (+ difficulty (layer-membership iprec))))))


(defmacro achievers (pred-pos)
   `(fg-added-by (aref (gethash (car, pred-pos) *factgraph-table*) (cdr ,pred-pos))))

(defun num-achievers (pred-pos)
  (remove-if #'(lambda (achiever)
		 (find 'decrease (gaction-costs achiever) :key #'car))
	     (flg-changed-by (aref (gethash (car pred-pos) *fluentgraph-table*) (cdr pred-pos)))))


(defmacro action-applied-inlayer (gaction layer)
   `(logbitp (gaction-int ,gaction) (gethash ,layer *action-layers*)))


(defun select-rplan-action (goal n-layer)
  (let ((current-relevant-action nil))
    (dolist (i-achiever (achievers goal) current-relevant-action)
      (when (action-applied-inlayer i-achiever n-layer)
	(let* ((this-diff (setf (gaction-h-difficulty i-achiever)
				(instaction-difficulty i-achiever))))
	  
	  (when (or (null current-relevant-action)
		    (< this-diff (gaction-h-difficulty current-relevant-action)))
	    (setf current-relevant-action i-achiever) 
	    ))))))


(defun remaining-numgoal (numgoal needed last-achieved)
  (when (> needed last-achieved)
    (cond ((numberp (second numgoal))
	   (list (car numgoal) (- needed last-achieved) (third numgoal)))
	  ((numberp (third numgoal))
	   (list (car numgoal) (second numgoal) (- needed last-achieved))))))


;;It returns the list of actions selected in the action-layer for achieving a numeric goal
;; together with a remaining numeric goal if it is not completely satisfied
(defun select-rplan-numeric-actions (goal n-layer)
  (let* ((numvar (if (numberp (second goal)) (third goal) (second goal)))
	 (achieved-in-loop (numvar-predpos-state-value numvar (gethash n-layer *achieved-layers*)))
	 (needed-layer-value (solve-numtask-exp goal))
	 (relevant-actions nil))
    (dolist (i-achiever (num-achievers numvar) (cons relevant-actions 
						     (remaining-numgoal goal needed-layer-value achieved-in-loop)))
      (when (and (< achieved-in-loop needed-layer-value) 
		 (action-applied-inlayer i-achiever n-layer))
	(let ((changing-cost (find numvar (gaction-costs i-achiever) :test #'equal :key #'second))) 
	  (push i-achiever relevant-actions)
	  (setf achieved-in-loop (cond ((eq 'increase (car changing-cost))
					(+ achieved-in-loop (third changing-cost)))
				       ((and (eq 'assign (car changing-cost))
					     (> (third changing-cost) achieved-in-loop))
					(third changing-cost)))))))))


;; The expression of numgoal is true when (> expression 0)
(defun achieved-in-numeric-graph (layer achiev-graph numgoal)
  (let ((goal-layer (gethash layer achiev-graph)))
    (cond ((numberp (second numgoal))
	   (> (funcall (car numgoal) (second numgoal) (numvar-state-value 
						       (car (third numgoal)) goal-layer (cdr (third numgoal))))
	      0))
	  ((numberp (third numgoal))
	  (> (funcall (car numgoal) (numvar-state-value (car (second numgoal)) goal-layer (cdr (second numgoal)))
		      (third numgoal))
	      0))
	  )))



(defun layer-num-achieved (precfun)
  (let ((flg (fluent-of-precfun precfun)))
    (gethash (solve-numtask-exp precfun) (flg-range-layers flg))))
    
		 
;; It takes the sub-goals from actions preconditions and push them in thier corresponding
;; layer of goals to achiev
(defun insert-prec-subgoals (action nlayer plangraph)
  (dolist (iprec-goal (gaction-preconds action))
    (unless (or (fact-is-member iprec-goal 0)
		(achieved-in-graph (1- nlayer) *achieved-layers* iprec-goal))
      (let ((layer (layer-membership iprec-goal)))
	(push iprec-goal (gethash layer (pg-goal-layers plangraph)))))))


(defun insert-numprec-subgoals (action nlayer plangraph)
  (dolist (iprec-numgoal (gaction-precfuns action))
    (cond ((listp iprec-numgoal)
	  (unless (or (= 0 (layer-num-achieved iprec-numgoal))
		      (achieved-in-numeric-graph (1- nlayer) *achieved-layers* iprec-numgoal))
	    (let ((layer (layer-num-achieved iprec-numgoal)))
	      (push iprec-numgoal (gethash layer (pg-numgoal-layers plangraph))))))
	  ((and (numberp iprec-numgoal) 
		(< iprec-numgoal 0))
	   (format t "Say-Warning Unsolvable numeric goal!")))))



			 
(defun create-achieved-membership (num-layers achieved-layers)
  (let ((ifact-layer nil))
    (dotimes (i-layer (+ num-layers 1) achieved-layers)
      (setf ifact-layer (gethash i-layer achieved-layers))
      (cond ((null ifact-layer)
	     (setf (gethash i-layer achieved-layers)
		   (copy-state (problem-patterns *current-problem*))))
	    (t (reset-state-facts ifact-layer))))))


(defun h-difficult-fraction (relevant-actions)
  (let ((num-actions (length relevant-actions))
	(difficult-sum (apply #'+ (mapcar #'cadr relevant-actions))))
    (if (> difficult-sum 0)
	(+ num-actions (/ (- 1 (float (/ 1 difficult-sum))) 2))
      num-actions)))


(defun h-pondered-difficulty (relevant-actions)
  (apply #'+ (mapcar #'(lambda (i-relevantaction)
			 (let ((i-diff (cadr i-relevantaction)))
			   (float (/ (+ 1 i-diff) (+ 2 i-diff)))))
		     relevant-actions)))


(defun update-numvars-membership (numvars-membership new-value pred-fun pos ilayer)
  (let* ((numvar-layer-table (gethash pred-fun numvars-membership)))
    (when (hash-table-p numvar-layer-table)
      (maphash (lambda (point layer-array)
		 (when (and (>= new-value point)
			    (< ilayer (aref layer-array pos)))
		   (setf (aref layer-array pos) ilayer)))
	     numvar-layer-table))))



;; Statements to do with the select action for the relaxded plan
;; This is done for literals and numeric subgoals
(defun with-selected-gpaction (selected-rplan-action nlayer pg focus-layer)
  (push selected-rplan-action (pg-relaxed-plan pg))
  (insert-prec-subgoals selected-rplan-action nlayer pg)
  (insert-numprec-subgoals selected-rplan-action nlayer pg)
   
  (apply-to-achiev selected-rplan-action (gethash nlayer *achieved-layers*))
  (apply-to-achiev selected-rplan-action (gethash (1- nlayer) *achieved-layers*))
  (when (= focus-layer nlayer)
      (apply-to-achiev selected-rplan-action *focus-goals*))
  (when *get-achieved-goals*
      (apply-to-achiev selected-rplan-action (gethash nlayer *h-achieved-layers*))))



(defun init-plangraph (node goals) 
  (when (expand-relaxedplangraph (snode-state node) goals)
    (create-achieved-membership (num-layers-graph) *achieved-layers*)
    (make-plangraph)
     
     ))


(defun find-helpful-byte (focus-goals)
  (let ((helpful-byte 0))
    (maphash #'(lambda (pred statemap)
		 (cond ((simple-bit-vector-p statemap)
			(dotimes (ibit (length statemap))
			  (when (= 1 (sbit statemap ibit))
			    (dolist (i-achiever (fg-added-by (fact-of-predpos (cons pred ibit))))
			      (when (= 1 (gaction-layer i-achiever))
				(set-instance-map helpful-byte (gaction-int i-achiever) 1))))))
		       (t
			(dotimes (ipos (length statemap))
			  (when (and (numberp (aref statemap ipos))
				     (> (aref statemap ipos) 0))
			    (dolist (i-achiever (num-achievers (cons pred ipos)))
			      (when (and (= 1 (gaction-layer i-achiever))
					 (>= (third (cost-of-numvar-gaction i-achiever (cons pred ipos)))
					     (aref statemap ipos)))
				(set-instance-map helpful-byte (gaction-int i-achiever) 1))))))))
	     
	     focus-goals)
    helpful-byte))


(defun metric-relaxedplan (node goals &key (focus-layer 1))
  (let ((inst-goals (get-instantiated-goals goals))
	(pg (init-plangraph node goals)))
    (when (plangraph-p pg)
      (when *get-achieved-goals*
	(setf (pg-h-achieved pg) (create-achieved-membership (num-layers-graph) *h-achieved-layers*))
 	(setf *h-achieved-goals* *h-achieved-layers*))
      
      (dolist (igoal inst-goals)
	(let ((layer (layer-membership igoal)))
	  (push igoal (gethash layer (pg-goal-layers pg)))))

      (dotimes (i (num-layers-graph))
	(let ((i-layer (- (num-layers-graph) i)))
	  (dolist (igoal (gethash i-layer (pg-goal-layers pg)))
	    (unless (achieved-in-graph i-layer *achieved-layers* igoal)
	      (with-selected-gpaction (select-rplan-action igoal i-layer)
		i-layer pg focus-layer)))
 	  (dolist (inumgoal (gethash i-layer (pg-numgoal-layers pg)))
 	    (unless (achieved-in-numeric-graph i-layer *achieved-layers* inumgoal)
	      (let ((solving-numlayer (select-rplan-numeric-actions inumgoal i-layer)))
		(dolist (i-selected (car solving-numlayer))
		  (with-selected-gpaction i-selected i-layer pg focus-layer))
		(when (cdr solving-numlayer)
		  (push inumgoal (gethash (1- i-layer) (pg-numgoal-layers pg))))
		)))))

	      
      (setf (snode-helpful-byte node) (find-helpful-byte *focus-goals*))
      (setf (snode-applicable-byte node) (gethash 1 *action-layers*))
      pg
	  )))




(defun h-metric-rxplan (node &optional (goals (problem-goals *current-problem*)))
  (let* ((pg (metric-relaxedplan node goals))
	 (relaxed-plan nil) (applied-rp-state nil) (h-metric nil))
    (cond ((plangraph-p pg)
	   (setf relaxed-plan (pg-relaxed-plan pg))
	   
	   (cond ((not (null (problem-metric *current-problem*)))
		  (setf applied-rp-state (apply-relaxed-plan relaxed-plan))
		  (setf h-metric (operand-funeval (cadr (problem-metric *current-problem*))
						  applied-rp-state 'nil)))
		 (t
		  (setf h-metric (length relaxed-plan))))
	   (values h-metric relaxed-plan ))
	  (t (values most-positive-fixnum nil nil)))))


;; @@ Cambiar en las demas heurísticas
(defun h-relaxedplan (node &optional (goals (problem-goals *current-problem*)))
  (let* ((pg (metric-relaxedplan node goals)))
    (cond ((plangraph-p pg)
	   (values (length (pg-relaxed-plan pg)) 
		   (pg-relaxed-plan pg)))
	  (t (values most-positive-fixnum nil nil)))))


(defun h-relaxedplan-plus (node &optional (problem *current-problem*))
  (let* ((goals (problem-goals problem))
	 (pg (metric-relaxedplan node goals)))
    (cond ((plangraph-p pg)
	   (values (h-difficult-fraction (pg-relaxed-plan pg)) 
		   (pg-relaxed-plan pg) 
		   ))
	  (t (values most-positive-fixnum nil nil)))))



;;=============================================================
;;Extras for Printing and Developing
;;=============================================================
;; (defun dev-fg-showvar (pred-pos var)
;;   (let ((fg (fact-of-predpos

(defun gaction-fullprint (ga)
  (format t "~% <Gaction>: ~a " (gaction-planaction ga))
  (format t "~%   precs: ~a " (gaction-preconds ga))
  (format t "~%   adds: ~a " (gaction-preconds ga))
  (format t "~%   dels: ~a " (gaction-preconds ga))
  (format t "~%   prec-h-gcount: ~a " (gaction-prec-h-gcount ga)))
