;; A tentative inital parser.  I don't want to waste time on this first
;; Based on pddl2ipps translator functions in IPSS
;; Td. 04.12.2005


(defvar *say-parse-errors* nil)

(defmacro negative-var-p (var)
  `(when (listp ,var) 
    (when (find (car ,var) (dom-negative-vars *pspace*)) t)))

(defmacro functor-p (var)
  `(when (and (listp ,var) (hkey-present (car ,var) (dom-functors *pspace*))) t))


(defmacro find-argument (pddl-def argument)
  `(cdr (assoc ,argument ,pddl-def)))


(defmacro negate-var (functor)
  `(intern (format nil "_~a" (symbol-name ,functor))))


(defun find-all-argument (pddl-def argument)
  (remove-if-not #'(lambda (definition)
		     (eq (car definition) argument))
		 pddl-def))

(defun show-parse-errors ()
  (dolist (ierr *say-parse-errors*)
    (format t "~% Parse Error: ~a" ierr)))


(defun convert-pddl-var (pddl-var)
  (intern (string-left-trim '(#\?) (format nil "~a" pddl-var))))


;; It checks if an argument of the literal is a cons of var-param or a domain constant
(defun literal-arg-p (arg arg-var)
  (cond ((consp arg))
	(t 
	 (cond ((find arg (gethash (cdr arg-var) (dom-constants *pspace*))) t)
	       (t (push (format nil "! ~a is not a domain constant" arg) *say-parse-errors*)))
	       )))
	


(defun convert-literal (literal parameters &optional (litype :predicate))
  (let ((litschema (cond ((eq litype :predicate) (dom-predicates *pspace*))
			 ((eq litype :functors) (dom-functors *pspace*))))
	(literal-args (mapcar #'(lambda (var)
				  (let ((say-var (convert-pddl-var var)))
				    (cond ((assoc say-var parameters))
					  (t var))))
			      (cdr literal))))
    (cond ((hkey-present (car literal) litschema)
	   (when (every #'literal-arg-p literal-args (gethash (car literal) litschema)) 
	     (cons (car literal) literal-args)))
	  (t (push (format nil " - ~a - is not a ~a in the domain" (car literal) litype) *say-parse-errors*) 
	     nil)
	  )))


(defun get-inherit-types (dtype types-def)
  (let ((dtype-def (car (find-argument types-def dtype)))
	(result-types (list dtype)))
      (dolist (this-type dtype-def result-types)
	(if (find-argument types-def this-type)
	    (setf result-types (append result-types 
				       (get-inherit-types this-type types-def)))
	    (setf result-types (append result-types (list this-type)))))))


(defun buildall-inherit-types (types-def)
  (let ((inherit-types nil))
    (dolist (itype types-def inherit-types)
      (when (listp itype)
	(push (list (car itype) (get-inherit-types (car itype) types-def)) inherit-types)))))



(defun translate-types-def (types)
  (let ((read-types nil) (sub-types nil) 
	(top-types (remove '- types)))
  (do* ((prev-type (car types) type)
	(rest-types (cdr types) (cdr rest-types))
	(type (car rest-types) (car rest-types)))
      ((null rest-types) nil)
    (cond ((eq prev-type '-)
	   (push (cons type (list (reverse read-types))) sub-types)
	   (setf read-types nil)
	   (setf rest-types (cdr rest-types))
	   (setf type (car rest-types)))
	  (t
	   (push prev-type read-types)
	   (setf top-types (remove prev-type top-types)))))
  (if sub-types
      (reverse sub-types)
    (append top-types read-types))
))


;; Like translate-types-def. In this case some definitions could be repeat, so i put them in a 
;; hash table directly
(defun translate-objects-def (objects-def)
  (let ((ht-objects (make-hash-table :test #'eq))
	(read-objects nil))
  (do* ((prev-object (car objects-def) object)
	(rest-objects (cdr objects-def) (cdr rest-objects))
	(object (car rest-objects) (car rest-objects)))
      ((null rest-objects) ht-objects)
    (cond ((eq prev-object '-)
	   (setf (gethash object ht-objects)
		 (append (gethash object ht-objects) (reverse read-objects)))
	   (setf read-objects nil)
	   (setf rest-objects (cdr rest-objects))
	   (setf object (car rest-objects)))
	  (t
	   (push prev-object read-objects))))))


(defun compute-params (params-def)
  (let ((translated-params (translate-types-def params-def))
	(params-types nil))
    (dolist (i-param translated-params (nreverse params-types))
      (dolist (i-var (cadr i-param))
	(push (cons (convert-pddl-var i-var) (car i-param))
	      params-types)))))


(defun parse-predicates (preds-def)
  (let ((thash (make-hash-table :test #'eq)))
    (mapc #'(lambda (ipred) 
	      (setf (gethash (car ipred) thash) (compute-params (cdr ipred))))
	  preds-def)
    thash))


(defun convert-function (function parameters functors)
  (mapcar #'(lambda (formula) (cond ((or (atom formula) (numberp formula))
				     formula)
				    ((gethash (car formula) functors)
				     (convert-literal formula parameters :functors))
				    (t
				     (convert-function formula parameters functors))))
	  function))


(defun parse-preconds (preconds this-action predicates functors)
  (cond ((null preconds) nil)
	((or (and (eq (car preconds) 'over)
		  (eq (cadr preconds) 'all))
	     (and (eq (car preconds) 'at)
		  (or (eq (cadr preconds) 'start)
		      (eq (cadr preconds) 'end))))
	 (parse-preconds (caddr preconds) this-action predicates functors))
	
	((member (car preconds) '(and or))
	  (dolist (iprecond (cdr preconds))
	    (parse-preconds iprecond this-action predicates functors)))

	((hkey-present (car preconds) predicates)
	 (push (convert-literal preconds (action-parameters this-action))
	       (action-preconditions this-action)))
	
	(t
	 (push (convert-function preconds (action-parameters this-action) functors)
	       (action-precond-funs this-action)))
))



(defun parse-effects (effects this-action predicates functors)
  (cond ((null effects) nil)
	((or (and (eq (car effects) 'over)
		  (eq (cadr effects) 'all))
	     (and (eq (car effects) 'at)
		  (or (eq (cadr effects) 'start)
		      (eq (cadr effects) 'end))))
	 (parse-effects (caddr effects) this-action predicates functors))

	((member (car effects) '(and or))
	 (dolist (ieffect (cdr effects))
	   (parse-effects ieffect this-action predicates functors)))
			 
	((eq (car effects) 'not)
	 (if (hkey-present (caadr effects) predicates)
	     (push (convert-literal (cadr effects) (action-parameters this-action))
		   (action-dels this-action))))

	((hkey-present (car effects) predicates)
	 (push (convert-literal effects (action-parameters this-action))
		   (action-adds this-action)))
	
	((member (car effects) '(increase decrease assign))
	 (push (convert-function effects (action-parameters this-action) functors)
	       (action-costs this-action)))
	
	(t nil)
))


(defun build-action (action-def predicates functors this-action)
    (let* ((params (compute-params (getf action-def :parameters)))
	   (preconds-def (append (getf action-def :precondition)
				 (getf action-def :condition)))
	   (effects-def (getf action-def :effect)))
      (setf (action-parameters this-action) params)
      (parse-preconds preconds-def this-action predicates functors)
      (parse-effects effects-def this-action predicates functors))
    this-action)



(defun parse-actions (actions-def predicates functors )
  (let ((this-action nil))
    (dolist (iaction (reverse actions-def))
      (setf this-action (make-action :name (getf iaction :action)))
      (push (build-action iaction predicates functors this-action) (dom-actions *pspace*))
)))


;; ==================================================================
;; Additional parse-domain functions to deal with metrics
;; I put this in here rather than metric file because is easier 
;; to understand what i am doing here

(defun set-metric-total-time()
  (when (and (find :fluents (dom-requirements *pspace*))
	     (not (hkey-present 'total-time (dom-functors *pspace*))))
    (setf (gethash 'total-time (dom-functors *pspace*)) nil)))


(defun get-real-functors ()
  (let ((functors (dom-functors *pspace*))
	(real-functors nil))
    (dolist (iaction (dom-actions *pspace*) real-functors)
      (dolist (icost (action-costs iaction))
	(when (hkey-present (car (second icost)) functors)
	  (unless (find (car (second icost)) real-functors)
	    (push (car (second icost)) real-functors)))))))


;; Calculate positive or negative vars for a generic precond-function
(defun extract-signed-vars (precfun sign functors)
  (cond ((numberp precfun)
	 nil)
	((hkey-present (car precfun) functors)
	 (unless sign
	   (list (car precfun))))
	((eq '- (car precfun))
	 (append (extract-signed-vars (second precfun) sign functors)
		 (extract-signed-vars (third precfun) (not sign) functors)))
	(t
	 (append (extract-signed-vars (second precfun) sign functors)
		 (extract-signed-vars (third precfun) sign functors)))))

;; The step 1 changes (exp <[<=] exp')
(defun transform-lnf-step1 (prec-fun)
  (cond ((eq '< (car prec-fun))
	 (list '> (third prec-fun) (second prec-fun)))
	((eq '<= (car prec-fun)) 
	 (list '>= (third prec-fun) (second prec-fun)))
	(t prec-fun)))


;; The step 2 changes (exp >[>=] exp') to (exp - exp' >[>=] 0)
(defun transform-lnf-step2 (prec-fun)
  (cond ((or (eq '> (car prec-fun))
	     (eq '>= (car prec-fun)))
	 (list (car prec-fun) (list '- (second prec-fun) (third prec-fun)) 0))
	(t prec-fun)))


;; This set of functions translate preconditions-functions to de LNF (linear function)
;; in their normal-form explained by Hoffman.
(defun translate-precfuns()
  (let ((actions (dom-actions *pspace*)))
    (dolist (iaction actions)
      (unless (null (action-precond-funs iaction))
	(setf (action-precond-funs iaction)
	      (mapcar #'(lambda (i-precfn)
			  (transform-lnf-step2 (transform-lnf-step1 i-precfn)))
		      (action-precond-funs iaction)))))))




;; Calculates the positive and negative vars for relaxing the planning graph with costs
;; This should be made after the LNF transformation
(defun separate-functors-torelax ()
  (let ((positive-vars nil) 
	(negative-vars nil)
	(functors (dom-functors *pspace*))
	(real-functors (get-real-functors))) 
    (dolist (iaction (dom-actions *pspace*))
      (dolist (i-precfun (action-precond-funs iaction))
	(setf positive-vars 
	      (union positive-vars (extract-signed-vars (second i-precfun) 'nil functors)))
	(setf negative-vars 
	      (union negative-vars (extract-signed-vars (second i-precfun) 't functors)))))
    (setf (dom-real-functors *pspace*) real-functors)
    (setf (dom-positive-vars *pspace*)
	  (mapcan #'(lambda (x) 
		      (when 
			  (find x real-functors) (list x))) 
		  positive-vars))
    (setf (dom-negative-vars *pspace*)
	  (mapcan #'(lambda (x) 
		      (when (find x real-functors) 
			(list x))) 
		  negative-vars))))


;; It takes negative vars and introduce artificial variables to hold the opposite values
;; in the state for the relaxing phase
(defun create-artificial-vars()
  (dolist (ivar (dom-negative-vars *pspace*))
    (let ((inegvar (negate-var ivar)))
      (push inegvar (dom-artificial-vars *pspace*))
      (setf (gethash inegvar (dom-functors *pspace*)) (gethash ivar (dom-functors *pspace*)))
      )))
   

(defun change-to-negvar (expression)
  (cond ((find (car expression) (dom-negative-vars *pspace*))
	 (cons (negate-var (car expression)) (cdr expression)))
	(t (list '* -1 expression))))


(defun artificial-fun-exp (expression sign-exp)
  (cond ((numberp expression)
	 (* expression (if sign-exp 1 -1)))
	((functor-p expression)
	 (cond ((not sign-exp)
		(cond ((negative-var-p expression)
		       (cons (negate-var (car expression)) (cdr expression)))
		      (t (list '* -1 expression))))
	       (t expression)))
	((eq (car expression) '-)
	 (list '+ (artificial-fun-exp (second expression) sign-exp) 
	       (artificial-fun-exp (third expression) (not sign-exp))))
	(t
	 (list (car expression) (artificial-fun-exp (second expression) sign-exp) 
	       (artificial-fun-exp (third expression) t)))))


(defun update-for-artificial-preconds()
  (dolist (iaction (dom-actions *pspace*))
    (dolist (iprecfun (action-precond-funs iaction))
      (setf (second iprecfun) (artificial-fun-exp (second iprecfun) t)))))


(defun insert-artificial-costs()
  (let ((temp-costs nil))
    (dolist (iaction (dom-actions *pspace*))
      (dolist (icost (action-costs iaction))
	(when (find (car (second icost)) (dom-negative-vars *pspace*))
	  (cond ((eq 'increase (car icost))
		 (push (list 'decrease (change-to-negvar (second icost)) (third icost)) temp-costs))
		((eq 'decrease (car icost))
		 (push (list 'increase (change-to-negvar (second icost)) (third icost)) temp-costs))
		((eq 'assign (car icost))
		 (push (list 'assign (change-to-negvar (second icost)) (list '* -1  (third icost))) temp-costs)))
	  (unless (find (negate-var (car (second icost))) (dom-real-functors *pspace*))
	    (push (negate-var (car (second icost))) (dom-real-functors *pspace*)))))
      (setf (action-costs iaction)
	    (append (action-costs iaction) temp-costs))
      (setf temp-costs nil))))
			      
		      

;; ==================================================================


;; The domain structure is assigned to the global variable *pspace*
(defun parse-domain (arguments)
  (setf *say-parse-errors* nil)
  (let* ((domain (find-argument arguments 'domain))
	 (requirements (find-argument arguments :requirements))
	 (dtypes (translate-types-def (find-argument arguments :types)))
	 (inheritypes (buildall-inherit-types dtypes))
	 (predicates (parse-predicates (find-argument arguments :predicates)))
	 (functors (parse-predicates (find-argument arguments :functions)))
	 (constants (translate-objects-def (find-argument arguments :constants)))
	 (actions-def (append (find-all-argument arguments :action)
			  (find-all-argument arguments :durative-action)))
	 )
    (setf *pspace* (make-dom :name domain
			     :requirements requirements
			     :domtypes dtypes
			     :inheritypes inheritypes
			     :constants constants
			     :predicates predicates
			     :functors functors
                             ))
;;     (parse-operators actions-def)
    (parse-actions actions-def predicates functors)

    (create-predicate-table)
    (collect-top-inertia)
    (translate-precfuns)
    (set-metric-total-time)
    (separate-functors-torelax)
    (create-artificial-vars)
    (update-for-artificial-preconds)
    (insert-artificial-costs)
))


(defun read-pddl-domain (file)
    (with-open-file (istream file :direction :input)
	(let ((pddl-def (read istream)))
	  (if (listp pddl-def)
	      (cond ((eq 'domain (car (second pddl-def)))
		     (parse-domain (cdr pddl-def))
		     (unless (null *say-parse-errors*)
		       (setf *pspace* nil)
		       (show-parse-errors))
		     ))))))


(defun read-file (file)
  (with-open-file (istream file :direction :input)
    (read istream)))