;; SAYPHI Planner 
;; Problem Parser
;; Td. 04.12.2005
;; ________________________________________________________________________________________
;; 

(defparameter *facts* nil)
(defparameter *init-facts* nil)
(defparameter *fluents* nil)
(defparameter *init-fluents* nil)


;; This function creates real instances of objects for domain constants defined in domain file
(defun update-constant-objects (objects)
  (maphash #'(lambda (type constants)
	       (setf (gethash type objects)
		     (remove-duplicates (append (gethash type objects) constants))))
	   (dom-constants *pspace*)))


(defun get-inheritobjects (abs-type inheritypes objects)
  (let ((inheritobjects nil)
	(subtypes (get-vals abs-type inheritypes)))
    (dolist (isubtype subtypes inheritobjects)
      (setf inheritobjects (append inheritobjects (gethash isubtype objects))))))


(defun build-patterns (predicates functors objects)
  (let ((pattern (make-hash-table :test #'eq)))
    (maphash 
     #'(lambda (keypred argspred)
	 (let* ((count-objlist (mapcar #'(lambda (var) (length (gethash (cdr var) objects)))
				       argspred))
		(bitmap-len (apply #'* count-objlist)))
	   (setf (gethash keypred pattern) (make-array bitmap-len :element-type 'bit 
						       :initial-element 0))))
     predicates)
    (when (hash-table-p functors)
      (maphash 
       #'(lambda (keyfun argsfun)
	   (let* ((count-objlist (mapcar #'(lambda (var) (length (gethash (cdr var) objects)))
					 argsfun))
		  (map-len (if (null count-objlist) 
			       1
			     (apply #'* count-objlist))))
	     (setf (gethash keyfun pattern) (make-array map-len :initial-element nil))))
       functors))
    pattern
))


(defun translate-state (state-def objects predicates functors state-pattern)
  (let ((hash-state (copy-state state-pattern))
	(lit-pos nil))
    (when (hkey-present 'total-time (dom-functors *pspace*))
      (setf (svref (gethash 'total-time hash-state) 0) 0))
    (dolist (iliteral state-def hash-state)
      (cond ((eq '= (car  iliteral))
	     (let ((ifunctor (cadr iliteral))
		   (func-val (third iliteral)))
	       (if (cdr ifunctor)
		   (setf lit-pos (map-position 
				  (literal-args-pos (cdr ifunctor) 
						    (gethash (car ifunctor) functors) objects)))
		 (setf lit-pos 0))
	       (setf (svref (gethash (car ifunctor) hash-state) lit-pos) func-val)))
	    (t
	     (if (cdr iliteral)
		 (setf lit-pos (map-position 
				(literal-args-pos (cdr iliteral) 
						  (gethash (car iliteral) predicates) objects)))
	       (setf lit-pos 0))
	     (setf (sbit (gethash (car iliteral) hash-state) lit-pos) 1))))
    ))
	  

(defun translate-goals (goals-def objects predicates pattern)
  (let ((hash-goals  (make-hash-table :test #'eq))
	(goals (if (eq 'and (car goals-def))
		   (cdr goals-def)
		 goals-def))
	(lit-pos nil)
	(map-len 0))
    (dolist (igoal goals)
      (unless (gethash (car igoal) hash-goals)
	(setf map-len (array-dimension (gethash (car igoal) pattern) 0))
	(setf (gethash (car igoal) hash-goals) (make-array map-len :element-type 'bit 
							   :initial-element 0)))
      (if (cdr igoal)
	  (setf lit-pos (map-position (literal-args-pos 
				       (cdr igoal) (gethash (car igoal) predicates) objects)))
	(setf lit-pos 0))
      (setf (sbit (gethash (car igoal) hash-goals) lit-pos) 1))
    hash-goals))


;; It appends the inheritobjects to the existing list of objects because some domains definitions
;; has an open hierarchy that defines objects of abstract types.
(defun buildall-inheritobjects (objects inheritypes)
  (update-constant-objects objects)
  (let ((inheritobjects (make-hash-table :test #'eq)))
    (maphash #'(lambda (ob-type inst-objects)
		 (setf (gethash ob-type inheritobjects) inst-objects))
	     objects)
    (dolist (itype inheritypes inheritobjects)
      (setf (gethash (car itype) inheritobjects)
	    (remove-duplicates (append (gethash (car itype) inheritobjects)
				       (get-inheritobjects (car itype) inheritypes objects)))))))


;; Creates a hash table with the conses-poslist (position . type)
;; for domain constants. This is stored in problem because objects may have constants and real objects
(defun build-constant-poslist (objects)
  (let ((constant-poslist (make-hash-table :test #'eq)))
    (maphash (lambda (key constants)
	       (dolist (i-const constants)
		 (setf (gethash i-const constant-poslist)
		       (cons (position i-const (gethash key objects)) key))))
	     (dom-constants *pspace*))
    constant-poslist))


;; It sets in state-patterns numvars to 0 after init state is created
;; This is performed since some of the numvars (= (pred) val) may not be defined in the state definition
(defun pattern-functor-update (patterns state)
  (maphash (lambda (key varmap)
	     (unless (simple-bit-vector-p varmap)
	       (let ((pattern-map (gethash key patterns)))
		 (dotimes (iposvar (length varmap))
		   (when (numberp (aref varmap iposvar))
		     (setf (aref pattern-map iposvar) 0))))))
	   state))


;; (defun create-fluent(fluent value)
;;   (let ((fluent-int (1+ (hash-table-count *fluents*))))
;;     (setf (gethash fluent *fluents*)
;; 	  (make-fluent :literal fluent-int
;; 		       :int (1+ (hash-table-count *fluents*))
;; 		       :init-value value))
;;     ))



(defun literal-predpos (literal &optional (fluent nil))
  (let ((predicates (if fluent (dom-functors *pspace*)
			(dom-predicates *pspace*)))
	(objects (problem-inheritobjects *current-problem*)))
    (cons (car literal)
	  (map-position (literal-args-pos (cdr literal) (gethash (car literal) predicates) objects)))))
  
(defun create-fact (fact)
  (setf (gethash fact *facts*)
	(make-fact :literal fact
		   :int (1+ (hash-table-count *facts*))
		   :predpos (literal-predpos fact)
		   ))
  )


(defun create-fluent (fluent)
  (setf (gethash (car fluent) *fluents*)
	(make-fluent :literal (car fluent)
		     :int  (1+ (hash-table-count *fluents*))
		     :predpos (literal-predpos (car fluent) t)
		     :init-value (second fluent))))

			       


(defun parse-initial-state (state-def)
  (setf *facts* (make-hash-table :test #'equal))
  (setf *init-facts* (make-hash-table :test #'equal))
  (setf *fluents* (make-hash-table :test #'equal))
  (setf *init-fluents* (make-hash-table :test #'equal))
  (create-fluent '((total-time) 0))
  (dolist (iliteral state-def)
    (cond ((eq '= (car  iliteral))
	   (unless (gethash (cadr iliteral) *fluents*)
	     (setf (gethash (cadr iliteral) *init-fluents*) (create-fluent (cdr iliteral)))
	     ))
	  (t
	   (unless (gethash iliteral *facts*)
	     (setf (gethash iliteral *init-facts*) (create-fact iliteral)))))))

(defun parse-initial-goals (literal-goals)
  (dolist (igoal literal-goals)
    (unless (gethash igoal *facts*)
      (create-fact igoal))))
  
  
(defun parse-problem (problem-def)
  (let* ((prob-domain (car (find-argument problem-def :domain)))
	 (problem (car (find-argument problem-def 'problem)))
	 (objects (translate-objects-def (find-argument problem-def :objects)))
	 (inheritypes (dom-inheritypes *pspace*))
	 (predicates (dom-predicates *pspace*))
	 (functors (dom-functors *pspace*))
	 (inheritobjects (buildall-inheritobjects objects inheritypes))
	 (patterns (build-patterns predicates functors inheritobjects))
	 (init-def (find-argument problem-def :init))  
	 (init-state (translate-state init-def inheritobjects predicates functors patterns))
	 (goals-def (find-argument problem-def :goal))
	 (goals (translate-goals (car goals-def) inheritobjects predicates patterns))
	 (metric (find-argument problem-def :metric)))
    (pattern-functor-update patterns init-state)
    (setf *current-problem* (make-say-problem :name problem
					  :domain prob-domain
					  :objects objects
					  :inheritobjects inheritobjects
					  :patterns patterns
					  :init-state init-state
					  :goals goals
					  :metric metric
					  :lit-init-state init-def
					  :lit-goals (if (eq 'and (caar goals-def))
							 (cdar goals-def)
						       goals-def)
					  :constant-poslist (build-constant-poslist inheritobjects)
					  ))
    ;;for new representation
    (parse-initial-state init-def)
    (parse-initial-goals (problem-lit-goals *current-problem*))
    ))


(defun read-pddl-problem (file)
  (cond ((dom-p *pspace*)
	 (with-open-file (istream file :direction :input)
	   (let ((pddl-def (read istream)))
	     (if (listp pddl-def)
		 (cond ((eq 'problem (car (second pddl-def)))
			(parse-problem (cdr pddl-def)))
		       (t
			(error (msg-error 'not_pddl_problem 'read-pddl-problem))))
		 (error (msg-error 'bad_list_construction 'read-pddl-problem))))))
	(t (error "SAYPHI Error: There is no loaded domain! [read-pddl-problem]"))))
