;; SAYPHI Planner 
;; Forward Guided Search Algorithms
;; 23.10.2007 Rocio Garcia Duran
;; _________________________________________________________________________________________

;; rocio
(defvar *use-rnpc* nil
  "If t, it will ask a learned model by RNPC how to make a decision")

(defvar *use-ribl* nil
  "If t, it will ask a learned model by RIBL how to make a decision. The global variable *use-rnpc* must be true too to translate the examples to mtarff format")

;; (defvar *training-rnpc-p* nil
;;   "If t, it is obtaining the training set and getting rules with variables such as <a>.
;;    Otherwise, it is executing RNPC and variables will be a, b, ... for mtarff files")
;; 
;; (defvar *mtarff-directory* "/home/rocio/trabajo4/ipss/tmp/"
;;   "For the translation of PDL crs into relational Weka")
;; 
;; (defvar *first-time-mtarff-p* nil
;;   "If t, it will create the first mtarff file. This must be exactly the same that the classifier mtarff file.
;;    If not, it will not compute the relational distance.")
;; 
;; (defvar *var-code* 96
;;   "Char code of A")
;; 
;; (defvar *var-code-i* 0
;;   "to multiplicate the name of variables. a, b, ..., a1, b1, ..., a2, b2, ...")
;; 
;; (defvar *rnpc-run* 0
;;   "Number of classifier to execute the rnpc")


(defun relational-hill-climbing (init-node h-function cost-fn search-options &key (minimize-h t) 
				(problem *current-problem*)
				(helpful *say-helpful*))
  (declare (ignore cost-fn search-options))
  (let ((sort-function (if minimize-h #'< #'>))
	(visited nil)
	(node init-node)
	(next-node nil)
	(rnpc-list nil)
	)
    (store-h-extras node h-function)

    (do* ((stop-this (stop-search node problem) (stop-search node problem)))
	 (stop-this (when-stop (cdr stop-this) node problem))
      (expand-state node :helpful helpful)
      ;; rocio
      (setf rnpc-list (rnpc-classify node))
      (print rnpc-list)
      (map-rnpc node h-function rnpc-list)
      ;; fin rocio

      (setf next-node (hc-next-node node visited sort-function))
      (cond ((snode-p next-node)
	     (push node visited)
	     (setf node next-node))
	    (t
	     (restore-nonhelpful node h-function)
	     (when (> *say-output* 1) (format t "{~% Restoring NON-Helpful Actions!!}"))
	     (setf next-node (hc-next-node node visited sort-function))
	     (when (snode-p next-node)
	       (push node visited))
	     (setf node next-node)))
      (when (snode-p node)
	(print-search-node node nil))
      )))


(defun rnpc-classify (node)
  (let ((node-state nil)  ; state of the current node
	(node-goals nil)  ; goals of the current node
	(node-ops   nil)) ; posible operators to execute in the current node
    (setf node-state (pp-state (snode-state node) 'list)) 
    (setf node-goals (pp-state (target-goals (snode-state node)) 'list))
    (setf node-ops nil)
    ;; transtale 
    (translate node-state node-goals)
    ;; rnpc-classify
    (dolist (inode (snode-children node))
      (unless (member (car (snode-plan-action inode)) node-ops) 
	(setf node-ops (cons (car (snode-plan-action inode)) node-ops)))) 
;    (format t " POSIBLES OPERADORES: ~a" node-ops)) 

;;   (if *use-ribl* ;;rocio
;;       (shell "java Probando2 0000000001 /home/rocio/asayphi/domains/blocksworld/classifierRIBL/ blocksworld.mtarff /home/rocio/asayphi/domains/blocksworld/test/ blocksworld.mtarff weka.rl.core.distance.MinkowskiRIBL /home/rocio/asayphi/domains/blocksworld/classifierRIBL/ > /home/rocio/asayphi/domains/blocksworld/test/salida-test")
;;       (shell (format nil "java Probando2 000001 /home/rocio/asayphi/domains/blocksworld/executions/run~a/ blocksworld.mtarff /home/rocio/asayphi/domains/blocksworld/test/ blocksworld.mtarff weka.rl.core.distance.MinkowskiRIBL /home/rocio/asayphi/domains/blocksworld/executions/run~a/ ~a > /home/rocio/asayphi/domains/blocksworld/test/salida-test" *rnpc-run* *rnpc-run* *rnpc-run*)))
;;   (let ((salida nil))
;;     (with-open-file (stream (concatenate 'string *mtarff-directory* "salida-test")
;; 			    :direction :input)
;;       (setq salida (read stream)))
;;     (some #'(lambda (rnpc-class)
;; 	      (car (member rnpc-class node-ops :key #'p4::rule-name)))
;; 	  salida)))

  '(LIFT LOAD DROP DRIVE UNLOAD)))


(defun map-rnpc (node h-function rnpc-list)
  (let ((rnpc-op nil))
    ;; probar uno a uno los operadores de la lista hasta el primero que se pueda ejecutar
    (dolist (iop rnpc-list)
      (unless rnpc-op ;; si es vacia
	(when (some #'(lambda (node-op) (eq iop (car (snode-plan-action node-op))))
		    (snode-children node))
	  (setf rnpc-op iop))))
    ;; para todos los hijos del nodo, si su operador es el elegido por rnpc -> calcula la heurística; si no, infinito
    (dolist (inode (snode-children node))
      (if (eq (car (snode-plan-action inode)) rnpc-op)
	  (store-h-extras inode h-function)
	  (setf (snode-h-value inode) most-positive-fixnum)
	  ))))

    
(defun translate (state goal)
  nil)