;; SAYPHI Planner 
;; Support function for command user interface
;; Td. 10.01.2006
;; ________________________________________________________________________________________
;; 

(defvar *problem-dir* "probs/")
(defvar *result-dir* "result/")
(defvar *domain-subdirs* '("probs" "result" "rules" "cases"))
(defvar *problem-file* "")


(defvar *say-solution-format* :values)
(defvar *experiment-train* nil)
(defvar *experiment-results* nil)
(defvar *experiment-tests* nil)
(defvar *experiment-outputs* nil)
(defvar *experiment-study* nil)


;;Variables to control the experiment of each domain
(defvar *this-domain* nil)
(defvar *this-domain-file* nil)
(defvar *this-timeout* nil)
(defvar *this-date* nil)




(defun say-plan (domain-dir domain-file problem-file)
  (say-domain domain-dir domain-file)
  (prob problem-file))


(defun say-domain (domain-dir &optional (domain-file "domain.pddl"))
  (setf *domain-dir* (concatenate 'string *domains-dir* domain-dir "/"))
  (read-pddl-domain (concatenate 'string *domain-dir* domain-file))
  (format t "~% ~a domain loaded!" domain-dir)
  t)


(defun prob (problem-file)
  (cond ((null *domain-dir*)
	 (error (msg-error 'no_domain_dir 'problem)))
	(t
	 (read-pddl-problem (concatenate 'string *domain-dir* *problem-dir* problem-file))
	 (setf *problem-file* problem-file)
	 (format t "~% ~a problem loaded!" problem-file)
	 t)))


(defun prob-fromset (problem-file)
  (setf *problem-file* problem-file)
  (cond ((null *domain-dir*)
	 (error (msg-error 'no_domain_dir 'problem)))
	(t
	 (read-pddl-problem (concatenate 'string *domain-dir* "probsets/" problem-file))
	 (format t "~% ~a problem loaded!" problem-file)
	 t)))


(defun sayphi-compile (file-name)
  (compile-file (concatenate 'string *system-dir* "planner/" file-name  *say-lisp-extension*)
       :output-file (concatenate 'string *bin-dir* file-name *bin-extension*)))


(defun sld (file-name)
  (load (concatenate 'string *system-dir* "planner/" file-name *say-lisp-extension*)))


(defun lcbr (file-name)
  (load (concatenate 'string *system-dir* "cbr/" file-name *say-lisp-extension*)))


;;Function para cargar dominio y fichero con ruta absoluta
(defun sayphi (domain-path prob-path &key (path "")
	                                  (plan-args nil))
    (read-pddl-domain (concatenate 'string path domain-path))
    (format t "~% domain loaded!")
    (read-pddl-problem (concatenate 'string path prob-path))
    (setf *problem-file* (filename-of-stringpath prob-path))
    (format t "~% ~a problem loaded!" *problem-file*)
    (apply #'plan plan-args)
)


(defun filename-of-stringpath (strpath)
    (subseq strpath (1+ (position #\/ strpath :from-end t))))
 

;; Retorna una lista ordenada de los archivos del directorio
(defun dirfiles-sorted (path wild-char)
   (let* ((dirfiles (directory (concatenate 'string path wild-char)))
          (file-names (mapcar #'(lambda (x-name)
				 (file-namestring x-name))
			      dirfiles))
          (sorted-names (sort file-names #'string-lessp)))
       (mapcar #'(lambda (x-name)
		   (concatenate 'string path x-name)) sorted-names)))
 


(defun get-problem-filename ()
  (cond ((find #\. *problem-file*)
	 (subseq *problem-file* 0 (- (length *problem-file*) 5)))
	(t *problem-file*)))


(defun get-problem-subset ()
  (let ((prob-name (get-problem-filename)))
    (when (> (length prob-name) 8)
      (subseq prob-name (- (length prob-name) 6) (- (length prob-name) 4)))))



(defun format-groupprob (number)
  (cond ((< number 10)
	 (format nil "0~d" number))
	(t
	 (format nil "~d" number))))
	      


(defun format-plotfield(value)
  (cond ((floatp value)
	 (format nil "~4$" value))
	(t
	 (format nil "~a" value))))
	


;;======================================================================
;; Basic Functions - Missing in SBCL or Common Lisp
;;======================================================================

;; It converts a string into a integer
(defun s_string-to-integer (strnumber)
  (let ((num 0) (decimal 0) (cut-num nil))
    (when (stringp strnumber)
      (dotimes (i (length strnumber) num)
 (setf decimal (- (char-int (aref strnumber (- (1- (length strnumber)) i))) 48)) 
 (cond ((and (not cut-num) 
      (>= decimal 0)
      (<= decimal 9))
        (setf num (+ num (* (expt 10 i) decimal))))
       (t (setf cut-num t))
       ))
 
)))

;;======================================================================
;; Functions for dealing with SBCL Analysis
;;======================================================================

;; It runs the form and stores in problem plist the consed bytes of the form
(defmacro say-consed-bytes (byte-cons-property form)
  ` (let ((trace-string (make-array 300 :element-type 'character 
				    :fill-pointer 3 :adjustable t))
	  (bytes-consed 0))
      (with-output-to-string 
	  (*trace-output* trace-string)
	#+SBCL (prog1 
		   (time ,form)
		 (setf bytes-consed (subseq trace-string (+ 6 (search "and" trace-string :from-end t))
					    (- (search "bytes" trace-string :from-end t) 1)))
		 (setf (get (problem-plist *current-problem*) ,byte-cons-property) 
		       (s_string-to-integer (remove #\, bytes-consed))
		       ))
	#-SBCL (time ,form)
	)))


(defun say-used-memory (room-space-property)
  (let ((trace-string (make-array 100 :element-type 'character 
				  :fill-pointer 3 :adjustable t))
	(memory-used 0))
     (with-output-to-string 
	(*standard-output* trace-string)
       (progn 
	 (room nil)
 	 (setf memory-used (subseq trace-string (+ 3 (search ":" trace-string ))
 				   (- (search "bytes" trace-string) 1)))
	 (setf (get (problem-plist *current-problem*) room-space-property) 
 	       (s_string-to-integer (remove #\, memory-used))
 	      )
))))
    
  


;;=======================================================================
;; Macro Utilities for Developing & Debugging
;;=======================================================================

(defmacro s_action_val (action &rest vars)
  `(let ((x_action (find ,action (dom-actions *pspace*) :key #'action-name)))
    (format t "~% <Action: ~a>" (action-name x_action))
    ,@(mapcar #'(lambda (x) `(format t "~% ~a: ~a" (quote ,x) (,x x_action))) vars)))
    
(defmacro s_allaction_val (&rest vars)
  `(dolist (iaction (dom-actions *pspace*))
     (format t "~%~% <Action: ~a>" (action-name iaction))
     ,@(mapcar #'(lambda (x) `(format t "~% ~a: ~a" (quote ,x) (,x iaction))) vars)))

;;It Gets the number of instances of each predicate in the state
(defun s_predschema_count ()
  (let ((pattern (problem-search-tree *current-problem*)))
    (format t "~% == STATE PREDICATES SCHEMAS ==")
    (maphash #'(lambda (pred varmap)
		 (cond ((bit-vector-p varmap)
			(format t "~% ~a: ~d (~a)" pred (length varmap) (count 1 varmap)))
		       (t
			(format t "~% ~a: ~d" pred (length varmap)))))
	     (snode-state pattern))))


(defmacro sg_allaction_val (&rest vars)
  `(dolist (iaction *actions*)
     (format t "~%~% <Gaction: ~a>" (gaction-planaction iaction))
     ,@(mapcar #'(lambda (x) `(format t "~% ~a: ~a" (quote ,x) (,x iaction))) vars)))


