
;; SAPYPHI PLANNER
;; Functions for ouput & printing
;; Td. 04.01.2006
;; ___________________________________________________________________________________


(defmacro say-indent (depth)
   `(let ((indent-len (if (> ,depth 20) 20 ,depth)))
      (make-sequence 'string indent-len :initial-element #\ )))

(defun binary-int (int)
  (setf *print-base* 2 
	*print-radix* t)
  (write int)
  (setf *print-base* 10 
	*print-radix* nil))
 

(defun pp-hash-table (htable)
  (maphash #'(lambda (key value)
	       (format t "~%~a :~a" key value)) htable))

(defun pp-predicate-pos (pred-pos)
  (let* ((predicates (dom-predicates *pspace*))
	 (objects (problem-inheritobjects *current-problem*))
	 (poslist (objpos-from-mappos (cdr pred-pos) 
				      (gethash (car pred-pos) predicates) objects))
	 (inst (cons (car pred-pos) (instances-from-poslist poslist objects))))
    (format t "~a"  inst)))


(defun pp-predicate-state (pred pred-bitmap args objects &optional (output 'printed))
  (let ((pred-inst (realinstances-from-bitmap pred-bitmap args objects))
	(result nil))
    (mapcar #'(lambda (inst)
		(if (eq output 'list)
		    (push (cons pred inst) result)
		    (format t "~%    ~a"  (cons pred inst))))
	    pred-inst)
    result))
  
(defun pp-functor-state (pred varsmap args objects &optional (output 'printed))
  (let ((result nil))
    (dotimes (i (length varsmap))
	(when (numberp (aref varsmap i))
	    (let ((var-inst (realinst-from-mappos i args objects)))
	      (if (eq output 'list)
		  (push (list '= (cons pred var-inst)  (aref varsmap i)) result)
		  (format t "~%    ~a = ~a "  (cons pred var-inst)  (aref varsmap i))))))
    result))

;; If output=printed, it will print the state
;; if output=list, it will return a list
(defun pp-state (state &optional (output 'printed))
  (let ((predicates (dom-predicates *pspace*))
	(functors (dom-functors *pspace*))
	(objects (problem-inheritobjects *current-problem*))
	(result nil))
    (maphash #'(lambda (pred bitmap)
		 (let ((state (if (simple-bit-vector-p bitmap)
				  (pp-predicate-state pred bitmap (gethash pred predicates) objects output)
				  (pp-functor-state pred bitmap (gethash pred functors) objects output))))
		   (if (eq output 'list)
		       (setq result (append state result)))))
	     state)
    (if (eq output 'list)
	result)))



	 
(defun say-pp-solution (solution &optional (evaltrace nil))
  (let ((planaction-counter 1) (out-trace ""))
    (cond ((> *say-output* 0)
	   (if (solution-found solution)
	       (format t "~% Solution: ")
	     (format t "~% Path When Stopped: "))
	   (dolist (i-node (solution-path solution) solution)
	     (if evaltrace
	       (setf out-trace (format nil "{g:~2$ , h:~2$ , f:~2$}" 
				       (snode-g-value i-node) (snode-h-value i-node) (snode-f-value i-node))) 
	       (setf out-trace ""))
	     (format t "~%     ~d: ~a [~a]  ~a" planaction-counter (snode-plan-action i-node) (snode-cost i-node) out-trace)
	     (incf planaction-counter)))
	  (t
	   solution))))

(defun pp-relaxed-plan (relaxed-plan)
  (let ((objects (problem-inheritobjects *current-problem*))
	(action-counter 0))
    (dolist (i-rxstep relaxed-plan)
      (incf action-counter)
      (format t "~%     Rx~d: ~a ~a [~a]" 
	      action-counter (caar i-rxstep) (instance-plan-action (car i-rxstep) objects)
	      (cdr i-rxstep)))))


(defun pp-achieved-graph (achiev-graph)
  (format t "~% >> -- Achieved Graph -- ")
  (maphash #'(lambda (layer ht-proplayer)
	       (format t "~% >>>> Layer ~a" layer)
	       (pp-hash-table ht-proplayer))
	   achiev-graph))


(defun pp-solution-gptrace (solution problem)
  (let ((objects (problem-inheritobjects problem))
	(planaction-counter 1))
    (dolist (i-node (solution-path solution) solution)
      (let ((plan-action (snode-applied-action i-node)))
	
	       (format t "~% ~d: ~a ~a" planaction-counter (car plan-action)
		       (instance-plan-action plan-action objects))
	       (incf planaction-counter)
	       (format t  "~%    Relaxed-plan:")
	       (pp-relaxed-plan (snode-relaxed-plan i-node))
	       (format t  "~%    G1 goals:~%")
	       (pp-state (snode-focus-goals i-node))
	       (format t "~%   _______________________________________________________")
	       )
      )))



(defun print-search-node (node open-list)
  (declare (ignore open-list))
  (cond ((= *say-output* 3)
	 (when (snode-p (snode-parent node))
	   (format t "~%~% EVALUATIONS:")
	   (dolist (i-sibling (snode-children (snode-parent node)))
	     (format t "~% ~a" i-sibling)))
	 (format t "~% ----------------------------------------------")
	 (format t "~% [~d  ~d h:~10,2f q:~d] ~a <~a>" 
		 (snode-number node) (snode-depth node) (snode-h-value node) 
		 (snode-f-value node)
		 (say-indent (snode-depth node)) (snode-applied-action node))
	 (format t "~% ----------------------------------------------"))
	((= *say-output* 2)
	 (format t "~% [~d ~t~d h:~2$  g:~d  f:~d]{~d}[~d] ~a [~a]" 
		 (snode-number node) (snode-depth node) (snode-h-value node) (snode-g-value node)
		 (snode-f-value node)
		 (if (snode-p (snode-parent node))
		   (length (snode-children (snode-parent node))) 0)
		 (get (problem-plist *current-problem*) :node-evaluated)
;; 		 (say-indent (snode-depth node)) 
		 (snode-plan-action node) (snode-cost node)))
    ))



(defun sayout-initialize ()
  (declare (special *current-problem*))
  (when (> *say-output* 1)
    (format t "Instantiating Actions... ")))

(defun sayout-search (stream alg heuristic cost)
  (when (> *say-output* 1)
    (format stream "~%Planner Parameters")
    (format stream "~%   Algorithm: ~a" alg)
    (format stream "~%   Heuristic Function: ~a" heuristic)
    (unless (null cost)
      (format stream "~%   Cost Function: ~a" cost)) 
    (format stream "~%_____________________________________________________________")
    (format t "~%Planning...~%~%")))
 

(defun pp-welcome ()
  (format t "~%~%~%")
  (format t "~% _______________________________________________")
  (format t "~%                                                ")
  (format t "~%              SAYPHI PLANNER v 2.0              ")
  (format t "~%                                                ")
  (format t "~%        UNIVERSIDAD CARLOS III DE MADRID        ")
  (format t "~%         PLANNING & LEARNING GROUP (PLG)        ")
  (format t "~%                                                ")
  (format t "~% _______________________________________________~% "))  


(defun pp-action-instance-byte (action)
  (format t "~% == Applicable Actions Schema ==")
  (dotimes (ibit (action-num-instances action))
    (when (logbitp ibit (action-instance-byte action))
      (format t "~% ~d ~a" ibit (cons (action-name action) (instance-plan-action (cons (action-name action) ibit)))))))



(defun pp-depth-indent (depth)
  (let ((indent " "))
    (dotimes (i depth indent)
      (setf indent (concatenate 'string indent " ")))))


(defun pp-search-tree (stream node &optional (depth nil))
  (let ((max-depth (if depth depth 100)))
    (cond ((<= (snode-depth node) max-depth)
      (format stream "~%~a~a~d[~a] ~a [~a] {~2$,~2$,~2$}"
	      (pp-depth-indent (snode-depth node))
	      (if (snode-closed node) "!" " ")
	      (snode-number node) (snode-depth node) (snode-plan-action node) (snode-cost node)
	      (snode-g-value node) (snode-h-value node) (snode-f-value node))
	   (dolist (i-child (snode-children node))
	     (pp-search-tree stream i-child max-depth)))
	  (t
	   (format t "--cut!"))
	  )))
  
;; For debuggin, only pass the list of nodes, like open-list
(defun pp-node-list (node-list &key (stream t))
  (format stream "~%>> Node List: ")
  (dolist (inode node-list)
    (format stream "~a " (snode-number inode))))
