;;******************************************************************************************
;; This file contains the callbacks functions for the ipss interface.
;; Raquel 05/2005
;;******************************************************************************************



;; --------------------------------------------------------------------------------------------
;; muestra la ventana pasada por parametro
;; --------------------------------------------------------------------------------------------
;; show window
(defun show-window (window-label)
   (declare (special widgets))
   (gtk_widget_show_all (gethash window-label widgets))
   1)

;; --------------------------------------------------------------------------------------------
;; oculta la ventana pasada por parametro
;; --------------------------------------------------------------------------------------------
;; hide window
(defun hide-window (window-label)
   (declare (special widgets))
   (gtk_widget_hide_on_delete (gethash window-label widgets))
    1) 

;; --------------------------------------------------------------------------------------------
;; Callback Functions (and other aux functions)
;;---------------------------------------------------------------------------------------------
;; --------------------------------------------------------------------------------------------
;; window vl-w-lr
;; Button LEARN
;; signal clicked
;; 
;; hamlet-learn parameters:
;;                          (domain OK
;;                     &key (probset "learning-1") OK
;;			    (learn-time-bound 200) OK
;;			    (eager-mode 'lazy-total) OK
;;			    (learning-type 'dynamic) OK
;;			    (re-init-p t) OK
;;			    (linearp nil)
;;    			    (probabilities nil)
;;			    (use-probabilities nil)
;;			    (cr-file nil) OK
;;			    (output-file "learn") OK
;;			    (make-span-p nil) 
;;			    (max-MS most-positive-fixnum)
;;			    (include-resources-p nil)
;;			    (num-op 500)
;;			    (act-dist 10000)
;;			    (cost-bound nil) OK
;;			    (cost-type nil)) OK
;; --------------------------------------------------------------------------------------------

;;---------------------------------------------------------------------------------------------
;; Ejecuta el aprendizaje de reglas
;;---------------------------------------------------------------------------------------------
(defun execute-learn()
;(gtk_button_released (gethash "button-lr-learn" widgets))
;(gtk_main_iteration)
(declare (special widgets))


;; almacena en variables los datos que el usuario a indicado en la interfaz para la realizacion del aprendizaje
(let  ((learn-time-bound (read-from-string (gtk_entry_get_text (gethash "time-bound-entry" widgets))))
      (eager-mode (read-from-string (gtk_combo_box_get_active_text (gethash "eager-mode-entry" widgets)))) 
      (learning-type (read-from-string 
                     (gtk_combo_box_get_active_text (gethash "learning-type-entry" widgets))))
      (cr-file (gtk_entry_get_text (gethash "output-rules-entry" widgets)))
      (cost-bound (if (= 1 (gtk_toggle_button_get_active (gethash "checkbutton-cost-bound" widgets)))
                      (read-from-string (gtk_entry_get_text (gethash "cost-bound-entry" widgets)))
                      nil))
      (cost-type   (read-from-string (gtk_entry_get_text (gethash "cost-type-entry" widgets))))
      (domain)
      (probset))
     

   ;; hamlet variables
   (setf *utility-threshold* (read-from-string 
                    (gtk_entry_get_text (gethash "spin-utility-threshold-entry" widgets))))
   (setf *hamlet-output-level* (read-from-string 
                    (gtk_entry_get_text (gethash "spin-verbosity-entry" widgets))))


   ;domain and probset 
   ;; carga los ficheros de dominio y de pruebas y muestra error en caso de que alguno no se haya indicado
   (if (or (and (eq *language* 'es) (setq domain (read-obligatory-field "domain-entry" "v1-w-lr" "No se ha especificado el dominio"))) (and (eq *language* 'en) (setq domain (read-obligatory-field "domain-entry" "v1-w-lr" "Domain is necessary"))))
     (if (or (and (eq *language* 'es) (setq probset  (read-obligatory-field "probset-entry" "v1-w-lr" "No se ha especificado un problema o conjunto de problemas"))) (and (eq *language* 'en) (setq probset  (read-obligatory-field "probset-entry" "v1-w-lr" "Probset is necessary"))))

       ;; si no se han dado problemas se comienza con la evaluacion de las siguientes expresiones
       (progn 
         ;rules file
         (if (equal cr-file "")
           (let ((cr-file-default  (format nil "rules-~(~a-~a-~d.~a~)" eager-mode 
                                         learning-type learn-time-bound *lisp-extension*)))
            (gtk_entry_set_text (gethash "output-rules-entry" widgets) cr-file-default)
            (setq cr-file  cr-file-default)))
	 
         (setf *valid-domain* 0)
	 (setf *valid-problem* 0)
         (setf *problem-number* 0)
	 (setf *valid-control-rules* 0)
         (handler-case (hamlet-learn (read-from-string domain)
                               :probset (read-from-string probset)
			    :learn-time-bound learn-time-bound
			    :eager-mode  eager-mode
			    :learning-type learning-type
			    :re-init-p (= 1 (gtk_toggle_button_get_active (gethash "checkbutton-re-init-p" widgets)))
;;			    :linearpile-default  (format n
;;    			    :probabilities
;;			    :use-probabilities
			    :cr-file cr-file
;;			    :output-file 
;;			    :make-span-p
;;			    :max-MS 
;;			    :include-resources-p
;;			    :num-op
;;			    :act-dist 
			    :cost-bound cost-bound
			    :cost-type cost-type)

           ;; en caso de error durante el aprendizaje
	   (error (signal)  
		(format t "~d~%" signal)
		(if (eq *valid-domain* 0)
			(error-dialog nil (delete_invalid_chars (format nil "Se ha producido un error en el dominio~%~d~%" signal)))
		(if (eq *valid-problem* 0)
			(error-dialog nil (delete_invalid_chars (format nil "Se ha producido un error en el problema: ~d~%~d~%" *problem-number* signal)))
		(error-dialog nil (delete_invalid_chars (format nil "~d~%" signal)))))
	   )
	  )
	 (if (eq *valid-control-rules* 1)
	   (information-dialog nil (format nil "El aprendizaje de reglas ha concluido satisfactoriamente"))
	 )
           1)   
          ;;else probset 
          0)
         ;;else domain
         0)))

	  
;; --------------------------------------------------------------------------------------------
;; to read a text entry with an obligatory field
;; it returns a string or NIL and it opens a dialog for error information
;; --------------------------------------------------------------------------------------------
(defun read-obligatory-field (widget-entry-name parent-window message)
  (declare (special widgets))
  (let ((text (gtk_entry_get_text (gethash widget-entry-name widgets))))
   (if (equal text "")
     (let ((dialog  (gtk_message_dialog_new (gethash parent-window widgets) 1 3 2 
                    message)))
        (gtk_dialog_run dialog)
        (gtk_widget_destroy dialog))
     text)))

;; --------------------------------------------------------------------------------------------
;; error dialog
;; --------------------------------------------------------------------------------------------
(defun error-dialog (parent-window message)
  (declare (special widgets))
  (let ((dialog  (gtk_message_dialog_new (gethash parent-window widgets) 1 3 2 
                    message)))
        (gtk_dialog_run dialog)
        (gtk_widget_destroy dialog)))

;; --------------------------------------------------------------------------------------------
;; information dialog
;; --------------------------------------------------------------------------------------------
(defun information-dialog (parent-window message)
    (declare (special widgets))
      (let ((dialog  (gtk_message_dialog_new (gethash parent-window widgets) 1 0 1
			message)))
	        (gtk_dialog_run dialog)
		        (gtk_widget_destroy dialog)))


;; --------------------------------------------------------------------------------------------
;; to show a dialog with a question message
;; it returns -9 -> no
;;            -8 -> yes
;;            -4 -> message window closed
;; --------------------------------------------------------------------------------------------
(defun show-question (parent-window message)
  (declare (special widgets))
  (let* ((dialog  (gtk_message_dialog_new (gethash parent-window widgets) 1 2 4 
                    message))
        (result (gtk_dialog_run dialog)))
   (gtk_widget_destroy dialog)
  result))

;; --------------------------------------------------------------------------------------------
;; window vl-w-lr
;; ToggleButton checkbutton-stop-after-problem-p
;; signal toggled
;; --------------------------------------------------------------------------------------------
(defun set_stop-after-problem-p ()
  (declare (special widgets))  
  (if (= 1 (gtk_toggle_button_get_active (gethash "checkbutton-stop-after-problem-p" widgets)))
       (progn (interface-trace t "*stop-after-problem-p* has been set to True" ) (setf *stop-after-problem-p* t))
       (progn (interface-trace t "*stop-after-problem-p* has been set to False") (setf *stop-after-problem-p*  nil)))
  1) 

;; --------------------------------------------------------------------------------------------
;; window vl-w-ipss
;; ToggleButton IPSScheckbutton-stop-after-problem-p
;; signal toggled
;; --------------------------------------------------------------------------------------------
(defun set_IPSSstop-after-problem-p ()
  (declare (special widgets))
  (if (= 1 (gtk_toggle_button_get_active (gethash "IPSScheckbutton-stop-after-problem-p" widgets)))
       (progn (interface-trace t "*stop-after-problem-p* has been set to True") (setf *stop-after-problem-p* t))
       (progn (interface-trace t "*stop-after-problem-p* has been set to False") (setf *stop-after-problem-p*  nil)))
  1) 

;; --------------------------------------------------------------------------------------------
;; window vl-w-ipss
;; ToggleButton IPSScheckbutton-store-fired-rules-p
;; signal toggled
;; --------------------------------------------------------------------------------------------
(defun set_IPSSstore-fired-rules-p ()
  (declare (special widgets))  
  (if (= 1 (gtk_toggle_button_get_active (gethash "IPSScheckbutton-store-fired-rules" widgets)))
       (progn (interface-trace t "*interface-store-fired-crs-p* has been set to True" ) (setf *interface-store-fired-crs-p* t))
       (progn (interface-trace t "*interface-store-fired-crs-p* has been set to False") (setf *interface-store-fired-crs-p* nil)))
  1)

;; --------------------------------------------------------------------------------------------
;; window vl-w-ipss
;; ToggleButton IPSScheckbutton-store-plans-p
;; signal toggled
;; --------------------------------------------------------------------------------------------
(defun set_IPSSstore-plans-p ()
  (declare (special widgets))  
  (if (= 1 (gtk_toggle_button_get_active (gethash "IPSScheckbutton-store-plans" widgets)))
       (progn (interface-trace t "*store-plans-p* has been set to True" ) (setf *store-plans-p* t))
       (progn (interface-trace t "*store-plans-p* has been set to False") (setf *store-plans-p* nil)))
  1)


;; --------------------------------------------------------------------------------------------
;; window vl-w-lr
;; ToggleButton checkbutton-cost-bound
;; signal toggled
;; --------------------------------------------------------------------------------------------
(defun set_cost-bound ()
  (declare (special widgets))  
  (if (= 1 (gtk_toggle_button_get_active (gethash "checkbutton-cost-bound" widgets)))
       (progn (interface-trace t "cost-bound is going to be introduced" ) 
              (gtk_widget_set_sensitive (gethash "cost-bound-entry" widgets) 1))                         
       (progn (interface-trace t "There is not cost-bound") 
              (gtk_widget_set_sensitive (gethash "cost-bound-entry" widgets) 0)))
  1) 


;; --------------------------------------------------------------------------------------------
;; window vl-w-ipss
;; ToggleButton IPSScheckbutton-cost-bound
;; signal toggled
;; --------------------------------------------------------------------------------------------
(defun set_IPSScost-bound ()
  (declare (special widgets))  
  (if (= 1 (gtk_toggle_button_get_active (gethash "IPSScheckbutton-cost-bound" widgets)))
       (progn (interface-trace t "cost-bound is going to be introduced" ) 
              (gtk_widget_set_sensitive (gethash "IPSScost-bound-entry" widgets) 1))                         
       (progn (interface-trace t "There is not cost-bound") 
              (gtk_widget_set_sensitive (gethash "IPSScost-bound-entry" widgets) 0)))
  1) 

;; --------------------------------------------------------------------------------------------
;; window vl-w-ipss
;; Opciones de Metric FF y Best-first
;; signal toggled
;; --------------------------------------------------------------------------------------------

(defun set_metric-best-first ()
  (declare (special widgets))  
  (if (eq 'best-first (read-from-string (gtk_combo_box_get_active_text (gethash "METRICalgorithm-entry" widgets))))
       (progn  
              (gtk_widget_set_sensitive (gethash "METRICpoda-entry" widgets) 1)
              (gtk_widget_set_sensitive (gethash "METRICmax-tam-entry" widgets) 1)
	      )                         
       (progn 
	      (gtk_combo_box_set_active (gethash "METRICpoda-entry" widgets) 3)
	      (gtk_widget_set_sensitive (gethash "METRICpoda-entry" widgets) 0)		
              (gtk_widget_set_sensitive (gethash "METRICmax-tam-entry" widgets) 0)
	      ))
  1) 


;; --------------------------------------------------------------------------------------------
;; window vl-w-ipss
;; Opciones de LPG y Incremental
;; signal toggled
;; --------------------------------------------------------------------------------------------

(defun set_lpg-incremental ()
  (declare (special widgets))  
  (if (eq 'INCREMENTAL (read-from-string (gtk_combo_box_get_active_text (gethash "LPGmodoejecucion-entry" widgets))))
       (progn  
             (gtk_widget_set_sensitive (gethash "LPGnumber-solutions-entry" widgets) 1)
	     )
       (progn 
              (gtk_widget_set_sensitive (gethash "LPGnumber-solutions-entry" widgets) 0)
	      ))
  1)

;; --------------------------------------------------------------------------------------------
;; window vl-w-ipss
;; Opciones de LPG y best-first
;; signal toggled
;; --------------------------------------------------------------------------------------------

(defun set_lpg-best-first ()
  (declare (special widgets))  
    (if (= 1 (gtk_toggle_button_get_active (gethash "LPGcheckbutton-nobestfirst" widgets)))
       (progn  
              (gtk_widget_set_sensitive (gethash "LPGcheckbutton-onlybestfirst" widgets) 0))
       (progn  
              (gtk_widget_set_sensitive (gethash "LPGcheckbutton-onlybestfirst" widgets) 1)))

    (if (= 1 (gtk_toggle_button_get_active (gethash "LPGcheckbutton-onlybestfirst" widgets)))
       (progn  
              (gtk_widget_set_sensitive (gethash "LPGcheckbutton-nobestfirst" widgets) 0))
       (progn  
              (gtk_widget_set_sensitive (gethash "LPGcheckbutton-nobestfirst" widgets) 1)))

  1) 


;; --------------------------------------------------------------------------------------------
;; any window
;; button Browse
;; signal cliked
;; To select a file from a file browser
;; entry-label -->entry_text label associated 
;; selection-type --> it can be domain, probset, prob or file
;; --------------------------------------------------------------------------------------------
(defun select (entry-label selection-type)
  (declare (special widgets))  
  ;; se crea una nueva pantalla de dialogo para la seleccion del archivo si devuelve -2 es aceptar, -1 cancelacion
  (let ((dialog-widget (gtk_file_chooser_dialog_new "Select File" (gethash "v1-w-lr" widgets) 
                                                    0 "gtk-ok" -2 "gtk-cancel" -1  nil)))

  ;; se asigna a la variable planner el planificador en el que nos encontramos: IPSS, SAYPHI o METRIC-FF
  (let  ((planner (read-from-string (gtk_combo_box_get_active_text (gethash "combobox1" widgets)))))
 	(if (eq planner 'IPSS)
		(setf *world-path* (format nil "~adomains/" *my-prodigy-pathname*)) 0)
	(if (eq planner 'SAYPHI) 
		(setf *world-path* (format nil "~adomains/" *my-sayphi-pathname*)) 0)
	(if (eq planner 'METRIC-FF) 
		(setf *world-path* (format nil "~adomains/" *my-metricff-pathname*)) 0)
	(if (eq planner 'SGPLAN) 
		(setf *world-path* (format nil "~adomains/" *my-sgplan-pathname*)) 0)
	(if (eq planner 'LPG) 
		(setf *world-path* (format nil "~adomains/" *my-lpg-pathname*)) 0)
	)
				    				    
 ;; directory      
  ;; se asigna el path actual a mostrar en la ventana de seleccion de archivo
  (case selection-type
    (domain  (gtk_file_chooser_set_current_folder dialog-widget *world-path*))
    (probset (gtk_file_chooser_set_current_folder dialog-widget 
               (concatenate 'string *domain-directory* "probsets/")))
    (prob    (gtk_file_chooser_set_current_folder dialog-widget 
               (concatenate 'string *domain-directory* "probs/")))
    (file    (gtk_file_chooser_set_current_folder dialog-widget *domain-directory*)))
          
  
  ;; si el usuario ha pulsado aceptar  
  (if (= (gtk_dialog_run dialog-widget) -2)

    ;; se guarda en la variable file el archivo que haya elegido el usuario
    (let ((file (gtk_file_chooser_get_filename dialog-widget))
	   (is-file)
	   (file-name))

	 ;; se hace un switch del tipo de seleccion
         (case selection-type 

	  ;; si es de tipo file se almacena en la tabla hash la direccion del fichero correspondiente
          (file (gtk_entry_set_text (gethash entry-label widgets)
                           (format nil "~a.~a" (pathname-name file) (pathname-type file)))
		(setf *reglas* file)
	  )
         
          ;; si es de tipo dominio 
          (domain    
	    
                ;;se ejecutan varias comprobaciones
	  	(handler-case
			
			;; se asigna 1 a la variable is-file si el fichero existe o 0 en caso contrario
			(if (probe-file file)
			     (setf is-file 1)
			     (setf is-file 0)	
			)
                ;; si se produce un error de seal se asigna valor 0 a la variable is-file
		(error (signal)
			(setf is-file 0)))
		
		;; si el fichero existe
		(if (eq is-file 1)	
				
			(progn
				(setf *dominio* file)
				(setf file-name (concatenate 'string (pathname-name file) "." (pathname-type file)))			
                                        ;; se le asigna a la variable file el path donde se encuentra el fichero
			    		(setq file (subseq file 0 (- (length file) (+ (length file-name) 1))))
			)
			(error-dialog nil (format nil "File ~a isn't a IPSS or PDDL domain file" (pathname-name file)))

			;; se elimina la opcion de elegir la carpeta donde est incluido el dominio.
			;;(setf *dominio* (concatenate 'string file "/domain.pddl"))				 
			)
                        
                        ;; se resetea el dominio, asignando el nombre del directorio del dominio
			(reset-domain-variables nil (format nil "~a" (pathname-name file)))

                        ;; se asigna al campo de texto del dominio el nombre del dominio elegido
			(gtk_entry_set_text (gethash entry-label widgets) (format nil "~a" (pathname-name file)))
	  )

          ;; si es de tipo conjunto de problemas
          (probset 
                      ;; se comprueba si el ultimo directorio del path se corresponde con "probstes"
                      (if (equal (car (last (pathname-directory file))) "probsets")
                         (setf *interface-probset-p* nil)
                         (setf *interface-probset-p* t))

			 ;; se asigna al campo de texto del conjunto de pruebas el nombre del archivo de problemas
                         (gtk_entry_set_text (gethash entry-label widgets)
                                    (format nil "~a" (pathname-name file))))

          ;; si es de tipo problema
          (prob 
		      (setf *problemas* file)
		      ;; se comprueba si el ultimo directorio del path se corresponde con "probs"
                      (if (equal (car (last (pathname-directory file))) "probs")
                         (setf *interface-probset-p* nil)
                         (setf *interface-probset-p* t))

                         ;; se asigna al campo de texto de problema el nombre del archivo del problema
                         (gtk_entry_set_text (gethash entry-label widgets)
                                    (format nil "~a" (pathname-name file)))))))

   (gtk_widget_destroy dialog-widget))
  
1)
;; --------------------------------------------------------------------------------------------
;; window "v1-w-trans"
;; button Browse
;; signal cliked
;; To select a file from a file browser
;; entry-label -->entry_text label associated 
;; selection-type --> it can be 'domain-check1, 'domain-check2 'domain-check3 'probset or 'prob
;; --------------------------------------------------------------------------------------------
(defun select-to-translate (entry-label selection-type &optional (filename-tocheck nil)
(entry-label-domain nil))
(declare (special widgets))

  ;; se crea una nueva pantalla de dialogo para la seleccion del archivo si devuelve -2 es aceptar, -1 cancelacion
  (let ((dialog-widget (gtk_file_chooser_dialog_new "Select File" (gethash "v1-w-trans" widgets) 
                                                    0 "gtk-ok" -2 "gtk-cancel" -1  nil)))

    ;; directory      
    ;; se asigna el path de inicio de la pantalla de seleccion
    (case selection-type
                
;;-----------------------------------------inicio parte david montero-------------------------------------------------------

;; en lugar de mostrar el directorio temporal, se muestra el path general del programa para que el usuario elija.
;; y dependiendo de que boton de buscar se pulsa se mostrar el path inicial de sayphi o de ipss
        ;;(domain-check (gtk_file_chooser_set_current_folder dialog-widget *ipss-temporal-dir*))
	
	  ((or domain-check1 domain-check2) (gtk_file_chooser_set_current_folder dialog-widget (concatenate 'string *my-sayphi-pathname* "domains/")))

	  (domain-check3 (gtk_file_chooser_set_current_folder dialog-widget (concatenate 'string *my-prodigy-pathname* "domains/")))


;; en lugar de mostar desde el directorio temporal, se muestra desde el directorio que eligio el usuario
;;        (prob  (gtk_file_chooser_set_current_folder dialog-widget 
;;						  (concatenate 'string *ipss-temporal-dir* (gtk_entry_get_text (gethash entry-label-domain widgets)) "/probs/")))

(prob  (gtk_file_chooser_set_current_folder dialog-widget 
						  (concatenate 'string *domain-directory-trans* "/probs/")))


;;-----------------------------------------fin parte david montero----------------------------------------------------------

        (probset  (gtk_file_chooser_set_current_folder dialog-widget 
						     (concatenate 'string *ipss-temporal-dir* (gtk_entry_get_text (gethash entry-label-domain widgets)) "/probsets/"))))

      ;; si el usuario ha pulsado en aceptar
      (if (= (gtk_dialog_run dialog-widget) -2)

        ;; se guarda en la variable file el nombre del fichero que ha elegido el usuario
	(let ((file (gtk_file_chooser_get_filename dialog-widget))
		(is-file)
		(file-name))

           (case selection-type 
             
             ;; si es de tipo check dominio
             ((or domain-check1 domain-check2 domain-check3) (interface-trace t (concatenate 'string file "/" filename-tocheck))
			     ;(if filename-tocheck 
				 ;(if (probe-file (pathname (concatenate 'string file "/" filename-tocheck)))
				  ;   (progn
				   ;  (print "Existo!!!!") 

                ;; se comprueba si el fichero elegido existe
		(handler-case
			(if (probe-file file)
			     (setf is-file 1)
			     (setf is-file 0)	
			)
		(error (signal)
			(setf is-file 0)))
		
                ;; si el fichero existe
		(if (eq is-file 1)
			(progn

;;--------------------------------------------inicio david montero---------------------------------------------------------

;; esto es para que el dominio ya no sea el temporal en lo de la traduccion, ya no se sobreescribe como antes
				(setf *domain-directory-trans* "/")
				(dolist (parte (pathname-directory file))
					(if (not (equal parte ':ABSOLUTE))  
					(setf *domain-directory-trans* (concatenate 'string *domain-directory-trans* parte "/")))
				)

;;--------------------------------------------fin david montero---------------------------------------------------------

			 	;;se asigna a la variable file-name el nombre del fichero
				(setf file-name (concatenate 'string (pathname-name file) "." (pathname-type file)))

                                ;; si el nombre del fichero es domain.pddl o domain.lisp
				(if (or (equal "domain.pddl" file-name)(equal "domain.lisp" file-name))

                                ;; se le asigna al campo de texto correspondiente el nombre del directorio del dominio
			    	(gtk_entry_set_text (gethash entry-label widgets) 
			    		(format nil "~a" (car (last (pathname-directory file)))))
			    	(error-dialog nil (format nil "File ~a isn't a PDDL domain file" 	(pathname-name file)))
				)
			)		 
			;; si el fichero no existe se asigna el nombre del fichero elegido 
			(gtk_entry_set_text (gethash entry-label widgets) (format nil "~a" (pathname-name file)))
		)

;;esto se supone que es el tratamiento del error pero esta mal colocado
			
				     ;(error-dialog nil (format nil "File ~a/~a doesn't exist" file filename-tocheck)))
				 ;(if  (probe-directory (pathname (concatenate 'string file "/")))  
				      ;(gtk_entry_set_text (gethash entry-label widgets) 
					;		  (format nil "~a" (pathname-name file)))
		                      ;(error-dialog nil (format nil "Directory ~a doesn't exist" (pathname-name file))))
	    )


            ;; si el tipo de seleccion es de tipo problema se comprueba si el directorio es "probs"
	    (prob  (if (equal (car (last (pathname-directory file))) "probs")

;;-----------------------------------inicio parte david montero-------------------------------------------------------------

;; se almacena el directorio del archivo de pruebas en la variable *probs-directory* para cargarlo cuando se pulse traducir.
			(progn
				(setf *probs-directory* "/")
				(dolist (parte (pathname-directory file))
					(if (not (equal parte ':ABSOLUTE))  
					(setf *probs-directory* (concatenate 'string *probs-directory* parte "/")))
				)

;;-----------------------------------fin parte david montero-------------------------------------------------------------

                      ;; se asigna al campo de texto correspondiente el nombre del archivo
		       (gtk_entry_set_text (gethash entry-label widgets)
					   (format nil "~a*" (pathname-name file))))))

            ;; si el tipo de seleccion es de tipo conjunto de problemas se comprueba si el directorio es "probsets"
	    (probset  (if (equal (car (last (pathname-directory file))) "probsets")

;;-----------------------------------inicio parte david montero-------------------------------------------------------------

;; se almacena el directorio del archivo de pruebas en la variable *probset-directory* para cargarlo cuando se pulse traducir.
			(progn
				(setf *probset-directory* "/")
				(dolist (parte (pathname-directory file))
					(if (not (equal parte ':ABSOLUTE))  
					(setf *probset-directory* (concatenate 'string *probset-directory* parte "/")))
				)

;;-----------------------------------fin parte david montero-------------------------------------------------------------

			  ;; se asigna al campo de texto correspondiente el nombre del archivo
			  (gtk_entry_set_text (gethash entry-label widgets)
					      (format nil "~a" (pathname-name file)))))))))
	
    (gtk_widget_destroy dialog-widget))

  1)

;; --------------------------------------------------------------------------------------------
;; window vl-w-lr
;; ComboBox learning-type
;; signal changed
;; --------------------------------------------------------------------------------------------
(defun learning-type-if-EBL ()
 (declare (special widgets))
 (let ((learning-type (read-from-string 
                     (gtk_combo_box_get_active_text (gethash "learning-type-entry" widgets)))))

    (if (eq learning-type 'ebl)
     (progn
      ;;The consistent learning type is eager-partial
      (gtk_combo_box_set_active (gethash "eager-mode-entry" widgets) 3)
      (gtk_widget_set_sensitive (gethash "eager-mode-entry" widgets) 0)
      ;; show a label and a combo box to set the utility threshold
      (show-window "label-utility-threshold")
      (show-window "spin-utility-threshold-entry"))
     (progn
      (gtk_widget_set_sensitive (gethash "eager-mode-entry" widgets) 1)
      (hide-window "label-utility-threshold")
      (hide-window "spin-utility-threshold-entry"))))
 1)
 

;; --------------------------------------------------------------------------------------------
;; window vl-w-ipss
;; signal changed
;; controla si se produce algun cambio en la eleccion del planificador
;; --------------------------------------------------------------------------------------------
 (defun planner-interface ()
 (declare (special widgets))

 ;; lee el tipo de planificador elegido
 (let ((planner-name (read-from-string 
                     (gtk_combo_box_get_active_text (gethash "combobox1" widgets)))))

   ;; si el planificador es ipss
   (if (eq planner-name 'ipss)
     (progn
        (hide-window "SAYPHIheuristic-label")
        (hide-window "SAYPHIheuristic-entry")
        (hide-window "SAYPHIalgorithm-label")
        (hide-window "SAYPHIalgorithm-entry")

	(hide-window "LPGmodoejecucion-label")
	(hide-window "LPGmodoejecucion-entry")
	(hide-window "LPGsearchsteps-label")
	(hide-window "LPGsearchsteps-entry")
	(hide-window "LPGrestarts-label")
	(hide-window "LPGrestarts-entry")
	(hide-window "LPGrepeats-label")
	(hide-window "LPGrepeats-entry")
	(hide-window "LPGseed-label")
	(hide-window "LPGseed-entry")
	(hide-window "LPGtime-bound-local-label")
	(hide-window "LPGtime-bound-local-entry")
	(hide-window "LPGnumber-solutions-label")
	(hide-window "LPGnumber-solutions-entry")
	(hide-window "LPGnoise-label")
	(hide-window "LPGnoise-entry")
        ;;(hide-window "LPGcheckbutton-save-plans")
        ;;(hide-window "LPGcheckbutton-verbose")
        (hide-window "LPGcheckbutton-static-noise")
        ;;(hide-window "LPGcheckbutton-maxnoise")
        (hide-window "LPGcheckbutton-lowmemory")
        (hide-window "LPGcheckbutton-nobestfirst")
        (hide-window "LPGcheckbutton-onlybestfirst")
        (hide-window "LPGcheckbutton-timesteps")


	(hide-window "METRICheuristic-entry")
	(hide-window "METRICalgorithm-entry")
	(hide-window "METRICpoda-entry")
	(hide-window "METRICpoda-label")
	(hide-window "METRICw-g-entry")
	(hide-window "METRICw-h-entry")
	(hide-window "METRICw-g-label")
	(hide-window "METRICw-h-label")
	(hide-window "METRICmax-tam-label")
	(hide-window "METRICmax-tam-entry")
	(hide-window "METRICcheckbutton-metric-optimice")

        (show-window "IPSSspin-verbosity-entry")
        (show-window "IPSSverbosity-label")
        (show-window "label70")
        (show-window "button-rules")
	(show-window "IPSSoutput-rules-entry")
	(show-window "IPSScost-type-label")
	(show-window "IPSScheckbutton-multiple-sols-p")
        (show-window "IPSScheckbutton-backtrack-with-costs-p")
	(show-window "IPSScheckbutton-cost-bound")
        (show-window "IPSScheckbutton-stop-after-problem-p")
	(show-window "IPSScheckbutton-store-fired-rules")
        (show-window "IPSScheckbutton-store-plans")
        (show-window "IPSScheckbutton-run-without-p")
        (show-window "IPSScost-type-entry")
        (show-window "IPSScost-bound-entry")
        (show-window "button-ipss-show-search-tree")))

    ;; si el planificador es sayphi
    (if (eq planner-name 'sayphi)
     (progn
        (show-window "SAYPHIheuristic-label")
        (show-window "SAYPHIheuristic-entry") 
        (show-window "SAYPHIalgorithm-label")
        (show-window "SAYPHIalgorithm-entry")        
        (show-window "IPSSspin-verbosity-entry")
        (show-window "IPSSverbosity-label")
	(hide-window "label70")
        (hide-window "button-rules")

	(hide-window "IPSSoutput-rules-entry")
	(hide-window "IPSScost-type-label")
        (hide-window "IPSScheckbutton-multiple-sols-p")
	(hide-window "IPSScheckbutton-backtrack-with-costs-p")
	(hide-window "IPSScheckbutton-cost-bound")
        (hide-window "IPSScheckbutton-stop-after-problem-p")
	(hide-window "IPSScheckbutton-store-fired-rules")
        (hide-window "IPSScheckbutton-store-plans")
        (hide-window "IPSScheckbutton-run-without-p")
        (hide-window "IPSScost-type-entry")
        (hide-window "IPSScost-bound-entry")

	(hide-window "LPGmodoejecucion-label")
	(hide-window "LPGmodoejecucion-entry")
	(hide-window "LPGsearchsteps-label")
	(hide-window "LPGsearchsteps-entry")
	(hide-window "LPGrestarts-label")
	(hide-window "LPGrestarts-entry")
	(hide-window "LPGrepeats-label")
	(hide-window "LPGrepeats-entry")
	(hide-window "LPGseed-label")
	(hide-window "LPGseed-entry")
	(hide-window "LPGtime-bound-local-label")
	(hide-window "LPGtime-bound-local-entry")
	(hide-window "LPGnumber-solutions-label")
	(hide-window "LPGnumber-solutions-entry")
	(hide-window "LPGnoise-label")
	(hide-window "LPGnoise-entry")
        ;;(hide-window "LPGcheckbutton-save-plans")
        ;;(hide-window "LPGcheckbutton-verbose")
        (hide-window "LPGcheckbutton-static-noise")
        ;;(hide-window "LPGcheckbutton-maxnoise")
        (hide-window "LPGcheckbutton-lowmemory")
        (hide-window "LPGcheckbutton-nobestfirst")
        (hide-window "LPGcheckbutton-onlybestfirst")
        (hide-window "LPGcheckbutton-timesteps")

	(hide-window "METRICheuristic-entry")
	(hide-window "METRICalgorithm-entry")
	(hide-window "METRICw-g-entry")
	(hide-window "METRICw-h-entry")
	(hide-window "METRICw-g-label")
	(hide-window "METRICw-h-label")
	(hide-window "METRICpoda-entry")
	(hide-window "METRICpoda-label")
	(hide-window "METRICmax-tam-label")
	(hide-window "METRICmax-tam-entry")
	(hide-window "METRICcheckbutton-metric-optimice")
        (hide-window "button-ipss-show-search-tree")))

    ;; si el planificador es metric-ff
    (if (eq planner-name 'metric-ff)
      (progn	
        (hide-window "SAYPHIheuristic-entry")
        (hide-window "SAYPHIalgorithm-entry")

	(hide-window "IPSScost-type-label")
        (hide-window "IPSScheckbutton-multiple-sols-p")
	(hide-window "IPSScheckbutton-backtrack-with-costs-p")
	(hide-window "IPSScheckbutton-cost-bound")
        (hide-window "IPSScheckbutton-stop-after-problem-p")
	(hide-window "IPSScheckbutton-store-fired-rules")
        (hide-window "IPSScheckbutton-store-plans")
        (hide-window "IPSScheckbutton-run-without-p")
        (hide-window "IPSScost-type-entry")
        (hide-window "IPSScost-bound-entry")
        (hide-window "IPSSspin-verbosity-entry")
        (hide-window "IPSSverbosity-label")

	(hide-window "LPGmodoejecucion-label")
	(hide-window "LPGmodoejecucion-entry")
	(hide-window "LPGsearchsteps-label")
	(hide-window "LPGsearchsteps-entry")
	(hide-window "LPGrestarts-label")
	(hide-window "LPGrestarts-entry")
	(hide-window "LPGrepeats-label")
	(hide-window "LPGrepeats-entry")
	(hide-window "LPGseed-label")
	(hide-window "LPGseed-entry")
	(hide-window "LPGtime-bound-local-label")
	(hide-window "LPGtime-bound-local-entry")
	(hide-window "LPGnumber-solutions-label")
	(hide-window "LPGnumber-solutions-entry")
	(hide-window "LPGnoise-label")
	(hide-window "LPGnoise-entry")
        ;;(hide-window "LPGcheckbutton-save-plans")
        ;;(hide-window "LPGcheckbutton-verbose")
        (hide-window "LPGcheckbutton-static-noise")
        ;;(hide-window "LPGcheckbutton-maxnoise")
        (hide-window "LPGcheckbutton-lowmemory")
        (hide-window "LPGcheckbutton-nobestfirst")
        (hide-window "LPGcheckbutton-onlybestfirst")
        (hide-window "LPGcheckbutton-timesteps")

	(hide-window "button-ipss-show-search-tree")
        (show-window "SAYPHIheuristic-label")
        (show-window "SAYPHIalgorithm-label")
        (show-window "label70")
        (show-window "button-rules")
	(show-window "IPSSoutput-rules-entry")
	(show-window "METRICheuristic-entry")
	(show-window "METRICalgorithm-entry")
	(show-window "METRICw-g-entry")
	(show-window "METRICw-h-entry")
	(show-window "METRICw-g-label")
	(show-window "METRICw-h-label")
	(show-window "METRICpoda-entry")
	(show-window "METRICpoda-label")
	(show-window "METRICmax-tam-label")
	(show-window "METRICmax-tam-entry")
	(show-window "METRICcheckbutton-metric-optimice")))

    ;; si el planificador es sgplan
    (if (eq planner-name 'sgplan)
      (progn	
        (hide-window "SAYPHIheuristic-entry")
        (hide-window "SAYPHIalgorithm-entry")

	(hide-window "IPSScost-type-label")
        (hide-window "IPSScheckbutton-multiple-sols-p")
	(hide-window "IPSScheckbutton-backtrack-with-costs-p")
	(hide-window "IPSScheckbutton-cost-bound")
        (hide-window "IPSScheckbutton-stop-after-problem-p")
	(hide-window "IPSScheckbutton-store-fired-rules")
        (hide-window "IPSScheckbutton-store-plans")
        (hide-window "IPSScheckbutton-run-without-p")
        (hide-window "IPSScost-type-entry")
        (hide-window "IPSScost-bound-entry")
        (hide-window "IPSSspin-verbosity-entry")
        (hide-window "IPSSverbosity-label")
	(hide-window "button-ipss-show-search-tree")

        (hide-window "SAYPHIheuristic-label")
        (hide-window "SAYPHIalgorithm-label")
        (hide-window "label70")
        (hide-window "button-rules")
	(hide-window "IPSSoutput-rules-entry")

	(hide-window "LPGmodoejecucion-label")
	(hide-window "LPGmodoejecucion-entry")
	(hide-window "LPGsearchsteps-label")
	(hide-window "LPGsearchsteps-entry")
	(hide-window "LPGrestarts-label")
	(hide-window "LPGrestarts-entry")
	(hide-window "LPGrepeats-label")
	(hide-window "LPGrepeats-entry")
	(hide-window "LPGseed-label")
	(hide-window "LPGseed-entry")
	(hide-window "LPGtime-bound-local-label")
	(hide-window "LPGtime-bound-local-entry")
	(hide-window "LPGnumber-solutions-label")
	(hide-window "LPGnumber-solutions-entry")
	(hide-window "LPGnoise-label")
	(hide-window "LPGnoise-entry")
        ;;(hide-window "LPGcheckbutton-save-plans")
        ;;(hide-window "LPGcheckbutton-verbose")
        (hide-window "LPGcheckbutton-static-noise")
        ;;(hide-window "LPGcheckbutton-maxnoise")
        (hide-window "LPGcheckbutton-lowmemory")
        (hide-window "LPGcheckbutton-nobestfirst")
        (hide-window "LPGcheckbutton-onlybestfirst")
        (hide-window "LPGcheckbutton-timesteps")

	(hide-window "METRICheuristic-entry")
	(hide-window "METRICalgorithm-entry")
	(hide-window "METRICw-g-entry")
	(hide-window "METRICw-h-entry")
	(hide-window "METRICw-g-label")
	(hide-window "METRICw-h-label")
	(hide-window "METRICpoda-entry")
	(hide-window "METRICpoda-label")
	(hide-window "METRICmax-tam-label")
	(hide-window "METRICmax-tam-entry")
	(hide-window "METRICcheckbutton-metric-optimice")))

    ;; si el planificador es lpg
    (if (eq planner-name 'lpg)
      (progn	
        (hide-window "SAYPHIheuristic-entry")
        (hide-window "SAYPHIalgorithm-entry")

	(hide-window "IPSScost-type-label")
        (hide-window "IPSScheckbutton-multiple-sols-p")
	(hide-window "IPSScheckbutton-backtrack-with-costs-p")
	(hide-window "IPSScheckbutton-cost-bound")
        (hide-window "IPSScheckbutton-stop-after-problem-p")
	(hide-window "IPSScheckbutton-store-fired-rules")
        (hide-window "IPSScheckbutton-store-plans")
        (hide-window "IPSScheckbutton-run-without-p")
        (hide-window "IPSScost-type-entry")
        (hide-window "IPSScost-bound-entry")
        (hide-window "IPSSspin-verbosity-entry")
        (hide-window "IPSSverbosity-label")
	(hide-window "button-ipss-show-search-tree")

        (hide-window "SAYPHIheuristic-label")
        (hide-window "SAYPHIalgorithm-label")
        (hide-window "label70")
        (hide-window "button-rules")
	(hide-window "IPSSoutput-rules-entry")

	(show-window "LPGmodoejecucion-label")
	(show-window "LPGmodoejecucion-entry")
	(show-window "LPGsearchsteps-label")
	(show-window "LPGsearchsteps-entry")
	(show-window "LPGrestarts-label")
	(show-window "LPGrestarts-entry")
	(show-window "LPGrepeats-label")
	(show-window "LPGrepeats-entry")
	(show-window "LPGseed-label")
	(show-window "LPGseed-entry")
	(show-window "LPGtime-bound-local-label")
	(show-window "LPGtime-bound-local-entry")
	(show-window "LPGnumber-solutions-label")
	(show-window "LPGnumber-solutions-entry")
	(show-window "LPGnoise-label")
	(show-window "LPGnoise-entry")
        ;;(show-window "LPGcheckbutton-save-plans")
        ;;(show-window "LPGcheckbutton-verbose")
        (show-window "LPGcheckbutton-static-noise")
        ;;(show-window "LPGcheckbutton-maxnoise")
        (show-window "LPGcheckbutton-lowmemory")
        (show-window "LPGcheckbutton-nobestfirst")
        (show-window "LPGcheckbutton-onlybestfirst")
        (show-window "LPGcheckbutton-timesteps")

	(hide-window "METRICheuristic-entry")
	(hide-window "METRICalgorithm-entry")
	(hide-window "METRICw-g-entry")
	(hide-window "METRICw-h-entry")
	(hide-window "METRICw-g-label")
	(hide-window "METRICw-h-label")
	(hide-window "METRICpoda-entry")
	(hide-window "METRICpoda-label")
	(hide-window "METRICmax-tam-label")
	(hide-window "METRICmax-tam-entry")
	(hide-window "METRICcheckbutton-metric-optimice")))
    )


 1)
 
;; --------------------------------------------------------------------------------------------
;; window vl-w-lr
;; inicializa la pantalla de aprendizaje de reglas
;; --------------------------------------------------------------------------------------------
(defun initialize-v1-w-lr ()
(declare (special widgets))

;; oculta el resto de pantallas
(hide-window "fixed-v1-ipss")
(hide-window "fixed-v1-lm")
(hide-window "fixed-v1-trans")

;; muestra la pantalla de aprendizaje de reglas
(show-window "fixed-v1-lr")


 ;; comprueba si el modo activo es EBL para activar sus campos especiales
 (if (not (eq 'ebl (read-from-string (gtk_combo_box_get_active_text (gethash "learning-type-entry" widgets)))))
   (progn   (hide-window "label-utility-threshold")
            (hide-window "spin-utility-threshold-entry")))
(interface-trace t "v1-w-lr initialized")
1)



;; --------------------------------------------------------------------------------------------
;; window vl-w-ipss
;; inicializa la pantalla de planificacion
;; --------------------------------------------------------------------------------------------
(defun initialize-v1-w-ipss ()

;; oculta el resto de pantallas
(hide-window "fixed-v1-lr")
(hide-window "fixed-v1-lm")
(hide-window "fixed-v1-trans")

;; muestra la pantalla de planificacion
(show-window "fixed-v1-ipss")

;; comprueba si el planificador elegido es SAYPHI para mostrar los datos del planificador SAYPHI.
(if (eq 'sayphi (read-from-string (gtk_combo_box_get_active_text (gethash "combobox1" widgets))))
     (progn
        (show-window "SAYPHIheuristic-label")
        (show-window "SAYPHIheuristic-entry") 
        (show-window "SAYPHIalgorithm-label")
        (show-window "SAYPHIalgorithm-entry")        
        (show-window "IPSSspin-verbosity-entry")
        (show-window "IPSSverbosity-label")
	(hide-window "label70")
        (hide-window "button-rules")

	(hide-window "IPSSoutput-rules-entry")
	(hide-window "IPSScost-type-label")
        (hide-window "IPSScheckbutton-multiple-sols-p")
	(hide-window "IPSScheckbutton-backtrack-with-costs-p")
	(hide-window "IPSScheckbutton-cost-bound")
        (hide-window "IPSScheckbutton-stop-after-problem-p")
	(hide-window "IPSScheckbutton-store-fired-rules")
        (hide-window "IPSScheckbutton-store-plans")
        (hide-window "IPSScheckbutton-run-without-p")
        (hide-window "IPSScost-type-entry")
        (hide-window "IPSScost-bound-entry")

	(hide-window "LPGmodoejecucion-label")
	(hide-window "LPGmodoejecucion-entry")
	(hide-window "LPGsearchsteps-label")
	(hide-window "LPGsearchsteps-entry")
	(hide-window "LPGrestarts-label")
	(hide-window "LPGrestarts-entry")
	(hide-window "LPGrepeats-label")
	(hide-window "LPGrepeats-entry")
	(hide-window "LPGseed-label")
	(hide-window "LPGseed-entry")
	(hide-window "LPGtime-bound-local-label")
	(hide-window "LPGtime-bound-local-entry")
	(hide-window "LPGnumber-solutions-label")
	(hide-window "LPGnumber-solutions-entry")
	(hide-window "LPGnoise-label")
	(hide-window "LPGnoise-entry")
        ;;(hide-window "LPGcheckbutton-save-plans")
        ;;(hide-window "LPGcheckbutton-verbose")
        (hide-window "LPGcheckbutton-static-noise")
        ;;(hide-window "LPGcheckbutton-maxnoise")
        (hide-window "LPGcheckbutton-lowmemory")
        (hide-window "LPGcheckbutton-nobestfirst")
        (hide-window "LPGcheckbutton-onlybestfirst")
        (hide-window "LPGcheckbutton-timesteps")

	(hide-window "METRICheuristic-entry")
	(hide-window "METRICalgorithm-entry")
	(hide-window "METRICw-g-entry")
	(hide-window "METRICw-h-entry")
	(hide-window "METRICw-g-label")
	(hide-window "METRICw-h-label")
	(hide-window "METRICpoda-entry")
	(hide-window "METRICpoda-label")
	(hide-window "METRICmax-tam-label")
	(hide-window "METRICmax-tam-entry")
	(hide-window "METRICcheckbutton-metric-optimice")
        (hide-window "button-ipss-show-search-tree")))

;; comprueba si el planificador elegido es IPSS para mostrar los datos del planificador IPSS.
(if (eq 'ipss (read-from-string (gtk_combo_box_get_active_text (gethash "combobox1" widgets))))
   (progn	

        (hide-window "SAYPHIheuristic-label")
        (hide-window "SAYPHIheuristic-entry")
        (hide-window "SAYPHIalgorithm-label")
        (hide-window "SAYPHIalgorithm-entry")

        (show-window "IPSSspin-verbosity-entry")
        (show-window "IPSSverbosity-label")

	(hide-window "LPGmodoejecucion-label")
	(hide-window "LPGmodoejecucion-entry")
	(hide-window "LPGsearchsteps-label")
	(hide-window "LPGsearchsteps-entry")
	(hide-window "LPGrestarts-label")
	(hide-window "LPGrestarts-entry")
	(hide-window "LPGrepeats-label")
	(hide-window "LPGrepeats-entry")
	(hide-window "LPGseed-label")
	(hide-window "LPGseed-entry")
	(hide-window "LPGtime-bound-local-label")
	(hide-window "LPGtime-bound-local-entry")
	(hide-window "LPGnumber-solutions-label")
	(hide-window "LPGnumber-solutions-entry")
	(hide-window "LPGnoise-label")
	(hide-window "LPGnoise-entry")
        ;;(hide-window "LPGcheckbutton-save-plans")
        ;;(hide-window "LPGcheckbutton-verbose")
        (hide-window "LPGcheckbutton-static-noise")
        ;;(hide-window "LPGcheckbutton-maxnoise")
        (hide-window "LPGcheckbutton-lowmemory")
        (hide-window "LPGcheckbutton-nobestfirst")
        (hide-window "LPGcheckbutton-onlybestfirst")
        (hide-window "LPGcheckbutton-timesteps")


	(hide-window "METRICheuristic-entry")
	(hide-window "METRICalgorithm-entry")
	(hide-window "METRICw-g-entry")
	(hide-window "METRICw-h-entry")
	(hide-window "METRICw-g-label")
	(hide-window "METRICw-h-label")
	(hide-window "METRICpoda-entry")
	(hide-window "METRICpoda-label")
	(hide-window "METRICmax-tam-label")
	(hide-window "METRICmax-tam-entry")
	(hide-window "METRICcheckbutton-metric-optimice")
        (show-window "label70")
        (show-window "button-rules")
	(show-window "IPSSoutput-rules-entry")
	(show-window "IPSScost-type-label")
	(show-window "IPSScheckbutton-multiple-sols-p")
        (show-window "IPSScheckbutton-backtrack-with-costs-p")
	(show-window "IPSScheckbutton-cost-bound")
        (show-window "IPSScheckbutton-stop-after-problem-p")
	(show-window "IPSScheckbutton-store-fired-rules")
        (show-window "IPSScheckbutton-store-plans")
        (show-window "IPSScheckbutton-run-without-p")
        (show-window "IPSScost-type-entry")
        (show-window "IPSScost-bound-entry")
        (show-window "button-ipss-show-search-tree")))


;; comprueba si el planificador elegido es metric-ff para mostrar los datos del planificador metric-ff.
(if (eq 'metric-ff (read-from-string (gtk_combo_box_get_active_text (gethash "combobox1" widgets))))
   (progn	
        (hide-window "SAYPHIheuristic-label")
        (hide-window "SAYPHIheuristic-entry")
        (hide-window "SAYPHIalgorithm-label")
        (hide-window "SAYPHIalgorithm-entry")

	(hide-window "IPSScheckbutton-multiple-sols-p")
        (hide-window "IPSScheckbutton-backtrack-with-costs-p")
	(hide-window "IPSScheckbutton-cost-bound")
        (hide-window "IPSScheckbutton-stop-after-problem-p")
	(hide-window "IPSScheckbutton-store-fired-rules")
        (hide-window "IPSScheckbutton-store-plans")
        (hide-window "IPSScheckbutton-run-without-p")
        (hide-window "IPSScost-type-entry")
        (hide-window "IPSScost-bound-entry")
	(hide-window "IPSScost-type-label")
        (hide-window "IPSSspin-verbosity-entry")
        (hide-window "IPSSverbosity-label")
	(hide-window "button-ipss-show-search-tree")
        (show-window "label70")
        (show-window "button-rules")
	(show-window "IPSSoutput-rules-entry")

	(hide-window "LPGmodoejecucion-label")
	(hide-window "LPGmodoejecucion-entry")
	(hide-window "LPGsearchsteps-label")
	(hide-window "LPGsearchsteps-entry")
	(hide-window "LPGrestarts-label")
	(hide-window "LPGrestarts-entry")
	(hide-window "LPGrepeats-label")
	(hide-window "LPGrepeats-entry")
	(hide-window "LPGseed-label")
	(hide-window "LPGseed-entry")
	(hide-window "LPGtime-bound-local-label")
	(hide-window "LPGtime-bound-local-entry")
	(hide-window "LPGnumber-solutions-label")
	(hide-window "LPGnumber-solutions-entry")
	(hide-window "LPGnoise-label")
	(hide-window "LPGnoise-entry")
        ;;(hide-window "LPGcheckbutton-save-plans")
        ;;(hide-window "LPGcheckbutton-verbose")
        (hide-window "LPGcheckbutton-static-noise")
        ;;(hide-window "LPGcheckbutton-maxnoise")
        (hide-window "LPGcheckbutton-lowmemory")
        (hide-window "LPGcheckbutton-nobestfirst")
        (hide-window "LPGcheckbutton-onlybestfirst")
        (hide-window "LPGcheckbutton-timesteps")

        (show-window "SAYPHIheuristic-label")
        (show-window "SAYPHIalgorithm-label")
	(show-window "METRICheuristic-entry")
	(show-window "METRICalgorithm-entry")
	(show-window "METRICw-g-entry")
	(show-window "METRICw-h-entry")
	(show-window "METRICw-g-label")
	(show-window "METRICw-h-label")
	(show-window "METRICpoda-entry")
	(show-window "METRICpoda-label")
	(show-window "METRICmax-tam-label")
	(show-window "METRICmax-tam-entry")
	(show-window "METRICcheckbutton-metric-optimice")))

;; comprueba si el planificador elegido es sgplan para mostrar los datos del planificador sgplan.
(if (eq 'sgplan (read-from-string (gtk_combo_box_get_active_text (gethash "combobox1" widgets))))
   (progn	
        (hide-window "SAYPHIheuristic-entry")
        (hide-window "SAYPHIalgorithm-entry")
	(hide-window "IPSScost-type-label")
        (hide-window "IPSScheckbutton-multiple-sols-p")
	(hide-window "IPSScheckbutton-backtrack-with-costs-p")
	(hide-window "IPSScheckbutton-cost-bound")
        (hide-window "IPSScheckbutton-stop-after-problem-p")
	(hide-window "IPSScheckbutton-store-fired-rules")
        (hide-window "IPSScheckbutton-store-plans")
        (hide-window "IPSScheckbutton-run-without-p")
        (hide-window "IPSScost-type-entry")
        (hide-window "IPSScost-bound-entry")
        (hide-window "IPSSspin-verbosity-entry")
        (hide-window "IPSSverbosity-label")
	(hide-window "button-ipss-show-search-tree")
        (hide-window "SAYPHIheuristic-label")
        (hide-window "SAYPHIalgorithm-label")
        (hide-window "label70")
        (hide-window "button-rules")
	(hide-window "IPSSoutput-rules-entry")

	(hide-window "LPGmodoejecucion-label")
	(hide-window "LPGmodoejecucion-entry")
	(hide-window "LPGsearchsteps-label")
	(hide-window "LPGsearchsteps-entry")
	(hide-window "LPGrestarts-label")
	(hide-window "LPGrestarts-entry")
	(hide-window "LPGrepeats-label")
	(hide-window "LPGrepeats-entry")
	(hide-window "LPGseed-label")
	(hide-window "LPGseed-entry")
	(hide-window "LPGtime-bound-local-label")
	(hide-window "LPGtime-bound-local-entry")
	(hide-window "LPGnumber-solutions-label")
	(hide-window "LPGnumber-solutions-entry")
	(hide-window "LPGnoise-label")
	(hide-window "LPGnoise-entry")
        ;;(hide-window "LPGcheckbutton-save-plans")
        ;;(hide-window "LPGcheckbutton-verbose")
        (hide-window "LPGcheckbutton-static-noise")
        ;;(hide-window "LPGcheckbutton-maxnoise")
        (hide-window "LPGcheckbutton-lowmemory")
        (hide-window "LPGcheckbutton-nobestfirst")
        (hide-window "LPGcheckbutton-onlybestfirst")
        (hide-window "LPGcheckbutton-timesteps")

	(hide-window "METRICheuristic-entry")
	(hide-window "METRICalgorithm-entry")
	(hide-window "METRICw-g-entry")
	(hide-window "METRICw-h-entry")
	(hide-window "METRICw-g-label")
	(hide-window "METRICw-h-label")
	(hide-window "METRICpoda-entry")
	(hide-window "METRICpoda-label")
	(hide-window "METRICmax-tam-label")
	(hide-window "METRICmax-tam-entry")
	(hide-window "METRICcheckbutton-metric-optimice")))

(if (eq 'lpg (read-from-string (gtk_combo_box_get_active_text (gethash "combobox1" widgets))))
      (progn	
        (hide-window "SAYPHIheuristic-entry")
        (hide-window "SAYPHIalgorithm-entry")

	(hide-window "IPSScost-type-label")
        (hide-window "IPSScheckbutton-multiple-sols-p")
	(hide-window "IPSScheckbutton-backtrack-with-costs-p")
	(hide-window "IPSScheckbutton-cost-bound")
        (hide-window "IPSScheckbutton-stop-after-problem-p")
	(hide-window "IPSScheckbutton-store-fired-rules")
        (hide-window "IPSScheckbutton-store-plans")
        (hide-window "IPSScheckbutton-run-without-p")
        (hide-window "IPSScost-type-entry")
        (hide-window "IPSScost-bound-entry")
        (hide-window "IPSSspin-verbosity-entry")
        (hide-window "IPSSverbosity-label")
	(hide-window "button-ipss-show-search-tree")

        (hide-window "SAYPHIheuristic-label")
        (hide-window "SAYPHIalgorithm-label")
        (hide-window "label70")
        (hide-window "button-rules")
	(hide-window "IPSSoutput-rules-entry")

	(show-window "LPGmodoejecucion-label")
	(show-window "LPGmodoejecucion-entry")
	(show-window "LPGsearchsteps-label")
	(show-window "LPGsearchsteps-entry")
	(show-window "LPGrestarts-label")
	(show-window "LPGrestarts-entry")
	(show-window "LPGrepeats-label")
	(show-window "LPGrepeats-entry")
	(show-window "LPGseed-label")
	(show-window "LPGseed-entry")
	(show-window "LPGtime-bound-local-label")
	(show-window "LPGtime-bound-local-entry")
	(show-window "LPGnumber-solutions-label")
	(show-window "LPGnumber-solutions-entry")
	(show-window "LPGnoise-label")
	(show-window "LPGnoise-entry")
        ;;(show-window "LPGcheckbutton-save-plans")
        ;;(show-window "LPGcheckbutton-verbose")
        (show-window "LPGcheckbutton-static-noise")
        ;;(show-window "LPGcheckbutton-maxnoise")
        (show-window "LPGcheckbutton-lowmemory")
        (show-window "LPGcheckbutton-nobestfirst")
        (show-window "LPGcheckbutton-onlybestfirst")
        (show-window "LPGcheckbutton-timesteps")

	(hide-window "METRICheuristic-entry")
	(hide-window "METRICalgorithm-entry")
	(hide-window "METRICw-g-entry")
	(hide-window "METRICw-h-entry")
	(hide-window "METRICw-g-label")
	(hide-window "METRICw-h-label")
	(hide-window "METRICpoda-entry")
	(hide-window "METRICpoda-label")
	(hide-window "METRICmax-tam-label")
	(hide-window "METRICmax-tam-entry")
	(hide-window "METRICcheckbutton-metric-optimice")))

(interface-trace t "v1-w-ipss initialized")
1)


;; --------------------------------------------------------------------------------------------
;; window vl-w-lm
;; inicializa la pantalla de aprendizaje de macros
;; --------------------------------------------------------------------------------------------
(defun initialize-v1-w-lm ()

;; oculta el resto de pantallas
(hide-window "fixed-v1-lr")
(hide-window "fixed-v1-ipss")
(hide-window "fixed-v1-trans")

;; muestra la pantalla correspondiente al aprendizaje de macros
(show-window "fixed-v1-lm")
(interface-trace t "v1-w-lm initialized")
1)


;; --------------------------------------------------------------------------------------------
;; window vl-w-trans
;; inicializa la pantalla de aprendizaje de traduccion
;; --------------------------------------------------------------------------------------------
(defun initialize-v1-w-trans ()

;; oculta el resto de pantallas
(hide-window "fixed-v1-lr")
(hide-window "fixed-v1-ipss")
(hide-window "fixed-v1-lm")

;; muestra la pantalla correspondiente a la pantalla de traduccion
(show-window "fixed-v1-trans")
(interface-trace t "v1-w-lm initialized")
1)


;; --------------------------------------------------------------------------------------------
;; borra del string pasado por parametro todos los acentos
;; --------------------------------------------------------------------------------------------

(defun delete_invalid_chars (str)
  (let ((long (length str))
	(char_c)
	(new_string ""))
    
    ;; recorremos toda la cadena de texto str
    (dotimes (i long new_string)
    	(progn
	  ;; guardamos en char_c el caracter actual de str
	  (setf char_c (char-code (elt str i)))

	  ;; si char_c tiene un codigo ascii mayor que 128
	  (if (> char_c 128)

                ;; si el caracter actual es un acento se cambia por su vocal sin acento en la variable new_string
		(if (eq char_c (char-code '#\)) (setf new_string (concatenate 'string new_string "a"))
	      	(if (eq char_c (char-code '#\)) (setf new_string (concatenate 'string new_string "e"))
	      	(if (eq char_c (char-code '#\)) (setf new_string (concatenate 'string new_string "i"))
	      	(if (eq char_c (char-code '#\)) (setf new_string (concatenate 'string new_string "o"))
	      	(if (eq char_c (char-code '#\)) (setf new_string (concatenate 'string new_string "u")))))))

	      ;; si no se corresponde con ningun acento el caracter se aade a la variable new_string
	      (setf new_string (concatenate 'string new_string (string (elt str i))))
	  )
    )
  )
  )
  )


;; --------------------------------------------------------------------------------------------
;; window vl-w-ipss
;; ejecuta la planificacion IPSS
;; --------------------------------------------------------------------------------------------

(defun execute-plan-ipss()
(declare (special widgets))

(if (not(eq *planner* 'ipss))
    ;;(load (concatenate 'string *my-prodigy-pathname* "loader.lisp"))
    0)

(setf *planner* 'ipss)
(let  (

;(planner (read-from-string (gtk_combo_box_get_active_text (gethash "combobox1" widgets))))

;; se asigna a la variable time-bound el valor del campo "limite tiempo"
(time-bound (read-from-string (gtk_entry_get_text (gethash "IPSStime-bound-entry" widgets))))
       
       ;; se asigna a la variable cr-file el fichero de reglas de salida indicado
       (cr-file (gtk_entry_get_text (gethash "IPSSoutput-rules-entry" widgets)))

       ;; se asigna a la variable cost-bound el valor del campo "limite coste" si este esta seleccionado
       (cost-bound (if (= 1 (gtk_toggle_button_get_active (gethash "IPSScheckbutton-cost-bound" widgets)))
                       (read-from-string (gtk_entry_get_text (gethash "IPSScost-bound-entry" widgets)))
                      nil))

      ;; se asigna a la variable cost-type el valor del campo "Tipo coste"
      (cost-type   (read-from-string (gtk_entry_get_text (gethash "IPSScost-type-entry" widgets))))
      (domain)
      (probset))
 ;     (print planner) 

   (setf *si-solucion* nil)
   ;; variables de hamlet
   (setf *hamlet-output-level* (read-from-string 
                    (gtk_entry_get_text (gethash "IPSSspin-verbosity-entry" widgets))))

   ;;dominio y conjunto de problemas
   (if  (or (and (eq *language* 'es) (setq domain (read-obligatory-field "IPSSdomain-entry" "v1-w-lr" "No se ha especificado el dominio"))) (and (eq *language* 'en) (setq domain (read-obligatory-field "IPSSdomain-entry" "v1-w-lr" "Domain is necessary"))))
     (if (or (and (eq *language* 'es) (setq probset  (read-obligatory-field "IPSSproblem-entry" "v1-w-lr" "No se ha especificado un problema o conjunto de problemas"))) (and (eq *language* 'en) (setq probset  (read-obligatory-field "IPSSproblem-entry" "v1-w-lr" "Problem or probset is necessary"))))

	
       ;; se ejecutan las siguientes acciones
       (progn 
	 (print (format nil "vamos a ver el dominio ~a~% y los problemas ~a~%" domain probset))
         ;rules file
         (if (equal cr-file "") (setq cr-file nil))
	
	(setf *valid-planner* 0) 
        (setf *valid-domain* 0)
	(setf *valid-problem* 0)
	(setf *problem-number* 0)
	(setf *unsolved-problems* nil)
	(setf path-dominio "/")
	(dolist (parte (butlast (pathname-directory *dominio*)))
					(if (not (equal parte ':ABSOLUTE))  
					(setf path-dominio (concatenate 'string path-dominio parte "/"))))
	(setf path-problemas "/")
	(dolist (parte (pathname-directory *problemas*))
					(if (not (equal parte ':ABSOLUTE))  
					(setf path-problemas (concatenate 'string path-problemas parte "/"))))

	(setf nombre-dominio (last (pathname-directory *dominio*)))
	(setf archivo-dominio (format nil "~a.~a" (pathname-name *dominio*) (pathname-type *dominio*)))
	(setf archivo-problemas (format nil "~a" (pathname-name *problemas*) ))
	
	(print "path del domino")
	(print path-dominio)
	(print "nombre del dominio")
	(print nombre-dominio)
	(print "archivo del dominio")
	(print archivo-dominio)
	(print "path del problema")
	(print path-problemas)
	(print "archivo del problema")
	(print archivo-problemas)
	(print probset)
	(setf *problem-path* path-problemas)
	(print "empieza ejecucion hamlet")

        (handler-case 

		;; se ejecuta hamlet-test de la clase load-problems.lisp y se le pasan por parametros las variables
 		;; correspondientes a la interfaz IPSS
		(hamlet-test (read-from-string (car nombre-dominio)) (if *interface-probset-p* probset (read-from-string probset))

                           :test-time-bound time-bound
;;			 (refinep nil)
			 :run-without-p (= 1 (gtk_toggle_button_get_active (gethash "IPSScheckbutton-run-without-p" widgets)))
			 :multiple-sols-p (= 1 (gtk_toggle_button_get_active (gethash "IPSScheckbutton-multiple-sols-p" widgets)))
;;			 (linearp nil)
			 :cr-file cr-file
;;			 (probabilities nil) 
;;			 (use-probabilities nil)
;;			 (output-file nil)
			 :probset-p *interface-probset-p*   
			 :backtrack-with-costs-p  (= 1 (gtk_toggle_button_get_active (gethash "IPSScheckbutton-backtrack-with-costs-p" widgets)))
;;			 (make-span-p nil)
;;			 (max-MS most-positive-fixnum)
;;			 (include-resources-p nil)
;;			 (num-op 500)
;;			 (act-dist 10000)
;;			 (output-format 'lisp)
			 :store-fired-rules-p *interface-store-fired-crs-p*
			 :cost-bound cost-bound
			 :cost-type cost-type)

		;; se comprueba si se ha producido algun error y se muestra por pantalla el error sucedido
		(error (signal)
		     	(format t "~d~%" signal)
			(if (eq *valid-domain* 0)
			(error-dialog nil (delete_invalid_chars (format nil "Se ha producido un error en el dominio~%~d~%" signal)))
			(if (eq *valid-problem* 0)
			  (error-dialog nil (delete_invalid_chars (format nil "Se ha producido un error en el problema: ~d~%~d~%" *problem-number* signal)))
			(error-dialog nil (delete_invalid_chars (format nil "~d~%")))))
		 )
	  )
	  
	  ;; muestra los problemas que no se han solucionado
	  (print *unsolved-problems*)

	  ;; si existe algun problema que no se ha solucionado
	  (if (> (length *unsolved-problems*) 0)

		
	    	(if *interface-probset-p*
		  (progn
		    ;; almacena en la variable string-unsolved todos los problemas no solucionados
		    (setf string-unsolved "")
		    (dolist (problem *unsolved-problems*)
		      (setf string-unsolved (concatenate 'string string-unsolved (format nil "~a" problem)", "))
		      )
		   
		   ;; Se muestra una pantalla de error donde se indica para que problemas no se ha encontrado solucion
		   (if (eq *language* 'es)
		   	(error-dialog nil (format nil "No se ha encontrado solucion para los problemas: ~% ~a" string-unsolved))
			(error-dialog nil (format nil "No found solution for problems: ~% ~a" string-unsolved)))
		   
		   )
		  
		  ;; Se muestra una pantalla indicando que no se ha encontrado solucion al problema
		  (if (eq *language* 'es)
		  	(error-dialog nil (format nil "No se ha encontrado solucion al problema"))
			(error-dialog nil (format nil "No found solution for problem")))
		  )

		;; si el la planificacion ha sido valida
	  	(if (eq *valid-planner* 1)
		        
			;; Se muestra una pantalla indicando que la planificacion ha sido correcta
			(if (eq *language* 'es)
				(progn
					(setf *si-solucion* t)
	      				(information-dialog nil (format nil "La planificacion ha concluido satisfactoriamente"))
				)
				(progn
					(setf *si-solucion* t)
					(information-dialog nil (format nil "The planner has finished satisfactorily"))))
	  ))
	  
          1)   
          ;;else probset 
          0)
         ;;else domain
         0))
)


;; --------------------------------------------------------------------------------------------
;; window vl-w-ipss
;; ejecuta la planificacion SAYPHI
;; --------------------------------------------------------------------------------------------

(defun execute-plan-sayphi()
(declare (special widgets))
(if (not(eq *planner* 'sayphi))
    (progn
    (print 8)
    (load (concatenate 'string *my-sayphi-pathname* "loader.lisp"))
    (print 9)
    )
    0)
(setf *planner* 'sayphi)
(let  (

;(planner (read-from-string (gtk_combo_box_get_active_text (gethash "combobox1" widgets))))

;; obtiene de la interfaz los valores de la heuristica y del algortimo elegidos
(*say-heuristic* (read-from-string (gtk_combo_box_get_active_text (gethash "SAYPHIheuristic-entry" widgets))))
(*say-algorithm* (read-from-string (gtk_combo_box_get_active_text (gethash "SAYPHIalgorithm-entry" widgets))))

;; obtiene el tiempo limite de aprendizaje
(time-bound (read-from-string (gtk_entry_get_text (gethash "IPSStime-bound-entry" widgets))))
       
      (domain)
      (probset))
   (setf *si-solucion* nil)
 ;     (print planner) 

   (setf *say-output* (read-from-string 
                    (gtk_entry_get_text (gethash "IPSSspin-verbosity-entry" widgets))))

   ;; dominio y conjunto de pruebas
   (if  (or (and (eq *language* 'es) (setq domain (read-obligatory-field "IPSSdomain-entry" "v1-w-lr" "No se ha especificado el dominio"))) (and (eq *language* 'en) (setq domain (read-obligatory-field "IPSSdomain-entry" "v1-w-lr" "Domain is necessary"))))
     (if (or (and (eq *language* 'es) (setq probset  (read-obligatory-field "IPSSproblem-entry" "v1-w-lr" "No se ha especificado un problema o conjunto de problemas"))) (and (eq *language* 'en) (setq probset  (read-obligatory-field "IPSSproblem-entry" "v1-w-lr" "Problem or probset is necessary"))))
       (progn 
	 (print "1")
	 (print "2")
	 (print "3")
	 (setf domain *dominio*)
	 (setf probset *problemas*)
       
	(setf *valid-planner* 0) 
        (setf *valid-domain* 0)
	(setf *valid-problem* 0)
	(setf *problem-number* 0)
	(setf *unsolved-problems* nil)
        (setf *say-timeout* 0)        
        
        (print *say-heuristic*)
        (print *say-algorithm*)
        
        (print domain)
	(print probset)
	;;(say-domain domain)
	;;(print (concatenate 'string probset ".pddl"))
	;;(prob (concatenate 'string probset ".pddl"))
	(if (equal *say-algorithm* 'dfbnb)
		(setf *say-algorithm* 'hc-bnb))       
 
        (print time-bound)
	(setf lista (list ':timeout time-bound  ':heuristic *say-heuristic* ':algorithm *say-algorithm*))
	(print lista)

	(handler-case (sayphi domain probset :plan-args lista)
		(error (signal)  
		     	(format t "~d~%" signal)
			(if (eq *valid-domain* 0)
			(setf error (delete_invalid_chars (format nil "Se ha producido un error en el dominio~%~d~%" signal)))
			(if (eq *valid-problem* 0)
			  (set error (delete_invalid_chars (format nil "Se ha producido un error en el problema: ~d~%~d~%" *problem-number* signal)))
			(setf error (delete_invalid_chars (format nil "~d~%")))))
		 ))



	;;(sayphi domain probset :plan-args lista)
	(if (and (solution-p *say-solution*) (solution-found *say-solution*))
			(if (eq *language* 'es)
				(progn
					(setf *si-solucion* t)
	      				(information-dialog nil (format nil "La planificacion ha concluido
 satisfactoriamente"))
				)
				(progn
					(setf *si-solucion* t)
					(information-dialog nil (format nil "The planner has finished satisfactorily"))))
			(if (eq *language* 'es)
	      			(information-dialog nil (format nil "No se ha encontrado solucion al problema"))
				(information-dialog nil (format nil "No found solution for problem")))
	  )
          1)   
          ;;else probset 
          0)
         ;;else domain
         0))
)


;; --------------------------------------------------------------------------------------------
;; window vl-w-ipss
;; ejecuta la planificacion METRIC-FF
;; --------------------------------------------------------------------------------------------

(defun execute-plan-metric-ff()
(declare (special widgets))
(if (not(eq *planner* 'metric-ff))
	(setf *planner* 'metric-ff)
)
(let ((opciones)
      (wg (gtk_entry_get_text (gethash "METRICw-g-entry" widgets)))
      (wh (gtk_entry_get_text (gethash "METRICw-h-entry" widgets)))
      (max-tam (gtk_entry_get_text (gethash "METRICmax-tam-entry" widgets)))
      (metric-opt (= 1 (gtk_toggle_button_get_active (gethash "METRICcheckbutton-metric-optimice" widgets))))
      (domain)
      (probset)
      (solucion 0)
      (rules (gtk_entry_get_text (gethash "IPSSoutput-rules-entry" widgets)))
      (contenido)
      (correcta)
      (time-bound (read-from-string (gtk_entry_get_text (gethash "IPSStime-bound-entry" widgets))))
      (ff-heuristic (read-from-string (gtk_combo_box_get_active_text (gethash "METRICheuristic-entry" widgets))))
      (ff-algorithm (read-from-string (gtk_combo_box_get_active_text (gethash "METRICalgorithm-entry" widgets))))
      (ff-poda (read-from-string (gtk_combo_box_get_active_text (gethash "METRICpoda-entry" widgets)))))
      

   (setf *si-solucion* nil)
   ;;dominio y conjunto de problemas
   (if  (or (and (eq *language* 'es) (setq domain (read-obligatory-field "IPSSdomain-entry" "v1-w-lr" "No se ha especificado el dominio"))) (and (eq *language* 'en) (setq domain (read-obligatory-field "IPSSdomain-entry" "v1-w-lr" "Domain is necessary"))))
     (if (or (and (eq *language* 'es) (setq probset  (read-obligatory-field "IPSSproblem-entry" "v1-w-lr" "No se ha especificado un problema o conjunto de problemas"))) (and (eq *language* 'en) (setq probset  (read-obligatory-field "IPSSproblem-entry" "v1-w-lr" "Problem or probset is necessary"))))

(progn

(if (or (equal rules "") (equal rules "nil") )
	(setf *reglas* nil)
)

(if (not (null *reglas*))
    	(setf opciones (concatenate 'string opciones "-r " *reglas* " "))
)

(if (equal 'best-first ff-algorithm)
      (progn
      	(setf opciones (concatenate 'string opciones "-E "))
	(case ff-poda
		(stop-at-first-solution (setf opciones (concatenate 'string opciones "-P 0 ")))
		(Anytime-WA*-prunning-by-g-values (setf opciones (concatenate 'string opciones "-P 1 ")))
		(Anytime-WA*-pruning-by-g-values-and-using-helpful-actions-prune (setf opciones (concatenate 'string opciones "-P 2 ")))
		(Anytime-WA*-prunning-by-f-values (setf opciones (concatenate 'string opciones "-P 3 ")))
		(Anytime-WA*-pruning-by-f-values-and-using-helpful-actions-prune (setf opciones (concatenate 'string opciones "-P 4 ")))
	)
	(if (and (numberp max-tam) (>= 0 max-tam)) (setf opciones (concatenate 'string opciones "-m ~d " max-tam)))
      )
)

(case ff-heuristic
	(Metric-FF (setf opciones (concatenate 'string opciones "-V 0 ")))
        (graphplan-with-fixed-cost-step (setf opciones (concatenate 'string opciones "-V 1 ")))
        (cheapest-applicable-action-with-cost>0 (setf opciones (concatenate 'string opciones "-V 2 ")))
        (cheapest-applicable-action-with-cost>=0 (setf opciones (concatenate 'string opciones "-V 3 ")))
        (cheapest-applicable-action-with-cost>=0-including-delayed-actions (setf opciones (concatenate 'string opciones "-V 4 ")))
        (cheapest-applicable-action-with-cost>0-including-delayed-actions (setf opciones (concatenate 'string opciones "-V 5 ")))
)

(if (and (numberp wg) (>= wg 0))
	(setf opciones (concatenate 'string opciones "-g " wg " "))
)

(if (and (numberp wh) (>= wh 0))
	(setf opciones (concatenate 'string opciones "-h " wh " "))
)
(if metric-opt
	(setf opciones (concatenate 'string opciones "-O "))
)
	(setf opciones (concatenate 'string opciones (format nil "-c ~a " time-bound)))


;;(print (cl-ppcre:scan "(a)*b" "xaaabd"))
;;(run-shell-command (concatenate 'string *my-metricff-pathname* "system/planner/ff -o " *dominio* " -f " *problemas* " " opciones " | grep \"\\([0-9]\\+: \\|total time\\|total cost\\|egmentation\\|FALSE\\|step[ ]\\+0:\\)\"" "| sed -e \"s/ff: goal can be simplified to FALSE. No plan will solve it/ nil/\"" "| sed -e \"s/step $/ ((*finish*/\"" "| sed -e \"s/step[ ]\\+0:/ ((/\"" "| sed -e \"s/[0-9]\\+:/)(/\" " "| sed -e \"s/\\([0-9]\\+.[0-9]\\+\\) seconds total time/)) \\1/\"" " | sed -e \"s/total cost:[ ]\\([0-9]\\+.[0-9]\\+\\)/ \\1\\)/\""  "> " *my-metricff-pathname* "ff-output.lisp" ))

(run-shell-command (concatenate 'string *my-metricff-pathname* "system/planner/ff -o " *dominio* " -f " *problemas* " " opciones "> " *my-metricff-pathname* "ff-output.lisp" ))

	(with-open-file (stream (concatenate 'string *my-metricff-pathname* "ff-output.lisp") :direction :input :if-does-not-exist nil)
		(with-open-file (stream1 (concatenate 'string *my-metricff-pathname* "ff-output2.lisp") :direction :output :if-exists :supersede :if-does-not-exist :create)
		
		    (if stream
			(do ((leido (read-line stream nil 'eof)
			    (read-line stream nil 'eof)))
			    ((eq leido 'eof))
			    ;;comprobaciones de la linea
			    (progn
	        	   	 (setf correcta (cl-ppcre::scan "([0-9]+:|total time|total cost|egmentation|FALSE|step[ ]+0:)" leido))
			   	 (if correcta
					(format stream1 "~a~%" leido)
			   	 )
			    )))
		)

	)


	(with-open-file (stream (concatenate 'string *my-metricff-pathname* "ff-output2.lisp") :direction :input :if-does-not-exist nil)
		(with-open-file (stream1 (concatenate 'string *my-metricff-pathname* "ff-output.lisp") :direction :output :if-exists :supersede :if-does-not-exist :create)
		
		    (if stream
			(do ((leido (read-line stream nil 'eof)
			    (read-line stream nil 'eof)))
			    ((eq leido 'eof))
			    ;;comprobaciones de la linea
			    (progn
	        	   	 (setf correcta (cl-ppcre::scan "step[ ]+0:" leido))
				 (if correcta
					(progn
						(setf leido (cl-ppcre::regex-replace "step[ ]+0:" leido "((("))
						(setf solucion (+ solucion 1))
					)
				 )
				 (setf correcta (cl-ppcre::scan "[0-9]+:" leido))
				 (if correcta
					(setf leido (cl-ppcre::regex-replace "[0-9]+:" leido ")("))
				 )
				 (setf correcta (cl-ppcre::scan "seconds total time" leido))
				 (if correcta
			
					(setf leido (cl-ppcre::register-groups-bind (v1) ("([0-9]+\.[0-9]+) seconds total time" leido) (format nil ")) ~a " v1)))
				 )
				 (setf correcta (cl-ppcre::scan "total cost:[ ]+[0-9]+\.[0-9]+" leido))
				 (if correcta

					(setf leido (cl-ppcre::register-groups-bind (v1) ("total cost:[ ]+([0-9]+\.[0-9]+)" leido) (format nil "~a)" v1)))
				 )
				 (format stream1 "~a~%" leido) 
			    )))
		)
	)

	(with-open-file (stream (concatenate 'string *my-metricff-pathname* "ff-output.lisp") :direction :input :if-does-not-exist nil)
		(with-open-file (stream1 (concatenate 'string *my-metricff-pathname* "ff-output2.lisp") :direction :output :if-exists :supersede :if-does-not-exist :create)

		    (if stream
			(do ((leido (read-line stream nil 'eof)
			    (read-line stream nil 'eof)))
			    ((eq leido 'eof))
			    ;;comprobaciones de la linea
			    (setf contenido (concatenate 'string contenido (format nil "~a" leido)))
			))
		    (setf contenido (cl-ppcre::regex-replace "\\)\\)\\)" contenido ")" ))
		    (format stream1 "(~a)" contenido)
		)

	)


	;;(shell (concatenate 'string "rm " *my-metricff-pathname* "ff-output*.lisp"))
	(print (format nil "el contenido es: ~%"))
	(print contenido)
	(if (or (null contenido) (equal contenido "") (= solucion 0))
			(if (eq *language* 'es)
	      			(information-dialog nil (format nil "No se ha encontrado solucion al problema"))
				(information-dialog nil (format nil "No found solution for problem")))
			(if (eq *language* 'es)
				(progn
					(setf *si-solucion* t)
	      				(information-dialog nil (format nil "La planificacion ha concluido satisfactoriamente"))
					
				)
				(progn
					(setf *si-solucion* t)
					(information-dialog nil (format nil "The planner has finished satisfactorily"))))
	  )
	1);;close progn
	0);;else probset
	0);;else domain
)
)



;; --------------------------------------------------------------------------------------------
;; window vl-w-ipss
;; ejecuta la planificacion SGPLAN
;; --------------------------------------------------------------------------------------------

(defun execute-plan-sgplan()
(declare (special widgets))
(if (not(eq *planner* 'sgplan))
	(setf *planner* 'sgplan)
)
(let ((domain)
      (probset)
      (solucion 0)
      (correcta)
      (contenido)
      (time-bound (read-from-string (gtk_entry_get_text (gethash "IPSStime-bound-entry" widgets)))))
      
   (setf *si-solucion* nil)
   ;;dominio y conjunto de problemas
   (if  (or (and (eq *language* 'es) (setq domain (read-obligatory-field "IPSSdomain-entry" "v1-w-lr" "No se ha especificado el dominio"))) (and (eq *language* 'en) (setq domain (read-obligatory-field "IPSSdomain-entry" "v1-w-lr" "Domain is necessary"))))
     (if (or (and (eq *language* 'es) (setq probset  (read-obligatory-field "IPSSproblem-entry" "v1-w-lr" "No se ha especificado un problema o conjunto de problemas"))) (and (eq *language* 'en) (setq probset  (read-obligatory-field "IPSSproblem-entry" "v1-w-lr" "Problem or probset is necessary"))))

(progn

;;(print (cl-ppcre:scan "(a)*b" "xaaabd"))
;;(run-shell-command (concatenate 'string *my-metricff-pathname* "system/planner/ff -o " *dominio* " -f " *problemas* " " opciones " | grep \"\\([0-9]\\+: \\|total time\\|total cost\\|egmentation\\|FALSE\\|step[ ]\\+0:\\)\"" "| sed -e \"s/ff: goal can be simplified to FALSE. No plan will solve it/ nil/\"" "| sed -e \"s/step $/ ((*finish*/\"" "| sed -e \"s/step[ ]\\+0:/ ((/\"" "| sed -e \"s/[0-9]\\+:/)(/\" " "| sed -e \"s/\\([0-9]\\+.[0-9]\\+\\) seconds total time/)) \\1/\"" " | sed -e \"s/total cost:[ ]\\([0-9]\\+.[0-9]\\+\\)/ \\1\\)/\""  "> " *my-metricff-pathname* "ff-output.lisp" ))

(run-shell-command (concatenate 'string *my-sgplan-pathname* "system/planner/sgplan522 -o " *dominio* " -f " *problemas* " -cputime " (format nil "~a " time-bound)  "> " *my-sgplan-pathname* "sgplan-output.lisp" ))

	(with-open-file (stream (concatenate 'string *my-sgplan-pathname* "sgplan-output.lisp") :direction :input :if-does-not-exist nil)
		(with-open-file (stream1 (concatenate 'string *my-sgplan-pathname* "sgplan-output2.lisp") :direction :output :if-exists :supersede :if-does-not-exist :create)
		
		    (if stream
			(do ((leido (read-line stream nil 'eof)
			    (read-line stream nil 'eof)))
			    ((eq leido 'eof))
			    ;;comprobaciones de la linea
			    (progn
	        	   	 (setf correcta (cl-ppcre::scan "([0-9]+:|\;[ ]+Time|NrActions|MetricValue)" leido))
			   	 (if correcta
					(format stream1 "~a~%" leido)
			   	 )
			    )))
		)
	)

	(with-open-file (stream (concatenate 'string *my-sgplan-pathname* "sgplan-output2.lisp") :direction :input :if-does-not-exist nil)
		(with-open-file (stream1 (concatenate 'string *my-sgplan-pathname* "sgplan-output.lisp") :direction :output :if-exists :supersede :if-does-not-exist :create)
		
		    (if stream
			(do ((leido (read-line stream nil 'eof)
			    (read-line stream nil 'eof)))
			    ((eq leido 'eof))
			    ;;comprobaciones de la linea
			    (progn
				 (setf correcta (cl-ppcre::scan "\;[ ]+Time [0-9]+\.[0-9]+" leido))
				 (if correcta
					(setf leido (cl-ppcre::register-groups-bind (v1) ("\;[ ]+Time ([0-9]+\.[0-9]+)" leido) (format nil "( ~a " v1)))
  				 )


				 (setf correcta (cl-ppcre::scan "\;[ ]+NrActions [0-9]+" leido))
				 (if correcta
					(setf leido (cl-ppcre::register-groups-bind (v1) ("\;[ ]+NrActions ([0-9]+)" leido) (format nil "~a " v1)))
  				 )


				 (setf correcta (cl-ppcre::scan "\;[ ]+MetricValue [0-9]+" leido))
				 (if correcta
					(setf leido (cl-ppcre::register-groups-bind (v1) ("\;[ ]+MetricValue ([0-9]+\.[0-9]+)" leido) (format nil " ~a (" v1)))
  				 )
				

				 (setf correcta (cl-ppcre::scan "\;[ ]+MetricValue" leido))
				 (if correcta
					(setf leido (cl-ppcre::regex-replace "\;[ ]+MetricValue" leido "0.0 ("))
  				 )

	        	   	 (setf correcta (cl-ppcre::scan "[0-9]+\.[0-9]+:" leido))
				 (if correcta
					(progn
						(setf leido (cl-ppcre::regex-replace "[0-9]+\.[0-9]+:" leido "" ))
						(setf solucion (+ solucion 1))
					)
				 )

	        	   	 (setf correcta (cl-ppcre::scan "\[[0-9]+\.[0-9]+\]" leido))
				 (if correcta
					(progn
						(setf leido (cl-ppcre::regex-replace "\[[0-9]+\.[0-9]+\]" leido "" ))
					)
				 )

	        	   	 (setf correcta (cl-ppcre::scan "\[[0-9]+\]" leido))
				 (if correcta
					(progn
						(setf leido (cl-ppcre::regex-replace "\[[0-9]+\]" leido "" ))
					)
				 )

				 (format stream1 "~a~%" leido) 
			    )))
		)
	)

	(with-open-file (stream (concatenate 'string *my-sgplan-pathname* "sgplan-output.lisp") :direction :input :if-does-not-exist nil)
		(with-open-file (stream1 (concatenate 'string *my-sgplan-pathname* "sgplan-output2.lisp") :direction :output :if-exists :supersede :if-does-not-exist :create)

		    (if stream
			(do ((leido (read-line stream nil 'eof)
			    (read-line stream nil 'eof)))
			    ((eq leido 'eof))
			    ;;comprobaciones de la linea
			    (setf contenido (concatenate 'string contenido (format nil "~a" leido)))
			))
		    (setf contenido (concatenate 'string contenido (format nil "))")))
		    (format stream1 "~a" contenido)
		)

	)

	(setf contenido (file-to-string "sgplan-output2.lisp" *my-sgplan-pathname*))
	;;(shell (concatenate 'string "rm " *my-sgplan-pathname* "sgplan-output*.lisp"))
	(print (format nil "el contenido es: ~%"))
	(print contenido)
	(if (or (null contenido) (equal contenido "") (= solucion 0))
			(if (eq *language* 'es)
	      			(information-dialog nil (format nil "No se ha encontrado solucion al problema"))
				(information-dialog nil (format nil "No found solution for problem")))
			(if (eq *language* 'es)
				(progn
					(setf *si-solucion* t)
		      			(information-dialog nil (format nil "La planificacion ha concluido satisfactoriamente"))
				)
				(progn
					(setf *si-solucion* t)
					(information-dialog nil (format nil "The planner has finished satisfactorily"))))
	  )
	1);;close progn
	0);;else probset
	0);;else domain
)
)



;; --------------------------------------------------------------------------------------------
;; window vl-w-ipss
;; ejecuta la planificacion LPG
;; --------------------------------------------------------------------------------------------

(defun execute-plan-lpg()
(declare (special widgets))
(if (not(eq *planner* 'lpg))
	(setf *planner* 'lpg)
)
(let ((opciones)
      ;;(noout (= 0 (gtk_toggle_button_get_active (gethash "LPGcheckbutton-save-plans" widgets))))
      ;;(voff  (= 1 (gtk_toggle_button_get_active (gethash "LPGcheckbutton-verbose" widgets))))
      (static-noise (= 1 (gtk_toggle_button_get_active (gethash "LPGcheckbutton-static-noise" widgets))))
      ;;(max-noise (= 1 (gtk_toggle_button_get_active (gethash "LPGcheckbutton-maxnoise" widgets))))
      (low-memory (= 1 (gtk_toggle_button_get_active (gethash "LPGcheckbutton-lowmemory" widgets))))
      (nobestfirst (= 1 (gtk_toggle_button_get_active (gethash "LPGcheckbutton-nobestfirst" widgets))))
      (onlybestfirst (= 1 (gtk_toggle_button_get_active (gethash "LPGcheckbutton-onlybestfirst" widgets))))
      (time-steps (= 1 (gtk_toggle_button_get_active (gethash "LPGcheckbutton-timesteps" widgets))))
      (time-bound (read-from-string (gtk_entry_get_text (gethash "IPSStime-bound-entry" widgets))))      
      (time-bound-local (read-from-string (gtk_entry_get_text (gethash "LPGtime-bound-local-entry" widgets))))      
      (search-steps (read-from-string (gtk_entry_get_text (gethash "LPGsearchsteps-entry" widgets))))      
      (number-solutions (read-from-string (gtk_entry_get_text (gethash "LPGnumber-solutions-entry" widgets))))
      (restarts (read-from-string (gtk_entry_get_text (gethash "LPGrestarts-entry" widgets))))
      (repeats (read-from-string (gtk_entry_get_text (gethash "LPGrepeats-entry" widgets))))
      (seed (read-from-string (gtk_entry_get_text (gethash "LPGseed-entry" widgets))))
      (noise (read-from-string (gtk_entry_get_text (gethash "LPGnoise-entry" widgets))))      
      (modo-ejecucion (read-from-string (gtk_combo_box_get_active_text (gethash "LPGmodoejecucion-entry" widgets))))
      (domain)
      (nombre-directorio "temporales/")
      (probset)
      (solucion 0)
      (correcta)
      (contenido)
      (comando))
      
  (setf *si-solucion* nil)
  ;;dominio y conjunto de problemas
   (if  (or (and (eq *language* 'es) (setq domain (read-obligatory-field "IPSSdomain-entry" "v1-w-lr" "No se ha especificado el dominio"))) (and (eq *language* 'en) (setq domain (read-obligatory-field "IPSSdomain-entry" "v1-w-lr" "Domain is necessary"))))
     (if (or (and (eq *language* 'es) (setq probset  (read-obligatory-field "IPSSproblem-entry" "v1-w-lr" "No se ha especificado un problema o conjunto de problemas"))) (and (eq *language* 'en) (setq probset  (read-obligatory-field "IPSSproblem-entry" "v1-w-lr" "Problem or probset is necessary"))))


(progn
(print (format nil "hasta aqui si~%"))
(setf comando (concatenate 'string "rm " *my-lpg-pathname* nombre-directorio "temporal.SOL"))
(SHELL comando)
(if (equal 'incremental modo-ejecucion)
	(setf opciones (concatenate 'string (format nil "-n ~d " number-solutions) "-noout "))
)
(if (equal 'speed modo-ejecucion)
	(setf opciones (concatenate 'string "-speed " "-noout "))
)
(if (equal 'quality modo-ejecucion)
	(setf opciones (concatenate 'string "-quality " "-out " *my-lpg-pathname* nombre-directorio "temporal "))
)


(setf opciones (concatenate 'string opciones (format nil "-cputime ~s -cputime_localsearch ~s -search_steps ~s -restarts ~s -repeats ~s -seed ~s -noise ~s " time-bound time-bound-local search-steps restarts repeats seed noise)))

;(if noout
;	(if (not (equal 'quality modo-ejecucion))
;		(setf opciones (concatenate 'string opciones "-noout "))
;	)
;	(progn
;		(setf nombre-directorio "planes_obtenidos/")
;		(setf opciones (concatenate 'string opciones "-out " *my-lpg-pathname* nombre-directorio (pathname-name *problemas*) " " ))
;	)
;)

;;(if voff (setf opciones (concatenate 'string opciones "-v off ")))
(if static-noise (setf opciones (concatenate 'string opciones "-static_noise ")))
;;(if max-noise (setf opciones (concatenate 'string opciones "-static_noise ")))
(if low-memory (setf opciones (concatenate 'string opciones "-lowmemory ")))
(if nobestfirst (setf opciones (concatenate 'string opciones "-nobestfirst ")))
(if onlybestfirst (setf opciones (concatenate 'string opciones "-onlybestfirst ")))
(if time-steps (setf opciones (concatenate 'string opciones "-timesteps ")))

(print (format nil "~s~%" opciones))

(run-shell-command (concatenate 'string *my-lpg-pathname* "system/planner/lpg-td-1.0 -o " *dominio* " -f " *problemas* " " opciones "> " *my-lpg-pathname* "lpg-output.lisp" ))

	(with-open-file (stream (concatenate 'string *my-lpg-pathname* "lpg-output.lisp") :direction :input :if-does-not-exist nil)
		(with-open-file (stream1 (concatenate 'string *my-lpg-pathname* "lpg-output2.lisp") :direction :output :if-exists :supersede :if-does-not-exist :create)
		
		    (if stream
			(do ((leido (read-line stream nil 'eof)
			    (read-line stream nil 'eof)))
			    ((eq leido 'eof))
			    ;;comprobaciones de la linea
			    (progn
				 (if (equal 'quality modo-ejecucion)
	        	  	 	 (setf correcta (cl-ppcre::scan "(Total time|Actions|Execution cost)" leido))
	        	  	 	 (setf correcta (cl-ppcre::scan "([0-9]+:|Plan computed|Total time|Actions|Execution cost|Number of actions)" leido))	
				 )
			   	 (if correcta
					(format stream1 "~a~%" leido)
			   	 )
			    )))
		)
	)

(with-open-file (stream (concatenate 'string *my-lpg-pathname* "lpg-output2.lisp") :direction :input :if-does-not-exist nil)
		(with-open-file (stream1 (concatenate 'string *my-lpg-pathname* "lpg-output.lisp") :direction :output :if-exists :supersede :if-does-not-exist :create)
		
		    (if stream
			(do ((leido (read-line stream nil 'eof)
			    (read-line stream nil 'eof)))
			    ((eq leido 'eof))
			    ;;comprobaciones de la linea
			    (progn
	        	   	 (setf correcta (cl-ppcre::scan "Number of actions" leido))
				 (if correcta
						(setf leido (cl-ppcre::regex-replace "Number of actions[ ]+:[ ]+[0-9]+" leido "("))
				 )
				 (setf correcta (cl-ppcre::scan "Plan computed" leido))
				 (if correcta
						(progn
						(setf leido (cl-ppcre::regex-replace "Plan computed:" leido "(("))
						(setf solucion (+ solucion 1))
						)
				)
				(format stream1 "~a~%" leido) 
			    )))
		)
	)

	(with-open-file (stream (concatenate 'string *my-lpg-pathname* "lpg-output.lisp") :direction :input :if-does-not-exist nil)
		(with-open-file (stream1 (concatenate 'string *my-lpg-pathname* "lpg-output2.lisp") :direction :output :if-exists :supersede :if-does-not-exist :create)
		
		    (if stream
			(do ((leido (read-line stream nil 'eof)
			    (read-line stream nil 'eof)))
			    ((eq leido 'eof))
			    ;;comprobaciones de la linea
			    (progn
	        	   	 (setf correcta (cl-ppcre::scan "[0-9]+\.[0-9]+:" leido))
				 (if correcta
					     (progn
						(setf leido (cl-ppcre::regex-replace "[0-9]+\.[0-9]+:" leido ""))
						(setf leido (cl-ppcre::regex-replace "[D:[0-9]+\.[0-9]+;[ ]+C:[0-9]+\.[0-9]+]" leido ""))
					      )
				 )
				(format stream1 "~a~%" leido) 
			    )))
		)
	)

	(with-open-file (stream (concatenate 'string *my-lpg-pathname* "lpg-output2.lisp") :direction :input :if-does-not-exist nil)
		(with-open-file (stream1 (concatenate 'string *my-lpg-pathname* "lpg-output.lisp") :direction :output :if-exists :supersede :if-does-not-exist :create)
		
		    (if stream
			(progn
			(do ((leido (read-line stream nil 'eof)
			    (read-line stream nil 'eof)))
			    ((eq leido 'eof))
			    ;;comprobaciones de la linea
			    (progn
	        	   	 (setf correcta (cl-ppcre::scan "Total time:" leido))
				 (if correcta
						(setf leido (cl-ppcre::regex-replace "Total time:[ ]+" leido ")"))
				 )
				 (setf correcta (cl-ppcre::scan "Actions:" leido))
				 (if correcta
						(setf leido (cl-ppcre::regex-replace "Actions:[ ]+" leido ""))
				)
				 (setf correcta (cl-ppcre::scan "Execution cost:" leido))
				 (if correcta
						(progn
						(setf leido (cl-ppcre::regex-replace "Execution cost:[ ]+" leido ""))
						(setf leido (concatenate 'string leido ")"))
						)
				)

				(format stream1 "~a~%" leido) 
			    ))
			  (format stream1 ")~%")
			  ))
			    
		)
	)	

	(if (equal 'quality modo-ejecucion)
		(with-open-file (stream (concatenate 'string *my-lpg-pathname* nombre-directorio  "temporal.SOL") :direction :input :if-does-not-exist nil)
			(with-open-file (stream1 (concatenate 'string *my-lpg-pathname* "lpg-output2.lisp") :direction :output :if-exists :supersede :if-does-not-exist :create)
		
				(if stream
					(do ((leido (read-line stream nil 'eof)
					    (read-line stream nil 'eof)))
					    ((eq leido 'eof))
					    ;;comprobaciones de la linea
					    (progn
		        	  	 	 (setf correcta (cl-ppcre::scan "[0-9]+:" leido))
						 (if correcta
							(progn
								(setf solucion (+ solucion 1))
								(setf leido (cl-ppcre::regex-replace "[[0-9]+\.[0-9]+]" leido ""))
								(setf leido (cl-ppcre::regex-replace "[0-9]+\.[0-9]+:" leido ""))
								(setf leido (cl-ppcre::regex-replace "[D:[0-9]+\.[0-9]+;[ ]+C:[0-9]+\.[0-9]+]" leido ""))
								(setf leido (cl-ppcre::regex-replace "[0-9]+:" leido ""))
									
								(setf leido (cl-ppcre::regex-replace "[[0-9]+]" leido ""))
							)

						 )
					   	 (if correcta
							(format stream1 "~a~%" leido)
					   	 )
					    )))
			)
		)	
	)

	(if (equal 'quality modo-ejecucion)
		(with-open-file (stream (concatenate 'string *my-lpg-pathname* "lpg-output2.lisp") :direction :input :if-does-not-exist nil)
			(with-open-file (stream2 (concatenate 'string *my-lpg-pathname* "lpg-output.lisp") :direction :input :if-does-not-exist nil)
		
				(if stream
					(progn
						(setf contenido "(((")
						(do ((leido (read-line stream nil 'eof)
						    (read-line stream nil 'eof)))
						    ((eq leido 'eof))
						    ;;comprobaciones de la linea
						    (setf contenido (concatenate 'string contenido leido " "))
						)
					)
				)
				(if stream2
					(progn
						(do ((leido (read-line stream2 nil 'eof)
						    (read-line stream2 nil 'eof)))
						    ((eq leido 'eof))
						    ;;comprobaciones de la linea
						    (setf contenido (concatenate 'string contenido leido " "))
						)
					)

				)

			)
		)
	)

	(if (equal 'quality modo-ejecucion)
		(with-open-file (stream (concatenate 'string *my-lpg-pathname* "lpg-output.lisp") :direction :output :if-exists :supersede :if-does-not-exist :create)
			(progn
				(format stream contenido)					
				(setf contenido "")
			)
		)
	)



	(with-open-file (stream (concatenate 'string *my-lpg-pathname* "lpg-output.lisp") :direction :input :if-does-not-exist nil)
		   (if stream
			(do ((leido (read-line stream nil 'eof)
			    (read-line stream nil 'eof)))
			    ((eq leido 'eof))
			    ;;comprobaciones de la linea
			    (setf contenido (concatenate 'string contenido (format nil "~a " leido)))
			))
	)


	;;(shell (concatenate 'string "rm " *my-lpg-pathname* "lpg-output*.lisp"))
	(print (format nil "el contenido es: ~%"))
	(print contenido)
	(if (or (null contenido) (equal contenido "") (= solucion 0))
			(if (eq *language* 'es)
	      			(information-dialog nil (format nil "No se ha encontrado solucion al problema"))
				(information-dialog nil (format nil "No found solution for problem")))
			(if (eq *language* 'es)
				(progn
					(setf *si-solucion* t)
	      				(information-dialog nil (format nil "La planificacion ha concluido satisfactoriamente"))
				)
				(progn
					(setf *si-solucion* t)
					(information-dialog nil (format nil "The planner has finished satisfactorily"))))
	  )


	1);;close progn
	0);;else probset
	0);;else domain
)
)







;; --------------------------------------------------------------------------------------------
;; window vl-w-ipss
;; ejecuta una planificacion
;; --------------------------------------------------------------------------------------------

(defun execute-plan()
	(declare (special widgets))
	(gtk_button_released (gethash "button-ipss-plan" widgets))
	(gtk_main_iteration)
	(let  ((planner (read-from-string (gtk_combo_box_get_active_text (gethash "combobox1" widgets)))))
 		(if (eq planner 'IPSS) (execute-plan-ipss) 0)
		(if (eq planner 'SAYPHI) (execute-plan-sayphi) 0)
 		(if (eq planner 'METRIC-FF) (execute-plan-metric-ff) 0)
	        (if (eq planner 'SGPLAN) (execute-plan-sgplan) 0)
	        (if (eq planner 'LPG) (execute-plan-LPG) 0)
	)
)	 
	          
(defun destroy-all ()
 (gtk_main_quit)
  0
;; (exit)
)


;; --------------------------------------------------------------------------------------------
;; window vl-w-trans
;; Button translate domain (pddl-domain-to-prodigy)
;; signal clicked
;; --------------------------------------------------------------------------------------------
(defun translate-domain-pddl2prodigy ()
(let ((real-world-path *world-path*)
      (domain))
  ;; cambiamos el path actual por el path de sayphi
  (setf *world-path* (format nil "~adomains/" *my-sayphi-pathname*))
  
  ;; si no se ha especificado el dominio se muestra un mensaje de error
  (if (setq domain (concatenate 'string (read-obligatory-field "TRANSpddldomain-entry1"  nil "PDDL domain is necessary") "/" "domain.pddl"))
    
   (progn 

    ;; se resetean las variables del dominio
    ;;(reset-domain-variables nil (format nil "~a" (car (last (pathname-directory domain)))))

    (let () 
         (handler-case (progn

;;----------------------------------------inicio parte david montero---------------------------------------------------------

				;; almacenamos el path completo del fichero temporal donde se almacenara la traduccion
				(setq *fichero-temporal* (concatenate 'string *ipss-temporal-dir* "domain.lisp"))

;;----------------------------------------fin parte david david montero------------------------------------------------------

				;; se carga el fichero domain.pddl para que se realice la conversion a ipss automaticamente
				(load (concatenate 'string *domain-directory-trans* "domain.pddl"))
				

;;----------------------------------------inicio parte david montero---------------------------------------------------------

        		;;	(if (eq 1 (choose-file-to-trans-ipss-domain *domain-directory* (format nil "~a/domain.lisp" *ipss-temporal-dir*))) (let ()
			;;	(put-file "textview-pddl" "domain.pddl" *domain-directory-trans*)
             		;;	(put-file "textview-prodigy" "domain.lisp" *ipss-temporal-dir*)
            		;;	(show-window "V6-SHOW2FILES")))

;;----------------------------------------fin parte david david montero------------------------------------------------------

		(put-file "textview-pddl" "domain.pddl" *domain-directory-trans*)
  	      (put-file "textview-prodigy" "domain.lisp" *ipss-temporal-dir*)
        	(show-window "V6-SHOW2FILES")


)
		;; si se produce algun error
	    	(error (signal)
			(error-dialog nil (delete_invalid_chars (format nil "~d~%" signal)))
		)
	    )
	    
	    
	    )))
  ;; se restaura el path actual
  (setf *world-path* real-world-path)
  ;; se elimina el archivo temporal
  )
  0)
 
;; --------------------------------------------------------------------------------------------
;; window vl-w-trans
;; Button translate problems (pddl-problem-to-prodigy)
;; signal clicked
;; --------------------------------------------------------------------------------------------

(defun translate-problems-pddl2prodigy ()
(declare (special widgets))
(let ((real-world-path *world-path*)
      (domain)
      (problems)
      (all-together (= 1 (gtk_toggle_button_get_active (gethash "TRANScheckbutton-all-together" widgets)))))
  ;; se actualiza el path actual y se pone como el temporal
  (setf *world-path* *ipss-temporal-dir*) 

;; se comprueba si se han indicado el fichero de dominio y de conjunto de pruebas
(if (setq domain (read-obligatory-field "TRANSpddldomain-entry2" nil "Domain name is necessary"))
  (if  (setq problems (read-obligatory-field "TRANSpddlproblem-entry" nil "Problems file prefix is necessary"))
     (progn
	    ;; se realiza la traduccion de los problemas

;;-----------------------------------------------inicio de david montero----------------------------------------------------

;;            (handler-case (translate-problems (concatenate 'string *ipss-temporal-dir*  domain "/probs/") 

;; en lugar de cargar el directorio temporal, se carga el directorio de pruebas elegido por el usuario.
	      (handler-case (translate-problems (concatenate 'string *probs-directory*) 
                          :domain (read-from-string domain)
                          :file-prefix problems
                          :output-file-name "from-pddl"
                          :all-together-p all-together)


;;--------------------------------------------------fin de david montero----------------------------------------------------
		;; se comprueba si ha habido errores
		(error (signal)
			(error-dialog nil (delete_invalid_chars (format nil "~d~%" signal)))	
		)
	    )
              ;; se muestra el conjunto de problemas transformado a prodigy
              (if all-together
               (progn
		(setq *fichero-temporal* (concatenate 'string *ipss-temporal-dir* "from-pddl.lisp"))
                (put-file "textview-file" (concatenate 'string "from-pddl.lisp") 
                                          (concatenate 'string *ipss-temporal-dir*))
                (show-window "V7-SHOW1FILE"))))))
 (setf *world-path* real-world-path))
0)




;; --------------------------------------------------------------------------------------------
;; window vl-w-trans
;; Button translate problems (prodigy-problems-to-pddl)
;; signal clicked
;; --------------------------------------------------------------------------------------------
(defun translate-problems-prodigy2pddl ()

(let ((real-world-path *world-path*)
      (domain)
      (problems))

  ;; modifica el path actual por el temporal
  (setf *world-path* *ipss-temporal-dir*) 

;; comprueba si se han introducido los ficheros de dominio y de problemas
(if (setq domain (read-obligatory-field "TRANSprodigydomain-entry" nil "Domain name is necessary"))
  (if  (setq problems (read-obligatory-field "TRANSprodigyproblem-entry" nil "Probset is necessary"))

     (progn
	    ;; se realiza la conversion del fichero de prodigy a pddl
            (handler-case (translate-problems-to-pddl (concatenate 'string *ipss-temporal-dir*  domain "/probsets/"
                                                             problems ".lisp")
                                        :domain (read-from-string domain)
                                        :output-directory (concatenate 'string *ipss-temporal-dir*  domain "/probsets/")) 

		;; se tratan los posibles errores
	    	(error (signal)
	    		(error-dialog nil (delete_invalid_chars (format nil "~d~%" signal)))
	    	)
	    )
	
	      ;; se muestran los problemas traducidos
         ;;   (put-file "textview-file" (concatenate 'string "temp-problems-0-1-" domain ".pddl") 
         ;;                             (concatenate 'string *ipss-temporal-dir* domain "/probsets/"))
         ;;   (show-window "V7-SHOW1FILE")
    )))

 ;; se restaura el path actual
 (setf *world-path* real-world-path))

0)


;; --------------------------------------------------------------------------------------------
;; Function that reads a file and return a string with the file content
;; --------------------------------------------------------------------------------------------
(defun file-to-string (filename &optional (directory *domain-directory*))

;; abre el flujo en el fichero indicado por parametro
(with-open-file (stream (concatenate 'string directory filename) :direction :input :if-does-not-exist nil)
(let ((result ""))

    ;; lee todas las lineas del fichero y las concatena una a una a la variable result
    (if stream
	(do ((leido (read-line stream nil 'eof)
		    (read-line stream nil 'eof)))
	    ((eq leido 'eof))
             (setq result (concatenate 'string result (format nil "~a~%" leido)))))
result)))


;; --------------------------------------------------------------------------------------------
;; Function that puts the text of a file in a textview with name text-view-name
;; --------------------------------------------------------------------------------------------
(defun put-file (text-view-name filename &optional (directory *domain-directory*))
(declare (special widgets))  

  ;; almacena en la variable file-string el contenido del fichero
  (let* ((file-string (file-to-string filename directory))

	;; obtiene el text-view y el text-buffer del mismo
        (text-view (gethash text-view-name widgets))
        (text-buffer (gtk_text_view_get_buffer text-view)))
   
   (gtk_text_view_set_wrap_mode text-view 2)
   ;; copia en el buffer del text view el contenido del fichero
   (gtk_text_buffer_set_text text-buffer file-string (length file-string))
   
   ;; copia en el text-view el contenido del buffer
   (gtk_text_view_set_buffer text-view text-buffer)))


;; --------------------------------------------------------------------------------------------
;; window vl-w-lm
;; Button LEARN
;; signal clicked
;; --------------------------------------------------------------------------------------------
(defun execute-learn-macros ()
  (declare (special widgets))  

  (gtk_button_released (gethash "button-lm-learn" widgets))

  ;; se obtiene el contenido del primer paso de planificacion, si esta vacio se pone un 1 por defecto 
  (let ((first-plan-step (if (equal "" (gtk_entry_get_text (gethash "LMfirstplanstep-entry" widgets)))
			     1
			     (read-from-string (gtk_entry_get_text (gethash "LMfirstplanstep-entry" widgets)))))

	;; se obtiene el contenido del ultimo paso de planificacion, si esta vacio se le asigna el ultimo paso
	(last-plan-step (if (equal "" (gtk_entry_get_text (gethash "LMlastplanstep-entry" widgets)))
			    (length (prodigy-result-solution *prodigy-result*))
			    (read-from-string (gtk_entry_get_text (gethash "LMlastplanstep-entry" widgets))))))

    (setf *macros-range* 0)
    (setf *valid-macros* 0)
    
    (handler-case
      ;; se ejecuta el aprendizaje del macro 
      (learn-macro *domain-name* 
		 (= 1 (gtk_toggle_button_get_active (gethash  "LMcheckbutton-install-macro-p" widgets)))
		 (= 1 (gtk_toggle_button_get_active (gethash  "LMcheckbutton-printp" widgets)))
		 first-plan-step
		 last-plan-step)

      ;; se controlan los errores
      (error (signal)  
	     (format t "~d~%" signal)
		(if (eq *macros-range* 0)
			(error-dialog nil (delete_invalid_chars (format nil "Los pasos del plan introducidos deben estar en el rango [1, ~d]." (length (prodigy-result-solution *prodigy-result*)))))
		(error-dialog nil (delete_invalid_chars (format nil "~d~%" signal))))
		)
      )

      ;; se indica al usuario que el aprendizaje de macros se ha completado correctamente
      (if (eq *valid-macros* 1)
	(information-dialog nil (format nil "El aprendizaje de macros ha finalizado satisfactoriamente"))
	)
    0))


;; --------------------------------------------------------------------------------------------
;; window vl-w-lm
;; Button SEE MACROS
;; signal clicked
;; --------------------------------------------------------------------------------------------
(defun see-macros ()          
  (put-file "textview-macros" "macros.lisp")
  (show-window "V5-MACROS")
0)

;;; Function that interfaces "easily" to Hamlet
(defun learn-cr (node)
  (declare (type p4::nexus node))
  (declare (special compute-state-p))
  (declare (special learn-goal-rule-p))  
  (let* ((learnp (and (not (eq node (find-node 2)))
		      (not (eq node (find-node 3)))))
	 (best-childs (car (getf (p4::nexus-plist node) 'best-childs-info)))
	 (state (or (get-state-for-node node)
		    (getf (p4::nexus-plist node) 'state)))
	 (applied-child (if (and learnp
				 (p4::a-or-b-node-p node))
			    (find-if #'p4::applied-op-node-p
				     (p4::nexus-children node)
				     :from-end t)))
	 (les-substitutions nil)
	 (goal-regressed-state-p nil)
	 (goals-alist nil)
	 (new-les nil)
	 (bad-children nil))
    (when learnp
      (dolist (child (p4::nexus-children node))
	(if (not (member child best-childs))
	    (push child bad-children)))
      (if (and bad-children *inductionp*)
	  (repair-overgeneralization node node
				     (list (cons nil bad-children)) 0)))
    (dolist (best-child best-childs)
      (let ((regressed-state (get-regressed-state node best-child learnp))
	    (les-goals (learn-only-optimal-path best-child compute-state-p)))
	;; if there are two or more ways to achieve a certain
	;; goal, we pick up the first one
	(dolist (goal-state (cdr les-goals))
	  (pushnew goal-state goals-alist :test #'equal :key #'car))
	(dolist (le (car les-goals))
	  (push le new-les))
	;; if there are two or more ways to achieve a certain
	;; goal, we pick up the first one
	(when (and (p4::goal-node-p node)
		   (not goal-regressed-state-p))
	  (setq goal-regressed-state-p t)
	  (pushnew (cons (get-target-goal node) regressed-state)
		   goals-alist :test #'equal :key #'car))
	(if (and learnp
		 (or (not (p4::goal-node-p best-child))
		     (not applied-child)
		     (clobbersp applied-child regressed-state)))
	    (dolist (le-goals (learn-decision node best-child state
					      regressed-state t
					      learn-goal-rule-p))
	      (push (cons *substitution* le-goals) les-substitutions)))))
    (dolist (le-substitution les-substitutions (cons new-les goals-alist))
      (let ((le (cadr le-substitution)))
	(push le new-les)
	(if (and *inductionp* *use-preferred-preconds-p*)
	    (set-preferred-preconds le (cddr le-substitution) goals-alist
				    (car le-substitution))))))
  (save-cr))

;;; Function that prints the interface trace if the *interface-printp* variable is true
(defun interface-trace (&rest args)
  (if *interface-printp* 
         (progn
              (format t "~%interface-trace: ")     
              (apply #'format args)
              (format t "~%"))))

(defun fill-meta-predicate-options nil
  (case (get-valor-combobox)
      ((current-goal candidate-goal target-goal true-in-state)
       (setf (get-valor-siguiente-combobox)
	     (get-all-goals)))
    ((current-operator candidate-operator applicable-op)
     (setf (get-valor-siguiente-combobox)
	     (get-all-operators)))
    (type-of-object
     (setf (get-valor-siguiente-combobox)
	   (get-all-types :top-type)))
    (lisp-function)))

;; some-candidate-goals
;; prior-goal

(defun get-all-types (type)
  (dolist (subtype (p4::all-subtypes type *current-problem-space*))
      ))


;;------------------------------------------------inicio de almacenamiento de planes-----------------------------------------

;;-----------------------------------------------------------------------
;; david montero
;; funcion que a la que se llama desde la interfaz para almacenar
;; los planes actuales en un fichero
;;-----------------------------------------------------------------------

(defun save_plans()
 (let ((dialog-widget (gtk_file_selection_new)))
   (gtk_file_selection_set_filename dialog-widget (concatenate 'string *domain-directory* "plan.xml"))
  (if (= (gtk_dialog_run dialog-widget) -5)
    (let ((file (gtk_file_selection_get_filename dialog-widget)))
      (save_plans_file file)))
   (gtk_widget_destroy dialog-widget))
0
)


;;-----------------------------------------------------------------------
;; david montero
;; funcion que recibe un fichero donde almacenar los planes y lo abre
;;-----------------------------------------------------------------------

(defun save_plans_file (filename)

  ;; se comprueba si el fichero existe y se le pregunta al usuario si lo quiere sobreescribir
  (let ((is-file 1) (continuar 1))
    (if (probe-file filename)
	(setf is-file 1)
	(setf is-file 0)
    )
  
    (if (= is-file 1)
	(let ((result 
	    (if (eq *language* 'es)
		(show-question nil (format nil "El fichero ~a ya existe y se va a sobreescribir ~&~a" filename "Desea continuar?"))
		(show-question nil (format nil "The file ~a exists and it going to be overwritten ~&~a" filename "Would you like to continue?"))
	    ))) 
    ;; yes
          (if (= result -8)	
		(setf continuar 1)
		(setf continuar 0)
          )
	)
    )
    (if (= continuar 1)
	    (with-open-file (stream filename :direction :output
				:if-exists :supersede
				:if-does-not-exist :create)
                         (format stream "<?xml version=\"1.0\" encoding=\"ISO-8859-1\"?>~%")
                         (save_plans_to_stream stream)
            )
    )
)
)


;;-----------------------------------------------------------------------
;; david montero
;; funcion que comienza a dividir los problemas uno a uno del plan
;;-----------------------------------------------------------------------

(defun save_plans_to_stream (stream &optional (ini *estructura-arbol*))
	(let ((contenido) (longitud 0) (contador 0))
		(case *planner*
		(ipss
			(print "MUESTRO LA ESTRUCTURA DEL ARBOL")
			(print ini)
			;; De momento no funciona con los probsets!!!!
		        ;;(if (eq (length ini) 1) (save-results-plan stream (car ini)) 
			(progn 
				(format stream "<allplans>~%")
				(format stream "<problem>~%") 
				(save-this-problem stream (car ini))
				(format stream "<result type=\"plancost\" value=\"~d\" />~%" (nth 1(nth 2 (car (cdr ini)))))
				(format stream "<result type=\"planlength\" value=\"~d\" />~%" (nth 1 (nth 2 (car (cdr ini)))))
				(format stream "<result type=\"totalTime\" value=\"~d\" />~%" (nth 1 (nth 0 (car (cdr ini)))))
				(format stream "</problem>~%")
				(format stream "</allplans>~%"))		

		)
		(sayphi
		        (if (solution-found *say-solution*)
				(progn
					(format stream "<allplans>~%")
					(format stream "<problem>~%") 
					(format stream "<plan>~%")


					(dolist (i-node (solution-path *say-solution*) *say-solution*)
						     (let ((plan-action (snode-applied-action i-node)))
						       	(format stream "<operator name=\"~a\" atributes=\"~a\" />~%" (car (gaction-planaction plan-action)) (cdr (gaction-planaction plan-action)))))			
			
					(format stream "<result type=\"plancost\" value=\"~d\" />~%" (solution-total-cost *say-solution*))
					(format stream "<result type=\"planlength\" value=\"~d\" />~%" (solution-depth *say-solution*))
					(format stream "<result type=\"totalTime\" value=\"~d\" />~%" (solution-total-time *say-solution*))
					(format stream "</plan>~%")
					(format stream "</problem>~%")		
					(format stream "</allplans>~%")			
				)
			)
		)
		(sgplan
			(setf contenido (file-to-string "sgplan-output2.lisp" *my-sgplan-pathname*))
			(setf contenido (read-from-string contenido))
			(format stream "<allplans>~%")
			(format stream "<problem>~%") 
			(format stream "<plan>~%")
			(dolist (p (car (last contenido)))

				(format stream "<operator name=\"~a\" atributes=\"~a\" />~%" (car p) (cdr p))
			)
			(format stream "<result type=\"plancost\" value=\"~d\" />~%" (nth 2 contenido))

			(format stream "<result type=\"planlength\" value=\"~d\" />~%" (nth 1 contenido))
			(format stream "<result type=\"totalTime\" value=\"~d\" />~%" (car contenido))
			(format stream "</plan>~%")
			(format stream "</problem>~%")		
			(format stream "</allplans>~%")
		)	
		(metric-ff
	              (setf contenido (file-to-string "ff-output2.lisp" *my-metricff-pathname*))
   		      (setf contenido (read-from-string contenido))
			(format stream "<allplans>~%")
			(format stream "<problem>~%") 
			(dolist (p contenido)
				(incf contador)
				(if (or (and (> (length contenido) 1)  (< contador (- (length contenido) 2))) (= 1 (length contenido)))
	
					(progn
						(format stream "<plan>~%")
						(setf longitud 0)
						(dolist (pl (car p))
							(format stream "<operator name=\"~a\" atributes=\"~a\" />~%" (car pl) (cdr pl))
							(incf longitud)
						)
						(format stream "<result type=\"plancost\" value=\"~d\" />~%" (nth 2 p))
						(format stream "<result type=\"planlength\" value=\"~d\" />~%" longitud)
						(format stream "<result type=\"totalTime\" value=\"~d\" />~%" (nth 1 p))
						(format stream "</plan>~%")				
					)
				)
			)
			(format stream "</problem>~%")		
			(format stream "</allplans>~%")

		)	
		(lpg


			(with-open-file (stream (concatenate 'string *my-lpg-pathname* "lpg-output.lisp") :direction :input :if-does-not-exist nil)

			    (if stream
				(do ((leido (read-line stream nil 'eof)
				    (read-line stream nil 'eof)))
				    ((eq leido 'eof))
				    ;;comprobaciones de la linea
				    (setf contenido (concatenate 'string contenido (format nil "~a " leido)))
			))
		
		      )
   		      (setf contenido (read-from-string contenido))
			(format stream "<allplans>~%")
			(format stream "<problem>~%") 
			(dolist (p contenido)
				(format stream "<plan>~%")
				(setf longitud 0)
				(dolist (pl (car p))
					(format stream "<operator name=\"~a\" atributes=\"~a\" />~%" (car pl) (cdr pl))
					(incf longitud)
				)
				(format stream "<result type=\"plancost\" value=\"~d\" />~%" (nth 3 p))
					(format stream "<result type=\"planlength\" value=\"~d\" />~%" (nth 2 p))
					(format stream "<result type=\"totalTime\" value=\"~d\" />~%" (nth 1 p))
					(format stream "</plan>~%")				

			)
			(format stream "</problem>~%")		
			(format stream "</allplans>~%")
		)
	)
	)
		
)


;;-----------------------------------------------------------------------
;; david montero
;; funcion que guarda en el fichero el numero de problema y realiza una 
;; llamada para almacenar los planes de dicho problema
;;-----------------------------------------------------------------------

(defun save-this-problem (stream problem)
	(progn 
	       (save-plans-stream stream (cdr problem))
	)
)


;;-----------------------------------------------------------------------
;; david montero
;; funcion que comienza a dividir los planes uno a uno
;;-----------------------------------------------------------------------

(defun save-plans-stream (stream plans)
	(if (eq (length plans) 1) (save-this-plan-stream stream (car plans))  (progn (save-this-plan-stream stream (car plans)) (save-plans-stream stream (cdr plans))))
)

;;-----------------------------------------------------------------------
;; david montero
;; funcion que almacena en el fichero stream el plan pasado por parametro
;;-----------------------------------------------------------------------

(defun save-this-plan-stream (stream plan)
	(progn 
		;;(format stream "~t -~a~%" (car plan))
		(format stream "<plan>~%")
		(dolist (p (cdr plan))
			(format stream "<operator name=\"~a\" atributes=\"~a\" />~%" (car p) (cdr p))
		)
		(format stream "</plan>~%")
	)
)

;;-----------------------------------------------------------------------
;; david montero
;; funcion que almacena en el fichero stream los resultados pasados por parametro
;;----------------------------------------------------------------------

(defun save-results-plan (stream resultados)
	(let ()
		(format stream "<results>~%") 
		(dolist (r resultados)
			(case (car r)
			(4
				(format stream "<result type=\"plancost\" value=\"~d\" />~%" (car (cdr r)))
			)
			(3  
				(format stream "<result type=\"planlength\" value=\"~d\" />~%" (car (cdr r)))
			)
			(2
				(format stream "<result type=\"bestcost\" value=\"~d\" />~%" (car (cdr r)))
			)
			(1
				(format stream "<result type=\"totalNodes\" value=\"~d\" />~%" (car (cdr r)))
			)
			(0
				(format stream "<result type=\"totalTime\" value=\"~d\" />~%" (car (cdr r)))
			)
			)
		)
		(format stream "</results>~%") 
		(format stream "</allplans>~%") 
	)
 )

;;----------------------------------------------fin de almacenamiento de planes---------------------------------------------


;;-----------------------------------------inicio de almacenamiento del arbol de busqueda------------------------------------


;;-----------------------------------------------------------------------
;; david montero
;; funcion que a la que se llama desde la interfaz para almacenar
;; el arbol de busqueda actual en un fichero
;;-----------------------------------------------------------------------

(defun save_tree_structure ()
 (let ((dialog-widget (gtk_file_selection_new)))
   (print *estructura-arbol2*)
   (print *estructura-arbol3*)
   (gtk_file_selection_set_filename dialog-widget (concatenate 'string *domain-directory* "tree.xml"))
  (if (= (gtk_dialog_run dialog-widget) -5)
    (let ((file (gtk_file_selection_get_filename dialog-widget)))
     (save_tree_file file)))
   (gtk_widget_destroy dialog-widget))
0
)


;;-----------------------------------------------------------------------
;; david montero
;; funcion que recibe un fichero donde almacenar el arbol y lo abre
;;----------------------------------------------------------------------

(defun save_tree_file (filename)

  ;; se comprueba si el fichero existe y se le pregunta al usuario si lo quiere sobreescribir
  (let ((is-file 1) (continuar 1))
    (if (probe-file filename)
	(setf is-file 1)
	(setf is-file 0)
    )
  
    (if (= is-file 1)
	(let ((result 
	    (if (eq *language* 'es)
		(show-question nil (format nil "El fichero ~a ya existe y se va a sobreescribir ~&~a" filename "Desea continuar?"))
		(show-question nil (format nil "The file ~a exists and it going to be overwritten ~&~a" filename "Would you like to continue?"))
	    ))) 
    ;; yes
          (if (= result -8)	
		(setf continuar 1)
		(setf continuar 0)
          )
	)
    )
    (if (= continuar 1)
	    (with-open-file (stream filename :direction :output
				:if-exists :supersede
				:if-does-not-exist :create)
		;; para guardarlo en html
	        ;;(save_tree_to_stream stream *estructura-arbol2*))
		;; para guardarlo en xml
		(save_tree_to_stream stream *estructura-arbol3*)
		(cerrar-etiquetas stream -2)
		(format stream "</arbol>"))
    )
)
)

;;-----------------------------------------------------------------------
;; david montero
;; funcion que inicia y finaliza el codigo html del fichero stream
;;-----------------------------------------------------------------------

(defun save_tree_to_stream (stream tree)
	(print tree)	

	;; para almacenar el arbol en html
	;;(progn
  	;; (format stream "<html>~%~t<head>Arbol para el dominio ~a~%~t</head>~%<body>~%" *domain-name*)
	;; (save_this_tree stream tree -1)
        ;; (format stream "</body>~%~t</html>")
	;;)

	;; para almacenar el arbol en xml
	(progn
  	 (format stream "<?xml version=\"1.0\" encoding=\"ISO-8859-1\"?>~%")
	 (format stream "<arbol>~%")
	 (save_this_tree_xml stream tree -1)
	)

)


;;-----------------------------------------------------------------------
;; david montero
;; funcion que recorre todos los nodos del arbol y realiza la llamada
;; para su correcto almacenamiento indicando la profundidad del nodo
;;-----------------------------------------------------------------------

(defun save_this_tree_xml (stream tree depth)
	(if (not (null tree))
	(if (= (length tree) 1) 
		(write_this_node_xml stream (car tree) depth)
		(progn
			(write_this_node_xml stream (car tree) depth)
			(save_this_tree_xml stream (cdr tree) 
				(if (= (length (cdr tree)) 1)
					(let ()
						(if (numberp (car (car (cdr tree))))
							(car (car (cdr tree)))
							(- depth 2)
                                        	) 
					)
                                        (let ()
					(if (numberp (car (car (cdr tree))))
						(car (car (cdr tree)))
						(- depth 2)
					))))))))


;;-----------------------------------------------------------------------
;; david montero
;; funcion que almacena el nodo actual en el flujo stream
;;-----------------------------------------------------------------------

(defun write_this_node_xml (stream node depth)
	(if (numberp (car node))
		(progn 
			(if (and *actual-lvl-arbol* (<= depth *actual-lvl-arbol*))
				(cerrar-etiquetas stream depth);
			)
			(setf *tipo-lvl* (append *tipo-lvl* (list (list (car (cdr node)) depth))))
			(setf *actual-lvl-arbol* depth)
			;;(print *tipo-lvl*)
			(print (format nil "~%~%"))
			;;(if (eq (car (cdr node)) 'ex) (format stream "<nodo tipo="exito">") (format stream "<nodo tipo="error">"))
			;;(dolist (nodo (cdr (cdr node)))
			;;	(format stream "~a~%" nodo)
			;;)


			;;(if (eq (car (cdr node)) 'ex) (format stream "<nodo tipo=\"exito\" value =\"~a\">~%" (cdr (cdr node))) (format stream "<nodo tipo=\"fallo\" value =\"~a\">~%" (cdr (cdr node))))

			(if (eq (car (cdr node)) 'ex) (format stream "<nodo tipo=\"~a\" exito=\"SI\" " (car (cdr (cdr node)))) (format stream "<nodo tipo=\"~a\" exito=\"NO\" " (car (cdr (cdr node)))))

			(let ((longitud) 
			     (posicion 1)
			     (atributo 1)
			     (h))
			     (setf h (concatenate 'string "( " (car (cdr (cdr (cdr node)))) " )" ) )
			     (setf longitud (length (read-from-string h)))
			     (dolist (nodo (read-from-string h))
				     (print (format nil "entro ~a ~d ~d~%" nodo posicion longitud))
				     (cond	
					((and (listp nodo) (not (null nodo)))
						(dolist (nodohijo nodo)
							(if (= atributo 1)
								(format stream "operator=\"~a\" " nodohijo)
								(format stream "atributo~d=\"~a\" " (- atributo 1) nodohijo)
							)
						   	(setf atributo (+ 1 atributo))				     
						)
					)
					(t
						(if (= posicion (- longitud 1))
							(format stream "num-nodo=\"~d\" " nodo)
							(if (= posicion longitud)
								;;preguntar que es esto!!!!!!!
								(format stream "num=\"~a\" " nodo)
								(if (= atributo 1)
									(format stream "operator=\"~a\" " nodo)
									(format stream "atributo~d=\"~a\" " (- atributo 1) nodo)
								)
							)
						)
						(setf atributo (+ 1 atributo))
					)
				     )
				   (setf posicion (+ 1 posicion))				     
			     )
			)
			(format stream ">~%")
		)
		(progn
			(dolist (nodo node)
				;;(print nodo)
			        (let ((longitud) 
				     (posicion 1)
				     (atributo 1)
				     (h))
				     (setf h (concatenate 'string "( " nodo " )" ) )
				     (setf longitud (length (read-from-string h)))

				     (format stream "<nodo tipo=\"~a\" exito=\"untried\" " (car (read-from-string h)))
				     (dolist (nodo2 (cdr (read-from-string h)))
					     (print (format nil "entro ~a ~d ~d~%" nodo2 posicion longitud))
					     (cond
						((and (listp nodo2) (not (null nodo2)))
							(dolist (nodohijo nodo2)
								(if (= atributo 1)
									(format stream "operator=\"~a\" " nodohijo)
									(format stream "atributo~d=\"~a\" " (- atributo 1) nodohijo)
								)
							   	(setf atributo (+ 1 atributo))				     
							)
						)
						(t
							(if (< posicion longitud)
								(if (= atributo 1)
									(format stream "operator=\"~a\" " nodo2)
									(format stream "atributo~d=\"~a\" " (- atributo 1) nodo2)
								)
							)
							(setf atributo (+ 1 atributo))
						)
					     )
					   (setf posicion (+ 1 posicion))				     
				     )
				)
			(format stream "> </nodo>~%")
		))))


;;-----------------------------------------------------------------------
;; david montero
;; funcion que cierra todas las etiquetas abiertas en el arbol de busqueda
;;-----------------------------------------------------------------------
(defun cerrar-etiquetas (stream depth)
	(if (and *tipo-lvl* (<= depth *actual-lvl-arbol*)) 
		 (let ((ultimo))
			(setf ultimo (car (last *tipo-lvl*)))
			(if (and (<= depth *actual-lvl-arbol*) (eq (car ultimo) 'er)) 
				(progn
					;;(format stream "</error~d>~%" (car (last ultimo)))
					;;(format stream "</error>~%")
					(format stream "</nodo>~%")
					(setf *actual-lvl-arbol* (car (last ultimo)))
					(setf *tipo-lvl* (butlast *tipo-lvl*))
					;;(print (format nil "~d --> ~d ~%" *actual-lvl-arbol* depth))
					(cerrar-etiquetas stream depth)))
			(if (and (< depth *actual-lvl-arbol*) (eq (car ultimo) 'ex)) 
				(progn
					;;(format stream "</exito~d>~%" (car (last ultimo)))
					;;(format stream "</exito>~%")
					(format stream "</nodo>~%")
					(setf *actual-lvl-arbol* (car (last ultimo)))
					(setf *tipo-lvl* (butlast *tipo-lvl*))
					;;(print (format nil "~d --> ~d ~%" *actual-lvl-arbol* depth))
					(cerrar-etiquetas stream depth)))
		)
	)
)

;;-----------------------------------------------------------------------
;; david montero
;; funcion que recorre todos los nodos del arbol y realiza la llamada
;; para su correcto almacenamiento indicando la profundidad del nodo
;;-----------------------------------------------------------------------

(defun save_this_tree (stream tree depth)
	(if (not (null tree))
	(if (= (length tree) 1) 
		(write_this_node stream (car tree) depth)
		(progn
			(write_this_node stream (car tree) depth)
			(save_this_tree stream (cdr tree) 
				(if (= (length (cdr tree)) 1) 
					(let ()
						(if (numberp (car (car (cdr tree))))
							(car (car (cdr tree)))
							(- depth 2)
                                        	) 
					)
                                        (let ()
					(if (numberp (car (car (cdr tree))))
						(car (car (cdr tree)))
						(- depth 2)
					))))))))


;;-----------------------------------------------------------------------
;; david montero
;; funcion que almacena el nodo actual en el flujo stream con la 
;; tabulacin correspondiente a su profundidad depth
;;-----------------------------------------------------------------------

(defun write_this_node (stream node depth)
	(if (numberp (car node))
		(dolist (nodo (cdr node))
			(format stream "<pre>")
			(do ((x -1 (+ 1 x))) ((> x depth)) 
				(format stream "  "))
			(format stream "~a</pre>" nodo)
		)
		(dolist (nodo node)
			(format stream "<pre>")
			(do ((x -1 (+ 1 x))) ((> x depth)) 
				(format stream "  "))
			(format stream "~a</pre>" nodo))))

;;-----------------------------------------fin de almacenamiento del arbol de busqueda---------------------------------------








;;-------------------------------------inicio de almacenamiento de dominios traducidos--------------------------------------

;;-----------------------------------------------------------------------------------------------------
;; david montero
;; funcion que le da al usuario la opcion de elegir un fichero de salida para la pantalla de traduccion
;; recibe el dominio donde se encuentra el fichero a traducir y el temporal donde esta almacenada la traduccion
;;-----------------------------------------------------------------------------------------------------

(defun choose-file-to-trans-ipss-domain (dominio-fichero-a-convertir temporal-usado)

 (let ((valor 0)(dialog-widget (gtk_file_selection_new)))
   (gtk_file_selection_set_filename dialog-widget (concatenate 'string dominio-fichero-a-convertir "salida.lisp"))
  (if (= (gtk_dialog_run dialog-widget) -5)
    (let ((file (gtk_file_selection_get_filename dialog-widget)))
      (save_domain_file file temporal-usado)
      (setq valor 1)))
   (gtk_widget_destroy dialog-widget)
valor))


;;-----------------------------------------------------------------------------------------------------
;; david montero
;; funcion que le da al usuario la opcion de elegir un fichero de salida para la pantalla de traduccion
;; en este caso todas las variables necesarias esan almacenadas anteriormente
;;-----------------------------------------------------------------------------------------------------

(defun save_trans_domain ()
 (let ((valor 0)(dialog-widget (gtk_file_selection_new)))
   (gtk_file_selection_set_filename dialog-widget (concatenate 'string *domain-directory* "salida.lisp"))
  (if (= (gtk_dialog_run dialog-widget) -5)
    (let ((file (gtk_file_selection_get_filename dialog-widget)))
      (save_domain_file file *fichero-temporal*)
      (setq valor 1)))
   (gtk_widget_destroy dialog-widget)
valor))


;;--------------------------------------------------------------------------
;; david montero
;; funcion que abre los stream de los dos ficheros recibidos por parametro
;;--------------------------------------------------------------------------

(defun save_domain_file (filename temporal)
  (with-open-file (stream filename :direction :output
				:if-exists :supersede
				:if-does-not-exist :create)
  		(with-open-file (stream2 temporal :direction :input)
			(write_stream1_to_stream2 stream2 stream))))


;;--------------------------------------------------------------------------
;; david montero
;; funcion que escribe el contenido del stream1 en stream2
;;--------------------------------------------------------------------------

(defun write_stream1_to_stream2 (stream1 stream2)
  (let ((regla))
	(loop while (setf regla (read-line stream1 nil)) do
		(format stream2 "~a~%" regla))))

;;-------------------------------------fin de almacenamiento de dominios traducidos------------------------------------------


;;------------------------------------------inicio de ayuda online-----------------------------------------------------------

;;------------------------------------------------------------------------------------------------------------
;; david montero
;; funcion que abre un navegador firefox y lo redirige al manual online de pltool en la web de la asignatura
;;------------------------------------------------------------------------------------------------------------

(defun show-help-online ()
(SHELL "firefox http://www.plg.inf.uc3m.es/pltool-es.php &")
)

;;------------------------------------------fin de ayuda online--------------------------------------------------------------


;;------------------------------------------inicio de experimenter-----------------------------------------------------------

;;------------------------------------------------------------------------------------------------------------
;; david montero
;; funcion que crea un nuevo archivo de experimentacion
;;------------------------------------------------------------------------------------------------------------

(defun new-file-experimenter()
	(declare (special widgets))
	(let ((dialog-widg (gtk_file_selection_new)))
        (gtk_file_selection_set_filename dialog-widg (concatenate 'string *my-experimenter-pathname* "nuevo.exp"))
	  (if (= (gtk_dialog_run dialog-widg) -5)
		(progn 
		(start-experimenter)
		(setf *fichero-experimenter* (gtk_file_selection_get_filename dialog-widg))
		)
		(show-window "EXPERIMENTER")
		)
	   (gtk_widget_destroy dialog-widg)		
1))


;;------------------------------------------------------------------------------------------------------------
;; david montero
;; funcion que muestra la pantalla de eleccion de dominios en el experimenter, solo si antes se ha elegido
;; un nuevo fichero para el experimenter
;;------------------------------------------------------------------------------------------------------------

(defun show-domain-experimenter()
	(if *fichero-experimenter*
		(progn
		    (setf *dominio-experimenter* nil)
		    (setf *problema-experimenter* nil)
		    (setf *reglas-experimenter* nil)
		    (show-window "domain-experimenter")
		1)
		(progn 
			(if (eq *language* 'es)
	      			(error-dialog nil (format nil "Debe elegir un fichero de experimento antes"))
				(error-dialog nil (format nil "Must select a file experimenter first"))
			)
		1)))


;;------------------------------------------------------------------------------------------------------------
;; david montero
;; funcion que muestra la pantalla de eleccion de planificadores en el experimenter, solo si antes se ha elegido
;; un nuevo fichero para el experimenter
;;------------------------------------------------------------------------------------------------------------

(defun show-algoritm-experimenter()
	(if *fichero-experimenter* 
		(progn
			(start-window-algoritm-experimenter)
		1)
		(progn
			(if (eq *language* 'es)
	      			(error-dialog nil (format nil "Debe elegir un fichero de experimento antes"))
				(error-dialog nil (format nil "Must select a file experimenter first"))
			)
		1)))


;;------------------------------------------------------------------------------------------------------------
;; david montero
;; funcion que inicializa las variables del experimenter
;;------------------------------------------------------------------------------------------------------------

(defun start-experimenter()
	(declare (special widgets))
	(let ((window1 (gethash "scrolledwindow1Experimenter" widgets))
	      (window2 (gethash "scrolledwindow2Experimenter" widgets)))

		;; si existe algun arbol mostrado o bien en los dominios o bien en los planificadores, se elimina
		(if *arbol-dominios*
		      (progn (gtk_container_remove window1 *arbol-dominios*)
	     	      (gtk_widget_destroy *arbol-dominios*)))
		(if *arbol-algoritmos*
		      (progn (gtk_container_remove window2 *arbol-algoritmos*)
		     (gtk_widget_destroy *arbol-algoritmos*)))

	      (setf *arbol-dominios* (start_domains_experimenter))
 	      (gtk_container_add window1 *arbol-dominios*)
	      (setf *arbol-algoritmos* (start_algortims_experimenter))
 	      (gtk_container_add window2 *arbol-algoritmos*)
	      (setf *fichero-experimenter* nil)
	      (setf *tipo-dominio* nil)
	      (setf *dominio-experimenter* nil)
	      (setf *problema-experimenter* nil)
	      (setf *lista-dominios* nil)
	      (setf *lista-algoritmos* nil)
	      (setf *veces-experimenter* 1)
	      (setf *modo-ejecucion* 1)
	      (show-domains-experimenter)
	      1))


;;------------------------------------------------------------------------------------------------------------
;; david montero
;; funcion que inicializa el arbol gtk_tree_view que se utilizara en la parte de los dominios del experimenter
;;------------------------------------------------------------------------------------------------------------

(defun start_domains_experimenter ()
(let ((view (gtk_tree_view_new))
      (col (gtk_tree_view_column_new))
      (renderer (gtk_cell_renderer_text_new)))
  (gtk_tree_view_column_set_title col "Domains and problems")

  (gtk_tree_view_column_pack_start col renderer 1)
  
  (gtk_tree_view_column_add_attribute col renderer "markup" 0)

;;  /* pack tree view column into tree view */
  (gtk_tree_view_append_column view col)

  (gtk_tree_view_set_model view nil)

  view
))


;;------------------------------------------------------------------------------------------------------------
;; david montero
;; funcion que inicializa el arbol gtk_tree_view que se utilizara en la parte de los planificadores del experimenter
;;------------------------------------------------------------------------------------------------------------

(defun start_algortims_experimenter ()
(let ((view (gtk_tree_view_new))
      (col (gtk_tree_view_column_new))
      (renderer (gtk_cell_renderer_text_new)))

  (gtk_tree_view_column_set_title col "Algoritms")

  (gtk_tree_view_column_pack_start col renderer 1)
  
  (gtk_tree_view_column_add_attribute col renderer "markup" 0)

;;  /* pack tree view column into tree view */
  (gtk_tree_view_append_column view col)

  (gtk_tree_view_set_model view nil)

  view
))


;;------------------------------------------------------------------------------------------------------------
;; david montero
;; funcion que realiza el tratamiento de busqueda al aadir un nuevo dominio y problema
;;------------------------------------------------------------------------------------------------------------

(defun select-experimenter (entry-label selection-type)
  (declare (special widgets))  
  ;; se crea una nueva pantalla de dialogo para la seleccion del archivo si devuelve -2 es aceptar, -1 cancelacion
  (let ((dialog-widget (gtk_file_chooser_dialog_new "Select File" (gethash "domain-experimenter" widgets) 
                                                    0 "gtk-ok" -2 "gtk-cancel" -1  nil)))


 ;; directory      
  ;; se asigna el path actual a mostrar en la ventana de seleccion de archivo
  (case selection-type
    (domain  (gtk_file_chooser_set_current_folder dialog-widget *truename*))

    (prob    
		(let ((aux "/")) 
				(dolist (parte (pathname-directory *dominio-experimenter*))
					(if (not (equal parte ':ABSOLUTE))  
						(setf aux (concatenate 'string aux parte "/")))
				)
		(gtk_file_chooser_set_current_folder dialog-widget aux)))

    (rules
		(let ((aux "/")) 
				(dolist (parte (pathname-directory *dominio-experimenter*))
					(if (not (equal parte ':ABSOLUTE))  
						(setf aux (concatenate 'string aux parte "/")))
				)
		(gtk_file_chooser_set_current_folder dialog-widget aux))
    )
  )
          
  
  ;; si el usuario ha pulsado aceptar y si se va a seleccionar un problema ya se ha especificado un dominio 
  (if (and (or (eq selection-type 'domain)  (not (null *tipo-dominio*))) (= (gtk_dialog_run dialog-widget) -2))

    ;; se guarda en la variable file el archivo que haya elegido el usuario
    (let ((file (gtk_file_chooser_get_filename dialog-widget)) (is-file) (file-name))

	 ;; se hace un switch del tipo de seleccion
         (case selection-type 

	  ;; si es de tipo file se almacena en la tabla hash la direccion del fichero correspondiente
          (file (gtk_entry_set_text (gethash entry-label widgets)
                           (format nil "~a.~a" (pathname-name file) (pathname-type file))))
         

	  ;;si es de tipo reglas
	  (rules
	  	(handler-case
			
			;; se asigna 1 a la variable is-file si el fichero existe o 0 en caso contrario
			(if (probe-file file)
			     (setf is-file 1)
			     (setf is-file 0)	
			)
                ;; si se produce un error de seal se asigna valor 0 a la variable is-file
		(error (signal)
			(setf is-file 0)))

		(if (eq is-file 1)	
			(progn
				(setf file-name (concatenate 'string (pathname-name file) "." (pathname-type file)))

				;; se asigna a la variable *reglas-experimenter* el dominio elegido
				(setf *reglas-experimenter* file)

		            	;; se asigna al campo de texto del dominio el nombre del dominio elegido
				(gtk_entry_set_text (gethash entry-label widgets) (format nil "~a" (pathname-name file)))
			)

			;; se muestra un mensaje de error para que se elija un fichero correcto
			(if (eq *language* 'es)
	      			(error-dialog nil (format nil "El fichero ~a no es un fichero de dominio IPSS o PDDL" (pathname-name file)))
				(error-dialog nil (format nil "File ~a isn't a IPSS or PDDL domain file" (pathname-name file)))
			))
                        
                        ;; se resetea el dominio, asignando el nombre del directorio del dominio
			(reset-domain-variables nil (format nil "~a" (pathname-name file)))
	  )

          ;; si es de tipo dominio 
          (domain    
	    
                ;;se ejecutan varias comprobaciones
	  	(handler-case
			
			;; se asigna 1 a la variable is-file si el fichero existe o 0 en caso contrario
			(if (probe-file file)
			     (setf is-file 1)
			     (setf is-file 0)	
			)
                ;; si se produce un error de seal se asigna valor 0 a la variable is-file
		(error (signal)
			(setf is-file 0)))
		
		;; si el fichero existe
		(if (eq is-file 1)	
			(progn
				(setf file-name (concatenate 'string (pathname-name file) "." (pathname-type file)))
				;; si el archivo seleccionado es domain.lisp o domain.pddl 
				(if (or (equal (pathname-type file) "lisp")
					(equal (pathname-type file) "pddl"))


					(progn
						;; se le asigna a la variable *tipo-dominio* el tipo de dominio elegido
						(if (equal "domain.lisp" file-name) (setf *tipo-dominio* 'lisp) (setf *tipo-dominio* 'pddl))
						;; se asigna a la variable *dominio-experimenter* el dominio elegido
						(setf *dominio-experimenter* file)
	                                        ;; se le asigna a la variable file el path donde se encuentra el fichero
				    		(setq file (subseq file 0 (- (length file) (+ (length file-name) 1))))

            			            	;; se asigna al campo de texto del dominio el nombre del dominio elegido
						(gtk_entry_set_text (gethash entry-label widgets) (format nil "~a" (pathname-name file)))
					)
				
                                ;; si no es ninguno de los anteriores se muestra un mensaje de error
			(if (eq *language* 'es)
	      			(error-dialog nil (format nil "El fichero ~a no es un fichero de dominio IPSS o PDDL" (pathname-name file)))
				(error-dialog nil (format nil "File ~a isn't a IPSS or PDDL domain file" (pathname-name file)))
			)))
			;; se muestra un mensaje de error para que se elija un fichero correcto

			(if (eq *language* 'es)
	      			(error-dialog nil (format nil "El fichero ~a no existe" (pathname-name file)))
				(error-dialog nil (format nil "File ~a not exists" (pathname-name file)))
			))
                        
                        ;; se resetea el dominio, asignando el nombre del directorio del dominio
			(reset-domain-variables nil (format nil "~a" (pathname-name file))))

          ;; si es de tipo problema
          (prob

                ;;se ejecutan varias comprobaciones
	  	(handler-case
			;; se asigna 1 a la variable is-file si el fichero existe o 0 en caso contrario
			(if (probe-file file)
			     (setf is-file 1)
			     (setf is-file 0)	
			)
                ;; si se produce un error de seal se asigna valor 0 a la variable is-file
		(error (signal)
			(setf is-file 0)))
		(if (eq is-file 1)
		      (progn 
 	       	      ;; mirar tambien el probset para SAYPHI!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
		      ;; se comprueba si el ultimo directorio del path se corresponde con "probs"
                      (if (equal (car (last (pathname-directory file))) "probs")
                         (setf *interface-probset-p* nil)
                         (setf *interface-probset-p* t))
			 (progn
		         (print (format nil "entro~%"))
			 (print *tipo-dominio*)
			 (case *tipo-dominio*
				(lisp 
					(if (equal (pathname-type file) "lisp") 
						(progn 
							 (setf *problema-experimenter* file)
       					                  (gtk_entry_set_text (gethash entry-label widgets) (format nil "~a" (pathname-name file)))							 
						)
						(if (eq *language* 'es)
				      			(error-dialog nil (format nil "Debe elegir un problema tipo lisp"))
							(error-dialog nil (format nil "You must choose a problem type lisp"))
						)

					)
				)
				(pddl
					(if (equal (pathname-type file) "pddl") 
						(progn 
							  (setf *problema-experimenter* file)
       					                  (gtk_entry_set_text (gethash entry-label widgets) (format nil "~a" (pathname-name file)))							 
						)
						(if (eq *language* 'es)
				      			(error-dialog nil (format nil "Debe elegir un problema tipo pddl"))
							(error-dialog nil (format nil "You must choose a problem type ppdl"))
						)
					)

				)
				(otherwise 

					(if (eq *language* 'es)
			      			(error-dialog nil (format nil "Primero elije un dominio"))
						(error-dialog nil (format nil "Choose a domain first"))
					)
				) 

			 ))
		      )
		      (if (eq *language* 'es)
	      			(error-dialog nil (format nil "El fichero ~a no existe" (pathname-name file)))
				(error-dialog nil (format nil "File ~a not exists" (pathname-name file)))
		      )
		)

		      )))
		(if (and (or (eq selection-type 'prob) (eq selection-type 'rules)) (null *tipo-dominio*)) (error-dialog nil (format nil "Choose a domain first"))))

   (gtk_widget_destroy dialog-widget))
  
1)


;;------------------------------------------------------------------------------------------------------------
;; david montero
;; funcion que al pulsar el boton ok en la pantalla de aadir un nuevo dominio al experimenter
;; lo aade al arbol mostrado, actualiza la variable *lista-dominios* y resetea las variables necesarias
;;------------------------------------------------------------------------------------------------------------

(defun ok-domain-experimenter ()
	(declare (special widgets))
	(if (or (null *dominio-experimenter*) (null *problema-experimenter*))
	(progn
		      (if (eq *language* 'es)
	      			(error-dialog nil (format nil "Debe elegir un dominio y un problema antes"))
				(error-dialog nil (format nil "Must choose a domain and one problem first"))
		      )
	1)
	(progn
		(print *reglas-experimenter*)
		(setf *lista-dominios* (append *lista-dominios* (list (list *tipo-dominio* *dominio-experimenter* *problema-experimenter* *reglas-experimenter*))))
		(print *lista-dominios*)
		(setf *tipo-dominio* nil)
		(setf *dominio-experimenter* nil)
		(setf *problema-experimenter* nil)
		(setf *reglas-experimenter* nil)
		(gtk_entry_set_text (gethash "EXPERIMENTERdomain-entry" widgets) (format nil ""))
		(gtk_entry_set_text (gethash "EXPERIMENTERproblem-entry" widgets) (format nil ""))
		(gtk_entry_set_text (gethash "EXPERIMENTERrules-entry" widgets) (format nil ""))
		(hide-window "domain-experimenter")
		(show-domains-experimenter)
	1))
)


;;------------------------------------------------------------------------------------------------------------
;; david montero
;; funcion que al pulsar el boton de ok en la pantalla de eleccion de modo lo asigna a la variable *modo-ejecucion*
;;------------------------------------------------------------------------------------------------------------

(defun ok-mode-experimenter ()
	(declare (special widgets))
	(let ((modo (read-from-string (gtk_combo_box_get_active_text (gethash "EXPERIMENTERtipe-mode-combobox" widgets)))) 
	      (guardar-planes (read-from-string (gtk_combo_box_get_active_text (gethash "EXPERIMENTERsave-plans-combobox" widgets)))))
		(print modo)
		(print guardar-planes)
		(case modo
			(Domains-first (setf *modo-ejecucion* 2))
			(Planners-first (setf *modo-ejecucion* 1))
		)
		(case guardar-planes
			(Save-plans (setf *almacenar-planes* t))
			(Not-save-plans (setf *almacenar-planes* nil))
		
		)
		(cancel-mode-experimenter)
	1))


;;------------------------------------------------------------------------------------------------------------
;; david montero
;; funcion que al pulsar el boton ok en la pantalla de aadir un nuevo planificador al experimenter
;; lo aade al arbol mostrado, actualiza la variable *lista-algoritmos* y resetea las variables necesarias
;;------------------------------------------------------------------------------------------------------------

(defun ok-algoritm-experimenter ()
	(declare (special widgets))
	(let ((limite-tiempo)
	      (verbosity)
	      (max-tam)
	      (wg)
              (wh)
	      (heuristica-metric)
	      (algoritmo-metric)
	      (anytime-metric)
              (metric-opt)
	      (tipo-coste)
	      (limite-coste)
              (multiples-sol)
              (sin-reglas)
              (backtrack)
              (parar-cada-problema)
              (almacenar-reglas)
              (almacenar-planes)
              (heuristica-sayphi)
              (algoritmo-sayphi)
	      (modo-ejecucion)
	      (limite-tiempo-local)
	      (search-steps)
	      (number-solutions)
	      (restarts)
	      (repeats)
	      (seed)
	      (noise)
	      (static-noise)
	      (lowmemory)
	      (nobestfirst)
	      (onlybestfirst)
	      (timesteps)
	      (name (gtk_entry_get_text (gethash "EXPERIMENTERplanner-name-entry" widgets)))
	      (planner-name (read-from-string (gtk_combo_box_get_active_text (gethash "EXPERIMENTERcombobox1" widgets)))))

	(progn
		(case planner-name
		    (ipss
			(progn
			(setf limite-tiempo (read-from-string (gtk_entry_get_text (gethash "EXPERIMENTERIPSStime-bound-entry" widgets))))
			(setf verbosity (read-from-string (gtk_entry_get_text (gethash "EXPERIMENTERIPSSspin-verbosity-entry" widgets))))
		        (setf limite-coste (if (= 1 (gtk_toggle_button_get_active (gethash "EXPERIMENTERIPSScheckbutton-cost-bound" widgets)))
                                 (read-from-string (gtk_entry_get_text (gethash "EXPERIMENTERIPSScost-bound-entry" widgets)))
                                 nil))
		        (setf tipo-coste   (read-from-string (gtk_entry_get_text (gethash "EXPERIMENTERIPSScost-type-entry" widgets))))
			(setf multiples-sol (= 1 (gtk_toggle_button_get_active (gethash "EXPERIMENTERIPSScheckbutton-multiple-sols-p" widgets))))
			(setf sin-reglas (= 1 (gtk_toggle_button_get_active (gethash "EXPERIMENTERIPSScheckbutton-run-without-p" widgets))))
			(setf backtrack (= 1 (gtk_toggle_button_get_active (gethash "EXPERIMENTERIPSScheckbutton-backtrack-with-costs-p" widgets))))
			(setf parar-cada-problema (= 1 (gtk_toggle_button_get_active (gethash "EXPERIMENTERIPSScheckbutton-stop-after-problem-p" widgets))))
			(setf almacenar-reglas (= 1 (gtk_toggle_button_get_active (gethash "EXPERIMENTERIPSScheckbutton-store-fired-rules" widgets))))
			(setf almacenar-planes (= 1 (gtk_toggle_button_get_active (gethash "EXPERIMENTERIPSScheckbutton-store-plans" widgets))))
			(setf *lista-algoritmos* (append *lista-algoritmos* (list (list planner-name limite-tiempo verbosity limite-coste tipo-coste multiples-sol sin-reglas backtrack parar-cada-problema almacenar-reglas almacenar-planes name))))))

		   (sayphi
                     (progn
			(setf limite-tiempo (read-from-string (gtk_entry_get_text (gethash "EXPERIMENTERIPSStime-bound-entry" widgets))))
			(setf verbosity (read-from-string (gtk_entry_get_text (gethash "EXPERIMENTERIPSSspin-verbosity-entry" widgets))))
			(setf heuristica-sayphi (read-from-string (gtk_combo_box_get_active_text (gethash "EXPERIMENTERSAYPHIheuristic-entry" widgets))))
			(setf algoritmo-sayphi (read-from-string (gtk_combo_box_get_active_text (gethash "EXPERIMENTERSAYPHIalgorithm-entry" widgets))))
			(setf *lista-algoritmos* (append *lista-algoritmos* (list (list planner-name limite-tiempo verbosity heuristica-sayphi algoritmo-sayphi name))))))

		   (metric-ff
                     (progn
			(setf limite-tiempo (read-from-string (gtk_entry_get_text (gethash "EXPERIMENTERIPSStime-bound-entry" widgets))))
			(setf max-tam (read-from-string (gtk_entry_get_text (gethash "EXPERIMENTERMETRICmax-tam-entry" widgets))))
			(setf wg (read-from-string (gtk_entry_get_text (gethash "EXPERIMENTERMETRICw-g-entry" widgets))))
			(setf wh (read-from-string (gtk_entry_get_text (gethash "EXPERIMENTERMETRICw-h-entry" widgets))))
		        (setf metric-opt (= 1 (gtk_toggle_button_get_active (gethash "EXPERIMENTERMETRICcheckbutton-metric-optimice" widgets))))
			(setf heuristica-metric (read-from-string (gtk_combo_box_get_active_text (gethash "EXPERIMENTERMETRICheuristic-entry" widgets))))
			(setf algoritmo-metric (read-from-string (gtk_combo_box_get_active_text (gethash "EXPERIMENTERMETRICalgorithm-entry" widgets))))
			(setf anytime-metric (read-from-string (gtk_combo_box_get_active_text (gethash "EXPERIMENTERMETRICpoda-entry" widgets))))
			(setf *lista-algoritmos* (append *lista-algoritmos* (list (list planner-name limite-tiempo max-tam wg wh metric-opt heuristica-metric algoritmo-metric anytime-metric name))))))

		   (lpg
                     (progn
			(setf modo-ejecucion (read-from-string (gtk_combo_box_get_active_text (gethash "EXPERIMENTERLPGmodoejecucion-entry" widgets))))
			(setf limite-tiempo (read-from-string (gtk_entry_get_text (gethash "EXPERIMENTERIPSStime-bound-entry" widgets))))
			(setf limite-tiempo-local (read-from-string (gtk_entry_get_text (gethash "EXPERIMENTERLPGtime-bound-local-entry" widgets))))
			(setf search-steps (read-from-string (gtk_entry_get_text (gethash "EXPERIMENTERLPGsearchsteps-entry" widgets))))
			(setf number-solutions (read-from-string (gtk_entry_get_text (gethash "EXPERIMENTERLPGnumber-solutions-entry" widgets))))
			(setf restarts (read-from-string (gtk_entry_get_text (gethash "EXPERIMENTERLPGrestarts-entry" widgets))))
			(setf repeats (read-from-string (gtk_entry_get_text (gethash "EXPERIMENTERLPGrepeats-entry" widgets))))
			(setf seed (read-from-string (gtk_entry_get_text (gethash "EXPERIMENTERLPGseed-entry" widgets))))
			(setf noise (read-from-string (gtk_entry_get_text (gethash "EXPERIMENTERLPGnoise-entry" widgets))))
		        ;;(setf save-plans (= 1 (gtk_toggle_button_get_active (gethash "EXPERIMENTERLPGcheckbutton-save-plans" widgets))))
		        ;;(setf verbose (= 1 (gtk_toggle_button_get_active (gethash "EXPERIMENTERLPGcheckbutton-verbose" widgets))))
		        (setf static-noise (= 1 (gtk_toggle_button_get_active (gethash "EXPERIMENTERLPGcheckbutton-static-noise" widgets))))
		        (setf lowmemory (= 1 (gtk_toggle_button_get_active (gethash "EXPERIMENTERLPGcheckbutton-lowmemory" widgets))))
		        (setf nobestfirst (= 1 (gtk_toggle_button_get_active (gethash "EXPERIMENTERLPGcheckbutton-nobestfirst" widgets))))
		        (setf onlybestfirst (= 1 (gtk_toggle_button_get_active (gethash "EXPERIMENTERLPGcheckbutton-onlybestfirst" widgets))))
		        (setf timesteps (= 1 (gtk_toggle_button_get_active (gethash "EXPERIMENTERLPGcheckbutton-timesteps" widgets))))

			(setf *lista-algoritmos* (append *lista-algoritmos* (list (list planner-name modo-ejecucion limite-tiempo limite-tiempo-local search-steps number-solutions restarts repeats seed noise static-noise lowmemory nobestfirst onlybestfirst timesteps name))))))

		   (sgplan
                     (progn
			(setf limite-tiempo (read-from-string (gtk_entry_get_text (gethash "EXPERIMENTERIPSStime-bound-entry" widgets))))
			(setf *lista-algoritmos* (append *lista-algoritmos* (list (list planner-name limite-tiempo name))))))

		)
		(print planner-name)
		(print *lista-algoritmos*)
		(hide-window "algoritm-experimenter")
		(show-domains-experimenter)
	1))
)

;;------------------------------------------------------------------------------------------------------------
;; david montero
;; funcion que al pulsar el boton cancel en la pantalla de aadir un nuevo dominio al experimenter,
;; resetea las variables correspondientes y cierra la pantalla
;;------------------------------------------------------------------------------------------------------------

(defun cancel-domain-experimenter ()
	(declare (special widgets))
	(progn
		(setf *tipo-dominio* nil)
		(setf *dominio-experimenter* nil)
		(setf *problema-experimenter* nil)
		(gtk_entry_set_text (gethash "EXPERIMENTERdomain-entry" widgets) (format nil ""))
		(gtk_entry_set_text (gethash "EXPERIMENTERproblem-entry" widgets) (format nil ""))
		(gtk_entry_set_text (gethash "EXPERIMENTERrules-entry" widgets) (format nil ""))
		(hide-window "domain-experimenter")
	1)
)


;;------------------------------------------------------------------------------------------------------------
;; david montero
;; funcion que al pulsar el boton cancel en la pantalla de eleccion de modo del experimenter cierra la pantalla
;;------------------------------------------------------------------------------------------------------------

(defun cancel-mode-experimenter ()
	(declare (special widgets))
	(progn
		(hide-window "mode-experimenter")
	1)
)


;;------------------------------------------------------------------------------------------------------------
;; david montero
;; funcion que al pulsar el boton cancel en la pantalla de eleccion de repeticiones del experimenter cierra la pantalla
;;------------------------------------------------------------------------------------------------------------

(defun cancel-repetitions-experimenter ()
	(declare (special widgets))
	(progn
		(hide-window "repetitions-experimenter")
	1)
)

;;------------------------------------------------------------------------------------------------------------
;; david montero
;; funcion que al pulsar el boton cancel en la pantalla de aadir un nuevo planificador al experimenter,
;; cierra la pantalla sin generar error
;;------------------------------------------------------------------------------------------------------------

(defun cancel-algoritm-experimenter ()
	(declare (special widgets))
	(progn
		(hide-window "algoritm-experimenter")
	1)
)


;;------------------------------------------------------------------------------------------------------------
;; david montero
;; funcion que muestra en la pantalla del experimenter el arbol de los dominios y de los planificadores aadidos
;;------------------------------------------------------------------------------------------------------------

(defun show-domains-experimenter ()
(declare (special widgets))
(let ((window (gethash "scrolledwindow1Experimenter" widgets)) 
      (window1 (gethash "scrolledwindow2Experimenter" widgets)))

  ;; se resetean los arboles actuales de los dominios y de los planificadores
  (if *arbol-dominios*
      (progn (gtk_container_remove window *arbol-dominios*)
	     (gtk_widget_destroy *arbol-dominios*)))
  (if *arbol-algoritmos*
      (progn (gtk_container_remove window1 *arbol-algoritmos*)
	     (gtk_widget_destroy *arbol-algoritmos*)))

  ;; se crean los nuevos arboles de dominios y de planificadores
  (setf *arbol-dominios* (create_view_domains_experimenter))
  (setf *arbol-algoritmos* (create_view_algoritms_experimenter))

  ;; se aaden a la pantalla los arboles creados
  (gtk_container_add window *arbol-dominios*)
  (gtk_container_add window1 *arbol-algoritmos*)
  (show-window "EXPERIMENTER")
))

;;------------------------------------------------------------------------------------------------------------
;; david montero
;; funcion que muestra la pantalla de eleccion de modo del experimenter por dominios o por planificadores
;;------------------------------------------------------------------------------------------------------------

(defun show-mode-experimenter ()
(declare (special widgets))
(if (null *fichero-experimenter*)
		(progn 
			(if (eq *language* 'es)
	      			(error-dialog nil (format nil "Debe tener abierto un fichero de experimento antes"))
				(error-dialog nil (format nil "Must be open one file experimenter first"))
			)
		1)
		(progn
			  (show-window "mode-experimenter")
		)
)
)

;;------------------------------------------------------------------------------------------------------------
;; david montero
;; funcion que muestra la pantalla de eleccion de repeticiones del experimenter por defecto una repeticion
;;------------------------------------------------------------------------------------------------------------

(defun show-repetitions-experimenter ()
			(declare (special widgets))
(if (null *fichero-experimenter*)
		(progn 
			(if (eq *language* 'es)
	      			(error-dialog nil (format nil "Debe tener abierto un fichero de experimento antes"))
				(error-dialog nil (format nil "Must be open one file experimenter first"))
			)
		1)
		(progn
  			(show-window "repetitions-experimenter")
		)
)
)

;;------------------------------------------------------------------------------------------------------------
;; david montero
;; funcion que crea el arbol a mostrar en la parte de los dominios del experimenter
;;------------------------------------------------------------------------------------------------------------

(defun create_view_domains_experimenter ()
(let ((view (gtk_tree_view_new))
      (col (gtk_tree_view_column_new))
      (renderer (gtk_cell_renderer_text_new))
      (model))

      (if (eq *language* 'es)
	  (gtk_tree_view_column_set_title col "Dominios y problemas")
	  (gtk_tree_view_column_set_title col "Domains and problems")
      )

  (gtk_tree_view_column_pack_start col renderer 1)
  
  (gtk_tree_view_column_add_attribute col renderer "markup" 0)

;;  /* pack tree view column into tree view */
  (gtk_tree_view_append_column view col)


  (setq model (create_and_fill_domain_experimenter))
	
  (gtk_tree_view_set_model view model)

;; /* destroy model automatically with view */
  (g_object_unref model)
  view
))

;;------------------------------------------------------------------------------------------------------------
;; david montero
;; funcion que crea el arbol a mostrar en la parte de los planificadores del experimenter
;;------------------------------------------------------------------------------------------------------------

(defun create_view_algoritms_experimenter ()
(let ((view (gtk_tree_view_new))
      (col (gtk_tree_view_column_new))
      (renderer (gtk_cell_renderer_text_new))
      (model))

      (if (eq *language* 'es)
	  (gtk_tree_view_column_set_title col "Planificadores")
	  (gtk_tree_view_column_set_title col "Planners")
      )

  (gtk_tree_view_column_pack_start col renderer 1)
  
  (gtk_tree_view_column_add_attribute col renderer "markup" 0)

;;  /* pack tree view column into tree view */
  (gtk_tree_view_append_column view col)


  (setq model (create_and_fill_algoritm_experimenter))
	
  (gtk_tree_view_set_model view model)

;; /* destroy model automatically with view */
  (g_object_unref model)
  view
))


;;------------------------------------------------------------------------------------------------------------
;; david montero
;; funcion que genera el modelo del arbol a mostrar en la parte de los dominios del experimenter
;;------------------------------------------------------------------------------------------------------------

(defun create_and_fill_domain_experimenter ()
(let ((toplevel)
      (child)
      (domain 0)
      (treestore (gtk_tree_store_newv 1  (make-array '1 :initial-contents '(64)))))

(declare (type GtkTreeIter toplevel))
(declare (type GtkTreeIter child))

;; para todos los dominios que existen en la variable *lista-dominios*
(dolist (dominio *lista-dominios*)
	(setq domain (1+ domain))
	(setq toplevel (gtk_tree_store_append treestore nil))

	;; se aade un nuevo dominio al arbol

      (if (eq *language* 'es)
	(gtk_tree_store_set treestore toplevel 0 (format nil "<span foreground=~s>~a ~a</span>" "darkgreen" "Dominio y problema" domain)  -1)

	(gtk_tree_store_set treestore toplevel 0 (format nil "<span foreground=~s>~a ~a</span>" "darkgreen" "Domain and problem" domain)  -1)
      )

	       ;; se indica el tipo de dominio
	       (setq child (gtk_tree_store_append treestore toplevel))

      (if (eq *language* 'es)
	       (gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Tipo Dominio" (car dominio)) -1)

	       (gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Type Domain" (car dominio)) -1)
      )
		
	       ;; se indica el path del fichero del dominio
	       (setq child (gtk_tree_store_append treestore toplevel))

      (if (eq *language* 'es)
	       (gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Doinio" (cadr dominio)) -1)

	       (gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Domain" (cadr dominio)) -1)
      )

	       ;; se indica el path del fichero de problemas
	       (setq child (gtk_tree_store_append treestore toplevel))

      (if (eq *language* 'es)
	       (gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Problema" (nth 2 dominio)) -1)

	       (gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Problem" (nth 2 dominio)) -1)
      )

	       ;; se indica el fichero de reglas
	       (setq child (gtk_tree_store_append treestore toplevel))

      (if (eq *language* 'es)
	       (gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Reglas" (car (last dominio)) ) -1)

	       (gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Rules" (car (last dominio)) ) -1)
      )

)
 treestore
))


;;------------------------------------------------------------------------------------------------------------
;; david montero
;; funcion que genera el modelo del arbol a mostrar en la parte de los planificadores del experimenter
;;------------------------------------------------------------------------------------------------------------

(defun create_and_fill_algoritm_experimenter ()
(let ((toplevel)
      (child)
      (planner 0)
      (treestore (gtk_tree_store_newv 1  (make-array '1 :initial-contents '(64)))))

(declare (type GtkTreeIter toplevel))
(declare (type GtkTreeIter child))

;; para todos los planificadores en la variable *lista-algortimos*
(dolist (algoritmo *lista-algoritmos*)
	(setq planner (1+ planner))
	(setq toplevel (gtk_tree_store_append treestore nil))
	
	;; se aade un nuevo planificador al arbol

      (if (eq *language* 'es)
	(if (or (equal "nil" (car (last algoritmo))) (equal "" (car (last algoritmo))))
	(gtk_tree_store_set treestore toplevel 0 (format nil "<span foreground=~s>~a ~a</span>" "darkgreen" "Planificador" planner)  -1)
	(gtk_tree_store_set treestore toplevel 0 (format nil "<span foreground=~s>~a </span>" "darkgreen" (car (last algoritmo)))  -1)
	)
	(if (or (equal "nil" (car (last algoritmo))) (equal "" (car (last algoritmo))))
	(gtk_tree_store_set treestore toplevel 0 (format nil "<span foreground=~s>~a ~a</span>" "darkgreen" "Planner" planner)  -1)
	(gtk_tree_store_set treestore toplevel 0 (format nil "<span foreground=~s>~a </span>" "darkgreen" (car (last algoritmo)))  -1)
	)
      )
	       (setq child (gtk_tree_store_append treestore toplevel))

	       ;; se comprueba el tipo de planificador
	       (case (car algoritmo)

		;; si el planificador es ipss
		(ipss
		      (if (eq *language* 'es)
			(progn
			 (gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Tipo  Algoritmo" (car algoritmo)) -1)
		         (setq child (gtk_tree_store_append treestore toplevel))
			 (gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Nombre Planificador" (nth 11 algoritmo)) -1)
		         (setq child (gtk_tree_store_append treestore toplevel))
			 (gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Limite tiempo" (nth 1 algoritmo)) -1)
      	                 (setq child (gtk_tree_store_append treestore toplevel))
			 (gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Verbosity" (nth 2 algoritmo)) -1)
      	                 (setq child (gtk_tree_store_append treestore toplevel))
			 (gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Limite Coste" (nth 3 algoritmo)) -1)
      	                 (setq child (gtk_tree_store_append treestore toplevel))
			 (gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Tipo Coste" (nth 4 algoritmo)) -1)
      	                 (setq child (gtk_tree_store_append treestore toplevel))
			 (gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Multiples Soluciones" (if (null (nth 5 algoritmo)) "No" "Yes")) -1)
      	                 (setq child (gtk_tree_store_append treestore toplevel))
			 (gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Ejecutar sin reglas" (if (null (nth 6 algoritmo)) "No" "Yes")) -1)
      	                 (setq child (gtk_tree_store_append treestore toplevel))
			 (gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Backtrack con costes" (if (null (nth 7 algoritmo)) "No" "Yes")) -1)
      	                 (setq child (gtk_tree_store_append treestore toplevel))
			 (gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Parar en cada problema" (if (null (nth 8 algoritmo)) "No" "Yes")) -1)
      	                 (setq child (gtk_tree_store_append treestore toplevel))
			 (gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Almacenar reglas" (if (null (nth 9 algoritmo)) "No" "Yes")) -1)
      	                 (setq child (gtk_tree_store_append treestore toplevel))
			 (gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Almacenar planes" (if (null (nth 10 algoritmo)) "No" "Yes")) -1)
			)

			;; se aaden las caracteristicas del planificador IPSS concreto
			(progn
			 (gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Type Algoritm" (car algoritmo)) -1)
		         (setq child (gtk_tree_store_append treestore toplevel))
			 (gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Planner Name" (nth 11 algoritmo)) -1)
		         (setq child (gtk_tree_store_append treestore toplevel))
			 (gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Time Bound" (nth 1 algoritmo)) -1)
      	                 (setq child (gtk_tree_store_append treestore toplevel))
			 (gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Verbosity" (nth 2 algoritmo)) -1)
      	                 (setq child (gtk_tree_store_append treestore toplevel))
			 (gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Cost Bound" (nth 3 algoritmo)) -1)
      	                 (setq child (gtk_tree_store_append treestore toplevel))
			 (gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Cost Type" (nth 4 algoritmo)) -1)
      	                 (setq child (gtk_tree_store_append treestore toplevel))
			 (gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Multiples Solutions" (if (null (nth 5 algoritmo)) "No" "Yes")) -1)
      	                 (setq child (gtk_tree_store_append treestore toplevel))
			 (gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Run Without Rules" (if (null (nth 6 algoritmo)) "No" "Yes")) -1)
      	                 (setq child (gtk_tree_store_append treestore toplevel))
			 (gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Backtrack With Cost" (if (null (nth 7 algoritmo)) "No" "Yes")) -1)
      	                 (setq child (gtk_tree_store_append treestore toplevel))
			 (gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Stop After Problem" (if (null (nth 8 algoritmo)) "No" "Yes")) -1)
      	                 (setq child (gtk_tree_store_append treestore toplevel))
			 (gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Store Fired Rules" (if (null (nth 9 algoritmo)) "No" "Yes")) -1)
      	                 (setq child (gtk_tree_store_append treestore toplevel))
			 (gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Store Plans" (if (null (nth 10 algoritmo)) "No" "Yes")) -1))))

		;; si el planificador es sayphi
		(sayphi

		      ;; se aaden las caracteristicas del planificador SAYPHI
		      (if (eq *language* 'es)

 			(progn
			 (gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Tipo Algoritmo" (car algoritmo)) -1)
		         (setq child (gtk_tree_store_append treestore toplevel))
			 (gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Nombre Planificador" (nth 5 algoritmo)) -1)
		         (setq child (gtk_tree_store_append treestore toplevel))
			 (gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Limite Coste" (nth 1 algoritmo)) -1)
      	                 (setq child (gtk_tree_store_append treestore toplevel))
			 (gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Verbosity" (nth 2 algoritmo)) -1)
      	                 (setq child (gtk_tree_store_append treestore toplevel))
			 (gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Heuristica" (nth 3 algoritmo)) -1)
      	                 (setq child (gtk_tree_store_append treestore toplevel))
			 (gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Algoritmo" (nth 4 algoritmo)) -1))


			(progn
			 (gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Type Algoritm" (car algoritmo)) -1)
		         (setq child (gtk_tree_store_append treestore toplevel))
			 (gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Planner Name" (nth 5 algoritmo)) -1)
		         (setq child (gtk_tree_store_append treestore toplevel))
			 (gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Time Bound" (nth 1 algoritmo)) -1)
      	                 (setq child (gtk_tree_store_append treestore toplevel))
			 (gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Verbosity" (nth 2 algoritmo)) -1)
      	                 (setq child (gtk_tree_store_append treestore toplevel))
			 (gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Heuristic" (nth 3 algoritmo)) -1)
      	                 (setq child (gtk_tree_store_append treestore toplevel))
			 (gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Algoritm" (nth 4 algoritmo)) -1))))		

		;; si el planificador es metric-ff
		(metric-ff
		      (if (eq *language* 'es)
			(progn
			 (gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Tipo  Algoritmo" (car algoritmo)) -1)
		         (setq child (gtk_tree_store_append treestore toplevel))
			 (gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Nombre Planificador" (nth 9 algoritmo)) -1)
		         (setq child (gtk_tree_store_append treestore toplevel))
			 (gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Limite tiempo" (nth 1 algoritmo)) -1)
      	                 (setq child (gtk_tree_store_append treestore toplevel))
			 (gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Tamanio maximo en lista abierta" (nth 2 algoritmo)) -1)
      	                 (setq child (gtk_tree_store_append treestore toplevel))
			 (gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Peso de la funcion g" (nth 3 algoritmo)) -1)
      	                 (setq child (gtk_tree_store_append treestore toplevel))
			 (gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Peso de la heuristica" (nth 4 algoritmo)) -1)
      	                 (setq child (gtk_tree_store_append treestore toplevel))
			 (gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Optimizacion de metrica" (if (null (nth 5 algoritmo)) "No" "Yes")) -1)
      	                 (setq child (gtk_tree_store_append treestore toplevel))
			 (gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Heuristica" (nth 6 algoritmo)) -1)
      	                 (setq child (gtk_tree_store_append treestore toplevel))
			 (gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Algoritmo" (nth 7 algoritmo)) -1)
      	                 (setq child (gtk_tree_store_append treestore toplevel))
			 (gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Anytime" (nth 8 algoritmo)) -1))

			;; se aaden las caracteristicas del planificador IPSS concreto
			(progn
			 (gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Type Algoritm" (car algoritmo)) -1)
		         (setq child (gtk_tree_store_append treestore toplevel))
			 (gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Planner Name" (nth 9 algoritmo)) -1)
		         (setq child (gtk_tree_store_append treestore toplevel))
			 (gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Time Bound" (nth 1 algoritmo)) -1)
      	                 (setq child (gtk_tree_store_append treestore toplevel))
			 (gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Maximun size of the open list" (nth 2 algoritmo)) -1)
      	                 (setq child (gtk_tree_store_append treestore toplevel))
			 (gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Weight of g function" (nth 3 algoritmo)) -1)
      	                 (setq child (gtk_tree_store_append treestore toplevel))
			 (gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Weight of heuristic" (nth 4 algoritmo)) -1)
      	                 (setq child (gtk_tree_store_append treestore toplevel))
			 (gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Optimization of metric" (if (null (nth 5 algoritmo)) "No" "Yes")) -1)
      	                 (setq child (gtk_tree_store_append treestore toplevel))
			 (gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Heuristic" (nth 6 algoritmo)) -1)
      	                 (setq child (gtk_tree_store_append treestore toplevel))
			 (gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Algorithm" (nth 7 algoritmo)) -1)
      	                 (setq child (gtk_tree_store_append treestore toplevel))
			 (gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Anytime" (nth 8 algoritmo)) -1))))


		;; si el planificador es sgplan
		(sgplan
		      (if (eq *language* 'es)
			(progn
			 (gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Tipo  Algoritmo" (car algoritmo)) -1)
		         (setq child (gtk_tree_store_append treestore toplevel))
			 (gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Nombre Planificador" (nth 2 algoritmo)) -1)
		         (setq child (gtk_tree_store_append treestore toplevel))
			 (gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Limite tiempo" (nth 1 algoritmo)) -1))

			;; se aaden las caracteristicas del planificador IPSS concreto
			(progn
			 (gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Type Algoritm" (car algoritmo)) -1)
		         (setq child (gtk_tree_store_append treestore toplevel))
			 (gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Planner Name" (nth 2 algoritmo)) -1)
		         (setq child (gtk_tree_store_append treestore toplevel))
			 (gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Time Bound" (nth 1 algoritmo)) -1))))

		;; si el planificador es lpg
		(lpg
		      (if (eq *language* 'es)
			(progn
			 (gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Tipo  Algoritmo" (car algoritmo)) -1)
		         (setq child (gtk_tree_store_append treestore toplevel))
			 (gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Nombre Planificador" (car (last algoritmo))) -1)
		         (setq child (gtk_tree_store_append treestore toplevel))
			 (gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Modo ejecucion" (nth 1 algoritmo)) -1)
		         (setq child (gtk_tree_store_append treestore toplevel))
			 (gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Limite tiempo" (nth 2 algoritmo)) -1)
		         (setq child (gtk_tree_store_append treestore toplevel))
			 (gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Limite tiempo local" (nth 3 algoritmo)) -1)
		         (setq child (gtk_tree_store_append treestore toplevel))
			 (gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Pasos en la busqueda" (nth 4 algoritmo)) -1)
		         (setq child (gtk_tree_store_append treestore toplevel))
			 (gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Numero de soluciones" (nth 5 algoritmo)) -1)
		         (setq child (gtk_tree_store_append treestore toplevel))
			 (gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Reinicios" (nth 6 algoritmo)) -1)
		         (setq child (gtk_tree_store_append treestore toplevel))
			 (gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Repeticiones" (nth 7 algoritmo)) -1)
		         (setq child (gtk_tree_store_append treestore toplevel))
			 (gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Semilla" (nth 8 algoritmo)) -1)
		         (setq child (gtk_tree_store_append treestore toplevel))
			 (gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Ruido" (nth 9 algoritmo)) -1)
		         (setq child (gtk_tree_store_append treestore toplevel))
			 ;;(gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Almacenar planes" (nth 10 algoritmo)) -1)
		         ;;(setq child (gtk_tree_store_append treestore toplevel))
			 ;;(gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Sin verbose" (nth 11 algoritmo)) -1)
		         ;;(setq child (gtk_tree_store_append treestore toplevel))
			 (gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Ruido statico" (nth 10 algoritmo)) -1)
		         (setq child (gtk_tree_store_append treestore toplevel))
			 (gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Memoria baja" (nth 11 algoritmo)) -1)
		         (setq child (gtk_tree_store_append treestore toplevel))
			 (gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "No mejor primero" (nth 12 algoritmo)) -1)
		         (setq child (gtk_tree_store_append treestore toplevel))
			 (gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Solo mejor primero" (nth 13 algoritmo)) -1)
		         (setq child (gtk_tree_store_append treestore toplevel))
			 (gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Tiempo en cada paso" (nth 14 algoritmo)) -1))

			(progn
			 (gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Algorithm Type" (car algoritmo)) -1)
		         (setq child (gtk_tree_store_append treestore toplevel))
			 (gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Planner Name" (car (last algoritmo))) -1)
		         (setq child (gtk_tree_store_append treestore toplevel))
			 (gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Execution mode" (nth 1 algoritmo)) -1)
		         (setq child (gtk_tree_store_append treestore toplevel))
			 (gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Time bound" (nth 2 algoritmo)) -1)
		         (setq child (gtk_tree_store_append treestore toplevel))
			 (gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Time bound local" (nth 3 algoritmo)) -1)
		         (setq child (gtk_tree_store_append treestore toplevel))
			 (gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Search steps" (nth 4 algoritmo)) -1)
		         (setq child (gtk_tree_store_append treestore toplevel))
			 (gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Number solutions" (nth 5 algoritmo)) -1)
		         (setq child (gtk_tree_store_append treestore toplevel))
			 (gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Restarts" (nth 6 algoritmo)) -1)
		         (setq child (gtk_tree_store_append treestore toplevel))
			 (gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Repeats" (nth 7 algoritmo)) -1)
		         (setq child (gtk_tree_store_append treestore toplevel))
			 (gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Seed" (nth 8 algoritmo)) -1)
		         (setq child (gtk_tree_store_append treestore toplevel))
			 (gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Noise" (nth 9 algoritmo)) -1)
		         (setq child (gtk_tree_store_append treestore toplevel))
			 ;;(gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Save plans" (nth 10 algoritmo)) -1)
		         ;;(setq child (gtk_tree_store_append treestore toplevel))
			 ;;(gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Verbose off" (nth 11 algoritmo)) -1)
		         ;;(setq child (gtk_tree_store_append treestore toplevel))
			 (gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Static noise" (nth 10 algoritmo)) -1)
		         (setq child (gtk_tree_store_append treestore toplevel))
			 (gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Low memory" (nth 11 algoritmo)) -1)
		         (setq child (gtk_tree_store_append treestore toplevel))
			 (gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "No best first" (nth 12 algoritmo)) -1)
		         (setq child (gtk_tree_store_append treestore toplevel))
			 (gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Only best first" (nth 13 algoritmo)) -1)
		         (setq child (gtk_tree_store_append treestore toplevel))
			 (gtk_tree_store_set treestore child 0  (format nil "<span foreground=~s>~a:</span> ~a" "red" "Time steps" (nth 14 algoritmo)) -1))))

)


)
 treestore
))


;;------------------------------------------------------------------------------------------------------------
;; david montero
;; funcion que abre la pantalla de eleccion de planificador en el experimenter
;;------------------------------------------------------------------------------------------------------------

(defun start-window-algoritm-experimenter()
	(declare (special widgets))
	(progn
	     (show-window "algoritm-experimenter")
	     (planner-interface-experimenter)	
	     (set_IPSScost-bound-EXPERIMENTER)	    
	1))


;;------------------------------------------------------------------------------------------------------------
;; david montero
;; funcion que modifica los atributos a elegir dependiendo del planificador elegido en la pantalla del experimenter
;;------------------------------------------------------------------------------------------------------------

(defun planner-interface-experimenter ()
 (declare (special widgets))

 ;; lee el tipo de planificador elegido
 (let ((planner-name (read-from-string 
                     (gtk_combo_box_get_active_text (gethash "EXPERIMENTERcombobox1" widgets)))))

   ;; si el planificador es ipss se oculta todo lo relativo a SAYPHI y se muestra todo lo de IPSS
   (if (eq planner-name 'ipss)
     (progn
        (hide-window "EXPERIMENTERSAYPHIheuristic-label")
        (hide-window "EXPERIMENTERSAYPHIheuristic-entry")
        (hide-window "EXPERIMENTERSAYPHIalgorithm-label")
        (hide-window "EXPERIMENTERSAYPHIalgorithm-entry")

	(hide-window "EXPERIMENTERLPGmodoejecucion-label")
	(hide-window "EXPERIMENTERLPGmodoejecucion-entry")
	(hide-window "EXPERIMENTERLPGsearchsteps-label")
	(hide-window "EXPERIMENTERLPGsearchsteps-entry")
	(hide-window "EXPERIMENTERLPGrestarts-label")
	(hide-window "EXPERIMENTERLPGrestarts-entry")
	(hide-window "EXPERIMENTERLPGrepeats-label")
	(hide-window "EXPERIMENTERLPGrepeats-entry")
	(hide-window "EXPERIMENTERLPGseed-label")
	(hide-window "EXPERIMENTERLPGseed-entry")
	(hide-window "EXPERIMENTERLPGtime-bound-local-label")
	(hide-window "EXPERIMENTERLPGtime-bound-local-entry")
	(hide-window "EXPERIMENTERLPGnumber-solutions-label")
	(hide-window "EXPERIMENTERLPGnumber-solutions-entry")
	(hide-window "EXPERIMENTERLPGnoise-label")
	(hide-window "EXPERIMENTERLPGnoise-entry")
        ;;(hide-window "EXPERIMENTERLPGcheckbutton-save-plans")
        ;;(hide-window "EXPERIMENTERLPGcheckbutton-verbose")
        (hide-window "EXPERIMENTERLPGcheckbutton-static-noise")
        ;;(hide-window "EXPERIMENTERLPGcheckbutton-maxnoise")
        (hide-window "EXPERIMENTERLPGcheckbutton-lowmemory")
        (hide-window "EXPERIMENTERLPGcheckbutton-nobestfirst")
        (hide-window "EXPERIMENTERLPGcheckbutton-onlybestfirst")
        (hide-window "EXPERIMENTERLPGcheckbutton-timesteps")


	(hide-window "EXPERIMENTERMETRICheuristic-entry")
	(hide-window "EXPERIMENTERMETRICalgorithm-entry")
	(hide-window "EXPERIMENTERMETRICpoda-entry")
	(hide-window "EXPERIMENTERMETRICpoda-label")
	(hide-window "EXPERIMENTERMETRICw-g-entry")
	(hide-window "EXPERIMENTERMETRICw-h-entry")
	(hide-window "EXPERIMENTERMETRICw-g-label")
	(hide-window "EXPERIMENTERMETRICw-h-label")
	(hide-window "EXPERIMENTERMETRICmax-tam-label")
	(hide-window "EXPERIMENTERMETRICmax-tam-entry")
	(hide-window "EXPERIMENTERMETRICcheckbutton-metric-optimice")
        (show-window "EXPERIMENTERIPSSspin-verbosity-entry")
        (show-window "EXPERIMENTERIPSSverbosity-label")
        (show-window "EXPERIMENTERlabel70")
        (show-window "EXPERIMENTERbutton-rules")
	(show-window "EXPERIMENTERIPSSoutput-rules-entry")
	(show-window "EXPERIMENTERIPSScost-type-label")
	(show-window "EXPERIMENTERIPSScheckbutton-multiple-sols-p")
        (show-window "EXPERIMENTERIPSScheckbutton-backtrack-with-costs-p")
	(show-window "EXPERIMENTERIPSScheckbutton-cost-bound")
        (show-window "EXPERIMENTERIPSScheckbutton-stop-after-problem-p")
	(show-window "EXPERIMENTERIPSScheckbutton-store-fired-rules")
        (show-window "EXPERIMENTERIPSScheckbutton-store-plans")
        (show-window "EXPERIMENTERIPSScheckbutton-run-without-p")
        (show-window "EXPERIMENTERIPSScost-type-entry")
        (show-window "EXPERIMENTERIPSScost-bound-entry")
        (show-window "EXPERIMENTERbutton-ipss-show-search-tree")))

    ;; si el planificador es sayphi se oculta todo lo relativo a IPSS y se muestra todo lo de SAYPHI
    (if (eq planner-name 'sayphi)
     (progn
        (show-window "EXPERIMENTERSAYPHIheuristic-label")
        (show-window "EXPERIMENTERSAYPHIheuristic-entry") 
        (show-window "EXPERIMENTERSAYPHIalgorithm-label")
        (show-window "EXPERIMENTERSAYPHIalgorithm-entry")        
        (show-window "EXPERIMENTERIPSSspin-verbosity-entry")
        (show-window "EXPERIMENTERIPSSverbosity-label")
	(hide-window "EXPERIMENTERlabel70")
        (hide-window "EXPERIMENTERbutton-rules")
	(hide-window "EXPERIMENTERIPSSoutput-rules-entry")
	(hide-window "EXPERIMENTERIPSScost-type-label")
        (hide-window "EXPERIMENTERIPSScheckbutton-multiple-sols-p")
	(hide-window "EXPERIMENTERIPSScheckbutton-backtrack-with-costs-p")
	(hide-window "EXPERIMENTERIPSScheckbutton-cost-bound")
        (hide-window "EXPERIMENTERIPSScheckbutton-stop-after-problem-p")
	(hide-window "EXPERIMENTERIPSScheckbutton-store-fired-rules")
        (hide-window "EXPERIMENTERIPSScheckbutton-store-plans")
        (hide-window "EXPERIMENTERIPSScheckbutton-run-without-p")
        (hide-window "EXPERIMENTERIPSScost-type-entry")
        (hide-window "EXPERIMENTERIPSScost-bound-entry")

	(hide-window "EXPERIMENTERLPGmodoejecucion-label")
	(hide-window "EXPERIMENTERLPGmodoejecucion-entry")
	(hide-window "EXPERIMENTERLPGsearchsteps-label")
	(hide-window "EXPERIMENTERLPGsearchsteps-entry")
	(hide-window "EXPERIMENTERLPGrestarts-label")
	(hide-window "EXPERIMENTERLPGrestarts-entry")
	(hide-window "EXPERIMENTERLPGrepeats-label")
	(hide-window "EXPERIMENTERLPGrepeats-entry")
	(hide-window "EXPERIMENTERLPGseed-label")
	(hide-window "EXPERIMENTERLPGseed-entry")
	(hide-window "EXPERIMENTERLPGtime-bound-local-label")
	(hide-window "EXPERIMENTERLPGtime-bound-local-entry")
	(hide-window "EXPERIMENTERLPGnumber-solutions-label")
	(hide-window "EXPERIMENTERLPGnumber-solutions-entry")
	(hide-window "EXPERIMENTERLPGnoise-label")
	(hide-window "EXPERIMENTERLPGnoise-entry")
        ;;(hide-window "EXPERIMENTERLPGcheckbutton-save-plans")
        ;;(hide-window "EXPERIMENTERLPGcheckbutton-verbose")
        (hide-window "EXPERIMENTERLPGcheckbutton-static-noise")
        ;;(hide-window "EXPERIMENTERLPGcheckbutton-maxnoise")
        (hide-window "EXPERIMENTERLPGcheckbutton-lowmemory")
        (hide-window "EXPERIMENTERLPGcheckbutton-nobestfirst")
        (hide-window "EXPERIMENTERLPGcheckbutton-onlybestfirst")
        (hide-window "EXPERIMENTERLPGcheckbutton-timesteps")

	(hide-window "EXPERIMENTERMETRICheuristic-entry")
	(hide-window "EXPERIMENTERMETRICalgorithm-entry")
	(hide-window "EXPERIMENTERMETRICw-g-entry")
	(hide-window "EXPERIMENTERMETRICw-h-entry")
	(hide-window "EXPERIMENTERMETRICw-g-label")
	(hide-window "EXPERIMENTERMETRICw-h-label")
	(hide-window "EXPERIMENTERMETRICpoda-entry")
	(hide-window "EXPERIMENTERMETRICpoda-label")
	(hide-window "EXPERIMENTERMETRICmax-tam-label")
	(hide-window "EXPERIMENTERMETRICmax-tam-entry")
	(hide-window "EXPERIMENTERMETRICcheckbutton-metric-optimice")
        (hide-window "EXPERIMENTERbutton-ipss-show-search-tree")))

    (if (eq planner-name 'metric-ff)
      (progn	
        (hide-window "EXPERIMENTERSAYPHIheuristic-entry")
        (hide-window "EXPERIMENTERSAYPHIalgorithm-entry")
	(hide-window "EXPERIMENTERIPSScost-type-label")
        (hide-window "EXPERIMENTERIPSScheckbutton-multiple-sols-p")
	(hide-window "EXPERIMENTERIPSScheckbutton-backtrack-with-costs-p")
	(hide-window "EXPERIMENTERIPSScheckbutton-cost-bound")
        (hide-window "EXPERIMENTERIPSScheckbutton-stop-after-problem-p")
	(hide-window "EXPERIMENTERIPSScheckbutton-store-fired-rules")
        (hide-window "EXPERIMENTERIPSScheckbutton-store-plans")
        (hide-window "EXPERIMENTERIPSScheckbutton-run-without-p")
        (hide-window "EXPERIMENTERIPSScost-type-entry")
        (hide-window "EXPERIMENTERIPSScost-bound-entry")
        (hide-window "EXPERIMENTERIPSSspin-verbosity-entry")
        (hide-window "EXPERIMENTERIPSSverbosity-label")
	(hide-window "EXPERIMENTERbutton-ipss-show-search-tree")

	(hide-window "EXPERIMENTERLPGmodoejecucion-label")
	(hide-window "EXPERIMENTERLPGmodoejecucion-entry")
	(hide-window "EXPERIMENTERLPGsearchsteps-label")
	(hide-window "EXPERIMENTERLPGsearchsteps-entry")
	(hide-window "EXPERIMENTERLPGrestarts-label")
	(hide-window "EXPERIMENTERLPGrestarts-entry")
	(hide-window "EXPERIMENTERLPGrepeats-label")
	(hide-window "EXPERIMENTERLPGrepeats-entry")
	(hide-window "EXPERIMENTERLPGseed-label")
	(hide-window "EXPERIMENTERLPGseed-entry")
	(hide-window "EXPERIMENTERLPGtime-bound-local-label")
	(hide-window "EXPERIMENTERLPGtime-bound-local-entry")
	(hide-window "EXPERIMENTERLPGnumber-solutions-label")
	(hide-window "EXPERIMENTERLPGnumber-solutions-entry")
	(hide-window "EXPERIMENTERLPGnoise-label")
	(hide-window "EXPERIMENTERLPGnoise-entry")
        ;;(hide-window "EXPERIMENTERLPGcheckbutton-save-plans")
        ;;(hide-window "EXPERIMENTERLPGcheckbutton-verbose")
        (hide-window "EXPERIMENTERLPGcheckbutton-static-noise")
        ;;(hide-window "EXPERIMENTERLPGcheckbutton-maxnoise")
        (hide-window "EXPERIMENTERLPGcheckbutton-lowmemory")
        (hide-window "EXPERIMENTERLPGcheckbutton-nobestfirst")
        (hide-window "EXPERIMENTERLPGcheckbutton-onlybestfirst")
        (hide-window "EXPERIMENTERLPGcheckbutton-timesteps")

        (show-window "EXPERIMENTERSAYPHIheuristic-label")
        (show-window "EXPERIMENTERSAYPHIalgorithm-label")
        (show-window "EXPERIMENTERlabel70")
        (show-window "EXPERIMENTERbutton-rules")
	(show-window "EXPERIMENTERIPSSoutput-rules-entry")
	(show-window "EXPERIMENTERMETRICheuristic-entry")
	(show-window "EXPERIMENTERMETRICalgorithm-entry")
	(show-window "EXPERIMENTERMETRICw-g-entry")
	(show-window "EXPERIMENTERMETRICw-h-entry")
	(show-window "EXPERIMENTERMETRICw-g-label")
	(show-window "EXPERIMENTERMETRICw-h-label")
	(show-window "EXPERIMENTERMETRICpoda-entry")
	(show-window "EXPERIMENTERMETRICpoda-label")
	(show-window "EXPERIMENTERMETRICmax-tam-label")
	(show-window "EXPERIMENTERMETRICmax-tam-entry")
	(show-window "EXPERIMENTERMETRICcheckbutton-metric-optimice")))

    ;; si el planificador es sgplan
    (if (eq planner-name 'sgplan)
      (progn	
        (hide-window "EXPERIMENTERSAYPHIheuristic-entry")
        (hide-window "EXPERIMENTERSAYPHIalgorithm-entry")
	(hide-window "EXPERIMENTERIPSScost-type-label")
        (hide-window "EXPERIMENTERIPSScheckbutton-multiple-sols-p")
	(hide-window "EXPERIMENTERIPSScheckbutton-backtrack-with-costs-p")
	(hide-window "EXPERIMENTERIPSScheckbutton-cost-bound")
        (hide-window "EXPERIMENTERIPSScheckbutton-stop-after-problem-p")
	(hide-window "EXPERIMENTERIPSScheckbutton-store-fired-rules")
        (hide-window "EXPERIMENTERIPSScheckbutton-store-plans")
        (hide-window "EXPERIMENTERIPSScheckbutton-run-without-p")
        (hide-window "EXPERIMENTERIPSScost-type-entry")
        (hide-window "EXPERIMENTERIPSScost-bound-entry")
        (hide-window "EXPERIMENTERIPSSspin-verbosity-entry")
        (hide-window "EXPERIMENTERIPSSverbosity-label")
	(hide-window "EXPERIMENTERbutton-ipss-show-search-tree")
        (hide-window "EXPERIMENTERSAYPHIheuristic-label")
        (hide-window "EXPERIMENTERSAYPHIalgorithm-label")
        (hide-window "EXPERIMENTERlabel70")
        (hide-window "EXPERIMENTERbutton-rules")
	(hide-window "EXPERIMENTERIPSSoutput-rules-entry")

	(hide-window "EXPERIMENTERLPGmodoejecucion-label")
	(hide-window "EXPERIMENTERLPGmodoejecucion-entry")
	(hide-window "EXPERIMENTERLPGsearchsteps-label")
	(hide-window "EXPERIMENTERLPGsearchsteps-entry")
	(hide-window "EXPERIMENTERLPGrestarts-label")
	(hide-window "EXPERIMENTERLPGrestarts-entry")
	(hide-window "EXPERIMENTERLPGrepeats-label")
	(hide-window "EXPERIMENTERLPGrepeats-entry")
	(hide-window "EXPERIMENTERLPGseed-label")
	(hide-window "EXPERIMENTERLPGseed-entry")
	(hide-window "EXPERIMENTERLPGtime-bound-local-label")
	(hide-window "EXPERIMENTERLPGtime-bound-local-entry")
	(hide-window "EXPERIMENTERLPGnumber-solutions-label")
	(hide-window "EXPERIMENTERLPGnumber-solutions-entry")
	(hide-window "EXPERIMENTERLPGnoise-label")
	(hide-window "EXPERIMENTERLPGnoise-entry")
        ;;(hide-window "EXPERIMENTERLPGcheckbutton-save-plans")
        ;;(hide-window "EXPERIMENTERLPGcheckbutton-verbose")
        (hide-window "EXPERIMENTERLPGcheckbutton-static-noise")
        ;;(hide-window "EXPERIMENTERLPGcheckbutton-maxnoise")
        (hide-window "EXPERIMENTERLPGcheckbutton-lowmemory")
        (hide-window "EXPERIMENTERLPGcheckbutton-nobestfirst")
        (hide-window "EXPERIMENTERLPGcheckbutton-onlybestfirst")
        (hide-window "EXPERIMENTERLPGcheckbutton-timesteps")

	(hide-window "EXPERIMENTERMETRICheuristic-entry")
	(hide-window "EXPERIMENTERMETRICalgorithm-entry")
	(hide-window "EXPERIMENTERMETRICw-g-entry")
	(hide-window "EXPERIMENTERMETRICw-h-entry")
	(hide-window "EXPERIMENTERMETRICw-g-label")
	(hide-window "EXPERIMENTERMETRICw-h-label")
	(hide-window "EXPERIMENTERMETRICpoda-entry")
	(hide-window "EXPERIMENTERMETRICpoda-label")
	(hide-window "EXPERIMENTERMETRICmax-tam-label")
	(hide-window "EXPERIMENTERMETRICmax-tam-entry")
	(hide-window "EXPERIMENTERMETRICcheckbutton-metric-optimice")))

 (if (eq planner-name 'lpg)
      (progn	
        (hide-window "EXPERIMENTERSAYPHIheuristic-entry")
        (hide-window "EXPERIMENTERSAYPHIalgorithm-entry")
	(hide-window "EXPERIMENTERIPSScost-type-label")
        (hide-window "EXPERIMENTERIPSScheckbutton-multiple-sols-p")
	(hide-window "EXPERIMENTERIPSScheckbutton-backtrack-with-costs-p")
	(hide-window "EXPERIMENTERIPSScheckbutton-cost-bound")
        (hide-window "EXPERIMENTERIPSScheckbutton-stop-after-problem-p")
	(hide-window "EXPERIMENTERIPSScheckbutton-store-fired-rules")
        (hide-window "EXPERIMENTERIPSScheckbutton-store-plans")
        (hide-window "EXPERIMENTERIPSScheckbutton-run-without-p")
        (hide-window "EXPERIMENTERIPSScost-type-entry")
        (hide-window "EXPERIMENTERIPSScost-bound-entry")
        (hide-window "EXPERIMENTERIPSSspin-verbosity-entry")
        (hide-window "EXPERIMENTERIPSSverbosity-label")
	(hide-window "EXPERIMENTERbutton-ipss-show-search-tree")
        (hide-window "EXPERIMENTERSAYPHIheuristic-label")
        (hide-window "EXPERIMENTERSAYPHIalgorithm-label")
        (hide-window "EXPERIMENTERlabel70")
        (hide-window "EXPERIMENTERbutton-rules")
	(hide-window "EXPERIMENTERIPSSoutput-rules-entry")

	(show-window "EXPERIMENTERLPGmodoejecucion-label")
	(show-window "EXPERIMENTERLPGmodoejecucion-entry")
	(show-window "EXPERIMENTERLPGsearchsteps-label")
	(show-window "EXPERIMENTERLPGsearchsteps-entry")
	(show-window "EXPERIMENTERLPGrestarts-label")
	(show-window "EXPERIMENTERLPGrestarts-entry")
	(show-window "EXPERIMENTERLPGrepeats-label")
	(show-window "EXPERIMENTERLPGrepeats-entry")
	(show-window "EXPERIMENTERLPGseed-label")
	(show-window "EXPERIMENTERLPGseed-entry")
	(show-window "EXPERIMENTERLPGtime-bound-local-label")
	(show-window "EXPERIMENTERLPGtime-bound-local-entry")
	(show-window "EXPERIMENTERLPGnumber-solutions-label")
	(show-window "EXPERIMENTERLPGnumber-solutions-entry")
	(show-window "EXPERIMENTERLPGnoise-label")
	(show-window "EXPERIMENTERLPGnoise-entry")
        ;;(show-window "EXPERIMENTERLPGcheckbutton-save-plans")
        ;;(show-window "EXPERIMENTERLPGcheckbutton-verbose")
        (show-window "EXPERIMENTERLPGcheckbutton-static-noise")
        ;;(show-window "EXPERIMENTERLPGcheckbutton-maxnoise")
        (show-window "EXPERIMENTERLPGcheckbutton-lowmemory")
        (show-window "EXPERIMENTERLPGcheckbutton-nobestfirst")
        (show-window "EXPERIMENTERLPGcheckbutton-onlybestfirst")
        (show-window "EXPERIMENTERLPGcheckbutton-timesteps")

	(hide-window "EXPERIMENTERMETRICheuristic-entry")
	(hide-window "EXPERIMENTERMETRICalgorithm-entry")
	(hide-window "EXPERIMENTERMETRICw-g-entry")
	(hide-window "EXPERIMENTERMETRICw-h-entry")
	(hide-window "EXPERIMENTERMETRICw-g-label")
	(hide-window "EXPERIMENTERMETRICw-h-label")
	(hide-window "EXPERIMENTERMETRICpoda-entry")
	(hide-window "EXPERIMENTERMETRICpoda-label")
	(hide-window "EXPERIMENTERMETRICmax-tam-label")
	(hide-window "EXPERIMENTERMETRICmax-tam-entry")
	(hide-window "EXPERIMENTERMETRICcheckbutton-metric-optimice")))
    )
 1)


;; --------------------------------------------------------------------------------------------
;; window vl-w-ipss
;; Opciones de LPG y Incremental en el EXPERIMENTER
;; signal toggled
;; --------------------------------------------------------------------------------------------

(defun set_lpg-incremental-EXPERIMENTER ()
  (declare (special widgets))  
  (if (eq 'INCREMENTAL (read-from-string (gtk_combo_box_get_active_text (gethash "EXPERIMENTERLPGmodoejecucion-entry" widgets))))
       (progn  
             (gtk_widget_set_sensitive (gethash "EXPERIMENTERLPGnumber-solutions-entry" widgets) 1)
	     )
       (progn 
              (gtk_widget_set_sensitive (gethash "EXPERIMENTERLPGnumber-solutions-entry" widgets) 0)
	      ))
  1)

;; --------------------------------------------------------------------------------------------
;; window vl-w-ipss
;; Opciones de LPG y best-first en el EXPERIMENTER
;; signal toggled
;; --------------------------------------------------------------------------------------------

(defun set_lpg-best-first-EXPERIMENTER ()
  (declare (special widgets))  
    (if (= 1 (gtk_toggle_button_get_active (gethash "EXPERIMENTERLPGcheckbutton-nobestfirst" widgets)))
       (progn  
              (gtk_widget_set_sensitive (gethash "EXPERIMENTERLPGcheckbutton-onlybestfirst" widgets) 0))
       (progn  
              (gtk_widget_set_sensitive (gethash "EXPERIMENTERLPGcheckbutton-onlybestfirst" widgets) 1)))

    (if (= 1 (gtk_toggle_button_get_active (gethash "EXPERIMENTERLPGcheckbutton-onlybestfirst" widgets)))
       (progn  
              (gtk_widget_set_sensitive (gethash "EXPERIMENTERLPGcheckbutton-nobestfirst" widgets) 0))
       (progn  
              (gtk_widget_set_sensitive (gethash "EXPERIMENTERLPGcheckbutton-nobestfirst" widgets) 1)))

  1) 





;; --------------------------------------------------------------------------------------------
;; window "algoritm-experimenter"
;; ToggleButton EXPERIMENTERIPSScheckbutton-cost-bound
;; signal toggled
;; --------------------------------------------------------------------------------------------
(defun set_IPSScost-bound-EXPERIMENTER ()
  (declare (special widgets))
  (if (= 1 (gtk_toggle_button_get_active (gethash "EXPERIMENTERIPSScheckbutton-cost-bound" widgets)))
       (progn (interface-trace t "cost-bound is going to be introduced" ) 
              (gtk_widget_set_sensitive (gethash "EXPERIMENTERIPSScost-bound-entry" widgets) 1))                         
       (progn (interface-trace t "There is not cost-bound") 
              (gtk_widget_set_sensitive (gethash "EXPERIMENTERIPSScost-bound-entry" widgets) 0)))
  1) 


;; --------------------------------------------------------------------------------------------
;; window "algoritm-experimenter"
;; Opciones de Metric FF y Best-first
;; signal toggled
;; --------------------------------------------------------------------------------------------

(defun set_metric-best-first-experimenter ()
  (declare (special widgets))  
  (if (eq 'best-first (read-from-string (gtk_combo_box_get_active_text (gethash "EXPERIMENTERMETRICalgorithm-entry" widgets))))
       (progn  
              (gtk_widget_set_sensitive (gethash "EXPERIMENTERMETRICpoda-entry" widgets) 1)
              (gtk_widget_set_sensitive (gethash "EXPERIMENTERMETRICmax-tam-entry" widgets) 1))                         
       (progn 
	      (gtk_combo_box_set_active (gethash "EXPERIMENTERMETRICpoda-entry" widgets) 3)
	      (gtk_widget_set_sensitive (gethash "EXPERIMENTERMETRICpoda-entry" widgets) 0)		
              (gtk_widget_set_sensitive (gethash "EXPERIMENTERMETRICmax-tam-entry" widgets) 0)))

  1)


;;------------------------------------------------------------------------------------------------------------
;; david montero
;; funcion que guarda todo el estado actual del experimento en el archivo abierto
;;------------------------------------------------------------------------------------------------------------

(defun save-file-experimenter()
	(if (null *fichero-experimenter*)	
		(progn 
			(if (eq *language* 'es)
	      			(error-dialog nil (format nil "Debe tener abierto un fichero de experimento antes"))
				(error-dialog nil (format nil "Must be open one file experimenter first"))
			)
		1)
		(progn
		(with-open-file (stream *fichero-experimenter* :direction :output
				:if-exists :supersede
				:if-does-not-exist :create)
			(format stream "(")
			(save-domains-list stream)
  			(save-algoritm-list stream)
			(format stream " )")
			(if (eq *language* 'es)
	      			(information-dialog nil (format nil "Archivo guardado correctamente"))
				(information-dialog nil (format nil "The file has been saved satisfactorily")))
			)
		1)))


;;------------------------------------------------------------------------------------------------------------
;; david montero
;; funcion que guarda todo el estado actual del experimento en el archivo que indique el usuario
;;------------------------------------------------------------------------------------------------------------

(defun save-as-file-experimenter()
        (if (not (null *fichero-experimenter*))
		(let ((dialog-widg (gtk_file_selection_new)))
                        (print (format nil "hasta aqui si~%"))
	        	(gtk_file_selection_set_filename dialog-widg (concatenate 'string *my-experimenter-pathname* "nuevo.exp"))
			(if (= (gtk_dialog_run dialog-widg) -5)

			  ;; se comprueba si el fichero existe y se le pregunta al usuario si lo quiere sobreescribir
			  (let ((is-file 1) (continuar 1) (file (gtk_file_selection_get_filename dialog-widg)))
			    (if (probe-file file)	
				(setf is-file 1)
				(setf is-file 0)
			    )
  
			    (if (= is-file 1)
				(let ((result 
				    (if (eq *language* 'es)
					(show-question nil (format nil "El fichero ~a ya existe y se va a sobreescribir ~&~a" file "Desea continuar?"))
					(show-question nil (format nil "The file ~a exists and it going to be overwritten ~&~a" file "Would you like to continue?"))
				    ))) 
				    ;; yes
			            (if (= result -8)	
					(setf continuar 1)
					(setf continuar 0)
		                    )
				)
    			    )
			    (if (= continuar 1)
				(progn 
					(setf *fichero-experimenter* file)
					(with-open-file (stream *fichero-experimenter* :direction :output
						:if-exists :supersede
						:if-does-not-exist :create)
						(format stream "(")
						(save-domains-list stream)
			  			(save-algoritm-list stream)
						(format stream " )")
						(if (eq *language* 'es)
				      			(information-dialog nil (format nil "Archivo guardado correctamente"))	
							(information-dialog nil (format nil "The file has been saved satisfactorily")))
					)
				)
			   )
			 )
			)
			(gtk_widget_destroy dialog-widg)
		1)	
	(if (eq *language* 'es) (error-dialog nil (format nil "No existe ningun fichero abierto"))(error-dialog nil (format nil "There aren`t any file open")))))

;;------------------------------------------------------------------------------------------------------------
;; david montero
;; funcion auxiliar que almacena en el stream pasado por parametro todos los dominios incluidos en el experimento
;;------------------------------------------------------------------------------------------------------------

(defun save-domains-list (stream)
	(progn 
	(format stream "(")
	(dolist (dominio *lista-dominios*)
		(format stream "(~a " (nth 0 dominio))
		(print (nth 1 dominio))
		(format stream "~s " (nth 1 dominio))
		(print (nth 2 dominio))
		(format stream "~s " (nth 2 dominio))
		(print (nth 3 dominio))
		(format stream "~s) " (nth 3 dominio))
	)
	(format stream ")")
	)
)

;;------------------------------------------------------------------------------------------------------------
;; david montero
;; funcion auxiliar que almacena en el stream pasado por parametro todos los algoritmos incluidos en el experimento
;;------------------------------------------------------------------------------------------------------------

(defun save-algoritm-list (stream)
	(let ((longitud) (pos)) 
	(format stream "(")
	(dolist (dominio *lista-algoritmos*)
		(setf longitud (length dominio))
		(setf pos 1)
		(format stream " (")
		(dolist (atributo dominio)
			(if (= pos longitud)
				(format stream "~s " atributo)
				(format stream "~a " atributo)
			)	
			(setf pos (+ 1 pos))
		)
		(format stream " )")
	)
	(format stream ")")
	)
)

;;------------------------------------------------------------------------------------------------------------
;; david montero
;; funcion que abre un dialogo para abrir un nuevo archivo del tipo experimenter
;;------------------------------------------------------------------------------------------------------------

(defun open-file-experimenter()
	;;(declare (special widgets))
	(let ((dialog-widg (gtk_file_selection_new)))
        (gtk_file_selection_set_filename dialog-widg (concatenate 'string *my-experimenter-pathname*))
	  (if (= (gtk_dialog_run dialog-widg) -5)
		(progn 
		(setf *fichero-experimenter* (gtk_file_selection_get_filename dialog-widg))
		(start-experimenter-from-file)
		)
		(show-window "EXPERIMENTER")
		)
	   (gtk_widget_destroy dialog-widg)		
1))

;;------------------------------------------------------------------------------------------------------------
;; david montero
;; funcion que carga y muestra por pantalla el archivo abierto del experimenter
;;------------------------------------------------------------------------------------------------------------

(defun start-experimenter-from-file()
	(let ((aux))
		(handler-case
			(with-open-file (stream *fichero-experimenter* :direction :input)
				(setf aux (read stream))
				(setf *lista-dominios* (car aux))
				(setf *lista-algoritmos* (nth 1 aux))
				(setf *veces-experimenter* 1)
				(setf *modo-ejecucion* 1)
				(show-domains-experimenter))
		(error (signal)  

			(if (eq *language* 'es)
	      			(error-dialog nil (delete_invalid_chars (format nil "~d~%. Se ha producido un error al abrir el fichero ~a" signal *fichero-experimenter*)))
				(error-dialog nil (delete_invalid_chars (format nil "~d~%. It was one error when try to open the file ~a" signal *fichero-experimenter*)))
			))
		)
))

;;------------------------------------------------------------------------------------------------------------
;; david montero
;; funcion que abre la pantalla de eleccion de nuevo dominio a eliminar en el experimenter
;;------------------------------------------------------------------------------------------------------------

(defun show-delete-domain-experimenter()
	(declare (special widgets))
		(if *lista-dominios* 
			(progn
			(show-window "delete-domain-experimenter")  
			1)
			(progn 
				(if (eq *language* 'es)
		      			(error-dialog nil (format nil "No hay ningun dominio incluido"))
					(error-dialog nil (format nil "There aren't any domain included"))
				)
			1)
		)	 
)

;;------------------------------------------------------------------------------------------------------------
;; david montero
;; funcion que realiza la eliminacion del dominio elegido en el experimenter
;;------------------------------------------------------------------------------------------------------------

(defun finish-delete-domain()
	(declare (special widgets))
	(let ((valor (read-from-string (gtk_entry_get_text (gethash "delete-domain-entry" widgets)))))
	     (if (not (numberp valor)) 
		(if (eq *language* 'es) (error-dialog nil (format nil "Debe indicar el numero de dominio")) (error-dialog nil (format nil "You must indicate one domain number")))
		(if (and (> valor 0) (<= valor (length *lista-dominios*)))
			(let 
				((result (if (eq *language* 'en)(show-question nil (format nil "~a~d~a" "domain " valor " at the domain list is going to be deleted" "Would you like to continue?")) (show-question nil (format nil "~a~d~a" "El dominio " valor " de la lista de dominios va a ser eliminado" "Quiere continuar?")))))
				(if (= result -8) 
					(progn 
						(setf *lista-dominios* (delete-from-list *lista-dominios* valor 1))
						(hide-window "delete-domain-experimenter")
						(show-domains-experimenter))
					)
			)  
		
			(if (eq *language* 'es) (error-dialog nil (format nil "Debe indicar un numero de dominio correcto")) (error-dialog nil (format nil "You must indicate one correct domain number")))
		)
	     )
	1)
)

;;------------------------------------------------------------------------------------------------------------
;; david montero
;; funcion que asigna a la variable *veces-experimenter* el numero de veces que ha indicado el usuario que se debe
;; repetir el experimento
;;------------------------------------------------------------------------------------------------------------

(defun finish-repetitions-experimenter()
	(declare (special widgets))
	(let ((valor (read-from-string (gtk_entry_get_text (gethash "repetitions-experimenter-entry" widgets)))))
	     (if (not (numberp valor)) 
		(if (eq *language* 'es) (error-dialog nil (format nil "Debe indicar el numero de repeticiones")) (error-dialog nil (format nil "You must indicate one number for repetitions")))
		(if (> valor 0)
			(progn
				(setf *veces-experimenter* valor)
				(hide-window "repetitions-experimenter")
			)  
			(if (eq *language* 'es) (error-dialog nil (format nil "Debe indicar un numero de repeticiones correcto")) (error-dialog nil (format nil "You must indicate one correct repetitions number")))
		)
	     )
	1)
)

;;------------------------------------------------------------------------------------------------------------
;; david montero
;; funcion que muestra la pantalla de eleccion de planificador a eliminar
;;------------------------------------------------------------------------------------------------------------

(defun show-delete-planner-experimenter()
	(declare (special widgets))
		(if *lista-algoritmos* 
			(progn
			(show-window "delete-planner-experimenter")  
			1)
			(progn 
			(if (eq *language* 'es)
				(error-dialog nil (format nil "No hay ningun planificador incluido"))
				(error-dialog nil (format nil "There aren`t any planner included" ))
			)
			1)
		)	 
)

;;------------------------------------------------------------------------------------------------------------
;; david montero
;; funcion que se encarga de eliminar el planificador elegido de la lista de planificadores del experimento
;;------------------------------------------------------------------------------------------------------------

(defun finish-delete-planner()
	(declare (special widgets))
	(let ((valor (read-from-string (gtk_entry_get_text (gethash "delete-planner-entry" widgets)))))
	     (if (not (numberp valor)) 
		(if (eq *language* 'es) (error-dialog nil (format nil "Debe indicar el numero de planificador")) (error-dialog nil (format nil "You must indicate one planner number")))
		(if (and (> valor 0) (<= valor (length *lista-algoritmos*)))
			(let 
				((result (if (eq *language* 'en)(show-question nil (format nil "~a~d~a" "Planner " valor " at the planner list is going to be deleted" "Would you like to continue?")) (show-question nil (format nil "~a~d~a" "El planificador " valor " de la lista de planificadores va a ser eliminado" "Quiere continuar?")))))
				(if (= result -8) 
					(progn 
						(setf *lista-algoritmos* (delete-from-list *lista-algoritmos* valor 1))
						(hide-window "delete-planner-experimenter")
						(show-domains-experimenter))
					)
			)  
		
			(if (eq *language* 'es) (error-dialog nil (format nil "Debe indicar un numero de planificador correcto")) (error-dialog nil (format nil "You must indicate one correct planner number")))
		)
	     )
	1)
)


;;------------------------------------------------------------------------------------------------------------
;; david montero
;; funcion auxiliar que elimina de una lista el elemento de la posicion indicada, devolviendo la nueva lista
;;------------------------------------------------------------------------------------------------------------

(defun delete-from-list (lista posicion posicion-actual)
	(if lista
		(if (= posicion posicion-actual)   (delete-from-list (cdr lista) posicion (+ posicion-actual 1))
			(append (list (car lista)) (delete-from-list (cdr lista) posicion (+ posicion-actual 1)))
		)
	)
)

;;------------------------------------------------------------------------------------------------------------
;; david montero
;; funcion que se encarga de realizar la ejecucion del experimenter
;;------------------------------------------------------------------------------------------------------------

(defun start-execute-experimenter ()
	(if *fichero-experimenter*
	(let ((domain 0)(plan 0) (fichero-csv) (hora (multiple-value-list (get-decoded-time))))
        (setf hora (concatenate 'string (format nil "~d-~d-~d-~d-~d-~d" (nth 3 hora) (nth 4 hora) (nth 5 hora) (nth 2 hora) (nth 1 hora) (nth 0 hora) )))
        (SHELL (concatenate 'string "mkdir " *my-experimenter-pathname* (pathname-name *fichero-experimenter*) "-" hora "/"))
	(setf fichero-csv  (concatenate 'string *my-experimenter-pathname* (pathname-name *fichero-experimenter*) "-" hora "/" (format nil "resultados-~a.csv" (pathname-name *fichero-experimenter*))))
	(setf *datos-experimenter-tiempo* nil)
	(setf *datos-experimenter-expandidos* nil)
	(setf *datos-experimenter-longitud* nil)
	(setf *datos-experimenter-coste* nil)
	(setf *numero-dominios-resueltos* nil)
	(setf *datos-experimenter-veces-ganadas-tiempo* nil)
	(setf *datos-experimenter-veces-ganadas-expandidos* nil)	

        (with-open-file (stream fichero-csv :direction :output
				:if-exists :supersede
				:if-does-not-exist :create)
	(format stream "dominio problema planificador atributos tiempo tiempo-acumulado nodos nodos-acumulados coste coste-acumulado longitud longitud-acumulada resuelto num-resueltos~%")

	(case *modo-ejecucion*

	(1
	(do ((x 1 (+ 1 x))) ((> x *veces-experimenter*)) 
	  (setf domain 0)
	  (dolist (dominio *lista-dominios*)
		(progn
		(print (format nil "toda la lista de dominios~%"))
		(print *lista-dominios*)
		(print (format nil "dominio actual~%~%~%~%~%"))
		(print dominio)
		(setf domain (+ domain 1))
		(setf plan 1)
		(if (= x 1)
			(progn 
				(setf *datos-experimenter-tiempo-actuales* nil)
				(setf *datos-experimenter-expandidos-actuales* nil)
				(setf *datos-experimenter-coste-actuales* nil)
				(setf *datos-experimenter-longitud-actuales* nil)
			)
		)
		(dolist (planificador *lista-algoritmos*)
			(if (and (= domain 1) (= x 1))
				(progn
					(setf *datos-experimenter-tiempo* (actualiza-lista 0 *datos-experimenter-tiempo* plan 1))
					(setf *datos-experimenter-expandidos* (actualiza-lista 0 *datos-experimenter-expandidos* plan 1))					
					(setf *datos-experimenter-coste* (actualiza-lista 0 *datos-experimenter-coste* plan 1))
					(setf *datos-experimenter-longitud* (actualiza-lista 0 *datos-experimenter-longitud* plan 1))					
					(setf *datos-experimenter-veces-ganadas-tiempo* (actualiza-lista 0 *datos-experimenter-veces-ganadas-tiempo* plan 1))
					(setf *datos-experimenter-veces-ganadas-expandidos* (actualiza-lista 0 *datos-experimenter-veces-ganadas-expandidos* plan 1))
					(setf *numero-dominios-resueltos* (actualiza-lista 0 *numero-dominios-resueltos* plan 1))
				)
			)

			(let  ((tipo-planificador (car planificador)))
;;			(setf *datos-experimenter-tiempo-actuales* (actualiza-lista 0 *datos-experimenter-tiempo-actuales* plan 1))
;;			(setf *datos-experimenter-expandidos-actuales* (actualiza-lista 0 *datos-experimenter-expandidos-actuales* plan 1))

				(print (format nil "el numero de planificador es: ~d~%" plan))
				(if (and (eq tipo-planificador 'ipss) (equal (pathname-type (nth 1 dominio)) "lisp")) 						(ejecuta-ipss-experimenter (nth 1 dominio) (nth 2 dominio) (nth 3 dominio) (cdr planificador) domain plan stream hora x))
				(if (and (eq tipo-planificador 'sayphi) (equal (pathname-type (nth 1 dominio)) "pddl")) 
					(ejecuta-sayphi-experimenter (nth 1 dominio) (nth 2 dominio) (nth 3 dominio) (cdr planificador) domain plan stream hora x))	
				(if (and (eq tipo-planificador 'metric-ff) (equal (pathname-type (nth 1 dominio)) "pddl")) 
					(ejecuta-metric-experimenter (nth 1 dominio) (nth 2 dominio) (nth 3 dominio) (cdr planificador) domain plan stream hora x))	
				(if (and (eq tipo-planificador 'sgplan) (equal (pathname-type (nth 1 dominio)) "pddl")) 
					(ejecuta-sgplan-experimenter (nth 1 dominio) (nth 2 dominio) (nth 3 dominio) (cdr planificador) domain plan stream hora x))	
				(if (and (eq tipo-planificador 'lpg) (equal (pathname-type (nth 1 dominio)) "pddl")) 
					(ejecuta-lpg-experimenter (nth 1 dominio) (nth 2 dominio) (nth 3 dominio) (cdr planificador) domain plan stream hora x))	

				(setf plan (+ plan 1))
			)
		)
		(actualiza-mejores-experimenter)
		)
	    )
	)
	)
	(2 
	(do ((x 1 (+ 1 x))) ((> x *veces-experimenter*)) 
	    (setf plan 0)
	    (dolist (planificador *lista-algoritmos*)
		(progn
		(setf plan (+ plan 1))
		(setf domain 1)
		(dolist (dominio *lista-dominios*)
			(if (= x 1)
				(progn
					(setf *datos-experimenter-tiempo-actuales* nil)
					(setf *datos-experimenter-expandidos-actuales* nil)
					(setf *datos-experimenter-coste-actuales* nil)
					(setf *datos-experimenter-longitud-actuales* nil)
				)
			)
			(print (format nil "toda la lista de dominios~%"))
			(print *lista-dominios*)
			(print (format nil "dominio actual~%~%~%~%~%"))
			(print dominio)

			(if (and (= domain 1) (= x 1))
				(progn
					(setf *datos-experimenter-tiempo* (actualiza-lista 0 *datos-experimenter-tiempo* plan 1))
					(setf *datos-experimenter-expandidos* (actualiza-lista 0 *datos-experimenter-expandidos* plan 1))					
					(setf *datos-experimenter-coste* (actualiza-lista 0 *datos-experimenter-coste* plan 1))
					(setf *datos-experimenter-longitud* (actualiza-lista 0 *datos-experimenter-longitud* plan 1))					
					(setf *datos-experimenter-veces-ganadas-tiempo* (actualiza-lista 0 *datos-experimenter-veces-ganadas-tiempo* plan 1))
					(setf *datos-experimenter-veces-ganadas-expandidos* (actualiza-lista 0 *datos-experimenter-veces-ganadas-expandidos* plan 1))
					(setf *numero-dominios-resueltos* (actualiza-lista 0 *numero-dominios-resueltos* plan 1))
				)
			)

			(let  ((tipo-planificador (car planificador)))

				(print (format nil "el numero de planificador es: ~d~%" plan))
				(if (and (eq tipo-planificador 'ipss) (equal (pathname-type (nth 1 dominio)) "lisp")) 						(ejecuta-ipss-experimenter (nth 1 dominio) (nth 2 dominio) (nth 3 dominio) (cdr planificador) domain plan stream hora x))
				(if (and (eq tipo-planificador 'sayphi) (equal (pathname-type (nth 1 dominio)) "pddl")) 
					(ejecuta-sayphi-experimenter (nth 1 dominio) (nth 2 dominio) (nth 3 dominio) (cdr planificador) domain plan stream hora x))
				(if (and (eq tipo-planificador 'metric-ff) (equal (pathname-type (nth 1 dominio)) "pddl")) 
					(ejecuta-metric-experimenter (nth 1 dominio) (nth 2 dominio) (nth 3 dominio) (cdr planificador) domain plan stream hora x))	
				(if (and (eq tipo-planificador 'sgplan) (equal (pathname-type (nth 1 dominio)) "pddl")) 
					(ejecuta-sgplan-experimenter (nth 1 dominio) (nth 2 dominio) (nth 3 dominio) (cdr planificador) domain plan stream hora x))
				(if (and (eq tipo-planificador 'lpg) (equal (pathname-type (nth 1 dominio)) "pddl")) 
					(ejecuta-lpg-experimenter (nth 1 dominio) (nth 2 dominio) (nth 3 dominio) (cdr planificador) domain plan stream hora x))	
				(setf domain (+ domain 1))
			)
		)
		(actualiza-mejores-experimenter)
		)
	    )
	)))
	)
	;;parte donde se mostraban los resultados por pantalla, ahora solo se muestra un mensaje indicando que se ha 		terminado el experimenter
	;;(muestra-resultados-experimenter hora)
	(if (eq *language* 'es)
		(information-dialog nil (format nil "El experimento ha terminado"))
		(information-dialog nil (format nil "The experimenter h(user::prodigy-result-solutionsas finished")))
	(print *datos-experimenter-tiempo*)
	(print *datos-experimenter-expandidos*)
	(print *datos-experimenter-coste*)
	(print *datos-experimenter-longitud*)
	(print *datos-experimenter-veces-ganadas-tiempo*)
	(print *datos-experimenter-veces-ganadas-expandidos*)
	1)
	(progn 
			(if (eq *language* 'es)
	      			(error-dialog nil (format nil "Debe elegir un archivo de experimento antes"))
				(error-dialog nil (format nil "Must choose a file experimenter first"))
			)
	1)
	)
)

;;------------------------------------------------------------------------------------------------------------
;; david montero
;; funcion auxiliar que ejecuta una planificacion con prodigy en el experimenter. Recibe como parametros:
;; -dominio
;; -problema
;; -reglas
;; -los atributos del planificador
;; -el numero de dominio dentro del experimenter
;; -el numero de planificador dentro del experimenter
;; -el stream del fichero csv donde se almacenaran los resultados del experimenter
;; -la hora en la que se comenzo la ejecucion del experimenter (para el almacenamiento en la carpeta correcta)
;; -el numero de ejecucion dentro del experimenter (la vez)
;;------------------------------------------------------------------------------------------------------------

(defun ejecuta-ipss-experimenter (dominio problema reglas atributos numero-dominio numero-plan fichero-csv hora vez-experimenter)
	(let ((time-bound (nth 0 atributos))
	      (verbosity (nth 1 atributos))
	      (cost-bound (nth 2 atributos)) 
	      (cost-type (nth 3 atributos))	
	      (multiple-sols-p (nth 4 atributos))
	      (run-without-p (nth 5 atributos))
	      (backtrack-with-costs-p (nth 6 atributos))
	      (parar-cada-problema (nth 7 atributos))
	      (almacenar-reglas (nth 8 atributos))
	      (almacenar-planes (nth 9 atributos))
	      (domain (car (last (pathname-directory dominio))))
	      (probset (pathname-name problema))
	      (tiempo-total-actual (nth (- numero-plan 1) *datos-experimenter-tiempo*))
	      (numero-expandidos-actual (nth (- numero-plan 1) *datos-experimenter-expandidos*))
	      (numero-dominios-resueltos (nth (- numero-plan 1) *numero-dominios-resueltos*))
	      (coste-total-actual (nth (- numero-plan 1) *datos-experimenter-coste*))
	      (longitud-total-actual (nth (- numero-plan 1) *datos-experimenter-longitud*))
	      (error)
	      (filename)
	      )

	      (setf filename (concatenate 'string *my-experimenter-pathname* (pathname-name *fichero-experimenter*) "-" hora "/" (car (last (pathname-directory dominio))) "-" (format nil "~d" numero-dominio) "-" (pathname-name problema) "-" (if (or (equal "nil" (car (last atributos))) (equal "" (car (last atributos)))) (format nil "~a~d" "IPPS" numero-plan) (car (last atributos))) "-" (format nil "~d" vez-experimenter) ".xml"))
	      (setf *planner* 'ipss)
	      (setf *hamlet-output-level* verbosity)
	      (if parar-cada-problema (setf *stop-after-problem-p* t))
	      (if almacenar-reglas (setf *interface-store-fired-crs-p* t))
	      (if almacenar-planes (setf *store-plans-p* t))
	      (if (equal (car (last (pathname-directory problema))) "probsets")

 		 ;; este codigo en la parte de planificacion normal esta repetido, hace lo mismo si es prob o
		 ;; probset. comprobarlo mas adelante
                 (setf *interface-probset-p* t)
                 (setf *interface-probset-p* nil))


	      (setf *valid-planner* 0)
	      (setf *valid-domain* 0)
	      (setf *valid-problem* 0)
	      (setf *problem-number* 0)
	      (setf *unsolved-problems* nil)


              (handler-case (hamlet-test (read-from-string domain) (if *interface-probset-p* probset (read-from-string probset))
                         :test-time-bound time-bound
			 :run-without-p run-without-p
			 :multiple-sols-p multiple-sols-p
			 :cr-file reglas
			 :probset-p *interface-probset-p*
			 :backtrack-with-costs-p  backtrack-with-costs-p
			 :store-fired-rules-p almacenar-reglas
			 :cost-bound cost-bound
			 :cost-type cost-type)
		(error (signal)  
		     	(format t "~d~%" signal)
			(if (eq *valid-domain* 0)
			(setf error (delete_invalid_chars (format nil "Se ha producido un error en el dominio~%~d~%" signal)))
			(if (eq *valid-problem* 0)
			  (set error (delete_invalid_chars (format nil "Se ha producido un error en el problema: ~d~%~d~%" *problem-number* signal)))
			(setf error (delete_invalid_chars (format nil "~d~%")))))
		 )
	  )
	   (if error
		(with-open-file (stream filename :direction :output
				:if-exists :supersede
				:if-does-not-exist :create)
	         (format stream "Se ha producido un error en esta ejecucion, el error es: ~%~a " error))
	      (progn
	      (create_view_and_model_plans)
	      (print filename)
	      (if *almacenar-planes*
              	(save_plans_file filename)
	      )

	      (setf *datos-experimenter-tiempo-actuales* (actualiza-lista (user::prodigy-result-time *prodigy-result*) *datos-experimenter-tiempo-actuales* numero-plan 1))
	      (if tiempo-total-actual 
		(setf *datos-experimenter-tiempo* (actualiza-lista (+ (user::prodigy-result-time *prodigy-result*) tiempo-total-actual)  *datos-experimenter-tiempo* numero-plan 1))
	 	(setf *datos-experimenter-tiempo* (actualiza-lista (user::prodigy-result-time *prodigy-result*) *datos-experimenter-tiempo* numero-plan 1)))
	   (if numero-dominios-resueltos 
		(setf *numero-dominios-resueltos* (actualiza-lista (+ 1 numero-dominios-resueltos)  *numero-dominios-resueltos* numero-plan 1))
		(setf *numero-dominios-resueltos* (actualiza-lista 1  *numero-dominios-resueltos* numero-plan 1)))

	      (setf *datos-experimenter-expandidos-actuales* (actualiza-lista (user::prodigy-result-nodes *prodigy-result*) *datos-experimenter-expandidos-actuales* numero-plan 1))
	      (if numero-expandidos-actual 
		(setf *datos-experimenter-expandidos* (actualiza-lista (+ (user::prodigy-result-nodes *prodigy-result*) numero-expandidos-actual)  *datos-experimenter-expandidos* numero-plan 1))
 		(setf *datos-experimenter-expandidos* (actualiza-lista (user::prodigy-result-nodes *prodigy-result*) *datos-experimenter-expandidos* numero-plan 1)))

	      (setf *datos-experimenter-coste-actuales* (actualiza-lista *best-cost* *datos-experimenter-coste-actuales* numero-plan 1))
	      (if coste-total-actual 
		(setf *datos-experimenter-coste* (actualiza-lista (+ *best-cost* coste-total-actual)  *datos-experimenter-coste* numero-plan 1))
	 	(setf *datos-experimenter-coste* (actualiza-lista *best-cost* *datos-experimenter-coste* numero-plan 1)))


	      (setf *datos-experimenter-longitud-actuales* (actualiza-lista (obtener-longitud) *datos-experimenter-longitud-actuales* numero-plan 1))
	      (if longitud-total-actual 
		(setf *datos-experimenter-longitud* (actualiza-lista (+ (obtener-longitud) longitud-total-actual)  *datos-experimenter-longitud* numero-plan 1))
 		(setf *datos-experimenter-longitud* (actualiza-lista (obtener-longitud) *datos-experimenter-longitud* numero-plan 1)))


	(if (or error (not (eq *valid-planner* 1)))
		(format fichero-csv "~a ~a ~a ~a ~a ~d ~a ~d ~a ~d ~a ~d ~a ~d~%" dominio problema (if (or (equal "nil" (car (last atributos))) (equal "" (car (last atributos)))) (format nil "~a~d" "IPPS" numero-plan) (car (last atributos))) (unir-atributos atributos 'IPSS) "na" (nth (- numero-plan 1) *datos-experimenter-tiempo*) "na" (nth (- numero-plan 1) *datos-experimenter-expandidos*) "na" (nth (- numero-plan 1) *datos-experimenter-coste*) "na" (nth (- numero-plan 1) *datos-experimenter-longitud*) "no" (nth (- numero-plan 1) *numero-dominios-resueltos*))
		 

	      	(format fichero-csv "~a ~a ~a ~a ~d ~d ~d ~d ~d ~d ~d ~d ~a ~d~%" dominio problema (if (or (equal "nil" (car (last atributos))) (equal "" (car (last atributos)))) (format nil "~a~d" "IPPS" numero-plan) (car (last atributos))) (unir-atributos atributos 'IPSS)  (user::prodigy-result-time *prodigy-result*) (nth (- numero-plan 1) *datos-experimenter-tiempo*) (user::prodigy-result-nodes *prodigy-result*) (nth (- numero-plan 1) *datos-experimenter-expandidos*) *best-cost* (nth (- numero-plan 1) *datos-experimenter-coste*) (obtener-longitud) (nth (- numero-plan 1) *datos-experimenter-longitud*) "si" (nth (- numero-plan 1) *numero-dominios-resueltos*))
	)
	      (print (format nil "dominio: ~a~% problema: ~a" domain probset))
	       )
	   )
	1)
)

;;------------------------------------------------------------------------------------------------------------
;; david montero
;; funcion que devuelve la menor longitud de un conjunto de planes solucion en prodigy
;;------------------------------------------------------------------------------------------------------------

(defun obtener-longitud ()
	(let ((total (length (user::prodigy-result-solutions P4::*prodigy-result*)))
	      (valor))
		(if (= total 1) 
			(setf valor (length (nth 0 (user::prodigy-result-solutions P4::*prodigy-result*))))
			(if total
			(progn
				(do ((x 0 (+ 1 x))) ((> x total)) 
				(if valor 
				   (if (< (length (nth x (user::prodigy-result-solutions P4::*prodigy-result*))) valor)
					   (setf valor (length (nth x (user::prodigy-result-solutions P4::*prodigy-result*)))) 
				    )
				   (setf valor (length (nth x (user::prodigy-result-solutions P4::*prodigy-result*)))) 
				  ))
			)
			)
		)
	(print valor)
	(if valor
	valor 0))
)

;;------------------------------------------------------------------------------------------------------------
;; david montero
;; funcion auxiliar que ejecuta una planificacion con sayphi en el experimenter. Recibe como parametros:
;; -dominio
;; -problema
;; -reglas
;; -los atributos del planificador
;; -el numero de dominio dentro del experimenter
;; -el numero de planificador dentro del experimenter
;; -el stream del fichero csv donde se almacenaran los resultados del experimenter
;; -la hora en la que se comenzo la ejecucion del experimenter (para el almacenamiento en la carpeta correcta)
;; -el numero de ejecucion dentro del experimenter (la vez)
;;------------------------------------------------------------------------------------------------------------


(defun ejecuta-sayphi-experimenter (dominio problema reglas atributos numero-dominio numero-plan fichero-csv hora vez-experimenter)
	(let ((time-bound (nth 0 atributos))
	      (verbosity (nth 1 atributos))
	      (heuristic (nth 2 atributos)) 
	      (algoritm (nth 3 atributos))
	      (domain (car (last (pathname-directory dominio))))
	      (probset (pathname-name problema))
	      (tiempo-total-actual (nth (- numero-plan 1) *datos-experimenter-tiempo*))
	      (numero-expandidos-actual (nth (- numero-plan 1) *datos-experimenter-expandidos*))
	      (coste-total-actual (nth (- numero-plan 1) *datos-experimenter-coste*))
	      (longitud-total-actual (nth (- numero-plan 1) *datos-experimenter-longitud*))
	      (numero-dominios-resueltos (nth (- numero-plan 1) *numero-dominios-resueltos*))

	      (error)
	      (filename (concatenate 'string *my-experimenter-pathname* (pathname-name *fichero-experimenter*) "-" hora "/" (car (last (pathname-directory dominio))) "-" (format nil "~d" numero-dominio) "-" (pathname-name problema) "-" (if (or (equal "nil" (car (last atributos))) (equal "" (car (last atributos)))) (format nil "~a~d" "SAYPHI" numero-plan) (car (last atributos)))  "-" (format nil "~d" vez-experimenter) ".xml"))
	      )
	      (if (not(eq *planner* 'sayphi))
    			(load (concatenate 'string *my-sayphi-pathname* "loader.lisp"))
		0)
	      (print dominio)
	      (print problema)
	      (setf *planner* 'sayphi)
	      (setf *say-heuristic* heuristic)
	      (setf *say-algorithm* algoritm)
	      (setf *say-output* verbosity)

	      (setf *valid-planner* 0) 
	      (setf *valid-domain* 0)
	      (setf *valid-problem* 0)
	      (setf *problem-number* 0)
	      (setf *unsolved-problems* nil)
	      (setf *say-timeout* 0)   
	      
	      (progn 
	        (print domain)
		(if (equal *say-algorithm* 'dfbnb) (setf *sayphi-algorithm* 'hc-bnb))
		;;(say-domain domain)
		
		;;(print (concatenate 'string probset ".pddl"))
		;;(prob (concatenate 'string probset ".pddl"))
	      )
	      
	      	(setf lista (list ':timeout time-bound  ':heuristic *say-heuristic* ':algorithm *say-algorithm*))

                (handler-case (sayphi dominio problema :plan-args lista)
		(error (signal)  
		     	(format t "~d~%" signal)
			(if (eq *valid-domain* 0)
			(setf error (delete_invalid_chars (format nil "Se ha producido un error en el dominio~%~d~%" signal)))
			(if (eq *valid-problem* 0)
			  (set error (delete_invalid_chars (format nil "Se ha producido un error en el problema: ~d~%~d~%" *problem-number* signal)))
			(setf error (delete_invalid_chars (format nil "~d~%")))))
		 )
	  )
	(if error
	      (with-open-file (stream filename :direction :output
				:if-exists :supersede
				:if-does-not-exist :create)
         (format stream "Se ha producido un error en esta ejecucion, el error es: ~%~a " error))
	      (progn
	      (create_view_and_model_plans)
	      (print filename)
	      (if *almacenar-planes*
              	(save_plans_file filename)
	      )
	      
            (if (and (solution-p *say-solution*) (solution-found *say-solution*))
	     (progn
	      (setf *datos-experimenter-tiempo-actuales* (actualiza-lista (solution-total-time *say-solution*) *datos-experimenter-tiempo-actuales* numero-plan 1))
	      (if tiempo-total-actual 
		(setf *datos-experimenter-tiempo* (actualiza-lista (+ (solution-total-time *say-solution*) tiempo-total-actual)  *datos-experimenter-tiempo* numero-plan 1))
	 	(setf *datos-experimenter-tiempo* (actualiza-lista (solution-total-time *say-solution*) *datos-experimenter-tiempo* numero-plan 1)))

	      (setf *datos-experimenter-expandidos-actuales* (actualiza-lista (solution-num-nodes *say-solution*) *datos-experimenter-expandidos-actuales* numero-plan 1))
	      (if numero-expandidos-actual 
		(setf *datos-experimenter-expandidos* (actualiza-lista (+ (solution-num-nodes *say-solution*) numero-expandidos-actual)  *datos-experimenter-expandidos* numero-plan 1))
 		(setf *datos-experimenter-expandidos* (actualiza-lista (solution-num-nodes *say-solution*) *datos-experimenter-expandidos* numero-plan 1)))


;; en la siguiente version de sayphi donde pone 0 se pone el coste!!!!!!
	      (setf *datos-experimenter-coste-actuales* (actualiza-lista 0 *datos-experimenter-coste-actuales* numero-plan 1))
	      (if coste-total-actual 
		(setf *datos-experimenter-coste* (actualiza-lista (+ 0 coste-total-actual)  *datos-experimenter-coste* numero-plan 1))
	 	(setf *datos-experimenter-coste* (actualiza-lista 0 *datos-experimenter-coste* numero-plan 1)))

	      (if numero-dominios-resueltos 
		(setf *numero-dominios-resueltos* (actualiza-lista (+ 1 numero-dominios-resueltos)  *numero-dominios-resueltos* numero-plan 1))
		(setf *numero-dominios-resueltos* (actualiza-lista 1  *numero-dominios-resueltos* numero-plan 1)))


	      (setf *datos-experimenter-longitud-actuales* (actualiza-lista (solution-depth *say-solution*) *datos-experimenter-longitud-actuales* numero-plan 1))
	      (if longitud-total-actual 
		(setf *datos-experimenter-longitud* (actualiza-lista (+ (solution-depth *say-solution*) longitud-total-actual)  *datos-experimenter-longitud* numero-plan 1))
 		(setf *datos-experimenter-longitud* (actualiza-lista (solution-depth *say-solution*) *datos-experimenter-longitud* numero-plan 1)))
	    )
	    (progn
	      (setf *datos-experimenter-tiempo-actuales* (actualiza-lista 0 *datos-experimenter-tiempo-actuales* numero-plan 1))
	      (if tiempo-total-actual 
		(setf *datos-experimenter-tiempo* (actualiza-lista (+ 0 tiempo-total-actual)  *datos-experimenter-tiempo* numero-plan 1))
	 	(setf *datos-experimenter-tiempo* (actualiza-lista 0 *datos-experimenter-tiempo* numero-plan 1)))

	      (setf *datos-experimenter-expandidos-actuales* (actualiza-lista 0 *datos-experimenter-expandidos-actuales* numero-plan 1))
	      (if numero-expandidos-actual 
		(setf *datos-experimenter-expandidos* (actualiza-lista (+ 0 numero-expandidos-actual)  *datos-experimenter-expandidos* numero-plan 1))
 		(setf *datos-experimenter-expandidos* (actualiza-lista 0 *datos-experimenter-expandidos* numero-plan 1)))

	      (if numero-dominios-resueltos 
		(setf *numero-dominios-resueltos* (actualiza-lista (+ 0 numero-dominios-resueltos)  *numero-dominios-resueltos* numero-plan 1))
		(setf *numero-dominios-resueltos* (actualiza-lista 0  *numero-dominios-resueltos* numero-plan 1)))

	      (setf *datos-experimenter-coste-actuales* (actualiza-lista 0 *datos-experimenter-coste-actuales* numero-plan 1))
	      (if coste-total-actual 
		(setf *datos-experimenter-coste* (actualiza-lista (+ 0 coste-total-actual)  *datos-experimenter-coste* numero-plan 1))
	 	(setf *datos-experimenter-coste* (actualiza-lista 0 *datos-experimenter-coste* numero-plan 1)))


	      (setf *datos-experimenter-longitud-actuales* (actualiza-lista 0 *datos-experimenter-longitud-actuales* numero-plan 1))
	      (if longitud-total-actual 
		(setf *datos-experimenter-longitud* (actualiza-lista (+ 0 longitud-total-actual)  *datos-experimenter-longitud* numero-plan 1))
 		(setf *datos-experimenter-longitud* (actualiza-lista 0 *datos-experimenter-longitud* numero-plan 1)))


	    )
	   )
	)      
      )
      (if (or error (not (solution-found *say-solution*)) (not (solution-stop-reason *say-solution*))) 
	      (format fichero-csv "~a ~a ~a ~a ~a ~d ~a ~d ~a ~d ~a ~d ~a ~d~%" dominio problema (if (or (equal "nil" (car (last atributos))) (equal "" (car (last atributos)))) (format nil "~a~d" "SAYPHI" numero-plan) (car (last atributos))) (unir-atributos atributos 'SAYPHI) "na" (nth (- numero-plan 1) *datos-experimenter-tiempo*) "na" (nth (- numero-plan 1) *datos-experimenter-expandidos*) "na" (nth (- numero-plan 1) *datos-experimenter-coste*) "na" (nth (- numero-plan 1) *datos-experimenter-longitud*) "no" (nth (- numero-plan 1) *numero-dominios-resueltos*))

;;en la siguiente version de sayphi donde pone "na" poner el coste
	      (format fichero-csv "~a ~a ~a ~a ~d ~d ~a ~d ~d ~d ~d ~d~%" dominio problema (if (or (equal "nil" (car (last atributos))) (equal "" (car (last atributos)))) (format nil "~a~d" "SAYPHI" numero-plan) (car (last atributos))) (unir-atributos atributos 'SAYPHI) (solution-total-time *say-solution*) (nth (- numero-plan 1) *datos-experimenter-tiempo*) (solution-num-nodes *say-solution*) (nth (- numero-plan 1) *datos-experimenter-expandidos*) "na" (nth (- numero-plan 1) *datos-experimenter-coste*) (solution-depth *say-solution*) (nth (- numero-plan 1) *datos-experimenter-longitud*) "si" (nth (- numero-plan 1) *numero-dominios-resueltos*))
)
	      (print (format nil "dominio: ~a~% problema: ~a" domain probset))
	1)
)


;;------------------------------------------------------------------------------------------------------------
;; david montero
;; funcion auxiliar que ejecuta una planificacion con metric-ff en el experimenter. Recibe como parametros:
;; -dominio
;; -problema
;; -reglas
;; -los atributos del planificador
;; -el numero de dominio dentro del experimenter
;; -el numero de planificador dentro del experimenter
;; -el stream del fichero csv donde se almacenaran los resultados del experimenter
;; -la hora en la que se comenzo la ejecucion del experimenter (para el almacenamiento en la carpeta correcta)
;; -el numero de ejecucion dentro del experimenter (la vez)
;;------------------------------------------------------------------------------------------------------------


(defun ejecuta-metric-experimenter (dominio problema reglas atributos numero-dominio numero-plan fichero-csv hora vez-experimenter)
	(let ((time-bound (nth 0 atributos))
	      (max-tam (nth 1 atributos))
	      (wg (nth 2 atributos)) 
	      (wh (nth 3 atributos))
	      (metric-opt (nth 4 atributos))
	      (ff-heuristic (nth 5 atributos))
	      (ff-algorithm (nth 6 atributos))
	      (ff-poda (nth 7 atributos))
	      (domain dominio)
	      (probset problema)
	      (rules reglas)
	      ;;(error)
	      (opciones)
	      (contenido)
	      (solucion 0)
	      (tiempo)
	      (coste)
              (longitud)
	      (tiempo-total-actual (nth (- numero-plan 1) *datos-experimenter-tiempo*))
	      (numero-expandidos-actual (nth (- numero-plan 1) *datos-experimenter-expandidos*))
	      (coste-total-actual (nth (- numero-plan 1) *datos-experimenter-coste*))
	      (longitud-total-actual (nth (- numero-plan 1) *datos-experimenter-longitud*))
	      (numero-dominios-resueltos (nth (- numero-plan 1) *numero-dominios-resueltos*))
	      (planaction-counter 0)
	      (bestplanaction-counter -1)
	      (plan-counter 1)
	      (correcta)
	      (filename (concatenate 'string *my-experimenter-pathname* (pathname-name *fichero-experimenter*) "-" hora "/" (car (last (pathname-directory dominio))) "-" (format nil "~d" numero-dominio) "-" (pathname-name problema) "-" (if (or (equal "nil" (car (last atributos))) (equal "" (car (last atributos)))) (format nil "~a~d" "METRIC-FF" numero-plan) (car (last atributos))) "-" (format nil "~d" vez-experimenter) ".xml"))
	      )
		(if (not(eq *planner* 'metric-ff))
		      (setf *planner* 'metric-ff)		
		0)
	      (print dominio)
	      (print problema)
	      (progn

	      (setf *dominio* domain)
	      (setf *problemas* probset)
	      (if (or (equal rules "") (equal rules "nil") )
			(setf *reglas* nil)
			(setf *reglas* rules)
	      )

	      (if (not (null *reglas*))
		    	(setf opciones (concatenate 'string opciones "-r " *reglas* " "))
	       )

	       (if (equal 'best-first ff-algorithm)
		      (progn
		      	(setf opciones (concatenate 'string opciones "-E "))
			(case ff-poda
				(stop-at-first-solution (setf opciones (concatenate 'string opciones "-P 0 ")))
				(Anytime-WA*-prunning-by-g-values (setf opciones (concatenate 'string opciones "-P 1 ")))
				(Anytime-WA*-pruning-by-g-values-and-using-helpful-actions-prune (setf opciones (concatenate 'string opciones "-P 2 ")))
				(Anytime-WA*-prunning-by-f-values (setf opciones (concatenate 'string opciones "-P 3 ")))
				(Anytime-WA*-pruning-by-f-values-and-using-helpful-actions-prune (setf opciones (concatenate 'string opciones "-P 4 ")))
			)
			(if (and (numberp max-tam) (>= 0 max-tam)) (setf opciones (concatenate 'string opciones "-m ~d " max-tam)))
	   	      )
	      )

	      (case ff-heuristic
		(Metric-FF (setf opciones (concatenate 'string opciones "-V 0 ")))
	        (graphplan-with-fixed-cost-step (setf opciones (concatenate 'string opciones "-V 1 ")))
	        (cheapest-applicable-action-with-cost>0 (setf opciones (concatenate 'string opciones "-V 2 ")))
	        (cheapest-applicable-action-with-cost>=0 (setf opciones (concatenate 'string opciones "-V 3 ")))
	        (cheapest-applicable-action-with-cost>=0-including-delayed-actions (setf opciones (concatenate 'string opciones "-V 4 ")))
        	(cheapest-applicable-action-with-cost>0-including-delayed-actions (setf opciones (concatenate 'string opciones "-V 5 ")))
	      )

	      (if (and (numberp wg) (>= wg 0))
		(setf opciones (concatenate 'string opciones (format nil "-g ~d " wg)))
	      )

	      (if (and (numberp wh) (>= wh 0))
		(setf opciones (concatenate 'string opciones (format nil "-h ~d " wh)))
	      )

	      (if metric-opt
		(setf opciones (concatenate 'string opciones "-O "))
	      )
		(setf opciones (concatenate 'string opciones (format nil "-c ~a " time-bound)))

;;(run-shell-command (concatenate 'string *my-metricff-pathname* "system/planner/ff -o " *dominio* " -f " *problemas* " " opciones " | grep \"\\([0-9]\\+: \\|total time\\|total cost\\|egmentation\\|FALSE\\|step[ ]\\+0:\\)\"" "| sed -e \"s/ff: goal can be simplified to FALSE. No plan will solve it/ nil/\"" "| sed -e \"s/step $/ ((*finish*/\"" "| sed -e \"s/step[ ]\\+0:/ ((/\"" "| sed -e \"s/[0-9]\\+:/)(/\" " "| sed -e \"s/\\([0-9]\\+.[0-9]\\+\\) seconds total time/)) \\1/\"" " | sed -e \"s/total cost:/ /\""  "> " *my-metricff-pathname* "ff-output.lisp" ))

(run-shell-command (concatenate 'string *my-metricff-pathname* "system/planner/ff -o " *dominio* " -f " *problemas* " " opciones "> " *my-metricff-pathname* "ff-output.lisp" ))

	(with-open-file (stream (concatenate 'string *my-metricff-pathname* "ff-output.lisp") :direction :input :if-does-not-exist nil)
		(with-open-file (stream1 (concatenate 'string *my-metricff-pathname* "ff-output2.lisp") :direction :output :if-exists :supersede :if-does-not-exist :create)
		
		    (if stream
			(do ((leido (read-line stream nil 'eof)
			    (read-line stream nil 'eof)))
			    ((eq leido 'eof))
			    ;;comprobaciones de la linea
			    (progn
	        	   	 (setf correcta (cl-ppcre::scan "([0-9]+:|total time|total cost|egmentation|FALSE|step[ ]+0:)" leido))
			   	 (if correcta
					(format stream1 "~a~%" leido)
			   	 )
			    )))
		)

	)


	(with-open-file (stream (concatenate 'string *my-metricff-pathname* "ff-output2.lisp") :direction :input :if-does-not-exist nil)
		(with-open-file (stream1 (concatenate 'string *my-metricff-pathname* "ff-output.lisp") :direction :output :if-exists :supersede :if-does-not-exist :create)
		
		    (if stream
			(do ((leido (read-line stream nil 'eof)
			    (read-line stream nil 'eof)))
			    ((eq leido 'eof))
			    ;;comprobaciones de la linea
			    (progn
	        	   	 (setf correcta (cl-ppcre::scan "step[ ]+0:" leido))
				 (if correcta
					(progn
						(setf leido (cl-ppcre::regex-replace "step[ ]+0:" leido "((("))
						(setf solucion (+ solucion 1))
					)
				 )
				 (setf correcta (cl-ppcre::scan "[0-9]+:" leido))
				 (if correcta
					(setf leido (cl-ppcre::regex-replace "[0-9]+:" leido ")("))
				 )
				 (setf correcta (cl-ppcre::scan "seconds total time" leido))
				 (if correcta
			
					(setf leido (cl-ppcre::register-groups-bind (v1) ("([0-9]+\.[0-9]+) seconds total time" leido) (format nil ")) ~a " v1)))
				 )
				 (setf correcta (cl-ppcre::scan "total cost:[ ]+[0-9]+\.[0-9]+" leido))
				 (if correcta

					(setf leido (cl-ppcre::register-groups-bind (v1) ("total cost:[ ]+([0-9]+\.[0-9]+)" leido) (format nil "~a)" v1)))
				 )
				 (format stream1 "~a~%" leido) 
			    )))
		)
	)

	(with-open-file (stream (concatenate 'string *my-metricff-pathname* "ff-output.lisp") :direction :input :if-does-not-exist nil)
		(with-open-file (stream1 (concatenate 'string *my-metricff-pathname* "ff-output2.lisp") :direction :output :if-exists :supersede :if-does-not-exist :create)

		    (if stream
			(do ((leido (read-line stream nil 'eof)
			    (read-line stream nil 'eof)))
			    ((eq leido 'eof))
			    ;;comprobaciones de la linea
			    (setf contenido (concatenate 'string contenido (format nil "~a" leido)))
			))
		    (setf contenido (cl-ppcre::regex-replace "\\)\\)\\)" contenido ")" ))
		    (format stream1 "(~a)" contenido)
		)

	)
	);;close progn

	      (progn
	      (if (> solucion 0)
		(progn
		      (create_view_and_model_plans)
		      (print filename)
		)
	      )
	      (if (and *almacenar-planes* (> solucion 0))
              	(save_plans_file filename)
	      )
	      )
              (setf contenido (file-to-string "ff-output2.lisp" *my-metricff-pathname*))

	     ;;(shell (concatenate 'string "rm " *my-metricff-pathname* "ff-output*.lisp"))
            (if (and (not (null contenido)) (not (eq contenido "")) (> solucion 0))
	     (progn
	      ;;(setf contenido (concatenate 'string "(" contenido ")"))
	      (setf contenido (read-from-string contenido))
		(if (= 1 (length contenido))
			(progn
				(setf bestplanaction-counter 0)
				(dolist (plan (car (car contenido)))
					(incf bestplanaction-counter))
				(setf tiempo (car (cdr (car contenido))))
				(setf coste  (car (last (cdr (car contenido)))))
				(setf longitud bestplanaction-counter)
			)
			(progn
				(dolist (plan contenido)
					(if (<= plan-counter (- (length contenido) 2))
						(progn
							(dolist (plan1 (car plan))
								(incf planaction-counter)
							)
							(if (or (< bestplanaction-counter 0) (< planaction-counter bestplanaction-counter))
								(setf bestplanaction-counter planaction-counter)
							)
							(setf planaction-counter 0)
						)
					)
					(incf plan-counter)
				)
				(setf tiempo (nth (- (length contenido) 2) contenido))
				(setf coste  (nth (- (length contenido) 1) contenido))
				(setf longitud bestplanaction-counter)
			)
		)
		(print (format nil "valores de las constantes~%"))
		(print tiempo)
		(print coste)
		(print longitud)
	      (setf *datos-experimenter-tiempo-actuales* (actualiza-lista tiempo *datos-experimenter-tiempo-actuales* numero-plan 1))
	      (if tiempo-total-actual 
		(setf *datos-experimenter-tiempo* (actualiza-lista (+ tiempo tiempo-total-actual)  *datos-experimenter-tiempo* numero-plan 1))
	 	(setf *datos-experimenter-tiempo* (actualiza-lista tiempo *datos-experimenter-tiempo* numero-plan 1)))

;;si se puede obtener el numero de nodos expandidos en metric FF donde pone 0 poner ese calculo!!!!!
	      (setf *datos-experimenter-expandidos-actuales* (actualiza-lista 0 *datos-experimenter-expandidos-actuales* numero-plan 1))
	      (if numero-expandidos-actual 
		(setf *datos-experimenter-expandidos* (actualiza-lista (+ 0 numero-expandidos-actual)  *datos-experimenter-expandidos* numero-plan 1))
 		(setf *datos-experimenter-expandidos* (actualiza-lista 0 *datos-experimenter-expandidos* numero-plan 1)))

	      (setf *datos-experimenter-coste-actuales* (actualiza-lista coste *datos-experimenter-coste-actuales* numero-plan 1))
	      (if coste-total-actual 
		(setf *datos-experimenter-coste* (actualiza-lista (+ coste coste-total-actual)  *datos-experimenter-coste* numero-plan 1))
	 	(setf *datos-experimenter-coste* (actualiza-lista coste *datos-experimenter-coste* numero-plan 1)))

	      (if numero-dominios-resueltos 
		(setf *numero-dominios-resueltos* (actualiza-lista (+ 1 numero-dominios-resueltos)  *numero-dominios-resueltos* numero-plan 1))
		(setf *numero-dominios-resueltos* (actualiza-lista 1  *numero-dominios-resueltos* numero-plan 1)))


	      (setf *datos-experimenter-longitud-actuales* (actualiza-lista longitud *datos-experimenter-longitud-actuales* numero-plan 1))
	      (if longitud-total-actual 
		(setf *datos-experimenter-longitud* (actualiza-lista (+ longitud longitud-total-actual)  *datos-experimenter-longitud* numero-plan 1))
 		(setf *datos-experimenter-longitud* (actualiza-lista longitud *datos-experimenter-longitud* numero-plan 1)))
	    )
	    (progn
	      (setf *datos-experimenter-tiempo-actuales* (actualiza-lista 0 *datos-experimenter-tiempo-actuales* numero-plan 1))
	      (if tiempo-total-actual 
		(setf *datos-experimenter-tiempo* (actualiza-lista (+ 0 tiempo-total-actual)  *datos-experimenter-tiempo* numero-plan 1))
	 	(setf *datos-experimenter-tiempo* (actualiza-lista 0 *datos-experimenter-tiempo* numero-plan 1)))

	      (setf *datos-experimenter-expandidos-actuales* (actualiza-lista 0 *datos-experimenter-expandidos-actuales* numero-plan 1))
	      (if numero-expandidos-actual 
		(setf *datos-experimenter-expandidos* (actualiza-lista (+ 0 numero-expandidos-actual)  *datos-experimenter-expandidos* numero-plan 1))
		(setf *datos-experimenter-expandidos* (actualiza-lista 0 *datos-experimenter-expandidos* numero-plan 1)))

	      (setf *datos-experimenter-coste-actuales* (actualiza-lista 0 *datos-experimenter-coste-actuales* numero-plan 1))
	      (if coste-total-actual 
		(setf *datos-experimenter-coste* (actualiza-lista (+ 0 coste-total-actual)  *datos-experimenter-coste* numero-plan 1))
	 	(setf *datos-experimenter-coste* (actualiza-lista 0 *datos-experimenter-coste* numero-plan 1)))

	      (if numero-dominios-resueltos 
		(setf *numero-dominios-resueltos* (actualiza-lista (+ 0 numero-dominios-resueltos)  *numero-dominios-resueltos* numero-plan 1))
		(setf *numero-dominios-resueltos* (actualiza-lista 0  *numero-dominios-resueltos* numero-plan 1)))


	      (setf *datos-experimenter-longitud-actuales* (actualiza-lista 0 *datos-experimenter-longitud-actuales* numero-plan 1))
	      (if longitud-total-actual 
		(setf *datos-experimenter-longitud* (actualiza-lista (+ 0 longitud-total-actual)  *datos-experimenter-longitud* numero-plan 1))
 		(setf *datos-experimenter-longitud* (actualiza-lista 0 *datos-experimenter-longitud* numero-plan 1)))
	    )
	   )
      (if (or (null contenido)  (eq contenido "") (= solucion 0)) 
	      (progn
	      (print (format nil "entro1~%"))
	      (format fichero-csv "~a ~a ~a ~a ~a ~d ~a ~d ~a ~d ~a ~d ~a ~d~%" dominio problema (if (or (equal "nil" (car (last atributos))) (equal "" (car (last atributos)))) (format nil "~a~d" "METRIC-FF" numero-plan) (car (last atributos))) (unir-atributos atributos 'METRICFF) "na" (nth (- numero-plan 1) *datos-experimenter-tiempo*) "na" (nth (- numero-plan 1) *datos-experimenter-expandidos*) "na" (nth (- numero-plan 1) *datos-experimenter-coste*) "na" (nth (- numero-plan 1) *datos-experimenter-longitud*) "no" (nth (- numero-plan 1) *numero-dominios-resueltos*)))
	      (progn
	      (print (format nil "entro2~%"))
	      (format fichero-csv "~a ~a ~a ~a ~d ~d ~d ~d ~d ~d ~d ~d ~a ~d~%" dominio problema (if (or (equal "nil" (car (last atributos))) (equal "" (car (last atributos)))) (format nil "~a~d" "METRIC-FF" numero-plan) (car (last atributos))) (unir-atributos atributos 'METRICFF) tiempo (nth (- numero-plan 1) *datos-experimenter-tiempo*) 0 (nth (- numero-plan 1) *datos-experimenter-expandidos*) coste (nth (- numero-plan 1) *datos-experimenter-coste*) longitud (nth (- numero-plan 1) *datos-experimenter-longitud*) "si" (nth (- numero-plan 1) *numero-dominios-resueltos*))
))
	(print (format nil "dominio: ~a~% problema: ~a" domain probset))
	1)

)


;;------------------------------------------------------------------------------------------------------------
;; david montero
;; funcion auxiliar que ejecuta una planificacion con sgplan en el experimenter. Recibe como parametros:
;; -dominio
;; -problema
;; -reglas
;; -los atributos del planificador
;; -el numero de dominio dentro del experimenter
;; -el numero de planificador dentro del experimenter
;; -el stream del fichero csv donde se almacenaran los resultados del experimenter
;; -la hora en la que se comenzo la ejecucion del experimenter (para el almacenamiento en la carpeta correcta)
;; -el numero de ejecucion dentro del experimenter (la vez)
;;------------------------------------------------------------------------------------------------------------


(defun ejecuta-sgplan-experimenter (dominio problema reglas atributos numero-dominio numero-plan fichero-csv hora vez-experimenter)
	(let ((time-bound (nth 0 atributos))
	      (domain dominio)
	      (probset problema)
	      ;;(error)
	      (contenido)
	      (solucion 0)
	      (tiempo)
	      (coste)
              (longitud)
	      (correcta)
	      (tiempo-total-actual (nth (- numero-plan 1) *datos-experimenter-tiempo*))
	      (numero-expandidos-actual (nth (- numero-plan 1) *datos-experimenter-expandidos*))
	      (coste-total-actual (nth (- numero-plan 1) *datos-experimenter-coste*))
	      (longitud-total-actual (nth (- numero-plan 1) *datos-experimenter-longitud*))
	      (numero-dominios-resueltos (nth (- numero-plan 1) *numero-dominios-resueltos*))
	      (filename (concatenate 'string *my-experimenter-pathname* (pathname-name *fichero-experimenter*) "-" hora "/" (car (last (pathname-directory dominio))) "-" (format nil "~d" numero-dominio) "-" (pathname-name problema) "-" (if (or (equal "nil" (car (last atributos))) (equal "" (car (last atributos)))) (format nil "~a~d" "SGPLAN" numero-plan) (car (last atributos))) "-" (format nil "~d" vez-experimenter) ".xml"))
	      )
		(if (not(eq *planner* 'sgplan))
		      (setf *planner* 'sgplan)		
		0)
	      (print dominio)
	      (print problema)
	      (progn

	      (setf *dominio* domain)
	      (setf *problemas* probset)

;;(run-shell-command (concatenate 'string *my-metricff-pathname* "system/planner/ff -o " *dominio* " -f " *problemas* " " opciones " | grep \"\\([0-9]\\+: \\|total time\\|total cost\\|egmentation\\|FALSE\\|step[ ]\\+0:\\)\"" "| sed -e \"s/ff: goal can be simplified to FALSE. No plan will solve it/ nil/\"" "| sed -e \"s/step $/ ((*finish*/\"" "| sed -e \"s/step[ ]\\+0:/ ((/\"" "| sed -e \"s/[0-9]\\+:/)(/\" " "| sed -e \"s/\\([0-9]\\+.[0-9]\\+\\) seconds total time/)) \\1/\"" " | sed -e \"s/total cost:/ /\""  "> " *my-metricff-pathname* "ff-output.lisp" ))

(run-shell-command (concatenate 'string *my-sgplan-pathname* "system/planner/sgplan522 -o " *dominio* " -f " *problemas* " " (format nil "-cputime ~a " time-bound) "> " *my-sgplan-pathname* "sgplan-output.lisp"))
	(print 1)
	(with-open-file (stream (concatenate 'string *my-sgplan-pathname* "sgplan-output.lisp") :direction :input :if-does-not-exist nil)
		(with-open-file (stream1 (concatenate 'string *my-sgplan-pathname* "sgplan-output2.lisp") :direction :output :if-exists :supersede :if-does-not-exist :create)
		
		    (if stream
			(do ((leido (read-line stream nil 'eof)
			    (read-line stream nil 'eof)))
			    ((eq leido 'eof))
			    ;;comprobaciones de la linea
			    (progn
	        	   	 (setf correcta (cl-ppcre::scan "([0-9]+:|\;[ ]+Time|NrActions|MetricValue)" leido))
			   	 (if correcta
					(format stream1 "~a~%" leido)
			   	 )
			    )))
		)
	)
	(print 2)
	(with-open-file (stream (concatenate 'string *my-sgplan-pathname* "sgplan-output2.lisp") :direction :input :if-does-not-exist nil)
		(with-open-file (stream1 (concatenate 'string *my-sgplan-pathname* "sgplan-output.lisp") :direction :output :if-exists :supersede :if-does-not-exist :create)
		
		    (if stream
			(do ((leido (read-line stream nil 'eof)
			    (read-line stream nil 'eof)))
			    ((eq leido 'eof))
			    ;;comprobaciones de la linea
			    (progn
				 (setf correcta (cl-ppcre::scan "\;[ ]+Time [0-9]+\.[0-9]+" leido))
				 (if correcta
					(setf leido (cl-ppcre::register-groups-bind (v1) ("\;[ ]+Time ([0-9]+\.[0-9]+)" leido) (format nil "( ~a " v1)))
  				 )


				 (setf correcta (cl-ppcre::scan "\;[ ]+NrActions [0-9]+" leido))
				 (if correcta
					(setf leido (cl-ppcre::register-groups-bind (v1) ("\;[ ]+NrActions ([0-9]+)" leido) (format nil "~a " v1)))
  				 )

				 (setf correcta (cl-ppcre::scan "\;[ ]+MetricValue [0-9]+" leido))
				 (if correcta
					(setf leido (cl-ppcre::register-groups-bind (v1) ("\;[ ]+MetricValue ([0-9]+\.[0-9]+)" leido) (format nil " ~a (" v1)))
  				 )
				
				 (setf correcta (cl-ppcre::scan "\;[ ]+MetricValue" leido))
				 (if correcta
					(setf leido (cl-ppcre::regex-replace "\;[ ]+MetricValue" leido "0.0 ("))
  				 )

	        	   	 (setf correcta (cl-ppcre::scan "[0-9]+\.[0-9]+:" leido))
				 (if correcta
					(progn
						(setf leido (cl-ppcre::regex-replace "[0-9]+\.[0-9]+:" leido "" ))
						(setf solucion (+ solucion 1))
					)
				 )

	        	   	 (setf correcta (cl-ppcre::scan "\[[0-9]+\.[0-9]+\]" leido))
				 (if correcta
					(progn
						(setf leido (cl-ppcre::regex-replace "\[[0-9]+\.[0-9]+\]" leido "" ))
					)
				 )

	        	   	 (setf correcta (cl-ppcre::scan "\[[0-9]+\]" leido))
				 (if correcta
					(progn
						(setf leido (cl-ppcre::regex-replace "\[[0-9]+\]" leido "" ))
					)
				 )

				 (format stream1 "~a~%" leido) 
			    )))
		)
	)
	(print 3)
	(with-open-file (stream (concatenate 'string *my-sgplan-pathname* "sgplan-output.lisp") :direction :input :if-does-not-exist nil)
		(with-open-file (stream1 (concatenate 'string *my-sgplan-pathname* "sgplan-output2.lisp") :direction :output :if-exists :supersede :if-does-not-exist :create)

		    (if stream
			(do ((leido (read-line stream nil 'eof)
			    (read-line stream nil 'eof)))
			    ((eq leido 'eof))
			    ;;comprobaciones de la linea
			    (setf contenido (concatenate 'string contenido (format nil "~a" leido)))
			))
		    (setf contenido (concatenate 'string contenido (format nil "))")))
		    (format stream1 "~a" contenido)
		)

	)
	);;close progn

	      (progn
	      (if (> solucion 0)
		(progn
		      (create_view_and_model_plans)
		      (print filename)
		)
	      )
	      (if (and *almacenar-planes* (> solucion 0))
              	(save_plans_file filename)
	      )
	      )
	      (print 4)
              (setf contenido (file-to-string "sgplan-output2.lisp" *my-sgplan-pathname*))
	      (print 5)

	     ;;(shell (concatenate 'string "rm " *my-sgplan-pathname* "sgplan-output*.lisp"))
	      
            (if (and (not (null contenido)) (not (eq contenido "")) (> solucion 0))
	     (progn
	      (setf contenido (read-from-string contenido))
		(setf tiempo (nth 0 contenido))
		(setf coste  (nth 2 contenido))
		(setf longitud (nth 1 contenido))
		
		(print (format nil "valores de las constantes~%"))
		(print tiempo)
		(print coste)
		(print longitud)
	      (setf *datos-experimenter-tiempo-actuales* (actualiza-lista tiempo *datos-experimenter-tiempo-actuales* numero-plan 1))
	      (if tiempo-total-actual 
		(setf *datos-experimenter-tiempo* (actualiza-lista (+ tiempo tiempo-total-actual)  *datos-experimenter-tiempo* numero-plan 1))
	 	(setf *datos-experimenter-tiempo* (actualiza-lista tiempo *datos-experimenter-tiempo* numero-plan 1)))

;;si se puede obtener el numero de nodos expandidos en metric FF donde pone 0 poner ese calculo!!!!!
	      (setf *datos-experimenter-expandidos-actuales* (actualiza-lista 0 *datos-experimenter-expandidos-actuales* numero-plan 1))
	      (if numero-expandidos-actual 
		(setf *datos-experimenter-expandidos* (actualiza-lista (+ 0 numero-expandidos-actual)  *datos-experimenter-expandidos* numero-plan 1))
 		(setf *datos-experimenter-expandidos* (actualiza-lista 0 *datos-experimenter-expandidos* numero-plan 1)))

	      (setf *datos-experimenter-coste-actuales* (actualiza-lista coste *datos-experimenter-coste-actuales* numero-plan 1))
	      (if coste-total-actual 
		(setf *datos-experimenter-coste* (actualiza-lista (+ coste coste-total-actual)  *datos-experimenter-coste* numero-plan 1))
	 	(setf *datos-experimenter-coste* (actualiza-lista coste *datos-experimenter-coste* numero-plan 1)))

	      (if numero-dominios-resueltos 
		(setf *numero-dominios-resueltos* (actualiza-lista (+ 1 numero-dominios-resueltos)  *numero-dominios-resueltos* numero-plan 1))
		(setf *numero-dominios-resueltos* (actualiza-lista 1  *numero-dominios-resueltos* numero-plan 1)))


	      (setf *datos-experimenter-longitud-actuales* (actualiza-lista longitud *datos-experimenter-longitud-actuales* numero-plan 1))
	      (if longitud-total-actual 
		(setf *datos-experimenter-longitud* (actualiza-lista (+ longitud longitud-total-actual)  *datos-experimenter-longitud* numero-plan 1))
 		(setf *datos-experimenter-longitud* (actualiza-lista longitud *datos-experimenter-longitud* numero-plan 1)))
	    )
	    (progn
	      (setf *datos-experimenter-tiempo-actuales* (actualiza-lista 0 *datos-experimenter-tiempo-actuales* numero-plan 1))
	      (if tiempo-total-actual 
		(setf *datos-experimenter-tiempo* (actualiza-lista (+ 0 tiempo-total-actual)  *datos-experimenter-tiempo* numero-plan 1))
	 	(setf *datos-experimenter-tiempo* (actualiza-lista 0 *datos-experimenter-tiempo* numero-plan 1)))

	      (setf *datos-experimenter-expandidos-actuales* (actualiza-lista 0 *datos-experimenter-expandidos-actuales* numero-plan 1))
	      (if numero-expandidos-actual 
		(setf *datos-experimenter-expandidos* (actualiza-lista (+ 0 numero-expandidos-actual)  *datos-experimenter-expandidos* numero-plan 1))
		(setf *datos-experimenter-expandidos* (actualiza-lista 0 *datos-experimenter-expandidos* numero-plan 1)))

	      (setf *datos-experimenter-coste-actuales* (actualiza-lista 0 *datos-experimenter-coste-actuales* numero-plan 1))
	      (if coste-total-actual 
		(setf *datos-experimenter-coste* (actualiza-lista (+ 0 coste-total-actual)  *datos-experimenter-coste* numero-plan 1))
	 	(setf *datos-experimenter-coste* (actualiza-lista 0 *datos-experimenter-coste* numero-plan 1)))

	      (if numero-dominios-resueltos 
		(setf *numero-dominios-resueltos* (actualiza-lista (+ 0 numero-dominios-resueltos)  *numero-dominios-resueltos* numero-plan 1))
		(setf *numero-dominios-resueltos* (actualiza-lista 0  *numero-dominios-resueltos* numero-plan 1)))


	      (setf *datos-experimenter-longitud-actuales* (actualiza-lista 0 *datos-experimenter-longitud-actuales* numero-plan 1))
	      (if longitud-total-actual 
		(setf *datos-experimenter-longitud* (actualiza-lista (+ 0 longitud-total-actual)  *datos-experimenter-longitud* numero-plan 1))
 		(setf *datos-experimenter-longitud* (actualiza-lista 0 *datos-experimenter-longitud* numero-plan 1)))
	    )
	   )
      (if (or (null contenido)  (eq contenido "") (= solucion 0)) 
	      (progn
	      (print (format nil "entro1~%"))
	      (format fichero-csv "~a ~a ~a ~a ~a ~d ~a ~d ~a ~d ~a ~d ~a ~d~%" dominio problema (if (or (equal "nil" (car (last atributos))) (equal "" (car (last atributos)))) (format nil "~a~d" "SGPLAN" numero-plan) (car (last atributos))) (unir-atributos atributos 'SGPLAN) "na" (nth (- numero-plan 1) *datos-experimenter-tiempo*) "na" (nth (- numero-plan 1) *datos-experimenter-expandidos*) "na" (nth (- numero-plan 1) *datos-experimenter-coste*) "na" (nth (- numero-plan 1) *datos-experimenter-longitud*) "no" (nth (- numero-plan 1) *numero-dominios-resueltos*)))
	      (progn
	      (print (format nil "entro2~%"))
	      (format fichero-csv "~a ~a ~a ~a ~d ~d ~d ~d ~d ~d ~d ~d ~a ~d~%" dominio problema (if (or (equal "nil" (car (last atributos))) (equal "" (car (last atributos)))) (format nil "~a~d" "SGPLAN" numero-plan) (car (last atributos))) (unir-atributos atributos 'SGPLAN) tiempo (nth (- numero-plan 1) *datos-experimenter-tiempo*) 0 (nth (- numero-plan 1) *datos-experimenter-expandidos*) coste (nth (- numero-plan 1) *datos-experimenter-coste*) longitud (nth (- numero-plan 1) *datos-experimenter-longitud*) "si" (nth (- numero-plan 1) *numero-dominios-resueltos*))
))
	(print (format nil "dominio: ~a~% problema: ~a" domain probset))
	1)

)



;;------------------------------------------------------------------------------------------------------------
;; david montero
;; funcion auxiliar que ejecuta una planificacion con lpg en el experimenter. Recibe como parametros:
;; -dominio
;; -problema
;; -reglas
;; -los atributos del planificador
;; -el numero de dominio dentro del experimenter
;; -el numero de planificador dentro del experimenter
;; -el stream del fichero csv donde se almacenaran los resultados del experimenter
;; -la hora en la que se comenzo la ejecucion del experimenter (para el almacenamiento en la carpeta correcta)
;; -el numero de ejecucion dentro del experimenter (la vez)
;;------------------------------------------------------------------------------------------------------------

(defun ejecuta-lpg-experimenter (dominio problema reglas atributos numero-dominio numero-plan fichero-csv hora vez-experimenter)

	(let ((modo-ejecucion (nth 0 atributos))

	      (time-bound (nth 1 atributos))
	      (time-bound-local (nth 2 atributos)) 
	      (search-steps (nth 3 atributos))
	      (number-solutions (nth 4 atributos))
	      (restarts (nth 5 atributos))
	      (repeats (nth 6 atributos))
	      (seed (nth 7 atributos))
	      (noise (nth 8 atributos))
	      ;;(noout (nth 9 atributos))
	      ;;(voff (nth 10 atributos))
	      (static-noise (nth 9 atributos))
	      (low-memory (nth 10 atributos))
	      (nobestfirst (nth 11 atributos))
	      (onlybestfirst (nth 12 atributos))
	      (time-steps (nth 13 atributos))
	      (domain dominio)
	      (probset problema)
	      ;;(rules reglas)
	      ;;(error)
	      (opciones)
	      (contenido)
	      (solucion 0)
	      (tiempo)
	      (coste)
              (longitud)
	      (correcta)
	      (comando)
	      (nombre-directorio "temporales/")
	      (tiempo-total-actual (nth (- numero-plan 1) *datos-experimenter-tiempo*))
	      (numero-expandidos-actual (nth (- numero-plan 1) *datos-experimenter-expandidos*))
	      (coste-total-actual (nth (- numero-plan 1) *datos-experimenter-coste*))
	      (longitud-total-actual (nth (- numero-plan 1) *datos-experimenter-longitud*))
	      (numero-dominios-resueltos (nth (- numero-plan 1) *numero-dominios-resueltos*))
	      (planaction-counter 0)
	      (bestplanaction-counter -1)
	      (plan-counter 1)
	      (filename (concatenate 'string *my-experimenter-pathname* (pathname-name *fichero-experimenter*) "-" hora "/" (car (last (pathname-directory dominio))) "-" (format nil "~d" numero-dominio) "-" (pathname-name problema) "-" (if (or (equal "nil" (car (last atributos))) (equal "" (car (last atributos)))) (format nil "~a~d" "LPG" numero-plan) (car (last atributos))) "-" (format nil "~d" vez-experimenter) ".xml"))
	      )
		(if (not(eq *planner* 'lpg))
		      (setf *planner* 'lpg)		
		0)
	      (print dominio)
	      (print problema)
	      (progn

	      (setf *dominio* domain)
	      (setf *problemas* probset)


(if (equal 'incremental modo-ejecucion)
	(setf opciones (concatenate 'string (format nil "-n ~d " number-solutions) " -noout "))
)
(if (equal 'speed modo-ejecucion)
	(setf opciones (concatenate 'string "-speed " "-noout "))
)
(if (equal 'quality modo-ejecucion)
	(setf opciones (concatenate 'string "-quality " "-out " *my-lpg-pathname* nombre-directorio (pathname-name *problemas*) " "))
)
;;(if (and (equal 'quality modo-ejecucion) (not noout))
;;	(setf opciones (concatenate 'string opciones "-out " *my-lpg-pathname* nombre-directorio "temporal "))
;;	(setf opciones (concatenate 'string opciones "-noout "))
;;)


(setf opciones (concatenate 'string opciones (format nil "-cputime ~s -cputime_localsearch ~s -search_steps ~s -restarts ~s -repeats ~s -seed ~s -noise ~s " time-bound time-bound-local search-steps restarts repeats seed noise)))

;(if (not noout)
;	(if (not (equal 'quality modo-ejecucion))
;		(setf opciones (concatenate 'string opciones "-noout "))
;	)
;	(progn
;		(setf nombre-directorio "planes_obtenidos/")
;		(setf opciones (concatenate 'string opciones "-out " *my-lpg-pathname* nombre-directorio (pathname-name *problemas*) " " ))
;	)
;)

(setf comando (concatenate 'string "rm " *my-lpg-pathname* nombre-directorio "temporal.SOL" ))
(SHELL comando)

;;(if voff (setf opciones (concatenate 'string opciones "-v off ")))
(if static-noise (setf opciones (concatenate 'string opciones "-static_noise ")))
;;(if max-noise (setf opciones (concatenate 'string opciones "-static_noise ")))
(if low-memory (setf opciones (concatenate 'string opciones "-lowmemory ")))
(if nobestfirst (setf opciones (concatenate 'string opciones "-nobestfirst ")))
(if onlybestfirst (setf opciones (concatenate 'string opciones "-onlybestfirst ")))
(if time-steps (setf opciones (concatenate 'string opciones "-timesteps ")))


(run-shell-command (concatenate 'string *my-lpg-pathname* "system/planner/lpg-td-1.0 -o " *dominio* " -f " *problemas* " " opciones "> " *my-lpg-pathname* "lpg-output.lisp" ))


	(with-open-file (stream (concatenate 'string *my-lpg-pathname* "lpg-output.lisp") :direction :input :if-does-not-exist nil)
		(with-open-file (stream1 (concatenate 'string *my-lpg-pathname* "lpg-output2.lisp") :direction :output :if-exists :supersede :if-does-not-exist :create)
		
		    (if stream
			(do ((leido (read-line stream nil 'eof)
			    (read-line stream nil 'eof)))
			    ((eq leido 'eof))
			    ;;comprobaciones de la linea
			    (progn
	        	    	(if (equal 'quality modo-ejecucion)
	        	  	 	 (setf correcta (cl-ppcre::scan "(Total time|Actions|Execution cost)" leido))
	        	  	 	 (setf correcta (cl-ppcre::scan "([0-9]+:|Plan computed|Total time|Actions|Execution cost|Number of actions)" leido))	
				 )
			   	 (if correcta
					(format stream1 "~a~%" leido)
			   	 )
			    )))
		)
	)



	(with-open-file (stream (concatenate 'string *my-lpg-pathname* "lpg-output2.lisp") :direction :input :if-does-not-exist nil)
		(with-open-file (stream1 (concatenate 'string *my-lpg-pathname* "lpg-output.lisp") :direction :output :if-exists :supersede :if-does-not-exist :create)
		
		    (if stream
			(do ((leido (read-line stream nil 'eof)
			    (read-line stream nil 'eof)))
			    ((eq leido 'eof))
			    ;;comprobaciones de la linea
			    (progn
	        	   	 (setf correcta (cl-ppcre::scan "Number of actions" leido))
				 (if correcta
						(setf leido (cl-ppcre::regex-replace "Number of actions[ ]+:[ ]+[0-9]+" leido "("))
				 )
				 (setf correcta (cl-ppcre::scan "Plan computed" leido))
				 (if correcta
						(progn
						(setf leido (cl-ppcre::regex-replace "Plan computed:" leido "(("))
						(setf solucion (+ solucion 1))
						)
				)
				(format stream1 "~a~%" leido) 
			    )))
		)
	)


	(with-open-file (stream (concatenate 'string *my-lpg-pathname* "lpg-output.lisp") :direction :input :if-does-not-exist nil)
		(with-open-file (stream1 (concatenate 'string *my-lpg-pathname* "lpg-output2.lisp") :direction :output :if-exists :supersede :if-does-not-exist :create)
		
		    (if stream
			(do ((leido (read-line stream nil 'eof)
			    (read-line stream nil 'eof)))
			    ((eq leido 'eof))
			    ;;comprobaciones de la linea
			    (progn
	        	   	 (setf correcta (cl-ppcre::scan "[0-9]+\.[0-9]+:" leido))
				 (if correcta
					     (progn
						(setf leido (cl-ppcre::regex-replace "[0-9]+\.[0-9]+:" leido ""))
						(setf leido (cl-ppcre::regex-replace "[D:[0-9]+\.[0-9]+;[ ]+C:[0-9]+\.[0-9]+]" leido ""))
					      )
				 )
				(format stream1 "~a~%" leido) 
			    )))
		)
	)


	(with-open-file (stream (concatenate 'string *my-lpg-pathname* "lpg-output2.lisp") :direction :input :if-does-not-exist nil)
		(with-open-file (stream1 (concatenate 'string *my-lpg-pathname* "lpg-output.lisp") :direction :output :if-exists :supersede :if-does-not-exist :create)
		
		    (if stream
			(progn
			(do ((leido (read-line stream nil 'eof)
			    (read-line stream nil 'eof)))
			    ((eq leido 'eof))
			    ;;comprobaciones de la linea
			    (progn
	        	   	 (setf correcta (cl-ppcre::scan "Total time:" leido))
				 (if correcta
						(setf leido (cl-ppcre::regex-replace "Total time:[ ]+" leido ")"))
				 )
				 (setf correcta (cl-ppcre::scan "Actions:" leido))
				 (if correcta
						(setf leido (cl-ppcre::regex-replace "Actions:[ ]+" leido ""))
				)
				 (setf correcta (cl-ppcre::scan "Execution cost:" leido))
				 (if correcta
						(progn
						(setf leido (cl-ppcre::regex-replace "Execution cost:[ ]+" leido ""))
						(setf leido (concatenate 'string leido ")"))
						)
				)

				(format stream1 "~a~%" leido) 
			    ))
			  (format stream1 ")~%")
			  ))
			    
		)
	)



	(if (equal 'quality modo-ejecucion)
		(with-open-file (stream (concatenate 'string *my-lpg-pathname* nombre-directorio "temporal.SOL") :direction :input :if-does-not-exist nil)
			(with-open-file (stream1 (concatenate 'string *my-lpg-pathname* "lpg-output2.lisp") :direction :output :if-exists :supersede :if-does-not-exist :create)
		
				(if stream
					(do ((leido (read-line stream nil 'eof)
					    (read-line stream nil 'eof)))
					    ((eq leido 'eof))
					    ;;comprobaciones de la linea
					    (progn
		        	  	 	 (setf correcta (cl-ppcre::scan "[0-9]+:" leido))
						 (if correcta
							(progn
								(setf solucion (+ solucion 1))
								(setf leido (cl-ppcre::regex-replace "[[0-9]+\.[0-9]+]" leido ""))
								(setf leido (cl-ppcre::regex-replace "[0-9]+\.[0-9]+:" leido ""))
								(setf leido (cl-ppcre::regex-replace "[D:[0-9]+\.[0-9]+;[ ]+C:[0-9]+\.[0-9]+]" leido ""))
								(setf leido (cl-ppcre::regex-replace "[0-9]+:" leido ""))
									
								(setf leido (cl-ppcre::regex-replace "[[0-9]+]" leido ""))
							)

						 )
					   	 (if correcta
							(format stream1 "~a~%" leido)
					   	 )
					    )))
			)
		)	
	)

	(if (equal 'quality modo-ejecucion)
		(with-open-file (stream (concatenate 'string *my-lpg-pathname* "lpg-output2.lisp") :direction :input :if-does-not-exist nil)
			(with-open-file (stream2 (concatenate 'string *my-lpg-pathname* "lpg-output.lisp") :direction :input :if-does-not-exist nil)
		
				(if stream
					(progn
						(setf contenido "(((")
						(do ((leido (read-line stream nil 'eof)
						    (read-line stream nil 'eof)))
						    ((eq leido 'eof))
						    ;;comprobaciones de la linea
						    (setf contenido (concatenate 'string contenido leido " "))
						)
					)
				)
				(if stream2
					(progn
						(do ((leido (read-line stream2 nil 'eof)
						    (read-line stream2 nil 'eof)))
						    ((eq leido 'eof))
						    ;;comprobaciones de la linea
						    (setf contenido (concatenate 'string contenido leido " "))
						)
					)

				)

			)
		)
	)

	(if (equal 'quality modo-ejecucion)
		(with-open-file (stream (concatenate 'string *my-lpg-pathname* "lpg-output.lisp") :direction :output :if-exists :supersede :if-does-not-exist :create)
			(progn
				(format stream contenido)					
				(setf contenido "")
			)
		)
	)




	(with-open-file (stream (concatenate 'string *my-lpg-pathname* "lpg-output.lisp") :direction :input :if-does-not-exist nil)
		    (if stream
			(do ((leido (read-line stream nil 'eof)
			    (read-line stream nil 'eof)))
			    ((eq leido 'eof))
			    ;;comprobaciones de la linea
			    (setf contenido (concatenate 'string contenido (format nil "~a " leido)))
			))
	)
	);;close progn

	      (progn
	      (if (> solucion 0)
		(progn
		      (create_view_and_model_plans)
		      (print filename)
		)
	      )
	      (if (and *almacenar-planes* (> solucion 0))
              	(save_plans_file filename)
	      )
	      )
              (setf contenido (file-to-string "lpg-output.lisp" *my-lpg-pathname*))
		
	      (print (format nil "EL CONTENIDO ES ~%~%~%~%~% ~s~%~%~%~%~%" contenido))

      	     ;;(shell (concatenate 'string "rm " *my-lpg-pathname* "lpg-output*.lisp"))
            (if (and (not (null contenido)) (not (eq contenido "")) (> solucion 0))
	     (progn
	      ;;(setf contenido (concatenate 'string "(" contenido ")"))
	      (setf contenido (read-from-string contenido))
		(if (= 1 (length contenido))
			(progn
				(setf bestplanaction-counter 0)
				(dolist (plan (car (car contenido)))
					(incf bestplanaction-counter))
				(setf tiempo (car (cdr (car contenido))))
				(setf coste  (car (last (cdr (car contenido)))))
				(setf longitud bestplanaction-counter)
			)
			(progn
				(dolist (plan contenido)
						(progn
							(dolist (plan1 (car plan))
								(incf planaction-counter)
							)
							(if (or (< bestplanaction-counter 0) (< planaction-counter bestplanaction-counter))
								(setf bestplanaction-counter planaction-counter)
							)
							(setf planaction-counter 0)
						)
						(incf plan-counter)
				)
				(setf tiempo (car (cdr (car (last contenido)))))
				(setf coste  (nth 1 (cdr (car (last contenido)))))
				(setf longitud bestplanaction-counter)
			)
		)
		(print (format nil "valores de las constantes~%"))
		(print tiempo)
		(print coste)
		(print longitud)
	      (setf *datos-experimenter-tiempo-actuales* (actualiza-lista tiempo *datos-experimenter-tiempo-actuales* numero-plan 1))
	      (if tiempo-total-actual 
		(setf *datos-experimenter-tiempo* (actualiza-lista (+ tiempo tiempo-total-actual)  *datos-experimenter-tiempo* numero-plan 1))
	 	(setf *datos-experimenter-tiempo* (actualiza-lista tiempo *datos-experimenter-tiempo* numero-plan 1)))

;;si se puede obtener el numero de nodos expandidos en LPG donde pone 0 poner ese calculo!!!!!
	      (setf *datos-experimenter-expandidos-actuales* (actualiza-lista 0 *datos-experimenter-expandidos-actuales* numero-plan 1))
	      (if numero-expandidos-actual 
		(setf *datos-experimenter-expandidos* (actualiza-lista (+ 0 numero-expandidos-actual)  *datos-experimenter-expandidos* numero-plan 1))
 		(setf *datos-experimenter-expandidos* (actualiza-lista 0 *datos-experimenter-expandidos* numero-plan 1)))

	      (setf *datos-experimenter-coste-actuales* (actualiza-lista coste *datos-experimenter-coste-actuales* numero-plan 1))
	      (if coste-total-actual 
		(setf *datos-experimenter-coste* (actualiza-lista (+ coste coste-total-actual)  *datos-experimenter-coste* numero-plan 1))
	 	(setf *datos-experimenter-coste* (actualiza-lista coste *datos-experimenter-coste* numero-plan 1)))

	      (if numero-dominios-resueltos 
		(setf *numero-dominios-resueltos* (actualiza-lista (+ 1 numero-dominios-resueltos)  *numero-dominios-resueltos* numero-plan 1))
		(setf *numero-dominios-resueltos* (actualiza-lista 1  *numero-dominios-resueltos* numero-plan 1)))


	      (setf *datos-experimenter-longitud-actuales* (actualiza-lista longitud *datos-experimenter-longitud-actuales* numero-plan 1))
	      (if longitud-total-actual 
		(setf *datos-experimenter-longitud* (actualiza-lista (+ longitud longitud-total-actual)  *datos-experimenter-longitud* numero-plan 1))
 		(setf *datos-experimenter-longitud* (actualiza-lista longitud *datos-experimenter-longitud* numero-plan 1)))
	    )
	    (progn
	      (setf *datos-experimenter-tiempo-actuales* (actualiza-lista 0 *datos-experimenter-tiempo-actuales* numero-plan 1))
	      (if tiempo-total-actual 
		(setf *datos-experimenter-tiempo* (actualiza-lista (+ 0 tiempo-total-actual)  *datos-experimenter-tiempo* numero-plan 1))
	 	(setf *datos-experimenter-tiempo* (actualiza-lista 0 *datos-experimenter-tiempo* numero-plan 1)))

	      (setf *datos-experimenter-expandidos-actuales* (actualiza-lista 0 *datos-experimenter-expandidos-actuales* numero-plan 1))
	      (if numero-expandidos-actual 
		(setf *datos-experimenter-expandidos* (actualiza-lista (+ 0 numero-expandidos-actual)  *datos-experimenter-expandidos* numero-plan 1))
		(setf *datos-experimenter-expandidos* (actualiza-lista 0 *datos-experimenter-expandidos* numero-plan 1)))

	      (setf *datos-experimenter-coste-actuales* (actualiza-lista 0 *datos-experimenter-coste-actuales* numero-plan 1))
	      (if coste-total-actual 
		(setf *datos-experimenter-coste* (actualiza-lista (+ 0 coste-total-actual)  *datos-experimenter-coste* numero-plan 1))
	 	(setf *datos-experimenter-coste* (actualiza-lista 0 *datos-experimenter-coste* numero-plan 1)))

	      (if numero-dominios-resueltos 
		(setf *numero-dominios-resueltos* (actualiza-lista (+ 0 numero-dominios-resueltos)  *numero-dominios-resueltos* numero-plan 1))
		(setf *numero-dominios-resueltos* (actualiza-lista 0  *numero-dominios-resueltos* numero-plan 1)))


	      (setf *datos-experimenter-longitud-actuales* (actualiza-lista 0 *datos-experimenter-longitud-actuales* numero-plan 1))
	      (if longitud-total-actual 
		(setf *datos-experimenter-longitud* (actualiza-lista (+ 0 longitud-total-actual)  *datos-experimenter-longitud* numero-plan 1))
 		(setf *datos-experimenter-longitud* (actualiza-lista 0 *datos-experimenter-longitud* numero-plan 1)))
	    )
	   )
      (if (or (null contenido)  (eq contenido "") (= solucion 0)) 
	      (progn
	      (print (format nil "entro1~%"))
	      (format fichero-csv "~a ~a ~a ~a ~a ~d ~a ~d ~a ~d ~a ~d ~a ~d~%" dominio problema (if (or (equal "nil" (car (last atributos))) (equal "" (car (last atributos)))) (format nil "~a~d" "LPG" numero-plan) (car (last atributos))) (unir-atributos atributos 'LPG) "na" (nth (- numero-plan 1) *datos-experimenter-tiempo*) "na" (nth (- numero-plan 1) *datos-experimenter-expandidos*) "na" (nth (- numero-plan 1) *datos-experimenter-coste*) "na" (nth (- numero-plan 1) *datos-experimenter-longitud*) "no" (nth (- numero-plan 1) *numero-dominios-resueltos*)))
	      (progn
	      (print (format nil "entro2~%"))
	      (format fichero-csv "~a ~a ~a ~a ~d ~d ~d ~d ~d ~d ~d ~d ~a ~d~%" dominio problema (if (or (equal "nil" (car (last atributos))) (equal "" (car (last atributos)))) (format nil "~a~d" "LPG" numero-plan) (car (last atributos))) (unir-atributos atributos 'LPG) tiempo (nth (- numero-plan 1) *datos-experimenter-tiempo*) 0 (nth (- numero-plan 1) *datos-experimenter-expandidos*) coste (nth (- numero-plan 1) *datos-experimenter-coste*) longitud (nth (- numero-plan 1) *datos-experimenter-longitud*) "si" (nth (- numero-plan 1) *numero-dominios-resueltos*))
))
	(print (format nil "dominio: ~a~% problema: ~a" domain probset))
	1)

)





;;------------------------------------------------------------------------------------------------------------
;; david montero
;; funcion que intercambia el valor de una posicion en una lista por el nuevo valor indicado
;;------------------------------------------------------------------------------------------------------------

(defun actualiza-lista (nuevo-valor lista posicion posicion-actual)
		(if (> posicion posicion-actual) (append (list (car lista)) (actualiza-lista nuevo-valor (cdr lista) posicion (+ 1 posicion-actual)))  (append (list nuevo-valor) (cdr lista)))
)


;;------------------------------------------------------------------------------------------------------------
;; david montero
;; funcion que actualiza los mejores valores obtenidos para el tiempo, numero de nodos expandidos y tiempo.....
;;------------------------------------------------------------------------------------------------------------

(defun actualiza-mejores-experimenter ()
	(let ((mejor-en-tiempo)
	      (mejor-en-expandidos)
	      (tiempo-minimo)
              (expandidos-minimo)
	      (numero 1))
		(dolist (tiempo *datos-experimenter-tiempo-actuales*)
			(if (and tiempo tiempo-minimo) 
				(if (< tiempo tiempo-minimo)
					(progn 
						(setf mejor-en-tiempo (- numero 1))
						(setf tiempo-minimo tiempo)
					)
				)
				(if (and tiempo (null tiempo-minimo))
					(progn 
						(setf mejor-en-tiempo (- numero 1))
						(setf tiempo-minimo tiempo)
					)
				)
			)			
			(setf numero (+ 1 numero))
		)
		(setf numero 1)
		(dolist (expandidos *datos-experimenter-expandidos-actuales*)
			(if (and expandidos expandidos-minimo) 
				(if (< expandidos expandidos-minimo)
					(progn 
						(setf mejor-en-expandidos (- numero 1))
						(setf expandidos-minimo expandidos)
					)
				)
				(if (and expandidos (null expandidos-minimo))

					(progn 
						(setf mejor-en-expandidos (- numero 1))
						(setf expandidos-minimo expandidos)
					)

				)
			)			
			(setf numero (+ 1 numero))
		)
		(if (not (null *datos-experimenter-tiempo-actuales*))
		(progn
		(setf *datos-experimenter-veces-ganadas-tiempo* (actualiza-lista (if (nth mejor-en-tiempo *datos-experimenter-veces-ganadas-tiempo*) (+ (nth mejor-en-tiempo *datos-experimenter-veces-ganadas-tiempo*) 1) 1) *datos-experimenter-veces-ganadas-tiempo* mejor-en-tiempo 1))
		)
		)

		(if (not (null *datos-experimenter-expandidos-actuales*))
		(progn
		(setf *datos-experimenter-veces-ganadas-expandidos* (actualiza-lista (if (nth mejor-en-expandidos *datos-experimenter-veces-ganadas-expandidos*) (+ (nth mejor-en-expandidos *datos-experimenter-veces-ganadas-expandidos*) 1) 1) *datos-experimenter-veces-ganadas-expandidos* mejor-en-expandidos 1))
		)
		)
	)
)

;;------------------------------------------------------------------------------------------------------------
;; david montero
;; funcion que muestra por pantalla los resultados que se han conseguido con el experimenter, recibe la hora 
;; en la que se lanzo el experimenter para saber en que carpeta se encuentran los resultados
;;------------------------------------------------------------------------------------------------------------

(defun muestra-resultados-experimenter (hora)
	(declare (special widgets))
	(let ((filename (concatenate 'string *my-experimenter-pathname* (pathname-name *fichero-experimenter*) "-" hora "/" "resultados.txt")) (plan 1))
		(with-open-file (stream filename :direction :output
				:if-exists :supersede
				:if-does-not-exist :create)
		 (format stream "RESULTADOS OBTENIDOS~%~%~%")
		 (dolist (planner *datos-experimenter-tiempo*)
			(format stream "Datos del planificador ~d:~%~%" plan)
			(format stream "~t Tiempo tardado en resolver todos los problemas: ~d~%" (nth (- plan 1) *datos-experimenter-tiempo*))
			(format stream "~t Numero de nodos totales expandidos: ~d~%" (nth (- plan 1) *datos-experimenter-expandidos*))
			(format stream "~t Veces en los que el planificador ha sido mejor en tiempo: ~d~%" (nth (- plan 1) *datos-experimenter-veces-ganadas-tiempo*))
			(format stream "~t Veces en los que el planificador ha sido mejor en expansion de nodos: ~d~%~%" (nth (- plan 1) *datos-experimenter-veces-ganadas-expandidos*))
			(setf plan (+ plan 1))
		 ))		
	(put-file "textview-file" (concatenate 'string "resultados.txt") 
                                          (concatenate 'string *my-experimenter-pathname* (pathname-name *fichero-experimenter*) "-" hora "/"))		
                (show-window "V7-SHOW1FILE")
	)
)

;;------------------------------------------------------------------------------------------------------------
;; david montero
;; funcion que devuelve un string con todos los atributos que ha usado un planificador en el experimenter
;;------------------------------------------------------------------------------------------------------------

(defun unir-atributos (atributos planificador)
	(case planificador
	 (IPSS  (format nil "TimeBound:~a||Verbosity:~a||CostBound:~a||CostType:~a||MultiplesSol:~a||WithoutRules:~a||Bactrack:~a||StopAfterProblem:~a||SaveRules:~a||SavePlans:~a" (nth 0 atributos) (nth 1 atributos) (nth 2 atributos) (nth 3 atributos) (nth 4 atributos) (nth 5 atributos) (nth 6 atributos) (nth 7 atributos) (nth 8 atributos)(nth 9 atributos)))
	 (SAYPHI (format nil "TimeBound:~a||Verbosity:~a||Heuristic:~a||Algoritm:~a" (nth 0 atributos) (nth 1 atributos) (nth 2 atributos) (nth 3 atributos)))
	 (METRICFF (format nil "TimeBound:~a||Max-tam:~a||Wg:~a||Wh:~a||Optimice-metric:~a||Heuristic:~a||Algorithm:~a||Anytime:~a" (nth 0 atributos) (nth 1 atributos) (nth 2 atributos) (nth 3 atributos) (nth 4 atributos) (nth 5 atributos) (nth 6 atributos) (nth 7 atributos)))
	 (SGPLAN (format nil "TimeBound:~a" (nth 0 atributos)))
	 (LPG (format nil "ModoEjecucion:~a||TimeBound:~a||TimeBoundLocal:~a||SearchSteps:~a||NumberSolutions:~a||Restarts:~a||Repeats:~a||Seed:~a||Noise:~a||SavePlans:~a||VerboseOff:~a||StaticNoise:~a||LowMemory:~a||NoBestFirst:~a||OnlyBestFirst:~a||TimeSteps:~a" (nth 0 atributos) (nth 1 atributos) (nth 2 atributos) (nth 3 atributos) (nth 4 atributos) (nth 5 atributos) (nth 6 atributos) (nth 7 atributos) (nth 8 atributos) (nth 9 atributos) (nth 10 atributos) (nth 11 atributos) (nth 12 atributos) (nth 13 atributos) (nth 14 atributos) (nth 15 atributos)))
	
        )
)


;;------------------------------------------fin de experimenter--------------------------------------------------------------
